parser grammar SpecificationParser;

/**
 imported grammar rules
   langExpr
   langId
   langType
   langModifier
   langStatement
   startSpec - the rule entering the lexer into specification mode
   endSpec - the rule exiting the lexer from specification mode
 exported grammar rules for PVL
   valContractClause        - contract clause
   valStatement             - proof guiding statement
   valReserved              - reserved identifiers
 exported grammar rules for other languages
   valEmbedContract         - sequence of contract clauses embedded in sequence of comments
   valEmbedContractBlock    - sequence of contract clauses embedded in one comment
   valEmbedStatementBlock   - sequence of valStatements embedded in a comment
 */

valExpressionList
    : langExpr
    | langExpr ',' valExpressionList
    ;

valContractClause
 : 'modifies' valExpressionList ';'
 | 'accessible' valExpressionList ';'
 | 'assigns' valExpressionList ';'
 | 'assumes' langExpr ';'
 | 'requires' langExpr ';'
 | 'ensures' langExpr ';'
 | 'decreases' langExpr ';'
 | ('behavior' | 'behaviour') langId? ':' valContractClause+
 | ('normal_behavior' | 'normal_behaviour') langId? ':' valContractClause+
 | ('exceptional_behavior' | 'exceptional_behaviour') langId? ':' valContractClause+
 | 'signals' '(' langType (langId)? ')' langExpr ';'
 ;

valLoopClause
  : 'loop_invariant' langExpr ';'
  | 'loop_variant' langExpr ';'
  | 'loop_assigns' valExpressionList ';'
  | 'for' langId ':' valLoopClause
  ;

valStatement
 : 'assert' langExpr ';'
 | 'ghost' langStatement
 | 'for' langId ':' valStatement
 ;

valImpOp: '==>';
valEqOp:  '<==>';

valIndexExpression: '*' | langExpr '..' langExpr;

valPrimary
    : langId labelList '(' valExpressionList? ')'
    | '[' langExpr ']' langExpr
    | ('\\forall' | '\\exists') declArgList ';' langExpr
    | '\\old' '(' langExpr ')'
    | '\\at'  '(' langExpr ',' label ')'
    | '\\fresh' '(' langExpr ')'
    | '\\nothing'
    | '\\everything'
    | '\\result'
    | ('\\abs' | '\\sqrt' | '\\ceil' | '\\floor' | '\\real_abs' | '\\exact' | '\\round_error') '(' langExpr ')'
    | ('\\cos' | '\\sin' | '\\tan' | '\\cosh' | '\\sinh' | '\\tanh' | '\\log' | '\\log10' | '\\exp' | '\\acos' | '\\asin' | '\\atan' | '\\acosh' | '\\asinh' | '\\atanh') '(' langExpr ')'
    | ('\\pow' | '\\atan2' | '\\hypot') '(' langExpr ',' langExpr ')'
    | '\\pi' | '\\e'
    ;

valReserved
 : (VAL_ASSERT | VAL_BOOL | VAL_PREDICATE | VAL_REAL | VAL_INTEGER
    | VAL_PURE | VAL_AXIOM | VAL_AXIOMATIC | VAL_LEMMA | VAL_LOGIC
    | VAL_MODIFIES | VAL_ACCESSIBLE | VAL_ASSIGNS | VAL_ASSUMES | VAL_REQUIRES | VAL_ENSURES | VAL_INVARIANT
    | VAL_LOOP_INVARIANT | VAL_LOOP_ASSIGNS | VAL_LOOP_VARIANT | VAL_SIGNALS | VAL_DECREASES
    | VAL_ASSUME | VAL_GHOST | VAL_SIGNALS
    | VAL_BEHAVIOR | VAL_BEHAVIOUR | VAL_EXC_BEHAVIOR | VAL_EXC_BEHAVIOUR | VAL_NORMAL_BEHAVIOR | VAL_NORMAL_BEHAVIOUR)
 | LANG_ID_ESCAPE
 | '\\result'
 | '\\nothing'
 | 'true'
 | 'false'
 ;

label
 : (LABEL_HERE | LABEL_PRE | LABEL_OLD) | langId
 ;

labelList
  : '{' label (',' label)* '}'
  ;

valType
 : ('integer' | 'real' | 'predicate' | 'bool')
 ;

valDeclaration
 : 'predicate' langId labelList? '(' valArgList? ')' ('=' langExpr)? ';'
 | 'inductive' langId labelList? '(' valArgList? ')' '{' caseDefinition+ '}'
 | 'invariant' langId ':' langExpr ';'
 | 'lemma' langId labelList? ':' langExpr ';'
 | 'logic' ('integer' | 'real') langId labelList? '(' valArgList ')' ('=' langExpr)? ';'
 | 'axiom' langId labelList? ':' langExpr ';'
 | 'axiomatic' langId '{' valDeclaration+ '}'
 | valContractClause* 'ghost' langDecl // Maybe need to remove this one?
 ;

// Example: "//@+ CheckArithOverflow = no"
krakatoaOption
  : LineStartOptionSpec KRAKATOA_PROPERTY '=' langId EndSpec
  ;

caseDefinition
  : 'case' langId labelList? ':' langExpr ';'
  ;

valModifier
 :('pure')| langModifier
 ;

valArgList
 : valArg
 | valArg ',' valArgList
 ;

valArg
 : langType langId ('['']')?
 ;

// Unlike valArgList, declArgList allows for multiple declarations of one type, e.g. 'int a, b'
// This is allowed in things such as a \forall clause, but not in the parameters of a method.
declArgList
 : declArg (',' declArg)*
 ;

declArg
 : langType (langId ('[' ']')?)+
 ;

valEmbedContract: valEmbedContractBlock+;

valEmbedContractBlock
 : startSpec valContractClause* endSpec
 | {specLevel>0}? valContractClause+
 ;

valEmbedLoopSpec: valEmbedLoopSpecBlock+;

valEmbedLoopSpecBlock
  : startSpec valLoopClause* endSpec
  | {specLevel>0}? valLoopClause+
  ;

valEmbedStatementBlock
 : startSpec valStatement* endSpec
 | {specLevel>0}? valStatement+
 ;

valEmbedDeclarationBlock
 : startSpec valDeclaration* endSpec
 ;

valEmbedModifiers
 : startSpec valModifier* endSpec
 | {specLevel>0}? valModifier+
 ;

startSpec
 : LineStartSpec {specLevel++}
 | BlockStartSpecImmediate {specLevel++;}
 ;

endSpec
 : EndSpec {specLevel--;}
 | LineCommentStartInSpec {specLevel--;}
 ;
