parser grammar SpecParser;

options { tokenVocab=JavaPlusLexer; }
import JavaPlusParser;

//@parser::members {
//    public int specLevel = 0;
//}

/**
 imported grammar rules
   langExpr
   langId
   langType
   langModifier
   langStatement
   startSpec - the rule entering the lexer into specification mode
   endSpec - the rule exiting the lexer from specification mode
 exported grammar rules for PVL
   valContractClause        - contract clause
   valStatement             - proof guiding statement
   valReserved              - reserved identifiers
 exported grammar rules for other languages
   valEmbedContract         - sequence of contract clauses embedded in sequence of comments
   valEmbedContractBlock    - sequence of contract clauses embedded in one comment
   valEmbedStatementBlock   - sequence of valStatements embedded in a comment
 */
langExpr: expression;
langId: javaIdentifier;
langType: typeType;
langModifier: classOrInterfaceModifier;
langStatement: blockStatement;

startSpec
    : startSpecLine
    | startSpecBlock
    ;

startSpecLine
    : LineStartSpecImmediate {specLevel++;}
    ;
startSpecBlock
    : BlockStartSpecImmediate {specLevel++;}
    ;

endSpec
    : EndSpec {specLevel--;}
    | ContinuedLineCommentStop {specLevel--;}
    | BlockCommentStopInLineComment {specLevel--;}
    ;

valEmbedModifiers
 : startSpec valModifier* endSpec
 | {specLevel>0}? valSpecModifier+
 ;


valEmbedContract: valEmbedContractBlock+;

valEmbedContractStart: startSpec valEmbedContractBlock+;
valEmbedModifiersEnd: valModifier+ endSpec;


valModifier
 : valSpecModifier
 | langModifier
 ;
valSpecModifier
 : ('pure' | 'nullable' | 'nullable_by_default' | 'non_null'
    | 'ghost' | 'helper' | 'monitored' | 'model' | 'instance'
    | 'spec_public' | 'spec_protected')
 | mathModifier
 ;
mathModifier
 : 'spec_java_math' | 'spec_safe_math' | 'spec_bigint_math' | 'code_java_math' | 'code_safe_math' | 'code_bigint_math'
 ;

valEmbedContractBlock
 : valLinewiseContractBlock
 | startSpec ('also')? valSpecCase ('also' valSpecCase)* endSpec
 | {specLevel>0}? ('also')? valSpecCase ('also' valSpecCase)*
 ;

valLinewiseContractBlock
 : valEmbedAlso? valLinewiseSpecCase (valLinewiseAdditionalSpecCase)*
 | valLinewiseAlsoBehaviorClause (valLinewiseAdditionalSpecCase)*
 ;

valEmbedAlso
 : startSpec 'also' endSpec
 ;

valSpecCase
 : valContractClause+
 | valBehaviorClause
 ;

valLinewiseSpecCase
 : (startSpec valContractClause endSpec)+
 | valLinewiseBehaviorClause
 ;

valLinewiseAdditionalSpecCase
 : valEmbedAlso valLinewiseSpecCase
 | valLinewiseAlsoBehaviorClause
 ;

valContractClause
 : ('modifies' | 'assigns' | 'assignable' | 'modifiable') storeRefList ';'
 | 'accessible' storeRefList ';'
 | 'forall' ('non_null' | 'nullable')? typeSpec quantifiedVarDeclarator ';'
 | 'old' ('non_null' | 'nullable')? typeSpec specVariableDeclarator (',' specVariableDeclarator)* ';'
 | ('requires' | 'pre') langExpr ';'
 | ('ensures' | 'post') langExpr ';'
 | 'diverges' langExpr ';'
 | ('signals' | 'exsures') '(' langType langId? ')' langExpr? ';'
 | 'signals_only' ( langType (',' langType)* | '\\nothing') ';'
 | 'measured_by' langExpr ('if' langExpr)? ';'
 ;

valEmbedLoopSpec: valEmbedLoopSpecBlock+;

valEmbedLoopSpecBlock
 : startSpec valLoopClause* endSpec
 | {specLevel>0}? valLoopClause+
 ;

valLoopClause
 : ('loop_invariant' | 'maintaining') langExpr ';'
   | ('decreases' | 'decreasing' | 'loop_decreases') langExpr ';'
   | ('loop_modifies' | 'loop_assigns') valExpressionList ';'
 ;

valBehaviorClause
 : privacy?
   ('behavior' | 'behaviour'
    | 'normal_behavior' | 'normal_behaviour'
    | 'exceptional_behavior' | 'exceptional_behaviour')
   valContractClause*
 ;

valLinewiseBehaviorClause
 : startSpec privacy?
   ('behavior' | 'behaviour'
    | 'normal_behavior' | 'normal_behaviour'
    | 'exceptional_behavior' | 'exceptional_behaviour')
   endSpec (startSpec valContractClause endSpec)*
 ;

valLinewiseAlsoBehaviorClause
 : startSpec 'also' privacy?
   ('behavior' | 'behaviour'
    | 'normal_behavior' | 'normal_behaviour'
    | 'exceptional_behavior' | 'exceptional_behaviour')
   endSpec (startSpec valContractClause endSpec)*
 ;

privacy
 : 'public' | 'protected' | 'private'
 ;

valEmbedStatementBlock
 : startSpec valOrLangStatement* endSpec
 | {specLevel>0}? valStatement+
 ;

valOrLangStatement
 : valStatement | langStatement
 ;

valEmbedDeclarationBlock
 : startSpec valDeclaration* endSpec
 ;


valExpressionList
    : langExpr (',' langExpr)*
    ;

typeSpec
 : ownershipModifier* langType
 | '\\TYPE' ('[' ']')*
 ;

ownershipModifier
 : '\\rep' | '\\peer' | '\\readonly'
 | 'rep' | 'peer' | 'readonly'
 ;

valStatement
 : 'assert' langExpr (':' langExpr)?';'
 | 'assume' langExpr (':' langExpr)? ';'
 | 'set' langExpr ';'
 | 'unreachable'
 | valBlock
 ;

valBlock
 : '{' valOrLangStatement* '}'
 ;

valEqOp: '<==>' | '<=!=>';
valImpOp: '==>' | '<==';

valIndexExpression: '*' | langExpr '..' langExpr;

valPrimary
    : '\\result'
    | ('\\old' '(' langExpr ( ',' langId )? ')'
      | '\\pre' '(' langExpr ')')
    | '\\not_assigned' '(' storeRefList ')'
    | '\\not_modified' '(' storeRefList ')'
    | '\\only_accessed' '(' storeRefList ')'
    | '\\only_assigned' '(' storeRefList ')'
    | '\\only_called' '(' langId (',' langId)* ')'
    | '\\only_captured' '(' storeRefList ')'
    | '\\fresh' '(' expressionList ')'
    | '\\reach' '(' langExpr ')'
    | '\\duration' '(' langExpr ')'
    | '\\space' '(' langExpr ')'
    | '\\working_space' '(' langExpr ')'
    | '\\nonnullelements' '(' langExpr ')'
    | '\\typeof' '(' langExpr ')'
    | '\\elemtype' '(' langExpr ')'
    | '\\type' '(' langType ')'
    | '\\lockset'
    | '\\max' '(' langExpr ')'
    | '\\is_initialized' '(' langType ')'
    | '\\invariant_for' '(' langExpr ')'
    | '(' '\\lblneg' langId langExpr ')'
    | '(' '\\lblpos' langId langExpr ')'
    | '(' specQuantifiedExpression ')'
    | specQuantifiedExpression
    | '\\count'
    | '\\nothing'
    | informalDescription
    ;

specQuantifiedExpression
    : quantifier quantifiedVarDecls ';' (langExpr ';')? langExpr
    ;

quantifier
    : '\\forall' | '\\exists'
    | '\\max' | '\\min'
    | '\\num_of' | '\\product' | '\\sum'
    ;
    
quantifiedVarDecls 
    : ('non_null' | 'nullable')? langType quantifiedVarDeclarator (',' quantifiedVarDeclarator)*
    ;
quantifiedVarDeclarator 
    : langId ('[' ']')*
    ;
specVariableDeclarator
    : langId ('[' ']')* ('=' specInitializer)?
    ;
specArrayInitializer
    : '{' (specInitializer (',' specInitializer)* (',')? )? '}'
    ;
specInitializer 
    : expression
    | specArrayInitializer
    ;
setComprehension 
    : '{' ('non_null' | 'nullable')? langType quantifiedVarDeclarator '|' expression '&&' expression '}'
    ;
storeRefList 
    : storeRefKeyword 
    | storeRef ( ',' storeRef )*
    ;
storeRef 
    : storeRefExpression
    | informalDescription
    ;
storeRefExpression 
    : storeRefName ( storeRefNameSuffix )*
    ;
storeRefName 
    : langId | 'super' | 'this'
    ;
storeRefNameSuffix 
    : '.' langId 
    | '.' 'this' 
    | '[' specArrayRefExpr ']' 
    | '.' '*'
    ;
specArrayRefExpr 
    : expression
    | (expression '..' expression
        | FLOAT_LITERAL FLOAT_LITERAL) // workaround for 0..5 being lexed as 0. and .5
    | '*'
    ;
storeRefKeyword 
    : '\\nothing' | '\\everything' | '\\not_specified'
    ;


valReserved
 : ( ASSERT
    | VAL_PURE | VAL_AXIOM | VAL_MODIFIES
    | VAL_ACCESSIBLE | VAL_REQUIRES | VAL_ENSURES | VAL_LOOP_INVARIANT
    | VAL_ASSUME
    | VAL_GHOST
    | VAL_SIGNALS )
 | LANG_ID_ESCAPE
 | '\\result'
 | BOOL_LITERAL
 ;

valDeclaration
 : valModifier* 'invariant' langExpr ';'
   | valModifier* 'initially' langExpr ';'
   | valModifier* 'readable' langId 'if' langExpr ';'
   | valModifier* 'writable' langId 'if' langExpr ';'
   | valModifier* 'monitors_for' langId '=' valExpressionList ';'
   | 'axiom' langExpr ';'
   | (('also')? valSpecCase ('also' valSpecCase)*)? valModifier* 'pure' valModifier*
     langType langId formalParameters methodBody
   | valModifier* memberDeclaration
 ;

informalDescription
    : InformalDescStart InformalDescContent* InformalDescEnd
    ;
