import buildinfo.BuildInfo
import intermediateRepresentation.IRProgram
import languages.{IProgram, IProgramGenerator}
import languages.Krakatoa.KrakatoaProgram
import languages.OpenJML.OpenJMLProgram
import languages.VerCors.VerCorsProgram

import java.io.PrintWriter
import java.nio.file.{Files, Paths}
import java.util.regex.Pattern
import scala.annotation.tailrec
import org.rogach.scallop._

object Main {

  // map lowercase tool names to respective class/object
  val tools: Map[String, IProgramGenerator[_<:IProgram]]
        = Map("vercors" -> VerCorsProgram,
              "krakatoa" -> KrakatoaProgram,
              "openjml" -> OpenJMLProgram)

  // whether to print detailed information upon errors
  var debugMode: Boolean = false

  // command line parameter handling using Scallop for nice parsing, help message, etc.
  class Conf(arguments: Seq[String]) extends ScallopConf(arguments) {
    version(s"${BuildInfo.name} ${BuildInfo.version} (c) 2022 University of Twente")
    banner("""
             |Usage: AnnotationTranslator -s INPUT_LANG -t OUTPUT_LANG [OPTION]... FILES...
             |
             |Translates JML-like specification annotations from INPUT_LANG to OUTPUT_LANG
             |Options:
             |""".stripMargin)
    footer("""
      |Exit code:   0 for success,
      |          1- 9 for unexpected terminations,
      |         10-19 for errors in invoking this tool,
      |         20-29 for errors translating to the intermediate representation,
      |         30-39 for errors translating from IR to target language,
      |         40-49 for errors in post-processing
      |         """.stripMargin)
    appendDefaultToDescription = true

    val srcLanguage = opt[String](descr = "tool/language of the input file(s)", required = true)
    val targetLanguage = opt[String](descr = "tool/language of the output file(s)", required = true)
    val debug = opt[Boolean](descr = "print more verbose error messages", default = Some(false))
    val outputDir = opt[String](default = Some(Paths.get(".", "translated").toString),
                                descr = s"directory where to place files generated by translator")
    val files = trailArg[List[String]](descr = "files to translate", required = true)
    verify()

    override def onError(e: Throwable): Unit = {
      e match {
        case exceptions.ScallopException(message) =>
          System.err.println("[Error] Error using command line interface: " + e.getMessage)
          System.err.println("[Error] Use -h for help")
          sys.exit(10)
        case other => super.onError(other)
      }
    }

  }


  // extract the name of the given IProgram class/object
  private def getName(program: IProgramGenerator[_]): String = {
    // class name ends in $, so drop last character
    program.getClass.getSimpleName.dropRight(1)
  }

  def main(args: Array[String]): Unit = {
    // parse command line arguments
    val conf: Conf = new Conf(args)
    if(!tools.keySet.contains(conf.srcLanguage.apply().toLowerCase)) {
      System.err.println(s"[Error] Invalid value for ${conf.srcLanguage.name}, must be from (${tools.keys.mkString(", ")})")
      sys.exit(10)
    }
    if(!tools.keySet.contains(conf.targetLanguage.apply().toLowerCase)) {
      System.err.println(s"[Error] Invalid value for ${conf.targetLanguage.name}, must be from (${tools.keys.mkString(", ")})")
      sys.exit(10)
    }

    val src: IProgramGenerator[_ <: IProgram] = tools.apply(conf.srcLanguage.apply().toLowerCase)
    val target: IProgramGenerator[_ <: IProgram] = tools.apply(conf.targetLanguage.apply().toLowerCase)

    val outputDir = Paths.get(conf.outputDir.apply(), getName(target))
    Files.createDirectories(outputDir)
    val files = conf.files.apply()
    debugMode = conf.debug.apply()

    var maxErr = 0
    for (srcFile <- files) {
      val res = translateFile(srcFile, outputDir.toString, src, target)
      if(res > maxErr) {
        maxErr = res
      }
    }
    sys.exit(maxErr)
  }

  def translateFile(srcFile: String, outputDir: String,
                    src: IProgramGenerator[_ <: IProgram], target: IProgramGenerator[_ <: IProgram]): Int = {
    var parsed: IProgram = null
    var ir: IRProgram = null
    var res: IProgram = null
    val filename = srcFile.split(Pattern.quote(System.getProperty("file.separator"))).last
    val outputPath = Paths.get(outputDir, filename).toString

    println(s"Converting file \"$srcFile\" from ${getName(src)} to ${getName(target)}, "
            + s"storing result at \"$outputPath\"")
    try {
      parsed = src.fromFile(srcFile)
    } catch {
      case e: Throwable =>
        System.err.println(s"[Error] could not parse file $srcFile")
        if (debugMode) {
          throw e
        } else {
          System.err.println(e)
          return 21
        }
    }
    try {
      ir = parsed.toIRProgram
    } catch {
      case e: Throwable =>
        System.err.println(s"[Error] could not transform file $srcFile to Intermediate Representation")
        if (debugMode) {
          throw e
        } else {
          System.err.println(e)
          return 22
        }
    }
    try {
      res = target.fromIRProgram(ir)
    } catch {
      case e: Throwable =>
        System.err.println(s"[Error] could not transform file $srcFile to ${getName(target)}")
        if (debugMode) {
          throw e
        } else {
          System.err.println(e)
          return 30
        }
    }
    val writer = new PrintWriter(outputPath, "UTF-8")
    try {
      writer.write(res.toText)
      writer.close()
      return 0
    } catch {
      case e: Throwable =>
        writer.close()
        System.err.println(s"[Error] could not write output file to $outputPath")
        if (debugMode) {
          throw e
        } else {
          System.err.println(e)
          return 40
        }
    }
  }

  def runOpenJMLParser(args: Array[String]) : Int = {
    val text = "/* a preamble */\n package some.packageName;\n import some.other.packge.*; // a description\n import anoter.packge;\n class T { // some comment\n int x =3; /*@ requires x > 0; @*/ void test();\n }"

    val openjml = OpenJMLProgram.fromText(text)
    println(openjml.toText)
    openjml.toIRProgram
    0
  }

  def runVerCorsParser(args: Array[String]) : Int = {
    val input = List(
      "// -*- tab-width:2 ; indent-tabs-mode:nil -*-\n//:: cases JavaArrayExamples\n//:: suite puptol\n//:: tools silicon\n//:: verdict Pass\n\n/*\n  Answer to question 4 of the FMSE exam on April 16, 2015.\n*/\nclass Exercise4 {\n  \n  /*@\n     context_everywhere a!=null && b != null && res != null;\n     context_everywhere a.length==b.length && res.length==2*a.length;\n     context_everywhere (\\forall* int i ; 0 <= i && i < a.length ; Perm(a[i],1));\n     context_everywhere (\\forall* int i ; 0 <= i && i < b.length ; Perm(b[i],1));\n     context_everywhere (\\forall* int i ; 0 <= i && i < res.length ; Perm(res[i],1));\n     ensures (\\forall int i ; 0 <= i && i < a.length ; res[2*i]==a[i] );\n     ensures (\\forall int i ; 0 <= i && i < b.length ; res[2*i+1]==b[i] );\n   */\n  void zip(int[] a, int[] b, int[] res) {\n    int k = 0;\n    /*@\n      loop_invariant 0 <= k ** k%2 ==0 ** k <= res.length;\n      loop_invariant (\\forall int i ; 0 <= i && i < k/2 ; res[2*i]==a[i] );\n      loop_invariant (\\forall int i ; 0 <= i && i < k/2 ; res[2*i+1]==b[i] );      \n     */\n    while (k < res.length) {\n      res[k] = a[k/2];\n      res[k+1] = b[k/2];\n      k = k+2;\n    }\n  }\n}\n\nclass JavaArrayExamples {\n  /*@\n    context_everywhere a != null;\n    context (\\forall* int i ; 0 <= i && i < a.length ; Perm(a[i],write));\n  @*/\n  public void shift_drf(int a[]){\n    for(int i=0;i<a.length;i++)\n    /*@\n      requires Perm(a[i],1\\2);\n      requires i==0 ==> Perm(a[i],1\\2);\n      requires i < a.length-1 ==> Perm(a[i+1],1\\2);\n      ensures  Perm(a[i],write);\n    @*/\n    {\n      //@ ghost S1:if(i>0){ recv 0 < i ** i < a.length ** Perm(a[i],1\\2) from S2,1; }\n      S2:if (i < a.length-1) {\n        a[i]=a[i+1];\n        //@ send 0 <= i ** i < a.length - 1 ** Perm(a[i+1],1\\2) to S1,1;\n      }\n    }\n  }\n  \n  /*@ \n    context_everywhere ar!=null;\n    context_everywhere (\\forall* int k ; 0 <= k && k < ar.length ; Perm(ar[k],1));\n    ensures (\\forall int k ; 0 <= k && k < ar.length ; ar[k]==0 ) ;\n  @*/\n  public void zero_array(int ar[]){\n    int i=0;\n    int N=ar.length;\n    //@ loop_invariant 0<= i && i<=N && N==ar.length ;\n    //@ loop_invariant (\\forall int k ; 0 <= k && k < i ; ar[k]==0 ) ;\n    while(i<N){\n      ar[i]=0;\n      i++;\n    }\n  }\n\t\n  /*@ \n      context_everywhere ar != null ** M>0 ** N > 0 ** M * N == ar.length;\n      context   (\\forall* int k ; 0 <= k && k < ar.length ; Perm(ar[k],write));\n      ensures   (\\forall  int k ; 0 <= k && k < ar.length ; ar[k]==0 ) ;\n   */\n  public void zero_array_nested(int ar[],int M,int N){\n    for(int i=0;i<M;i++)\n    /*@\n      context (\\forall* int k ; i*N <= k && k < (i+1)*N ; Perm(ar[k],write));\n      ensures (\\forall  int k ; i*N <= k && k < (i+1)*N ; ar[k]==0 ) ;\n    @*/\n    {\n      for(int j=0;j<N;j++)\n      /*@\n\t\tcontext 0 <= i && i < M;\n        context Perm(ar[i*N+j],write);\n        ensures  ar[i*N+j]==0;\n      @*/\n      {\n        ar[i*N+j]=0;\n      }\n    }\n  }\n\n  /*@ \n    context_everywhere ar != null ** M>0 ** N > 0 ** M * N == ar.length;\n    context (\\forall* int k ; 0 <= k && k < ar.length ; Perm(ar[k],write));\n    ensures (\\forall  int k ; 0 <= k && k < ar.length ; ar[k]==0 ) ;\n  @*/\n  public void zero_array_smart_nested(int ar[],int M,int N) {\n    for(int i=0;i<M;i++) {\n      for(int j=0;j<N;j++)\n      /*@\n        context Perm(ar[i*N+j],write);\n        ensures ar[i*N+j]==0;\n      @*/\n      {\n        ar[i*N+j]=0;\n      }\n    }\n  }\n\n}\n"
    )
    input.foreach(VerCorsProgram.fromText)
    0
  }

  def runKrakatoaParser(args: Array[String]) : Int = {
    val input = List(
      """/* McCarthy's ``91'' function. */

      public class MacCarthy {

      /*@ decreases 101-n ;
        @ behavior less_than_101:
        @   assumes n <= 100;
        @   ensures \result == 91;
        @ behavior greater_than_100:
        @   assumes n >= 101;
        @   ensures \result == n - 10;
        @*/
      public static int f91(int n) {
        if (n <= 100) {
          return f91(f91(n + 11));
        }
        else
          return n - 10;
      }

    }

    /*
    Local Variables:
    compile-command: "make MacCarthy.why3ml"
    End:
    */""""""
    )
    input.foreach(KrakatoaProgram.fromText)
    0
  }

}
