package generic.syntaxTree


trait Expression extends TreeNode with Initializer with LambdaBody with IndexExpression

trait NestedInvocationExpression extends Expression {
  val left: Expression
  val dot: StringNode = ConstantStringNode(".")
  val right: Expression

  override def stringRepresentation: String = {
    var res = left.toString
    res += dot.toString
    res += right.toString
    res
  }

  def copyWhitespace(node: NestedInvocationExpression): Unit = {
    dot.copyWhitespace(node.dot)
    super.copyWhitespace(node)
  }
}

trait BinaryExpression extends Expression {
  val left: Expression
  val operator: StringNode
  val right: Expression

  override def stringRepresentation: String = {
    var res = left.toString
    res += operator.toString
    res += right.toString
    res
  }
}

trait InstanceOfExpression extends Expression {
  val left: Expression
  val instanceofKeyword: StringNode = ConstantStringNode("instanceof")
  val right: TypeIndication

  override def stringRepresentation: String = {
    var res = left.toString
    res += instanceofKeyword.toString
    res += right.toString
    res
  }
  
  def copyWhitespace(node: InstanceOfExpression): Unit = {
    instanceofKeyword.copyWhitespace(node.instanceofKeyword)
    super.copyWhitespace(node)
  }
}

trait ArrayIndexExpression extends Expression {
  val arrayExpression: Expression
  val openBracket: StringNode = ConstantStringNode("[")
  val indexExpression: IndexExpression
  val closeBracket: StringNode = ConstantStringNode("]")

  override def stringRepresentation: String = {
    var res = arrayExpression.toString
    res += openBracket.toString
    res += indexExpression.toString
    res += closeBracket.toString
    res
  }
  
  def copyWhitespace(node: ArrayIndexExpression): Unit = {
    openBracket.copyWhitespace(node.openBracket)
    closeBracket.copyWhitespace(node.closeBracket)
    super.copyWhitespace(node)
  }
}
trait IndexExpression extends TreeNode
trait IndexAllExpression extends IndexExpression {
  val star: StringNode = ConstantStringNode("*")

  override def stringRepresentation: String = star.toString

  def copyWhitespace(node: IndexAllExpression): Unit = {
    star.copyWhitespace(node.star)
    super.copyWhitespace(node)
  }
}
trait RangeExpression extends Expression with IndexExpression {
  val start: Expression
  val dots: StringNode = ConstantStringNode("..")
  val end: Expression

  override def stringRepresentation: String = start.toString + dots.toString + end.toString

  def copyWhitespace(node: RangeExpression): Unit = {
    dots.copyWhitespace(node.dots)
    super.copyWhitespace(node)
  }
}
trait PartialRangeExpression extends Expression with IndexExpression {
  val start: Option[Expression]
  val dots: StringNode = ConstantStringNode("..")
  val end: Option[Expression]

  override def stringRepresentation: String = {
    start.getOrElse("").toString + dots.toString + end.getOrElse("").toString
  }

  def copyWhitespace(node: PartialRangeExpression): Unit = {
    dots.copyWhitespace(node.dots)
    super.copyWhitespace(node)
  }
}
trait MethodCallExpression extends Expression {
  val name: StringNode
  val arguments: Arguments

  override def stringRepresentation: String = {
    var res = name.toString
    res += arguments.toString
    res
  }
}
trait NewExpression extends Expression {
  val newKeyword: StringNode = ConstantStringNode("new")
  var typeIndication: TypeIndication
  
  def copyWhitespace(node: NewExpression): Unit = {
    newKeyword.copyWhitespace(node.newKeyword)
    super.copyWhitespace(node)
  }
}
trait NewObjectExpression extends NewExpression {
  var typeArguments: Option[TypeArgumentClause]
  val arguments: Arguments
  val body: Option[Body]

  override def stringRepresentation: String = {
    var res = newKeyword.toString
    res += typeArguments.getOrElse("").toString
    res += typeIndication.toString
    res += arguments.toString
    res += body.getOrElse("").toString
    res
  }
}
trait NewArrayExpression extends NewExpression {
  val dimensions: List[Dimension]
  val initializer: Option[ArrayInitializer]

  override def stringRepresentation: String = {
    var res = newKeyword.toString
    res += typeIndication.toString
    res += dimensions.mkString
    res += initializer.getOrElse("").toString
    res
  }
}
trait Dimension extends TreeNode {
  val openBracket: StringNode = ConstantStringNode("[")
  val sizeExpression: Option[Expression]
  val closeBracket: StringNode = ConstantStringNode("]")

  override def stringRepresentation: String = {
    var res = openBracket.toString
    res += sizeExpression.getOrElse("").toString
    res += closeBracket.toString
    res
  }

  def copyWhitespace(node: Dimension): Unit = {
    openBracket.copyWhitespace(node.openBracket)
    closeBracket.copyWhitespace(node.closeBracket)
    super.copyWhitespace(node)
  }
}

trait CastExpression extends Expression {
  val openParenthesis: StringNode = ConstantStringNode("(")
  val annotations: List[Annotation]
  val typeIndications: List[TypeIndication]
  val closeParenthesis: StringNode = ConstantStringNode(")")
  val expression: Expression

  override def stringRepresentation: String = {
    var res = openParenthesis.toString
    res += annotations.mkString
    res += typeIndications.mkString
    res += closeParenthesis.toString
    res += expression.toString
    res
  }

  def copyWhitespace(node: CastExpression): Unit = {
    openParenthesis.copyWhitespace(node.openParenthesis)
    closeParenthesis.copyWhitespace(node.closeParenthesis)
    super.copyWhitespace(node)
  }
}
trait PostfixExpression extends Expression {
  val expression: Expression
  val postfix: StringNode

  override def stringRepresentation: String = expression.toString + postfix.toString
}
trait PrefixExpression extends Expression {
  val prefix: StringNode
  val expression: Expression

  override def stringRepresentation: String = prefix.toString + expression.toString
}
trait TernaryExpression extends Expression {
  val condition: Expression
  val questionmark: StringNode = ConstantStringNode("?")
  val thenExpression: Expression
  val colon: StringNode = ConstantStringNode(":")
  val elseExpression: Expression

  override def stringRepresentation: String = {
    var res = condition.toString
    res += questionmark.toString
    res += thenExpression.toString
    res += colon.toString
    res += elseExpression.toString
    res
  }

  def copyWhitespace(node: TernaryExpression): Unit = {
    questionmark.copyWhitespace(node.questionmark)
    colon.copyWhitespace(node.colon)
    super.copyWhitespace(node)
  }
}
trait LambdaExpression extends Expression {
  val hasParentheses: Boolean
  val openParenthesis: StringNode = ConstantStringNode("(")
  val parameters: List[Parameter]
  val closeParenthesis: StringNode = ConstantStringNode(")")
  val arrow: StringNode = ConstantStringNode("->")
  val body: LambdaBody

  override def stringRepresentation: String = {
    var res = ""
    if (hasParentheses) {
      res += openParenthesis.toString
    }
    res += parameters.mkString(",")
    if (hasParentheses) {
      res += closeParenthesis.toString
    }
    res += arrow.toString
    res += body.toString
    res
  }

  def copyWhitespace(node: LambdaExpression): Unit = {
    openParenthesis.copyWhitespace(node.openParenthesis)
    closeParenthesis.copyWhitespace(node.closeParenthesis)
    arrow.copyWhitespace(node.arrow)
    super.copyWhitespace(node)
  }
}
trait LambdaBody extends TreeNode


trait MethodReferenceExpression extends Expression {
  val left: Expression
  val doubleColon: StringNode = ConstantStringNode("::")
  val typeArgument: Option[TypeArgumentClause]
  val right: Expression // todo: express differently?

  override def stringRepresentation: String = {
    var res = left.toString
    res += doubleColon.toString
    res += typeArgument.getOrElse("").toString
    res += right.toString
    res
  }

  def copyWhitespace(node: MethodReferenceExpression): Unit = {
    doubleColon.copyWhitespace(node.doubleColon)
    super.copyWhitespace(node)
  }
}

trait ParExpression extends Expression {
  val openParenthesis: StringNode = ConstantStringNode("(")
  val expression: Expression
  val closeParenthesis: StringNode = ConstantStringNode(")")

  override def stringRepresentation: String = {
    openParenthesis.toString +
      expression.toString +
      closeParenthesis.toString
  }

  def copyWhitespace(node: ParExpression): Unit = {
    openParenthesis.copyWhitespace(node.openParenthesis)
    closeParenthesis.copyWhitespace(node.closeParenthesis)
    super.copyWhitespace(node)
  }
}

trait Literal extends Expression {
  val value: StringNode

  override def stringRepresentation: String = value.toString
}

trait ClassName extends Expression {
  val typeIndication: TypeIndication
  val dot: StringNode = ConstantStringNode(".")
  val classKeyword: StringNode = ConstantStringNode("class")

  override def stringRepresentation: String = {
    var res = typeIndication.toString
    res += dot.toString
    res += classKeyword.toString
    res
  }

  def copyWhitespace(node: ClassName): Unit = {
    dot.copyWhitespace(node.dot)
    classKeyword.copyWhitespace(node.classKeyword)
    super.copyWhitespace(node)
  }
}

trait GenericInvocationExpression extends Expression {
  val typeArgumentClause: TypeArgumentClause
  val name: StringNode
  val arguments: Option[Arguments]

  override def stringRepresentation: String = {
    var res = typeArgumentClause.toString
    res += name.toString
    res += arguments.getOrElse("").toString
    res
  }
}

