package generic.syntaxTree

trait CompilationUnit extends TreeNode {
  val pkg: Option[PackageDeclaration]
  val imports: List[ImportDeclaration]
  val declarations: List[TypeDeclaration]

  override def stringRepresentation: String = {
    var res = pkg.getOrElse("").toString
    res += imports.mkString
    res += declarations.mkString
    res
  }
}
trait NonTranslatedCompilationUnit extends CompilationUnit with StringNode

trait PackageDeclaration extends TreeNode {
  val annotations: List[Annotation]
  val packageKeyword: StringNode = ConstantStringNode("package")
  val name: StringNode
  val semicolon: StringNode = ConstantStringNode(";")

  override def stringRepresentation: String = {
    var res = annotations.mkString
    res += packageKeyword.toString
    res += name.toString
    res += semicolon.toString
    res
  }

  def copyWhitespace(node: PackageDeclaration): Unit = {
    packageKeyword.copyWhitespace(node.packageKeyword)
    semicolon.copyWhitespace(node.semicolon)
    super.copyWhitespace(node)
  }
}
trait ImportDeclaration extends TreeNode {
  val importKeyword: StringNode = ConstantStringNode("import")
  var modifiers: List[Modifier] = Nil
  val name: StringNode
  val semicolon: StringNode = ConstantStringNode(";")

  override def stringRepresentation: String = {
    var res = importKeyword.toString
    res += modifiers.mkString
    res += name.toString
    res += semicolon.toString
    res
  }

  def copyWhitespace(node: ImportDeclaration): Unit = {
    importKeyword.copyWhitespace(node.importKeyword)
    semicolon.copyWhitespace(node.semicolon)
    super.copyWhitespace(node)
  }
}

trait TypeDeclaration extends TreeNode with Statement
trait NonTranslatedTypeDeclaration extends TypeDeclaration with StringNode

trait Modifier extends TreeNode
trait StaticModifier extends StringNode with Modifier {
  var text: String = "static"
}

trait GenericsSuperclassClause extends TreeNode {
  val extendsKeyword: StringNode = ConstantStringNode("extends")
  val superclasses: List[TypeIndication]

  override def stringRepresentation: String = {
    var res = extendsKeyword.toString
    res += superclasses.mkString("&")
    res
  }

  def copyWhitespace(node: GenericsSuperclassClause): Unit = {
    extendsKeyword.copyWhitespace(node.extendsKeyword)
    super.copyWhitespace(node)
  }
}
trait SuperclassClause extends TreeNode {
  val extendsKeyword: StringNode = ConstantStringNode("extends")
  val superclass: TypeIndication

  override def stringRepresentation: String = {
    var res = extendsKeyword.toString
    res += superclass.toString
    res
  }

  def copyWhitespace(node: SuperclassClause): Unit = {
    extendsKeyword.copyWhitespace(node.extendsKeyword)
    super.copyWhitespace(node)
  }
}
trait MultiSuperclassClause extends TreeNode {
  val extendsKeyword: StringNode = ConstantStringNode("extends")
  val superclasses: List[TypeIndication]

  override def stringRepresentation: String = {
    var res = extendsKeyword.toString
    res += superclasses.mkString(",")
    res
  }

  def copyWhitespace(node: MultiSuperclassClause): Unit = {
    extendsKeyword.copyWhitespace(node.extendsKeyword)
    super.copyWhitespace(node)
  }
}
trait InterfaceClause extends TreeNode {
  val implementsKeyword: StringNode = ConstantStringNode("implements")
  val interfaces: List[TypeIndication]

  override def stringRepresentation: String = {
    var res = implementsKeyword.toString
    res += interfaces.mkString(",")
    res
  }

  def copyWhitespace(node: InterfaceClause): Unit = {
    implementsKeyword.copyWhitespace(node.implementsKeyword)
    super.copyWhitespace(node)
  }
}

trait ClassDeclaration extends TypeDeclaration with MemberDeclaration {
  val classKeyword: StringNode = ConstantStringNode("class")
  val name: StringNode
  val typeParameters: Option[TypeParameterClause]
  val superclass: Option[SuperclassClause]
  val interfaces: Option[InterfaceClause]
  val body: Body

  override def stringRepresentation: String = {
    var res = modifiers.mkString
    res += classKeyword.toString
    res += name.toString
    res += typeParameters.getOrElse("").toString
    res += superclass.getOrElse("").toString
    res += interfaces.getOrElse("").toString
    res += body.toString
    res
  }

  def copyWhitespace(node: ClassDeclaration): Unit = {
    classKeyword.copyWhitespace(node.classKeyword)
    super.copyWhitespace(node)
  }
}

trait TypeParameterClause extends TreeNode {
  val openAngle: StringNode = ConstantStringNode("<")
  val parameters: List[TypeParameter]
  val closeAngle: StringNode = ConstantStringNode(">")

  override def stringRepresentation: String = {
    var res = openAngle.toString
    res += parameters.mkString(",")
    res += closeAngle.toString
    res
  }

  def copyWhitespace(node: TypeParameterClause): Unit = {
    openAngle.copyWhitespace(node.openAngle)
    closeAngle.copyWhitespace(node.closeAngle)
    super.copyWhitespace(node)
  }
}
trait TypeParameter extends TreeNode {
  val annotations: List[Annotation]
  val name: TypeIndication
  val superclass: Option[GenericsSuperclassClause]

  override def stringRepresentation: String = {
    var res = name.toString
    res += superclass.getOrElse("").toString
    res
  }
}

trait EnumDeclaration extends TypeDeclaration with MemberDeclaration {
  val enumKeyword: StringNode = ConstantStringNode("enum")
  val name: StringNode
  val interfaces: Option[InterfaceClause]
  val openBrace: StringNode = ConstantStringNode("{")
  val constants: List[EnumConstant]
  val comma: Option[StringNode]
  val semicolon: Option[StringNode]
  val declarations: List[MemberDeclaration]
  val closeBrace: StringNode = ConstantStringNode("}")

  override def stringRepresentation: String = {
    var res = modifiers.mkString
    res += enumKeyword.toString
    res += name.toString
    res += interfaces.getOrElse("").toString
    res += openBrace.toString
    res += constants.mkString(",")
    res += comma.getOrElse("").toString
    res += semicolon.getOrElse("").toString
    res += declarations.mkString
    res += closeBrace.toString
    res
  }

  def copyWhitespace(node: EnumDeclaration): Unit = {
    enumKeyword.copyWhitespace(node.enumKeyword)
    openBrace.copyWhitespace(node.openBrace)
    closeBrace.copyWhitespace(node.closeBrace)
    super.copyWhitespace(node)
  }
}
trait EnumConstant extends TreeNode {
  val annotations: List[Annotation]
  val name: StringNode
  val arguments: Option[Arguments]
  val body: Option[Body]

  override def stringRepresentation: String = {
    var res = annotations.mkString
    res += name.toString
    res += arguments.getOrElse("").toString
    res += body.getOrElse("").toString
    res
  }
}

trait InterfaceDeclaration extends TypeDeclaration with MemberDeclaration {
  val interfaceKeyword: StringNode = ConstantStringNode("interface")
  val name: StringNode
  val typeParameter: Option[TypeParameterClause]
  val superInterfaces: Option[MultiSuperclassClause]
  val body: Body

  override def stringRepresentation: String = {
    var res = modifiers.mkString
    res += interfaceKeyword.toString
    res += name.toString
    res += typeParameter.getOrElse("").toString
    res += superInterfaces.getOrElse("").toString
    res += body.toString
    res
  }

  def copyWhitespace(node: InterfaceDeclaration): Unit = {
    interfaceKeyword.copyWhitespace(node.interfaceKeyword)
    super.copyWhitespace(node)
  }
}

trait Body extends TreeNode {
  val openBrace: StringNode = ConstantStringNode("{")
  var declarations: List[MemberDeclaration]
  val closeBrace: StringNode = ConstantStringNode("}")

  override def stringRepresentation: String = {
    var res = openBrace.toString
    res += declarations.mkString
    res += closeBrace
    res
  }

  def copyWhitespace(node: Body): Unit = {
    openBrace.copyWhitespace(node.openBrace)
    closeBrace.copyWhitespace(node.closeBrace)
    super.copyWhitespace(node)
  }
}

trait MemberDeclaration extends TreeNode {
  var contractBefore: Option[Contract]
  var modifiers: List[Modifier]
}
trait NonTranslatedMemberDeclaration extends MemberDeclaration with StringNode

trait MethodDeclaration extends MemberDeclaration {
  var returnType: Option[TypeIndication]    // None for constructor
  var name: StringNode
  val openParenthesis: StringNode = ConstantStringNode("(")
  var parameters: List[Parameter]
  val closeParenthesis: StringNode = ConstantStringNode(")")
  var dimensions: Option[StringNode]
  var throws: Option[ThrowsClause]
  var default: Option[DefaultClause]
  var contractWithin: Option[Contract]
  var body: BlockOrSemicolon

  override def stringRepresentation: String = {
    var res = contractBefore.getOrElse("").toString
    res += modifiers.mkString
    res += returnType.getOrElse("").toString
    res += name.toString
    res += openParenthesis.toString
    res += parameters.mkString(",")
    res += closeParenthesis.toString
    res += dimensions.getOrElse("").toString
    res += throws.getOrElse("").toString
    res += default.getOrElse("").toString
    res += contractWithin.getOrElse("").toString
    res += body.toString
    res
  }

  def copyWhitespace(node: MethodDeclaration): Unit = {
    openParenthesis.copyWhitespace(node.openParenthesis)
    closeParenthesis.copyWhitespace(node.closeParenthesis)
    super.copyWhitespace(node)
  }
}
trait GenericMethodDeclaration extends MethodDeclaration {
  val typeParameters: TypeParameterClause

  override def stringRepresentation: String = {
    var res = super.stringRepresentation
    if (typeParameters != null) {
      res = typeParameters.toString + res
    }
    res
  }
}

trait ThrowsClause extends TreeNode {
  val throwsKeyword: StringNode = ConstantStringNode("throws")
  val exceptions: List[StringNode]

  override def stringRepresentation: String = {
    var res = throwsKeyword.toString
    res += exceptions.mkString(",")
    res
  }

  def copyWhitespace(node: ThrowsClause): Unit = {
    throwsKeyword.copyWhitespace(node.throwsKeyword)
    super.copyWhitespace(node)
  }
}

// todo: leave it up to tool whether to make it just StringNode or more sophisticated?
// should also support '...' as part of varargs
trait TypeIndication extends TreeNode with TypeArgument {
  var annotations: List[Annotation]
  var modifiers: List[Modifier] = Nil
  val typeAsString: String
}
// may have whitespace after it in case of a list of types, e.g. in "implements" clause
trait GenericTypeIndication extends TypeIndication {
  val typeParameters: TypeParameterClause
}

trait BlockOrSemicolon extends MemberDeclaration
trait FieldDeclaration extends MemberDeclaration with TreeNode {
  val fieldType: TypeIndication
  val declarators: List[VariableDeclarator]
  val semicolon: StringNode = ConstantStringNode(";")

  override def stringRepresentation: String = {
    var res = modifiers.mkString
    res += fieldType.toString
    res += declarators.mkString(",")
    res += semicolon.toString
    res
  }

  def copyWhitespace(node: FieldDeclaration): Unit = {
    semicolon.copyWhitespace(node.semicolon)
    super.copyWhitespace(node)
  }
}

trait VariableDeclarator extends TreeNode {
  val name: StringNode
  val equalSign: StringNode = ConstantStringNode("=")
  val initializer: Option[Initializer]

  override def stringRepresentation: String = {
    var res = name.toString
    if (initializer.nonEmpty) {
      res += equalSign.toString
      res += initializer.get.toString
    }
    res
  }

  def copyWhitespace(node: VariableDeclarator): Unit = {
    equalSign.copyWhitespace(node.equalSign)
    super.copyWhitespace(node)
  }
}
trait Initializer extends TreeNode
trait ArrayInitializer extends Initializer {
  val openBrace: StringNode = ConstantStringNode("{")
  val initializers: List[Initializer]
  val trailingComma: Option[StringNode]
  val closeBrace: StringNode = ConstantStringNode("}")

  override def stringRepresentation: String = {
    var res = openBrace.toString
    res += initializers.mkString(",")
    res += closeBrace.toString
    res
  }

  def copyWhitespace(node: ArrayInitializer): Unit = {
    openBrace.copyWhitespace(node.openBrace)
    closeBrace.copyWhitespace(node.closeBrace)
    super.copyWhitespace(node)
  }
}

trait TypeArgumentClause extends TreeNode {
  val openAngle: StringNode = ConstantStringNode("<")
  val parameters: List[TypeArgument]
  val closeAngle: StringNode = ConstantStringNode(">")

  override def stringRepresentation: String = {
    var res = openAngle.toString
    res += parameters.mkString(",")
    res += closeAngle.toString
    res
  }

  def copyWhitespace(node: TypeArgumentClause): Unit = {
    openAngle.copyWhitespace(node.openAngle)
    closeAngle.copyWhitespace(node.closeAngle)
    super.copyWhitespace(node)
  }
}
trait TypeArgument extends TreeNode
trait BoundedTypeArgument extends TypeArgument {
  val annotations: List[Annotation]
  val questionMark: StringNode = ConstantStringNode("?")
  val superKeyword: StringNode = ConstantStringNode("super")
  val typeSuper: Option[TypeIndication]
  val extendsKeyword: StringNode = ConstantStringNode("extends")
  val typeExtends: Option[TypeIndication]

  override def stringRepresentation: String = {
    var res = annotations.mkString
    res += questionMark.toString
    if (typeSuper.nonEmpty) {
      res += superKeyword.toString + typeSuper.toString
    }
    if (typeExtends.nonEmpty) {
      res += extendsKeyword.toString + typeExtends.toString
    }
    res
  }

  def copyWhitespace(node: BoundedTypeArgument): Unit = {
    questionMark.copyWhitespace(node.questionMark)
    superKeyword.copyWhitespace(node.superKeyword)
    extendsKeyword.copyWhitespace(node.extendsKeyword)
    super.copyWhitespace(node)
  }
}

trait Parameter extends TreeNode {
  var modifiers: List[Modifier] = Nil
  val parameterType: TypeIndication // can also contain '...' from varargs
  val name: StringNode // todo: enough? can contain "[]" for arrays

  override def stringRepresentation: String = {
    var res = modifiers.mkString
    res += parameterType.toString
    res += name.toString
    res
  }
}

trait VarargsParameter extends Parameter {
  var annotations: List[Annotation]
  val dots: StringNode = ConstantStringNode("...")

  override def stringRepresentation: String = {
    var res = modifiers.mkString
    res += parameterType.toString
    res += annotations.mkString
    res += dots.toString
    res += name.toString
    res
  }

  def copyWhitespace(node: VarargsParameter): Unit = {
    dots.copyWhitespace(node.dots)
    super.copyWhitespace(node)
  }
}

trait Annotation extends Modifier // todo: Is StringAnnotation enough, or more types?
trait StringAnnotation extends Annotation with StringNode

trait AnnotationTypeDeclaration extends TypeDeclaration with MemberDeclaration {
  val at: StringNode = ConstantStringNode("@")
  val interfaceKeyword: StringNode = ConstantStringNode("interface")
  val name: StringNode
  val body: Body

  override def stringRepresentation: String = {
    var res = at.toString
    res += interfaceKeyword.toString
    res += name.toString
    res += body.toString
    res
  }

  def copyWhitespace(node: AnnotationTypeDeclaration): Unit = {
    at.copyWhitespace(node.at)
    interfaceKeyword.copyWhitespace(node.interfaceKeyword)
    super.copyWhitespace(node)
  }
}

trait DefaultClause extends TreeNode {
  val defaultKeyword: StringNode = ConstantStringNode("default")
  val value: StringNode   // Todo: enough? Default clause is for annotations only

  override def stringRepresentation: String = {
    var res = defaultKeyword.toString
    res += value.toString
    res
  }

  def copyWhitespace(node: DefaultClause): Unit = {
    defaultKeyword.copyWhitespace(node.defaultKeyword)
    super.copyWhitespace(node)
  }
}

trait Statement extends TreeNode
trait NonTranslatedStatement extends Statement with StringNode
trait Block extends Statement with BlockOrSemicolon with LambdaBody {
  val static: Option[Modifier]
  val openBrace: StringNode = ConstantStringNode("{")
  val statements: List[Statement]
  val closeBrace: StringNode = ConstantStringNode("}")

  override def stringRepresentation: String = {
    var res = static.getOrElse("").toString
    res += openBrace.toString
    res += statements.mkString
    res += closeBrace.toString
    res
  }

  def copyWhitespace(node: Block): Unit = {
    openBrace.copyWhitespace(node.openBrace)
    closeBrace.copyWhitespace(node.closeBrace)
    super.copyWhitespace(node)
  }
}
trait LocalDeclaration extends Statement with ForInit {
  var modifiers: List[Modifier]
  val declType: TypeIndication
  val declarators: List[VariableDeclarator]
  var isStatement: Boolean
  val semicolon: StringNode = ConstantStringNode(";")

  override def stringRepresentation: String = {
    var res = modifiers.mkString
    res += declType.toString
    res += declarators.mkString(",")
    if(isStatement) {
      res += semicolon.toString
    }
    res
  }

  def copyWhitespace(node: LocalDeclaration): Unit = {
    semicolon.copyWhitespace(node.semicolon)
    super.copyWhitespace(node)
  }
}
trait JavaAssertion extends Statement {
  val assertKeyword: StringNode = ConstantStringNode("assert")
  val content: Expression
  val colon: StringNode = ConstantStringNode(":")
  val message: Option[Expression]
  val semicolon: StringNode = ConstantStringNode(";")

  override def stringRepresentation: String = {
    var res = assertKeyword.toString
    res += content.toString
    if (message.nonEmpty) {
      res += colon.toString
      res += message.get.toString
    }
    res += semicolon.toString
    res
  }

  def copyWhitespace(node: JavaAssertion): Unit = {
    assertKeyword.copyWhitespace(node.assertKeyword)
    colon.copyWhitespace(node.colon)
    semicolon.copyWhitespace(node.semicolon)
    super.copyWhitespace(node)
  }
}
trait IfStatement extends Statement {
  val ifKeyword: StringNode = ConstantStringNode("if")
  val openParenthesis: StringNode = ConstantStringNode("(")
  val condition: Expression
  val closeParenthesis: StringNode = ConstantStringNode(")")
  val thenStatement: Statement
  val elseKeyword: StringNode = ConstantStringNode("else")
  val elseStatement: Option[Statement]

  override def stringRepresentation: String = {
    var res = ifKeyword.toString
    res += openParenthesis.toString
    res += condition.toString
    res += closeParenthesis.toString
    res += thenStatement.toString
    if (elseStatement.nonEmpty) {
      res += elseKeyword.toString
      res += elseStatement.get.toString
    }
    res
  }

  def copyWhitespace(node: IfStatement): Unit = {
    ifKeyword.copyWhitespace(node.ifKeyword)
    openParenthesis.copyWhitespace(node.openParenthesis)
    closeParenthesis.copyWhitespace(node.closeParenthesis)
    elseKeyword.copyWhitespace(node.elseKeyword)
    super.copyWhitespace(node)
  }
}
trait Loop extends Statement {
  var contractBefore: List[LoopSpecOrLabel]
  val openParenthesis: StringNode = ConstantStringNode("(")
  val closeParenthesis: StringNode = ConstantStringNode(")")
  val body: Statement

  def copyWhitespace(node: Loop): Unit = {
    openParenthesis.copyWhitespace(node.openParenthesis)
    closeParenthesis.copyWhitespace(node.closeParenthesis)
    super.copyWhitespace(node)
  }
}
trait LoopWithContractWithin extends Loop {
  var contractWithin: List[LoopSpecOrLabel]
}
trait ForLoop extends LoopWithContractWithin {
  val forKeyword: StringNode = ConstantStringNode("for")
  val control: ForLoopControl

  override def stringRepresentation: String = {
    var res = contractBefore.mkString
    res += forKeyword.toString
    res += openParenthesis.toString
    res += control.toString
    res += closeParenthesis.toString
    res += contractWithin.mkString
    res += body.toString
    res
  }

  def copyWhitespace(node: ForLoop): Unit = {
    forKeyword.copyWhitespace(node.forKeyword)
    super.copyWhitespace(node)
  }
}
trait ForLoopControl extends TreeNode
trait ForInit extends TreeNode
trait ExpressionListForInit extends ForInit {
  val expressions: List[Expression]

  override def stringRepresentation: String = {
    expressions.mkString(",")
  }
}
trait RegularForLoopControl extends ForLoopControl {
  val init: Option[ForInit]
  val firstSemicolon: StringNode = ConstantStringNode(";")
  val condition: Option[Expression]
  val secondSemicolon: StringNode = ConstantStringNode(";")
  val update: List[Expression]

  override def stringRepresentation: String = {
    var res = init.getOrElse("").toString
    res += firstSemicolon.toString
    res += condition.getOrElse("").toString
    res += secondSemicolon.toString
    res += update.mkString(",")
    res
  }

  def copyWhitespace(node: RegularForLoopControl): Unit = {
    firstSemicolon.copyWhitespace(node.firstSemicolon)
    secondSemicolon.copyWhitespace(node.secondSemicolon)
    super.copyWhitespace(node)
  }
}
trait ForEachLoopControl extends ForLoopControl {
  var modifiers: List[Modifier] = Nil
  val controlType: TypeIndication
  val variable: StringNode
  val colon: StringNode = ConstantStringNode(":")
  val control: Expression

  override def stringRepresentation: String = {
    var res = modifiers.mkString
    res += controlType.toString
    res += variable.toString
    res += colon.toString
    res += control.toString
    res
  }

  def copyWhitespace(node: ForEachLoopControl): Unit = {
    colon.copyWhitespace(node.colon)
    super.copyWhitespace(node)
  }
}
trait WhileLoop extends LoopWithContractWithin {
  val whileKeyword: StringNode = ConstantStringNode("while")
  val condition: Expression

  override def stringRepresentation: String = {
    var res = contractBefore.mkString
    res += whileKeyword.toString
    res += openParenthesis.toString
    res += condition.toString
    res += closeParenthesis.toString
    res += contractWithin.mkString
    res += body.toString
    res
  }

  def copyWhitespace(node: WhileLoop): Unit = {
    whileKeyword.copyWhitespace(node.whileKeyword)
    super.copyWhitespace(node)
  }
}
trait DoLoop extends Loop {
  val doKeyword: StringNode = ConstantStringNode("do")
  val whileKeyword: StringNode = ConstantStringNode("while")
  val condition: Expression
  val semicolon: StringNode = ConstantStringNode(";")

  override def stringRepresentation: String = {
    var res = contractBefore.mkString
    res += doKeyword.toString
    res += body.toString
    res += whileKeyword.toString
    res += openParenthesis.toString
    res += condition.toString
    res += closeParenthesis.toString
    res += semicolon.toString
    res
  }

  def copyWhitespace(node: DoLoop): Unit = {
    doKeyword.copyWhitespace(node.doKeyword)
    whileKeyword.copyWhitespace(node.whileKeyword)
    semicolon.copyWhitespace(node.semicolon)
    super.copyWhitespace(node)
  }
}
trait TryBlock extends Statement {
  val tryKeyword: StringNode = ConstantStringNode("try")
  val resourceSpecification: Option[TryResourceSpecification]
  val tryBlock: Block
  val catchClauses: List[CatchClause]
  val finallyKeyword: StringNode = ConstantStringNode("finally")
  val finallyBlock: Option[Block]

  override def stringRepresentation: String = {
    var res = tryKeyword.toString
    res += resourceSpecification.getOrElse("").toString
    res += tryBlock.toString
    res += catchClauses.mkString
    if (finallyBlock.nonEmpty) {
      res += finallyKeyword.toString
      res += finallyBlock.get.toString
    }
    res
  }

  def copyWhitespace(node: TryBlock): Unit = {
    tryKeyword.copyWhitespace(node.tryKeyword)
    finallyKeyword.copyWhitespace(node.finallyKeyword)
    super.copyWhitespace(node)
  }
}
trait CatchClause extends TreeNode {
  val catchKeyword: StringNode = ConstantStringNode("catch")
  val openParenthesis: StringNode = ConstantStringNode("(")
  var modifiers: List[Modifier] = Nil
  val types: List[TypeIndication]
  val name: StringNode
  val closeParenthesis: StringNode = ConstantStringNode(")")
  val block: Block

  override def stringRepresentation: String = {
    var res = catchKeyword.toString
    res += openParenthesis.toString
    res += modifiers.mkString
    res += types.mkString("|")
    res += name.toString
    res += closeParenthesis.toString
    res += block.toString
    res
  }

  def copyWhitespace(node: CatchClause): Unit = {
    catchKeyword.copyWhitespace(node.catchKeyword)
    openParenthesis.copyWhitespace(node.openParenthesis)
    closeParenthesis.copyWhitespace(node.closeParenthesis)
    super.copyWhitespace(node)
  }
}
trait TryResourceSpecification extends TreeNode {
  val openParenthesis: StringNode = ConstantStringNode("(")
  val resources: List[TryResource]
  val trailingSemicolon: Option[StringNode]
  val closeParenthesis: StringNode = ConstantStringNode(")")

  override def stringRepresentation: String = {
    var res = openParenthesis.toString
    res += resources.mkString(";")
    res += closeParenthesis.toString
    res
  }

  def copyWhitespace(node: TryResourceSpecification): Unit = {
    openParenthesis.copyWhitespace(node.openParenthesis)
    closeParenthesis.copyWhitespace(node.closeParenthesis)
    super.copyWhitespace(node)
  }
}
trait TryResource extends TreeNode {
  var modifiers: List[Modifier]
  val resourceType: TypeIndication
  val variable: StringNode
  val equalsSign: StringNode = ConstantStringNode("=")
  val expression: Expression

  override def stringRepresentation: String = {
    var res = modifiers.mkString
    res += resourceType.toString
    res += variable.toString
    res += equalsSign.toString
    res += expression.toString
    res
  }

  def copyWhitespace(node: TryResource): Unit = {
    equalsSign.copyWhitespace(node.equalsSign)
    super.copyWhitespace(node)
  }
}

trait SwitchStatement extends Statement {
  val switchKeyword: StringNode = ConstantStringNode("switch")
  val openParenthesis: StringNode = ConstantStringNode("(")
  val condition: Expression
  val closeParenthesis: StringNode = ConstantStringNode(")")
  val openBrace: StringNode = ConstantStringNode("{")
  val cases: List[SwitchCaseBlock]
  val closeBrace: StringNode = ConstantStringNode("}")

  override def stringRepresentation: String = {
    var res = switchKeyword.toString
    res += openParenthesis.toString
    res += condition.toString
    res += closeParenthesis.toString
    res += openBrace.toString
    res += cases.mkString
    res += closeBrace.toString
    res
  }

  def copyWhitespace(node: SwitchStatement): Unit = {
    switchKeyword.copyWhitespace(node.switchKeyword)
    openParenthesis.copyWhitespace(node.openParenthesis)
    closeParenthesis.copyWhitespace(node.closeParenthesis)
    openBrace.copyWhitespace(node.openBrace)
    closeBrace.copyWhitespace(node.closeBrace)
    super.copyWhitespace(node)
  }
}
trait SwitchCaseBlock extends TreeNode {
  val labels: List[SwitchLabel]
  val statements: List[Statement]

  override def stringRepresentation: String = labels.mkString + statements.mkString
}
trait SwitchLabel extends TreeNode {
  val caseKeyword: StringNode = ConstantStringNode("case")
  val expression: Option[Expression]
  val defaultKeyword: StringNode = ConstantStringNode("default")
  val colon: StringNode = ConstantStringNode(":")

  override def stringRepresentation: String = {
    var res = precedingWhitespace
    if (expression.nonEmpty) {
      res += caseKeyword.toString
      res += expression.get.toString
    } else {
      res += defaultKeyword.toString
    }
    res += colon.toString
    res
  }

  def copyWhitespace(node: SwitchLabel): Unit = {
    caseKeyword.copyWhitespace(node.caseKeyword)
    defaultKeyword.copyWhitespace(node.defaultKeyword)
    colon.copyWhitespace(node.colon)
    super.copyWhitespace(node)
  }
}
trait SynchronizedStatement extends Statement {
  val synchronizedKeyword: StringNode = ConstantStringNode("synchronized")
  val openParenthesis: StringNode = ConstantStringNode("(")
  val condition: Expression
  val closeParenthesis: StringNode = ConstantStringNode(")")
  val body: Statement

  override def stringRepresentation: String = {
    var res = synchronizedKeyword.toString
    res += openParenthesis.toString
    res += condition.toString
    res += closeParenthesis.toString
    res += body.toString
    res
  }

  def copyWhitespace(node: SynchronizedStatement): Unit = {
    synchronizedKeyword.copyWhitespace(node.synchronizedKeyword)
    openParenthesis.copyWhitespace(node.openParenthesis)
    closeParenthesis.copyWhitespace(node.closeParenthesis)
    super.copyWhitespace(node)
  }
}
trait ReturnStatement extends Statement {
  val returnKeyword: StringNode = ConstantStringNode("return")
  val expression: Option[Expression]
  val semicolon: StringNode = ConstantStringNode(";")

  override def stringRepresentation: String = {
    var res = returnKeyword.toString
    res += expression.getOrElse("").toString
    res += semicolon.toString
    res
  }

  def copyWhitespace(node: ReturnStatement): Unit = {
    returnKeyword.copyWhitespace(node.returnKeyword)
    semicolon.copyWhitespace(node.semicolon)
    super.copyWhitespace(node)
  }
}
trait ThrowStatement extends Statement {
  val throwKeyword: StringNode = ConstantStringNode("throw")
  val expression: Expression
  val semicolon: StringNode = ConstantStringNode(";")

  override def stringRepresentation: String = {
    var res = throwKeyword.toString
    res += expression.toString
    res += semicolon.toString
    res
  }

  def copyWhitespace(node: ThrowStatement): Unit = {
    throwKeyword.copyWhitespace(node.throwKeyword)
    semicolon.copyWhitespace(node.semicolon)
    super.copyWhitespace(node)
  }
}
trait BreakStatement extends Statement {
  val breakKeyword: StringNode = ConstantStringNode("break")
  val label: Option[StringNode]
  val semicolon: StringNode = ConstantStringNode(";")

  override def stringRepresentation: String = {
    var res = breakKeyword.toString
    res += label.getOrElse("").toString
    res += semicolon.toString
    res
  }

  def copyWhitespace(node: BreakStatement): Unit = {
    breakKeyword.copyWhitespace(node.breakKeyword)
    semicolon.copyWhitespace(node.semicolon)
    super.copyWhitespace(node)
  }
}
trait ContinueStatement extends Statement {
  val continueKeyword: StringNode = ConstantStringNode("continue")
  val label: Option[StringNode]
  val semicolon: StringNode = ConstantStringNode(";")

  override def stringRepresentation: String = {
    var res = continueKeyword.toString
    res += label.getOrElse("").toString
    res += semicolon.toString
    res
  }

  def copyWhitespace(node: ContinueStatement): Unit = {
    continueKeyword.copyWhitespace(node.continueKeyword)
    semicolon.copyWhitespace(node.semicolon)
    super.copyWhitespace(node)
  }
}
trait ExpressionStatement extends Statement {
  val expression: Expression
  val semicolon: StringNode = ConstantStringNode(";")

  override def stringRepresentation: String = expression.toString + semicolon.toString

  def copyWhitespace(node: ExpressionStatement): Unit = {
    semicolon.copyWhitespace(node.semicolon)
    super.copyWhitespace(node)
  }
}
trait LabeledStatement extends Statement {
  val label: Label
  val contract: Option[Contract]
  val statement: Statement

  override def stringRepresentation: String = label.toString + statement.toString
}
trait Label extends TreeNode with LoopSpecOrLabel {
  val label: StringNode
  val colon: StringNode = ConstantStringNode(":")

  override def stringRepresentation: String = label.toString + colon.toString

  def copyWhitespace(node: Label): Unit = {
    colon.copyWhitespace(node.colon)
    super.copyWhitespace(node)
  }
}

trait Arguments extends TreeNode {
  val openParenthesis: StringNode = ConstantStringNode("(")
  val arguments: List[Expression]
  val closeParenthesis: StringNode = ConstantStringNode(")")

  override def stringRepresentation: String = {
    var res = openParenthesis.toString
    res += arguments.mkString(",")
    res += closeParenthesis.toString
    res
  }

  def copyWhitespace(node: Arguments): Unit = {
    openParenthesis.copyWhitespace(node.openParenthesis)
    closeParenthesis.copyWhitespace(node.closeParenthesis)
    super.copyWhitespace(node)
  }
}

