package generic.syntaxTree
import generic.{Helper, SourceLocation}

trait JmlModifierBlock extends Modifier {
  var startSpec: Option[StringNode]
  val modifiers: List[Modifier]
  var endSpec: Option[StringNode]

  override def stringRepresentation: String = {
    var res = startSpec.getOrElse("").toString
    res += modifiers.mkString
    res += endSpec.getOrElse("").toString
    res
  }
}

trait SpecTypeIndication extends TypeIndication
trait IntegerSpecTypeIndication extends SpecTypeIndication
trait RealSpecTypeIndication extends SpecTypeIndication

trait PureModifier extends StringNode with Modifier {
  var text: String = "pure"
}
trait NullableModifier extends StringNode with Modifier {
  var text: String = "nullable"
}
trait NullableByDefaultModifier extends StringNode with Modifier {
  var text: String = "nullable_by_default"
}
trait NonnullModifier extends StringNode with Modifier {
  var text: String = "non_null"
}
trait GhostModifier extends StringNode with Modifier {
  var text: String = "ghost"
}

trait SpecMethodModifier extends Modifier

trait HelperModifier extends StringNode with Modifier {
  var text: String = "helper"
}
trait MonitoredModifier extends StringNode with Modifier {
  var text: String = "monitored"
}
trait ModelModifier extends StringNode with Modifier {
  var text: String = "model"
}
trait InstanceModifier extends StringNode with Modifier {
  var text: String = "instance"
}

trait Contract extends TreeNode {
  var startSpec: Option[StringNode]
  var clauses: List[ContractClause]
  var endSpec: Option[StringNode]

  override def stringRepresentation: String =  {
    var res = startSpec.getOrElse("").toString
    res += clauses.mkString
    res += endSpec.getOrElse("").toString
    res
  }

  def moveSpecDelimToContract(): Unit = {
    if(startSpec.isEmpty && endSpec.isEmpty && clauses.nonEmpty) {
      val first = clauses.head
      val last = clauses.last
      if(first.startSpec.isDefined && first.startSpec.get.text.contains("/*")&& last.endSpec.isDefined && last.endSpec.get.text.contains("*/")) {
        precedingWhitespace += first.precedingWhitespace
        first.precedingWhitespace = ""
        startSpec = first.startSpec
        first.startSpec = None
        trailingWhitespace = last.trailingWhitespace + trailingWhitespace
        last.trailingWhitespace = ""
        endSpec = last.endSpec
        last.endSpec = None
      }
    }
  }

  /** Add a clause at the start of the contract, with startSpec handled properly
   *  Generic in "T", the language-specific sub-type of TreeNode
   *  @param clause: the clause to be added, of language type T
   *  @param wrapInSpec: whether to wrap the clause in spec, or not (e.g. in case the whole method is ghost)
   *  @param makeStringNode: constructor to create a StringNode that is of language type T, taking text, srcLocation, precedingWhitespace and trailingWhitespace
   */
  def addClauseBeforeContract[T <: TreeNode](clause: ContractClause with T, wrapInSpec: Boolean, makeStringNode: (String, SourceLocation, String, String) => StringNode with T): Unit = {
    moveSpecDelimToContract()
    val newWrapInSpec = if (wrapInSpec && this.startSpec.isDefined) false else wrapInSpec
    val specs = clauses.filter(x => x.isInstanceOf[SpecCase])
    if(specs.nonEmpty) {
      specs.map(x=>x.asInstanceOf[SpecCase]).foreach(x => x.addClauseBeforeSpec[T](clause, newWrapInSpec, makeStringNode))
    }
    if(specs != clauses || clauses.isEmpty) {
      if (startSpec.isDefined && startSpec.get.text.contains("/*")) {
        if (clauses.isEmpty) {
          clauses = List(clause)
        } else {
          if(!clauses.head.precedingWhitespace.contains(System.lineSeparator())){
            clauses.head.precedingWhitespace = System.lineSeparator() + "      " + clauses.head.precedingWhitespace
          }
          clauses = clauses.prepended(clause)
        }
      } else {
        if (clauses.isEmpty) {
          if (newWrapInSpec) {
            clause.startSpec = Some(makeStringNode("/*@", clause.srcLocation, System.lineSeparator() + "      ", " "))
            clause.endSpec = Some(makeStringNode("*/", clause.srcLocation, " ", ""))
          }
          clauses = List(clause)
        } else {
          val first = clauses.head
          val newlines = Helper.getNewlines(first)
          if(newlines.isEmpty) {
            first.precedingWhitespace = System.lineSeparator() + "      " + first.precedingWhitespace
          } else {
            clause.precedingWhitespace = newlines + clause.precedingWhitespace
          }
          if (newWrapInSpec) {
            clause.startSpec = Some(makeStringNode("/*@", clause.srcLocation, "", " "))
            if(first.startSpec.isDefined) {
              clause.startSpec.get.precedingWhitespace = Helper.getNewlines(first.startSpec.get)
            } else if(clause.precedingWhitespace.isEmpty) {
              clause.startSpec.get.precedingWhitespace = System.lineSeparator() + "      "
            }
            clause.endSpec = Some(makeStringNode("*/", clause.srcLocation, " ", ""))
          }
          clauses = clauses.prepended(clause)
        }
      }
    }
  }
  /** Add a clause at the end of the contract, with startSpec handled properly
   *  Generic in "T", the language-specific sub-type of TreeNode
   *  @param clause: the clause to be added, of language type T
   *  @param wrapInSpec: whether to wrap the clause in spec, or not (e.g. in case the whole method is ghost)
   *  @param makeStringNode: constructor to create a StringNode that is of language type T, taking text, srcLocation, precedingWhitespace and trailingWhitespace
   */
  def addClauseAfterContract[T <: TreeNode](clause: ContractClause with T, wrapInSpec: Boolean, makeStringNode: (String, SourceLocation, String, String) => StringNode with T): Unit = {
    moveSpecDelimToContract()

    val specs = clauses.filter(x => x.isInstanceOf[SpecCase])
    if(specs.nonEmpty) {
      specs.map(x=>x.asInstanceOf[SpecCase]).foreach(x => x.addClauseBeforeSpec[T](clause, wrapInSpec, makeStringNode))
    }
    if(specs != clauses || clauses.isEmpty) {
      if (endSpec.isDefined) {
        if (clauses.isEmpty) {
          clauses = List(clause)
        } else {
          clause.precedingWhitespace = System.lineSeparator() + "      " + clause.precedingWhitespace
          clauses = clauses.appended(clause)
        }
      } else {
        if (clauses.isEmpty) {
          if (wrapInSpec) {
            clause.startSpec = Some(makeStringNode("/*@", clause.srcLocation, System.lineSeparator() + "      ", " "))
            clause.endSpec = Some(makeStringNode("*/", clause.srcLocation, " ", ""))
          }
          clauses = List(clause)
        } else {
          val last = clauses.last
          clause.precedingWhitespace = System.lineSeparator() + "      " + clause.precedingWhitespace
          if(last.startSpec.isDefined) {
            clause.startSpec = last.startSpec
          }
          if(last.endSpec.isDefined) {
            clause.endSpec = last.endSpec
          }
          clauses = clauses.appended(clause)
        }
      }
    }
  }
}

trait SpecCase extends ContractClause {
  var hasAlso: Boolean
  val alsoKeyword: StringNode = ConstantStringNode("also")
  var clauses: List[ContractClause]

  override def stringRepresentation: String = {
    var res = ""
    if (hasAlso) {
      res += alsoKeyword.toString
    }
    res += clauses.mkString
    wrappedStringRepresentation(res)
  }

  def copyWhitespace(node: SpecCase): Unit = {
    alsoKeyword.copyWhitespace(node.alsoKeyword)
    super.copyWhitespace(node)
  }

  /** Add a clause at the start of the SpecCase, with startSpec handled properly
   *  Generic in "T", the language-specific sub-type of TreeNode
   *  @param clause: the clause to be added, of language type T
   *  @param wrapInSpec: whether to wrap the clause in spec, or not (e.g. in case the whole method is ghost)
   *  @param makeStringNode: constructor to create a StringNode that is of language type T, taking text, srcLocation, precedingWhitespace and trailingWhitespace
   */
  def addClauseBeforeSpec[T <: TreeNode](clause: ContractClause with T, wrapInSpec: Boolean, makeStringNode: (String, SourceLocation, String, String) => StringNode with T): Unit = {
    if(startSpec.isDefined) {
      if (clauses.isEmpty) {
        clauses = List(clause)
      } else {
        clauses.head.precedingWhitespace = System.lineSeparator() + "      " + clauses.head.precedingWhitespace
        clauses = clauses.prepended(clause)
      }
    } else {
      if(clauses.isEmpty) {
        if(wrapInSpec) {
          clause.startSpec = Some(makeStringNode("/*@", clause.srcLocation, System.lineSeparator() + "      ", " "))
          clause.endSpec = Some(makeStringNode("*/", clause.srcLocation, " ", ""))
        }
        clauses = List(clause)
      } else {
        val first = clauses.head
        if(first.precedingWhitespace.nonEmpty){
          clause.precedingWhitespace = first.precedingWhitespace + clause.precedingWhitespace
          first.precedingWhitespace = System.lineSeparator() + "      "
        }
        if(wrapInSpec) {
          clause.startSpec = Some(makeStringNode("/*@", clause.srcLocation, System.lineSeparator() + "      ", " "))
          if(first.startSpec.isDefined) {
            val prec = first.startSpec.get.precedingWhitespace
            first.startSpec.get.precedingWhitespace = clause.startSpec.get.precedingWhitespace
            clause.startSpec.get.precedingWhitespace = prec
          }
          clause.endSpec = Some(makeStringNode("*/", clause.srcLocation, " ", ""))
        }
        clauses = clauses.prepended(clause)
      }
    }
  }
  /** Add a clause at the end of the SpecCase, with startSpec handled properly
   *  Generic in "T", the language-specific sub-type of TreeNode
   *  @param clause: the clause to be added, of language type T
   *  @param wrapInSpec: whether to wrap the clause in spec, or not (e.g. in case the whole method is ghost)
   *  @param makeStringNode: constructor to create a StringNode that is of language type T, taking text, srcLocation, precedingWhitespace and trailingWhitespace
   */
  def addClauseAfterSpec[T <: TreeNode](clause: ContractClause with T, wrapInSpec: Boolean, makeStringNode: (String, SourceLocation, String, String) => StringNode with T): Unit = {
    if(endSpec.isDefined) {
      if (clauses.isEmpty) {
        clauses = List(clause)
      } else {
        clause.precedingWhitespace = System.lineSeparator() + "      " + clause.precedingWhitespace
        clauses = clauses.appended(clause)
      }
    } else {
      if(clauses.isEmpty) {
        if(wrapInSpec) {
          clause.startSpec = Some(makeStringNode("/*@", clause.srcLocation, System.lineSeparator() + "      ", " "))
          clause.endSpec = Some(makeStringNode("*/", clause.srcLocation, " ", ""))
        }
        clauses = List(clause)
      } else {
        val last = clauses.last
        clause.precedingWhitespace = System.lineSeparator() + "      " + clause.precedingWhitespace
        if(last.startSpec.isDefined) {
          clause.startSpec = last.startSpec
        }
        if(last.endSpec.isDefined) {
          clause.endSpec = last.endSpec
        }
        clauses = clauses.appended(clause)
      }
    }
  }
}

trait LoopSpecification extends LoopSpecOrLabel {
  val startSpec: Option[StringNode]
  val clauses: List[LoopClause]
  val endSpec: Option[StringNode]

  override def stringRepresentation: String = {
    startSpec.getOrElse("").toString + clauses.mkString + endSpec.getOrElse("").toString
  }
}

trait LoopSpecOrLabel extends TreeNode

trait ContractClause extends TreeNode {
  var startSpec: Option[StringNode]
  var endSpec: Option[StringNode]

  def wrappedStringRepresentation(contained: String): String = {
    startSpec.getOrElse("").toString + contained + endSpec.getOrElse("").toString
  }
}

trait LoopClause extends TreeNode

trait NonTranslatedContractClause extends ContractClause with StringNode {
  // a NonTranslatedContractClause should be a comment, and thus doesn't need start/endSpec
  this.startSpec = None
  this.endSpec = None

  override def stringRepresentation: String = {
    wrappedStringRepresentation(text)
  }
}
trait NonTranslatedLoopClause extends LoopClause with StringNode

trait AssignableClause extends ContractClause {
  val assignableKeyword: StringNode = ConstantStringNode("assignable")
  val expressions: List[Expression]
  val semicolon: StringNode = ConstantStringNode(";")

  override def stringRepresentation: String = {
    wrappedStringRepresentation(assignableKeyword.toString
                                + expressions.mkString(",")
                                + semicolon.toString)
  }

  def copyWhitespace(node: AssignableClause): Unit = {
    assignableKeyword.copyWhitespace(node.assignableKeyword)
    semicolon.copyWhitespace(node.semicolon)
    super.copyWhitespace(node)
  }
}

trait AccessibleClause extends ContractClause {
  val accessibleKeyword: StringNode = ConstantStringNode("accessible")
  val expressions: List[Expression]
  val semicolon: StringNode = ConstantStringNode(";")

  override def stringRepresentation: String = {
    wrappedStringRepresentation(accessibleKeyword.toString
                                + expressions.mkString(",")
                                + semicolon.toString)
  }

  def copyWhitespace(node: AccessibleClause): Unit = {
    accessibleKeyword.copyWhitespace(node.accessibleKeyword)
    semicolon.copyWhitespace(node.semicolon)
    super.copyWhitespace(node)
  }
}

trait ForallVariableDeclaration extends ContractClause {
  val forallKeyword: StringNode = ConstantStringNode("forall")
  val declaration: QuantifiedVarDeclaration
  val semicolon: StringNode = ConstantStringNode(";")

  override def stringRepresentation: String = {
    wrappedStringRepresentation(forallKeyword.toString + declaration.toString + semicolon.toString)
  }

  def copyWhitespace(node: ForallVariableDeclaration): Unit = {
    forallKeyword.copyWhitespace(node.forallKeyword)
    semicolon.copyWhitespace(node.semicolon)
    super.copyWhitespace(node)
  }
}

trait OldVariableDeclaration extends ContractClause {
  val oldKeyword: StringNode = ConstantStringNode("old")
  val modifiers: List[Modifier]
  val typeIndication: TypeIndication
  val declarators: List[VariableDeclarator]
  val semicolon: StringNode = ConstantStringNode(";")

  override def stringRepresentation: String = {
    var res = oldKeyword.toString
    res += typeIndication.toString
    res += modifiers.mkString
    res += declarators.mkString(",")
    res += semicolon.toString
    wrappedStringRepresentation(res)
  }

  def copyWhitespace(node: OldVariableDeclaration): Unit = {
    oldKeyword.copyWhitespace(node.oldKeyword)
    semicolon.copyWhitespace(node.semicolon)
    super.copyWhitespace(node)
  }
}

trait Precondition extends ContractClause {
  val requiresKeyword: StringNode = ConstantStringNode("requires")
  val expression: Expression
  val semicolon: StringNode = ConstantStringNode(";")

  override def stringRepresentation: String = {
    wrappedStringRepresentation(requiresKeyword.toString + expression.toString + semicolon.toString)
  }

  def copyWhitespace(node: Precondition): Unit = {
    requiresKeyword.copyWhitespace(node.requiresKeyword)
    semicolon.copyWhitespace(node.semicolon)
    super.copyWhitespace(node)
  }
}

trait Postcondition extends ContractClause {
  val ensuresKeyword: StringNode = ConstantStringNode("ensures")
  val expression: Expression
  val semicolon: StringNode = ConstantStringNode(";")

  override def stringRepresentation: String = {
    wrappedStringRepresentation(ensuresKeyword.toString + expression.toString + semicolon.toString)
  }

  def copyWhitespace(node: Postcondition): Unit = {
    ensuresKeyword.copyWhitespace(node.ensuresKeyword)
    semicolon.copyWhitespace(node.semicolon)
    super.copyWhitespace(node)
  }
}

trait LoopInvariant extends LoopClause {
  val invariantKeyword: StringNode = ConstantStringNode("loop_invariant")
  val expression: Expression
  val semicolon: StringNode = ConstantStringNode(";")

  override def stringRepresentation: String = {
    invariantKeyword.toString + expression.toString + semicolon.toString
  }

  def copyWhitespace(node: LoopInvariant): Unit = {
    invariantKeyword.copyWhitespace(node.invariantKeyword)
    semicolon.copyWhitespace(node.semicolon)
    super.copyWhitespace(node)
  }
}

trait LoopModifiesClause extends LoopClause {
  val loopModifiesKeyword: StringNode = ConstantStringNode("loop_modifies")
  val expressions: List[Expression]
  val semicolon: StringNode = ConstantStringNode(";")

  override def stringRepresentation: String = {
    loopModifiesKeyword.toString + expressions.mkString(",") + semicolon.toString
  }

  def copyWhitespace(node: LoopModifiesClause): Unit = {
    loopModifiesKeyword.copyWhitespace(node.loopModifiesKeyword)
    semicolon.copyWhitespace(node.semicolon)
    super.copyWhitespace(node)
  }
}

trait DivergeClause extends ContractClause {
  val divergesKeyword: StringNode = ConstantStringNode("diverges")
  val expression: Expression
  val semicolon: StringNode = ConstantStringNode(";")

  override def stringRepresentation: String = {
    wrappedStringRepresentation(divergesKeyword.toString + expression.toString + semicolon.toString)
  }

  def copyWhitespace(node: DivergeClause): Unit = {
    divergesKeyword.copyWhitespace(node.divergesKeyword)
    semicolon.copyWhitespace(node.semicolon)
    super.copyWhitespace(node)
  }
}

trait DecreasesClause extends LoopClause {
  val decreasesKeyword: StringNode = ConstantStringNode("decreases")
  val expression: Expression
  val semicolon: StringNode = ConstantStringNode(";")

  override def stringRepresentation: String = {
    decreasesKeyword.toString + expression.toString + semicolon.toString
  }

  def copyWhitespace(node: DecreasesClause): Unit = {
    decreasesKeyword.copyWhitespace(node.decreasesKeyword)
    semicolon.copyWhitespace(node.semicolon)
    super.copyWhitespace(node)
  }
}

trait ExceptionPostcondition extends ContractClause {
  val signalsKeyword: StringNode = ConstantStringNode("signals")
  val openParenthesis: StringNode = ConstantStringNode("(")
  val exceptionType: TypeIndication
  val exceptionName: Option[StringNode]
  val closeParenthesis: StringNode = ConstantStringNode(")")
  val expression: Option[Expression]
  val semicolon: StringNode = ConstantStringNode(";")

  override def stringRepresentation: String = {
    var res = signalsKeyword.toString
    res += openParenthesis.toString
    res += exceptionType.toString
    res += exceptionName.getOrElse("").toString
    res += closeParenthesis.toString
    res += expression.getOrElse("").toString
    res += semicolon.toString
    wrappedStringRepresentation(res)
  }

  def copyWhitespace(node: ExceptionPostcondition): Unit = {
    signalsKeyword.copyWhitespace(node.signalsKeyword)
    openParenthesis.copyWhitespace(node.openParenthesis)
    closeParenthesis.copyWhitespace(node.closeParenthesis)
    semicolon.copyWhitespace(node.semicolon)
    super.copyWhitespace(node)
  }
}

trait SignalsOnlyClause extends ContractClause {
  val signalsOnlyKeyword: StringNode = ConstantStringNode("signals_only")
  val exceptionTypes: List[TypeIndication]
  val nothingKeyword: StringNode = ConstantStringNode("\\nothing")
  val semicolon: StringNode = ConstantStringNode(";")

  override def stringRepresentation: String = {
    var res = signalsOnlyKeyword.toString
    if (exceptionTypes.nonEmpty) {
      res += exceptionTypes.mkString(",")
    } else {
      res += nothingKeyword.toString
    }
    res += semicolon.toString
    wrappedStringRepresentation(res)
  }

  def copyWhitespace(node: SignalsOnlyClause): Unit = {
    signalsOnlyKeyword.copyWhitespace(node.signalsOnlyKeyword)
    nothingKeyword.copyWhitespace(node.nothingKeyword)
    semicolon.copyWhitespace(node.semicolon)
    super.copyWhitespace(node)
  }
}

trait MeasuredByClause extends ContractClause {
  val measuredByKeyword: StringNode = ConstantStringNode("measured_by")
  val expression: Expression
  val ifKeyword: StringNode = ConstantStringNode("if")
  val predicate: Option[Expression]
  val semicolon: StringNode = ConstantStringNode(";")

  override def stringRepresentation: String = {
    var res = measuredByKeyword.toString
    res += expression.toString
    if(predicate.isDefined) {
      res += ifKeyword.toString
      res += predicate.toString
    }
    res += semicolon.toString
    wrappedStringRepresentation(res)
  }

  def copyWhitespace(node: MeasuredByClause): Unit = {
    measuredByKeyword.copyWhitespace(node.measuredByKeyword)
    ifKeyword.copyWhitespace(node.ifKeyword)
    semicolon.copyWhitespace(node.semicolon)
    super.copyWhitespace(node)
  }
}

trait GhostIO extends ContractClause {
  val argType: TypeIndication
  val name: StringNode
  val semicolon: StringNode = ConstantStringNode(";")
  
  def copyWhitespace(node: GhostIO): Unit = {
    semicolon.copyWhitespace(node.semicolon)
    super.copyWhitespace(node)
  }

}
trait GhostArgument extends GhostIO {
  val givenKeyword: StringNode = ConstantStringNode("given")

  override def stringRepresentation: String = {
    wrappedStringRepresentation(givenKeyword.toString
                                + argType.toString
                                + name.toString
                                + semicolon.toString)
  }

  def copyWhitespace(node: GhostArgument): Unit = {
    givenKeyword.copyWhitespace(node.givenKeyword)
    super.copyWhitespace(node)
  }
}

trait GhostReturn extends GhostIO {
  val yieldsKeyword: StringNode = ConstantStringNode("yields")

  override def stringRepresentation: String = {
    wrappedStringRepresentation(yieldsKeyword.toString
                                + argType.toString
                                + name.toString
                                + semicolon.toString)
  }

  def copyWhitespace(node: GhostReturn): Unit = {
    yieldsKeyword.copyWhitespace(node.yieldsKeyword)
    super.copyWhitespace(node)
  }
}


trait Behavior extends SpecCase {
  var privacy: Option[StringNode]
  val behaviorKeyword: StringNode = ConstantStringNode("behavior")

  override def stringRepresentation: String = {
    var res = startSpec.getOrElse("").toString
    if (hasAlso) {
      res += alsoKeyword.toString
    }
    res += privacy.getOrElse("").toString
    res += behaviorKeyword.toString
    res += clauses.mkString
    res += endSpec.getOrElse("").toString
    res
  }

  def copyWhitespace(node: Behavior): Unit = {
    behaviorKeyword.copyWhitespace(node.behaviorKeyword)
    super.copyWhitespace(node)
  }
}
trait NormalBehavior extends Behavior {
  override val behaviorKeyword: StringNode = ConstantStringNode("normal_behavior")
}
trait ExceptionalBehavior extends Behavior {
  override val behaviorKeyword: StringNode = ConstantStringNode("exceptional_behavior")
}

trait AssertStatement extends Statement {
  val assertKeyword: StringNode = ConstantStringNode("assert")
  val content: Expression
  val colon: StringNode = ConstantStringNode(":")
  val message: Option[Expression]
  val semicolon: StringNode = ConstantStringNode(";")

  override def stringRepresentation: String = {
    var res = assertKeyword.toString
    res += content.toString
    if (message.nonEmpty) {
      res += colon.toString
      res += message.get.toString
    }
    res += semicolon.toString
    res
  }

  def copyWhitespace(node: AssertStatement): Unit = {
    assertKeyword.copyWhitespace(node.assertKeyword)
    colon.copyWhitespace(node.colon)
    semicolon.copyWhitespace(node.semicolon)
    super.copyWhitespace(node)
  }
}
trait AssumeStatement extends Statement {
  val assumeKeyword: StringNode = ConstantStringNode("assume")
  val content: Expression
  val colon: StringNode = ConstantStringNode(":")
  val message: Option[Expression]
  val semicolon: StringNode = ConstantStringNode(";")

  override def stringRepresentation: String = {
    var res = assumeKeyword.toString
    res += content.toString
    if (message.nonEmpty) {
      res += colon.toString
      res += message.get.toString
    }
    res += semicolon.toString
    res
  }

  def copyWhitespace(node: AssumeStatement): Unit = {
    assumeKeyword.copyWhitespace(node.assumeKeyword)
    colon.copyWhitespace(node.colon)
    semicolon.copyWhitespace(node.semicolon)
    super.copyWhitespace(node)
  }
}
trait SetStatement extends Statement {
  val setKeyword: StringNode = ConstantStringNode("set")
  val assignment: Expression
  val semicolon: StringNode = ConstantStringNode(";")

  override def stringRepresentation: String = {
    var res = setKeyword.toString
    res += assignment.toString
    res += semicolon.toString
    res
  }

  def copyWhitespace(node: SetStatement): Unit = {
    setKeyword.copyWhitespace(node.setKeyword)
    semicolon.copyWhitespace(node.semicolon)
    super.copyWhitespace(node)
  }
}
trait UnreachableStatement extends Statement {
  val unreachableKeyword: StringNode = ConstantStringNode("unreachable")

  override def stringRepresentation: String = unreachableKeyword.toString

  def copyWhitespace(node: UnreachableStatement): Unit = {
    unreachableKeyword.copyWhitespace(node.unreachableKeyword)
    super.copyWhitespace(node)
  }
}

trait FoldStatement extends Statement {
  val foldKeyword: StringNode = ConstantStringNode("fold")
  val expression: Expression
  val semicolon: StringNode = ConstantStringNode(";")

  override def stringRepresentation: String = {
    foldKeyword.toString + expression.toString + semicolon.toString
  }

  def copyWhitespace(node: FoldStatement): Unit = {
    foldKeyword.copyWhitespace(node.foldKeyword)
    semicolon.copyWhitespace(node.semicolon)
    super.copyWhitespace(node)
  }
}

trait UnfoldStatement extends Statement {
  val unfoldKeyword: StringNode = ConstantStringNode("unfold")
  val expression: Expression
  val semicolon: StringNode = ConstantStringNode(";")

  override def stringRepresentation: String = {
    unfoldKeyword.toString + expression.toString + semicolon.toString
  }

  def copyWhitespace(node: UnfoldStatement): Unit = {
    unfoldKeyword.copyWhitespace(node.unfoldKeyword)
    semicolon.copyWhitespace(node.semicolon)
    super.copyWhitespace(node)
  }
}

trait GhostStatement extends Statement {
  var ghostKeyword: GhostModifier
  val statement: Statement

  override def stringRepresentation: String = {
    ghostKeyword.toString + statement.toString
  }

  def copyWhitespace(node: GhostStatement): Unit = {
    ghostKeyword.copyWhitespace(node.ghostKeyword)
    super.copyWhitespace(node)
  }
}

trait JmlStatementBlock extends Statement {
  var startSpec: Option[StringNode]
  val statements: List[Statement]
  var endSpec: Option[StringNode]

  override def stringRepresentation: String = {
    var res = startSpec.getOrElse("").toString
    res += statements.mkString
    res += endSpec.getOrElse("").toString
    res
  }
}

trait ResultExpression extends StringNode with Expression {
  private var _text: String = "\\result"
  def text: String = _text
  def text_=(t: String): Unit = {
    _text = t
  }
}

trait OldExpression extends Expression {
  val oldKeyword: StringNode = ConstantStringNode("\\old")
  val openParenthesis: StringNode = ConstantStringNode("(")
  val expression: Expression
  val comma: StringNode = ConstantStringNode(",")
  val label: Option[StringNode]
  val closeParenthesis: StringNode = ConstantStringNode(")")

  override def stringRepresentation: String = {
    var res = oldKeyword.toString
    res += openParenthesis.toString
    res += expression.toString
    if (label.nonEmpty) {
      res += comma.toString
      res += label.get.toString
    }
    res += closeParenthesis.toString
    res
  }

  def copyWhitespace(node: OldExpression): Unit = {
    oldKeyword.copyWhitespace(node.oldKeyword)
    openParenthesis.copyWhitespace(node.openParenthesis)
    comma.copyWhitespace(node.comma)
    closeParenthesis.copyWhitespace(node.closeParenthesis)
    super.copyWhitespace(node)
  }
}

trait ExpressionPredicate extends Expression {
  val keyword: StringNode
  val openParenthesis: StringNode = ConstantStringNode("(")
  val expression: Expression
  val closeParenthesis: StringNode = ConstantStringNode(")")

  override def stringRepresentation: String = {
    var res = keyword.toString
    res += openParenthesis.toString
    res += expression.toString
    res += closeParenthesis.toString
    res
  }

  def copyWhitespace(node: ExpressionPredicate): Unit = {
    keyword.copyWhitespace(node.keyword)
    openParenthesis.copyWhitespace(node.openParenthesis)
    closeParenthesis.copyWhitespace(node.closeParenthesis)
    super.copyWhitespace(node)
  }
}

trait PreExpression extends ExpressionPredicate {
  override val keyword = ConstantStringNode("\\pre")
}

trait StoreRefPredicate extends Expression {
  val keyword: StringNode
  val openParenthesis: StringNode = ConstantStringNode("(")
  val storeRefs: List[Expression]
  val closeParenthesis: StringNode = ConstantStringNode(")")

  override def stringRepresentation: String = {
    var res = keyword.toString
    res += openParenthesis.toString
    res += storeRefs.mkString(",")
    res += closeParenthesis.toString
    res
  }

  def copyWhitespace(node: StoreRefPredicate): Unit = {
    keyword.copyWhitespace(node.keyword)
    openParenthesis.copyWhitespace(node.openParenthesis)
    closeParenthesis.copyWhitespace(node.closeParenthesis)
    super.copyWhitespace(node)
  }
}

trait NotAssignedExpression extends StoreRefPredicate {
  override val keyword = ConstantStringNode("\\not_assigned")
}

trait NotModifiedExpression extends StoreRefPredicate {
  override val keyword = ConstantStringNode("\\not_modified")
}

trait OnlyAccessedExpression extends StoreRefPredicate {
  override val keyword = ConstantStringNode("\\only_accessed")
}

trait OnlyAssignedExpression extends StoreRefPredicate {
  override val keyword = ConstantStringNode("\\only_assigned")
}

trait FreshExpression extends Expression {
  val freshKeyword: StringNode = ConstantStringNode("\\fresh")
  val openParenthesis: StringNode = ConstantStringNode("(")
  val expressions: List[Expression]
  val closeParenthesis: StringNode = ConstantStringNode(")")

  override def stringRepresentation: String = {
    var res = freshKeyword.toString
    res += openParenthesis.toString
    res += expressions.mkString(",")
    res += closeParenthesis.toString
    res
  }

  def copyWhitespace(node: FreshExpression): Unit = {
    freshKeyword.copyWhitespace(node.freshKeyword)
    openParenthesis.copyWhitespace(node.openParenthesis)
    closeParenthesis.copyWhitespace(node.closeParenthesis)
    super.copyWhitespace(node)
  }
}

trait NonNullElementsExpression extends ExpressionPredicate{
  override val keyword: StringNode = ConstantStringNode("\\nonnullelements")
}

trait TypeOfExpression extends ExpressionPredicate{
  override val keyword: StringNode = ConstantStringNode("\\typeof")
}

trait TypeExpression extends Expression {
  val typeKeyword: StringNode = ConstantStringNode("\\type")
  val openParenthesis: StringNode = ConstantStringNode("(")
  val typeIndication: TypeIndication
  val closeParenthesis: StringNode = ConstantStringNode(")")

  override def stringRepresentation: String = {
    var res = typeKeyword.toString
    res += openParenthesis.toString
    res += typeIndication.toString
    res += closeParenthesis.toString
    res
  }

  def copyWhitespace(node: ExpressionPredicate): Unit = {
    typeKeyword.copyWhitespace(node.keyword)
    openParenthesis.copyWhitespace(node.openParenthesis)
    closeParenthesis.copyWhitespace(node.closeParenthesis)
    super.copyWhitespace(node)
  }
}

trait QuantifiedExpression extends Expression {
  val openParenthesis: StringNode = ConstantStringNode("(")
  val quantifier: StringNode
  val declaration: List[QuantifiedVarDeclaration]
  val firstSemicolon: StringNode = ConstantStringNode(";")
  val guard: Option[Expression]
  val secondSemicolon: StringNode = ConstantStringNode(";")
  val expression: Expression
  val closeParenthesis: StringNode = ConstantStringNode(")")

  override def stringRepresentation: String = {
    var res = openParenthesis.toString
    res += quantifier.toString
    res += declaration.mkString(",")
    res += firstSemicolon.toString
    if (guard.nonEmpty) {
      res += guard.get.toString
      res += secondSemicolon.toString
    }
    res += expression.toString
    res += closeParenthesis.toString
    res
  }

  def copyWhitespace(node: QuantifiedExpression): Unit = {
    openParenthesis.copyWhitespace(node.openParenthesis)
    quantifier.copyWhitespace(node.quantifier)
    firstSemicolon.copyWhitespace(node.firstSemicolon)
    secondSemicolon.copyWhitespace(node.secondSemicolon)
    closeParenthesis.copyWhitespace(node.closeParenthesis)
    super.copyWhitespace(node)
  }
}

trait ForallExpression extends QuantifiedExpression {
  override val quantifier: StringNode = ConstantStringNode("\\forall")
}
trait ExistsExpression extends QuantifiedExpression {
  override val quantifier: StringNode = ConstantStringNode("\\exists")
}
trait MaxQuantifiedExpression extends QuantifiedExpression {
  override val quantifier: StringNode = ConstantStringNode("\\max")
}
trait MinExpression extends QuantifiedExpression {
  override val quantifier: StringNode = ConstantStringNode("\\min")
}
trait NumOfExpression extends QuantifiedExpression {
  override val quantifier: StringNode = ConstantStringNode("\\num_of")
}
trait ProductExpression extends QuantifiedExpression {
  override val quantifier: StringNode = ConstantStringNode("\\product")
}
trait SumExpression extends QuantifiedExpression {
  override val quantifier: StringNode = ConstantStringNode("\\sum")
}

trait LetExpression extends Expression {
  val openParenthesis: StringNode = ConstantStringNode("(")
  val letKeyword: StringNode= ConstantStringNode("\\let")
  val declType: TypeIndication
  val declaration: VariableDeclarator
  val semicolon: StringNode = ConstantStringNode(";")
  val expression: Expression
  val closeParenthesis: StringNode = ConstantStringNode(")")

  override def stringRepresentation: String = {
    var res = openParenthesis.toString
    res += letKeyword.toString
    res += declType.toString
    res += declaration.toString
    res += semicolon.toString
    res += expression.toString
    res += closeParenthesis.toString
    res
  }

  def copyWhitespace(node: QuantifiedExpression): Unit = {
    openParenthesis.copyWhitespace(node.openParenthesis)
    letKeyword.copyWhitespace(node.quantifier)
    semicolon.copyWhitespace(node.firstSemicolon)
    closeParenthesis.copyWhitespace(node.closeParenthesis)
    super.copyWhitespace(node)
  }
}

trait QuantifiedVarDeclaration extends TreeNode {
  val declType: TypeIndication
  val declarators: List[StringNode]

  override def stringRepresentation: String = {
    var res = declType.toString
    res += declarators.mkString(",")
    res
  }
}
trait NonNullQuantifiedVarDeclaration extends QuantifiedVarDeclaration {
  val nonNullKeyword: StringNode = ConstantStringNode("non_null")

  override def stringRepresentation: String = nonNullKeyword.toString + super.stringRepresentation

  def copyWhitespace(node: NonNullQuantifiedVarDeclaration): Unit = {
    nonNullKeyword.copyWhitespace(node.nonNullKeyword)
    super.copyWhitespace(node)
  }
}
trait NullableQuantifiedVarDeclaration extends QuantifiedVarDeclaration {
  val nullableKeyword: StringNode = ConstantStringNode("nullable")

  override def stringRepresentation: String = nullableKeyword.toString + super.stringRepresentation

  def copyWhitespace(node: NullableQuantifiedVarDeclaration): Unit = {
    nullableKeyword.copyWhitespace(node.nullableKeyword)
    super.copyWhitespace(node)
  }
}

trait SetComprehension extends NewExpression {
  val openBrace: StringNode = ConstantStringNode("{")
  val declaration: QuantifiedVarDeclaration
  val bar: StringNode = ConstantStringNode("|")
  val containsExpression: Expression
  val and: StringNode = ConstantStringNode("&&")
  val filterExpression: Expression
  val closeBrace: StringNode = ConstantStringNode("}")

  override def stringRepresentation: String = {
    var res = openBrace.toString
    res += declaration.toString
    res += bar.toString
    res += containsExpression.toString
    res += and.toString
    res += filterExpression.toString
    res += closeBrace.toString
    res
  }

  def copyWhitespace(node: SetComprehension): Unit = {
    openBrace.copyWhitespace(node.openBrace)
    bar.copyWhitespace(node.bar)
    and.copyWhitespace(node.and)
    closeBrace.copyWhitespace(node.closeBrace)
    super.copyWhitespace(node)
  }
}

trait CardinalityExpression extends Expression {
  val leftBar: StringNode = ConstantStringNode("|")
  val expression: Expression
  val rightBar: StringNode = ConstantStringNode("|")

  override def stringRepresentation: String = {
    leftBar.toString + expression.toString + rightBar.toString
  }

  def copyWhitespace(node: CardinalityExpression): Unit = {
    leftBar.copyWhitespace(node.leftBar)
    rightBar.copyWhitespace(node.rightBar)
    super.copyWhitespace(node)
  }
}

trait OptionSomeExpression extends ExpressionPredicate{
  override val keyword: StringNode = ConstantStringNode("Some")
}

trait TypedADTLiteral extends Expression {
  val adtType: StringNode
  val openAngle: StringNode = ConstantStringNode("<")
  val innerType: TypeIndication
  val closeAngle: StringNode = ConstantStringNode(">")
  val openBrace: StringNode = ConstantStringNode("{")
  val values: List[Expression]
  val closeBrace: StringNode = ConstantStringNode("}")

  override def stringRepresentation: String = {
    adtType.toString + openAngle.toString + innerType.toString + closeAngle.toString
    + openBrace.toString + values.mkString(",") + closeBrace.toString
  }

  def copyWhitespace(node: TypedADTLiteral): Unit = {
    adtType.copyWhitespace(node.adtType)
    openAngle.copyWhitespace(node.openAngle)
    closeAngle.copyWhitespace(node.closeAngle)
    openBrace.copyWhitespace(node.openBrace)
    closeBrace.copyWhitespace(node.closeBrace)
    super.copyWhitespace(node)
  }
}
trait TypedSeqLiteral extends TypedADTLiteral {
  override val adtType: StringNode = ConstantStringNode("seq")
}
trait TypedSetLiteral extends TypedADTLiteral {
  override val adtType: StringNode = ConstantStringNode("set")
}
trait TypedBagLiteral extends TypedADTLiteral {
  override val adtType: StringNode = ConstantStringNode("bag")
}

// e.g. VerCors Perm(e, p), or Verifast e|->v
trait PermissionExpression extends Expression {
  val target: Expression
  val amount: Option[Expression]
}

// e.g. Verifast e|->v, VerCors PointsTo(x,val, p)
trait ValuedPermissionExpression extends PermissionExpression {
  override val amount: Some[Expression]
  val value: Expression
}

// e.g. VerCors p=perm(e)
trait CurrentPermissionExpression extends Expression {
  val target: Expression
}

trait SomePermissionPredicate extends PermissionExpression {
  override val amount: Option[Expression] = None
}

trait ScaledPredicateExpression extends Expression {
  val openBracket: StringNode = ConstantStringNode("[")
  val amount: Expression
  val closeBracket: StringNode = ConstantStringNode("]")
  val predicate: Expression

  override def stringRepresentation: String = {
    openBracket.toString + amount.toString + closeBracket.toString + predicate.toString
  }

  def copyWhitespace(node: ScaledPredicateExpression): Unit = {
    openBracket.copyWhitespace(node.openBracket)
    closeBracket.copyWhitespace(node.closeBracket)
    super.copyWhitespace(node)
  }
}

trait UnfoldingExpression extends Expression {
  val unfoldingKeyword: StringNode = ConstantStringNode("\\Unfolding")
  val predicate: Expression
  val inKeyword: StringNode = ConstantStringNode("\\in")
  val expression: Expression

  override def stringRepresentation: String = {
    unfoldingKeyword.toString + predicate.toString + inKeyword.toString + expression.toString
  }

  def copyWhitespace(node: UnfoldingExpression): Unit = {
    unfoldingKeyword.copyWhitespace(node.unfoldingKeyword)
    inKeyword.copyWhitespace(node.inKeyword)
    super.copyWhitespace(node)
  }
}

trait SeparatingConjunctOperator extends StringNode 
trait ImpliesOperator extends StringNode

trait SpecDeclarationBlock extends MemberDeclaration {
  val startSpec: Option[StringNode]
  val declarations: List[SpecDeclaration]
  val endSpec: Option[StringNode]

  override def stringRepresentation: String = {
    startSpec.getOrElse("").toString + declarations.mkString + endSpec.getOrElse("").toString
  }
}

trait SpecDeclaration extends MemberDeclaration
trait NonTranslatedSpecDeclaration extends SpecDeclaration with StringNode

trait GhostDeclaration extends SpecDeclaration {
  val ghost: GhostModifier
  val decl: MemberDeclaration

  override def stringRepresentation: String = ghost.toString + decl.toString
}
trait ModelDeclaration extends SpecDeclaration {
  val model: ModelModifier
  val decl: MemberDeclaration

  override def stringRepresentation: String = model.toString + decl.toString
}
trait InvariantDeclaration extends SpecDeclaration {
  val invariantKeyword: StringNode = ConstantStringNode("invariant")
  val expression: Expression
  val semicolon: StringNode = ConstantStringNode(";")

  override def stringRepresentation: String = {
    modifiers.mkString + invariantKeyword.toString + expression.toString + semicolon.toString
  }

  def copyWhitespace(node: InvariantDeclaration): Unit = {
    invariantKeyword.copyWhitespace(node.invariantKeyword)
    semicolon.copyWhitespace(node.semicolon)
    super.copyWhitespace(node)
  }
}

trait PredicateDeclaration extends SpecDeclaration {
  val tType: TypeIndication
  val name: StringNode
  val openParenthesis: StringNode = ConstantStringNode("(")
  val params: List[Parameter]
  val closeParenthesis: StringNode = ConstantStringNode(")")
  val equalSign: StringNode = ConstantStringNode("=")
  val expr: Option[Expression]
  val semicolon: StringNode = ConstantStringNode(";")

  override def stringRepresentation: String = {
    var res = tType.toString
    res += name.toString
    res += openParenthesis.toString
    res += params.mkString(",")
    res += closeParenthesis.toString
    if (expr.nonEmpty) {
      res += equalSign.toString
      res += expr.getOrElse("").toString
    }
    res += semicolon.toString
    res
  }
  
  def copyWhitespace(node: PredicateDeclaration): Unit = {
    openParenthesis.copyWhitespace(node.openParenthesis)
    closeParenthesis.copyWhitespace(node.closeParenthesis)
    equalSign.copyWhitespace(node.equalSign)
    semicolon.copyWhitespace(node.semicolon)
    super.copyWhitespace(node)
  }
}

trait InitiallyDeclaration extends SpecDeclaration {
  val initiallyKeyword: StringNode = ConstantStringNode("initially")
  val expression: Expression
  val semicolon: StringNode = ConstantStringNode(";")

  override def stringRepresentation: String = {
    modifiers.mkString + initiallyKeyword.toString + expression.toString + semicolon.toString
  }

  def copyWhitespace(node: InitiallyDeclaration): Unit = {
    initiallyKeyword.copyWhitespace(node.initiallyKeyword)
    semicolon.copyWhitespace(node.semicolon)
    super.copyWhitespace(node)
  }
}

trait ReadableDeclaration extends SpecDeclaration {
  val readableKeyword: StringNode = ConstantStringNode("readable")
  val identifier: StringNode
  val ifKeyword: StringNode = ConstantStringNode("if")
  val expression: Expression
  val semicolon: StringNode = ConstantStringNode(";")

  override def stringRepresentation: String = {
    var res = modifiers.mkString
    res += readableKeyword.toString
    res += identifier.toString
    res += ifKeyword.toString
    res += expression.toString
    res += semicolon.toString
    res
  }

  def copyWhitespace(node: ReadableDeclaration): Unit = {
    readableKeyword.copyWhitespace(node.readableKeyword)
    ifKeyword.copyWhitespace(node.ifKeyword)
    semicolon.copyWhitespace(node.semicolon)
    super.copyWhitespace(node)
  }
}

trait WritableDeclaration extends SpecDeclaration {
  val writable: StringNode = ConstantStringNode("writable")
  val identifier: StringNode
  val ifKeyword: StringNode = ConstantStringNode("if")
  val expression: Expression
  val semicolon: StringNode = ConstantStringNode(";")

  override def stringRepresentation: String = {
    var res = modifiers.mkString
    res += writable.toString
    res += identifier.toString
    res += ifKeyword.toString
    res += expression.toString
    res += semicolon.toString
    res
  }

  def copyWhitespace(node: WritableDeclaration): Unit = {
    writable.copyWhitespace(node.writable)
    ifKeyword.copyWhitespace(node.ifKeyword)
    semicolon.copyWhitespace(node.semicolon)
    super.copyWhitespace(node)
  }
}

trait MonitorsForDeclaration extends SpecDeclaration {
  val monitorsForKeyword: StringNode = ConstantStringNode("monitors_for")
  val identifier: StringNode
  val equals: StringNode = ConstantStringNode("=")
  val expressions: List[Expression]
  val semicolon: StringNode = ConstantStringNode(";")

  override def stringRepresentation: String = {
    var res = modifiers.mkString
    res += monitorsForKeyword.toString
    res += identifier.toString
    res += equals.toString
    res += expressions.mkString(",")
    res += semicolon.toString
    res
  }

  def copyWhitespace(node: MonitorsForDeclaration): Unit = {
    monitorsForKeyword.copyWhitespace(node.monitorsForKeyword)
    equals.copyWhitespace(node.equals)
    semicolon.copyWhitespace(node.semicolon)
    super.copyWhitespace(node)
  }
}

trait AxiomDeclaration extends SpecDeclaration {
  val axiomKeyword: StringNode = ConstantStringNode("axiom")
  val expression: Expression
  val semicolon: StringNode = ConstantStringNode(";")

  override def stringRepresentation: String = {
    axiomKeyword.toString + expression.toString + semicolon.toString
  }

  def copyWhitespace(node: AxiomDeclaration): Unit = {
    axiomKeyword.copyWhitespace(node.axiomKeyword)
    semicolon.copyWhitespace(node.semicolon)
    super.copyWhitespace(node)
  }
}

trait SpecMethodDeclaration extends MethodDeclaration with SpecDeclaration
