package languages.Krakatoa

import generic.{NotSupportedException, SourceLocation}
import languages.Krakatoa.syntaxTree as krakatoa
import intermediateRepresentation.syntaxTree as ir
import languages.Krakatoa.syntaxTree.EmptyStatement

/**
 * Translates from intermediate tree to Krakatoa-specific AST
 */
class FromIRVisitor {

  //TODO: May need to store specDeclarations so they can be moved to outside the class?
  var specDeclarations: List[ir.MemberDeclaration] = List.empty

  def copyWhitespace(from: ir.IrTreeNode, to: krakatoa.KrakatoaTreeNode): Unit = {
    to.precedingWhitespace = from.precedingWhitespace
    to.trailingWhitespace = from.trailingWhitespace
  }

  def visitStringNode(node: ir.StringNode): krakatoa.StringNode = {
    node match {
      case ir.StringNode(text, src, preceding, trailing) => krakatoa.StringNode(text, src, preceding, trailing)
    }
  }

  def visitCompilationUnit(node: ir.CompilationUnit): krakatoa.CompilationUnit = {
    node match {
      case ir.CompilationUnit(pkg, imports, decls, srcLocation, precedingWhitespace) =>
        var krakatoaPkg: Option[krakatoa.PackageDeclaration] = None
        if (pkg.isDefined) krakatoaPkg = Some(visitPackageDeclaration(pkg.get))
        val res = krakatoa.CompilationUnit(krakatoaPkg, imports.map(visitImportDeclaration), decls.map(visitTypeDeclaration), srcLocation, precedingWhitespace)
        copyWhitespace(node, res)
        res
    }
  }

  def visitPackageDeclaration(node: ir.PackageDeclaration): krakatoa.PackageDeclaration = {
    node match {
      case ir.PackageDeclaration(name, annotations, srcLocation, precedingWhitespace) =>
        val res = krakatoa.PackageDeclaration(visitStringNode(name), annotations.map(visitStringAnnotation), srcLocation, precedingWhitespace)
        res.packageKeyword.precedingWhitespace = node.packageKeyword.precedingWhitespace
        res.semicolon.precedingWhitespace = node.semicolon.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitImportDeclaration(node: ir.ImportDeclaration): krakatoa.ImportDeclaration = {
    node match {
      case ir.ImportDeclaration(name, srcLocation, precedingWhitespace) =>
        val res = krakatoa.ImportDeclaration(visitStringNode(name), srcLocation, precedingWhitespace)
        res.modifiers = node.modifiers.map(visitModifier)
        res.importKeyword.precedingWhitespace = node.importKeyword.precedingWhitespace
        res.semicolon.precedingWhitespace = node.semicolon.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitTypeDeclaration(node: ir.TypeDeclaration): krakatoa.TypeDeclaration = {
    node match {
      case decl: ir.AnnotationTypeDeclaration => visitAnnotationTypeDeclaration(decl)
      case decl: ir.ClassDeclaration => visitClassDeclaration(decl)
      case decl: ir.EnumDeclaration => visitEnumDeclaration(decl)
      case decl: ir.InterfaceDeclaration => visitInterfaceDeclaration(decl)
      case decl: ir.SpecDeclarationBlock => visitSpecDeclarationBlock(decl)
      case decl: ir.EmptyStatement => visitEmptyStatement(decl)
      case decl: ir.NonTranslatedTypeDeclaration => visitNonTranslatedTypeDeclaration(decl)
      case x =>
        println(s"Error: Unknown IR type declaration $x")
        null
    }
  }

  def visitNonTranslatedTypeDeclaration(decl: ir.NonTranslatedTypeDeclaration): krakatoa.NonTranslatedTypeDeclaration = {
    decl match {
      case ir.NonTranslatedTypeDeclaration(text, srcLocation, precedingWhitespace) =>
        val res = krakatoa.NonTranslatedTypeDeclaration(text, srcLocation, precedingWhitespace)
        copyWhitespace(decl, res)
        res
    }
  }

  def visitStringAnnotation(node: ir.StringAnnotation): krakatoa.StringAnnotation = {
    node match {
      case ir.StringAnnotation(text, srcLocation, precedingWhitespace) =>
        val res = krakatoa.StringAnnotation(text, srcLocation, precedingWhitespace)
        copyWhitespace(node, res)
        res
    }
  }

  def visitModifier(node: generic.syntaxTree.Modifier): krakatoa.Modifier = {
    node match {
      case mod: ir.Modifier => visitModifier(mod)
      case m =>
        println("Error unknown IR modifier $m")
        null
    }
  }

  def visitModifier(node: ir.Modifier): krakatoa.Modifier = {
    node match {
      case mod: ir.StaticModifier => visitStaticModifier(mod)
      case mod: ir.Annotation => visitAnnotation(mod)
      case mod: ir.ModifierStringNode => visitModifierStringNode(mod)
      case mod: ir.PureModifier => visitPureModifier(mod)
      case mod: ir.NullableModifier => visitNullableModifier(mod)
      case mod: ir.NonnullModifier => visitNonnullModifier(mod)
      case mod: ir.GhostModifier => visitGhostModifier(mod)
      case mod: ir.InstanceModifier => visitInstanceModifier(mod)
      case mod: ir.SpecMethodModifier => visitSpecMethodModifier(mod)
      case x =>
        println(s"Error: Unknown IR modifier $x")
        null
    }
  }
  def visitStaticModifier(node: ir.StaticModifier): krakatoa.StaticModifier = {
    node match {
      case ir.StaticModifier(srcLocation, precedingWhitespace) =>
        val res = krakatoa.StaticModifier(srcLocation, precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitNonnullModifier(node: ir.NonnullModifier): krakatoa.Modifier = {
    node match {
      case ir.NonnullModifier(srcLocation, precedingWhitespace) =>
        val res = krakatoa.ModifierStringNode("non_null", srcLocation, precedingWhitespace)
        copyWhitespace(node, res)
        res
    }
  }

  def visitNullableModifier(node: ir.NullableModifier): krakatoa.Modifier = {
    node match {
      case ir.NullableModifier(srcLocation, precedingWhitespace) =>
        val res = krakatoa.ModifierStringNode("nullable", srcLocation, precedingWhitespace)
        copyWhitespace(node, res)
        res
    }
  }

  def visitSpecMethodModifier(node: ir.SpecMethodModifier): krakatoa.Modifier = {
    node match {
      case ir.SpecMethodModifier(srcLocation, precedingWhitespace) =>
        val res = krakatoa.SpecMethodModifier(srcLocation, precedingWhitespace)
        copyWhitespace(node, res)
        res
    }
  }

  def visitAnnotationTypeDeclaration(node: ir.AnnotationTypeDeclaration): krakatoa.AnnotationTypeDeclaration = {
    node match {
      case ir.AnnotationTypeDeclaration(name, body, srcLocation, precedingWhitespace) =>
        val res = krakatoa.AnnotationTypeDeclaration(visitStringNode(name), visitBody(body), srcLocation, precedingWhitespace)
        res.at.precedingWhitespace = node.at.precedingWhitespace
        res.interfaceKeyword.precedingWhitespace = node.interfaceKeyword.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitClassDeclaration(node: ir.ClassDeclaration): krakatoa.ClassDeclaration = {
    node match {
      case ir.ClassDeclaration(name, typeParameters, superclass, interfaces, body, srcLocation, precedingWhitespace) =>
        val krakatoaTypeParameters = if (typeParameters.isEmpty) None else Some(visitTypeParameterClause(typeParameters.get))
        val krakatoaSuperclass = if (superclass.isEmpty) None else Some(visitSuperclassClause(superclass.get))
        val krakatoaInterfaces = if (interfaces.isEmpty) None else Some(visitInterfaceClause(interfaces.get))
        val res = krakatoa.ClassDeclaration(visitStringNode(name), krakatoaTypeParameters, krakatoaSuperclass, krakatoaInterfaces, visitBody(body), srcLocation, precedingWhitespace)
        res.classKeyword.precedingWhitespace = node.classKeyword.precedingWhitespace
        res.modifiers = node.modifiers.map(visitModifier)
        copyWhitespace(node, res)
        res
    }
  }

  def visitEnumDeclaration(node: ir.EnumDeclaration): krakatoa.EnumDeclaration = {
    node match {
      case ir.EnumDeclaration(name, interfaces, constants, declarations, comma, semicolon, srcLocation, precedingWhitespace) =>
        val krakatoaInterfaces = if (interfaces.isEmpty) None else Some(visitInterfaceClause(interfaces.get))
        val krakatoaComma = if (comma.isEmpty) None else Some(visitStringNode(comma.get))
        val krakatoaSemicolon = if (semicolon.isEmpty) None else Some(visitStringNode(semicolon.get))
        val res = krakatoa.EnumDeclaration(visitStringNode(name), krakatoaInterfaces, constants.map(visitEnumConstant), declarations.map(visitMemberDeclaration), krakatoaComma, krakatoaSemicolon, srcLocation, precedingWhitespace)
        res.modifiers = node.modifiers.map(visitModifier)
        res.enumKeyword.precedingWhitespace = node.enumKeyword.precedingWhitespace
        res.openBrace.precedingWhitespace = node.openBrace.precedingWhitespace
        res.closeBrace.precedingWhitespace = node.closeBrace.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitInterfaceDeclaration(node: ir.InterfaceDeclaration): krakatoa.InterfaceDeclaration = {
    node match {
      case ir.InterfaceDeclaration(name, typeParameter, superInterfaces, body, srcLocation, precedingWhitespace) =>
        val krakatoaTypeParameters = if (typeParameter.isEmpty) None else Some(visitTypeParameterClause(typeParameter.get))
        var krakatoaSuperclass = if (superInterfaces.isEmpty) None else Some(visitMultiSuperclassClause(superInterfaces.get))
        val res = krakatoa.InterfaceDeclaration(visitStringNode(name), krakatoaTypeParameters, krakatoaSuperclass, visitBody(body), srcLocation, precedingWhitespace)
        res.interfaceKeyword.precedingWhitespace = node.interfaceKeyword.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitEmptyStatement(node: ir.EmptyStatement): krakatoa.EmptyStatement = {
    val res = new EmptyStatement(node.srcLocation)
    copyWhitespace(node, res)
    res
  }

  def visitSpecDeclarationBlock(node: ir.SpecDeclarationBlock): krakatoa.SpecDeclarationBlock = {
    node match {
      case ir.SpecDeclarationBlock(startSpec, declarations, endSpec, srcLocation, precedingWhitespace) =>
        val res = krakatoa.SpecDeclarationBlock(startSpec.map(visitStringNode), declarations.flatMap(x => visitSpecDeclaration(x)), endSpec.map(visitStringNode), srcLocation, precedingWhitespace)
        copyWhitespace(node, res)
        res
    }
  }

  def visitAnnotation(node: ir.Annotation): krakatoa.Annotation = {
    node match {
      case ann: ir.StringAnnotation => visitStringAnnotation(ann)
      case ann =>
        println(s"Error: Unknown IR Annotation $ann")
        null
    }
  }

  def visitModifierStringNode(node: ir.ModifierStringNode): krakatoa.ModifierStringNode = {
    node match {
      case ir.ModifierStringNode(str, srcLocation, whitespace) =>
        val res = krakatoa.ModifierStringNode(str, srcLocation, whitespace)
        copyWhitespace(node, res)
        res
    }
  }

  def visitPureModifier(node: ir.PureModifier): krakatoa.PureModifier = {
    node match {
      case ir.PureModifier(srcLocation, precedingWhitespace) =>
        val res = krakatoa.PureModifier(srcLocation, precedingWhitespace)
        copyWhitespace(node, res)
        res
    }
  }

  def visitInstanceModifier(node: ir.InstanceModifier): krakatoa.ModifierStringNode = {
    node match {
      case ir.InstanceModifier(srcLocation, precedingWhitespace) =>
        System.err.println(s"[warning] Instance modifier not supported ${srcLocation.toString}, dropped.")
        val res = krakatoa.ModifierStringNode("", srcLocation)
        res.copyWhitespace(node)
        res
    }
  }

  def visitBody(node: ir.Body): krakatoa.Body = {
    node match {
      case ir.Body(declarations, srcLocation, precedingWhitespace) =>
        val res = krakatoa.Body(declarations.map(visitMemberDeclaration), srcLocation, precedingWhitespace)
        res.openBrace.precedingWhitespace = node.openBrace.precedingWhitespace
        res.closeBrace.precedingWhitespace = node.closeBrace.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitTypeParameterClause(node: ir.TypeParameterClause): krakatoa.TypeParameterClause = {
    node match {
      case ir.TypeParameterClause(parameters, srcLocation, precedingWhitespace) =>
        val res = krakatoa.TypeParameterClause(parameters.map(visitTypeParameter), srcLocation, precedingWhitespace)
        res.openAngle.precedingWhitespace = node.openAngle.precedingWhitespace
        res.closeAngle.precedingWhitespace = node.closeAngle.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitSuperclassClause(node: ir.SuperclassClause): krakatoa.SuperclassClause = {
    node match {
      case ir.SuperclassClause(superclass, srcLocation, precedingWhitespace) =>
        val res = krakatoa.SuperclassClause(visitTypeIndication(superclass), srcLocation, precedingWhitespace)
        res.extendsKeyword.precedingWhitespace = node.extendsKeyword.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitInterfaceClause(node: ir.InterfaceClause): krakatoa.InterfaceClause = {
    node match {
      case ir.InterfaceClause(interfaces, srcLocation, precedingWhitespace) =>
        val res = krakatoa.InterfaceClause(interfaces.map(visitTypeIndication), srcLocation, precedingWhitespace)
        res.implementsKeyword.precedingWhitespace = node.implementsKeyword.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitEnumConstant(node: ir.EnumConstant): krakatoa.EnumConstant = {
    node match {
      case ir.EnumConstant(name, annotations, arguments, body, srcLocation, precedingWhitespace) =>
        val krakatoaArgs = if (arguments.isEmpty) None else Some(visitArguments(arguments.get))
        val krakatoaBody = if (body.isEmpty) None else Some(visitBody(body.get))
        val res = krakatoa.EnumConstant(visitStringNode(name), annotations.map(visitStringAnnotation), krakatoaArgs, krakatoaBody, srcLocation, precedingWhitespace)
        copyWhitespace(node, res)
        res
    }
  }


  def visitMemberDeclaration(node: ir.MemberDeclaration): krakatoa.MemberDeclaration = {
    node match {
      case mem: ir.ClassDeclaration => visitClassDeclaration(mem)
      case mem: ir.EnumDeclaration => visitEnumDeclaration(mem)
      case mem: ir.InterfaceDeclaration => visitInterfaceDeclaration(mem)
      case mem: ir.MethodDeclaration => visitMethodDeclaration(mem)
      case mem: ir.BlockOrSemicolon => visitBlockOrSemicolon(mem)
      case mem: ir.FieldDeclaration => visitFieldDeclaration(mem)
      case mem: ir.AnnotationTypeDeclaration => visitAnnotationTypeDeclaration(mem)
      case mem: ir.SpecDeclarationBlock => visitSpecDeclarationBlock(mem)
      case m =>
        println(s"Error unknown IR MemberDeclaration $m")
        null
    }
  }

  def visitMultiSuperclassClause(node: ir.MultiSuperclassClause): krakatoa.MultiSuperclassClause = {
    node match {
      case ir.MultiSuperclassClause(superclasses, srcLocation, precedingWhitespace) =>
        val res = krakatoa.MultiSuperclassClause(superclasses.map(visitTypeIndication), srcLocation, precedingWhitespace)
        res.extendsKeyword.precedingWhitespace = node.extendsKeyword.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitSpecDeclaration(node: ir.SpecDeclaration): List[krakatoa.SpecDeclaration] = {
    node match {
      case decl: ir.PredicateDeclaration => List(visitPredicateDeclaration(decl))
      //case decl: ir.InductiveDatatypeDeclaration => List(visitInductiveDatatypeDeclaration(decl))
      case decl: ir.InvariantDeclaration => List(visitInvariantDeclaration(decl))
      //case decl: ir.LemmaDeclaration => List(visitLemmaDeclaration(decl))
      //case decl: ir.LogicVariableDeclaration => List(visitLogicVariableDeclaration(decl))
      case decl: ir.AxiomDeclaration => List(visitAxiomDeclaration(decl))
      //case decl: ir.AxiomaticDeclaration => visitAxiomaticDeclaration(decl)
      case decl: ir.GhostDeclaration => List(visitGhostDeclaration(decl))
      case d =>
        println(s"Error unknown IR SpecDeclaration $d")
        null
    }
  }

  def visitTypeParameter(node: ir.TypeParameter): krakatoa.TypeParameter = {
    node match {
      case ir.TypeParameter(annotations, name, superclass, srcLocation, precedingWhitespace) =>
        val krakatoaSuperclass = if (superclass.isEmpty) None else Some(visitGenericsSuperclassClause(superclass.get))
        val res = krakatoa.TypeParameter(annotations.map(visitStringAnnotation), visitTypeIndication(name), krakatoaSuperclass, srcLocation, precedingWhitespace)
        copyWhitespace(node, res)
        res
    }
  }

  def visitTypeIndication(node: ir.TypeIndication): krakatoa.TypeIndication = {
    node match {
      case ind: ir.SpecTypeIndication => visitSpecTypeIndication(ind)
      case ind: ir.NoType => visitNoType(ind)
      case ind: ir.StringGenericTypeIndication => visitStringGenericTypeIndication(ind)
      case ind: ir.StringTypeIndication => visitStringTypeIndication(ind)
      case i =>
        println(s"Error unknown IR TypeIndication $i")
        null
    }
  }

  def visitSpecTypeIndication(node: ir.SpecTypeIndication): krakatoa.SpecTypeIndication = {
    node match {
      case ind: ir.IntegerSpecTypeIndication => visitIntegerSpecTypeIndication(ind)
      case ind: ir.RealSpecTypeIndication => visitRealSpecTypeIndication(ind)
    }
  }

  def visitIntegerSpecTypeIndication(node: ir.IntegerSpecTypeIndication): krakatoa.IntegerSpecTypeIndication = {
    node match {
      case ir.IntegerSpecTypeIndication(_fullTypeString, anns, _typeAsString, src, whitespace) =>
        val res = krakatoa.IntegerSpecTypeIndication(annotations = anns.map(visitAnnotation), srcLocation = src, precedingWhitespace = whitespace)
        copyWhitespace(node, res)
        res
    }
  }

  def visitRealSpecTypeIndication(node: ir.RealSpecTypeIndication): krakatoa.RealSpecTypeIndication = {
    node match {
      case ir.RealSpecTypeIndication(_fullTypeString, anns, _typeAsString, src, whitespace) =>
        val res = krakatoa.RealSpecTypeIndication(annotations = anns.map(visitAnnotation), srcLocation = src, precedingWhitespace = whitespace)
        copyWhitespace(node, res)
        res
    }
  }

  def visitArguments(node: ir.Arguments): krakatoa.Arguments = {
    node match {
      case ir.Arguments(arguments, srcLocation, precedingWhitespace) =>
        val res = krakatoa.Arguments(arguments.map(visitExpression), srcLocation, precedingWhitespace)
        res.openParenthesis.precedingWhitespace = node.openParenthesis.precedingWhitespace
        res.closeParenthesis.precedingWhitespace = node.closeParenthesis.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitMethodDeclaration(node: ir.MethodDeclaration): krakatoa.MethodDeclaration = {
    node match {
      case ir.MethodDeclaration(name, returnType, parameters, dimensions, throws, default, contractWithin, body, srcLocation, precedingWhitespace) =>
        val contract = node.contractBefore.map(visitContract)
        val krakatoaReturn = if (returnType.isEmpty) None else Some(visitTypeIndication(returnType.get))
        val krakatoaDims = if (dimensions.isEmpty) None else Some(visitStringNode(dimensions.get))
        val krakatoaThrows = if (throws.isEmpty) None else Some(visitThrowsClause(throws.get))
        val krakatoaDefault = if (default.isEmpty) None else Some(visitDefaultClause(default.get))
        val krakatoaContractWithin = if (contractWithin.isEmpty) None else Some(visitContract(contractWithin.get))
        val res = krakatoa.MethodDeclaration(visitStringNode(name), krakatoaReturn, parameters.map(visitParameter), krakatoaDims, krakatoaThrows, krakatoaDefault, krakatoaContractWithin, visitBlockOrSemicolon(body), srcLocation, precedingWhitespace)
        res.contractBefore = contract
        res.modifiers = node.modifiers.map(visitModifier)
        res.openParenthesis.precedingWhitespace = node.openParenthesis.precedingWhitespace
        res.closeParenthesis.precedingWhitespace = node.closeParenthesis.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitFieldDeclaration(node: ir.FieldDeclaration): krakatoa.FieldDeclaration = {
    node match {
      case ir.FieldDeclaration(fieldType, declarators, srcLocation, precedingWhitespace) =>
        val res = krakatoa.FieldDeclaration(visitTypeIndication(fieldType), declarators.map(visitVariableDeclarator), srcLocation, precedingWhitespace)
        res.modifiers = node.modifiers.map(visitModifier)
        res.semicolon.precedingWhitespace = node.semicolon.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitBlockOrSemicolon(node: ir.BlockOrSemicolon): krakatoa.BlockOrSemicolon = {
    node match {
      case block: ir.Block => visitBlock(block)
      case semi: ir.EmptyStatement => visitEmptyStatement(semi)
      case o =>
        println(s"Error unknown IR BlockOrSemicolon $o")
        null
    }
  }

  def visitPredicateDeclaration(node: ir.PredicateDeclaration): krakatoa.SpecDeclaration = {
    node match {
      case ir.PredicateDeclaration(name, ttype, parameters, expression, srcLocation, precedingWhitespace) =>
        //TODO: check whether it's okay to ignore the ttype
        val res = krakatoa.PredicateDeclaration(visitStringNode(name), None, parameters.map(visitParameter), expression.map(visitExpression), srcLocation, precedingWhitespace)
        res.tType.precedingWhitespace = node.tType.precedingWhitespace
        res
    }
  }

  def visitInvariantDeclaration(node: ir.InvariantDeclaration): krakatoa.InvariantDeclaration = {
    node match {
      case ir.InvariantDeclaration(expression, srcLocation, precedingWhitespace) =>
        val nameInvariant = krakatoa.StringNode("", srcLocation, precedingWhitespace)
        val res = krakatoa.InvariantDeclaration(nameInvariant, visitExpression(expression), srcLocation, precedingWhitespace)
        res.modifiers = node.modifiers.map(visitModifier)
        res.invariantKeyword.precedingWhitespace = node.invariantKeyword.precedingWhitespace
        res.semicolon.precedingWhitespace = node.semicolon.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitAxiomDeclaration(node: ir.AxiomDeclaration): krakatoa.SpecDeclaration = {
    node match {
      case ir.AxiomDeclaration(expression, srcLocation, precedingWhitespace) =>
        val nameAxiom = krakatoa.StringNode("", srcLocation, precedingWhitespace)
        val res = krakatoa.AxiomDeclaration(nameAxiom, None, visitExpression(expression), srcLocation, precedingWhitespace)
        copyWhitespace(node, res)
        res.axiomKeyword.precedingWhitespace = node.axiomKeyword.precedingWhitespace
        res
    }
  }

  def visitGhostDeclaration(node: ir.GhostDeclaration): krakatoa.GhostDeclaration = {
    krakatoa.GhostDeclaration(visitMemberDeclaration(node.decl), node.srcLocation, node.precedingWhitespace)
  }

  def visitGhostModifier(node: ir.GhostModifier): krakatoa.GhostModifier = {
    node match {
      case ir.GhostModifier(srcLocation, precedingWhitespace) => krakatoa.GhostModifier(srcLocation, precedingWhitespace)
    }
  }

  def visitGenericsSuperclassClause(node: ir.GenericsSuperclassClause): krakatoa.GenericsSuperclassClause = {
    node match {
      case ir.GenericsSuperclassClause(superclasses, srcLocation, precedingWhitespace) =>
        val res = krakatoa.GenericsSuperclassClause(superclasses.map(visitTypeIndication), srcLocation, precedingWhitespace)
        res.extendsKeyword.precedingWhitespace = node.extendsKeyword.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitNoType(node: ir.NoType): krakatoa.NoType = {
    node match {
      case ir.NoType(srcLocation) =>
        val res = krakatoa.NoType(srcLocation)
        copyWhitespace(node, res)
        res
    }
  }

  def visitStringGenericTypeIndication(node: ir.StringGenericTypeIndication): krakatoa.StringGenericTypeIndication = {
    node match {
      case ir.StringGenericTypeIndication(fullTypeString, typeParameters, annotations, typeAsString, srcLocation, precedingWhitespace) =>
        val res = krakatoa.StringGenericTypeIndication(fullTypeString, visitTypeParameterClause(typeParameters), annotations.map(visitAnnotation), typeAsString, srcLocation, precedingWhitespace)
        res.modifiers = node.modifiers.map(visitModifier)
        copyWhitespace(node, res)
        res
    }
  }

  def visitStringTypeIndication(node: ir.StringTypeIndication): krakatoa.StringTypeIndication = {
    node match {
      case ir.StringTypeIndication(fullTypeString, annotations, typeAsString, srcLocation, precedingWhitespace, trailingWhitespace) =>
        val typeString = fullTypeString.replace("integer", "int")
        val res = krakatoa.StringTypeIndication(typeString, annotations.map(visitAnnotation), typeAsString.replace("integer", "int"), srcLocation, precedingWhitespace, trailingWhitespace)
        res.modifiers = node.modifiers.map(visitModifier)
        copyWhitespace(node, res)
        res
    }
  }

  def visitExpression(node: ir.Expression): krakatoa.Expression = {
    node match {
      case expr: ir.NestedInvocationExpression => visitNestedInvocationExpression(expr)
      case expr: ir.BinaryExpression => visitBinaryExpression(expr)
      case expr: ir.InstanceofExpression => visitInstanceofExpression(expr)
      case expr: ir.ArrayIndexExpression => visitArrayIndexExpression(expr)
      case expr: ir.MethodCallExpression => visitMethodCallExpression(expr)
      case expr: ir.NewObjectExpression => visitNewObjectExpression(expr)
      case expr: ir.NewArrayExpression => visitNewArrayExpression(expr)
      case expr: ir.CastExpression => visitCastExpression(expr)
      case expr: ir.PostfixExpression => visitPostfixExpression(expr)
      case expr: ir.PrefixExpression => visitPrefixExpression(expr)
      case expr: ir.TernaryExpression => visitTernaryExpression(expr)
      case expr: ir.LambdaExpression => visitLambdaExpression(expr)
      case expr: ir.MethodReferenceExpression => visitMethodReferenceExpression(expr)
      case expr: ir.ParExpression => visitParExpression(expr)
      case expr: ir.Literal => visitLiteral(expr)
      case expr: ir.ClassName => visitClassName(expr)
      case expr: ir.GenericInvocationExpression => visitGenericInvocationExpression(expr)
      case expr: ir.ResultExpression => visitResultExpression(expr)
      case expr: ir.OldExpression => visitOldExpression(expr)
      case expr: ir.FreshExpression => visitFreshExpression(expr)
      case expr: ir.ForallExpression => visitForallExpression(expr)
      case expr: ir.ExistsExpression => visitExistsExpression(expr)
      case expr: ir.StringNode => visitStringNode(expr)
      case e =>
        println(s"Error unknown IR Expression $e")
        null
    }
  }

  def visitResultExpression(node: ir.ResultExpression): krakatoa.ResultExpression = {
    node match {
      case ir.ResultExpression(srcLocation, precedingWhitespace) =>
        val res = krakatoa.ResultExpression(srcLocation, precedingWhitespace)
        copyWhitespace(node, res)
        res
    }
  }

  def visitThrowsClause(node: ir.ThrowsClause): krakatoa.Throws = {
    node match {
      case ir.ThrowsClause(exceptions, srcLocation, precedingWhitespace) =>
        val res = krakatoa.Throws(exceptions.map(visitStringNode), srcLocation, precedingWhitespace)
        res.throwsKeyword.precedingWhitespace = node.throwsKeyword.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitDefaultClause(node: ir.DefaultClause): krakatoa.DefaultClause = {
    node match {
      case ir.DefaultClause(value, srcLocation, precedingWhitespace) =>
        val res = krakatoa.DefaultClause(visitStringNode(value), srcLocation, precedingWhitespace)
        res.defaultKeyword.precedingWhitespace = node.defaultKeyword.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitContract(node: ir.Contract): krakatoa.Contract = {
    node match {
      case ir.Contract(clauses, srcLocation, precedingWhitespace) =>
        val res = krakatoa.Contract(clauses.map(visitContractClause), srcLocation)
        copyWhitespace(node, res)
        res
    }
  }

  def visitParameter(node: ir.Parameter): krakatoa.Parameter = {
    node match {
      case ir.Parameter(parameterType, name, srcLocation, precedingWhitespace) =>
        val modifiers = node.modifiers.map(visitModifier)
        val res = krakatoa.Parameter(visitTypeIndication(parameterType), visitStringNode(name), modifiers, srcLocation, precedingWhitespace)
        copyWhitespace(node, res)
        res
    }
  }

  def visitVariableDeclarator(node: ir.VariableDeclarator): krakatoa.VariableDeclarator = {
    node match {
      case ir.VariableDeclarator(name, initializer, srcLocation, precedingWhitespace) =>
        val krakatoaInit = if (initializer.isEmpty) None else Some(visitInitializer(initializer.get))
        val res = krakatoa.VariableDeclarator(visitStringNode(name), krakatoaInit, srcLocation, precedingWhitespace)
        res.equalSign.precedingWhitespace = node.equalSign.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitBlock(node: ir.Block): krakatoa.Block = {
    node match {
      case ir.Block(static, statements, srcLocation, precedingWhitespace) =>
        val krakatoaStatic = static.map(visitModifier)
        val res = krakatoa.Block(krakatoaStatic, statements.map(visitStatement), srcLocation, precedingWhitespace)
        res.openBrace.precedingWhitespace = node.openBrace.precedingWhitespace
        res.closeBrace.precedingWhitespace = node.closeBrace.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitContractClause(node: ir.ContractClause): krakatoa.ContractClause = {
    try {
      node match {
        case m: ir.AssignableClause =>
          val res = krakatoa.AssignsClause(expressions = m.expressions.map(visitExpression), m.srcLocation, m.precedingWhitespace)
          res.assignableKeyword.precedingWhitespace = m.assignableKeyword.precedingWhitespace
          res.startSpec = m.startSpec.map(visitStringNode)
          res.endSpec = m.endSpec.map(visitStringNode)
          res
        case m: ir.AccessibleClause =>
          val res = krakatoa.AccessibleClause(m.expressions.map(visitExpression), m.srcLocation, m.precedingWhitespace)
          res.accessibleKeyword.precedingWhitespace = m.accessibleKeyword.precedingWhitespace
          res.startSpec = m.startSpec.map(visitStringNode)
          res.endSpec = m.endSpec.map(visitStringNode)
          res
        case m: ir.Precondition =>
          //TODO: "requires" needs to become an "assumes" in behavior clauses
          val res = krakatoa.RequiresClause(visitExpression(m.expression), m.srcLocation, m.precedingWhitespace)
          res.requiresKeyword.precedingWhitespace = m.requiresKeyword.precedingWhitespace
          res.startSpec = m.startSpec.map(visitStringNode)
          res.endSpec = m.endSpec.map(visitStringNode)
          res
        case m: ir.Postcondition =>
          val res = krakatoa.EnsuresClause(visitExpression(m.expression), m.srcLocation, m.precedingWhitespace)
          res.ensuresKeyword.precedingWhitespace = m.ensuresKeyword.precedingWhitespace
          res.startSpec = m.startSpec.map(visitStringNode)
          res.endSpec = m.endSpec.map(visitStringNode)
          res
        case m: ir.MeasuredByClause =>
          val res = krakatoa.DecreasesClause(visitExpression(m.expression), m.srcLocation, m.precedingWhitespace)
          res.measuredByKeyword.precedingWhitespace = m.measuredByKeyword.precedingWhitespace
          res.startSpec = m.startSpec.map(visitStringNode)
          res.endSpec = m.endSpec.map(visitStringNode)
          res
        case m: ir.NormalBehavior =>
          val preconditions = m.clauses.filter(str => (str.isInstanceOf[ir.Precondition]))
          //behaviorPreconditions += m.behaviorName.text -> preconditions.map(p => p.asInstanceOf[ir.AssumesClause])
          val name = krakatoa.StringNode("", m.srcLocation, " ")
          val res = krakatoa.NormalBehavior(m.startSpec.map(visitStringNode), name, m.hasAlso, m.privacy.map(visitStringNode), m.clauses.map(visitContractClause), m.endSpec.map(visitStringNode), m.srcLocation, m.precedingWhitespace)
          res.behaviorKeyword.precedingWhitespace = m.behaviorKeyword.precedingWhitespace
          res.startSpec = m.startSpec.map(visitStringNode)
          res.endSpec = m.endSpec.map(visitStringNode)
          res
        case m: ir.ExceptionalBehavior =>
          val preconditions = m.clauses.filter(str => str.isInstanceOf[ir.Precondition])
          //behaviorPreconditions += m.behaviorName.text -> preconditions.map(p => p.asInstanceOf[ir.AssumesClause])
          val name = krakatoa.StringNode("", m.srcLocation, " ")
          val res = krakatoa.ExceptionalBehavior(m.startSpec.map(visitStringNode), name, m.hasAlso, m.privacy.map(visitStringNode), m.clauses.map(visitContractClause), m.endSpec.map(visitStringNode), m.srcLocation, m.precedingWhitespace)
          res.behaviorKeyword.precedingWhitespace = m.behaviorKeyword.precedingWhitespace
          res.startSpec = m.startSpec.map(visitStringNode)
          res.endSpec = m.endSpec.map(visitStringNode)
          res
        case m: ir.Behavior =>
          val preconditions = m.clauses.filter(str => str.isInstanceOf[ir.Precondition])
          //this.behaviorPreconditions += m.behaviorName.text -> preconditions
          val name = krakatoa.StringNode("", m.srcLocation, " ")
          val res = krakatoa.Behavior(m.startSpec.map(visitStringNode), name, m.hasAlso, m.privacy.map(visitStringNode), m.clauses.map(visitContractClause), m.endSpec.map(visitStringNode), m.srcLocation, m.precedingWhitespace)
          res.behaviorKeyword.precedingWhitespace = m.behaviorKeyword.precedingWhitespace
          res.startSpec = m.startSpec.map(visitStringNode)
          res.endSpec = m.endSpec.map(visitStringNode)
          res
        case m: ir.ExceptionalPostcondition =>
          val res = krakatoa.SignalsClause(visitTypeIndication(m.exceptionType), m.exceptionName.map(visitStringNode), Some(visitExpression(m.expression.get)), m.srcLocation, m.precedingWhitespace)
          res.signalsKeyword.precedingWhitespace = m.signalsKeyword.precedingWhitespace
          res.startSpec = m.startSpec.map(visitStringNode)
          res.endSpec = m.endSpec.map(visitStringNode)
          res
      }
    } catch {
      case p: NotSupportedException =>
        var text = "\n" + p.getMessage
        text += node.stringRepresentation.replace("\n", "\n//").replace("@", " ")
        val res = krakatoa.NonTranslatedContractClause(text, node.srcLocation, node.precedingWhitespace)
        res
    }
  }

  def visitInitializer(node: ir.Initializer): krakatoa.Initializer = {
    node match {
      case init: ir.ArrayInitializer => visitArrayInitializer(init)
      case init: ir.Expression => visitExpression(init)
      case i =>
        println(s"Error unknown IR Initializer $i")
        null
    }
  }

  def visitStatement(node: ir.Statement): krakatoa.Statement = {
    try {
      node match {
        case stat: ir.LocalDeclaration => visitLocalDeclaration(stat)
        case stat: ir.Block => visitBlock(stat)
        case stat: ir.IfStatement => visitIfStatement(stat)
        case stat: ir.ForLoop => visitForLoop(stat)
        case stat: ir.WhileLoop => visitWhileLoop(stat)
        case stat: ir.DoLoop => visitDoWhileLoop(stat)
        case stat: ir.TryBlock => visitTry(stat)
        case stat: ir.SwitchStatement => visitSwitchStatement(stat)
        case stat: ir.SynchronizedStatement => visitSynchronizedStatement(stat)
        case stat: ir.ReturnStatement => visitReturnStatement(stat)
        case stat: ir.ThrowStatement => visitThrowStatement(stat)
        case stat: ir.BreakStatement => visitBreakStatement(stat)
        case stat: ir.ContinueStatement => visitContinueStatement(stat)
        case stat: ir.EmptyStatement => visitEmptyStatement(stat)
        case stat: ir.ExpressionStatement => visitExpressionStatement(stat)
        case stat: ir.LabeledStatement => visitLabeledStatement(stat)
        case stat: ir.AssertStatement => visitAssertStatement(stat)
        case stat: ir.GhostStatement => visitGhostStatement(stat)
        case stat: ir.JavaAssertion => visitAssert(stat)
        case stat: ir.JmlStatementBlock => visitJmlStatementBlock(stat)
        case s =>
          println(s"Error unknown IR statement $s")
          null
      }
    } catch {
      case p: NotSupportedException =>
        var text = "\n" + p.getMessage
        text += node.stringRepresentation.replace("\n", "\n//").replace("@", " ")
        val res = krakatoa.NonTranslatedStatement(text, node.srcLocation, node.precedingWhitespace)
        res
    }
  }

  def visitJmlStatementBlock(block: ir.JmlStatementBlock): krakatoa.KrakatoaStatementBlock = {
    block match {
      case ir.JmlStatementBlock(_startSpec, statements, _endSpec, srcLocation, precedingWhitespace) =>
        krakatoa.KrakatoaStatementBlock(_startSpec.map(visitStringNode), statements.map(visitStatement), _endSpec.map(visitStringNode), srcLocation, precedingWhitespace)
    }
  }

  def visitArrayInitializer(node: ir.ArrayInitializer): krakatoa.ArrayInitializer = {
    node match {
      case ir.ArrayInitializer(initializers, trailingComma, srcLocation, precedingWhitespace) =>
        val krakatoaComma = if (trailingComma.isEmpty) None else Some(visitStringNode(trailingComma.get))
        val res = krakatoa.ArrayInitializer(initializers.map(visitInitializer), krakatoaComma, srcLocation, precedingWhitespace)
        res.openBrace.precedingWhitespace = node.openBrace.precedingWhitespace
        res.closeBrace.precedingWhitespace = node.closeBrace.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitNestedInvocationExpression(node: ir.NestedInvocationExpression): krakatoa.NestedInvocationExpression = {
    node match {
      case ir.NestedInvocationExpression(left, right, srcLocation, precedingWhitespace) =>
        val res = krakatoa.NestedInvocationExpression(visitExpression(left), visitExpression(right), srcLocation, precedingWhitespace)
        res.dot.precedingWhitespace = node.dot.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitBinaryExpression(node: ir.BinaryExpression): krakatoa.BinaryExpression = {
    node match {
      case ir.BinaryExpression(left, operator, right, srcLocation, precedingWhitespace) =>
        val res = krakatoa.BinaryExpression(visitExpression(left), visitStringNode(operator), visitExpression(right), srcLocation, precedingWhitespace)
        copyWhitespace(node, res)
        res
    }
  }

  def visitInstanceofExpression(node: ir.InstanceofExpression): krakatoa.InstanceofExpression = {
    node match {
      case ir.InstanceofExpression(left, right, srcLocation, precedingWhitespace) =>
        val res = krakatoa.InstanceofExpression(visitExpression(left), visitTypeIndication(right), srcLocation, precedingWhitespace)
        copyWhitespace(node, res)
        res
    }
  }

  def visitArrayIndexExpression(node: ir.ArrayIndexExpression): krakatoa.ArrayIndexExpression = {
    node match {
      case ir.ArrayIndexExpression(arrayExpression, indexExpression, srcLocation, precedingWhitespace) =>
        val res = krakatoa.ArrayIndexExpression(visitExpression(arrayExpression), visitIndexExpression(indexExpression), srcLocation, precedingWhitespace)
        res.openBracket.precedingWhitespace = node.openBracket.precedingWhitespace
        res.closeBracket.precedingWhitespace = node.closeBracket.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitIndexExpression(node: ir.IndexExpression): krakatoa.IndexExpression = {
    node match {
      //TODO: double check whether I need to support the IndexAllExpression and RangeExpression
      case idx: ir.Expression => visitExpression(idx)
      case i =>
        println(s"Error unknown IR IndexExpression $i")
        null
    }
  }

  def visitMethodCallExpression(node: ir.MethodCallExpression): krakatoa.MethodCallExpression = {
    node match {
      case ir.MethodCallExpression(name, arguments, srcLocation, precedingWhitespace) =>
        val res = krakatoa.MethodCallExpression(visitStringNode(name), visitArguments(arguments), srcLocation, precedingWhitespace)
        copyWhitespace(node, res)
        res
    }
  }

  def visitNewObjectExpression(node: ir.NewObjectExpression): krakatoa.NewObjectExpression = {
    node match {
      case ir.NewObjectExpression(typeArguments, typeIndication, arguments, body, srcLocation, precedingWhitespace) =>
        val typeArgs = if (typeArguments.isEmpty) None else Some(visitTypeArgumentClause(typeArguments.get))
        val krakatoaBody = if (body.isEmpty) None else Some(visitBody(body.get))
        val res = krakatoa.NewObjectExpression(typeArgs, visitTypeIndication(typeIndication), visitArguments(arguments), krakatoaBody, srcLocation, precedingWhitespace)
        res.newKeyword.precedingWhitespace = node.newKeyword.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitNewArrayExpression(node: ir.NewArrayExpression): krakatoa.NewArrayExpression = {
    node match {
      case ir.NewArrayExpression(typeIndication, dimensions, initializer, srcLocation, precedingWhitespace) =>
        val krakatoaInit = if (initializer.isEmpty) None else Some(visitArrayInitializer(initializer.get))
        val res = krakatoa.NewArrayExpression(visitTypeIndication(typeIndication), dimensions.map(visitDimension), krakatoaInit, srcLocation, precedingWhitespace)
        res.newKeyword.precedingWhitespace = node.newKeyword.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitDimension(node: ir.Dimension): krakatoa.Dimension = {
    node match {
      case ir.Dimension(sizeExpression, srcLocation, precedingWhitespace) =>
        val krakatoaSize = if (sizeExpression.isEmpty) None else Some(visitExpression(sizeExpression.get))
        val res = krakatoa.Dimension(krakatoaSize, srcLocation, precedingWhitespace)
        res.openBracket.precedingWhitespace = node.openBracket.precedingWhitespace
        res.closeBracket.precedingWhitespace = node.closeBracket.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitCastExpression(node: ir.CastExpression): krakatoa.CastExpression = {
    node match {
      case ir.CastExpression(annotations, typeIndications, expression, srcLocation, precedingWhitespace) =>
        val res = krakatoa.CastExpression(annotations.map(visitStringAnnotation), typeIndications.map(visitTypeIndication), visitExpression(expression), srcLocation, precedingWhitespace)
        res.openParenthesis.precedingWhitespace = node.openParenthesis.precedingWhitespace
        res.closeParenthesis.precedingWhitespace = node.closeParenthesis.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitPostfixExpression(node: ir.PostfixExpression): krakatoa.PostfixExpression = {
    node match {
      case ir.PostfixExpression(expression, postfix, srcLocation, precedingWhitespace) =>
        val res = krakatoa.PostfixExpression(visitExpression(expression), visitStringNode(postfix), srcLocation, precedingWhitespace)
        copyWhitespace(node, res)
        res
    }
  }

  def visitPrefixExpression(node: ir.PrefixExpression): krakatoa.PrefixExpression = {
    node match {
      case ir.PrefixExpression(expression, prefix, srcLocation, precedingWhitespace) =>
        val res = krakatoa.PrefixExpression(visitExpression(expression), visitStringNode(prefix), srcLocation, precedingWhitespace)
        copyWhitespace(node, res)
        res
    }
  }

  def visitTernaryExpression(node: ir.TernaryExpression): krakatoa.TernaryExpression = {
    node match {
      case ir.TernaryExpression(condition, thenExpression, elseExpression, srcLocation, precedingWhitespace) =>
        val res = krakatoa.TernaryExpression(visitExpression(condition), visitExpression(thenExpression), visitExpression(elseExpression), srcLocation, precedingWhitespace)
        res.questionmark.precedingWhitespace = node.questionmark.precedingWhitespace
        res.colon.precedingWhitespace = node.colon.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitLambdaExpression(node: ir.LambdaExpression): krakatoa.LambdaExpression = {
    node match {
      case ir.LambdaExpression(hasParentheses, parameters, body, srcLocation, precedingWhitespace) =>
        val res = krakatoa.LambdaExpression(hasParentheses, parameters.map(visitParameter), visitLambdaBody(body), srcLocation, precedingWhitespace)
        res.arrow.precedingWhitespace = node.arrow.precedingWhitespace
        res.openParenthesis.precedingWhitespace = node.openParenthesis.precedingWhitespace
        res.closeParenthesis.precedingWhitespace = node.closeParenthesis.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitLambdaBody(node: ir.LambdaBody): krakatoa.LambdaBody = {
    node match {
      case body: ir.Expression => visitExpression(body)
      case b =>
        println(s"Error unknown IR LambdaBody $b")
        null
    }
  }

  def visitMethodReferenceExpression(node: ir.MethodReferenceExpression): krakatoa.MethodReferenceExpression = {
    node match {
      case ir.MethodReferenceExpression(left, typeArgument, right, srcLocation, precedingWhitespace) =>
        val krakatoaArg = if (typeArgument.isEmpty) None else Some(visitTypeArgumentClause(typeArgument.get))
        val res = krakatoa.MethodReferenceExpression(visitExpression(left), krakatoaArg, visitExpression(right), srcLocation, precedingWhitespace)
        res.doubleColon.precedingWhitespace = node.doubleColon.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitParExpression(node: ir.ParExpression): krakatoa.ParExpression = {
    node match {
      case ir.ParExpression(expression, srcLocation, precedingWhitespace) =>
        val res = krakatoa.ParExpression(visitExpression(expression), srcLocation, precedingWhitespace)
        res.openParenthesis.precedingWhitespace = node.openParenthesis.precedingWhitespace
        res.closeParenthesis.precedingWhitespace = node.closeParenthesis.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitLiteral(node: ir.Literal): krakatoa.Literal = {
    node match {
      case ir.Literal(value, srcLocation, precedingWhitespace) =>
        val res = krakatoa.Literal(visitStringNode(value), srcLocation, precedingWhitespace)
        copyWhitespace(node, res)
        res
    }
  }

  def visitClassName(node: ir.ClassName): krakatoa.ClassName = {
    node match {
      case ir.ClassName(typeIndication, srcLocation, precedingWhitespace) =>
        val res = krakatoa.ClassName(visitTypeIndication(typeIndication), srcLocation, precedingWhitespace)
        res.classKeyword.precedingWhitespace = node.classKeyword.precedingWhitespace
        res.dot.precedingWhitespace = node.dot.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitGenericInvocationExpression(node: ir.GenericInvocationExpression): krakatoa.GenericInvocationExpression = {
    node match {
      case ir.GenericInvocationExpression(typeArgumentClause, name, arguments, srcLocation, precedingWhitespace) =>
        val krakatoaArgs = if (arguments.isEmpty) None else Some(visitArguments(arguments.get))
        val res = krakatoa.GenericInvocationExpression(visitTypeArgumentClause(typeArgumentClause), visitStringNode(name), krakatoaArgs, srcLocation, precedingWhitespace)
        copyWhitespace(node, res)
        res
    }
  }

  def visitTypeArgumentClause(node: ir.TypeArgumentClause): krakatoa.TypeArgumentClause = {
    node match {
      case ir.TypeArgumentClause(parameters, srcLocation, precedingWhitespace) =>
        val res = krakatoa.TypeArgumentClause(parameters.map(visitTypeArgument), srcLocation, precedingWhitespace)
        res.openAngle.precedingWhitespace = node.openAngle.precedingWhitespace
        res.closeAngle.precedingWhitespace = node.closeAngle.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitTypeArgument(node: ir.TypeArgument): krakatoa.TypeArgument = {
    node match {
      case arg: ir.BoundedTypeArgument => visitBoundedTypeArgument(arg)
      case arg: ir.TypeIndication => visitTypeIndication(arg)
      case a =>
        println(s"Error unknown IR TypeArgument $a")
        null
    }
  }

  def visitBoundedTypeArgument(node: ir.BoundedTypeArgument): krakatoa.BoundedTypeArgument = {
    node match {
      case ir.BoundedTypeArgument(annotations, typeSuper, typeExtends, srcLocation, precedingWhitespace) =>
        val krakatoaSuper = if (typeSuper.isEmpty) None else Some(visitTypeIndication(typeSuper.get))
        val krakatoaExtends = if (typeExtends.isEmpty) None else Some(visitTypeIndication(typeExtends.get))
        val res = krakatoa.BoundedTypeArgument(annotations.map(visitAnnotation), krakatoaSuper, krakatoaExtends, srcLocation, precedingWhitespace)
        res.superKeyword.precedingWhitespace = node.superKeyword.precedingWhitespace
        res.extendsKeyword.precedingWhitespace = node.extendsKeyword.precedingWhitespace
        res.questionMark.precedingWhitespace = node.questionMark.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitOldExpression(node: ir.OldExpression): krakatoa.OldExpression = {
    node match {
      case ir.OldExpression(expression, label, srcLocation, precedingWhitespace) =>
        val krakatoaLabel = if (label.isEmpty) None else Some(visitStringNode(label.get))
        val res = krakatoa.OldExpression(visitExpression(expression), krakatoaLabel, srcLocation, precedingWhitespace)
        res.oldKeyword.precedingWhitespace = node.oldKeyword.precedingWhitespace
        res.comma.precedingWhitespace = node.comma.precedingWhitespace
        res.openParenthesis.precedingWhitespace = node.openParenthesis.precedingWhitespace
        res.closeParenthesis.precedingWhitespace = node.closeParenthesis.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitFreshExpression(node: ir.FreshExpression): krakatoa.FreshExpression = {
    node match {
      case ir.FreshExpression(expressions, srcLocation, precedingWhitespace) =>
        val res = krakatoa.FreshExpression(expressions.map(visitExpression), srcLocation, precedingWhitespace)
        res.openParenthesis.precedingWhitespace = node.openParenthesis.precedingWhitespace
        res.closeParenthesis.precedingWhitespace = node.closeParenthesis.precedingWhitespace
        res.freshKeyword.precedingWhitespace = node.freshKeyword.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitQuantifiedVarDeclaration(node: ir.QuantifiedVarDeclaration): krakatoa.QuantifiedVarDeclaration = {
    node match {
      case ir.QuantifiedVarDeclaration(declType, declarators, srcLocation, precedingWhitespace) =>
        krakatoa.QuantifiedVarDeclaration(visitTypeIndication(declType), declarators.map(visitStringNode), srcLocation, precedingWhitespace)
    }
  }

  def visitForallExpression(node: ir.ForallExpression): krakatoa.ForallExpression = {
    node match {
      case ir.ForallExpression(decls, guard, expression, srcLocation, precedingWhitespace) =>
        //TODO Need to extract guard & expression from expression (a => b)
        //TODO: May need to split decls into nested foralls?
        val krakatoaDecls = decls.map(visitQuantifiedVarDeclaration)
        var krakatoaExpression: krakatoa.Expression = null
        if (guard.isDefined) {
          val krakatoaGuard = visitExpression(guard.get)
          val implied = visitExpression(expression)
          val op = krakatoa.StringNode("==>", srcLocation, " ")
          krakatoaExpression = krakatoa.BinaryExpression(krakatoaGuard, op, implied, srcLocation, "")
        } else {
          krakatoaExpression = visitExpression(expression)
        }
        val res = krakatoa.ForallExpression(krakatoaDecls, krakatoaExpression, srcLocation, precedingWhitespace)
        res.forallKeyword.precedingWhitespace = node.quantifier.precedingWhitespace
        res.semi.precedingWhitespace = node.firstSemicolon.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitExistsExpression(node: ir.ExistsExpression): krakatoa.ExistsExpression = {
    node match {
      case ir.ExistsExpression(decls, guard, expression, srcLocation, precedingWhitespace) =>
        val krakatoaDecls = decls.map(visitQuantifiedVarDeclaration)
        var krakatoaExpression: krakatoa.Expression = null
        if (guard.isDefined) {
          val krakatoaGuard = visitExpression(guard.get)
          val implied = visitExpression(expression)
          val op = krakatoa.StringNode("==>", srcLocation, " ")
          krakatoaExpression = krakatoa.BinaryExpression(krakatoaGuard, op, implied, srcLocation, "")
        } else {
          krakatoaExpression = visitExpression(expression)
        }
        val res = krakatoa.ExistsExpression(krakatoaDecls, krakatoaExpression, srcLocation, precedingWhitespace)
        res.existsKeyword.precedingWhitespace = node.quantifier.precedingWhitespace
        res.semi.precedingWhitespace = node.firstSemicolon.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitLocalDeclaration(node: ir.LocalDeclaration): krakatoa.LocalVariableDeclaration = {
    node match {
      case ir.LocalDeclaration(modifiers, declType, declarators, isStatement, srcLocation, precedingWhitespace) =>
        val res = krakatoa.LocalVariableDeclaration(modifiers.map(visitModifier), visitTypeIndication(declType), declarators.map(visitVariableDeclarator), isStatement, srcLocation, precedingWhitespace)
        res.semicolon.precedingWhitespace = node.semicolon.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitIfStatement(node: ir.IfStatement): krakatoa.IfStatement = {
    node match {
      case ir.IfStatement(condition, thenStatement, elseStatement, srcLocation, precedingWhitespace) =>
        val krakatoaElse = if (elseStatement.isEmpty) None else Some(visitStatement(elseStatement.get))
        val res = krakatoa.IfStatement(visitExpression(condition), visitStatement(thenStatement), krakatoaElse, srcLocation, precedingWhitespace)
        res.ifKeyword.precedingWhitespace = node.ifKeyword.precedingWhitespace
        res.elseKeyword.precedingWhitespace = node.elseKeyword.precedingWhitespace
        res.openParenthesis.precedingWhitespace = node.openParenthesis.precedingWhitespace
        res.closeParenthesis.precedingWhitespace = node.closeParenthesis.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitForLoop(node: ir.ForLoop): krakatoa.ForLoop = {
    node match {
      case ir.ForLoop(contractBefore, control, contractWithin, body, srcLocation, precedingWhitespace) =>
        val krakatoaContractBefore = if (contractBefore.isEmpty) Nil else contractBefore.map(visitLoopSpecOrLabel)
        val krakatoaContractWithin = List.empty
        val res = krakatoa.ForLoop(krakatoaContractBefore, krakatoaContractWithin, visitForLoopControl(control), visitStatement(body), srcLocation, precedingWhitespace)
        res.forKeyword.precedingWhitespace = node.forKeyword.precedingWhitespace
        res.openParenthesis.precedingWhitespace = node.openParenthesis.precedingWhitespace
        res.closeParenthesis.precedingWhitespace = node.closeParenthesis.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitForLoopControl(node: ir.ForLoopControl): krakatoa.ForLoopControl = {
    node match {
      case init: ir.ForEachLoopControl => visitForEachLoopControl(init)
      case init: ir.RegularForLoopControl => visitRegularForLoopControl(init)
    }
  }

  def visitRegularForLoopControl(node: ir.RegularForLoopControl): krakatoa.RegularForLoopControl = {
    node match {
      case ir.RegularForLoopControl(init, condition, update, srcLocation, precedingWhitespace) =>
        val krakatoaInit = if (init.isEmpty) None else Some(visitForInit(init.get))
        val krakatoaCondition = if (condition.isEmpty) None else Some(visitExpression(condition.get))
        val res = krakatoa.RegularForLoopControl(krakatoaInit, krakatoaCondition, update.map(visitExpression), srcLocation, precedingWhitespace)
        res.firstSemicolon.precedingWhitespace = node.firstSemicolon.precedingWhitespace
        res.secondSemicolon.precedingWhitespace = node.secondSemicolon.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitForInit(node: ir.ForInit): krakatoa.ForInit = {
    node match {
      case init: ir.ExpressionListForInit => visitExpressionListForInit(init)
      case init: ir.LocalDeclaration => visitLocalDeclaration(init)
      case i =>
        println(s"Error unknown IR ForInit $i")
        null
    }
  }

  def visitExpressionListForInit(node: ir.ExpressionListForInit): krakatoa.ExpressionListForInit = {
    node match {
      case ir.ExpressionListForInit(expressions, srcLocation, precedingWhitespace) =>
        val res = krakatoa.ExpressionListForInit(expressions.map(visitExpression), srcLocation, precedingWhitespace)
        copyWhitespace(node, res)
        res
    }
  }

  def visitForEachLoopControl(node: ir.ForEachLoopControl): krakatoa.ForEachLoopControl = {
    node match {
      case ir.ForEachLoopControl(controlType, variable, control, srcLocation, precedingWhitespace) =>
        val res = krakatoa.ForEachLoopControl(visitTypeIndication(controlType), visitStringNode(variable), visitExpression(control), srcLocation, precedingWhitespace)
        res.modifiers = node.modifiers.map(visitModifier)
        res.colon.precedingWhitespace = node.colon.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitWhileLoop(node: ir.WhileLoop): krakatoa.WhileLoop = {
    node match {
      case ir.WhileLoop(contractBefore, condition, contractWithin, body, srcLocation, precedingWhitespace) =>
        val krakatoaContractBefore = if (contractBefore.isEmpty) Nil else contractBefore.map(visitLoopSpecOrLabel)
        val krakatoaContractWithin = if (contractWithin.isEmpty) Nil else contractWithin.map(visitLoopSpecOrLabel)
        val res = krakatoa.WhileLoop(krakatoaContractBefore, krakatoaContractWithin, visitExpression(condition), visitStatement(body), srcLocation, precedingWhitespace)
        res.whileKeyword.precedingWhitespace = node.whileKeyword.precedingWhitespace
        res.openParenthesis.precedingWhitespace = node.openParenthesis.precedingWhitespace
        res.closeParenthesis.precedingWhitespace = node.closeParenthesis.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitLoopSpecOrLabel(node: ir.LoopSpecOrLabel): krakatoa.LoopSpecOrLabel = {
    node match {
      case l: ir.Label => visitLabel(l)
      case spec: ir.LoopSpecification => visitLoopSpecification(spec)
    }
  }

  def visitLoopSpecification(node: ir.LoopSpecification): krakatoa.LoopSpecification = {
    node match {
      case ir.LoopSpecification(startSpec, clauses, endSpec, srcLocation, precedingWhitespace) =>
        val krakatoaClauses = clauses.map(visitLoopClause)
        val krakatoaStart = startSpec.map(visitStringNode)
        val krakatoaEnd = endSpec.map(visitStringNode)
        val res = krakatoa.LoopSpecification(krakatoaStart, krakatoaClauses, krakatoaEnd, srcLocation, precedingWhitespace)
        copyWhitespace(node, res)
        res
    }
  }

  def visitLoopClause(node: ir.LoopClause): krakatoa.LoopClause = {
    try {
      node match {
        case spec: ir.LoopInvariant => visitLoopInvariantClause(spec)
        case spec: ir.DecreasesClause => visitDecreasesClause(spec)
        case spec: ir.AssignableClause => visitAssignsClause(spec)
        case s =>
          println(s"Error unknown IR LoopClause $s")
          null
      }
    } catch {
      case p: NotSupportedException =>
        var text = "\n" + p.getMessage
        text += node.stringRepresentation.replace("\n", "\n//").replace("@", " ")
        val res = krakatoa.NonTranslatedLoopClause(text, node.srcLocation, node.precedingWhitespace)
        res
    }
  }

  def visitLoopInvariantClause(node: ir.LoopInvariant): krakatoa.LoopInvariantClause = {
    node match {
      case ir.LoopInvariant(expression, srcLocation, precedingWhitespace) =>
        val res = krakatoa.LoopInvariantClause(visitExpression(expression), srcLocation, precedingWhitespace)
        res.invariantKeyword.precedingWhitespace = node.invariantKeyword.precedingWhitespace
        res.semicolon.precedingWhitespace = node.semicolon.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitDecreasesClause(node: ir.DecreasesClause): krakatoa.LoopVariantClause = {
    node match {
      case ir.DecreasesClause(expression, srcLocation, precedingWhitespace) =>
        val res = krakatoa.LoopVariantClause(visitExpression(expression), srcLocation, precedingWhitespace)
        res.decreasesKeyword.precedingWhitespace = node.decreasesKeyword.precedingWhitespace
        res.semicolon.precedingWhitespace = node.semicolon.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitAssignsClause(node: ir.AssignableClause): krakatoa.AssignsClause = {
    node match {
      case ir.AssignableClause(exprs, srcLocation, precedingWhitespace) =>
        val res = krakatoa.AssignsClause(expressions = exprs.map(visitExpression), srcLocation, precedingWhitespace)
        res.assignableKeyword.precedingWhitespace = node.assignableKeyword.precedingWhitespace
        res.semicolon.precedingWhitespace = node.semicolon.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitDoWhileLoop(node: ir.DoLoop): krakatoa.DoWhileLoop = {
    node match {
      case ir.DoLoop(contractBefore, body, condition, srcLocation, precedingWhitespace) =>
        val krakatoaContractBefore = if (contractBefore.isEmpty) Nil else contractBefore.map(visitLoopSpecOrLabel)
        val res = krakatoa.DoWhileLoop(krakatoaContractBefore, visitStatement(body), visitExpression(condition), srcLocation, precedingWhitespace)
        res.doKeyword.precedingWhitespace = node.doKeyword.precedingWhitespace
        res.whileKeyword.precedingWhitespace = node.whileKeyword.precedingWhitespace
        res.openParenthesis.precedingWhitespace = node.openParenthesis.precedingWhitespace
        res.closeParenthesis.precedingWhitespace = node.closeParenthesis.precedingWhitespace
        res.semicolon.precedingWhitespace = node.semicolon.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitTry(node: ir.TryBlock): krakatoa.Try = {
    node match {
      case ir.TryBlock(resourceSpecification, tryBlock, catchClauses, finallyBlock, srcLocation, precedingWhitespace) =>
        val krakatoaResSpec = if (resourceSpecification.isEmpty) None
        else Some(visitTryResourceSpecification(resourceSpecification.get))
        val krakatoaFinally = if (finallyBlock.isEmpty) None else Some(visitBlock(finallyBlock.get))
        val res = krakatoa.Try(visitBlock(tryBlock), krakatoaResSpec, catchClauses.map(visitCatchClause), krakatoaFinally, srcLocation, precedingWhitespace)
        res.tryKeyword.precedingWhitespace = node.tryKeyword.precedingWhitespace
        res.finallyKeyword.precedingWhitespace = node.finallyKeyword.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitTryResourceSpecification(node: ir.TryResourceSpecification): krakatoa.TryResourceSpecification = {
    node match {
      case ir.TryResourceSpecification(resources, trailingSemicolon, srcLocation, precedingWhitespace) =>
        val krakatoaSemi = if (trailingSemicolon.isEmpty) None else Some(visitStringNode(trailingSemicolon.get))
        val res = krakatoa.TryResourceSpecification(resources.map(visitTryResource), krakatoaSemi, srcLocation, precedingWhitespace)
        res.openParenthesis.precedingWhitespace = node.openParenthesis.precedingWhitespace
        res.closeParenthesis.precedingWhitespace = node.closeParenthesis.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitTryResource(node: ir.TryResource): krakatoa.TryResource = {
    node match {
      case ir.TryResource(modifiers, resourceType, variable, expression, srcLocation, precedingWhitespace) =>
        val res = krakatoa.TryResource(modifiers.map(visitModifier), visitTypeIndication(resourceType), visitStringNode(variable), visitExpression(expression), srcLocation, precedingWhitespace)
        res.equalsSign.precedingWhitespace = node.equalsSign.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitCatchClause(node: ir.CatchClause): krakatoa.Catch = {
    node match {
      case ir.CatchClause(types, name, block, srcLocation, precedingWhitespace) =>
        val res = krakatoa.Catch(node.modifiers.map(visitModifier), types.map(visitTypeIndication), visitStringNode(name), visitBlock(block), srcLocation, precedingWhitespace)
        res.catchKeyword.precedingWhitespace = node.catchKeyword.precedingWhitespace
        res.openParenthesis.precedingWhitespace = node.openParenthesis.precedingWhitespace
        res.closeParenthesis.precedingWhitespace = node.closeParenthesis.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitSwitchStatement(node: ir.SwitchStatement): krakatoa.Switch = {
    node match {
      case ir.SwitchStatement(condition, cases, srcLocation, precedingWhitespace) =>
        val res = krakatoa.Switch(visitExpression(condition), cases.map(visitSwitchCaseBlock), srcLocation, precedingWhitespace)
        res.openParenthesis.precedingWhitespace = node.openParenthesis.precedingWhitespace
        res.closeParenthesis.precedingWhitespace = node.closeParenthesis.precedingWhitespace
        res.openBrace.precedingWhitespace = node.openBrace.precedingWhitespace
        res.closeBrace.precedingWhitespace = node.closeBrace.precedingWhitespace
        res.switchKeyword.precedingWhitespace = node.switchKeyword.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitSwitchCaseBlock(node: ir.SwitchCaseBlock): krakatoa.SwitchCaseBlock = {
    node match {
      case ir.SwitchCaseBlock(labels, statements, srcLocation, precedingWhitespace) =>
        val res = krakatoa.SwitchCaseBlock(labels.map(visitSwitchLabel), statements.map(visitStatement), srcLocation, precedingWhitespace)
        copyWhitespace(node, res)
        res
    }
  }

  def visitSwitchLabel(node: ir.SwitchLabel): krakatoa.SwitchLabel = {
    node match {
      case ir.SwitchLabel(expression, srcLocation, precedingWhitespace) =>
        val krakatoaExpr = if (expression.isEmpty) None else Some(visitExpression(expression.get))
        val res = krakatoa.SwitchLabel(krakatoaExpr, srcLocation, precedingWhitespace)
        res.caseKeyword.precedingWhitespace = node.caseKeyword.precedingWhitespace
        res.defaultKeyword.precedingWhitespace = node.defaultKeyword.precedingWhitespace
        res.colon.precedingWhitespace = node.colon.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitSynchronizedStatement(node: ir.SynchronizedStatement): krakatoa.Synchronized = {
    node match {
      case ir.SynchronizedStatement(condition, body, srcLocation, precedingWhitespace) =>
        val res = krakatoa.Synchronized(visitExpression(condition), visitStatement(body), srcLocation, precedingWhitespace)
        res.synchronizedKeyword.precedingWhitespace = node.synchronizedKeyword.precedingWhitespace
        res.openParenthesis.precedingWhitespace = node.openParenthesis.precedingWhitespace
        res.closeParenthesis.precedingWhitespace = node.closeParenthesis.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitReturnStatement(node: ir.ReturnStatement): krakatoa.Return = {
    node match {
      case ir.ReturnStatement(expression, srcLocation, precedingWhitespace) =>
        val krakatoaExpr = if (expression.isEmpty) None else Some(visitExpression(expression.get))
        val res = krakatoa.Return(krakatoaExpr, srcLocation, precedingWhitespace)
        res.returnKeyword.precedingWhitespace = node.returnKeyword.precedingWhitespace
        res.semicolon.precedingWhitespace = node.semicolon.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitThrowStatement(node: ir.ThrowStatement): krakatoa.ThrowStatement = {
    node match {
      case ir.ThrowStatement(expression, srcLocation, precedingWhitespace) =>
        val res = krakatoa.ThrowStatement(visitExpression(expression), srcLocation, precedingWhitespace)
        res.throwKeyword.precedingWhitespace = node.throwKeyword.precedingWhitespace
        res.semicolon.precedingWhitespace = node.semicolon.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitBreakStatement(node: ir.BreakStatement): krakatoa.Break = {
    node match {
      case ir.BreakStatement(label, srcLocation, precedingWhitespace) =>
        val krakatoaLabel = if (label.isEmpty) None else Some(visitStringNode(label.get))
        val res = krakatoa.Break(krakatoaLabel, srcLocation, precedingWhitespace)
        res.breakKeyword.precedingWhitespace = node.breakKeyword.precedingWhitespace
        res.semicolon.precedingWhitespace = node.semicolon.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitContinueStatement(node: ir.ContinueStatement): krakatoa.Continue = {
    node match {
      case ir.ContinueStatement(label, srcLocation, precedingWhitespace) =>
        val krakatoaLabel = if (label.isEmpty) None else Some(visitStringNode(label.get))
        val res = krakatoa.Continue(krakatoaLabel, srcLocation, precedingWhitespace)
        res.continueKeyword.precedingWhitespace = node.continueKeyword.precedingWhitespace
        res.semicolon.precedingWhitespace = node.semicolon.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitExpressionStatement(node: ir.ExpressionStatement): krakatoa.ExpressionStatement = {
    node match {
      case ir.ExpressionStatement(expression, srcLocation, precedingWhitespace) =>
        val res = krakatoa.ExpressionStatement(visitExpression(expression), srcLocation, precedingWhitespace)
        res.semicolon.precedingWhitespace = node.semicolon.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitLabeledStatement(node: ir.LabeledStatement): krakatoa.LabeledStatement = {
    node match {
      case ir.LabeledStatement(label, statement, srcLocation, precedingWhitespace) =>
        val res = krakatoa.LabeledStatement(visitLabel(label), visitStatement(statement), None, srcLocation, precedingWhitespace)
        copyWhitespace(node, res)
        res
    }
  }

  def visitLabel(node: ir.Label): krakatoa.Label = {
    node match {
      case ir.Label(label, srcLocation, precedingWhitespace) =>
        val res = krakatoa.Label(visitStringNode(label), srcLocation, precedingWhitespace)
        res.colon.precedingWhitespace = node.colon.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitAssertStatement(node: ir.AssertStatement): krakatoa.AssertStatement = {
    node match {
      case ir.AssertStatement(content, message, srcLocation, precedingWhitespace) =>
        val krakatoaMsg = if (message.isEmpty) None else Some(visitExpression(message.get))
        val res = krakatoa.AssertStatement(visitExpression(content), krakatoaMsg, srcLocation, precedingWhitespace)
        res.assertKeyword.precedingWhitespace = node.assertKeyword.precedingWhitespace
        res.colon.precedingWhitespace = node.colon.precedingWhitespace
        res.semicolon.precedingWhitespace = node.semicolon.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitGhostStatement(node: ir.GhostStatement): krakatoa.Statement = {
    node match {
      case ir.GhostStatement(statement, srcLocation, precedingWhitespace) =>
        krakatoa.GhostStatement(visitStatement(statement), srcLocation, precedingWhitespace)
    }
  }

  def visitAssert(node: ir.JavaAssertion): krakatoa.Assert = {
    krakatoa.Assert(visitExpression(node.content), node.message.map(visitExpression), node.srcLocation, node.precedingWhitespace)
  }

}
