package languages.Krakatoa

import generated.{KrakatoaParser, KrakatoaParserBaseVisitor}
import languages.Krakatoa.syntaxTree.*
import generic.{NotSupportedException, SourceLocation}
import org.antlr.v4.runtime.tree.TerminalNode
import org.antlr.v4.runtime.{CommonTokenStream, ParserRuleContext, Token}

import scala.jdk.CollectionConverters.*

/**
 * Generates a Krakatoa-specific AST based on the parse tree
 * @param tokens
 * @param filename
 */
class KrakatoaVisitor(val tokens: CommonTokenStream, val filename: String) extends KrakatoaParserBaseVisitor[KrakatoaTreeNode] {
  def setNonTranslatedTreeNodeText[T <: generic.syntaxTree.StringNode](node: T,
                                                                       ctx: ParserRuleContext,
                                                                       exc: NotSupportedException): Unit = {
    val ws = getWhitespaceLeft(ctx)
    val asString = new ParserPrintVisitor(tokens, skipFirstHidden = true).visit(ctx)
    node.precedingWhitespace = ws
    node.text = s"// Warning: ${exc.construct} not supported by the Specification Translator. Respective parts commented out." + System.lineSeparator()
      + "// "
    node.text += asString.replace("\n", "\n//").replace("@", " ")
    System.err.println(exc.getMessage)
  }
  def getWhitespaceRight(ctx: ParserRuleContext): String = {
    val hiddenTokens = tokens.getHiddenTokensToRight(ctx.start.getTokenIndex)
    if (hiddenTokens == null) "" else removeAtSignFromComments(hiddenTokens.asScala.map(wsToString).mkString)
  }

  // Some multi-line comments contain @-signs later on, it can be confusing whether things after the @ should be
  // interpreted as specifications. This method will remove these @-signs to avoid confusion
  private def removeAtSignFromComments(string: String): String = {
    val pattern = """(\/\*[^@](.*?\r?\n)+\s+)@(.*)""".r
    pattern.replaceAllIn(string, m => removeAtSignFromComments(m.group(1) + m.group(3)))
  }

  def getWhitespaceRight(node: TerminalNode): String = {
    val hiddenTokens = tokens.getHiddenTokensToRight(node.getSymbol.getTokenIndex)
    if (hiddenTokens == null) "" else removeAtSignFromComments(hiddenTokens.asScala.map(_.getText).mkString)
  }

  def getWhitespaceRight(token: Token): String = {
    val hiddenTokens = tokens.getHiddenTokensToRight(token.getTokenIndex)
    if (hiddenTokens == null) "" else removeAtSignFromComments(hiddenTokens.asScala.map(_.getText).mkString)
  }

  def getWhitespaceLeft(token: Token): String = {
    val hiddenTokens = tokens.getHiddenTokensToLeft(token.getTokenIndex)
    if (hiddenTokens == null) "" else removeAtSignFromComments(hiddenTokens.asScala.map(_.getText).mkString)
  }
  
  def getWhitespaceLeft(node: TerminalNode): String = {
    val hiddenTokens = tokens.getHiddenTokensToLeft(node.getSymbol.getTokenIndex)
    if (hiddenTokens == null) "" else removeAtSignFromComments(hiddenTokens.asScala.map(_.getText).mkString)
  }

  def getWhitespaceLeft(ctx: ParserRuleContext): String = {
    val hiddenTokens = tokens.getHiddenTokensToLeft(ctx.start.getTokenIndex)
    if (hiddenTokens == null) "" else removeAtSignFromComments(hiddenTokens.asScala.map(wsToString).mkString)
  }

  def addWhitespaceRight[T <: KrakatoaTreeNode](node: T, ctx: ParserRuleContext): T = {
    node.trailingWhitespace = getWhitespaceRight(ctx)
    node
  }

  def addWhitespaceRight[T <: KrakatoaTreeNode](node: T, terminal: TerminalNode): T = {
    node.trailingWhitespace = getWhitespaceRight(terminal)
    node
  }

  def wsToString(token: Token): String = {
    token.getText
  }

  def printNode(ctx: ParserRuleContext): StringNode = {
    val src = getSourceLocation(ctx)
    val text = new ParserPrintVisitor(tokens).visit(ctx)
    StringNode(text, src)
  }

  def printNodeWithWS(ctx: ParserRuleContext): StringNode = {
    val src = getSourceLocation(ctx)
    val text = new ParserPrintVisitor(tokens).visit(ctx)
    StringNode(text, src, pTrailingWhitespace = getWhitespaceRight(ctx))
  }

  private def getSourceLocation(ctx: ParserRuleContext): SourceLocation = {
    SourceLocation(filename, ctx.start.getLine, ctx.stop.getLine)
  }
  
  private def getSourceLocation(node: TerminalNode): SourceLocation = {
    SourceLocation(filename, node.getSymbol.getLine)
  }

  private def getSourceLocation(token: Token) : SourceLocation = {
    SourceLocation(filename, token.getLine)
  }

  override def visitQualifiedName(ctx: KrakatoaParser.QualifiedNameContext): StringNode = printNode(ctx)

  override def visitCompilationUnit(ctx: KrakatoaParser.CompilationUnitContext): CompilationUnit = {
    val src = SourceLocation(filename, ctx.start.getLine, ctx.stop.getLine)
    var pkgDecl: Option[PackageDeclaration] = None
    if (ctx.packageDeclaration() != null) {
      pkgDecl = Some(visitPackageDeclaration(ctx.packageDeclaration()))
    }
    val imports = ctx.importDeclaration().asScala.map(visitImportDeclaration).toList
    val declarations = ctx.typeDeclaration().asScala.map(visitTypeDeclaration).toList
    val res = CompilationUnit(pkgDecl, imports, declarations, src)
    res.trailingWhitespace = getWhitespaceLeft(ctx.stop)
    res
  }

  override def visitPackageDeclaration(ctx: KrakatoaParser.PackageDeclarationContext): PackageDeclaration = {
    val src = getSourceLocation(ctx)
    val name = visitQualifiedName(ctx.qualifiedName())
    val annotations = ctx.annotation().asScala.map(visitAnnotation).toList
    val pkg = PackageDeclaration(name, annotations, src)
    pkg.packageKeyword.precedingWhitespace = getWhitespaceLeft(ctx.PACKAGE())
    pkg.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
    pkg
  }

  override def visitImportDeclaration(ctx: KrakatoaParser.ImportDeclarationContext): ImportDeclaration = {
    val name = visitQualifiedName(ctx.qualifiedName())
    if (ctx.DOT() != null) {
      name.text += getWhitespaceLeft(ctx.DOT()) + ctx.DOT().getText
      name.text += getWhitespaceLeft(ctx.MUL()) + ctx.MUL().getText
    }
    var modifiers: List[Modifier] = Nil
    if (ctx.STATIC() != null) {
      val static = StaticModifier(getSourceLocation(ctx.STATIC()), getWhitespaceLeft(ctx.STATIC()))
      modifiers = modifiers.appended(static)
    }
    val imp = ImportDeclaration(name, getSourceLocation(ctx))
    imp.modifiers = modifiers
    imp.importKeyword.precedingWhitespace = getWhitespaceLeft(ctx.IMPORT())
    imp.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
    imp
  }

  override def visitTypeDeclaration(ctx: KrakatoaParser.TypeDeclarationContext): TypeDeclaration = {
    var res: TypeDeclaration = null
    try {
      if (ctx.SEMI() != null) {
        res = new EmptyStatement(getSourceLocation(ctx.SEMI()), getWhitespaceLeft(ctx.SEMI()))
      } else if (ctx.krakatoaOption() != null) {
        //comment option since it cannot be translated
        val commented = NonTranslatedTypeDeclaration("", getSourceLocation(ctx.krakatoaOption()))
        val exc = NotSupportedException("KrakatoaOption", getSourceLocation(ctx.krakatoaOption()))
        val ws = getWhitespaceLeft(ctx.krakatoaOption())
        val asString = ctx.krakatoaOption().LineStartOptionSpec().getText + ctx.krakatoaOption().KRAKATOA_PROPERTY().getText + ctx.krakatoaOption().ASSIGN().getText + ctx.krakatoaOption().langId().getText + ctx.krakatoaOption().EndSpec().getText
        commented.precedingWhitespace = ws
        commented.trailingWhitespace = getWhitespaceRight(ctx.krakatoaOption())
        commented.text = s" Warning: ${exc.construct} not supported by the Specification Translator. Respective parts commented out." + System.lineSeparator()
          + "// "
        commented.text += asString.replace("\n", "\n//").replace("@", " ")
        commented.text = commented.text.replace("// //", "//")
        commented.text = commented.text.replace("///*", "/*")
        commented.text += "\n"
        System.err.println(exc)
        res = commented
      } else {
        val modifiers = ctx.classOrInterfaceModifier().asScala.map(visitClassOrInterfaceModifier).toList
        if (ctx.classDeclaration() != null) {
          val cls = visitClassDeclaration(ctx.classDeclaration())
          cls.modifiers = modifiers
          cls.srcLocation = getSourceLocation(ctx)
          res = cls
        } else if (ctx.enumDeclaration() != null) {
          val enumDecl = visitEnumDeclaration(ctx.enumDeclaration())
          enumDecl.modifiers = modifiers
          enumDecl.srcLocation = getSourceLocation(ctx)
          res = enumDecl
        } else if (ctx.interfaceDeclaration() != null) {
          val interface = visitInterfaceDeclaration(ctx.interfaceDeclaration())
          interface.modifiers = modifiers
          interface.srcLocation = getSourceLocation(ctx)
          res = interface
        } else if (ctx.annotationTypeDeclaration() != null) {
          val annotationType = visitAnnotationTypeDeclaration(ctx.annotationTypeDeclaration())
          annotationType.modifiers = modifiers
          annotationType.srcLocation = getSourceLocation(ctx)
          res = annotationType
        } else if (ctx.valEmbedDeclarationBlock() != null) {
          val specs = visitValEmbedDeclarationBlock(ctx.valEmbedDeclarationBlock())
          specs.srcLocation = getSourceLocation(ctx)
          res = specs
        } else {
          println(s"Unknown type declaration: ${ctx.getText}")
        }
      }
    } catch {
      case e: NotSupportedException =>
        val decl = NonTranslatedTypeDeclaration("", getSourceLocation(ctx))
        setNonTranslatedTreeNodeText(decl, ctx, e)
        res = decl
    }
    res
  }

  override def visitAnnotation(ctx: KrakatoaParser.AnnotationContext): StringAnnotation = {
    val src = getSourceLocation(ctx)
    val text = new ParserPrintVisitor(tokens).visit(ctx)
    StringAnnotation(text, src)
  }

  override def visitClassDeclaration(ctx: KrakatoaParser.ClassDeclarationContext): ClassDeclaration = {
    val name = visitJavaIdentifier(ctx.javaIdentifier())
    var sup: Option[SuperclassClause] = None
    if (ctx.EXTENDS() != null) {
      val ext = SuperclassClause(visitTypeType(ctx.typeType()), getSourceLocation(ctx.typeType()))
      ext.extendsKeyword.precedingWhitespace = getWhitespaceLeft(ctx.EXTENDS())
      sup = Option(ext)
    }
    var interfaces: Option[InterfaceClause] = None
    if (ctx.IMPLEMENTS() != null) {
      val int = InterfaceClause(visitTypeListAsList(ctx.typeList()), getSourceLocation(ctx.typeList()))
      int.implementsKeyword.precedingWhitespace = getWhitespaceLeft(ctx.IMPLEMENTS())
      interfaces = Option(int)
    }
    val typeParameterClause = if (ctx.typeParameters() != null)
                                Option(visitTypeParameters(ctx.typeParameters()))
                              else None
    val body = visitClassBody(ctx.classBody())
    val res = ClassDeclaration(name, typeParameterClause, sup, interfaces, body, getSourceLocation(ctx))
    res.classKeyword.precedingWhitespace = getWhitespaceLeft(ctx.CLASS())
    res
  }

  override def visitEnumDeclaration(ctx: KrakatoaParser.EnumDeclarationContext): EnumDeclaration = {
    val name = visitJavaIdentifier(ctx.javaIdentifier())
    var interfaces: Option[InterfaceClause] = None
    if (ctx.IMPLEMENTS() != null) {
      val int = InterfaceClause(visitTypeListAsList(ctx.typeList()), getSourceLocation(ctx.typeList()))
      int.implementsKeyword.precedingWhitespace = getWhitespaceLeft(ctx.IMPLEMENTS())
      interfaces = Option(int)
    }
    var constants: List[EnumConstant] = Nil
    if (ctx.enumConstants() != null) {
      constants = ctx.enumConstants().enumConstant().asScala.map(x=>addWhitespaceRight(visitEnumConstant(x),x)).toList
    }
    var comma: Option[StringNode] = None
    if (ctx.COMMA() != null) {
      comma = Option(StringNode(",", getSourceLocation(ctx.COMMA())))
      comma.get.precedingWhitespace = getWhitespaceLeft(ctx.COMMA())
    }
    var decls: List[MemberDeclaration] = Nil
    var semi: Option[StringNode] = None
    if (ctx.enumBodyDeclarations() != null) {
      semi = Option(StringNode(";", getSourceLocation(ctx.enumBodyDeclarations().SEMI())))
      semi.get.precedingWhitespace = getWhitespaceLeft(ctx.enumBodyDeclarations().SEMI())
      decls = ctx.enumBodyDeclarations().classBodyDeclaration().asScala.map(visitClassBodyDeclaration).toList
    }
    val res = EnumDeclaration(name, interfaces, constants, decls, comma, semi, getSourceLocation(ctx))
    res.enumKeyword.precedingWhitespace = getWhitespaceLeft(ctx.ENUM())
    res.openBrace.precedingWhitespace = getWhitespaceLeft(ctx.LBRACE())
    res.closeBrace.precedingWhitespace = getWhitespaceLeft(ctx.RBRACE())
    res
  }

  override def visitEnumConstant(ctx: KrakatoaParser.EnumConstantContext): EnumConstant = {
    val annotations = ctx.annotation().asScala.map(visitAnnotation).toList
    val name = visitJavaIdentifier(ctx.javaIdentifier())
    val args = if (ctx.arguments() != null) Option(visitArguments(ctx.arguments())) else None
    val body = if (ctx.classBody() != null) Option(visitClassBody(ctx.classBody())) else None
    val constant = EnumConstant(name, annotations, args, body, getSourceLocation(ctx))
    constant
  }

  override def visitArguments(ctx: KrakatoaParser.ArgumentsContext): Arguments = {
    val args = if (ctx.expressionList() == null) Nil else ctx.expressionList().expression().asScala.map(x=>addWhitespaceRight(visitExpression(x),x)).toList
    val arguments = Arguments(args, getSourceLocation(ctx))
    arguments.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.LPAREN())
    arguments.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.RPAREN())
    arguments
  }

  override def visitInterfaceDeclaration(ctx: KrakatoaParser.InterfaceDeclarationContext): InterfaceDeclaration = {
    val name = visitJavaIdentifier(ctx.javaIdentifier())
    var sup: Option[MultiSuperclassClause] = None
    if (ctx.EXTENDS() != null) {
      val ext = MultiSuperclassClause(visitTypeListAsList(ctx.typeList()), getSourceLocation(ctx.typeList()))
      ext.extendsKeyword.precedingWhitespace = getWhitespaceLeft(ctx.EXTENDS())
      sup = Option(ext)
    }
    val typeParameterClause = if (ctx.typeParameters() != null) Option(visitTypeParameters(ctx.typeParameters())) else None
    val body = visitInterfaceBody(ctx.interfaceBody())
    val res = InterfaceDeclaration(name, typeParameterClause, sup, body, getSourceLocation(ctx))
    res.interfaceKeyword.precedingWhitespace = getWhitespaceLeft(ctx.INTERFACE())
    res
  }

  override def visitInterfaceBody(ctx: KrakatoaParser.InterfaceBodyContext) : Body = {
    val decls = ctx.interfaceBodyDeclaration().asScala.map(visitInterfaceBodyDeclaration).toList
    val body = Body(decls, getSourceLocation(ctx))
    body.openBrace.precedingWhitespace = getWhitespaceLeft(ctx.LBRACE())
    body.closeBrace.precedingWhitespace = getWhitespaceLeft(ctx.RBRACE())
    body
  }

  override def visitInterfaceBodyDeclaration(ctx: KrakatoaParser.InterfaceBodyDeclarationContext): MemberDeclaration = {
    ctx.getAltNumber match {
      case 1 =>
        val decl = visitInterfaceMemberDeclaration(ctx.interfaceMemberDeclaration())
        val modifiers = ctx.modifier().asScala.map(visitModifier)
        decl.modifiers = decl.modifiers.prependedAll(modifiers)
        decl.srcLocation = getSourceLocation(ctx)
        decl
      case 2 =>
        new EmptyStatement(getSourceLocation(ctx), getWhitespaceLeft(ctx))
      case x =>
        println(s"Error: unexpected AltNum in InterfaceBodyDeclaration: $x")
        null
    }
  }

  override def visitInterfaceMemberDeclaration(ctx: KrakatoaParser.InterfaceMemberDeclarationContext): MemberDeclaration = {
    ctx.getAltNumber match {
      case 1 => visitConstDeclaration(ctx.constDeclaration())
      case 2 => visitInterfaceMethodDeclaration(ctx.interfaceMethodDeclaration())
      case 3 => visitGenericInterfaceMethodDeclaration(ctx.genericInterfaceMethodDeclaration())
      case 4 => visitInterfaceDeclaration(ctx.interfaceDeclaration())
      case 5 => visitAnnotationTypeDeclaration(ctx.annotationTypeDeclaration())
      case 6 => visitClassDeclaration(ctx.classDeclaration())
      case 7 => visitEnumDeclaration(ctx.enumDeclaration())
      case x =>
        println(s"Error: unexpected AltNum in InterfaceMemberDeclaration: $x")
        null
    }
  }

  override def visitConstDeclaration(ctx: KrakatoaParser.ConstDeclarationContext): FieldDeclaration = {
    val tType = visitTypeType(ctx.typeType())
    val decls = ctx.constantDeclarator().asScala.map(x=>addWhitespaceRight(visitConstantDeclarator(x), x)).toList
    val field = FieldDeclaration(tType, decls, getSourceLocation(ctx))
    field.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
    field
  }

  override def visitConstantDeclarator(ctx: KrakatoaParser.ConstantDeclaratorContext): VariableDeclarator = {
    val name = visitJavaIdentifier(ctx.javaIdentifier())
    for (i <- 0 until ctx.LBRACK().size()) {
      name.text += getWhitespaceLeft(ctx.LBRACK(i)) + ctx.LBRACK(i).getText
      name.text += getWhitespaceLeft(ctx.RBRACK(i)) + ctx.RBRACK(i).getText
    }
    val initializer = Option(visitVariableInitializer(ctx.variableInitializer()))
    val decl = VariableDeclarator(name, initializer, getSourceLocation(ctx))
    decl.equalSign.precedingWhitespace = getWhitespaceLeft(ctx.ASSIGN())
    decl
  }

  override def visitInterfaceMethodDeclaration(ctx: KrakatoaParser.InterfaceMethodDeclarationContext): MethodDeclaration = {
    val decl = new MethodDeclaration(null, None, Nil, None, None, None, None, null, getSourceLocation(ctx))
    setInterfaceMethodDeclarationFields(ctx, decl)
    decl
  }

  def setInterfaceMethodDeclarationFields(ctx: KrakatoaParser.InterfaceMethodDeclarationContext, decl: MethodDeclaration): Unit = {
    decl.name = visitJavaIdentifier(ctx.javaIdentifier())
    decl.returnType = None
    if (ctx.typeParameters() != null) {
      val typeParameters = visitTypeParameters(ctx.typeParameters())
      val annotations = ctx.annotation().asScala.map(visitAnnotation).toList
      val text = new ParserPrintVisitor(tokens).visit(ctx.typeTypeOrVoid())
      val typeIndication = new StringGenericTypeIndication(text, typeParameters, annotations, ctx.getText, getSourceLocation(ctx))
      decl.returnType = Option(typeIndication)
    } else {
      decl.returnType = Option(visitTypeTypeOrVoid(ctx.typeTypeOrVoid()))
    }
    decl.parameters = if (ctx.formalParameters().formalParameterList() != null) visitFormalParameterListAsList(ctx.formalParameters().formalParameterList()) else Nil
    var dimensions = ""
    for (x <- 0 until ctx.LBRACK().size()) {
      dimensions += getWhitespaceLeft(ctx.LBRACK(x)) + ctx.LBRACK(x).getText
      dimensions += getWhitespaceLeft(ctx.RBRACK(x)) + ctx.RBRACK(x).getText
    }
    decl.dimensions = if (dimensions=="") None
    else Option(StringNode(dimensions, getSourceLocation(ctx.LBRACK(0)))) // todo: better src?
    decl.throws = None
    if (ctx.THROWS() != null) {
      decl.throws = Option(Throws(visitQualifiedNameListAsList(ctx.qualifiedNameList()), getSourceLocation(ctx.qualifiedNameList())))
      decl.throws.get.throwsKeyword.precedingWhitespace = getWhitespaceLeft(ctx.THROWS())
    }
    decl.body = visitMethodBody(ctx.methodBody())
    decl.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.formalParameters().LPAREN())
    decl.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.formalParameters().RPAREN())
  }

  override def visitGenericInterfaceMethodDeclaration(ctx: KrakatoaParser.GenericInterfaceMethodDeclarationContext): GenericMethodDeclaration = {
    val typeParams = visitTypeParameters(ctx.typeParameters())
    val decl = GenericMethodDeclaration(typeParams, null, None, Nil, None, None, None, None, null, getSourceLocation(ctx))
    setInterfaceMethodDeclarationFields(ctx.interfaceMethodDeclaration(), decl)
    decl
  }

  override def visitAnnotationTypeDeclaration(ctx: KrakatoaParser.AnnotationTypeDeclarationContext): AnnotationTypeDeclaration = {
    val name = visitJavaIdentifier(ctx.javaIdentifier())
    val decls = ctx.annotationTypeBody().annotationTypeElementDeclaration().asScala.map(visitAnnotationTypeElementDeclaration).toList
    val body = Body(decls, getSourceLocation(ctx.annotationTypeBody()))
    AnnotationTypeDeclaration(name, body, getSourceLocation(ctx))
  }

  override def visitAnnotationTypeElementDeclaration(ctx: KrakatoaParser.AnnotationTypeElementDeclarationContext): MemberDeclaration = {
    ctx.getAltNumber match {
      case 1 =>
        val elems = visitAnnotationTypeElementRest(ctx.annotationTypeElementRest())
        val modifiers = ctx.modifier().asScala.map(visitModifier)
        elems.modifiers = elems.modifiers.prependedAll(modifiers)
        elems.srcLocation = getSourceLocation(ctx)
        elems
      case 2 =>
        EmptyStatement(getSourceLocation(ctx), getWhitespaceLeft(ctx))
      case x =>
        println(s"Error: unexpected AltNum in AnnotationTypeElementDeclaration: $x")
        null
    }
  }

  //TODO: Haven't completely finished this yet
  override def visitAnnotationTypeElementRest(ctx: KrakatoaParser.AnnotationTypeElementRestContext): MemberDeclaration = {
    ctx.getAltNumber match {
      case 1 =>
        val tType = visitTypeType(ctx.typeType())
        if (ctx.annotationMethodOrConstantRest().annotationConstantRest() != null) {
          val decl = visitVariableDeclaratorsAsList(ctx.annotationMethodOrConstantRest().annotationConstantRest().variableDeclarators())
          val semi = StringNode(";", getSourceLocation(ctx), getWhitespaceLeft(ctx.SEMI()))
          FieldDeclaration(tType, decl, getSourceLocation(ctx))
        } else if (ctx.annotationMethodOrConstantRest().annotationMethodRest() != null) {
          val name = visitJavaIdentifier(ctx.annotationMethodOrConstantRest().annotationMethodRest().javaIdentifier())
          val defaultValue = StringNode(ctx.annotationMethodOrConstantRest().annotationMethodRest().defaultValue().elementValue().getText, getSourceLocation(ctx.annotationMethodOrConstantRest().annotationMethodRest().defaultValue().elementValue()))
          val defaultClause = Some(DefaultClause(defaultValue, getSourceLocation(ctx.annotationMethodOrConstantRest().annotationMethodRest().defaultValue())))
          GenericMethodDeclaration(typeParameters=null, pName=name, pReturnType=Some(tType), pParameters=Nil, pDimensions=None, pThrows=None, pDefault=defaultClause, pContractWithin = None, pBody=null, getSourceLocation(ctx))
        } else {
          println(s"Error: unexpected thing in AnnotationMethodOrConstantRest")
          null
        }
      case 2 =>
        // TODO: ignoring optional ';' right now for all other cases
        visitClassDeclaration(ctx.classDeclaration())
      case 3 =>
        visitInterfaceDeclaration(ctx.interfaceDeclaration())
      case 4 =>
        visitEnumDeclaration(ctx.enumDeclaration())
      case 5 =>
        visitAnnotationTypeDeclaration(ctx.annotationTypeDeclaration())
      case x =>
        println(s"Error: unexpected AltNum in AnnotationTypeElementRest: $x")
        null
    }
  }

  override def visitJavaIdentifier(ctx: KrakatoaParser.JavaIdentifierContext): StringNode = {
    visitTextNode(ctx)
  }

  def visitTextNode(ctx: ParserRuleContext): StringNode = {
    val src = getSourceLocation(ctx)
    StringNode(ctx.getText, src, getWhitespaceLeft(ctx))
  }

  override def visitTypeType(ctx: KrakatoaParser.TypeTypeContext): StringTypeIndication = {
    val text = new ParserPrintVisitor(tokens).visit(ctx)
    StringTypeIndication(removeAtSignFromComments(text), Nil, ctx.getText, getSourceLocation(ctx))
  }

  def visitTypeTypeWithWS(ctx: KrakatoaParser.TypeTypeContext): StringTypeIndication = {
    val text = new ParserPrintVisitor(tokens).visit(ctx)
    StringTypeIndication(text, Nil, ctx.getText, getSourceLocation(ctx), pTrailingWhitespace = getWhitespaceRight(ctx))
  }

  def visitTypeListAsList(ctx: KrakatoaParser.TypeListContext): List[TypeIndication] = {
    val types = ctx.typeType().asScala.map(visitTypeTypeWithWS).toList
    types.last.trailingWhitespace = ""
    types
  }

  override def visitTypeParameters(ctx: KrakatoaParser.TypeParametersContext): TypeParameterClause = {
    val parameters = ctx.typeParameter().asScala.map(x=>addWhitespaceRight(visitTypeParameter(x), x)).toList
    val clause = TypeParameterClause(parameters, getSourceLocation(ctx))
    clause.openAngle.precedingWhitespace = getWhitespaceLeft(ctx.LT())
    clause.closeAngle.precedingWhitespace = getWhitespaceLeft(ctx.GT())
    parameters.last.trailingWhitespace = ""
    clause
  }

  override def visitTypeParameter(ctx: KrakatoaParser.TypeParameterContext): TypeParameter = {
    val name = visitJavaIdentifier(ctx.javaIdentifier())
    val typeName = new StringTypeIndication(name.text, Nil, ctx.javaIdentifier().getText, name.srcLocation)
    var annotations: List[StringAnnotation] = Nil
    var superclass: Option[GenericsSuperclassClause] = None
    if (ctx.EXTENDS() != null) {
      val sup = GenericsSuperclassClause(visitTypeBoundAsList(ctx.typeBound()), getSourceLocation(ctx.typeBound()))
      sup.extendsKeyword.precedingWhitespace = getWhitespaceLeft(ctx.EXTENDS())
      superclass = Option(sup)
      val typeAnnotations = ctx.annotation().asScala.filter(_.start.getTokenIndex < ctx.EXTENDS().getSymbol.getTokenIndex)
      annotations = typeAnnotations.map(visitAnnotation).toList
      val boundAnnotations = ctx.annotation().asScala.filter(_.start.getTokenIndex > ctx.EXTENDS().getSymbol.getTokenIndex)
      // add annotations before "typeBound" to first "typeType"
      sup.superclasses.head.annotations = sup.superclasses.head.annotations.prependedAll(boundAnnotations.map(visitAnnotation).toList)
    } else {
      annotations = ctx.annotation().asScala.map(visitAnnotation).toList
    }
    val param = TypeParameter(annotations, typeName, superclass, getSourceLocation(ctx))
    param
  }

  def visitTypeBoundAsList(ctx: KrakatoaParser.TypeBoundContext): List[TypeIndication] = {
    val types = ctx.typeType().asScala.map(visitTypeTypeWithWS).toList
    types.last.trailingWhitespace = ""
    types
  }

  override def visitClassBody(ctx: KrakatoaParser.ClassBodyContext): Body = {
    val decls = ctx.classBodyDeclaration().asScala.map(visitClassBodyDeclaration).toList
    val body = Body(decls, getSourceLocation(ctx))
    body.openBrace.precedingWhitespace = getWhitespaceLeft(ctx.LBRACE())
    body.closeBrace.precedingWhitespace = getWhitespaceLeft(ctx.RBRACE())
    body
  }

  override def visitClassBodyDeclaration(ctx: KrakatoaParser.ClassBodyDeclarationContext): MemberDeclaration = {
    var decl: MemberDeclaration = null
    ctx.getAltNumber match {
      case 1 =>
        decl = EmptyStatement(getSourceLocation(ctx.SEMI()), getWhitespaceLeft(ctx.SEMI()))
      case 2 =>
        decl = visitBlock(ctx.block())
        if (ctx.STATIC() != null) {
          val static = StaticModifier(getSourceLocation(ctx.STATIC()), getWhitespaceLeft(ctx.STATIC()))
          decl.modifiers = decl.modifiers.prepended(static)
        }
      case 3 =>
        decl = visitMemberDeclaration(ctx.memberDeclaration())
        decl.contractBefore = if (ctx.valEmbedContract() != null) Option(visitValEmbedContract(ctx.valEmbedContract())) else None
        decl.modifiers = decl.modifiers.prependedAll(ctx.modifier().asScala.map(visitModifier))
      case 4 =>
        decl = visitValEmbedDeclarationBlock(ctx.valEmbedDeclarationBlock())
      case 5 =>
        decl = visitValDeclaration(ctx.valDeclaration())
      case x =>
        println(s"Error: unexpected AltNum in ClassBodyDeclaration: $x")
    }
    decl.srcLocation = getSourceLocation(ctx)
    decl
  }

  override def visitModifier(ctx: KrakatoaParser.ModifierContext): Modifier = {
    ctx.getAltNumber match {
      case 1 => visitClassOrInterfaceModifier(ctx.classOrInterfaceModifier())
      case 6 => visitValEmbedModifiers(ctx.valEmbedModifiers())
      case _ => new ModifierStringNode(ctx.getText, getSourceLocation(ctx), getWhitespaceLeft(ctx))
    }
  }

  override def visitValEmbedModifiers(ctx: KrakatoaParser.ValEmbedModifiersContext): Modifier = {
    val modifiers = ctx.valModifier().asScala.map(visitValModifier).toList
    var startSpec: Option[StringNode] = None
    var endSpec: Option[StringNode] = None
    if (ctx.startSpec() != null) {
      startSpec = Some(StringNode(ctx.startSpec().getText, getSourceLocation(ctx.startSpec()), getWhitespaceLeft(ctx.startSpec())))
      endSpec = Some(StringNode(ctx.endSpec.getText, getSourceLocation(ctx.endSpec()), getWhitespaceLeft(ctx.endSpec())))
    }
    ModifierBlock(startSpec, modifiers, endSpec, getSourceLocation(ctx))
  }

  override def visitValModifier(ctx: KrakatoaParser.ValModifierContext): Modifier = {
    if (ctx.VAL_PURE() != null) {
      PureModifier(getSourceLocation(ctx), getWhitespaceLeft(ctx))
    } else {
      visitClassOrInterfaceModifier(ctx.langModifier().classOrInterfaceModifier())
    }
  }

  override def visitMemberDeclaration(ctx: KrakatoaParser.MemberDeclarationContext): MemberDeclaration = {
    try {
      ctx.getAltNumber match {
        case 1 => visitMethodDeclaration(ctx.methodDeclaration())
        case 2 => visitGenericMethodDeclaration(ctx.genericMethodDeclaration())
        case 3 => visitFieldDeclaration(ctx.fieldDeclaration())
        case 4 => visitConstructorDeclaration(ctx.constructorDeclaration())
        case 5 => visitGenericConstructorDeclaration(ctx.genericConstructorDeclaration())
        case 6 => visitInterfaceDeclaration(ctx.interfaceDeclaration())
        case 7 => visitAnnotationTypeDeclaration(ctx.annotationTypeDeclaration())
        case 8 => visitClassDeclaration(ctx.classDeclaration())
        case 9 => visitEnumDeclaration(ctx.enumDeclaration())
        case x =>
          println(s"Error: unexpected AltNum in MemberDeclaration: $x")
          null
      }
    } catch {
      case e: NotSupportedException =>
        val commented = NonTranslatedMemberDeclaration("", getSourceLocation(ctx))
        setNonTranslatedTreeNodeText(commented, ctx, e)
        commented
    }
  }

  override def visitFieldDeclaration(ctx: KrakatoaParser.FieldDeclarationContext): FieldDeclaration = {
    val fieldType = visitTypeType(ctx.typeType())
    val decls = visitVariableDeclaratorsAsList(ctx.variableDeclarators())
    val field = FieldDeclaration(fieldType, decls, getSourceLocation(ctx))
    field.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
    field
  }

  override def visitTypeTypeOrVoid(ctx: KrakatoaParser.TypeTypeOrVoidContext): TypeIndication = {
    if (ctx.typeType() != null) {
      visitTypeType(ctx.typeType())
    } else {
      new StringTypeIndication(ctx.VOID().getText, Nil, ctx.VOID().getText, getSourceLocation(ctx), getWhitespaceLeft(ctx.VOID()))
    }
  }

  def visitFormalParameterListAsList(ctx: KrakatoaParser.FormalParameterListContext): List[Parameter] = {
    var params:List[Parameter] = Nil
    if (ctx.formalParameter() != null) {
      params = ctx.formalParameter().asScala.map(visitFormalParameter).toList
    }
    if (ctx.lastFormalParameter() != null) {
      params.appended(visitLastFormalParameter(ctx.lastFormalParameter()))
    }
    params
  }

  override def visitFormalParameter(ctx: KrakatoaParser.FormalParameterContext): Parameter = {
    val modifiers = ctx.variableModifier().asScala.map(visitVariableModifier).toList
    val paramType = visitTypeType(ctx.typeType())
    val name = visitVariableDeclaratorId(ctx.variableDeclaratorId())
    val param = Parameter(paramType, name, modifiers, getSourceLocation(ctx))
    param
  }

  override def visitLastFormalParameter(ctx: KrakatoaParser.LastFormalParameterContext): VarargsParameter = {
    val modifiers = ctx.variableModifier().asScala.map(visitVariableModifier).toList
    val paramType = visitTypeType(ctx.typeType())
    val annotations = ctx.annotation().asScala.map(visitAnnotation).toList
    val name = visitVariableDeclaratorId(ctx.variableDeclaratorId())
    val param = VarargsParameter(paramType, annotations, name, getSourceLocation(ctx))
    param.modifiers = modifiers
    param.dots.precedingWhitespace = getWhitespaceLeft(ctx.ELLIPSIS())
    param
  }

  def setMethodDeclarationFields(ctx: KrakatoaParser.MethodDeclarationContext, decl: MethodDeclaration): Unit = {
    decl.name = visitJavaIdentifier(ctx.javaIdentifier())
    decl.returnType = Option(visitTypeTypeOrVoid(ctx.typeTypeOrVoid()))
    if (ctx.formalParameters().formalParameterList() != null) {
      decl.parameters = visitFormalParameterListAsList(ctx.formalParameters().formalParameterList())
    }
    var dimensions = ""
    for (x <- 0 until ctx.LBRACK().size()) {
      dimensions += getWhitespaceLeft(ctx.LBRACK(x)) + ctx.LBRACK(x).getText
      dimensions += getWhitespaceLeft(ctx.RBRACK(x)) + ctx.RBRACK(x).getText
    }
    if (dimensions != "") {
      decl.dimensions = Option(StringNode(dimensions, getSourceLocation(ctx.LBRACK(0))))
    }
    if (ctx.THROWS() != null) {
      decl.throws = Option(Throws(visitQualifiedNameListAsList(ctx.qualifiedNameList()), getSourceLocation(ctx.qualifiedNameList())))
      decl.throws.get.throwsKeyword.precedingWhitespace = getWhitespaceLeft(ctx.THROWS())
    }
    decl.body = visitMethodBody(ctx.methodBody())
    decl.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.formalParameters().LPAREN())
    decl.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.formalParameters().RPAREN())
  }

  override def visitMethodBody(ctx: KrakatoaParser.MethodBodyContext): BlockOrSemicolon = {
    if (ctx.SEMI() != null) {
      EmptyStatement(getSourceLocation(ctx.SEMI()), getWhitespaceLeft(ctx.SEMI()))
    } else {
      visitBlock(ctx.block())
    }
  }

  override def visitMethodDeclaration(ctx: KrakatoaParser.MethodDeclarationContext): MethodDeclaration = {
    val decl = new MethodDeclaration(null, null, Nil, None, None, None, None, null, getSourceLocation(ctx))
    setMethodDeclarationFields(ctx, decl)
    decl
  }

  override def visitGenericMethodDeclaration(ctx: KrakatoaParser.GenericMethodDeclarationContext): GenericMethodDeclaration = {
    val typeParams = visitTypeParameters(ctx.typeParameters())
    val decl = GenericMethodDeclaration(typeParams, null, null, Nil, None, None, None, None, null, getSourceLocation(ctx))
    setMethodDeclarationFields(ctx.methodDeclaration(), decl)
    decl
  }

  def setConstructorDeclarationFields(ctx: KrakatoaParser.ConstructorDeclarationContext, decl: MethodDeclaration): Unit = {
    decl.name = visitJavaIdentifier(ctx.javaIdentifier())
    if (ctx.formalParameters().formalParameterList() != null) {
      decl.parameters = visitFormalParameterListAsList(ctx.formalParameters().formalParameterList())
    }
    if (ctx.THROWS() != null) {
      decl.throws = Option(Throws(visitQualifiedNameListAsList(ctx.qualifiedNameList()), getSourceLocation(ctx.qualifiedNameList())))
      decl.throws.get.throwsKeyword.precedingWhitespace = getWhitespaceLeft(ctx.THROWS())
    }
    decl.body = visitBlock(ctx.constructorBody)
    decl.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.formalParameters().LPAREN())
    decl.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.formalParameters().RPAREN())
  }

  def visitQualifiedNameListAsList(ctx: KrakatoaParser.QualifiedNameListContext): List[StringNode] = {
    val names = ctx.qualifiedName().asScala.map(visitQualifiedNameWithWS).toList
    names.last.trailingWhitespace = ""
    names
  }

  override def visitConstructorDeclaration(ctx: KrakatoaParser.ConstructorDeclarationContext): MethodDeclaration = {
    val decl = GenericMethodDeclaration(typeParameters=null, pName=null, pReturnType=None, pParameters=Nil, pDimensions=None, pThrows=None, pDefault=None, pContractWithin=None, pBody=null, getSourceLocation(ctx))
    setConstructorDeclarationFields(ctx, decl)
    decl
  }

  override def visitGenericConstructorDeclaration(ctx: KrakatoaParser.GenericConstructorDeclarationContext): GenericMethodDeclaration = {
    val typeParameters = visitTypeParameters(ctx.typeParameters())
    val decl = new GenericMethodDeclaration(typeParameters, null, None, Nil, None, None, None, None, null, getSourceLocation(ctx))
    setConstructorDeclarationFields(ctx.constructorDeclaration(), decl)
    decl
  }

  override def visitBlock(ctx: KrakatoaParser.BlockContext): Block = {
    val statements = ctx.blockStatement().asScala.map(visitBlockStatement).toList
    val block = Block(None, statements, getSourceLocation(ctx))
    block.openBrace.precedingWhitespace = getWhitespaceLeft(ctx.LBRACE())
    block.closeBrace.precedingWhitespace = getWhitespaceLeft(ctx.RBRACE())
    block
  }
  
  override def visitBlockStatement(ctx: KrakatoaParser.BlockStatementContext): Statement = {
    ctx.getAltNumber match {
      case 1 =>
        val decl = visitLocalVariableDeclaration(ctx.localVariableDeclaration())
        decl.isStatement = true
        decl.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
        decl
      case 2 =>
        visitStatement(ctx.statement())
      case 3 =>
        visitLocalTypeDeclaration(ctx.localTypeDeclaration())
      case 4 =>
        visitValEmbedStatementBlock(ctx.valEmbedStatementBlock())
      case x =>
        println(s"Error: unexpected AltNum in BlockStatement: $x")
        null
    }
  }
  
  override def visitLocalVariableDeclaration(ctx: KrakatoaParser.LocalVariableDeclarationContext): LocalVariableDeclaration = {
    val modifiers = ctx.variableModifier().asScala.map(visitVariableModifier).toList
    val typeIndication = visitTypeType(ctx.typeType())
    val declarators = visitVariableDeclaratorsAsList(ctx.variableDeclarators())
    val decl = LocalVariableDeclaration(modifiers, typeIndication, declarators, isStatement = false, getSourceLocation(ctx))
    decl
  }

  override def visitVariableModifier(ctx: KrakatoaParser.VariableModifierContext): Modifier = {
    if (ctx.annotation() != null) {
      visitAnnotation(ctx.annotation())
    } else {
      new ModifierStringNode(ctx.FINAL().getText, getSourceLocation(ctx), getWhitespaceLeft(ctx))
    }
  }

  override def visitStatement(ctx: KrakatoaParser.StatementContext): Statement = {
    try {
      ctx.getAltNumber match {
        case 1 => //blockLabel = block
          visitBlock(ctx.block())
        case 2 => // assert expr (: expr)? ;
          val content = visitExpression(ctx.expression(0))
          val message = if (ctx.COLON() != null) Option(visitExpression(ctx.expression(1))) else None
          val assertion = AssertStatement(content, message, getSourceLocation(ctx))
          assertion.assertKeyword.precedingWhitespace = getWhitespaceLeft(ctx.ASSERT())
          if (ctx.COLON() != null) {
            assertion.colon.precedingWhitespace = getWhitespaceLeft(ctx.COLON())
          }
          assertion.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
          assertion
        case 3 => // IF expr statement (ELSE statement)?
          val condition = visitExpression(ctx.parExpression().expression())
          val thenStatement = visitStatement(ctx.statement(0))
          val elseStatement = if (ctx.ELSE() != null) Option(visitStatement(ctx.statement(1))) else None
          val ifStatement = IfStatement(condition, thenStatement, elseStatement, getSourceLocation(ctx))
          ifStatement.ifKeyword.precedingWhitespace = getWhitespaceLeft(ctx.IF())
          ifStatement.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.parExpression().LPAREN())
          ifStatement.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.parExpression().RPAREN())
          if (ctx.ELSE() != null) {
            ifStatement.elseKeyword.precedingWhitespace = getWhitespaceLeft(ctx.ELSE())
          }
          ifStatement
        case 4 => // contract? FOR (control) statement
          val contract = if (ctx.valEmbedLoopSpec() != null) visitValEmbedLoopSpecAsList(ctx.valEmbedLoopSpec()) else Nil
          val control = visitForControl(ctx.forControl())
          val statement = visitStatement(ctx.statement(0))
          val loop = ForLoop(contract, Nil, control, statement, getSourceLocation(ctx))
          loop.forKeyword.precedingWhitespace = getWhitespaceLeft(ctx.FOR())
          loop.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.LPAREN())
          loop.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.RPAREN())
          loop
        case 5 => // contract? WHILE expr statement
          val contract = if (ctx.valEmbedLoopSpec() != null) visitValEmbedLoopSpecAsList(ctx.valEmbedLoopSpec()) else Nil
          val expr = visitExpression(ctx.parExpression().expression())
          val statement = visitStatement(ctx.statement(0))
          val loop = WhileLoop(contract, Nil, expr, statement, getSourceLocation(ctx))
          loop.whileKeyword.precedingWhitespace = getWhitespaceLeft(ctx.WHILE())
          loop.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.parExpression().LPAREN())
          loop.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.parExpression().RPAREN())
          loop
        case 6 => // DO statement WHILE expr ;
          val contract = if (ctx.valEmbedLoopSpec() != null) visitValEmbedLoopSpecAsList(ctx.valEmbedLoopSpec()) else Nil
          val expr = visitExpression(ctx.parExpression().expression())
          val body = visitStatement(ctx.statement(0))
          val loop = DoWhileLoop(contract, body, expr, getSourceLocation(ctx))
          loop.doKeyword.precedingWhitespace = getWhitespaceLeft(ctx.DO())
          loop.whileKeyword.precedingWhitespace = getWhitespaceLeft(ctx.WHILE())
          loop.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.parExpression().LPAREN())
          loop.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.parExpression().RPAREN())
          loop.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
          loop
        case 7 => // TRY block (catch+ finally? | finally)
          val block = visitBlock(ctx.block())
          val catchClause = ctx.catchClause().asScala.map(visitCatchClause).toList
          val finallyBlock = if (ctx.finallyBlock() != null) Option(visitBlock(ctx.finallyBlock().block())) else None
          val tryBlock = Try(block, None, catchClause, finallyBlock, getSourceLocation(ctx))
          if (ctx.finallyBlock() != null) {
            tryBlock.finallyKeyword.precedingWhitespace = getWhitespaceLeft(ctx.finallyBlock().FINALLY())
          }
          tryBlock
        case 8 => //TRY resource block catch* finally?
          val resource = visitResourceSpecification(ctx.resourceSpecification())
          val block = visitBlock(ctx.block())
          val catchClause = ctx.catchClause().asScala.map(visitCatchClause).toList
          val finallyBlock = if (ctx.finallyBlock() != null) Option(visitBlock(ctx.finallyBlock().block())) else None
          val tryBlock = Try(block, Option(resource), catchClause, finallyBlock, getSourceLocation(ctx))
          tryBlock.tryKeyword.precedingWhitespace = getWhitespaceLeft(ctx.TRY())
          if (ctx.finallyBlock() != null) {
            tryBlock.finallyKeyword.precedingWhitespace = getWhitespaceLeft(ctx.finallyBlock().FINALLY())
          }
          tryBlock
        case 9 => //SWITCH expr { blockGroup* label* }
          val expr = visitExpression(ctx.parExpression().expression())
          var blocks = ctx.switchBlockStatementGroup().asScala.map(visitSwitchBlockStatementGroup).toList
          if (!ctx.switchLabel().isEmpty) {
            val lastBlock = SwitchCaseBlock(ctx.switchLabel().asScala.map(visitSwitchLabel).toList, Nil, getSourceLocation(ctx.switchLabel(0)))
            blocks = blocks.appended(lastBlock)
          }
          val switch = Switch(expr, blocks, getSourceLocation(ctx))
          switch.switchKeyword.precedingWhitespace = getWhitespaceLeft(ctx.SWITCH())
          switch.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.parExpression.LPAREN())
          switch.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.parExpression.RPAREN())
          switch.openBrace.precedingWhitespace = getWhitespaceLeft(ctx.LBRACE())
          switch.closeBrace.precedingWhitespace = getWhitespaceLeft(ctx.RBRACE())
          switch
        case 10 => // SYNCHRONIZED expr block
          val expr = visitExpression(ctx.parExpression().expression())
          val block = visitBlock(ctx.block())
          val synchronized = Synchronized(expr, block, getSourceLocation(ctx))
          synchronized.synchronizedKeyword.precedingWhitespace = getWhitespaceLeft(ctx.SYNCHRONIZED())
          synchronized.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.LBRACE())
          synchronized.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.RBRACE())
          synchronized
        case 11 => // RETURN expr? ;
          val expr = if (ctx.expression() != null) Option(visitExpression(ctx.expression(0))) else None
          val returnStatement = Return(expr, getSourceLocation(ctx))
          returnStatement.returnKeyword.precedingWhitespace = getWhitespaceLeft(ctx.RETURN())
          returnStatement.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
          returnStatement
        case 12 => // THROW expr ;
          val expr = visitExpression(ctx.expression(0))
          val throwStatement = ThrowStatement(expr, getSourceLocation(ctx))
          throwStatement.throwKeyword.precedingWhitespace = getWhitespaceLeft(ctx.THROW())
          throwStatement.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
          throwStatement
        case 13 => // BREAK id? ;
          val id = if (ctx.javaIdentifier() != null) Option(visitJavaIdentifier(ctx.javaIdentifier())) else None
          val breakStatement = Break(id, getSourceLocation(ctx))
          breakStatement.breakKeyword.precedingWhitespace = getWhitespaceLeft(ctx.BREAK())
          breakStatement.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
          breakStatement
        case 14 => // CONTINUE id? ;
          val id = if (ctx.javaIdentifier() != null) Option(visitJavaIdentifier(ctx.javaIdentifier())) else None
          val continue = Continue(id, getSourceLocation(ctx))
          continue.continueKeyword.precedingWhitespace = getWhitespaceLeft(ctx.CONTINUE())
          continue.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
          continue
        case 15 => // SEMI
          val semi = new EmptyStatement(getSourceLocation(ctx), getWhitespaceLeft(ctx.SEMI))
          semi
        case 16 => // expr ;
          val expr = ExpressionStatement(visitExpression(ctx.expression(0)), getSourceLocation(ctx))
          expr.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
          expr
        case 17 => // label=id : statement
          val labelName = visitJavaIdentifier(ctx.javaIdentifier())
          val label = Label(labelName, getSourceLocation(ctx.javaIdentifier()))
          label.colon.precedingWhitespace = getWhitespaceLeft(ctx.COLON())
          val statement = visitStatement(ctx.statement(0))
          LabeledStatement(label, statement, None, getSourceLocation(ctx))
        case 18 => // valStatement
          visitValStatement(ctx.valStatement())
        case x =>
          println(s"Error: unexpected AltNum in Statement: $x")
          null
      }
    } catch {
      case e: NotSupportedException =>
        val commented = NonTranslatedStatement("", getSourceLocation(ctx))
        setNonTranslatedTreeNodeText(commented, ctx, e)
        commented
    }
  }

  override def visitResourceSpecification(ctx: KrakatoaParser.ResourceSpecificationContext): TryResourceSpecification = {
    val semi = if (ctx.SEMI() != null) Option(StringNode(ctx.SEMI.getText, getSourceLocation(ctx.SEMI()), getWhitespaceLeft(ctx.SEMI()))) else None
    val resources = ctx.resources().resource().asScala.map(x=>addWhitespaceRight(visitResource(x),x)).toList
    val res = TryResourceSpecification(resources, semi, getSourceLocation(ctx))
    res.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.LPAREN())
    res.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.RPAREN())
    res
  }

  override def visitResource(ctx: KrakatoaParser.ResourceContext): TryResource = {
    val modifiers = ctx.variableModifier().asScala.map(visitVariableModifier).toList
    val rType = visitClassOrInterfaceType(ctx.classOrInterfaceType())
    val variable = visitVariableDeclaratorId(ctx.variableDeclaratorId())
    val expr = visitExpression(ctx.expression())
    val res = TryResource(modifiers, rType, variable, expr, getSourceLocation(ctx))
    res.equalsSign.precedingWhitespace = getWhitespaceLeft(ctx.ASSIGN())
    res
  }

  override def visitClassOrInterfaceType(ctx: KrakatoaParser.ClassOrInterfaceTypeContext): TypeIndication = {
    val text = new ParserPrintVisitor(tokens).visit(ctx)
    new StringTypeIndication(text, Nil, ctx.getText, getSourceLocation(ctx))
  }

  override def visitSwitchBlockStatementGroup(ctx: KrakatoaParser.SwitchBlockStatementGroupContext): SwitchCaseBlock = {
    SwitchCaseBlock(ctx.switchLabel().asScala.map(visitSwitchLabel).toList, ctx.blockStatement().asScala.map(visitBlockStatement).toList, getSourceLocation(ctx))
  }

  override def visitSwitchLabel(ctx: KrakatoaParser.SwitchLabelContext): SwitchLabel = {
    if (ctx.CASE() != null) {
      var cond: Expression = null
      if (ctx.expression() != null) {
        cond = visitExpression(ctx.expression())
      } else {
        cond = visitJavaIdentifier(ctx.javaIdentifier())
      }
      val label = SwitchLabel(Option(cond), getSourceLocation(ctx))
      label.caseKeyword.precedingWhitespace = getWhitespaceLeft(ctx.CASE())
      label.colon.precedingWhitespace = getWhitespaceLeft(ctx.COLON())
      label
    } else {
      val label = SwitchLabel(None, getSourceLocation(ctx))
      label.defaultKeyword.precedingWhitespace = getWhitespaceLeft(ctx.DEFAULT())
      label.colon.precedingWhitespace = getWhitespaceLeft(ctx.COLON())
      label
    }
  }

  override def visitForControl(ctx: KrakatoaParser.ForControlContext): ForLoopControl = {
    if (ctx.enhancedForControl() != null) {
      visitEnhancedForControl(ctx.enhancedForControl())
    } else {
      val initializer = if (ctx.forInit() != null) Option(visitForInit(ctx.forInit())) else None
      val condition = if (ctx.expression() != null) Option(visitExpression(ctx.expression())) else None
      val update = if (ctx.expressionList() != null) ctx.expressionList().expression().asScala.map(x=>addWhitespaceRight(visitExpression(x), x)).toList else Nil
      val control = RegularForLoopControl(initializer, condition, update, getSourceLocation(ctx))
      control.firstSemicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI(0))
      control.secondSemicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI(1))
      control
    }
  }

  override def visitCatchClause(ctx: KrakatoaParser.CatchClauseContext): Catch = {
    val modifiers = ctx.variableModifier().asScala.map(visitVariableModifier).toList
    val types = visitCatchTypeAsList(ctx.catchType())
    val name = visitJavaIdentifier(ctx.javaIdentifier())
    val block = visitBlock(ctx.block())
    val catchClause = Catch(Nil, types, name, block, getSourceLocation(ctx))
    catchClause.catchKeyword.precedingWhitespace = getWhitespaceLeft(ctx.CATCH())
    catchClause.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.LPAREN())
    catchClause.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.RPAREN())
    catchClause.modifiers = modifiers
    catchClause
  }

  def visitCatchTypeAsList(ctx: KrakatoaParser.CatchTypeContext): List[TypeIndication] = {
    ctx.qualifiedName().asScala.map(visitQualifiedNameWithWSAsTypeIndication).toList
  }

  def visitQualifiedNameWithWS(ctx: KrakatoaParser.QualifiedNameContext): StringNode = printNodeWithWS(ctx)

  def visitQualifiedNameWithWSAsTypeIndication(ctx: KrakatoaParser.QualifiedNameContext): TypeIndication = {
    val text = new ParserPrintVisitor(tokens).visit(ctx)
    new StringTypeIndication(text, Nil, ctx.getText, getSourceLocation(ctx),
      pTrailingWhitespace=getWhitespaceRight(ctx))
  }

  override def visitEnhancedForControl(ctx: KrakatoaParser.EnhancedForControlContext): ForEachLoopControl = {
    val modifiers = ctx.variableModifier().asScala.map(visitVariableModifier).toList
    val cType = visitTypeType(ctx.typeType())
    val variable = visitVariableDeclaratorId(ctx.variableDeclaratorId())
    val control = visitExpression(ctx.expression())
    val ctl = ForEachLoopControl(cType, variable, control, getSourceLocation(ctx))
    ctl.modifiers = modifiers
    ctl.colon.precedingWhitespace = getWhitespaceLeft(ctx.COLON())
    ctl
  }

  override def visitForInit(ctx: KrakatoaParser.ForInitContext): ForInit = {
    if (ctx.localVariableDeclaration() != null) {
      visitLocalVariableDeclaration(ctx.localVariableDeclaration())
    } else {
      val exprs = ctx.expressionList().expression().asScala.map(x=>addWhitespaceRight(visitExpression(x),x)).toList
      ExpressionListForInit(exprs, getSourceLocation(ctx))
    }
  }

  def visitVariableDeclaratorsAsList(ctx: KrakatoaParser.VariableDeclaratorsContext): List[VariableDeclarator] = {
    val decls = ctx.variableDeclarator().asScala.map(visitVariableDeclarator).toList
    decls.last.trailingWhitespace = ""
    decls
  }

  override def visitVariableDeclarator(ctx: KrakatoaParser.VariableDeclaratorContext): VariableDeclarator = {
    val name = visitVariableDeclaratorId(ctx.variableDeclaratorId())
    val initializer = if (ctx.ASSIGN() == null) None
    else Option(visitVariableInitializer(ctx.variableInitializer()))
    val decl = VariableDeclarator(name, initializer, getSourceLocation(ctx))
    if (ctx.ASSIGN() != null) {
      decl.equalSign.precedingWhitespace = getWhitespaceLeft(ctx.ASSIGN())
    }
    decl
  }

  override def visitVariableInitializer(ctx: KrakatoaParser.VariableInitializerContext): Initializer = {
    ctx.getAltNumber match {
      case 1 => visitArrayInitializer(ctx.arrayInitializer())
      case 2 => visitExpression(ctx.expression())
      case x =>
        println(s"Error: unexpected AltNum in VariableInitializer: $x")
        null
    }
  }

  override def visitArrayInitializer(ctx: KrakatoaParser.ArrayInitializerContext): ArrayInitializer = {
    val initializers = ctx.variableInitializer().asScala
      .map(x=>addWhitespaceRight(visitVariableInitializer(x),x)).toList
    var comma: Option[StringNode] = None
    if (ctx.variableInitializer().size() == ctx.COMMA().size() && ctx.COMMA().size() > 0) {
      val trailingComma: StringNode = StringNode(",", getSourceLocation(ctx.COMMA().asScala.last))
      trailingComma.precedingWhitespace = getWhitespaceLeft(ctx.COMMA().asScala.last)
      comma = Option(trailingComma)
    }
    val ini = ArrayInitializer(initializers, comma, getSourceLocation(ctx))
    ini.openBrace.precedingWhitespace = getWhitespaceLeft(ctx.LBRACE())
    ini.closeBrace.precedingWhitespace = getWhitespaceLeft(ctx.RBRACE())
    ini
  }

  override def visitVariableDeclaratorId(ctx: KrakatoaParser.VariableDeclaratorIdContext): StringNode = {
    printNode(ctx)
  }
  
  override def visitLocalTypeDeclaration(ctx: KrakatoaParser.LocalTypeDeclarationContext): Statement = {
    if (ctx.classDeclaration() != null) {
      val classDecl = visitClassDeclaration(ctx.classDeclaration())
      classDecl.modifiers = classDecl.modifiers.prependedAll(ctx.classOrInterfaceModifier().asScala.map(visitClassOrInterfaceModifier))
      classDecl
    } else if (ctx.interfaceDeclaration() != null) {
      val intDecl = visitInterfaceDeclaration(ctx.interfaceDeclaration())
      intDecl.modifiers = intDecl.modifiers.prependedAll(ctx.classOrInterfaceModifier().asScala.map(visitClassOrInterfaceModifier))
      intDecl
    } else {
      val semi = new EmptyStatement(getSourceLocation(ctx.SEMI()), getWhitespaceLeft(ctx.SEMI()))
      semi
    }
  }
  
  override def visitClassOrInterfaceModifier(ctx: KrakatoaParser.ClassOrInterfaceModifierContext): Modifier = {
    if (ctx.annotation() != null) {
      visitAnnotation(ctx.annotation())
    } else if(ctx.STATIC() != null) {
      StaticModifier(getSourceLocation(ctx), getWhitespaceLeft(ctx))
    } else {
      new ModifierStringNode(ctx.getText, getSourceLocation(ctx), getWhitespaceLeft(ctx))
    }
  }
  
  override def visitValEmbedStatementBlock(ctx: KrakatoaParser.ValEmbedStatementBlockContext): KrakatoaStatementBlock = {
    val statements = ctx.valStatement().asScala.map(visitValStatement).toList
    var startSpec: Option[StringNode] = None
    var endSpec: Option[StringNode] = None
    if (ctx.startSpec() != null) {
      startSpec = Some(StringNode(ctx.startSpec().getText, getSourceLocation(ctx.startSpec()), getWhitespaceLeft(ctx.startSpec())))
      endSpec = Some(StringNode(ctx.endSpec().getText, getSourceLocation(ctx.endSpec()), getWhitespaceLeft(ctx.endSpec())))
    }
    KrakatoaStatementBlock(startSpec, statements, endSpec, getSourceLocation(ctx))
  }
  
  override def visitValEmbedContract(ctx: KrakatoaParser.ValEmbedContractContext): Contract = {
    val cases = ctx.valEmbedContractBlock().asScala.flatMap(visitValEmbedContractBlockAsList).toList
    val res = Contract(cases, getSourceLocation(ctx))
    if (cases.head.startSpec.isDefined && cases.head.startSpec.get.text.contains("/*")) {
      res.startSpec = cases.head.startSpec
      res.endSpec = cases.last.endSpec
      cases.head.startSpec = None
      cases.last.endSpec = None
    }
    res
  }

  def visitValEmbedContractBlockAsList(ctx: KrakatoaParser.ValEmbedContractBlockContext): List[ContractClause] = {
    val cases = ctx.valContractClause().asScala.map(visitValContractClause).toList
    if (ctx.startSpec() != null) {
      cases.head.startSpec = Some(StringNode(ctx.startSpec().getText, getSourceLocation(ctx.startSpec()), getWhitespaceLeft(ctx.startSpec())))
      cases.last.endSpec = Some(StringNode(ctx.endSpec().getText, getSourceLocation(ctx.endSpec()), getWhitespaceLeft(ctx.endSpec())))
    }
    cases
  }

  override def visitValContractClause(ctx: KrakatoaParser.ValContractClauseContext): ContractClause = {
    try {
      ctx.getAltNumber match {
        case 1 =>
          val expressions = visitValExpressionListAsList(ctx.valExpressionList())
          val res = AssignsClause(StringNode("modifies", getSourceLocation(ctx.VAL_MODIFIES())), expressions, getSourceLocation(ctx))
          res.assignableKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_ASSIGNS())
          res.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
          res
        case 2 =>
          val expressions = visitValExpressionListAsList(ctx.valExpressionList())
          val res = AccessibleClause(expressions, getSourceLocation(ctx))
          res.accessibleKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_ACCESSIBLE())
          res.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
          res
        case 3 =>
          val expressions = visitValExpressionListAsList(ctx.valExpressionList())
          val res = AssignsClause(expressions=expressions, srcLocation=getSourceLocation(ctx))
          res.assignableKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_ASSIGNS())
          res.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
          res
        case 4 =>
          val expr = visitExpression(ctx.langExpr().expression())
          val res = AssumesClause(expr, getSourceLocation(ctx))
          res.assumesKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_ASSUMES())
          res.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
          res
        case 5 =>
          val expr = visitExpression(ctx.langExpr().expression())
          val res = RequiresClause(expr, getSourceLocation(ctx))
          res.requiresKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_REQUIRES())
          res.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
          res
        case 6 =>
          val expr = visitExpression(ctx.langExpr().expression())
          val res = EnsuresClause(expr, getSourceLocation(ctx))
          res.ensuresKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_ENSURES())
          res.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
          res
        case 7 =>
          val expr = visitExpression(ctx.langExpr().expression())
          val res = DecreasesClause(expr, getSourceLocation(ctx))
          res.measuredByKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_DECREASES())
          res.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
          res
        case 8 =>
          var label: StringNode = null
          if (!ctx.langId().javaIdentifier().isEmpty) {
            label = visitJavaIdentifier(ctx.langId().javaIdentifier())
          }
          val clauses = ctx.valContractClause().asScala.map(visitValContractClause).toList
          val whitespace = if ctx.VAL_BEHAVIOR() != null then getWhitespaceLeft(ctx.VAL_BEHAVIOR()) else getWhitespaceLeft(ctx.VAL_BEHAVIOUR())
          Behavior(None, label, false, None, clauses, None, getSourceLocation(ctx), whitespace)
        case 9 =>
          var label: StringNode = null
          if (ctx.langId().javaIdentifier() != null) {
            label = visitJavaIdentifier(ctx.langId().javaIdentifier())
          }
          val clauses = ctx.valContractClause().asScala.map(visitValContractClause).toList
          val whitespace = if ctx.VAL_NORMAL_BEHAVIOR() != null then getWhitespaceLeft(ctx.VAL_NORMAL_BEHAVIOR()) else getWhitespaceLeft(ctx.VAL_NORMAL_BEHAVIOUR())
          NormalBehavior(None, label, false, None, clauses, None, getSourceLocation(ctx), whitespace)
        case 10 =>
          var label: StringNode = null
          if (ctx.langId().javaIdentifier() != null) {
            label = visitJavaIdentifier(ctx.langId().javaIdentifier())
          }
          val clauses = ctx.valContractClause().asScala.map(visitValContractClause).toList
          val whitespace = if ctx.VAL_EXC_BEHAVIOR() != null then getWhitespaceLeft(ctx.VAL_EXC_BEHAVIOR()) else getWhitespaceLeft(ctx.VAL_EXC_BEHAVIOUR())
          ExceptionalBehavior(None, label, false, None, clauses, None, getSourceLocation(ctx), whitespace)
        case 11 =>
          val tType = visitTypeType(ctx.langType().typeType())
          val name = Option(ctx.langId()).map(x => visitJavaIdentifier(x.javaIdentifier()))
          val expr = visitExpression(ctx.langExpr().expression())
          val res = SignalsClause(tType, name, Some(expr), getSourceLocation(ctx))
          res.signalsKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_SIGNALS())
          res.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.LPAREN())
          res.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.RPAREN())
          res.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
          res
      }
    } catch {
      case e: NotSupportedException =>
        val commented = NonTranslatedContractClause("", getSourceLocation(ctx))
        setNonTranslatedTreeNodeText(commented, ctx, e)
        commented
    }
  }

  def visitValEmbedLoopSpecAsList(ctx: KrakatoaParser.ValEmbedLoopSpecContext): List[LoopSpecification] = {
    ctx.valEmbedLoopSpecBlock().asScala.map(visitValEmbedLoopSpecBlock).toList
  }

  override def visitValEmbedLoopSpecBlock(ctx: KrakatoaParser.ValEmbedLoopSpecBlockContext): LoopSpecification = {
    val clauses = ctx.valLoopClause().asScala.map(visitValLoopClause).toList
    val openSpec = Option(ctx.startSpec()).map(x=>StringNode(x.getText, getSourceLocation(x), getWhitespaceLeft(x)))
    val endSpec = Option(ctx.endSpec()).map(x=>StringNode(x.getText, getSourceLocation(x), getWhitespaceLeft(x)))
    LoopSpecification(openSpec, clauses, endSpec, getSourceLocation(ctx))
  }

  override def visitValLoopClause(ctx: KrakatoaParser.ValLoopClauseContext): LoopClause = {
    ctx.getAltNumber match {
      case 1 =>
        val expr = visitExpression(ctx.langExpr.expression())
        val clause = LoopInvariantClause(expr, getSourceLocation(ctx))
        clause.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
        clause.invariantKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_LOOP_INVARIANT())
        clause
      case 2 =>
        val expr = visitExpression(ctx.langExpr.expression())
        val clause = LoopVariantClause(expr, getSourceLocation(ctx))
        clause.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
        clause.decreasesKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_LOOP_VARIANT())
        clause
      case 3 =>
        val keyword = StringNode("loop_assigns", null)
        val exprs = visitValExpressionListAsList(ctx.valExpressionList())
        val assigns = AssignsClause(keyword, exprs, getSourceLocation(ctx))
        assigns.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
        assigns.assignableKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_LOOP_ASSIGNS)
        assigns
      case 4 =>
        val clause = visitValLoopClause(ctx.valLoopClause())
        val label = visitJavaIdentifier(ctx.langId().javaIdentifier())
        val forClause = ForClause(label, clause, getSourceLocation(ctx))
        forClause.forKeyword.precedingWhitespace = getWhitespaceLeft(ctx.FOR())
        forClause
    }
  }

  def visitValExpressionListAsList(ctx: KrakatoaParser.ValExpressionListContext): List[Expression] = {
    var res: List[Expression] = List(visitExpression(ctx.langExpr().expression()))
    if (ctx.valExpressionList() != null) {
      res = res.appendedAll(visitValExpressionListAsList(ctx.valExpressionList()))
    }
    res
  }

  override def visitValEmbedDeclarationBlock(ctx: KrakatoaParser.ValEmbedDeclarationBlockContext): SpecDeclarationBlock = {
    val decls = ctx.valDeclaration().asScala.map(visitValDeclaration).toList
    val startSpec = StringNode(ctx.startSpec().getText, getSourceLocation(ctx.startSpec()), getWhitespaceLeft(ctx.startSpec()))
    val endSpec = StringNode(ctx.endSpec().getText, getSourceLocation(ctx.endSpec()), getWhitespaceLeft(ctx.endSpec()))
    SpecDeclarationBlock(Some(startSpec), decls, Some(endSpec), getSourceLocation(ctx))
  }

  override def visitValDeclaration(ctx: KrakatoaParser.ValDeclarationContext): SpecDeclaration = {
    ctx.getAltNumber match {
      case 1 => // predicate
        val name = visitJavaIdentifier(ctx.langId().javaIdentifier())
        var labels : Option[LabelList] = None
        if (ctx.labelList() != null) {
          labels = Some(visitLabelList(ctx.labelList()))
        }
        var args: List[Parameter] = null
        if (ctx.valArgList() != null) {
          args = visitValArgListAsList(ctx.valArgList())
        }
        var expr: Option[Expression] = None
        if (ctx.langExpr() != null) {
          expr = Some(visitExpression(ctx.langExpr().expression()))
        }
        val decl = PredicateDeclaration(name, labels, args, expr, getSourceLocation(ctx))
        decl.tType.precedingWhitespace = getWhitespaceLeft(ctx.VAL_PREDICATE())
        decl.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.LPAREN())
        decl.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.RPAREN())
        decl.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
        if (ctx.ASSIGN() != null) then decl.equalSign.precedingWhitespace = getWhitespaceLeft(ctx.ASSIGN())
        decl
      case 2 => // inductive data type
        val name = visitJavaIdentifier(ctx.langId().javaIdentifier())
        var labels : Option[LabelList] = None
        if (ctx.labelList() != null) {
          labels = Some(visitLabelList(ctx.labelList()))
        }
        var args: List[Parameter] = null
        if (ctx.valArgList() != null) {
          args = visitValArgListAsList(ctx.valArgList())
        }
        val cases = ctx.caseDefinition().asScala.map(visitCaseDefinition).toList
        val decl = InductiveDatatypeDeclaration(name, labels, args, cases, getSourceLocation(ctx))
        decl.inductiveKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_INDUCTIVE())
        decl.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.LPAREN())
        decl.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.RPAREN())
        decl.openCurlyParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.LBRACE())
        decl.closeCurlyParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.RBRACE())
        decl
      case 3 => // invariant
        val name = visitJavaIdentifier(ctx.langId().javaIdentifier())
        val expr = visitExpression(ctx.langExpr().expression())
        val decl = InvariantDeclaration(name, expr, getSourceLocation(ctx))
        decl.invariantKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_INVARIANT())
        decl.colon.precedingWhitespace = getWhitespaceLeft(ctx.COLON())
        decl.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
        decl
      case 4 => // lemma
        val name = visitJavaIdentifier(ctx.langId().javaIdentifier())
        var labels: Option[LabelList] = None
        if (ctx.labelList() != null) {
          labels = Some(visitLabelList(ctx.labelList()))
        }
        val expr = visitExpression(ctx.langExpr().expression())
        val decl = LemmaDeclaration(name, labels, expr, getSourceLocation(ctx))
        decl.lemmaKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_LEMMA())
        decl.colon.precedingWhitespace = getWhitespaceLeft(ctx.COLON())
        decl.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
        decl
      case 5 => // logic int/real
        val tType = if ctx.VAL_INTEGER() != null then IntegerSpecTypeIndication(srcLocation=getSourceLocation(ctx), getWhitespaceLeft(ctx.VAL_INTEGER())) else RealSpecTypeIndication(srcLocation=getSourceLocation(ctx), getWhitespaceLeft(ctx.VAL_REAL()))
        val name = visitJavaIdentifier(ctx.langId().javaIdentifier())
        var labels: Option[LabelList] = None
        if (ctx.labelList() != null) {
          labels = Some(visitLabelList(ctx.labelList()))
        }
        val args: List[Parameter] = visitValArgListAsList(ctx.valArgList())
        var expr: Option[Expression] = None
        if (ctx.langExpr() != null) {
          expr = Some(visitExpression(ctx.langExpr().expression()))
        }
        val decl = LogicVariableDeclaration(tType, name, labels, args, expr, getSourceLocation(ctx))
        decl.logicKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_LOGIC())
        decl.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.LPAREN())
        decl.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.RPAREN())
        if (ctx.ASSIGN() != null) decl.equals.precedingWhitespace = getWhitespaceLeft(ctx.ASSIGN())
        decl.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
        decl
      case 6 => // axiom
        val name = visitJavaIdentifier(ctx.langId().javaIdentifier())
        var labels: Option[LabelList] = None
        if (ctx.labelList() != null) {
          labels = Some(visitLabelList(ctx.labelList()))
        }
        val expr = visitExpression(ctx.langExpr().expression())
        val decl = AxiomDeclaration(name, labels, expr, getSourceLocation(ctx))
        decl.axiomKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_AXIOM())
        decl.colon.precedingWhitespace = getWhitespaceLeft(ctx.COLON())
        decl.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
        decl
      case 7 => // axiomatic
        val name = visitJavaIdentifier(ctx.langId().javaIdentifier())
        val decls: List[SpecDeclaration] = ctx.valDeclaration().asScala.map(visitValDeclaration).toList
        val decl = AxiomaticDeclaration(name, decls, getSourceLocation(ctx))
        decl.axiomaticKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_AXIOMATIC())
        decl.openCurlyParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.LBRACE())
        decl.closeCurlyParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.RBRACE())
        decl
      case 8 => // ghost code
        val decl = visitMemberDeclaration(ctx.langDecl().memberDeclaration())
        val res = GhostDeclaration(decl, getSourceLocation(ctx), getWhitespaceLeft(ctx.VAL_GHOST()))
        res
    }
  }

  def visitExpression(ctx: KrakatoaParser.ExpressionContext): Expression = {
    visit(ctx).asInstanceOf[Expression]
  }

  override def visitPrimaryExpression(ctx: KrakatoaParser.PrimaryExpressionContext): Expression = {
    if (ctx.LPAREN() != null) {
      val expr = ParExpression(visitExpression(ctx.expression()), getSourceLocation(ctx))
      expr.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.LPAREN())
      expr.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.RPAREN())
      expr
    } else if (ctx.THIS() != null) {
      val name = StringNode(ctx.THIS().getText, getSourceLocation(ctx))
      Literal(name, getSourceLocation(ctx), getWhitespaceLeft(ctx.THIS()))
    } else if (ctx.SUPER() != null) {
      val name = StringNode(ctx.SUPER().getText, getSourceLocation(ctx))
      Literal(name, getSourceLocation(ctx), getWhitespaceLeft(ctx.SUPER()))
    } else if (ctx.literal() != null) {
      val name = StringNode(ctx.literal().getText, getSourceLocation(ctx))
      Literal(name, getSourceLocation(ctx), getWhitespaceLeft(ctx.literal()))
    } else if (ctx.javaIdentifier() != null) {
      visitJavaIdentifier(ctx.javaIdentifier())
    } else if (ctx.typeTypeOrVoid() != null) {
      val typeInd = visitTypeTypeOrVoid(ctx.typeTypeOrVoid())
      val expr = ClassName(typeInd, getSourceLocation(ctx))
      expr.dot.precedingWhitespace = getWhitespaceLeft(ctx.DOT())
      expr.classKeyword.precedingWhitespace = getWhitespaceLeft(ctx.CLASS())
      expr
    } else {
        val typeParam = visitNonWildcardTypeArguments(ctx.nonWildcardTypeArguments())
        if (ctx.explicitGenericInvocationSuffix() != null) {
          if (ctx.explicitGenericInvocationSuffix().SUPER() != null) {
            if (ctx.explicitGenericInvocationSuffix().superSuffix().javaIdentifier() != null) {
              var text = ctx.explicitGenericInvocationSuffix().SUPER().getText
              text += getWhitespaceLeft(ctx.explicitGenericInvocationSuffix().superSuffix().DOT())
              text += ctx.explicitGenericInvocationSuffix().superSuffix().DOT().getText
              text += getWhitespaceLeft(ctx.explicitGenericInvocationSuffix().superSuffix().javaIdentifier())
              text += ctx.explicitGenericInvocationSuffix().superSuffix().javaIdentifier().getText
              val name = StringNode(text, getSourceLocation(ctx), getWhitespaceLeft(ctx))
              val arguments = if (ctx.explicitGenericInvocationSuffix().superSuffix().arguments()!=null)
                Some(visitArguments(ctx.explicitGenericInvocationSuffix()
                  .superSuffix().arguments()))
              else None
              GenericInvocationExpression(typeParam, name, arguments, getSourceLocation(ctx))

            } else {
              val sup = ctx.explicitGenericInvocationSuffix().SUPER()
              val name = StringNode(sup.getText, getSourceLocation(sup), getWhitespaceLeft(sup))
              val arguments = visitArguments(ctx.explicitGenericInvocationSuffix().superSuffix().arguments())
              GenericInvocationExpression(typeParam, name, Some(arguments), getSourceLocation(ctx))
            }
          } else {
            val name = visitJavaIdentifier(ctx.explicitGenericInvocationSuffix().javaIdentifier())
            val arguments = visitArguments(ctx.explicitGenericInvocationSuffix().arguments())
            GenericInvocationExpression(typeParam, name, Some(arguments), getSourceLocation(ctx))
          }
        } else {
          val name = StringNode(ctx.THIS().getText, getSourceLocation(ctx.THIS()),
            getWhitespaceLeft(ctx.THIS()))
          val arguments = visitArguments(ctx.arguments())
          GenericInvocationExpression(typeParam, name, Some(arguments), getSourceLocation(ctx))
        }
    }
  }

  override def visitValPrimaryExpression(ctx: KrakatoaParser.ValPrimaryExpressionContext): Expression
  = visitValPrimary(ctx.valPrimary())

  override def visitBinaryExpression(ctx: KrakatoaParser.BinaryExpressionContext): Expression = {
    val left = visitExpression(ctx.expression(0))
    val right = visitExpression(ctx.expression(1))
    var op: StringNode = null
    if (ctx.bop != null) {
      op = StringNode(ctx.bop.getText, getSourceLocation(ctx.bop), getWhitespaceLeft(ctx.bop))
    } else if (ctx.impOp() != null) {
      op = StringNode(ctx.impOp().getText, getSourceLocation(ctx.impOp()), getWhitespaceLeft(ctx.impOp()))
    } else if (ctx.eqOp() != null) {
      op = StringNode(ctx.eqOp().getText, getSourceLocation(ctx.eqOp()), getWhitespaceLeft(ctx.eqOp()))
    } else{
      var text = ""
      if (ctx.GT().size() > 0){
        for (x <- ctx.GT().asScala) {
          text += x.getText
        }
        op = StringNode(text, getSourceLocation(ctx.GT(0)), getWhitespaceLeft(ctx.GT(0)))
      } else {
        for (x <- ctx.LT().asScala) {
          text += x.getText
        }
        op = StringNode(text, getSourceLocation(ctx.LT(0)), getWhitespaceLeft(ctx.LT(0)))
      }
    }
    BinaryExpression(left, op, right, getSourceLocation(ctx))
  }

  override def visitInstanceofExpression(ctx: KrakatoaParser.InstanceofExpressionContext): Expression = {
    val left = visitExpression(ctx.expression())
    val right = visitTypeType(ctx.typeType())
    val res = InstanceofExpression(left, right, getSourceLocation(ctx))
    res.instanceofKeyword.precedingWhitespace = getWhitespaceLeft(ctx.INSTANCEOF())
    res
  }

  override def visitValPrimary(ctx: KrakatoaParser.ValPrimaryContext): Expression = {
    ctx.altNum match {
      case 1 => // langId labelList '(' expressionList? ')'
        val name = visitJavaIdentifier(ctx.langId().javaIdentifier())
        val labels = visitLabelList(ctx.labelList())
        val exprs = visitValExpressionListAsList(ctx.valExpressionList())
        LabeledExpression(name, labels, exprs, getSourceLocation(ctx))
      case 2 => // '[' expr ']' expr
        null //TODO
      case 3 => // ('\forall' | '\exists') declArgList ';' expr
        val declList = visitDeclArgListAsList(ctx.declArgList())
        val expr = visitExpression(ctx.langExpr(0).expression())
        if (ctx.FORALL() != null) {
          val qExpr = ForallExpression(declList, expr, getSourceLocation(ctx), getWhitespaceLeft(ctx.FORALL()))
          qExpr.semi.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
          qExpr
        } else {
          val qExpr = ExistsExpression(declList, expr, getSourceLocation(ctx), getWhitespaceLeft(ctx.EXISTS()))
          qExpr.semi.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
          qExpr
        }
      case 4 => // '\old(' expr ')'
        val expr = visitExpression(ctx.langExpr(0).expression())
        val old = OldExpression(expr, None, getSourceLocation(ctx))
        old.oldKeyword.precedingWhitespace = getWhitespaceLeft(ctx.OLD())
        old.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.LPAREN())
        old.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.RPAREN())
        old
      case 5 => // '\at(' expr ',' label ')'
        val expr = visitExpression(ctx.langExpr(0).expression())
        val label = visitLabel(ctx.label())
        val atExpr = AtExpression(expr, Some(label), getSourceLocation(ctx))
        atExpr.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.LPAREN())
        atExpr.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.RPAREN())
        atExpr.oldKeyword.precedingWhitespace = getWhitespaceLeft(ctx.AT())
        atExpr
      case 6 => // '\fresh(' expr ')'
        val expression = visitExpression(ctx.langExpr(0).expression())
        val expr = FreshExpression(List(expression), getSourceLocation(ctx))
        expr.freshKeyword.precedingWhitespace = getWhitespaceLeft(ctx.FRESH())
        expr.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.LPAREN())
        expr.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.RPAREN())
        expr
      case 7 => // \nothing
          StringNode(ctx.NOTHING().getText, getSourceLocation(ctx.NOTHING()), getWhitespaceLeft(ctx.NOTHING()))
      case 8 => // \everything
          StringNode(ctx.EVERYTHING().getText, getSourceLocation(ctx.EVERYTHING()), getWhitespaceLeft(ctx.EVERYTHING()))
      case 9 => // \result
        return ResultExpression(getSourceLocation(ctx.RESULT()), getWhitespaceLeft(ctx.RESULT()))
      case (10 | 11 | 12 | 13 | 14) => // built-in functions such as \abs(x), cos(x), \pow(x,y) and \pi
        throw new NotSupportedException("built-in mathematical functions", getSourceLocation(ctx))
      case _ =>
        println("Unknown valPrimary expression")
        null
    }
  }

  def visitDeclArgListAsList(ctx: KrakatoaParser.DeclArgListContext): List[QuantifiedVarDeclaration] = {
    ctx.declArg().asScala.map(visitDeclArg).toList
  }

  override def visitDeclArg(ctx: KrakatoaParser.DeclArgContext): QuantifiedVarDeclaration = {
    val tType = visitTypeType(ctx.langType().typeType())
    var decls: List[StringNode] = Nil
    for (i <- 0 to ctx.langId().size()-1) {
      var text = visitJavaIdentifier(ctx.langId(i).javaIdentifier()).text
      if (ctx.LBRACK(i) != null) {
        text = text.concat(ctx.LBRACK(i).getText + ctx.RBRACK(i).getText)
      }
      val decl = StringNode(text, getSourceLocation(ctx.langId(i)), getWhitespaceLeft(ctx.langId(i)))
      decls = decls.appended(decl)
    }
    QuantifiedVarDeclaration(tType, decls, getSourceLocation(ctx))
  }

  override def visitMethodCall(ctx: KrakatoaParser.MethodCallContext): Expression = {
    var name: StringNode = null
    if (ctx.javaIdentifier() != null) {
      name = visitJavaIdentifier(ctx.javaIdentifier())
    } else if (ctx.THIS() != null) {
      name = StringNode(ctx.THIS().getText, getSourceLocation(ctx.THIS()))
    } else { // SUPER call
      name = StringNode(ctx.SUPER().getText, getSourceLocation(ctx.SUPER()))
    }
    var args : List[Expression] = Nil
    if (ctx.expressionList() != null) {
      args = args.appendedAll(visitExpressionListAsList(ctx.expressionList()))
    }
    MethodCallExpression(name, Arguments(args, getSourceLocation(ctx)), getSourceLocation(ctx))
  }

  def visitExpressionListAsList(ctx: KrakatoaParser.ExpressionListContext): List[Expression] = {
    ctx.expression().asScala.map(visitExpression).toList
  }

  override def visitNestedInvocationExpression(ctx: KrakatoaParser.NestedInvocationExpressionContext):
  Expression = {
    val left = visitExpression(ctx.expression())
    var right: Expression = null
    if (ctx.javaIdentifier() != null) {
      right = visitJavaIdentifier(ctx.javaIdentifier())
    } else if (ctx.methodCall() != null) {
      right = visitMethodCall(ctx.methodCall())
    } else if (ctx.THIS() != null) {
      right = StringNode(ctx.THIS().getText, getSourceLocation(ctx.THIS()),
        getWhitespaceLeft(ctx.THIS()))
    } else if (ctx.NEW() != null) {
      val creator = visitInnerCreator(ctx.innerCreator())
      creator.newKeyword.precedingWhitespace = getWhitespaceLeft(ctx.NEW())
      if (ctx.nonWildcardTypeArguments() != null) {
        creator.typeArguments = Some(visitNonWildcardTypeArguments(ctx.nonWildcardTypeArguments()))
      }
      right = creator
    } else if (ctx.SUPER() != null) {
      if (ctx.superSuffix().javaIdentifier() != null) {
        var text = ctx.SUPER().getText
        text += getWhitespaceLeft(ctx.superSuffix().DOT())
        text += ctx.superSuffix().DOT().getText
        text += getWhitespaceLeft(ctx.superSuffix().javaIdentifier())
        text += ctx.superSuffix().javaIdentifier().getText
        if (ctx.superSuffix().arguments()!=null) {
          val name = StringNode(text, getSourceLocation(ctx), getWhitespaceLeft(ctx.SUPER()))
          val args = visitArguments(ctx.superSuffix().arguments())
          right = MethodCallExpression(name, args, getSourceLocation(ctx))
        } else {
          right = StringNode(text, getSourceLocation(ctx), getWhitespaceLeft(ctx.SUPER()))
        }
      } else {
        val name = StringNode(ctx.SUPER().getText, getSourceLocation(ctx.SUPER()),
          getWhitespaceLeft(ctx.SUPER()))
        val arguments = visitArguments(ctx.superSuffix().arguments())
        MethodCallExpression(name, arguments, getSourceLocation(ctx))
      }
    } else {
      right = visitExplicitGenericInvocation(ctx.explicitGenericInvocation())
    }
    val expr = NestedInvocationExpression(left, right, getSourceLocation(ctx))
    expr.dot.precedingWhitespace = getWhitespaceLeft(ctx.DOT())
    expr
  }

  override def visitArrayIndexExpression(ctx: KrakatoaParser.ArrayIndexExpressionContext): Expression = {
    val arrExpr = visitExpression(ctx.expression())
    var index: IndexExpression = null
    if (ctx.indexExpression().expression() != null) {
      index = visitExpression(ctx.indexExpression().expression())
    } else {
      val idxExprCtx = ctx.indexExpression().valIndexExpression()
      if (idxExprCtx.MUL() != null) {
        val idx = IndexAllExpression(getSourceLocation(ctx))
        idx.star.precedingWhitespace = getWhitespaceLeft(idxExprCtx.MUL())
        index = idx
      } else {
        val start = visitExpression(idxExprCtx.langExpr(0).expression())
        val end = visitExpression(idxExprCtx.langExpr(1).expression())
        val range = RangeExpression(start, end, getSourceLocation(ctx))
        range.dots.precedingWhitespace = getWhitespaceLeft(idxExprCtx.RANGE_TO())
        index = range
      }
    }
    val arr = ArrayIndexExpression(arrExpr, index, getSourceLocation(ctx))
    arr.openBracket.precedingWhitespace = getWhitespaceLeft(ctx.LBRACK())
    arr.closeBracket.precedingWhitespace = getWhitespaceLeft(ctx.RBRACK())
    arr
  }

  //TODO: check if we need methodcall in JavaParser
//  override def visitMethodCallExpression(ctx: KrakatoaParser.MethodCallExpressionContext): Expression =
//    visitMethodCall(ctx.methodCall())

  override def visitNewExpression(ctx: KrakatoaParser.NewExpressionContext): Expression = {
    val newExpr = visitCreator(ctx.creator())
    newExpr.newKeyword.precedingWhitespace = getWhitespaceLeft(ctx.NEW())
    newExpr
  }

  override def visitCastExpression(ctx: KrakatoaParser.CastExpressionContext): Expression = {
    val annotations = ctx.annotation().asScala.map(visitAnnotation).toList
    val types = ctx.typeType().asScala.map(visitTypeTypeWithWS).toList
    val expr = visitExpression(ctx.expression())
    val cast = CastExpression(annotations, types, expr, getSourceLocation(ctx))
    cast.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.LPAREN())
    cast.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.RPAREN())
    cast
  }

  override def visitPostfixExpression(ctx: KrakatoaParser.PostfixExpressionContext): Expression = {
    val expr = visitExpression(ctx.expression())
    val op = StringNode(ctx.postfix.getText, getSourceLocation(ctx.postfix), getWhitespaceLeft(ctx.postfix))
    PostfixExpression(expr, op, getSourceLocation(ctx))
  }

  override def visitPrefixExpression(ctx: KrakatoaParser.PrefixExpressionContext): Expression = {
    val expr = visitExpression(ctx.expression())
    val op = StringNode(ctx.prefix.getText, getSourceLocation(ctx.prefix), getWhitespaceLeft(ctx.prefix))
    PrefixExpression(expr, op, getSourceLocation(ctx))
  }

  override def visitTernaryExpression(ctx: KrakatoaParser.TernaryExpressionContext): Expression = {
    val cond = visitExpression(ctx.expression(0))
    val thenExpr = visitExpression(ctx.expression(1))
    val elseExpr = visitExpression(ctx.expression(2))
    val tern = TernaryExpression(cond, thenExpr, elseExpr, getSourceLocation(ctx))
    tern.questionmark.precedingWhitespace = getWhitespaceLeft(ctx.QUESTION())
    tern.colon.precedingWhitespace = getWhitespaceLeft(ctx.COLON())
    tern
  }

  override def visitLamExpression(ctx: KrakatoaParser.LamExpressionContext): Expression =
    visitLambdaExpression(ctx.lambdaExpression())

  override def visitMethodRefExpression(ctx: KrakatoaParser.MethodRefExpressionContext): Expression = {
    if (ctx.expression() != null) {
      val left = visitExpression(ctx.expression())
      var typeArgumentClause: Option[TypeArgumentClause] = None
      if (ctx.typeArguments() != null) {
        typeArgumentClause = Some(visitTypeArguments(ctx.typeArguments()))
      }
      val right = visitJavaIdentifier(ctx.javaIdentifier())
      val expr = MethodReferenceExpression(left, typeArgumentClause, right, getSourceLocation(ctx))
      expr.doubleColon.precedingWhitespace = getWhitespaceLeft(ctx.COLONCOLON())
      expr
    } else if (ctx.typeType() != null) {
      val left = visitTypeType(ctx.typeType())
      var typeArgumentClause: Option[TypeArgumentClause] = None
      if (ctx.typeArguments() != null) {
        typeArgumentClause = Some(visitTypeArguments(ctx.typeArguments()))
      }
      val right = if (ctx.javaIdentifier()!=null) visitJavaIdentifier(ctx.javaIdentifier())
      else StringNode(ctx.NEW().getText, getSourceLocation(ctx.NEW()), getWhitespaceLeft(ctx.NEW()))
      val expr = MethodReferenceExpression(left, typeArgumentClause, right, getSourceLocation(ctx))
      expr.doubleColon.precedingWhitespace = getWhitespaceLeft(ctx.COLONCOLON())
      expr
    } else {
      val text = new ParserPrintVisitor(tokens).visit(ctx.classType())
      val left = new StringTypeIndication(text, Nil, ctx.classType().getText,
        getSourceLocation(ctx.classType()))
      var typeArgumentClause: Option[TypeArgumentClause] = None
      if (ctx.typeArguments() != null) {
        typeArgumentClause = Some(visitTypeArguments(ctx.typeArguments()))
      }
      val right = StringNode(ctx.NEW().getText, getSourceLocation(ctx.NEW()), getWhitespaceLeft(ctx.NEW()))
      val expr = MethodReferenceExpression(left, typeArgumentClause, right, getSourceLocation(ctx))
      expr.doubleColon.precedingWhitespace = getWhitespaceLeft(ctx.COLONCOLON())
      expr
    }
  }

  override def visitLambdaExpression(ctx: KrakatoaParser.LambdaExpressionContext): Expression = {
    val params = visitLambdaParametersAsList(ctx.lambdaParameters())
    val hasParentheses = (ctx.lambdaParameters().LPAREN() != null)
    val body = visitLambdaBody(ctx.lambdaBody())
    val expr = LambdaExpression(hasParentheses, params, body, getSourceLocation(ctx))
    if (hasParentheses) {
      expr.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.lambdaParameters().LPAREN())
      expr.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.lambdaParameters().RPAREN())
    }
    expr.arrow.precedingWhitespace = getWhitespaceLeft(ctx.ARROW())
    expr
  }

  def visitLambdaParametersAsList(ctx: KrakatoaParser.LambdaParametersContext): List[Parameter] = {
    ctx.altNum match {
      case 1 =>
        val name = visitJavaIdentifier(ctx.javaIdentifier(0))
        val noType = new NoType(getSourceLocation(ctx))
        List(Parameter(noType, name, Nil, getSourceLocation(ctx)))
      case 2 =>
        if (ctx.formalParameterList() != null) {
          visitFormalParameterListAsList(ctx.formalParameterList())
        } else {
          Nil
        }
      case 3 =>
        var list: List[Parameter] = Nil
        val noType = new NoType(getSourceLocation(ctx))
        for (ident <- ctx.javaIdentifier().asScala) {
          val name = visitJavaIdentifier(ident)
          val param = Parameter(noType, name, Nil, getSourceLocation(ident))
          param.trailingWhitespace = getWhitespaceRight(ident)
          list = list.appended(param)
        }
        list.last.trailingWhitespace = ""
        list
    }
  }

  override def visitLambdaBody(ctx: KrakatoaParser.LambdaBodyContext): LambdaBody = {
    ctx.altNum match {
      case 1 =>
        visitExpression(ctx.expression())
      case 2 =>
        visitBlock(ctx.block())
    }
  }

  override def visitCreator(ctx: KrakatoaParser.CreatorContext): NewExpression = {
    val typeIndication = visitCreatedName(ctx.createdName())
    if (ctx.arrayCreatorRest() != null) {
      var dimensions: List[Dimension] = Nil
      for (i <- 0 until ctx.arrayCreatorRest().LBRACK().size()) {
        var size: Option[Expression] = None
        if (i < ctx.arrayCreatorRest().expression().size()) {
          size = Some(visitExpression(ctx.arrayCreatorRest().expression(i)))
        }
        val dim = Dimension(size, getSourceLocation(ctx.arrayCreatorRest().LBRACK(i)))
        dim.openBracket.precedingWhitespace = getWhitespaceLeft(ctx.arrayCreatorRest().LBRACK(i))
        dim.closeBracket.precedingWhitespace = getWhitespaceLeft(ctx.arrayCreatorRest().RBRACK(i))
        dimensions = dimensions.appended(dim)
      }
      val initializer = if(ctx.arrayCreatorRest().arrayInitializer() != null)
        Some(visitArrayInitializer(ctx.arrayCreatorRest().arrayInitializer()))
      else None
      NewArrayExpression(typeIndication, dimensions, initializer, getSourceLocation(ctx))
    } else {
      val typeArgumentClause = if(ctx.nonWildcardTypeArguments() != null)
        Some(visitNonWildcardTypeArguments(ctx.nonWildcardTypeArguments()))
      else None
      val args = visitArguments(ctx.classCreatorRest().arguments())
      val body = if(ctx.classCreatorRest().classBody() == null) None
      else Some(visitClassBody(ctx.classCreatorRest().classBody()))
      NewObjectExpression(typeArgumentClause, typeIndication, args, body, getSourceLocation(ctx))
    }
  }

  override def visitCreatedName(ctx: KrakatoaParser.CreatedNameContext): TypeIndication = {
    val text = new ParserPrintVisitor(tokens).visit(ctx)
    new StringTypeIndication(text, Nil, ctx.getText, getSourceLocation(ctx))
  }

  override def visitInnerCreator(ctx: KrakatoaParser.InnerCreatorContext): NewObjectExpression = {
    val printer = new ParserPrintVisitor(tokens)
    var fulltext = printer.visit(ctx.javaIdentifier())
    var text = ctx.javaIdentifier().getText
    if (ctx.nonWildcardTypeArgumentsOrDiamond() != null) {
      fulltext += printer.visit(ctx.nonWildcardTypeArgumentsOrDiamond())
      text += ctx.nonWildcardTypeArgumentsOrDiamond().getText
    }
    val typeIndication = new StringTypeIndication(fulltext, Nil, text,
      getSourceLocation(ctx.javaIdentifier()))
    val args = visitArguments(ctx.classCreatorRest().arguments())
    val body = if(ctx.classCreatorRest().classBody() == null) None
    else Some(visitClassBody(ctx.classCreatorRest().classBody()))
    NewObjectExpression(None, typeIndication, args, body, getSourceLocation(ctx))
  }

  override def visitExplicitGenericInvocation(ctx: KrakatoaParser.ExplicitGenericInvocationContext):
  GenericInvocationExpression = {
    val typeParam = visitNonWildcardTypeArguments(ctx.nonWildcardTypeArguments())
    if (ctx.explicitGenericInvocationSuffix().SUPER() != null) {
      if (ctx.explicitGenericInvocationSuffix().superSuffix().javaIdentifier() != null) {
        var text = ctx.explicitGenericInvocationSuffix().SUPER().getText
        text += getWhitespaceLeft(ctx.explicitGenericInvocationSuffix().superSuffix().DOT())
        text += ctx.explicitGenericInvocationSuffix().superSuffix().DOT().getText
        text += getWhitespaceLeft(ctx.explicitGenericInvocationSuffix().superSuffix().javaIdentifier())
        text += ctx.explicitGenericInvocationSuffix().superSuffix().javaIdentifier().getText
        val name = StringNode(text, getSourceLocation(ctx), getWhitespaceLeft(ctx))
        val arguments = if (ctx.explicitGenericInvocationSuffix().superSuffix().arguments()!=null)
          Some(visitArguments(ctx.explicitGenericInvocationSuffix().superSuffix().arguments()))
        else None
        GenericInvocationExpression(typeParam, name, arguments, getSourceLocation(ctx))
      } else {
        val sup = ctx.explicitGenericInvocationSuffix().SUPER()
        val name = StringNode(sup.getText, getSourceLocation(sup), getWhitespaceLeft(sup))
        val arguments = visitArguments(ctx.explicitGenericInvocationSuffix().superSuffix().arguments())
        GenericInvocationExpression(typeParam, name, Some(arguments), getSourceLocation(ctx))
      }
    } else {
      val name = visitJavaIdentifier(ctx.explicitGenericInvocationSuffix().javaIdentifier())
      val arguments = visitArguments(ctx.explicitGenericInvocationSuffix().arguments())
      GenericInvocationExpression(typeParam, name, Some(arguments), getSourceLocation(ctx))
    }
  }

  override def visitNonWildcardTypeArguments(ctx: KrakatoaParser.NonWildcardTypeArgumentsContext): TypeArgumentClause = {
    val parameters = ctx.typeList().typeType().asScala
    val types = parameters.map(visitTypeTypeWithWS).toList
    val clause = TypeArgumentClause(types, getSourceLocation(ctx))
    clause.openAngle.precedingWhitespace = getWhitespaceLeft(ctx.LT())
    clause.closeAngle.precedingWhitespace = getWhitespaceLeft(ctx.GT())
    // WS before closeAngle already covered by trailing WS of last typeParameter
    types.last.trailingWhitespace = ""
    clause
  }

  override def visitTypeArguments(ctx: KrakatoaParser.TypeArgumentsContext): TypeArgumentClause = {
    val args = ctx.typeArgument().asScala.map(x=>addWhitespaceRight(visitTypeArgument(x),x)).toList
    val clause = TypeArgumentClause(args, getSourceLocation(ctx))
    clause.openAngle.precedingWhitespace = getWhitespaceLeft(ctx.LT())
    clause.closeAngle.precedingWhitespace = getWhitespaceLeft(ctx.GT())
    clause
  }

  override def visitTypeArgument(ctx: KrakatoaParser.TypeArgumentContext): TypeArgument = {
    ctx.altNum match {
      case 1 =>
        visitTypeType(ctx.typeType())
      case 2 =>
        val annotations = ctx.annotation().asScala.map(visitAnnotation).toList
        var superType: Option[TypeIndication] = None
        var extendsType: Option[TypeIndication] = None
        if (ctx.EXTENDS() != null) {
          extendsType = Some(visitTypeType(ctx.typeType()))
        }
        if (ctx.SUPER() != null) {
          superType = Some(visitTypeType(ctx.typeType()))
        }
        val arg = BoundedTypeArgument(annotations, superType, extendsType, getSourceLocation(ctx))
        arg.questionMark.precedingWhitespace = getWhitespaceLeft(ctx.QUESTION())
        if (ctx.EXTENDS() != null) {
          arg.extendsKeyword.precedingWhitespace = getWhitespaceLeft(ctx.EXTENDS())
        }
        if (ctx.SUPER() != null) {
          arg.superKeyword.precedingWhitespace = getWhitespaceLeft(ctx.EXTENDS())
        }
        arg
    }
  }

  override def visitLabelList(ctx: KrakatoaParser.LabelListContext): LabelList = {
    val labels = ctx.label().asScala.map(visitLabel).toList
    LabelList(labels, getSourceLocation(ctx))
  }

  override def visitLabel(ctx: KrakatoaParser.LabelContext): StringNode = {
    var label: StringNode = null;
    if (ctx.LABEL_HERE() != null) {
      label = StringNode(ctx.LABEL_HERE().getText, getSourceLocation(ctx.LABEL_HERE()))
    } else if (ctx.LABEL_OLD() != null) {
      label = StringNode(ctx.LABEL_OLD().getText, getSourceLocation(ctx.LABEL_OLD()))
    } else if (ctx.LABEL_PRE() != null) {
      label = StringNode(ctx.LABEL_PRE().getText, getSourceLocation(ctx.LABEL_PRE()))
    } else {
      label = visitJavaIdentifier(ctx.langId().javaIdentifier())
    }
    label
  }

  def visitValArgListAsList(ctx: KrakatoaParser.ValArgListContext): List[Parameter] = {
    var list = List(visitValArg(ctx.valArg()))
    if (ctx.valArgList() != null) {
      val args = visitValArgListAsList(ctx.valArgList())
      list = list.appendedAll(args)
    }
    list
  }

  override def visitValArg(ctx: KrakatoaParser.ValArgContext): Parameter = {
    //Note: these are written as "type id" and then optionally followed by "[]" to indicate an array
    // Currently these brackets [] are added as part of the name instead of the type!
    var text = new ParserPrintVisitor(tokens).visit(ctx)
    text = removeAtSignFromComments(text)
    var paramType: TypeIndication = null
    if (text.contains("integer")) {
      paramType = IntegerSpecTypeIndication(annotations = Nil, srcLocation = getSourceLocation(ctx), "")
    } else if (text.contains("real")) {
      paramType = RealSpecTypeIndication(annotations = Nil, srcLocation = getSourceLocation(ctx), "")
    } else {
      paramType = visitTypeType(ctx.langType().typeType())
    }
    val id = visitJavaIdentifier(ctx.langId().javaIdentifier()).text
    var name = id
    if (ctx.LBRACK() != null) {
      name = name.concat(ctx.LBRACK().getText+ctx.RBRACK().getText)
    }
    Parameter(paramType, StringNode(name, getSourceLocation(ctx.langId()), getWhitespaceLeft(ctx.langId())), Nil, getSourceLocation(ctx))
  }

  override def visitCaseDefinition(ctx: KrakatoaParser.CaseDefinitionContext): CaseDefinition = {
    val name = visitJavaIdentifier(ctx.langId().javaIdentifier())
    val labels = if (ctx.labelList() != null) Some(visitLabelList(ctx.labelList())) else None
    val expr = visitExpression(ctx.langExpr().expression())
    val res = CaseDefinition(name, labels, expr, getSourceLocation(ctx), getWhitespaceLeft(ctx.CASE()))
    res.colon.precedingWhitespace = getWhitespaceLeft(ctx.COLON())
    res.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
    res

  }

  override def visitValStatement(ctx: KrakatoaParser.ValStatementContext): Statement = {
    try {
      ctx.getAltNumber match {
        case 1 =>
          val expr = visitExpression(ctx.langExpr().expression())
          AssertStatement(expr, None, getSourceLocation(ctx), getWhitespaceLeft(ctx.ASSERT()))
        case 2 =>
          val statement = visitBlockStatement(ctx.langStatement().blockStatement())
          GhostStatement(statement, getSourceLocation(ctx), getWhitespaceLeft(ctx.VAL_GHOST()))
        case 3 =>
          val name = visitJavaIdentifier(ctx.langId().javaIdentifier())
          val statement = visitValStatement(ctx.valStatement())
          ForStatement(name, statement, getSourceLocation(ctx), getWhitespaceLeft(ctx.FOR()))
        case x =>
          println(s"Error: unexpected AltNum in ValStatement: $x")
          null
      }
    } catch {
      case e: NotSupportedException =>
        val commented = NonTranslatedStatement("", getSourceLocation(ctx))
        setNonTranslatedTreeNodeText(commented, ctx, e)
        commented
    }
  }

}
