package languages.Krakatoa

import generic.{NotSupportedException, SourceLocation, Helper}
import languages.Krakatoa.syntaxTree as krakatoa
import intermediateRepresentation.syntaxTree as ir
import intermediateRepresentation.syntaxTree.{BinaryExpression, EmptyStatement, OldExpression, StringNode}
import org.antlr.v4.runtime.ParserRuleContext

import scala.collection.mutable.Map

/**
 * Translates Krakatoa-specific AST to intermediate tree
 */
class ToIRVisitor {

  // Store specDeclarations so they can be moved to inside the class
  var specDeclarations: List[ir.SpecDeclarationBlock] = List.empty

  // Store classname so it can be used to identify constructors
  var className: String = ""

  // Store top-level requires clauses so they can be added to behaviors if necessary
  var requiresClauses: List[ir.Precondition] = List.empty

  // Store which assumes-clauses are part of which behavior
  var behaviorPreconditions: Map[String, List[krakatoa.AssumesClause]] = Map.empty[String, List[krakatoa.AssumesClause]]

  // Store which ensures-clause belongs to which behavior
  // This allows us to get which behavior an ensures-clause belongs to, needed for translating assumes-clauses
  var behaviorPostconditions: Map[krakatoa.EnsuresClause, String] = Map.empty[krakatoa.EnsuresClause, String]

  def copyWhitespace(from: krakatoa.KrakatoaTreeNode, to: ir.IrTreeNode): Unit = {
    to.precedingWhitespace = from.precedingWhitespace
    to.trailingWhitespace = from.trailingWhitespace
  }

  def visitStringNode(node: krakatoa.StringNode): ir.StringNode = {
    node match {
      case krakatoa.StringNode(text, src, pred, trail) => ir.StringNode(text, src, pred, trail)
    }
  }

  def visitCompilationUnit(node: krakatoa.CompilationUnit): ir.CompilationUnit = {
    node match {
      case krakatoa.CompilationUnit(pkg, imports, declarations, srcLocation, precedingWhitespace) =>
        var irPkg: Option[ir.PackageDeclaration] = None
        if (pkg.isDefined) irPkg = Some(visitPackageDeclaration(pkg.get))
        var decls = declarations.map(visitTypeDeclaration)
        val classes = decls.filter(a => a.isInstanceOf[ir.ClassDeclaration]).map(a => a.asInstanceOf[ir.ClassDeclaration])

        var specDeclarationsToAdd: List[ir.SpecDeclarationBlock] = List.empty
        for (d <- specDeclarations) {
          var decls: List[ir.SpecDeclaration] = List.empty
          for (e <- d.declarations) {
            // Invariants should not be copied to each class
            if (!e.isInstanceOf[ir.InvariantDeclaration]) decls = decls.appended(e)
          }
          if (!decls.isEmpty) {
            val res = ir.SpecDeclarationBlock(d.startSpec, decls, d.endSpec, d.srcLocation, d.precedingWhitespace)
            specDeclarationsToAdd = specDeclarationsToAdd.appended(res)
          }
        }
        for (c <- classes) {
          // Add SpecDeclarationBlocks to each class (these are e.g. the predicates, lemmas and axioms)
          for (d <- specDeclarationsToAdd) {
              if (!c.body.declarations.contains(d)) {
                c.body.declarations = c.body.declarations.prepended(d)
                c.body.declarations.head.precedingWhitespace = c.body.declarations.head.precedingWhitespace.prepended('\n')
              }
          }
        }
        // Remove SpecDeclarationBlocks from decls now that they have been moved to the classes
        decls = decls.filter(a => !a.isInstanceOf[ir.SpecDeclarationBlock] && !a.isInstanceOf[ir.SpecDeclaration])
        val res = ir.CompilationUnit(irPkg, imports.map(visitImportDeclaration), decls, srcLocation, precedingWhitespace)
        copyWhitespace(node, res)
        res
    }
  }

  def visitPackageDeclaration(node: krakatoa.PackageDeclaration): ir.PackageDeclaration = {
    node match {
      case krakatoa.PackageDeclaration(name, annotations, srcLocation, precedingWhitespace) =>
        val res = ir.PackageDeclaration(visitStringNode(name), annotations.map(visitStringAnnotation), srcLocation, precedingWhitespace)
        res.packageKeyword.precedingWhitespace = node.packageKeyword.precedingWhitespace
        res.semicolon.precedingWhitespace = node.semicolon.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitImportDeclaration(node: krakatoa.ImportDeclaration): ir.ImportDeclaration = {
    node match {
      case krakatoa.ImportDeclaration(name, srcLocation, precedingWhitespace) =>
        val res = ir.ImportDeclaration(visitStringNode(name), srcLocation, precedingWhitespace)
        res.modifiers = node.modifiers.map(visitModifier)
        res.importKeyword.precedingWhitespace = node.importKeyword.precedingWhitespace
        res.semicolon.precedingWhitespace = node.semicolon.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitTypeDeclaration(node: krakatoa.TypeDeclaration): ir.TypeDeclaration = {
    node match {
      case decl: krakatoa.AnnotationTypeDeclaration => visitAnnotationTypeDeclaration(decl)
      case decl: krakatoa.ClassDeclaration => visitClassDeclaration(decl)
      case decl: krakatoa.EnumDeclaration => visitEnumDeclaration(decl)
      case decl: krakatoa.InterfaceDeclaration => visitInterfaceDeclaration(decl)
      case decl: krakatoa.SpecDeclarationBlock => visitSpecDeclarationBlock(decl)
      case decl: krakatoa.EmptyStatement => visitEmptyStatement(decl)
      case decl: krakatoa.NonTranslatedTypeDeclaration => visitNonTranslatedTypeDeclaration(decl)
      case x =>
        println(s"Error: Unknown Krakatoa type declaration $x")
        null
    }
  }

  def visitNonTranslatedTypeDeclaration(decl: krakatoa.NonTranslatedTypeDeclaration): ir.NonTranslatedTypeDeclaration = {
    decl match {
      case krakatoa.NonTranslatedTypeDeclaration(text, srcLocation, precedingWhitespace) =>
        val res = ir.NonTranslatedTypeDeclaration(text, srcLocation, precedingWhitespace)
        copyWhitespace(decl, res)
        res
    }
  }

  def visitStringAnnotation(node: krakatoa.StringAnnotation): ir.StringAnnotation = {
    node match {
      case krakatoa.StringAnnotation(text, srcLocation, precedingWhitespace) =>
        val res = ir.StringAnnotation(text, srcLocation, precedingWhitespace)
        copyWhitespace(node, res)
        res
    }
  }

  def visitModifier(node: generic.syntaxTree.Modifier): ir.Modifier = {
    node match {
      case mod: krakatoa.Modifier => visitModifier(mod)
      case m =>
        println("Error unknown Krakatoa modifier $m")
        null
    }
  }

  def visitModifier(node: krakatoa.Modifier): ir.Modifier = {
    node match {
      case mod: krakatoa.Annotation => visitAnnotation(mod)
      case mod: krakatoa.StaticModifier => visitStaticModifier(mod)
      case mod: krakatoa.ModifierStringNode => visitModifierStringNode(mod)
      case mod: krakatoa.PureModifier => visitPureModifier(mod)
      case mod: krakatoa.SpecMethodModifier => visitSpecMethodModifier(mod)
      case x =>
        println(s"Error: Unknown Krakatoa modifier $x")
        null
    }
  }
  def visitStaticModifier(node: krakatoa.StaticModifier): ir.StaticModifier = {
    node match {
      case krakatoa.StaticModifier(srcLocation, precedingWhitespace) =>
        val res = ir.StaticModifier(srcLocation, precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  
  def visitSpecMethodModifier(node: krakatoa.SpecMethodModifier): ir.Modifier = {
    node match {
      case krakatoa.SpecMethodModifier(srcLocation, precedingWhitespace) =>
        val res = ir.SpecMethodModifier(srcLocation, precedingWhitespace)
        copyWhitespace(node, res)
        res
    }
  }

  def visitAnnotationTypeDeclaration(node: krakatoa.AnnotationTypeDeclaration): ir.AnnotationTypeDeclaration = {
    node match {
      case krakatoa.AnnotationTypeDeclaration(name, body, srcLocation, precedingWhitespace) =>
        val res = ir.AnnotationTypeDeclaration(visitStringNode(name), visitBody(body), srcLocation, precedingWhitespace)
        res.at.precedingWhitespace = node.at.precedingWhitespace
        res.interfaceKeyword.precedingWhitespace = node.interfaceKeyword.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitClassDeclaration(node: krakatoa.ClassDeclaration): ir.ClassDeclaration = {
    node match {
      case krakatoa.ClassDeclaration(name, typeParameters, superclass, interfaces, body, srcLocation, precedingWhitespace) =>
        val irTypeParameters = if (typeParameters.isEmpty) None else Some(visitTypeParameterClause(typeParameters.get))
        val irSuperclass = if (superclass.isEmpty) None else Some(visitSuperclassClause(superclass.get))
        val irInterfaces = if (interfaces.isEmpty) None else Some(visitInterfaceClause(interfaces.get))
        val irName = visitStringNode(name)
        className = irName.text
        val res = ir.ClassDeclaration(irName, irTypeParameters, irSuperclass, irInterfaces, visitBody(body), srcLocation, precedingWhitespace)
        res.classKeyword.precedingWhitespace = node.classKeyword.precedingWhitespace
        res.modifiers = node.modifiers.map(visitModifier)
        copyWhitespace(node, res)
        res
    }
  }

  def visitEnumDeclaration(node: krakatoa.EnumDeclaration): ir.EnumDeclaration = {
    node match {
      case krakatoa.EnumDeclaration(name, interfaces, constants, declarations, comma, semicolon, srcLocation, precedingWhitespace) =>
        val irInterfaces = if (interfaces.isEmpty) None else Some(visitInterfaceClause(interfaces.get))
        val irComma = if (comma.isEmpty) None else Some(visitStringNode(comma.get))
        val irSemicolon = if (semicolon.isEmpty) None else Some(visitStringNode(semicolon.get))
        val res = ir.EnumDeclaration(visitStringNode(name), irInterfaces, constants.map(visitEnumConstant), declarations.map(visitMemberDeclaration), irComma, irSemicolon, srcLocation, precedingWhitespace)
        res.modifiers = node.modifiers.map(visitModifier)
        res.enumKeyword.precedingWhitespace = node.enumKeyword.precedingWhitespace
        res.openBrace.precedingWhitespace = node.openBrace.precedingWhitespace
        res.closeBrace.precedingWhitespace = node.closeBrace.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitInterfaceDeclaration(node: krakatoa.InterfaceDeclaration): ir.InterfaceDeclaration = {
    node match {
      case krakatoa.InterfaceDeclaration(name, typeParameter, superInterfaces, body, srcLocation, precedingWhitespace) =>
        val irTypeParameters = if (typeParameter.isEmpty) None else Some(visitTypeParameterClause(typeParameter.get))
        var irSuperclass = if (superInterfaces.isEmpty) None else Some(visitMultiSuperclassClause(superInterfaces.get))
        val res = ir.InterfaceDeclaration(visitStringNode(name), irTypeParameters, irSuperclass, visitBody(body), srcLocation, precedingWhitespace)
        res.interfaceKeyword.precedingWhitespace = node.interfaceKeyword.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitEmptyStatement(node: krakatoa.EmptyStatement): ir.EmptyStatement = {
    val res = new EmptyStatement(node.srcLocation)
    copyWhitespace(node, res)
    res
  }

  def visitSpecDeclarationBlock(node: krakatoa.SpecDeclarationBlock): ir.SpecDeclarationBlock = {
    node match {
      case krakatoa.SpecDeclarationBlock(startSpec, declarations, endSpec, srcLocation, precedingWhitespace) =>
        val res = ir.SpecDeclarationBlock(startSpec.map(visitStringNode), declarations.flatMap(x=>visitSpecDeclaration(x)), endSpec.map(visitStringNode), srcLocation, precedingWhitespace)
        this.specDeclarations = this.specDeclarations.appended(res)
        copyWhitespace(node, res)
        res
    }
  }

  def visitAnnotation(node: krakatoa.Annotation): ir.Annotation = {
    node match {
      case ann: krakatoa.StringAnnotation => visitStringAnnotation(ann)
      case ann =>
        println(s"Error: Unknown Krakatoa Annotation $ann")
        null
    }
  }

  def visitModifierStringNode(node: krakatoa.ModifierStringNode): ir.ModifierStringNode = {
    node match {
      case krakatoa.ModifierStringNode(str, srcLocation, whitespace) =>
        val res = ir.ModifierStringNode(str, srcLocation, whitespace)
        copyWhitespace(node, res)
        res
    }
  }

  def visitPureModifier(node: krakatoa.PureModifier): ir.PureModifier = {
    node match {
      case krakatoa.PureModifier(srcLocation, precedingWhitespace) =>
        val res = ir.PureModifier(srcLocation, precedingWhitespace)
        copyWhitespace(node, res)
        res
    }
  }

  def visitBody(node: krakatoa.Body): ir.Body = {
    node match {
      case krakatoa.Body(declarations, srcLocation, precedingWhitespace) =>
        val res = ir.Body(declarations.map(visitMemberDeclaration), srcLocation, precedingWhitespace)
        res.openBrace.precedingWhitespace = node.openBrace.precedingWhitespace
        res.closeBrace.precedingWhitespace = node.closeBrace.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitTypeParameterClause(node: krakatoa.TypeParameterClause): ir.TypeParameterClause = {
   node match {
     case krakatoa.TypeParameterClause(parameters, srcLocation, precedingWhitespace) =>
       val res = ir.TypeParameterClause(parameters.map(visitTypeParameter), srcLocation, precedingWhitespace)
       res.openAngle.precedingWhitespace = node.openAngle.precedingWhitespace
       res.closeAngle.precedingWhitespace = node.closeAngle.precedingWhitespace
       copyWhitespace(node, res)
       res
   }
  }

  def visitSuperclassClause(node: krakatoa.SuperclassClause): ir.SuperclassClause = {
    node match {
      case krakatoa.SuperclassClause(superclass, srcLocation, precedingWhitespace) =>
        val res = ir.SuperclassClause(visitTypeIndication(superclass), srcLocation, precedingWhitespace)
        res.extendsKeyword.precedingWhitespace = node.extendsKeyword.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitInterfaceClause(node: krakatoa.InterfaceClause): ir.InterfaceClause = {
    node match {
      case krakatoa.InterfaceClause(interfaces, srcLocation, precedingWhitespace) =>
        val res = ir.InterfaceClause(interfaces.map(visitTypeIndication), srcLocation, precedingWhitespace)
        res.implementsKeyword.precedingWhitespace = node.implementsKeyword.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitEnumConstant(node: krakatoa.EnumConstant): ir.EnumConstant = {
    node match {
      case krakatoa.EnumConstant(name, annotations, arguments, body, srcLocation, precedingWhitespace) =>
        val irArgs = if (arguments.isEmpty) None else Some(visitArguments(arguments.get))
        val irBody = if (body.isEmpty) None else Some(visitBody(body.get))
        val res = ir.EnumConstant(visitStringNode(name), annotations.map(visitStringAnnotation), irArgs, irBody, srcLocation, precedingWhitespace)
        copyWhitespace(node, res)
        res
    }
  }

  def visitMemberDeclaration(node: krakatoa.MemberDeclaration): ir.MemberDeclaration = {
    node match {
      case mem: krakatoa.ClassDeclaration => visitClassDeclaration(mem)
      case mem: krakatoa.EnumDeclaration => visitEnumDeclaration(mem)
      case mem: krakatoa.InterfaceDeclaration => visitInterfaceDeclaration(mem)
      case mem: krakatoa.MethodDeclaration => visitMethodDeclaration(mem)
      case mem: krakatoa.BlockOrSemicolon => visitBlockOrSemicolon(mem)
      case mem: krakatoa.FieldDeclaration => visitFieldDeclaration(mem)
      case mem: krakatoa.AnnotationTypeDeclaration => visitAnnotationTypeDeclaration(mem)
      case mem: krakatoa.SpecDeclarationBlock => visitSpecDeclarationBlock(mem)
      case mem: krakatoa.NonTranslatedMemberDeclaration => visitNonTranslatedMemberDeclaration(mem)
      case m =>
        println(s"Error unknown Krakatoa MemberDeclaration $m")
        null
    }
  }

  def visitNonTranslatedMemberDeclaration(node: krakatoa.NonTranslatedMemberDeclaration): ir.NonTranslatedMemberDeclaration = {
    node match {
      case krakatoa.NonTranslatedMemberDeclaration(text, srcLocation, precedingWhitespace) =>
        ir.NonTranslatedMemberDeclaration(text, srcLocation, precedingWhitespace)
    }
  }

  def visitMultiSuperclassClause(node: krakatoa.MultiSuperclassClause): ir.MultiSuperclassClause = {
    node match {
      case krakatoa.MultiSuperclassClause(superclasses, srcLocation, precedingWhitespace) =>
        val res = ir.MultiSuperclassClause(superclasses.map(visitTypeIndication), srcLocation, precedingWhitespace)
        res.extendsKeyword.precedingWhitespace = node.extendsKeyword.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitSpecDeclaration(node: krakatoa.SpecDeclaration): List[ir.SpecDeclaration] = {
    try {
      node match {
        case decl: krakatoa.PredicateDeclaration => List(visitPredicateDeclaration(decl))
        case decl: krakatoa.InductiveDatatypeDeclaration => List(visitInductiveDatatypeDeclaration(decl))
        case decl: krakatoa.InvariantDeclaration => List(visitInvariantDeclaration(decl))
        case decl: krakatoa.LemmaDeclaration => List(visitLemmaDeclaration(decl))
        case decl: krakatoa.LogicVariableDeclaration => List(visitLogicVariableDeclaration(decl))
        case decl: krakatoa.AxiomDeclaration => List(visitAxiomDeclaration(decl))
        case decl: krakatoa.AxiomaticDeclaration => visitAxiomaticDeclaration(decl)
        case decl: krakatoa.GhostDeclaration => List(visitGhostDeclaration(decl))
        case d =>
          println(s"Error unknown Krakatoa SpecDeclaration $d")
          null
      }
    } catch {
      case p: NotSupportedException =>
        var text = "\n//" + p.getMessage
        text += node.stringRepresentation
        text = text.replace("\n", "\n//")
        text = text.replace("////", "//")
        text = text.replace("@", " ")
        text = text.stripSuffix("//")
        val res = ir.NonTranslatedSpecDeclaration(text, node.srcLocation, node.precedingWhitespace)
        List(res)
    }
  }

  def visitTypeParameter(node: krakatoa.TypeParameter): ir.TypeParameter = {
    node match {
      case krakatoa.TypeParameter(annotations, name, superclass, srcLocation, precedingWhitespace) =>
        val irSuperclass = if (superclass.isEmpty) None else Some(visitGenericsSuperclassClause(superclass.get))
        val res = ir.TypeParameter(annotations.map(visitStringAnnotation), visitTypeIndication(name), irSuperclass, srcLocation, precedingWhitespace)
        copyWhitespace(node, res)
        res
    }
  }

  def visitTypeIndication(node: krakatoa.TypeIndication): ir.TypeIndication = {
    node match {
      case ind: krakatoa.SpecTypeIndication => visitSpecTypeIndication(ind)
      case ind: krakatoa.NoType => visitNoType(ind)
      case ind: krakatoa.StringGenericTypeIndication => visitStringGenericTypeIndication(ind)
      case ind: krakatoa.StringTypeIndication => visitStringTypeIndication(ind)
      case i =>
        println(s"Error unknown Krakatoa TypeIndication $i")
        null
    }
  }

  def visitSpecTypeIndication(node: krakatoa.SpecTypeIndication): ir.TypeIndication = {
    node match {
      case int: krakatoa.IntegerSpecTypeIndication => visitIntegerSpecTypeIndication(int)
      case real: krakatoa.RealSpecTypeIndication => visitRealSpecTypeIndication(real)
      case o =>
        println(s"Error unknown Krakatoa SpecTypeIndication $o")
        null
    }
  }

  def visitIntegerSpecTypeIndication(node: krakatoa.IntegerSpecTypeIndication): ir.TypeIndication = {
    node match {
      case krakatoa.IntegerSpecTypeIndication(fullTypeString, annotations, typeAsString, srcLocation, precedingWhitespace) =>
        val res = ir.IntegerSpecTypeIndication(fullTypeString, annotations.map(visitAnnotation), typeAsString, srcLocation, precedingWhitespace)
        copyWhitespace(node, res)
        res
    }
  }

  def visitRealSpecTypeIndication(node: krakatoa.RealSpecTypeIndication): ir.TypeIndication = {
    node match {
      case krakatoa.RealSpecTypeIndication(fullTypeString, annotations, typeAsString, srcLocation, precedingWhitespace) =>
        val res = ir.RealSpecTypeIndication(fullTypeString, annotations.map(visitAnnotation), typeAsString, srcLocation, precedingWhitespace)
        copyWhitespace(node, res)
        res
    }
  }

  def visitArguments(node: krakatoa.Arguments): ir.Arguments = {
    node match {
      case krakatoa.Arguments(arguments, srcLocation, precedingWhitespace) =>
        val res = ir.Arguments(arguments.map(visitExpression), srcLocation, precedingWhitespace)
        res.openParenthesis.precedingWhitespace = node.openParenthesis.precedingWhitespace
        res.closeParenthesis.precedingWhitespace = node.closeParenthesis.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitMethodDeclaration(node: krakatoa.MethodDeclaration): ir.MethodDeclaration = {
    this.requiresClauses = List.empty
    node match {
      case krakatoa.MethodDeclaration(name, returnType, parameters, dimensions, throws, default, contractWithin, body, srcLocation, precedingWhitespace) =>
        var contract = node.contractBefore.map(visitContract)
        val irReturn = returnType.map(visitTypeIndication)
        val irDims = dimensions.map(visitStringNode)
        val irThrows = throws.map(visitThrowsClause)
        val irDefault = default.map(visitDefaultClause)
        var irContractWithin = contractWithin.map(visitContract)
        val irName = visitStringNode(name)
        // If this method is a constructor, we should remove the assignable clauses if it's not \nothing or \everything
        if (irName.text.equals(this.className)) {
          contract = contract.map(filterAssignableClauses)
          irContractWithin = irContractWithin.map(filterAssignableClauses)
        }
        val res = ir.MethodDeclaration(irName, irReturn, parameters.map(visitParameter), irDims, irThrows, irDefault, irContractWithin, visitBlockOrSemicolon(body), srcLocation, precedingWhitespace)
        res.contractBefore = contract
        res.modifiers = node.modifiers.map(visitModifier)
        res.openParenthesis.precedingWhitespace = node.openParenthesis.precedingWhitespace
        res.closeParenthesis.precedingWhitespace = node.closeParenthesis.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  /**
   * Given a contract, it will remove all assignable clauses that are not \everything or \nothing
   * Assignable clauses in behaviors are also filtered
   * the startSpec and endSpec will be copied if necessary when an assignable clause was removed
   * @param contract Contract
   * @return Contract where assignable clauses that are not \nothing or \everything have been removed
   */
  private def filterAssignableClauses(contract: ir.Contract): ir.Contract = {
    var startSpec: Option[ir.StringNode] = None
    var endSpec: Option[ir.StringNode] = None
    var filteredClauses: List[ir.ContractClause] = List.empty
    val contractClauses = contract.clauses
    for (c <- contractClauses) {
      if (c.isInstanceOf[ir.AssignableClause]) {
        val cl = c.asInstanceOf[ir.AssignableClause]
        if (cl.expressions.size == 1 && cl.expressions.head.isInstanceOf[StringNode]) {
          val expr = cl.expressions.head.asInstanceOf[StringNode]
          if (expr.text.equals("\\nothing") || expr.text.equals("\\everything")) {
            // Keep the assignable clause if it's \nothing or \everything
            filteredClauses = filteredClauses.appended(c)
          }
        }
      } else {
        // Keep clauses that are not assignable clauses
        var clause = c
        c match {
          // Filter assignable clauses from behaviors if they're not \nothing or \everything
          case ir.NormalBehavior(startSpec, hasAlso, privacy, clauses, endSpec, srcLocation, precedingWhitespace) =>
            val behaviorClauses = filterAssignableFromBehaviorClause(clauses)
            clause = ir.NormalBehavior(startSpec, hasAlso, privacy, behaviorClauses, endSpec, srcLocation, precedingWhitespace)
          case ir.ExceptionalBehavior(startSpec, hasAlso, privacy, clauses, endSpec, srcLocation, precedingWhitespace) =>
            val behaviorClauses = filterAssignableFromBehaviorClause(clauses)
            clause = ir.ExceptionalBehavior(startSpec, hasAlso, privacy, behaviorClauses, endSpec, srcLocation, precedingWhitespace)
          case ir.Behavior(startSpec, hasAlso, privacy, clauses, endSpec, srcLocation, precedingWhitespace) =>
            val behaviorClauses = filterAssignableFromBehaviorClause(clauses)
            clause = ir.Behavior(startSpec, hasAlso, privacy, behaviorClauses, endSpec, srcLocation, precedingWhitespace)
          case _ =>
        }
        filteredClauses = filteredClauses.appended(clause)
      }
    }
    val res = ir.Contract(filteredClauses, contract.srcLocation, contract.precedingWhitespace)
    res.startSpec = contract.startSpec
    res.endSpec = contract.endSpec
    res
  }

  /**
   * Removes complicated assignable clauses from all clauses inside a behavior
   * Complicated assignable clauses are anything that is not \nothing or \everything
   * @param clauses Clauses of a behavior
   * @return Clauses of the behavior without complicated assignable clauses
   */
  private def filterAssignableFromBehaviorClause(clauses: List[ir.ContractClause]): List[ir.ContractClause] = {
    var behaviorClauses: List[ir.ContractClause] = List.empty
    for (c <- clauses) {
      if (c.isInstanceOf[ir.AssignableClause]) {
        val cl = c.asInstanceOf[ir.AssignableClause]
        if (cl.expressions.size == 1 && cl.expressions.head.isInstanceOf[StringNode]) {
          val expr = cl.expressions.head.asInstanceOf[StringNode]
          if (expr.text.equals("\\nothing") || expr.text.equals("\\everything")) {
            behaviorClauses = behaviorClauses.appended(c)
          } else {
            // Do nothing, i.e., remove more complicated assignable clause
            // Do not need to do anything to copy startSpec/endSpec
          }
        } else {
          // Do nothing, i.e., remove more complicated assignable clause
        }
      } else {
        behaviorClauses = behaviorClauses.appended(c)
      }
    }
    behaviorClauses
  }

  def visitFieldDeclaration(node: krakatoa.FieldDeclaration): ir.FieldDeclaration = {
    node match {
      case krakatoa.FieldDeclaration(fieldType, declarators, srcLocation, precedingWhitespace) =>
        val res = ir.FieldDeclaration(visitTypeIndication(fieldType), declarators.map(visitVariableDeclarator), srcLocation, precedingWhitespace)
        res.modifiers = node.modifiers.map(visitModifier)
        res.semicolon.precedingWhitespace = node.semicolon.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitBlockOrSemicolon(node: krakatoa.BlockOrSemicolon): ir.BlockOrSemicolon = {
    node match {
      case block: krakatoa.Block => visitBlock(block)
      case semi: krakatoa.EmptyStatement => visitEmptyStatement(semi)
      case o =>
        println(s"Error unknown Krakatoa BlockOrSemicolon $o")
        null
    }
  }

  def visitPredicateDeclaration(node: krakatoa.PredicateDeclaration): ir.SpecDeclaration = {
    node match {
      case krakatoa.PredicateDeclaration(name, labels, parameters, expression, srcLocation, precedingWhitespace) =>
        if (!labels.isEmpty) {
          println(s"Warning: cannot translate predicate with labels ${node.name}")
          // Comment predicate because we cannot translate predicates with labels
          var text = s"// Warning: Predicates with labels are not supported by the Specification Translator. Respective parts commented out.\n" + "// "
          text += node.stringRepresentation.replace("\n", "\n//").replace("@", " ")
          val res = ir.NonTranslatedSpecDeclaration(text, srcLocation, node.tType.precedingWhitespace)
          return res
        }
        val ttype = ir.StringTypeIndication("predicate", Nil, "predicate", srcLocation, precedingWhitespace)
        val res = ir.PredicateDeclaration(visitStringNode(name), ttype, parameters.map(visitParameter), expression.map(visitExpression), srcLocation, precedingWhitespace)
        res.tType.precedingWhitespace = node.tType.precedingWhitespace
        res
    }
  }

  def visitInductiveDatatypeDeclaration(node: krakatoa.InductiveDatatypeDeclaration): ir.SpecDeclaration = {
    node match {
      /**
       * Semantics of an inductive predicate/datatype from ACSL documentation:
       * "The semantics of such a definition is that P is the least fixpoint of the cases, i.e. is the
       * smallest predicate (in the sense that it is false the most often) satisfying the propositions
       * p1, . . . ,pk. "
       */
      case krakatoa.InductiveDatatypeDeclaration(name, labels, parameters, caseDefinitions, srcLocation, precedingWhitespace) =>
        if (!labels.isEmpty) {
          println(s"Warning: cannot translate inductive datatype with labels $name")
          // Comment datatype declaration because we cannot translate datatypes with labels
          var text = s"// Warning: Inductive datatype declarations with labels are not supported by the Specification Translator. Respective parts commented out.\n" + "// "
          text += node.stringRepresentation.replace("\n", "\n//").replace("@", " ")
          val res = ir.NonTranslatedSpecDeclaration(text, srcLocation, precedingWhitespace)
          res
        } else {
          // Translating the cases into postconditions (if there are no labels)
          var contractClauses: List[ir.ContractClause] = List.empty
          for (c <- caseDefinitions) {
            if (c.labels.isEmpty) {
              val clause = ir.Postcondition(visitExpression(c.expression), c.srcLocation, c.precedingWhitespace)
              contractClauses = contractClauses.appended(clause)
            }
          }
          val ttype = ir.StringTypeIndication("predicate", Nil, "predicate", srcLocation, node.inductiveKeyword.precedingWhitespace)
          val res = ir.PredicateDeclaration(visitStringNode(name), ttype, parameters.map(visitParameter), None, srcLocation, precedingWhitespace)
          res.tType.precedingWhitespace = node.inductiveKeyword.precedingWhitespace
          res.contractBefore = Some(ir.Contract(contractClauses, node.srcLocation, precedingWhitespace))
          res.contractBefore.get.trailingWhitespace = res.contractBefore.get.trailingWhitespace + "\n"
          System.err.println(s"Warning: the inductive predicate ${node.name}'s case definitions have been translated into postconditions. Be aware that this may over-approximate things!")
          res.modifiers = res.modifiers.appended(ir.ModifierStringNode("public", node.srcLocation, " "))
          res
        }
    }
  }

  def visitInvariantDeclaration(node: krakatoa.InvariantDeclaration): ir.InvariantDeclaration = {
    node match {
      case krakatoa.InvariantDeclaration(name, expression, srcLocation, precedingWhitespace) =>
        val res = ir.InvariantDeclaration(visitExpression(expression), srcLocation, precedingWhitespace)
        //TODO: IR does not allow names for invariants. Is it okay to just leave this out?
        res.modifiers = node.modifiers.map(visitModifier)
        var visibilityModifier: Boolean = false
        for (m <- res.modifiers.filter(a => a.isInstanceOf[ir.ModifierStringNode]).map(a => a.asInstanceOf[ir.ModifierStringNode])) {
          if (m.text.equals("public") || m.text.equals("private") || m.text.equals("protected") || m.text.equals("package")) {
            visibilityModifier = true
          }
        }
        if (!visibilityModifier) res.modifiers = res.modifiers.prepended(ir.ModifierStringNode("public", node.srcLocation, " "))

        res.invariantKeyword.precedingWhitespace = node.invariantKeyword.precedingWhitespace
        res.semicolon.precedingWhitespace = node.semicolon.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitLemmaDeclaration(node: krakatoa.LemmaDeclaration): ir.SpecDeclaration = {
    node match {
      case krakatoa.LemmaDeclaration(name, labels, expression, srcLocation, precedingWhitespace) =>
        /* From ACSL documentation:
        * "Of course, a complete verification of an ACSL specification has to provide a proof for each lemma."
        *
        * We'll convert it into an axiom AND a warning indicating to the user that this should be proven separately, e.g. using a proof assistant
        */
        if (labels.isEmpty) {
          System.err.println(s"Warning: the axiom ${node.name} should be proven separately.")
          val res = ir.AxiomDeclaration(visitExpression(expression), srcLocation, precedingWhitespace)
          res.axiomKeyword.precedingWhitespace = node.lemmaKeyword.precedingWhitespace
          res
        } else {
          // Comment lemma because we cannot translate lemmas with labels
          var text = s"// Warning: Lemmas with labels not supported by the Specification Translator. Respective parts commented out.\n" + "// "
          text += node.stringRepresentation.replace("\n", "\n//").replace("@", " ")
          val res = ir.NonTranslatedSpecDeclaration(text, srcLocation, node.lemmaKeyword.precedingWhitespace)
          res
        }
    }
  }

  def visitLogicVariableDeclaration(node: krakatoa.LogicVariableDeclaration): ir.SpecDeclaration = {
    if (!node.labels.isEmpty) {
      // comment if there are labels, cannot translate
      throw new NotSupportedException("labels", node.srcLocation)
    }
    val name = visitStringNode(node.name)
    val returnType = visitTypeIndication(node.tType)
    var decl: ir.MemberDeclaration = null
    if (node.parameters.isEmpty) {
      val varDecl = ir.VariableDeclarator(name, node.expression.map(visitExpression), node.srcLocation, node.precedingWhitespace)
      decl = ir.FieldDeclaration(returnType, List(varDecl), node.srcLocation, node.precedingWhitespace)
    } else {
      val params = node.parameters.map(visitParameter)
      var body: ir.BlockOrSemicolon = ir.EmptyStatement(node.srcLocation, node.precedingWhitespace)
      if (!node.expression.isEmpty) {
        body = ir.Block(None, List(ir.ReturnStatement(node.expression.map(visitExpression), node.srcLocation, node.precedingWhitespace)), node.srcLocation, node.precedingWhitespace)
      }
      decl = ir.SpecMethodDeclaration(name, Some(returnType), params, None, body, node.srcLocation, node.precedingWhitespace)
      decl.modifiers = decl.modifiers.appended(ir.SpecMethodModifier(node.srcLocation, " "))
      decl.modifiers = decl.modifiers.appended(ir.PureModifier(node.srcLocation, " "))
      decl.modifiers = decl.modifiers.appended(ir.ModifierStringNode("public", node.srcLocation, " "))
      return decl.asInstanceOf[ir.SpecDeclaration]
    }
    val res = ir.GhostDeclaration(ir.GhostModifier(node.logicKeyword.srcLocation, node.logicKeyword.precedingWhitespace), decl, node.srcLocation, node.precedingWhitespace)
    res.modifiers = res.modifiers.appended(ir.ModifierStringNode("public", node.srcLocation, " "))
    res
  }


  def visitAxiomDeclaration(node: krakatoa.AxiomDeclaration): ir.SpecDeclaration = {
    node match {
      case krakatoa.AxiomDeclaration(name, labels, expression, srcLocation, precedingWhitespace) =>
        if (labels.isEmpty) {
          val res = ir.AxiomDeclaration(visitExpression(expression), srcLocation, precedingWhitespace)
          copyWhitespace(node, res)
          res.axiomKeyword.precedingWhitespace = node.axiomKeyword.precedingWhitespace
          res
        } else {
          // Comment axiom because we cannot translate axioms with labels
          var text = s"// Warning: Axioms with labels not supported by the Specification Translator. Respective parts commented out." + System.lineSeparator() + "// "
          text += node.stringRepresentation.replace("\n", "\n//").replace("@", " ")
          val res = ir.NonTranslatedSpecDeclaration(text, srcLocation, node.axiomKeyword.precedingWhitespace)
          res
        }
    }
  }

  def visitAxiomaticDeclaration(node: krakatoa.AxiomaticDeclaration): List[ir.SpecDeclaration] = {
    // According to the ACSL documentation:
    // "The axiomatic construct is solely a grouping construct, meant to organize declarations that
    // together define the behavior of a collection of types, predicates and logic functions."
    // Based on this, we conclude that the "axiomatic" has no influence on the proof that is generated compared to when they would be written outside of an axiomatic.
    // Therefore, we transform this into the IR by removing this "axiomatic" wrapper, and adding all the individual elements
    var res = node.declarations.flatMap(x=>visitSpecDeclaration(x))
    res.head.precedingWhitespace = node.precedingWhitespace
    res
  }

  def visitGhostDeclaration(node: krakatoa.GhostDeclaration): ir.GhostDeclaration = {
    ir.GhostDeclaration(visitGhostModifier(node.ghost), visitMemberDeclaration(node.decl), node.srcLocation, node.precedingWhitespace)
  }

  def visitGhostModifier(node: krakatoa.GhostModifier): ir.GhostModifier = {
    node match {
      case krakatoa.GhostModifier(srcLocation, precedingWhitespace) => ir.GhostModifier(srcLocation, precedingWhitespace)
    }
  }

  def visitGenericsSuperclassClause(node: krakatoa.GenericsSuperclassClause): ir.GenericsSuperclassClause = {
    node match {
      case krakatoa.GenericsSuperclassClause(superclasses, srcLocation, precedingWhitespace) =>
        val res = ir.GenericsSuperclassClause(superclasses.map(visitTypeIndication), srcLocation, precedingWhitespace)
        res.extendsKeyword.precedingWhitespace = node.extendsKeyword.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitNoType(node: krakatoa.NoType): ir.NoType = {
    node match {
      case krakatoa.NoType(srcLocation) =>
        val res = ir.NoType(srcLocation)
        copyWhitespace(node, res)
        res
    }
  }

  def visitStringGenericTypeIndication(node: krakatoa.StringGenericTypeIndication): ir.StringGenericTypeIndication = {
    node match {
      case krakatoa.StringGenericTypeIndication(fullTypeString, typeParameters, annotations, typeAsString, srcLocation, precedingWhitespace) =>
        val res = ir.StringGenericTypeIndication(fullTypeString, visitTypeParameterClause(typeParameters), annotations.map(visitAnnotation), typeAsString, srcLocation, precedingWhitespace)
        res.modifiers = node.modifiers.map(visitModifier)
        copyWhitespace(node, res)
        res
    }
  }

  def visitStringTypeIndication(node: krakatoa.StringTypeIndication): ir.StringTypeIndication = {
    node match {
      case krakatoa.StringTypeIndication(fullTypeString, annotations, typeAsString, srcLocation, precedingWhitespace, trailingWhitespace) =>
        val typeString = fullTypeString.replace("integer", "int")
        val res = ir.StringTypeIndication(typeString, annotations.map(visitAnnotation), typeAsString.replace("integer", "int"), srcLocation, precedingWhitespace, trailingWhitespace)
        res.modifiers = node.modifiers.map(visitModifier)
        copyWhitespace(node, res)
        res
    }
  }

  def visitExpression(node: krakatoa.Expression): ir.Expression = {
    node match {
      case expr: krakatoa.NestedInvocationExpression => visitNestedInvocationExpression(expr)
      case expr: krakatoa.BinaryExpression => visitBinaryExpression(expr)
      case expr: krakatoa.InstanceofExpression => visitInstanceofExpression(expr)
      case expr: krakatoa.ArrayIndexExpression => visitArrayIndexExpression(expr)
      case expr: krakatoa.MethodCallExpression => visitMethodCallExpression(expr)
      case expr: krakatoa.NewObjectExpression => visitNewObjectExpression(expr)
      case expr: krakatoa.NewArrayExpression => visitNewArrayExpression(expr)
      case expr: krakatoa.CastExpression => visitCastExpression(expr)
      case expr: krakatoa.PostfixExpression => visitPostfixExpression(expr)
      case expr: krakatoa.PrefixExpression => visitPrefixExpression(expr)
      case expr: krakatoa.TernaryExpression => visitTernaryExpression(expr)
      case expr: krakatoa.LambdaExpression => visitLambdaExpression(expr)
      case expr: krakatoa.MethodReferenceExpression => visitMethodReferenceExpression(expr)
      case expr: krakatoa.ParExpression => visitParExpression(expr)
      case expr: krakatoa.Literal => visitLiteral(expr)
      case expr: krakatoa.ClassName => visitClassName(expr)
      case expr: krakatoa.GenericInvocationExpression => visitGenericInvocationExpression(expr)
      case expr: krakatoa.ResultExpression => visitResultExpression(expr)
      case expr: krakatoa.OldExpression => visitOldExpression(expr)
      case expr: krakatoa.AtExpression => visitAtExpression(expr)
      case expr: krakatoa.FreshExpression => visitFreshExpression(expr)
      case expr: krakatoa.LabeledExpression => visitLabeledExpression(expr)
      case expr: krakatoa.ForallExpression => visitForallExpression(expr)
      case expr: krakatoa.ExistsExpression => visitExistsExpression(expr)
      case expr: krakatoa.StringNode => visitStringNode(expr)
      case e =>
        println(s"Error unknown Krakatoa Expression $e")
        null
    }
  }

  def visitResultExpression(node: krakatoa.ResultExpression): ir.ResultExpression = {
    node match {
      case krakatoa.ResultExpression(srcLocation, precedingWhitespace) =>
        ir.ResultExpression(srcLocation, precedingWhitespace)
    }
  }

  def visitThrowsClause(node: krakatoa.Throws): ir.ThrowsClause = {
    node match {
      case krakatoa.Throws(exceptions, srcLocation, precedingWhitespace) =>
        val res = ir.ThrowsClause(exceptions.map(visitStringNode), srcLocation, precedingWhitespace)
        res.throwsKeyword.precedingWhitespace = node.throwsKeyword.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitDefaultClause(node: krakatoa.DefaultClause): ir.DefaultClause = {
    node match {
      case krakatoa.DefaultClause(value, srcLocation, precedingWhitespace) =>
        val res = ir.DefaultClause(visitStringNode(value), srcLocation, precedingWhitespace)
        res.defaultKeyword.precedingWhitespace = node.defaultKeyword.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitContract(node: krakatoa.Contract): ir.Contract = {
    node match {
      case krakatoa.Contract(clauses, srcLocation, precedingWhitespace) =>
        var visitedClauses = clauses.map(visitContractClause)
        // If there are only top-level requires-clauses & behaviors, then the requires-clauses should be removed
        // Since the requires-clauses are copied to each behavior (as well as measuredBy-clauses)
        var onlyTopLevelRequires = true
        var hasBehavior = false
        for (r <- visitedClauses) {
          if (r.isInstanceOf[ir.Behavior]) hasBehavior = true
          if (!(r.isInstanceOf[ir.Precondition] || r.isInstanceOf[ir.Behavior])) onlyTopLevelRequires = false
        }
        if (hasBehavior) {
          if (onlyTopLevelRequires) {
            var startSpec: Option[StringNode] = None
            if (visitedClauses.head.isInstanceOf[ir.Precondition]) {
              startSpec = visitedClauses.head.startSpec
            }
            // Remove the top-level requires clauses
            visitedClauses = visitedClauses.filter(a => !a.isInstanceOf[ir.Precondition] && !a.isInstanceOf[ir.MeasuredByClause])

            // Copy startSpec of what used to be the first clause
            if (startSpec.isDefined) visitedClauses.head.startSpec = startSpec
          } else {
            // There are other clauses than requires on the top-level & there are other behaviors
            // This means that we need to put all these top-level clauses into a Behavior
            val clauses = visitedClauses.filter(a => !a.isInstanceOf[ir.Behavior])
            // Copy whitespace from subclause from other behavior
            val preWhitespace = visitedClauses.filter(a => a.isInstanceOf[ir.Behavior]).map(a => a.asInstanceOf[ir.Behavior]).last.clauses.head.precedingWhitespace
            val newStartSpec = node.clauses.head.startSpec.map(visitStringNode)
            clauses.head.startSpec = None
            clauses.head.precedingWhitespace = preWhitespace
            val specCase = ir.Behavior(newStartSpec, false, None, clauses, None, srcLocation, " ")
            visitedClauses = visitedClauses.filter(a => a.isInstanceOf[ir.Behavior]).prepended(specCase)
          }
        }

        var startSpec: Option[ir.StringNode] = None
        var endSpec: Option[ir.StringNode] = None
        if (node.startSpec.isDefined) {
          startSpec = node.startSpec.map(visitStringNode)
        } else {
          startSpec = visitedClauses.head.startSpec
        }
        if (node.endSpec.isDefined) {
          endSpec = node.endSpec.map(visitStringNode)
        } else {
          endSpec = visitedClauses.last.endSpec
        }
        if (visitedClauses.head.startSpec.isDefined && visitedClauses.head.startSpec.get.text.contains("/*")) {
          visitedClauses.head.startSpec = None
        }
        visitedClauses.last.endSpec = None
        val res = ir.Contract(visitedClauses, srcLocation)
        if (startSpec.isDefined && startSpec.get.text.contains("/*")) res.startSpec = startSpec
        res.endSpec = endSpec
        copyWhitespace(node, res)
        res
    }
  }

  def visitParameter(node: krakatoa.Parameter): ir.Parameter = {
    node match {
      case krakatoa.Parameter(parameterType, name, modifiers, srcLocation, precedingWhitespace) =>
        val res = ir.Parameter(visitTypeIndication(parameterType), visitStringNode(name), srcLocation, precedingWhitespace)
        res.modifiers = modifiers.map(visitModifier)
        copyWhitespace(node, res)
        res
    }
  }

  def visitVariableDeclarator(node: krakatoa.VariableDeclarator): ir.VariableDeclarator = {
    node match {
      case krakatoa.VariableDeclarator(name, initializer, srcLocation, precedingWhitespace) =>
        val irInit = if (initializer.isEmpty) None else Some(visitInitializer(initializer.get))
        val res = ir.VariableDeclarator(visitStringNode(name), irInit, srcLocation, precedingWhitespace)
        res.equalSign.precedingWhitespace = node.equalSign.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitBlock(node: krakatoa.Block): ir.Block = {
    node match {
      case krakatoa.Block(static, statements, srcLocation, precedingWhitespace) =>
        val irStatic = static.map(visitModifier)
        val res = ir.Block(irStatic, statements.map(visitStatement), srcLocation, precedingWhitespace)
        res.openBrace.precedingWhitespace = node.openBrace.precedingWhitespace
        res.closeBrace.precedingWhitespace = node.closeBrace.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitContractClause(node: krakatoa.ContractClause): ir.ContractClause = {
    try {
      node match {
        case m: krakatoa.AssignsClause =>
          val res = ir.AssignableClause(m.expressions.map(visitExpression), m.srcLocation, m.precedingWhitespace)
          res.assignableKeyword.precedingWhitespace = m.assignableKeyword.precedingWhitespace
          res.startSpec = m.startSpec.map(visitStringNode)
          res.endSpec = m.endSpec.map(visitStringNode)
          res
        case m: krakatoa.AccessibleClause =>
          val res = ir.AccessibleClause(m.expressions.map(visitExpression), m.srcLocation, m.precedingWhitespace)
          res.accessibleKeyword.precedingWhitespace = m.accessibleKeyword.precedingWhitespace
          res.startSpec = m.startSpec.map(visitStringNode)
          res.endSpec = m.endSpec.map(visitStringNode)
          res
        case m: krakatoa.AssumesClause =>
          null
          //TODO: Check how to properly deal with clauses that don't need to be translated
        case m: krakatoa.RequiresClause =>
          val res = ir.Precondition(visitExpression(m.expression), m.srcLocation, m.precedingWhitespace)
          res.requiresKeyword.precedingWhitespace = m.requiresKeyword.precedingWhitespace
          res.startSpec = m.startSpec.map(visitStringNode)
          res.endSpec = m.endSpec.map(visitStringNode)
          // Need to store RequiresClause if it is not part of a behavior
          requiresClauses = requiresClauses.appended(res)
          res
        case m: krakatoa.EnsuresClause =>
          var expr: ir.Expression = visitExpression(m.expression)
          if (behaviorPostconditions.contains(m)) {
            val assumes = behaviorPreconditions.get(behaviorPostconditions.get(m).get)
            if (assumes.isDefined && !assumes.get.isEmpty) {
              // Transform assumes A ensures B into ensures A ==> B
              // First combine all assumptions into one single expression
              var assumptions: ir.Expression = visitExpression(assumes.get.head.expression)
              for (a <- assumes.get.tail)
                val op = StringNode("&&", a.expression.srcLocation, " ")
                val right = visitExpression(a.expression)
                assumptions = ir.BinaryExpression(assumptions, op, right, a.expression.srcLocation)
              val op = StringNode("==>", m.expression.srcLocation, " ")
              val oldAssumptions = ir.OldExpression(assumptions, None, m.expression.srcLocation)
              expr = ir.BinaryExpression(oldAssumptions, op, expr, m.expression.srcLocation, m.expression.precedingWhitespace)
            }
          }
          val res = ir.Postcondition(expr, m.srcLocation, m.precedingWhitespace)
          res.ensuresKeyword.precedingWhitespace = m.ensuresKeyword.precedingWhitespace
          res.startSpec = m.startSpec.map(visitStringNode)
          res.endSpec = m.endSpec.map(visitStringNode)
          res
        case m: krakatoa.DecreasesClause =>
          val res = ir.MeasuredByClause(visitExpression(m.expression), None, m.srcLocation, m.precedingWhitespace)
          res.measuredByKeyword.precedingWhitespace = m.measuredByKeyword.precedingWhitespace
          res.semicolon.precedingWhitespace = m.semicolon.precedingWhitespace
          res.startSpec = m.startSpec.map(visitStringNode)
          res.endSpec = m.endSpec.map(visitStringNode)
          res
        case m: krakatoa.NormalBehavior =>
          val preconditions = m.clauses.filter(str => (str.isInstanceOf[krakatoa.AssumesClause]))
          behaviorPreconditions += m.behaviorName.text -> preconditions.map(p => p.asInstanceOf[krakatoa.AssumesClause])
          val postconditions = m.clauses.filter(str => (str.isInstanceOf[krakatoa.EnsuresClause]))
          for (p <- postconditions)
            behaviorPostconditions += p.asInstanceOf[krakatoa.EnsuresClause] -> m.behaviorName.text
          val clausesToVisit: List[krakatoa.ContractClause] = m.clauses.filter(str => (!str.isInstanceOf[krakatoa.AssumesClause]))
          var visitedClauses: List[ir.ContractClause] = clausesToVisit.map(visitContractClause)
          // Make precedingWhitespace appropriate for new clauses that will be added to the behavior
          val preWhitespace = getPrecedingWhitespaceForRequires(m.behaviorName, visitedClauses.head)
          for (c <- this.requiresClauses) {
            val precondition = ir.Precondition(c.expression, c.srcLocation, c.precedingWhitespace)
            if (precondition.startSpec.isDefined && precondition.startSpec.get.text.contains("/*@")) precondition.startSpec = None // Remove startSpec if it's /*@
            precondition.precedingWhitespace = preWhitespace
            visitedClauses = visitedClauses.prepended(precondition)
          }
          val res = ir.NormalBehavior(m.startSpec.map(visitStringNode), m.hasAlso, m.privacy.map(visitStringNode), visitedClauses, m.endSpec.map(visitStringNode), m.srcLocation, m.precedingWhitespace)
          res.behaviorKeyword.precedingWhitespace = m.behaviorKeyword.precedingWhitespace
          res
        case m: krakatoa.ExceptionalBehavior =>
          val preconditions = m.clauses.filter(str => str.isInstanceOf[krakatoa.AssumesClause])
          behaviorPreconditions += m.behaviorName.text -> preconditions.map(p => p.asInstanceOf[krakatoa.AssumesClause])
          var visitedClauses = m.clauses.map(visitContractClause)
          // Make precedingWhitespace appropriate for new clauses that will be added to the behavior
          val preWhitespace = getPrecedingWhitespaceForRequires(m.behaviorName, visitedClauses.head)
          for (c <- this.requiresClauses) {
            val precondition = ir.Precondition(c.expression, c.srcLocation, c.precedingWhitespace)
            if (precondition.startSpec.isDefined && precondition.startSpec.get.text.contains("/*@")) precondition.startSpec = None // Remove startSpec if it's /*@
            precondition.precedingWhitespace = preWhitespace
            visitedClauses = visitedClauses.prepended(precondition)
          }
          val res = ir.ExceptionalBehavior(m.startSpec.map(visitStringNode), m.hasAlso, m.privacy.map(visitStringNode), visitedClauses, m.endSpec.map(visitStringNode), m.srcLocation, m.precedingWhitespace)
          res.behaviorKeyword.precedingWhitespace = m.behaviorKeyword.precedingWhitespace
          res
        case m: krakatoa.Behavior =>
          val preconditions = m.clauses.filter(str => (str.isInstanceOf[krakatoa.AssumesClause]))
          behaviorPreconditions += m.behaviorName.text -> preconditions.map(p => p.asInstanceOf[krakatoa.AssumesClause])
          val postconditions = m.clauses.filter(str => (str.isInstanceOf[krakatoa.EnsuresClause]))
          for (p <- postconditions)
            behaviorPostconditions += p.asInstanceOf[krakatoa.EnsuresClause] -> m.behaviorName.text
          val clausesToVisit = m.clauses.filter(str => (!str.isInstanceOf[krakatoa.AssumesClause]))
          var visitedClauses = clausesToVisit.map(visitContractClause)
          // Make precedingWhitespace appropriate for new clauses that will be added to the behavior
          val preWhitespace = getPrecedingWhitespaceForRequires(m.behaviorName, visitedClauses.head)
          for (c <- this.requiresClauses) {
            val precondition = ir.Precondition(c.expression, c.srcLocation, c.precedingWhitespace)
            if (precondition.startSpec.isDefined && precondition.startSpec.get.text.contains("/*@")) precondition.startSpec = None // Remove startSpec if it's /*@
            precondition.precedingWhitespace = preWhitespace
            visitedClauses = visitedClauses.prepended(precondition)
          }
          val res = ir.Behavior(m.startSpec.map(visitStringNode), m.hasAlso, m.privacy.map(visitStringNode), visitedClauses, m.endSpec.map(visitStringNode), m.srcLocation, m.precedingWhitespace)
          res.behaviorKeyword.precedingWhitespace = m.behaviorKeyword.precedingWhitespace
          res
        case m: krakatoa.SignalsClause =>
          val res = ir.ExceptionalPostcondition(visitTypeIndication(m.exceptionType), m.exceptionName.map(visitStringNode), Some(visitExpression(m.expression.get)), m.srcLocation, m.precedingWhitespace)
          res.signalsKeyword.precedingWhitespace = m.signalsKeyword.precedingWhitespace
          res.startSpec = m.startSpec.map(visitStringNode)
          res.endSpec = m.endSpec.map(visitStringNode)
          res
        case m: krakatoa.NonTranslatedContractClause =>
          val res = ir.NonTranslatedContractClause(m.text, m.srcLocation, m.precedingWhitespace)
          res.startSpec = m.startSpec.map(visitStringNode)
          res.endSpec = m.endSpec.map(visitStringNode)
          res
      }
    } catch {
      case p:NotSupportedException =>
        var text = "\n//"+p.getMessage
        text += node.stringRepresentation
        text = text.replace("\n", "\n//")
        text = text.replace("@", " ")
        text = text.replace("////", "//")
        text = text.stripSuffix("//")
        val res = ir.NonTranslatedContractClause(text, node.srcLocation, node.precedingWhitespace)
        res
    }
  }

  /**
   * Get precedingWhitespace for a (top-level) requires clause that is added to a behavior clause
   * It will remove comments within the whitespace to avoid duplication
   */
  private def getPrecedingWhitespaceForRequires(behaviorName: krakatoa.StringNode, firstClause: ir.ContractClause): String = {
    var preWhitespace = behaviorName.precedingWhitespace + behaviorName.trailingWhitespace
    if (firstClause.isInstanceOf[ir.Precondition]) {
      preWhitespace = preWhitespace + firstClause.asInstanceOf[ir.Precondition].requiresKeyword.precedingWhitespace.replaceAll("\\n\\s*@\\s*//.*", "")
    } else if (firstClause.isInstanceOf[ir.Postcondition]) {
      preWhitespace = preWhitespace + firstClause.asInstanceOf[ir.Postcondition].ensuresKeyword.precedingWhitespace.replaceAll("\\n\\s*@\\s*//.*", "")
    } else if (firstClause.isInstanceOf[ir.AssignableClause]) {
      preWhitespace = preWhitespace + firstClause.asInstanceOf[ir.AssignableClause].assignableKeyword.precedingWhitespace.replaceAll("\\n\\s*@\\s*//.*", "")
    } else if (firstClause.isInstanceOf[ir.ExceptionalPostcondition]) {
      preWhitespace = preWhitespace + firstClause.asInstanceOf[ir.ExceptionalPostcondition].signalsKeyword.precedingWhitespace.replaceAll("\\n\\s*@\\s*//.*", "")
    }
    if (firstClause.startSpec.isDefined) {
      // Do not copy comments that are in the whitespace
      val startSpecWhitespace = Helper.getNewlines(firstClause.startSpec.get)
      preWhitespace = preWhitespace + startSpecWhitespace
    }
    preWhitespace
  }

  def visitInitializer(node: krakatoa.Initializer): ir.Initializer = {
    node match {
      case init: krakatoa.ArrayInitializer => visitArrayInitializer(init)
      case init: krakatoa.Expression => visitExpression(init)
      case i =>
        println(s"Error unknown Krakatoa Initializer $i")
        null
    }
  }
  
  def visitStatement(node: krakatoa.Statement): ir.Statement = {
    try {
      node match {
        case stat: krakatoa.LocalVariableDeclaration => visitLocalVariableDeclaration(stat)
        case stat: krakatoa.Block => visitBlock(stat)
        case stat: krakatoa.IfStatement => visitIfStatement(stat)
        case stat: krakatoa.ForLoop => visitForLoop(stat)
        case stat: krakatoa.WhileLoop => visitWhileLoop(stat)
        case stat: krakatoa.DoWhileLoop => visitDoWhileLoop(stat)
        case stat: krakatoa.Try => visitTry(stat)
        case stat: krakatoa.Switch => visitSwitch(stat)
        case stat: krakatoa.Synchronized => visitSynchronized(stat)
        case stat: krakatoa.Return => visitReturn(stat)
        case stat: krakatoa.ThrowStatement => visitThrowStatement(stat)
        case stat: krakatoa.Break => visitBreak(stat)
        case stat: krakatoa.Continue => visitContinue(stat)
        case stat: krakatoa.EmptyStatement => visitEmptyStatement(stat)
        case stat: krakatoa.ExpressionStatement => visitExpressionStatement(stat)
        case stat: krakatoa.LabeledStatement => visitLabeledStatement(stat)
        case stat: krakatoa.AssertStatement => visitAssertStatement(stat)
        case stat: krakatoa.GhostStatement => visitGhostStatement(stat)
        case stat: krakatoa.Assert => visitAssert(stat)
        case stat: krakatoa.ForStatement => visitForStatement(stat)
        case stat: krakatoa.KrakatoaStatementBlock => visitKrakatoaStatementBlock(stat)
        case stat: krakatoa.NonTranslatedStatement => visitNonTranslatedStatement(stat)
        case s =>
          println(s"Error unknown Krakatoa statement $s")
          null
      }
    } catch {
      case p: NotSupportedException =>
        var text = "\n//"+p.getMessage
        text += "\n//" + node.stringRepresentation.replace("\n", "\n//").replace("@", " ")
        val res = ir.NonTranslatedStatement(text, node.srcLocation, node.precedingWhitespace)
        res
    }
  }

  def visitNonTranslatedStatement(node: krakatoa.NonTranslatedStatement): ir.NonTranslatedStatement = {
    node match {
      case krakatoa.NonTranslatedStatement(text, srcLocation, precedingWhitespace) =>
        ir.NonTranslatedStatement(text, srcLocation, precedingWhitespace)
    }
  }

  def visitKrakatoaStatementBlock(block: krakatoa.KrakatoaStatementBlock): ir.JmlStatementBlock = {
    block match {
      case krakatoa.KrakatoaStatementBlock(_startSpec, statements, _endSpec, srcLocation, precedingWhitespace) =>
        ir.JmlStatementBlock(_startSpec.map(visitStringNode), statements.map(visitStatement), _endSpec.map(visitStringNode), srcLocation, precedingWhitespace)
    }
  }

  def visitArrayInitializer(node: krakatoa.ArrayInitializer): ir.ArrayInitializer = {
    node match {
      case krakatoa.ArrayInitializer(initializers, trailingComma, srcLocation, precedingWhitespace) =>
        val irComma = if (trailingComma.isEmpty) None else Some(visitStringNode(trailingComma.get))
        val res = ir.ArrayInitializer(initializers.map(visitInitializer), irComma, srcLocation, precedingWhitespace)
        res.openBrace.precedingWhitespace = node.openBrace.precedingWhitespace
        res.closeBrace.precedingWhitespace = node.closeBrace.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitNestedInvocationExpression(node: krakatoa.NestedInvocationExpression): ir.NestedInvocationExpression = {
    node match {
      case krakatoa.NestedInvocationExpression(left, right, srcLocation, precedingWhitespace) =>
        val res = ir.NestedInvocationExpression(visitExpression(left), visitExpression(right), srcLocation, precedingWhitespace)
        res.dot.precedingWhitespace = node.dot.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitBinaryExpression(node: krakatoa.BinaryExpression): ir.BinaryExpression = {
    node match {
      case krakatoa.BinaryExpression(left, operator, right, srcLocation, precedingWhitespace) =>
        val res = ir.BinaryExpression(visitExpression(left), visitStringNode(operator), visitExpression(right), srcLocation, precedingWhitespace)
        copyWhitespace(node, res)
        res
    }
  }

  def visitInstanceofExpression(node: krakatoa.InstanceofExpression): ir.InstanceofExpression = {
    node match {
      case krakatoa.InstanceofExpression(left, right, srcLocation, precedingWhitespace) =>
        val res = ir.InstanceofExpression(visitExpression(left), visitTypeIndication(right), srcLocation, precedingWhitespace)
        copyWhitespace(node, res)
        res
    }
  }

  def visitArrayIndexExpression(node: krakatoa.ArrayIndexExpression): ir.ArrayIndexExpression = {
    node match {
      case krakatoa.ArrayIndexExpression(arrayExpression, indexExpression, srcLocation, precedingWhitespace) =>
        val res = ir.ArrayIndexExpression(visitExpression(arrayExpression), visitIndexExpression(indexExpression), srcLocation, precedingWhitespace)
        res.openBracket.precedingWhitespace = node.openBracket.precedingWhitespace
        res.closeBracket.precedingWhitespace = node.closeBracket.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitIndexExpression(node: krakatoa.IndexExpression): ir.IndexExpression = {
    node match {
      case idx: krakatoa.IndexAllExpression => visitIndexAllExpression(idx)
      case idx: krakatoa.RangeExpression => visitRangeExpression(idx)
      case idx: krakatoa.Expression => visitExpression(idx)
      case i =>
        println(s"Error unknown Krakatoa IndexExpression $i")
        null
    }
  }

  def visitIndexAllExpression(node: krakatoa.IndexAllExpression): ir.IndexAllExpression = {
    node match {
      case krakatoa.IndexAllExpression(srcLocation, precedingWhitespace) =>
        val res = ir.IndexAllExpression(srcLocation, precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitRangeExpression(node: krakatoa.RangeExpression): ir.RangeExpression = {
    node match {
      case krakatoa.RangeExpression(start, end, srcLocation, precedingWhitespace) =>
        val res = ir.RangeExpression(visitExpression(start), visitExpression(end), srcLocation, precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitMethodCallExpression(node: krakatoa.MethodCallExpression): ir.MethodCallExpression = {
    node match {
      case krakatoa.MethodCallExpression(name, arguments, srcLocation, precedingWhitespace) =>
        val res = ir.MethodCallExpression(visitStringNode(name), visitArguments(arguments), srcLocation, precedingWhitespace)
        copyWhitespace(node, res)
        res
    }
  }

  def visitNewObjectExpression(node: krakatoa.NewObjectExpression): ir.NewObjectExpression = {
    node match {
      case krakatoa.NewObjectExpression(typeArguments, typeIndication, arguments, body, srcLocation, precedingWhitespace) =>
        val typeArgs = if (typeArguments.isEmpty) None else Some(visitTypeArgumentClause(typeArguments.get))
        val irBody = if (body.isEmpty) None else Some(visitBody(body.get))
        val res = ir.NewObjectExpression(typeArgs, visitTypeIndication(typeIndication), visitArguments(arguments), irBody, srcLocation, precedingWhitespace)
        res.newKeyword.precedingWhitespace = node.newKeyword.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitNewArrayExpression(node: krakatoa.NewArrayExpression): ir.NewArrayExpression = {
    node match {
      case krakatoa.NewArrayExpression(typeIndication, dimensions, initializer, srcLocation, precedingWhitespace) =>
        val irInit = if (initializer.isEmpty) None else Some(visitArrayInitializer(initializer.get))
        val res = ir.NewArrayExpression(visitTypeIndication(typeIndication), dimensions.map(visitDimension), irInit, srcLocation, precedingWhitespace)
        res.newKeyword.precedingWhitespace = node.newKeyword.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitDimension(node: krakatoa.Dimension): ir.Dimension = {
    node match {
      case krakatoa.Dimension(sizeExpression, srcLocation, precedingWhitespace) =>
        val irSize = if (sizeExpression.isEmpty) None else Some(visitExpression(sizeExpression.get))
        val res = ir.Dimension(irSize, srcLocation, precedingWhitespace)
        res.openBracket.precedingWhitespace = node.openBracket.precedingWhitespace
        res.closeBracket.precedingWhitespace = node.closeBracket.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitCastExpression(node: krakatoa.CastExpression): ir.CastExpression = {
    node match {
      case krakatoa.CastExpression(annotations, typeIndications, expression, srcLocation, precedingWhitespace) =>
        val res = ir.CastExpression(annotations.map(visitStringAnnotation), typeIndications.map(visitTypeIndication), visitExpression(expression), srcLocation, precedingWhitespace)
        res.openParenthesis.precedingWhitespace = node.openParenthesis.precedingWhitespace
        res.closeParenthesis.precedingWhitespace = node.closeParenthesis.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitPostfixExpression(node: krakatoa.PostfixExpression): ir.PostfixExpression = {
    node match {
      case krakatoa.PostfixExpression(expression, postfix, srcLocation, precedingWhitespace) =>
        val res = ir.PostfixExpression(visitExpression(expression), visitStringNode(postfix), srcLocation, precedingWhitespace)
        copyWhitespace(node, res)
        res
    }
  }

  def visitPrefixExpression(node: krakatoa.PrefixExpression): ir.PrefixExpression = {
    node match {
      case krakatoa.PrefixExpression(expression, prefix, srcLocation, precedingWhitespace) =>
        val res = ir.PrefixExpression(visitExpression(expression), visitStringNode(prefix), srcLocation, precedingWhitespace)
        copyWhitespace(node, res)
        res
    }
  }

  def visitTernaryExpression(node: krakatoa.TernaryExpression): ir.TernaryExpression = {
    node match {
      case krakatoa.TernaryExpression(condition, thenExpression, elseExpression, srcLocation, precedingWhitespace) =>
        val res = ir.TernaryExpression(visitExpression(condition), visitExpression(thenExpression), visitExpression(elseExpression), srcLocation, precedingWhitespace)
        res.questionmark.precedingWhitespace = node.questionmark.precedingWhitespace
        res.colon.precedingWhitespace = node.colon.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitLambdaExpression(node: krakatoa.LambdaExpression): ir.LambdaExpression = {
    node match {
      case krakatoa.LambdaExpression(hasParentheses, parameters, body, srcLocation, precedingWhitespace) =>
        val res = ir.LambdaExpression(hasParentheses, parameters.map(visitParameter), visitLambdaBody(body), srcLocation, precedingWhitespace)
        res.arrow.precedingWhitespace = node.arrow.precedingWhitespace
        res.openParenthesis.precedingWhitespace = node.openParenthesis.precedingWhitespace
        res.closeParenthesis.precedingWhitespace = node.closeParenthesis.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitLambdaBody(node: krakatoa.LambdaBody): ir.LambdaBody = {
    node match {
      case body: krakatoa.Expression => visitExpression(body)
      case b =>
        println(s"Error unknown Krakatoa LambdaBody $b")
        null
    }
  }

  def visitMethodReferenceExpression(node: krakatoa.MethodReferenceExpression): ir.MethodReferenceExpression = {
    node match {
      case krakatoa.MethodReferenceExpression(left, typeArgument, right, srcLocation, precedingWhitespace) =>
        val irArg = if (typeArgument.isEmpty) None else Some(visitTypeArgumentClause(typeArgument.get))
        val res = ir.MethodReferenceExpression(visitExpression(left), irArg, visitExpression(right), srcLocation, precedingWhitespace)
        res.doubleColon.precedingWhitespace = node.doubleColon.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitParExpression(node: krakatoa.ParExpression): ir.ParExpression = {
    node match {
      case krakatoa.ParExpression(expression, srcLocation, precedingWhitespace) =>
        val res = ir.ParExpression(visitExpression(expression), srcLocation, precedingWhitespace)
        res.openParenthesis.precedingWhitespace = node.openParenthesis.precedingWhitespace
        res.closeParenthesis.precedingWhitespace = node.closeParenthesis.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitLiteral(node: krakatoa.Literal): ir.Literal = {
    node match {
      case krakatoa.Literal(value, srcLocation, precedingWhitespace) =>
        val res = ir.Literal(visitStringNode(value), srcLocation, precedingWhitespace)
        copyWhitespace(node, res)
        res
    }
  }

  def visitClassName(node: krakatoa.ClassName): ir.ClassName = {
    node match {
      case krakatoa.ClassName(typeIndication, srcLocation, precedingWhitespace) =>
        val res = ir.ClassName(visitTypeIndication(typeIndication), srcLocation, precedingWhitespace)
        res.classKeyword.precedingWhitespace = node.classKeyword.precedingWhitespace
        res.dot.precedingWhitespace = node.dot.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitGenericInvocationExpression(node: krakatoa.GenericInvocationExpression): ir.GenericInvocationExpression = {
    node match {
      case krakatoa.GenericInvocationExpression(typeArgumentClause, name, arguments, srcLocation, precedingWhitespace) =>
        val irArgs = if (arguments.isEmpty) None else Some(visitArguments(arguments.get))
        val res = ir.GenericInvocationExpression(visitTypeArgumentClause(typeArgumentClause), visitStringNode(name), irArgs, srcLocation, precedingWhitespace)
        copyWhitespace(node, res)
        res
    }
  }

  def visitTypeArgumentClause(node: krakatoa.TypeArgumentClause): ir.TypeArgumentClause = {
    node match {
      case krakatoa.TypeArgumentClause(parameters, srcLocation, precedingWhitespace) =>
        val res = ir.TypeArgumentClause(parameters.map(visitTypeArgument), srcLocation, precedingWhitespace)
        res.openAngle.precedingWhitespace = node.openAngle.precedingWhitespace
        res.closeAngle.precedingWhitespace = node.closeAngle.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitTypeArgument(node: krakatoa.TypeArgument): ir.TypeArgument = {
    node match {
      case arg: krakatoa.BoundedTypeArgument => visitBoundedTypeArgument(arg)
      case arg: krakatoa.TypeIndication => visitTypeIndication(arg)
      case a =>
        println(s"Error unknown Krakatoa TypeArgument $a")
        null
    }
  }

  def visitBoundedTypeArgument(node: krakatoa.BoundedTypeArgument): ir.BoundedTypeArgument = {
    node match {
      case krakatoa.BoundedTypeArgument(annotations, typeSuper, typeExtends, srcLocation, precedingWhitespace) =>
        val irSuper = if (typeSuper.isEmpty) None else Some(visitTypeIndication(typeSuper.get))
        val irExtends = if (typeExtends.isEmpty) None else Some(visitTypeIndication(typeExtends.get))
        val res = ir.BoundedTypeArgument(annotations.map(visitAnnotation), irSuper, irExtends, srcLocation, precedingWhitespace)
        res.superKeyword.precedingWhitespace = node.superKeyword.precedingWhitespace
        res.extendsKeyword.precedingWhitespace = node.extendsKeyword.precedingWhitespace
        res.questionMark.precedingWhitespace = node.questionMark.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitOldExpression(node: krakatoa.OldExpression): ir.OldExpression = {
    node match {
      case krakatoa.OldExpression(expression, label, srcLocation, precedingWhitespace) =>
        val irLabel = if (label.isEmpty || (!label.isEmpty && label.get.text.equals("Pre"))) None else Some(visitStringNode(label.get))
        val res = ir.OldExpression(visitExpression(expression), irLabel, srcLocation, precedingWhitespace)
        res.oldKeyword.precedingWhitespace = node.oldKeyword.precedingWhitespace
        res.comma.precedingWhitespace = node.comma.precedingWhitespace
        res.openParenthesis.precedingWhitespace = node.openParenthesis.precedingWhitespace
        res.closeParenthesis.precedingWhitespace = node.closeParenthesis.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitAtExpression(node: krakatoa.AtExpression): ir.OldExpression = {
    val label = if (node.label.isEmpty || (!node.label.isEmpty && node.label.get.text.equals("Pre"))) then None else Option(visitStringNode(node.label.get))
    OldExpression(visitExpression(node.expression), label, node.srcLocation, node.precedingWhitespace)
  }

  def visitFreshExpression(node: krakatoa.FreshExpression): ir.FreshExpression = {
    node match {
      case krakatoa.FreshExpression(expressions, srcLocation, precedingWhitespace) =>
        val res = ir.FreshExpression(expressions.map(visitExpression), srcLocation, precedingWhitespace)
        res.openParenthesis.precedingWhitespace = node.openParenthesis.precedingWhitespace
        res.closeParenthesis.precedingWhitespace = node.closeParenthesis.precedingWhitespace
        res.freshKeyword.precedingWhitespace = node.freshKeyword.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitLabeledExpression(node: krakatoa.LabeledExpression): ir.Expression = {
    // LabeledExpression looks something like: name{labels+}(args?)
    if (!node.labels.labels.isEmpty && node.labels.labels.size == 1) {
      val label = node.labels.labels.head.text
      if (label.equals("Here")) {
        // Return a method call, name(args)
        val args = ir.Arguments(node.expressions.map(visitExpression), node.srcLocation)
        return ir.MethodCallExpression(visitStringNode(node.name), args, node.srcLocation, node.precedingWhitespace)
      } else if (label.equals("Old")) {
        // Return a method call evaluated in the pre-state, \old(name(args))
        val args = ir.Arguments(node.expressions.map(visitExpression), node.srcLocation)
        val expr = ir.MethodCallExpression(visitStringNode(node.name), args, node.srcLocation)
        val oldExpression = ir.OldExpression(expr, None, node.srcLocation, node.name.precedingWhitespace)
        return oldExpression
      } else {
        throw NotSupportedException(node.stringRepresentation, node.srcLocation)
      }
    }
    throw new NotSupportedException(node.stringRepresentation, node.srcLocation)
  }

  def visitQuantifiedVarDeclaration(node: krakatoa.QuantifiedVarDeclaration): ir.QuantifiedVarDeclaration = {
    node match {
      case krakatoa.QuantifiedVarDeclaration(declType, declarators, srcLocation, precedingWhitespace) =>
        ir.QuantifiedVarDeclaration(visitTypeIndication(declType), declarators.map(visitStringNode), srcLocation, precedingWhitespace)
    }
  }

  def visitForallExpression(node: krakatoa.ForallExpression): ir.ForallExpression = {
    node match {
      case krakatoa.ForallExpression(decls, expression, srcLocation, precedingWhitespace) =>
        //TODO Need to extract guard & expression from expression (a => b)
        //TODO: May need to split decls into nested foralls?
        val irDecls = decls.map(visitQuantifiedVarDeclaration)
        val irGuard = None
        val irExpression = visitExpression(expression)
        val res = ir.ForallExpression(irDecls, irGuard, irExpression, srcLocation, precedingWhitespace)
        res.quantifier.precedingWhitespace = node.forallKeyword.precedingWhitespace
        res.firstSemicolon.precedingWhitespace = node.semi.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitExistsExpression(node: krakatoa.ExistsExpression): ir.ExistsExpression = {
    node match {
      case krakatoa.ExistsExpression(decls, expression, srcLocation, precedingWhitespace) =>
        val irDecls = decls.map(visitQuantifiedVarDeclaration)
        val irExpression = visitExpression(expression)
        val res = ir.ExistsExpression(irDecls, None, irExpression, srcLocation, precedingWhitespace)
        res.quantifier.precedingWhitespace = node.existsKeyword.precedingWhitespace
        res.firstSemicolon.precedingWhitespace = node.semi.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitLocalVariableDeclaration(node: krakatoa.LocalVariableDeclaration): ir.LocalDeclaration = {
    node match {
      case krakatoa.LocalVariableDeclaration(modifiers, declType, declarators, isStatement, srcLocation, precedingWhitespace) =>
        val res = ir.LocalDeclaration(modifiers.map(visitModifier), visitTypeIndication(declType), declarators.map(visitVariableDeclarator), isStatement, srcLocation, precedingWhitespace)
        res.semicolon.precedingWhitespace = node.semicolon.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitIfStatement(node: krakatoa.IfStatement): ir.IfStatement = {
    node match {
      case krakatoa.IfStatement(condition, thenStatement, elseStatement, srcLocation, precedingWhitespace) =>
        val irElse = if (elseStatement.isEmpty) None else Some(visitStatement(elseStatement.get))
        val res = ir.IfStatement(visitExpression(condition), visitStatement(thenStatement), irElse, srcLocation, precedingWhitespace)
        res.ifKeyword.precedingWhitespace = node.ifKeyword.precedingWhitespace
        res.elseKeyword.precedingWhitespace = node.elseKeyword.precedingWhitespace
        res.openParenthesis.precedingWhitespace = node.openParenthesis.precedingWhitespace
        res.closeParenthesis.precedingWhitespace = node.closeParenthesis.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitForLoop(node: krakatoa.ForLoop): ir.ForLoop = {
    node match {
      case krakatoa.ForLoop(contractBefore, _, control, body, srcLocation, precedingWhitespace) =>
        val irContractBefore = if (contractBefore.isEmpty) Nil else contractBefore.map(visitLoopSpecOrLabel)
        val irContractWithin = List.empty
        val res = ir.ForLoop(irContractBefore, visitForLoopControl(control), irContractWithin, visitStatement(body), srcLocation, precedingWhitespace)
        res.forKeyword.precedingWhitespace = node.forKeyword.precedingWhitespace
        res.openParenthesis.precedingWhitespace = node.openParenthesis.precedingWhitespace
        res.closeParenthesis.precedingWhitespace = node.closeParenthesis.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitForLoopControl(node: krakatoa.ForLoopControl): ir.ForLoopControl = {
    node match {
      case init: krakatoa.ForEachLoopControl => visitForEachLoopControl(init)
      case init: krakatoa.RegularForLoopControl => visitRegularForLoopControl(init)
    }
  }

  def visitRegularForLoopControl(node: krakatoa.RegularForLoopControl): ir.RegularForLoopControl = {
    node match {
      case krakatoa.RegularForLoopControl(init, condition, update, srcLocation, precedingWhitespace) =>
        val irInit = if (init.isEmpty) None else Some(visitForInit(init.get))
        val irCondition = if (condition.isEmpty) None else Some(visitExpression(condition.get))
        val res = ir.RegularForLoopControl(irInit, irCondition, update.map(visitExpression), srcLocation, precedingWhitespace)
        res.firstSemicolon.precedingWhitespace = node.firstSemicolon.precedingWhitespace
        res.secondSemicolon.precedingWhitespace = node.secondSemicolon.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitForInit(node: krakatoa.ForInit): ir.ForInit = {
    node match {
      case init: krakatoa.ExpressionListForInit => visitExpressionListForInit(init)
      case init: krakatoa.LocalVariableDeclaration => visitLocalVariableDeclaration(init)
      case i =>
        println(s"Error unknown Krakatoa ForInit $i")
        null
    }
  }

  def visitExpressionListForInit(node: krakatoa.ExpressionListForInit): ir.ExpressionListForInit = {
    node match {
      case krakatoa.ExpressionListForInit(expressions, srcLocation, precedingWhitespace) =>
        val res = ir.ExpressionListForInit(expressions.map(visitExpression), srcLocation, precedingWhitespace)
        copyWhitespace(node, res)
        res
    }
  }

  def visitForEachLoopControl(node: krakatoa.ForEachLoopControl): ir.ForEachLoopControl = {
    node match {
      case krakatoa.ForEachLoopControl(controlType, variable, control, srcLocation, precedingWhitespace) =>
        val res = ir.ForEachLoopControl(visitTypeIndication(controlType), visitStringNode(variable), visitExpression(control), srcLocation, precedingWhitespace)
        res.modifiers = node.modifiers.map(visitModifier)
        res.colon.precedingWhitespace = node.colon.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitWhileLoop(node: krakatoa.WhileLoop): ir.WhileLoop = {
    node match {
      case krakatoa.WhileLoop(contractBefore, contractWithin, condition, body, srcLocation, precedingWhitespace) =>
        val irContractBefore = if (contractBefore.isEmpty) Nil else contractBefore.map(visitLoopSpecOrLabel)
        val irContractWithin = if (contractWithin.isEmpty) Nil else contractWithin.map(visitLoopSpecOrLabel)
        val res = ir.WhileLoop(irContractBefore, visitExpression(condition), irContractWithin, visitStatement(body), srcLocation, precedingWhitespace)
        res.whileKeyword.precedingWhitespace = node.whileKeyword.precedingWhitespace
        res.openParenthesis.precedingWhitespace = node.openParenthesis.precedingWhitespace
        res.closeParenthesis.precedingWhitespace = node.closeParenthesis.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitLoopSpecOrLabel(node: krakatoa.LoopSpecOrLabel): ir.LoopSpecOrLabel = {
    node match {
      case l: krakatoa.Label => visitLabel(l)
      case spec: krakatoa.LoopSpecification => visitLoopSpecification(spec)
    }
  }

  def visitLoopSpecification(node: krakatoa.LoopSpecification): ir.LoopSpecification = {
    node match {
      case krakatoa.LoopSpecification(startSpec, clauses, endSpec, srcLocation, precedingWhitespace) =>
        val irClauses = clauses.map(visitLoopClause)
        val irStart = startSpec.map(visitStringNode)
        val irEnd = endSpec.map(visitStringNode)
        val res = ir.LoopSpecification(irStart, irClauses, irEnd, srcLocation, precedingWhitespace)
        copyWhitespace(node, res)
        res
    }
  }

  def visitLoopClause(node: krakatoa.LoopClause): ir.LoopClause = {
    try {
      node match {
        case spec: krakatoa.LoopInvariantClause => visitLoopInvariantClause(spec)
        case spec: krakatoa.LoopVariantClause => visitLoopVariantClause(spec)
        case spec: krakatoa.AssignsClause => visitAssignsClause(spec)
        case spec: krakatoa.ForClause => visitForClause(spec)
        case s =>
          println(s"Error unknown Krakatoa LoopClause $s")
          null
      }
    } catch {
      case p: NotSupportedException =>
        var text = "\n//"+p.getMessage
        text += node.stringRepresentation
        text = text.replace("\n", "\n//")
        text = text.replace("@", " ")
        text = text.replace("////", "//")
        val res = ir.NonTranslatedLoopClause(text, node.srcLocation, node.precedingWhitespace)
        res
    }
  }

  def visitLoopInvariantClause(node: krakatoa.LoopInvariantClause): ir.LoopInvariant = {
    node match {
      case krakatoa.LoopInvariantClause(expression, srcLocation, precedingWhitespace) =>
        val res = ir.LoopInvariant(visitExpression(expression), srcLocation, precedingWhitespace)
        res.invariantKeyword.precedingWhitespace = node.invariantKeyword.precedingWhitespace
        res.semicolon.precedingWhitespace = node.semicolon.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitLoopVariantClause(node: krakatoa.LoopVariantClause): ir.DecreasesClause = {
    node match {
      case krakatoa.LoopVariantClause(expression, srcLocation, precedingWhitespace) =>
        val res = ir.DecreasesClause(visitExpression(expression), srcLocation, precedingWhitespace)
        res.decreasesKeyword.precedingWhitespace = node.decreasesKeyword.precedingWhitespace
        res.semicolon.precedingWhitespace = node.semicolon.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitAssignsClause(node: krakatoa.AssignsClause): ir.AssignableClause = {
    node match {
      case krakatoa.AssignsClause(assignableKeyword, expressions, srcLocation, precedingWhitespace) =>
        val res = ir.AssignableClause(expressions.map(visitExpression), srcLocation, precedingWhitespace)
        res.assignableKeyword.precedingWhitespace = node.assignableKeyword.precedingWhitespace
        res.semicolon.precedingWhitespace = node.semicolon.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitForClause(node: krakatoa.ForClause): ir.LoopClause = {
    //if invariant then add condition for "langId" behavior as a in a => langExpr
    node.contractClause match {
      case krakatoa.LoopInvariantClause(expression, srcLocation, precedingWhitespace) =>
        val leftExpr = getBehaviorPrecondition(node.label.text)
        val condExpr = BinaryExpression(visitExpression(leftExpr), ir.StringNode("==>", srcLocation, " "), visitExpression(expression), srcLocation, precedingWhitespace)
        val res = ir.LoopInvariant(condExpr, node.srcLocation, node.precedingWhitespace)
        res.invariantKeyword.precedingWhitespace = node.forKeyword.precedingWhitespace
        res
      case _ =>
        println(s"Warning: cannot translate ${node.stringRepresentation} within a for loop clause")
        null
    }
  }

  def visitDoWhileLoop(node: krakatoa.DoWhileLoop): ir.DoLoop = {
    node match {
      case krakatoa.DoWhileLoop(contractBefore, body, condition, srcLocation, precedingWhitespace) =>
        val irContractBefore = if (contractBefore.isEmpty) Nil else contractBefore.map(visitLoopSpecOrLabel)
        val res = ir.DoLoop(irContractBefore, visitStatement(body), visitExpression(condition), srcLocation, precedingWhitespace)
        res.doKeyword.precedingWhitespace = node.doKeyword.precedingWhitespace
        res.whileKeyword.precedingWhitespace = node.whileKeyword.precedingWhitespace
        res.openParenthesis.precedingWhitespace = node.openParenthesis.precedingWhitespace
        res.closeParenthesis.precedingWhitespace = node.closeParenthesis.precedingWhitespace
        res.semicolon.precedingWhitespace = node.semicolon.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitTry(node: krakatoa.Try): ir.TryBlock = {
    node match {
      case krakatoa.Try(tryBlock, resourceSpecification, catchClauses, finallyBlock, srcLocation, precedingWhitespace) =>
        val irResSpec = if (resourceSpecification.isEmpty) None
        else Some(visitTryResourceSpecification(resourceSpecification.get))
        val irFinally = if (finallyBlock.isEmpty) None else Some(visitBlock(finallyBlock.get))
        val res = ir.TryBlock(irResSpec, visitBlock(tryBlock), catchClauses.map(visitCatchClause), irFinally, srcLocation, precedingWhitespace)
        res.tryKeyword.precedingWhitespace = node.tryKeyword.precedingWhitespace
        res.finallyKeyword.precedingWhitespace = node.finallyKeyword.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitTryResourceSpecification(node: krakatoa.TryResourceSpecification): ir.TryResourceSpecification = {
    node match {
      case krakatoa.TryResourceSpecification(resources, trailingSemicolon, srcLocation, precedingWhitespace) =>
        val irSemi = if (trailingSemicolon.isEmpty) None else Some(visitStringNode(trailingSemicolon.get))
        val res = ir.TryResourceSpecification(resources.map(visitTryResource), irSemi, srcLocation, precedingWhitespace)
        res.openParenthesis.precedingWhitespace = node.openParenthesis.precedingWhitespace
        res.closeParenthesis.precedingWhitespace = node.closeParenthesis.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitTryResource(node: krakatoa.TryResource): ir.TryResource = {
    node match {
      case krakatoa.TryResource(modifiers, resourceType, variable, expression, srcLocation, precedingWhitespace) =>
        val res = ir.TryResource(modifiers.map(visitModifier), visitTypeIndication(resourceType), visitStringNode(variable), visitExpression(expression), srcLocation, precedingWhitespace)
        res.equalsSign.precedingWhitespace = node.equalsSign.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitCatchClause(node: krakatoa.Catch): ir.CatchClause = {
    node match {
      case krakatoa.Catch(modifiers, types, name, block, srcLocation, precedingWhitespace) =>
        val res = ir.CatchClause(types.map(visitTypeIndication), visitStringNode(name), visitBlock(block), srcLocation, precedingWhitespace)
        res.modifiers = modifiers.map(visitModifier)
        res.catchKeyword.precedingWhitespace = node.catchKeyword.precedingWhitespace
        res.openParenthesis.precedingWhitespace = node.openParenthesis.precedingWhitespace
        res.closeParenthesis.precedingWhitespace = node.closeParenthesis.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitSwitch(node: krakatoa.Switch): ir.SwitchStatement = {
    node match {
      case krakatoa.Switch(condition, cases, srcLocation, precedingWhitespace) =>
        val res = ir.SwitchStatement(visitExpression(condition), cases.map(visitSwitchCaseBlock), srcLocation, precedingWhitespace)
        res.openParenthesis.precedingWhitespace = node.openParenthesis.precedingWhitespace
        res.closeParenthesis.precedingWhitespace = node.closeParenthesis.precedingWhitespace
        res.openBrace.precedingWhitespace = node.openBrace.precedingWhitespace
        res.closeBrace.precedingWhitespace = node.closeBrace.precedingWhitespace
        res.switchKeyword.precedingWhitespace = node.switchKeyword.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitSwitchCaseBlock(node: krakatoa.SwitchCaseBlock): ir.SwitchCaseBlock = {
    node match {
      case krakatoa.SwitchCaseBlock(labels, statements, srcLocation, precedingWhitespace) =>
        val res = ir.SwitchCaseBlock(labels.map(visitSwitchLabel), statements.map(visitStatement), srcLocation, precedingWhitespace)
        copyWhitespace(node, res)
        res
    }
  }

  def visitSwitchLabel(node: krakatoa.SwitchLabel): ir.SwitchLabel = {
    node match {
      case krakatoa.SwitchLabel(expression, srcLocation, precedingWhitespace) =>
        val irExpr = if (expression.isEmpty) None else Some(visitExpression(expression.get))
        val res = ir.SwitchLabel(irExpr, srcLocation, precedingWhitespace)
        res.caseKeyword.precedingWhitespace = node.caseKeyword.precedingWhitespace
        res.defaultKeyword.precedingWhitespace = node.defaultKeyword.precedingWhitespace
        res.colon.precedingWhitespace = node.colon.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitSynchronized(node: krakatoa.Synchronized): ir.SynchronizedStatement = {
    node match {
      case krakatoa.Synchronized(condition, body, srcLocation, precedingWhitespace) =>
        val res = ir.SynchronizedStatement(visitExpression(condition), visitStatement(body), srcLocation, precedingWhitespace)
        res.synchronizedKeyword.precedingWhitespace = node.synchronizedKeyword.precedingWhitespace
        res.openParenthesis.precedingWhitespace = node.openParenthesis.precedingWhitespace
        res.closeParenthesis.precedingWhitespace = node.closeParenthesis.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitReturn(node: krakatoa.Return): ir.ReturnStatement = {
    node match {
      case krakatoa.Return(expression, srcLocation, precedingWhitespace) =>
        val irExpr = if (expression.isEmpty) None else Some(visitExpression(expression.get))
        val res = ir.ReturnStatement(irExpr, srcLocation, precedingWhitespace)
        res.returnKeyword.precedingWhitespace = node.returnKeyword.precedingWhitespace
        res.semicolon.precedingWhitespace = node.semicolon.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitThrowStatement(node: krakatoa.ThrowStatement): ir.ThrowStatement = {
    node match {
      case krakatoa.ThrowStatement(expression, srcLocation, precedingWhitespace) =>
        val res = ir.ThrowStatement(visitExpression(expression), srcLocation, precedingWhitespace)
        res.throwKeyword.precedingWhitespace = node.throwKeyword.precedingWhitespace
        res.semicolon.precedingWhitespace = node.semicolon.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitBreak(node: krakatoa.Break): ir.BreakStatement = {
    node match {
      case krakatoa.Break(label, srcLocation, precedingWhitespace) =>
        val irLabel = if (label.isEmpty) None else Some(visitStringNode(label.get))
        val res = ir.BreakStatement(irLabel, srcLocation, precedingWhitespace)
        res.breakKeyword.precedingWhitespace = node.breakKeyword.precedingWhitespace
        res.semicolon.precedingWhitespace = node.semicolon.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitContinue(node: krakatoa.Continue): ir.ContinueStatement = {
    node match {
      case krakatoa.Continue(label, srcLocation, precedingWhitespace) =>
        val irLabel = if (label.isEmpty) None else Some(visitStringNode(label.get))
        val res = ir.ContinueStatement(irLabel, srcLocation, precedingWhitespace)
        res.continueKeyword.precedingWhitespace = node.continueKeyword.precedingWhitespace
        res.semicolon.precedingWhitespace = node.semicolon.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitExpressionStatement(node: krakatoa.ExpressionStatement): ir.ExpressionStatement = {
    node match {
      case krakatoa.ExpressionStatement(expression, srcLocation, precedingWhitespace) =>
        val res = ir.ExpressionStatement(visitExpression(expression), srcLocation, precedingWhitespace)
        res.semicolon.precedingWhitespace = node.semicolon.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitLabeledStatement(node: krakatoa.LabeledStatement): ir.LabeledStatement = {
    node match {
      case krakatoa.LabeledStatement(label, statement, _, srcLocation, precedingWhitespace) =>
        val res = ir.LabeledStatement(visitLabel(label), visitStatement(statement), srcLocation, precedingWhitespace)
        copyWhitespace(node, res)
        res
    }
  }

  def visitLabel(node: krakatoa.Label): ir.Label = {
    node match {
      case krakatoa.Label(label, srcLocation, precedingWhitespace) =>
        val res = ir.Label(visitStringNode(label), srcLocation, precedingWhitespace)
        res.colon.precedingWhitespace = node.colon.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitAssertStatement(node: krakatoa.AssertStatement): ir.AssertStatement = {
    node match {
      case krakatoa.AssertStatement(content, message, srcLocation, precedingWhitespace) =>
        val irMsg = if (message.isEmpty) None else Some(visitExpression(message.get))
        val res = ir.AssertStatement(visitExpression(content), irMsg, srcLocation, precedingWhitespace)
        res.assertKeyword.precedingWhitespace = node.assertKeyword.precedingWhitespace
        res.colon.precedingWhitespace = node.colon.precedingWhitespace
        res.semicolon.precedingWhitespace = node.semicolon.precedingWhitespace
        copyWhitespace(node, res)
        res
    }
  }

  def visitGhostStatement(node: krakatoa.GhostStatement): ir.Statement = {
    node match {
      case krakatoa.GhostStatement(statement, srcLocation, precedingWhitespace) =>
        ir.GhostStatement(visitStatement(statement), srcLocation, precedingWhitespace)
    }
  }

  def visitAssert(node: krakatoa.Assert): ir.JavaAssertion = {
    ir.JavaAssertion(visitExpression(node.content), node.message.map(visitExpression), node.srcLocation, node.precedingWhitespace)
  }

  def getBehaviorPrecondition(label: String): krakatoa.Expression = {
    val option = this.behaviorPreconditions.get(label)
    val srcLocation = SourceLocation("", -1)
    val default = krakatoa.Literal(krakatoa.StringNode("true", srcLocation, " "), srcLocation)
    if (option.nonEmpty) {
      val preconditions = option.get
      if (preconditions.isEmpty) return default
      var res = preconditions.head.expression
      val i = 1
      val srcLocation = res.srcLocation
      val precedingWhitespace = res.precedingWhitespace
      while (i < preconditions.size) {
        val oldRes = res
        res = krakatoa.BinaryExpression(preconditions(i).expression, krakatoa.StringNode("&&", srcLocation), oldRes, srcLocation, precedingWhitespace)
      }
      return res
    }
    return  default
  }

  def visitForStatement(node: krakatoa.ForStatement): ir.Statement = {
    var res: ir.Statement = null
    val condition = visitExpression(getBehaviorPrecondition(node.label.text))
    if (node.statement.isInstanceOf[krakatoa.AssertStatement]) {
      val assertStatement = node.statement.asInstanceOf[krakatoa.AssertStatement]
      // If statement is assert, then change into "assert condition => expression"
      val assertExpr = visitExpression(assertStatement.content)
      val srcLocation = assertStatement.srcLocation
      val precedingWhitespace = assertStatement.precedingWhitespace
      val content = ir.BinaryExpression(condition, ir.StringNode("==>", srcLocation), assertExpr, srcLocation, precedingWhitespace)
      res = ir.AssertStatement(content, None, srcLocation, precedingWhitespace)
    } else if (node.statement.isInstanceOf[krakatoa.GhostStatement]) {
      // If statement is ghost code, change into "ghost if(condition) statement"
      val ghostStatement = node.statement.asInstanceOf[krakatoa.GhostStatement]
      val statement = visitStatement(ghostStatement.statement)
      res = ir.IfStatement(condition, statement, None, ghostStatement.srcLocation, ghostStatement.precedingWhitespace)
      //TODO: How to add "ghost" keyword ??
    } else {
      println(s"Error unknown Krakatoa ForStatement $node")
    }
    res
  }
}
