package languages.Krakatoa.syntaxTree
import generic.{SourceLocation, syntaxTree}

class CompilationUnit(override val pkg: Option[PackageDeclaration],
                           override val imports: List[ImportDeclaration],
                           override val declarations: List[TypeDeclaration],
                           var srcLocation: SourceLocation,
                           var precedingWhitespace: String = ""
                          ) extends generic.syntaxTree.CompilationUnit with KrakatoaTreeNode
object CompilationUnit {
  def unapply(compilationUnit: CompilationUnit): Some[(Option[PackageDeclaration], List[ImportDeclaration], List[TypeDeclaration], SourceLocation, String)] = {
    Some(compilationUnit.pkg, compilationUnit.imports, compilationUnit.declarations, compilationUnit.srcLocation, compilationUnit.precedingWhitespace)
  }
}

case class NonTranslatedCompilationUnit(var text: String,
                                       _srcLocation: SourceLocation,
                                       _precedingWhitespace: String="") extends CompilationUnit(None, Nil, Nil, _srcLocation, _precedingWhitespace) with generic.syntaxTree.NonTranslatedCompilationUnit
case class PackageDeclaration(override val name: StringNode,
                              override val annotations: List[StringAnnotation],
                              var srcLocation: SourceLocation,
                              var precedingWhitespace: String = ""
                             ) extends generic.syntaxTree.PackageDeclaration with KrakatoaTreeNode

case class ImportDeclaration(override val name: StringNode,
                             var srcLocation: SourceLocation,
                             var precedingWhitespace: String = ""
                            ) extends generic.syntaxTree.ImportDeclaration with KrakatoaTreeNode

trait TypeDeclaration extends generic.syntaxTree.TypeDeclaration with Statement
case class NonTranslatedTypeDeclaration(var text: String,
                                        var srcLocation: SourceLocation,
                                        var precedingWhitespace: String="") extends TypeDeclaration with generic.syntaxTree.NonTranslatedTypeDeclaration

trait Modifier extends generic.syntaxTree.Modifier with KrakatoaTreeNode
case class StaticModifier(var srcLocation: SourceLocation,
                          var precedingWhitespace: String="")
      extends generic.syntaxTree.StaticModifier with Modifier

case class ClassDeclaration(override val name: StringNode,
                            override val typeParameters: Option[TypeParameterClause],
                            override val superclass: Option[SuperclassClause],
                            override val interfaces: Option[InterfaceClause],
                            override val body: Body,
                            var srcLocation: SourceLocation,
                            var precedingWhitespace: String = ""
                           ) extends generic.syntaxTree.ClassDeclaration with TypeDeclaration with MemberDeclaration()

case class TypeParameterClause(override val parameters: List[TypeParameter],
                               var srcLocation: SourceLocation,
                               var precedingWhitespace: String = ""
                              ) extends generic.syntaxTree.TypeParameterClause with KrakatoaTreeNode

case class TypeParameter(override val annotations: List[StringAnnotation],
                         override val name: TypeIndication,
                         override val superclass: Option[GenericsSuperclassClause],
                         var srcLocation: SourceLocation,
                         var precedingWhitespace: String = ""
                        ) extends generic.syntaxTree.TypeParameter with KrakatoaTreeNode

case class EnumDeclaration(override val name: StringNode,
                           override val interfaces: Option[InterfaceClause],
                           override val constants: List[EnumConstant],
                           override val declarations: List[MemberDeclaration],
                           override val comma: Option[StringNode],
                           override val semicolon: Option[StringNode] = None,
                           var srcLocation: SourceLocation,
                           var precedingWhitespace: String = ""
                          ) extends generic.syntaxTree.EnumDeclaration with TypeDeclaration with MemberDeclaration()

case class EnumConstant(override val name: StringNode,
                        override val annotations: List[StringAnnotation],
                        override val arguments: Option[Arguments],
                        override val body: Option[Body],
                        var srcLocation: SourceLocation,
                        var precedingWhitespace: String = ""
                       ) extends generic.syntaxTree.EnumConstant with KrakatoaTreeNode

case class InterfaceDeclaration(override val name: StringNode,
                                override val typeParameter: Option[TypeParameterClause],
                                override val superInterfaces: Option[MultiSuperclassClause],
                                override val body: Body,
                                var srcLocation: SourceLocation,
                                var precedingWhitespace: String = ""
                               ) extends generic.syntaxTree.InterfaceDeclaration with TypeDeclaration with MemberDeclaration()

case class MultiSuperclassClause(override val superclasses: List[TypeIndication],
                                 var srcLocation: SourceLocation,
                                 var precedingWhitespace: String = ""
                                ) extends generic.syntaxTree.MultiSuperclassClause with KrakatoaTreeNode

trait TypeIndication extends generic.syntaxTree.TypeIndication with KrakatoaTreeNode with TypeArgument {
  override val modifiers: List[Modifier] = Nil
}

class StringTypeIndication(fullTypeString: String,
                           private var _annotations: List[Annotation] = List.empty,
                           override val typeAsString: String = "",
                           pSrcLocation: SourceLocation,
                           pPrecedingWhitespace: String = "",
                           pTrailingWhitespace: String = "") extends StringNode(fullTypeString, pSrcLocation, pPrecedingWhitespace, pTrailingWhitespace) with TypeIndication {
  def annotations: List[Annotation] = _annotations
  def annotations_=(x1: List[generic.syntaxTree.Annotation]): Unit = {
    if(x1.forall(x => x.isKrakatoaNode)) {
      _annotations = x1.asInstanceOf[List[Annotation]]
    } else {
      throw new ClassCastException("not a Krakatoa node")
    }
  }
}
object StringTypeIndication {
  def unapply(stringTypeIndication: StringTypeIndication): Some[(String, List[Annotation], String, SourceLocation, String, String)] = {
    Some(stringTypeIndication.text,
      stringTypeIndication.annotations,
      stringTypeIndication.typeAsString,
      stringTypeIndication.srcLocation,
      stringTypeIndication.precedingWhitespace,
      stringTypeIndication.trailingWhitespace)
  }
}

class StringGenericTypeIndication(fullTypeString: String,
                                  override val typeParameters: TypeParameterClause,
                                  annotations: List[Annotation],
                                  override val typeAsString: String,
                                  srcLocation: SourceLocation,
                                  precedingWhitespace: String="")
  extends StringTypeIndication(fullTypeString, annotations, typeAsString, srcLocation, precedingWhitespace) with generic.syntaxTree.GenericTypeIndication
object StringGenericTypeIndication {
  def unapply(stringTypeIndication: StringGenericTypeIndication): Some[(String, TypeParameterClause, List[Annotation], String, SourceLocation, String)] = {
    Some(stringTypeIndication.text, stringTypeIndication.typeParameters, stringTypeIndication.annotations, stringTypeIndication.typeAsString, stringTypeIndication.srcLocation, stringTypeIndication.precedingWhitespace)
  }
}

case class GenericsSuperclassClause(override val superclasses: List[TypeIndication],
                                    var srcLocation: SourceLocation,
                                    var precedingWhitespace: String = ""
                                   ) extends generic.syntaxTree.GenericsSuperclassClause with KrakatoaTreeNode

case class SuperclassClause(override val superclass: TypeIndication,
                            var srcLocation: SourceLocation,
                            var precedingWhitespace: String = ""
                           ) extends generic.syntaxTree.SuperclassClause with KrakatoaTreeNode

case class InterfaceClause(override val interfaces: List[TypeIndication],
                           var srcLocation: SourceLocation,
                           var precedingWhitespace: String = ""
                          ) extends generic.syntaxTree.InterfaceClause with KrakatoaTreeNode

case class Body(private var _declarations: List[MemberDeclaration],
                var srcLocation: SourceLocation,
                var precedingWhitespace: String = ""
               ) extends generic.syntaxTree.Body with KrakatoaTreeNode{
  def declarations: List[MemberDeclaration] = _declarations
  def declarations_=(x1: List[generic.syntaxTree.MemberDeclaration]): Unit = {
    if(x1.forall(x => x.isKrakatoaNode)) {
      _declarations = x1.asInstanceOf[List[MemberDeclaration]]
    } else {
      throw new ClassCastException("not a Krakatoa node")
    }
  }
}

trait MemberDeclaration(private var _contractBefore: Option[Contract] = None,
                        private var _modifiers: List[Modifier] = Nil
                       ) extends generic.syntaxTree.MemberDeclaration with KrakatoaTreeNode {
  def contractBefore: Option[Contract] = _contractBefore

  def contractBefore_=(x1: Option[generic.syntaxTree.Contract]): Unit = {
    if(x1.isEmpty || (x1.isDefined && x1.get.isKrakatoaNode)){
      _contractBefore = x1.asInstanceOf[Option[Contract]]
    } else {
      throw new ClassCastException("not a Krakatoa node")
    }
  }

  def modifiers: List[Modifier] = _modifiers

  def modifiers_=(x1: List[generic.syntaxTree.Modifier]): Unit = {
    if(x1.forall(x => x.isKrakatoaNode)) {
      _modifiers = x1.asInstanceOf[List[Modifier]]
    } else {
      throw new ClassCastException("not a Krakatoa node")
    }
  }
}

case class NonTranslatedMemberDeclaration(var text: String,
                                          var srcLocation: SourceLocation,
                                          var precedingWhitespace: String="")
  extends MemberDeclaration() with generic.syntaxTree.NonTranslatedMemberDeclaration


class MethodDeclaration(private var _name: StringNode,
                        private var _returnType: Option[TypeIndication],
                        private var _parameters: List[Parameter],
                        private var _dimensions: Option[StringNode],
                        private var _throws: Option[Throws],
                        private var _default: Option[DefaultClause],
                        private var _contractWithin: Option[Contract],
                        private var _body: BlockOrSemicolon, //TODO: Check whether I need to add contractWithin
                        var srcLocation: SourceLocation,
                        var precedingWhitespace: String = ""
                       ) extends generic.syntaxTree.MethodDeclaration with MemberDeclaration() {
  def name: StringNode = _name
  def name_=(x1: generic.syntaxTree.StringNode): Unit = {
    if(x1.isKrakatoaNode) {
      _name = x1.asInstanceOf[StringNode]
    } else {
      throw new ClassCastException("not a Krakatoa node")
    }
  }
  def returnType: Option[TypeIndication] = _returnType
  def returnType_=(x1: Option[generic.syntaxTree.TypeIndication]): Unit = {
    if(x1.isEmpty || (x1.isDefined && x1.get.isKrakatoaNode)) {
      _returnType = x1.asInstanceOf[Option[TypeIndication]]
    } else {
      throw new ClassCastException("not a Krakatoa node")
    }
  }
  def parameters: List[Parameter] = _parameters
  def parameters_=(x1: List[generic.syntaxTree.Parameter]): Unit = {
    if(x1.forall(x => x.isKrakatoaNode)) {
      _parameters = x1.asInstanceOf[List[Parameter]]
    } else {
      throw new ClassCastException("not a Krakatoa node")
    }
  }
  def dimensions: Option[StringNode] = _dimensions
  def dimensions_=(x1: Option[generic.syntaxTree.StringNode]): Unit = {
    if(x1.isEmpty || (x1.isDefined && x1.get.isKrakatoaNode)) {
      _dimensions = x1.asInstanceOf[Option[StringNode]]
    } else {
      throw new ClassCastException("not a Krakatoa node")
    }
  }
  def throws: Option[Throws] = _throws
  def throws_=(x1: Option[generic.syntaxTree.ThrowsClause]): Unit = {
    if(x1.isEmpty || (x1.isDefined && x1.get.isKrakatoaNode)) {
      _throws = x1.asInstanceOf[Option[Throws]]
    } else {
      throw new ClassCastException("not a Krakatoa node")
    }
  }
  def default: Option[DefaultClause] = _default
  def default_=(x1: Option[generic.syntaxTree.DefaultClause]): Unit = {
    if(x1.isEmpty || (x1.isDefined && x1.get.isKrakatoaNode)) {
      _default = x1.asInstanceOf[Option[DefaultClause]]
    } else {
      throw new ClassCastException("not a Krakatoa node")
    }
  }
  def contractWithin: Option[Contract] = _contractWithin
  def contractWithin_=(x1: Option[generic.syntaxTree.Contract]): Unit = {
    if(x1.isEmpty || (x1.isDefined && x1.get.isKrakatoaNode)) {
      _contractWithin = x1.asInstanceOf[Option[Contract]]
    } else {
      throw new ClassCastException("not a Krakatoa node")
    }
  }
  def body: BlockOrSemicolon = _body
  def body_=(x1: generic.syntaxTree.BlockOrSemicolon): Unit = {
    if(x1.isKrakatoaNode) {
      _body = x1.asInstanceOf[BlockOrSemicolon]
    } else {
      throw new ClassCastException("not a Krakatoa node")
    }
  }
}
object MethodDeclaration {
  def unapply(methodDeclaration: MethodDeclaration): Some[(StringNode, Option[TypeIndication], List[Parameter], Option[StringNode], Option[Throws], Option[DefaultClause], Option[Contract], BlockOrSemicolon, SourceLocation, String)] = {
    Some(methodDeclaration.name, methodDeclaration.returnType, methodDeclaration.parameters, methodDeclaration.dimensions, methodDeclaration.throws, methodDeclaration.default, methodDeclaration.contractWithin, methodDeclaration.body, methodDeclaration.srcLocation, methodDeclaration.precedingWhitespace)
  }
}

case class GenericMethodDeclaration(val typeParameters: TypeParameterClause,
                                    pName: StringNode,
                                    pReturnType: Option[TypeIndication],
                                    pParameters: List[Parameter],
                                    pDimensions: Option[StringNode],
                                    pThrows: Option[Throws],
                                    pDefault: Option[DefaultClause],
                                    pContractWithin: Option[Contract],
                                    pBody: BlockOrSemicolon,
                                    pSrcLocation: SourceLocation,
                                    pPrecedingWhitespace: String = "",
                                   ) extends MethodDeclaration(pName, pReturnType, pParameters, pDimensions, pThrows, pDefault, pContractWithin, pBody, pSrcLocation, pPrecedingWhitespace) with generic.syntaxTree.GenericMethodDeclaration

case class DefaultClause(override val value: StringNode,
                         var srcLocation: SourceLocation,
                         var precedingWhitespace: String = ""
                        ) extends generic.syntaxTree.DefaultClause with KrakatoaTreeNode

case class FieldDeclaration(override val fieldType: TypeIndication,
                            override val declarators: List[VariableDeclarator],
                            var srcLocation: SourceLocation,
                            var precedingWhitespace: String = ""
                           ) extends generic.syntaxTree.FieldDeclaration with MemberDeclaration()

case class VariableDeclarator(override val name: StringNode,
                              override val initializer: Option[Initializer],
                              var srcLocation: SourceLocation,
                              var precedingWhitespace: String = ""
                             ) extends generic.syntaxTree.VariableDeclarator with KrakatoaTreeNode

trait Initializer extends generic.syntaxTree.Initializer with KrakatoaTreeNode

case class ArrayInitializer(override val initializers: List[Initializer],
                            override val trailingComma: Option[StringNode],
                            var srcLocation: SourceLocation,
                            var precedingWhitespace: String = ""
                           ) extends generic.syntaxTree.ArrayInitializer with Initializer

case class Parameter(override val parameterType: TypeIndication,
                     override val name: StringNode,
                     override val modifiers: List[Modifier],
                     var srcLocation: SourceLocation,
                     var precedingWhitespace: String = ""
                    ) extends generic.syntaxTree.Parameter with KrakatoaTreeNode

case class VarargsParameter(override val parameterType: TypeIndication,
                            private var _annotations: List[Annotation],
                            override val name: StringNode,
                            var srcLocation: SourceLocation,
                            var precedingWhitespace: String="") extends generic.syntaxTree.VarargsParameter with KrakatoaTreeNode {
  override def annotations: List[Annotation] = _annotations
  def annotations_=(x1: List[generic.syntaxTree.Annotation]): Unit = {
    if(x1.forall(x => x.isKrakatoaNode)) {
      _annotations = x1.asInstanceOf[List[Annotation]]
    } else {
      throw new ClassCastException("Not a Krakatoa node")
    }
  }
}

case class Block(override val static: Option[Modifier],
                 override val statements: List[Statement],
                 var srcLocation: SourceLocation,
                 var precedingWhitespace: String = ""
                ) extends generic.syntaxTree.Block with Statement with BlockOrSemicolon with KrakatoaTreeNode with LambdaBody with MemberDeclaration()

class NoType(srcLocation: SourceLocation) extends StringTypeIndication("", Nil, "", srcLocation)
object NoType {
  def unapply(noType: NoType): Some[SourceLocation] = Some(noType.srcLocation)
}

trait BlockOrSemicolon extends generic.syntaxTree.BlockOrSemicolon with MemberDeclaration
trait Statement extends generic.syntaxTree.Statement with KrakatoaTreeNode

case class LocalVariableDeclaration(private var _modifiers: List[Modifier],
                                    override val declType: TypeIndication,
                                    override val declarators: List[VariableDeclarator],
                                    var isStatement: Boolean = true,
                                    var srcLocation: SourceLocation,
                                    var precedingWhitespace: String = ""
                                   ) extends generic.syntaxTree.LocalDeclaration with Statement with ForInit {
  override def modifiers: List[Modifier] = _modifiers
  def modifiers_=(x1: List[syntaxTree.Modifier]): Unit = {
    if(x1.forall(x => x.isKrakatoaNode)) {
      _modifiers = x1.asInstanceOf[List[Modifier]]
    } else {
      throw new ClassCastException("not a Krakatoa node")
    }
  }
}

case class Assert(override val content: Expression,
                  override val message: Option[Expression],
                  var srcLocation: SourceLocation,
                  var precedingWhitespace: String = ""
                 ) extends generic.syntaxTree.JavaAssertion with Statement

case class IfStatement(override val condition: Expression,
                       override val thenStatement: Statement,
                       override val elseStatement: Option[Statement],
                       var srcLocation : SourceLocation,
                       var precedingWhitespace: String = ""
                      ) extends generic.syntaxTree.IfStatement with Statement

trait Loop extends generic.syntaxTree.Loop with Statement {
  protected var _contractBefore: List[LoopSpecOrLabel]
  def contractBefore: List[LoopSpecOrLabel] = _contractBefore
  def contractBefore_=(x1: List[syntaxTree.LoopSpecOrLabel]): Unit = {
    if(x1.forall(x => x.isKrakatoaNode)) {
      _contractBefore = x1.asInstanceOf[List[LoopSpecOrLabel]]
    } else {
      throw new ClassCastException("Not a Krakatoa node")
    }
  }

  override val body: Statement
}

trait  LoopWithContractWithin extends Loop with generic.syntaxTree.LoopWithContractWithin {
  protected var _contractWithin: List[LoopSpecOrLabel]
  def contractWithin: List[LoopSpecOrLabel] = _contractWithin
  def contractWithin_=(x1: List[syntaxTree.LoopSpecOrLabel]): Unit = {
    if(x1.forall(x => x.isKrakatoaNode)) {
      _contractWithin = x1.asInstanceOf[List[LoopSpecOrLabel]]
    } else {
      throw new ClassCastException("Not a Krakatoa node")
    }
  }
}

case class ForLoop(private var pContractBefore: List[LoopSpecOrLabel],
                   private var pContractWithin: List[LoopSpecOrLabel] = List.empty,
                   override val control: ForLoopControl,
                   override val body: Statement,
                   var srcLocation: SourceLocation,
                   var precedingWhitespace: String = ""
                  ) extends generic.syntaxTree.ForLoop with LoopWithContractWithin {

  def _contractBefore: List[LoopSpecOrLabel] = pContractBefore
  def _contractBefore_=(x1: List[LoopSpecOrLabel]): Unit = {pContractBefore = x1}
  def _contractWithin: List[LoopSpecOrLabel] = pContractWithin
  def _contractWithin_=(x1: List[LoopSpecOrLabel]): Unit = {pContractWithin = x1}
}

trait ForInit extends generic.syntaxTree.ForInit with KrakatoaTreeNode
trait ForLoopControl extends generic.syntaxTree.ForLoopControl with KrakatoaTreeNode
case class ExpressionListForInit(override val expressions: List[Expression],
                                 var srcLocation: SourceLocation,
                                 var precedingWhitespace: String="") extends generic.syntaxTree.ExpressionListForInit with ForInit
case class RegularForLoopControl (override val init: Option[ForInit],
                                  override val condition: Option[Expression],
                                  override val update: List[Expression],
                                  var srcLocation: SourceLocation,
                                  var precedingWhitespace: String="") extends generic.syntaxTree.RegularForLoopControl with ForLoopControl
case class ForEachLoopControl (override val controlType: TypeIndication,
                               override val variable: StringNode,
                               override val control: Expression,
                               var srcLocation: SourceLocation,
                               var precedingWhitespace: String="") extends generic.syntaxTree.ForEachLoopControl with ForLoopControl {
  override val modifiers: List[Modifier] = Nil
}

case class WhileLoop(private var pContractBefore: List[LoopSpecOrLabel],
                     private var pContractWithin: List[LoopSpecOrLabel] = List.empty,
                     override val condition: Expression,
                     override val body: Statement,
                     var srcLocation: SourceLocation,
                     var precedingWhitespace: String = ""
                    ) extends generic.syntaxTree.WhileLoop with LoopWithContractWithin {
  def _contractBefore: List[LoopSpecOrLabel] = pContractBefore
  def _contractBefore_=(x1: List[LoopSpecOrLabel]): Unit = {pContractBefore = x1}
  def _contractWithin: List[LoopSpecOrLabel] = pContractWithin
  def  _contractWithin_=(x1: List[LoopSpecOrLabel]): Unit = {pContractWithin = x1}
}

case class DoWhileLoop(private var pContractBefore: List[LoopSpecOrLabel],
                       override val body: Statement,
                       override val condition: Expression,
                       var srcLocation: SourceLocation,
                       var precedingWhitespace: String = ""
                      ) extends generic.syntaxTree.DoLoop with Loop {
  def _contractBefore: List[LoopSpecOrLabel] = pContractBefore
  def _contractBefore_=(x1: List[LoopSpecOrLabel]): Unit = {pContractBefore = x1}
}

case class Try(override val tryBlock: Block,
               override val resourceSpecification: Option[TryResourceSpecification],
               override val catchClauses: List[Catch],
               override val finallyBlock: Option[Block],
               var srcLocation: SourceLocation,
               var precedingWhitespace: String = ""
              ) extends generic.syntaxTree.TryBlock with Statement

case class TryResourceSpecification(override val resources: List[TryResource],
                                    override val trailingSemicolon: Option[StringNode],
                                    var srcLocation: SourceLocation,
                                    var precedingWhitespace: String = ""
                                   ) extends generic.syntaxTree.TryResourceSpecification with KrakatoaTreeNode

case class TryResource(private var _modifiers: List[Modifier],
                       override val resourceType: TypeIndication,
                       override val variable: StringNode,
                       override val expression: Expression,
                       var srcLocation: SourceLocation,
                       var precedingWhitespace: String = ""
                      ) extends generic.syntaxTree.TryResource with KrakatoaTreeNode {
  override def modifiers: List[Modifier] = _modifiers
  def modifiers_=(x1: List[generic.syntaxTree.Modifier]): Unit = {
    if(x1.forall(x => x.isKrakatoaNode)) {
      _modifiers = x1.asInstanceOf[List[Modifier]]
    } else {
      throw new ClassCastException("Not a Krakatoa node")
    }
  }
}

case class Catch(private var _modifiers: List[Modifier],
                 override val types: List[TypeIndication],
                 override val name: StringNode,
                 override val block: Block,
                 var srcLocation: SourceLocation,
                 var precedingWhitespace: String = ""
                ) extends generic.syntaxTree.CatchClause with KrakatoaTreeNode {
  override def modifiers: List[Modifier] = _modifiers
}

case class Switch(override val condition: Expression,
                  override val cases: List[SwitchCaseBlock],
                  var srcLocation: SourceLocation,
                  var precedingWhitespace: String = ""
                 ) extends generic.syntaxTree.SwitchStatement with Statement

case class SwitchCaseBlock(override val labels: List[SwitchLabel],
                           override val statements: List[Statement],
                           var srcLocation: SourceLocation,
                           var precedingWhitespace: String = ""
                          ) extends generic.syntaxTree.SwitchCaseBlock with KrakatoaTreeNode

case class SwitchLabel(override val expression: Option[Expression],
                       var srcLocation: SourceLocation,
                       var precedingWhitespace: String = ""
                      ) extends generic.syntaxTree.SwitchLabel with KrakatoaTreeNode

case class Synchronized(override val condition: Expression,
                        override val body: Statement,
                        var srcLocation: SourceLocation,
                        var precedingWhitespace: String = ""
                       ) extends generic.syntaxTree.SynchronizedStatement with Statement

case class Return(override val expression: Option[Expression],
                  var srcLocation: SourceLocation,
                  var precedingWhitespace: String = ""
                 ) extends generic.syntaxTree.ReturnStatement with Statement

case class Throws(override val exceptions: List[StringNode],
                  var srcLocation: SourceLocation,
                  var precedingWhitespace: String = ""
                 ) extends generic.syntaxTree.ThrowsClause with Statement

case class ThrowStatement(override val expression: Expression,
                          var srcLocation: SourceLocation,
                          var precedingWhitespace: String="") extends generic.syntaxTree.ThrowStatement with Statement

case class Break(override val label: Option[StringNode],
                 var srcLocation: SourceLocation,
                 var precedingWhitespace: String = ""
                ) extends generic.syntaxTree.BreakStatement with Statement

case class Continue(override val label: Option[StringNode],
                    var srcLocation: SourceLocation,
                    var precedingWhitespace: String = ""
                   ) extends generic.syntaxTree.ContinueStatement with Statement

class EmptyStatement(_srcLocation: SourceLocation,
                    _precedingWhitespace: String = ""
                    ) extends StringNode(";", _srcLocation, _precedingWhitespace) with Statement with TypeDeclaration with BlockOrSemicolon with MemberDeclaration()
case class ExpressionStatement(override val expression: Expression,
                               var srcLocation: SourceLocation,
                               var precedingWhitespace: String = ""
                              ) extends generic.syntaxTree.ExpressionStatement with Statement

case class LabeledStatement(override val label: Label,
                            override val statement: Statement,
                            override val contract: Option[Contract] = None,
                            var srcLocation: SourceLocation,
                            var precedingWhitespace: String = ""
                           ) extends generic.syntaxTree.LabeledStatement with Statement

//case class LocalTypeDeclaration(val classDeclaration: Option[ClassDeclaration],
//                                val interfaceDeclaration: Option[InterfaceDeclaration],
//                                var srcLocation: SourceLocation,
//                                var precedingWhitespace: String = ""
//                               ) extends Statement with MemberDeclaration {
////  def modifiers: List[Modifier] = modifiers
////  def modifiers_=(x1:List[generic.syntaxTree.Modifier]): Unit = {
////    x1 match {
////      case mod: List[Modifier] => modifiers = mod
////      case _ => throw new ClassCastException("Not a Krakatoa node")
////    }
////  }
//
//  override def stringRepresentation: String = {
//    var res = modifiers.mkString
//    if (classDeclaration.isDefined) {
//      res += classDeclaration.toString
//    } else if (interfaceDeclaration.isDefined) {
//      res += interfaceDeclaration.toString
//    } else {
//      res += ";"
//      //TODO: This defaults to ";", what if it's not ";", class or interface declaration? Should add exception?
//    }
//    res
//  }
//}

trait TypeArgument extends generic.syntaxTree.TypeArgument with KrakatoaTreeNode

case class BoundedTypeArgument(override val annotations: List[Annotation],
                               override val typeSuper: Option[TypeIndication],
                               override val typeExtends: Option[TypeIndication],
                               var srcLocation: SourceLocation,
                               var precedingWhitespace: String="") extends generic.syntaxTree.BoundedTypeArgument with TypeArgument

case class TypeArgumentClause(override val parameters: List[TypeArgument],
                              var srcLocation: SourceLocation,
                              var precedingWhitespace: String = ""
                             ) extends generic.syntaxTree.TypeArgumentClause with KrakatoaTreeNode

trait Annotation extends generic.syntaxTree.Annotation with Modifier

case class StringAnnotation(var text: String,
                            var srcLocation: SourceLocation,
                            var precedingWhitespace: String = ""
                           ) extends generic.syntaxTree.StringAnnotation with Annotation

case class AnnotationTypeDeclaration(override val name: StringNode,
                                     override val body: Body,
                                     var srcLocation: SourceLocation,
                                     var precedingWhitespace: String = ""
                                    ) extends generic.syntaxTree.AnnotationTypeDeclaration with TypeDeclaration with MemberDeclaration()

case class Label(override val label: StringNode,
                 var srcLocation: SourceLocation,
                 var precedingWhitespace: String = ""
                ) extends generic.syntaxTree.Label with KrakatoaTreeNode with LoopSpecOrLabel

case class Arguments(override val arguments: List[Expression],
                     var srcLocation: SourceLocation,
                     var precedingWhitespace: String = ""
                    ) extends generic.syntaxTree.Arguments with KrakatoaTreeNode