package languages.OpenJML

import generic.{Helper, NotSupportedException, SourceLocation}
import intermediateRepresentation.syntaxTree as ir
import languages.OpenJML.syntaxTree as jml

import scala.collection.mutable.ListBuffer


class FromIRVisitor {

  class PermissionExpressionException(val expr: ir.PermissionExpression) extends Exception
  class MultiplePermissionsException() extends Exception
  class CurrentPermissionValueException() extends Exception


  // indicates whether new modifiers, etc need to be wrapped in specifications
  // or if they already are (e.g. inside ghost declarations)
  var wrapInSpec = true

  private var collectedPermissions: ListBuffer[ir.PermissionExpression] = ListBuffer()

  def visitStringNode(node: ir.StringNode): jml.StringNode = {
    node match {
      case imp: ir.ImpliesOperator => visitImpliesOperator(imp)
      case sep: ir.SeparatingConjunctOperator => visitSeparatingConjunctOperator(sep)
      case ir.StringNode(text, src, pred, trail) =>
        jml.StringNode(text, src, pred, trail)
    }
  }
  def visitImpliesOperator(node: ir.ImpliesOperator): jml.ImpliesOperator = {
    node match {
      case ir.ImpliesOperator(srcLocation, pred) =>
        val res = new jml.ImpliesOperator(srcLocation)
        res.copyWhitespace(node)
        res
    }
  }
  def visitSeparatingConjunctOperator(node: ir.SeparatingConjunctOperator): jml.StringNode = {
    node match {
      case ir.SeparatingConjunctOperator(text, srcLocation, pred) =>
        val res = jml.StringNode("&&", srcLocation)
        res.copyWhitespace(node)
        res
    }
  }
  def visitModifierStringNode(node: ir.ModifierStringNode): jml.ModifierStringNode = {
    node match {
      case ir.ModifierStringNode(text, src, pred) =>
        val res = jml.ModifierStringNode(text, src, pred)
        res.copyWhitespace(node)
        res
    }
  }

  def transformStartSpec(node: jml.StringNode): jml.StringNode = {
    val linepattern = """((.*\r?\n)*.*)//\s+@((.*\r?\n)*.*)""".r
    linepattern.findPrefixMatchOf(node.precedingWhitespace+node.text) match {
      case Some(m) =>
        node.precedingWhitespace = m.group(1)
        node.text = "//@" + m.group(3)
      case None =>
    }
    var blockpattern = """((.*\r?\n)*.*)\/*\s+@((.*\r?\n)*.*)""".r
    blockpattern.findPrefixMatchOf(node.precedingWhitespace+node.text) match {
      case Some(m) =>
        node.precedingWhitespace = m.group(1)
        node.text = "/*@" + m.group(3)
      case None =>
    }
    blockpattern = """((.*\r?\n)*.*/\*([^+-]|\r|\n)(.)*?)(\r?\n\s*)@((.*\r?\n)*.*)""".r
    blockpattern.findPrefixMatchOf(node.precedingWhitespace+node.text) match {
      case Some(m) =>
        node.precedingWhitespace = m.group(1) + "*/" + m.group(5)
        node.text = "/*@" + m.group(6)
      case None =>
    }
    node
  }

  def makeSpecPublic(modifiers: List[jml.Modifier], src: SourceLocation): List[jml.Modifier] = {
    def findPublic(modifier: jml.Modifier): Boolean = {
      modifier match {
        case s: jml.ModifierStringNode => s.text == "public" || s.text == "spec_public"
        case j: jml.JmlModifierBlock => j.modifiers.exists(findPublic)
        case _ => false
      }
    }
    if(!modifiers.exists(findPublic)) {
      var specPublic: jml.Modifier = jml.SpecPublicModifier(src, " ")
      if (wrapInSpec) {
        specPublic = jml.JmlModifierBlock(Some(jml.StringNode("/*@", src, " ")), List(specPublic),
                                        Some(jml.StringNode("*/", src, " ")), specPublic.srcLocation)
      }
      modifiers.appended(specPublic)
    } else {
      modifiers
    }
  }


  def visitCompilationUnit(node: ir.CompilationUnit): jml.CompilationUnit = {
    node match {
      case nt: ir.NonTranslatedCompilationUnit =>
        visitNonTranslatedCompilationUnit(nt)
      case ir.CompilationUnit(pkg, imports, declarations, srcLocation, precedingWhitespace) =>
        try
          var irPkg: Option[jml.PackageDeclaration] = None
          if (pkg.isDefined) {
            irPkg = Some(visitPackageDeclaration(pkg.get))
          }
          val res = jml.CompilationUnit(irPkg,
                                        imports.map(visitImportDeclaration),
                                        declarations.map(visitTypeDeclaration),
                                        srcLocation,
                                        precedingWhitespace)
          res.copyWhitespace(node)
          res
        catch {
          case e: NotSupportedException =>
            val text = Helper.commentUnknownNode(node, s"${e.construct} not supported, whole CompilationUnit commented out")
            jml.NonTranslatedCompilationUnit(text, node.srcLocation)
        }
    }
  }

  def visitNonTranslatedCompilationUnit(node: ir.NonTranslatedCompilationUnit): jml.NonTranslatedCompilationUnit = {
    node match {
      case ir.NonTranslatedCompilationUnit(text, srcLocation, precedingWhitespace) =>
        val res = jml.NonTranslatedCompilationUnit(text, srcLocation, precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitPackageDeclaration(node: ir.PackageDeclaration): jml.PackageDeclaration = {
    node match {
      case ir.PackageDeclaration(name, annotations, srcLocation, precedingWhitespace) =>
        val res = jml.PackageDeclaration(visitStringNode(name),
                                        annotations.map(visitStringAnnotation),
                                        srcLocation,
                                        precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitImportDeclaration(node: ir.ImportDeclaration): jml.ImportDeclaration = {
    node match {
      case ir.ImportDeclaration(name, srcLocation, precedingWhitespace) =>
        val res = jml.ImportDeclaration(visitStringNode(name),
                              srcLocation,
                              precedingWhitespace)
        res.modifiers = node.modifiers.map(visitModifier)
        res.copyWhitespace(node)
        res
    }
  }

  def visitTypeDeclaration(node: ir.TypeDeclaration): jml.TypeDeclaration = {
    try
      node match {
        case decl: ir.AnnotationTypeDeclaration => visitAnnotationTypeDeclaration(decl)
        case decl: ir.ClassDeclaration =>
          val cl = visitClassDeclaration(decl)
          var mod: jml.Modifier = jml.NullableByDefaultModifier(node.srcLocation, " ")
          if(wrapInSpec) {
            mod = jml.JmlModifierBlock(Some(jml.StringNode("/*@", mod.srcLocation, " ")), List(mod),
                                       Some(jml.StringNode("*/", mod.srcLocation, " ")), mod.srcLocation)
          }
          if(cl.modifiers.isEmpty) {
            mod.precedingWhitespace = cl.classKeyword.precedingWhitespace + mod.precedingWhitespace
            cl.classKeyword.precedingWhitespace = " "
          }
          cl.modifiers = cl.modifiers.appended(mod)
          cl
        case decl: ir.EnumDeclaration => visitEnumDeclaration(decl)
        case decl: ir.InterfaceDeclaration =>
          val inter = visitInterfaceDeclaration(decl)
          var mod: jml.Modifier = jml.NullableByDefaultModifier(node.srcLocation, " ")
          if(wrapInSpec) {
            mod = jml.JmlModifierBlock(Some(jml.StringNode("/*@", mod.srcLocation, " ")), List(mod),
              Some(jml.StringNode("*/", mod.srcLocation, " ")), mod.srcLocation)
          }
          if(inter.modifiers.isEmpty) {
            mod.precedingWhitespace = inter.interfaceKeyword.precedingWhitespace + mod.precedingWhitespace
            inter.interfaceKeyword.precedingWhitespace = " "
          }
          inter.modifiers = inter.modifiers.appended(mod)
          inter
        case decl: ir.EmptyStatement => visitEmptyStatement(decl)
        case decl: ir.SpecDeclarationBlock => visitSpecDeclarationBlockOnTypeLevel(decl)
        case decl: ir.NonTranslatedTypeDeclaration => visitNonTranslatedTypeDeclaration(decl)
        case _ => ???
      }
    catch {
      case e: NotSupportedException =>
        val text = Helper.commentUnknownNode(node, s"${e.construct} not supported, whole TypeDeclaration commented out")
        jml.NonTranslatedTypeDeclaration(text, node.srcLocation)
    }
  }

  def visitNonTranslatedTypeDeclaration(node: ir.NonTranslatedTypeDeclaration): jml.TypeDeclaration = {
    node match {
      case ir.NonTranslatedTypeDeclaration(text, srcLocation, precedingWhitespace) =>
        val res = jml.NonTranslatedTypeDeclaration(text,
                                                  srcLocation,
                                                  precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def rewriteSpecPublicModifier(modifier: jml.Modifier): jml.Modifier = {
    modifier match {
      case sp: jml.SpecPublicModifier =>
        sp.text = "public"
      case _ =>
    }
    modifier
  }

  def visitGhostDeclaration(node: ir.GhostDeclaration): jml.SpecDeclaration = {
    node match {
      case ir.GhostDeclaration(ghost, decl, srcLocation, precedingWhitespace) =>
        if(decl.isInstanceOf[ir.MethodDeclaration]) {
          val res = jml.ModelDeclaration(jml.ModelModifier(ghost.srcLocation), visitMemberDeclaration(decl), srcLocation)
          res.model.copyWhitespace(ghost)
          res.copyWhitespace(node)
          res.decl.modifiers = res.decl.modifiers.map(rewriteSpecPublicModifier)
          res
        } else {
          val res = jml.GhostDeclaration(visitGhostModifier(ghost),
                                        visitMemberDeclaration(decl),
                                        srcLocation,
                                        precedingWhitespace)
          res.copyWhitespace(node)
          res.decl.modifiers = res.decl.modifiers.map(rewriteSpecPublicModifier)
          res
        }
    }
  }
  def visitModelDeclaration(node: ir.ModelDeclaration): jml.ModelDeclaration = {
    node match {
      case ir.ModelDeclaration(model, decl, srcLocation, precedingWhitespace) =>
        val res = jml.ModelDeclaration(visitModelModifier(model),
                                        visitMemberDeclaration(decl),
                                        srcLocation,
                                        precedingWhitespace)
        res.decl.modifiers = res.decl.modifiers.map(rewriteSpecPublicModifier)
        res.copyWhitespace(node)
        res
    }
  }
  def visitModifier(node: generic.syntaxTree.Modifier): jml.Modifier = {
    node match {
      case mod: ir.Modifier => visitModifier(mod)
      case m =>
        println(s"Error unknown JML modifier $m")
        null
    }
  }
  def visitModifier(node: ir.Modifier): jml.Modifier = {
    node match {
      case mod: ir.StaticModifier => visitStaticModifier(mod)
      case mod: ir.Annotation => visitAnnotation(mod)
      case mod: ir.ModifierStringNode => visitModifierStringNode(mod)
      case mod: ir.PureModifier => visitPureModifier(mod)
      case mod: ir.NullableModifier => visitNullableModifier(mod)
      case mod: ir.NonnullModifier => visitNonnullModifier(mod)
      case mod: ir.GhostModifier => visitGhostModifier(mod)
      case mod: ir.HelperModifier => visitHelperModifier(mod)
      case mod: ir.MonitoredModifier => visitMonitoredModifier(mod)
      case mod: ir.ModelModifier => visitModelModifier(mod)
      case mod: ir.InstanceModifier => visitInstanceModifier(mod)
      case mod: ir.JmlModifierBlock => visitJmlModifierBlock(mod)
      case mod: ir.SpecMethodModifier => visitSpecMethodModifier(mod)
      case _ => ???
    }
    // todo: more?
  }
  def visitStaticModifier(node: ir.StaticModifier): jml.StaticModifier = {
    node match {
      case ir.StaticModifier(srcLocation, precedingWhitespace) =>
        val res = jml.StaticModifier(srcLocation, precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitGenericsSuperclassClause(node: ir.GenericsSuperclassClause): jml.GenericsSuperclassClause = {
    node match {
      case ir.GenericsSuperclassClause(superclasses, srcLocation, precedingWhitespace) =>
        val res = jml.GenericsSuperclassClause(superclasses.map(visitTypeIndication),
                                              srcLocation,
                                              precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitSuperclassClause(node: ir.SuperclassClause): jml.SuperclassClause = {
    node match {
      case ir.SuperclassClause(superclass, srcLocation, precedingWhitespace) =>
        val res = jml.SuperclassClause(visitTypeIndication(superclass),
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitMultiSuperclassClause(node: ir.MultiSuperclassClause): jml.MultiSuperclassClause = {
    node match {
      case ir.MultiSuperclassClause(superclasses, srcLocation, precedingWhitespace) =>
        val res = jml.MultiSuperclassClause(superclasses.map(visitTypeIndication),
                                            srcLocation,
                                            precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitInterfaceClause(node: ir.InterfaceClause): jml.InterfaceClause = {
    node match {
      case ir.InterfaceClause(interfaces, srcLocation, precedingWhitespace) =>
        val res = jml.InterfaceClause(interfaces.map(visitTypeIndication),
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitClassDeclaration(node: ir.ClassDeclaration): jml.ClassDeclaration = {
    node match {
      case ir.ClassDeclaration(name, typeParameters, superclass, interfaces, body, srcLocation, precedingWhitespace) =>
        val irTypeParameters = typeParameters.map(visitTypeParameterClause)
        val irSuperclass = superclass.map(visitSuperclassClause)
        val irInterfaces = interfaces.map(visitInterfaceClause)
        val res = jml.ClassDeclaration(visitStringNode(name),
                                      irTypeParameters,
                                      irSuperclass,
                                      irInterfaces,
                                      visitBody(body),
                                      srcLocation,
                                      precedingWhitespace)
        res.modifiers = node.modifiers.map(visitModifier)
        res.copyWhitespace(node)
        res
    }
  }

  def visitTypeParameterClause(node: ir.TypeParameterClause): jml.TypeParameterClause = {
    node match {
      case ir.TypeParameterClause(parameters, srcLocation, precedingWhitespace) =>
        val res = jml.TypeParameterClause(parameters.map(visitTypeParameter),
                                          srcLocation,
                                          precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitTypeParameter(node: ir.TypeParameter): jml.TypeParameter = {
    node match {
      case ir.TypeParameter(annotations, name, superclass, srcLocation, precedingWhitespace) =>
        val irSuperclass = superclass.map(visitGenericsSuperclassClause)
        val res = jml.TypeParameter(annotations.map(visitStringAnnotation),
                                    visitTypeIndication(name),
                                    irSuperclass,
                                    srcLocation,
                                    precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitEnumDeclaration(node: ir.EnumDeclaration): jml.EnumDeclaration = {
    node match {
      case ir.EnumDeclaration(name, interfaces, constants, declarations, comma, semicolon, srcLocation, precedingWhitespace) =>
        val irInterfaces = interfaces.map(visitInterfaceClause)
        val irComma = comma.map(visitStringNode)
        val irSemicolon = semicolon.map(visitStringNode)
        val res = jml.EnumDeclaration(visitStringNode(name),
                                      irInterfaces,
                                      constants.map(visitEnumConstant),
                                      declarations.map(visitMemberDeclaration),
                                      irComma,
                                      irSemicolon,
                                      srcLocation,
                                      precedingWhitespace)
        res.modifiers = node.modifiers.map(visitModifier)
        res.copyWhitespace(node)
        res
    }
  }
  def visitEnumConstant(node: ir.EnumConstant): jml.EnumConstant = {
    node match {
      case ir.EnumConstant(name, annotations, arguments, body, srcLocation, precedingWhitespace) =>
        val irArgs = arguments.map(visitArguments)
        val irBody = body.map(visitBody)
        val res = jml.EnumConstant(visitStringNode(name),
                                  annotations.map(visitStringAnnotation),
                                  irArgs,
                                  irBody,
                                  srcLocation,
                                  precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitInterfaceDeclaration(node: ir.InterfaceDeclaration): jml.InterfaceDeclaration = {
    node match {
      case ir.InterfaceDeclaration(name, typeParameter, superInterfaces, body, srcLocation, precedingWhitespace) =>
        val irTypeParameters = typeParameter.map(visitTypeParameterClause)
        val irSuperclass = superInterfaces.map(visitMultiSuperclassClause)
        val res = jml.InterfaceDeclaration(visitStringNode(name),
                                          irTypeParameters,
                                          irSuperclass,
                                          visitBody(body),
                                          srcLocation,
                                          precedingWhitespace)
        res.modifiers = node.modifiers.map(visitModifier)
        res.copyWhitespace(node)
        res
    }
  }

  def visitBody(node: ir.Body): jml.Body = {
    node match {
      case ir.Body(declarations, srcLocation, precedingWhitespace) =>
        val res = jml.Body(declarations.map(visitMemberDeclaration),
                          srcLocation,
                          precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitMemberDeclaration(node: ir.MemberDeclaration): jml.MemberDeclaration = {
    try
      node match {
        case member: ir.ClassDeclaration => visitClassDeclaration(member)
        case member: ir.EnumDeclaration => visitEnumDeclaration(member)
        case member: ir.InterfaceDeclaration => visitInterfaceDeclaration(member)
        case member: ir.GenericMethodDeclaration => visitGenericMethodDeclaration(member)
        case member: ir.MethodDeclaration => visitMethodDeclaration(member)
        case member: ir.BlockOrSemicolon => visitBlockOrSemicolon(member)
        case member: ir.FieldDeclaration => visitFieldDeclaration(member)
        case member: ir.AnnotationTypeDeclaration => visitAnnotationTypeDeclaration(member)
        case member: ir.SpecDeclaration => visitSpecDeclaration(member)
        case member: ir.SpecDeclarationBlock => visitSpecDeclarationBlock(member)
        case member: ir.NonTranslatedMemberDeclaration => visitNonTranslatedMemberDeclaration(member)
        case _ => ???
      }
    catch {
      case e: NotSupportedException =>
        val text = Helper.commentUnknownNode(node, s"${e.construct} not supported, whole MemberDeclaration commented out")
        jml.NonTranslatedMemberDeclaration(text, node.srcLocation)
    }
  }
  def visitNonTranslatedMemberDeclaration(node: ir.NonTranslatedMemberDeclaration): jml.NonTranslatedMemberDeclaration = {
    node match {
      case ir.NonTranslatedMemberDeclaration(text, srcLocation, precedingWhitespace) =>
        val res = jml.NonTranslatedMemberDeclaration(text, srcLocation, precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitMethodDeclaration(node: ir.MethodDeclaration): jml.MethodDeclaration = {
    node match {
      case ir.MethodDeclaration(name, returnType, parameters, dimensions, throws, default, contractWithin,
                                 body, srcLocation, precedingWhitespace) =>
        val irReturn = returnType.map(visitTypeIndication)
        val irDims = dimensions.map(visitStringNode)
        val irThrows = throws.map(visitThrowsClause)
        val irDefault = default.map(visitDefaultClause)
        val irContractWithin = contractWithin.map(visitContract)
        val res = new jml.MethodDeclaration(visitStringNode(name),
                                            irReturn,
                                            parameters.map(visitParameter),
                                            irDims,
                                            irThrows,
                                            irDefault,
                                            irContractWithin,
                                            visitBlockOrSemicolon(body),
                                            srcLocation,
                                            precedingWhitespace)
        res.contractBefore = if(node.contractBefore.isEmpty) None else Some(visitContract(node.contractBefore.get))
        res.modifiers = makeSpecPublic(node.modifiers.map(visitModifier), srcLocation)
        res.copyWhitespace(node)
        if(node.modifiers.isEmpty && res.modifiers.nonEmpty) {
          if(res.returnType.isDefined) {
            res.modifiers.head.precedingWhitespace = res.returnType.get.precedingWhitespace
            res.returnType.get.precedingWhitespace = " "
          } else {
            res.modifiers.head.precedingWhitespace = res.name.precedingWhitespace
            res.name.precedingWhitespace = " "
          }
        }
        res
    }
  }


  def visitGenericMethodDeclaration(node: ir.GenericMethodDeclaration): jml.GenericMethodDeclaration = {
    node match {
      case ir.GenericMethodDeclaration(pName, typeParameters, pReturnType,
                                        pParameters, pDimensions, pThrows, pDefault, pContractWithin,
                                        pBody, pSrcLocation, pPrecedingWhitespace) =>
        val irReturn = pReturnType.map(visitTypeIndication)
        val irDims = pDimensions.map(visitStringNode)
        val irThrows = pThrows.map(visitThrowsClause)
        val irDefault = pDefault.map(visitDefaultClause)
        val irContractWithin = pContractWithin.map(visitContract)
        val res = jml.GenericMethodDeclaration(visitStringNode(pName),
                                              visitTypeParameterClause(typeParameters),
                                              irReturn,
                                              pParameters.map(visitParameter),
                                              irDims,
                                              irThrows,
                                              irDefault,
                                              irContractWithin,
                                              visitBlockOrSemicolon(pBody),
                                              pSrcLocation,
                                              pPrecedingWhitespace)
        res.contractBefore = node.contractBefore.map(visitContract)
        res.modifiers = makeSpecPublic(node.modifiers.map(visitModifier),pSrcLocation)
        res.copyWhitespace(node)
        if(node.modifiers.isEmpty && res.modifiers.nonEmpty) {
          if(res.returnType.isDefined) {
            res.modifiers.head.precedingWhitespace = res.returnType.get.precedingWhitespace
            res.returnType.get.precedingWhitespace = " "
          } else {
            res.modifiers.head.precedingWhitespace = res.name.precedingWhitespace
            res.name.precedingWhitespace = " "
          }
        }
        res
    }
  }

  def visitThrowsClause(node: ir.ThrowsClause): jml.ThrowsClause = {
    node match {
      case ir.ThrowsClause(exceptions, srcLocation, precedingWhitespace) =>
        val res = jml.ThrowsClause(exceptions.map(visitStringNode),
                                  srcLocation,
                                  precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitTypeIndication(node: ir.TypeIndication): jml.TypeIndication = {
    node match {
      case indication: ir.SpecTypeIndication => visitSpecTypeIndication(indication)
      case indication: ir.NoType => visitNoType(indication)
      case indication: ir.StringGenericTypeIndication => visitStringGenericTypeIndication(indication)
      case indication: ir.StringTypeIndication => visitStringTypeIndication(indication)
      case _ => ???
    }
  }

  def visitSpecTypeIndication(node: ir.SpecTypeIndication): jml.SpecTypeIndication = {
    node match {
      case int: ir.IntegerSpecTypeIndication => visitIntegerSpecTypeIndication(int)
      case real: ir.RealSpecTypeIndication => visitRealSpecTypeIndication(real)
    }
  }

  def visitIntegerSpecTypeIndication(node: ir.IntegerSpecTypeIndication): jml.SpecTypeIndication = {
    node match {
      case ir.IntegerSpecTypeIndication(fullTypeString, anns, typeAsString, location, whitespace) =>
        val res = jml.IntegerSpecTypeIndication(annotations = anns.map(visitAnnotation), srcLocation = location, precedingWhitespace = whitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitRealSpecTypeIndication(node: ir.RealSpecTypeIndication): jml.SpecTypeIndication = {
    node match {
      case ir.RealSpecTypeIndication(fullTypeString, anns, typeAsString, location, whitespace) =>
        val res = jml.RealSpecTypeIndication(annotations = anns.map(visitAnnotation), srcLocation = location, precedingWhitespace = whitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitStringTypeIndication(node: ir.StringTypeIndication): jml.StringTypeIndication = {
    node match {
      case ir.StringTypeIndication(fullTypeString, annotations, typeAsString,
                                    pSrcLocation, pPrecedingWhitespace, pTrailingWhitespace) =>
        val res = new jml.StringTypeIndication(fullTypeString,
                                              annotations.map(visitAnnotation),
                                              typeAsString,
                                              pSrcLocation,
                                              pPrecedingWhitespace,
                                              pTrailingWhitespace)
        res.modifiers = node.modifiers.map(visitModifier)
        res.copyWhitespace(node)
        res
    }
  }

  def visitStringGenericTypeIndication(node: ir.StringGenericTypeIndication): jml.StringGenericTypeIndication = {
    node match {
      case ir.StringGenericTypeIndication(fullTypeString, typeParameters, annotations, typeAsString,
                                            pSrcLocation, pPrecedingWhitespace) =>
        val res = jml.StringGenericTypeIndication(fullTypeString,
                                                  visitTypeParameterClause(typeParameters),
                                                  annotations.map(visitAnnotation),
                                                  typeAsString,
                                                  pSrcLocation,
                                                  pPrecedingWhitespace)
        res.modifiers = node.modifiers.map(visitModifier)
        res.copyWhitespace(node)
        res
    }
  }
  def visitNoType(node: ir.NoType): jml.NoType = {
    node match {
      case ir.NoType(srcLocation) =>
        new jml.NoType(srcLocation)
    }
  }

  def visitBlockOrSemicolon(node: ir.BlockOrSemicolon): jml.BlockOrSemicolon = {
    node match {
      case block: ir.Block => visitBlock(block)
      case semi: ir.EmptyStatement => visitEmptyStatement(semi)
      case _ => ???
    }
  }

  def visitFieldDeclaration(node: ir.FieldDeclaration): jml.FieldDeclaration = {
    node match {
      case ir.FieldDeclaration(fieldType, declarators, srcLocation, precedingWhitespace) =>
        val res = jml.FieldDeclaration(visitTypeIndication(fieldType),
                                      declarators.map(visitVariableDeclarator),
                                      srcLocation,
                                      precedingWhitespace)
        res.modifiers = makeSpecPublic(node.modifiers.map(visitModifier), srcLocation)
        if(node.modifiers.isEmpty && res.modifiers.nonEmpty) {
          res.modifiers.head.precedingWhitespace = res.fieldType.precedingWhitespace
          res.fieldType.precedingWhitespace = " "
        }
        res.copyWhitespace(node)
        res
    }
  }

  def visitVariableDeclarator(node: ir.VariableDeclarator): jml.VariableDeclarator = {
    node match {
      case ir.VariableDeclarator(name, initializer, srcLocation, precedingWhitespace) =>
        val irInit = initializer.map(visitInitializer)
        val res = jml.VariableDeclarator(visitStringNode(name),
                                        irInit,
                                        srcLocation,
                                        precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitInitializer(node: ir.Initializer): jml.Initializer = {
    node match {
      case init: ir.ArrayInitializer => visitArrayInitializer(init)
      case init: ir.Expression => visitExpression(init)
      case _ => ???
    }
  }
  def visitArrayInitializer(node: ir.ArrayInitializer): jml.ArrayInitializer = {
    node match {
      case ir.ArrayInitializer(initializers, trailingComma, srcLocation, precedingWhitespace) =>
        val irComma = trailingComma.map(visitStringNode)
        val res = jml.ArrayInitializer(initializers.map(visitInitializer),
                                      irComma,
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitTypeArgumentClause(node: ir.TypeArgumentClause): jml.TypeArgumentClause = {
    node match {
      case ir.TypeArgumentClause(parameters, srcLocation, precedingWhitespace) =>
        val res = jml.TypeArgumentClause(parameters.map(visitTypeArgument),
                                        srcLocation,
                                        precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitTypeArgument(node: ir.TypeArgument): jml.TypeArgument = {
    node match {
      case arg: ir.BoundedTypeArgument => visitBoundedTypeArgument(arg)
      case arg: ir.TypeIndication => visitTypeIndication(arg)
      case _ => ???
    }
  }
  def visitBoundedTypeArgument(node: ir.BoundedTypeArgument): jml.BoundedTypeArgument = {
    node match {
      case ir.BoundedTypeArgument(annotations, typeSuper, typeExtends, srcLocation, precedingWhitespace) =>
        val irSuper = typeSuper.map(visitTypeIndication)
        val irExtends = typeExtends.map(visitTypeIndication)
        val res = jml.BoundedTypeArgument(annotations.map(visitAnnotation),
                                          irSuper,
                                          irExtends,
                                          srcLocation,
                                          precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitParameter(node: ir.Parameter): jml.Parameter = {
    node match {
      case varargs: ir.VarargsParameter =>
        visitVarargsParameter(varargs)
      case ir.Parameter(parameterType, name, srcLocation, precedingWhitespace) =>
        val res = jml.Parameter(visitTypeIndication(parameterType),
                                visitStringNode(name),
                                srcLocation,
                                precedingWhitespace)
        res.modifiers = node.modifiers.map(visitModifier)
        res.copyWhitespace(node)
        res
    }
  }
  def visitVarargsParameter(node: ir.VarargsParameter): jml.VarargsParameter = {
    node match {
      case ir.VarargsParameter(parameterType, annotations, name, srcLocation, precedingWhitespace) =>
        val res = jml.VarargsParameter(visitTypeIndication(parameterType),
                                      annotations.map(visitAnnotation),
                                      visitStringNode(name),
                                      srcLocation,
                                      precedingWhitespace)
        res.modifiers = node.modifiers.map(visitModifier)
        res.copyWhitespace(node)
        res
    }
  }

  def visitAnnotation(node: ir.Annotation): jml.Annotation = {
    node match {
      case ann: ir.StringAnnotation => visitStringAnnotation(ann)
      case _ => ???
    }
  }

  def visitStringAnnotation(node: ir.StringAnnotation): jml.StringAnnotation = {
    node match {
      case ir.StringAnnotation(text, srcLocation, precedingWhitespace) =>
        val res = jml.StringAnnotation(text,
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitAnnotationTypeDeclaration(node: ir.AnnotationTypeDeclaration): jml.AnnotationTypeDeclaration = {
    node match {
      case ir.AnnotationTypeDeclaration(name, body, srcLocation, precedingWhitespace) =>
        val res = jml.AnnotationTypeDeclaration(visitStringNode(name),
                                                visitBody(body),
                                                srcLocation,
                                                precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitDefaultClause(node: ir.DefaultClause): jml.DefaultClause = {
    node match {
      case ir.DefaultClause(value, srcLocation, precedingWhitespace) =>
        val res = jml.DefaultClause(visitStringNode(value),
                                    srcLocation,
                                    precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitStatement(node: ir.Statement): jml.Statement = {
    try
      node match {
        case local: ir.LocalDeclaration => visitLocalDeclaration(local)
        case block: ir.Block => visitBlock(block)
        case assert: ir.JavaAssertion => visitJavaAssertion(assert)
        case ifStmt: ir.IfStatement => visitIfStatement(ifStmt)
        case forLoop: ir.ForLoop => visitForLoop(forLoop)
        case whileLoop: ir.WhileLoop => visitWhileLoop(whileLoop)
        case doLoop: ir.DoLoop => visitDoLoop(doLoop)
        case tryBlock: ir.TryBlock => visitTryBlock(tryBlock)
        case switch: ir.SwitchStatement => visitSwitchStatement(switch)
        case sync: ir.SynchronizedStatement => visitSynchronizedStatement(sync)
        case ret: ir.ReturnStatement => visitReturnStatement(ret)
        case throwStmt: ir.ThrowStatement => visitThrowStatement(throwStmt)
        case break: ir.BreakStatement => visitBreakStatement(break)
        case continue: ir.ContinueStatement => visitContinueStatement(continue)
        case empty: ir.EmptyStatement => visitEmptyStatement(empty)
        case expr: ir.ExpressionStatement => visitExpressionStatement(expr)
        case labeled: ir.LabeledStatement => visitLabeledStatement(labeled)
        case assert: ir.AssertStatement => visitAssertStatement(assert)
        case assume: ir.AssumeStatement => visitAssumeStatement(assume)
        case set: ir.SetStatement => visitSetStatement(set)
        case unreach: ir.UnreachableStatement => visitUnreachableStatement(unreach)
        case ghost: ir.GhostStatement => visitGhostStatement(ghost)
        case jmlBlock: ir.JmlStatementBlock => visitJmlStatementBlock(jmlBlock)
        case nontrans: ir.NonTranslatedStatement => visitNonTranslatedStatement(nontrans)
        case _ => ???
        // todo: ClassDecl, etc.
      }
    catch {
      case e: NotSupportedException =>
        val text = Helper.commentUnknownNode(node, s"${e.construct} not supported, whole Statement commented out.")
        jml.NonTranslatedStatement(text, node.srcLocation)
      case e: PermissionExpressionException =>
        val text = Helper.commentUnknownNode(node, "statement about permissions commented out.")
        jml.NonTranslatedStatement(text, node.srcLocation)
    }
  }

//  def visitGhostStatement(node: ir.GhostStatement): jml.GhostStatement = {
//    node match {
//      case ir.GhostStatement(statement, srcLocation, precedingWhitespace) =>
//        jml.GhostStatement(visitStatement(statement), srcLocation, precedingWhitespace)
//    }
//  }

  def visitBlock(node: ir.Block): jml.Block = {
    node match {
      case ir.Block(static, statements, srcLocation, precedingWhitespace) =>
        val irStatic = static.map(visitModifier)
        val res = jml.Block(irStatic,
                            statements.map(visitStatement),
                            srcLocation,
                            precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitLocalDeclaration(node: ir.LocalDeclaration): jml.LocalDeclaration = {
    node match {
      case ir.LocalDeclaration(modifiers, declType, declarators, isStatement, srcLocation, precedingWhitespace) =>
        val res = jml.LocalDeclaration(modifiers.map(visitModifier),
                                      visitTypeIndication(declType),
                                      declarators.map(visitVariableDeclarator),
                                      isStatement,
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitJavaAssertion(node: ir.JavaAssertion): jml.JavaAssertion = {
    node match {
      case ir.JavaAssertion(content, message, srcLocation, precedingWhitespace) =>
        val irMsg = message.map(visitExpression)
        val res = jml.JavaAssertion(visitExpression(content),
                                    irMsg,
                                    srcLocation,
                                    precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitIfStatement(node: ir.IfStatement): jml.IfStatement = {
    node match {
      case ir.IfStatement(condition, thenStatement, elseStatement, srcLocation, precedingWhitespace) =>
        val irElse = elseStatement.map(visitStatement)
        val res = jml.IfStatement(visitExpression(condition),
                                  visitStatement(thenStatement),
                                  irElse,
                                  srcLocation,
                                  precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitForLoop(node: ir.ForLoop): jml.ForLoop = {
    node match {
      case ir.ForLoop(contractBefore, control, contractWithin, body, srcLocation, precedingWhitespace) =>
        // OpenJML does not support contract within, thus joined to contract before
        val irContractWithin = contractWithin.map(visitLoopSpecOrLabel)
        val irContractBefore = contractBefore.map(visitLoopSpecOrLabel).appendedAll(irContractWithin)
        val res = jml.ForLoop(irContractBefore,
                              visitForLoopControl(control),
                              Nil,
                              visitStatement(body),
                              srcLocation,
                              precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitForLoopControl(node: ir.ForLoopControl): jml.ForLoopControl = {
    node match {
      case init: ir.ForEachLoopControl => visitForEachLoopControl(init)
      case init: ir.RegularForLoopControl => visitRegularForLoopControl(init)
    }
  }
  def visitForInit(node: ir.ForInit): jml.ForInit = {
    node match {
      case init: ir.ExpressionListForInit =>
        visitExpressionListForInit(init)
      case init: ir.LocalDeclaration =>
        visitLocalDeclaration(init)
      case _ =>
        ???
    }
  }
  def visitExpressionListForInit(node: ir.ExpressionListForInit): jml.ExpressionListForInit = {
    node match {
      case ir.ExpressionListForInit(expressions, srcLocation, precedingWhitespace) =>
        val res = jml.ExpressionListForInit(expressions.map(visitExpression),
                                          srcLocation,
                                          precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitRegularForLoopControl(node: ir.RegularForLoopControl): jml.RegularForLoopControl = {
    node match {
      case ir.RegularForLoopControl(init, condition, update, srcLocation, precedingWhitespace) =>
        val irInit = init.map(visitForInit)
        val irCondition = condition.map(visitExpression)
        val res = jml.RegularForLoopControl(irInit,
                                            irCondition,
                                            update.map(visitExpression),
                                            srcLocation,
                                            precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitForEachLoopControl(node: ir.ForEachLoopControl): jml.ForEachLoopControl = {
    node match {
      case ir.ForEachLoopControl(controlType, variable, control, srcLocation, precedingWhitespace) =>
        val res = jml.ForEachLoopControl(visitTypeIndication(controlType),
                                        visitStringNode(variable),
                                        visitExpression(control),
                                        srcLocation,
                                        precedingWhitespace)
        res.modifiers = node.modifiers.map(visitModifier)
        res.copyWhitespace(node)
        res
    }
  }

  def visitWhileLoop(node: ir.WhileLoop): jml.WhileLoop = {
    node match {
      case ir.WhileLoop(contractBefore, condition, contractWithin, body, srcLocation, precedingWhitespace) =>
        // OpenJML does not support contract within, thus joined to contract before
        val irContractWithin = contractWithin.map(visitLoopSpecOrLabel)
        val irContractBefore = contractBefore.map(visitLoopSpecOrLabel).appendedAll(irContractWithin)
        val res = jml.WhileLoop(irContractBefore,
                                visitExpression(condition),
                                Nil,
                                visitStatement(body),
                                srcLocation,
                                precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitDoLoop(node: ir.DoLoop): jml.DoLoop = {
    node match {
      case ir.DoLoop(contractBefore, body, condition, srcLocation, precedingWhitespace) =>
        val irContractBefore = contractBefore.map(visitLoopSpecOrLabel)
        val res = jml.DoLoop(irContractBefore,
                            visitStatement(body),
                            visitExpression(condition),
                            srcLocation,
                            precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitTryBlock(node: ir.TryBlock): jml.TryBlock = {
    node match {
      case ir.TryBlock(resourceSpecification, tryBlock, catchClauses, finallyBlock, srcLocation, precedingWhitespace) =>
        val irResSpec = if (resourceSpecification.isEmpty) None
                        else Some(visitTryResourceSpecification(resourceSpecification.get))
        val irFinally = finallyBlock.map(visitBlock)
        val res = jml.TryBlock(irResSpec,
                              visitBlock(tryBlock),
                              catchClauses.map(visitCatchClause),
                              irFinally,
                              srcLocation,
                              precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitCatchClause(node: ir.CatchClause): jml.CatchClause = {
    node match {
      case ir.CatchClause(types, name, block, srcLocation, precedingWhitespace) =>
        val res = jml.CatchClause(types.map(visitTypeIndication),
                                  visitStringNode(name),
                                  visitBlock(block),
                                  srcLocation,
                                  precedingWhitespace)
        res.modifiers = node.modifiers.map(visitModifier)
        res.copyWhitespace(node)
        res
    }
  }
  def visitTryResourceSpecification(node: ir.TryResourceSpecification): jml.TryResourceSpecification = {
    node match {
      case ir.TryResourceSpecification(resources, trailingSemicolon, srcLocation, precedingWhitespace) =>
        val irSemi = trailingSemicolon.map(visitStringNode)
        val res = jml.TryResourceSpecification(resources.map(visitTryResource),
                                              irSemi,
                                              srcLocation,
                                              precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitTryResource(node: ir.TryResource): jml.TryResource = {
    node match {
      case ir.TryResource(modifiers, resourceType, variable, expression, srcLocation, precedingWhitespace) =>
        val res = jml.TryResource(modifiers.map(visitModifier),
                                  visitTypeIndication(resourceType),
                                  visitStringNode(variable),
                                  visitExpression(expression),
                                  srcLocation,
                                  precedingWhitespace)
        res.modifiers = node.modifiers.map(visitModifier)
        res.copyWhitespace(node)
        res
    }
  }

  def visitSwitchStatement(node: ir.SwitchStatement): jml.SwitchStatement = {
    node match {
      case ir.SwitchStatement(condition, cases, srcLocation, precedingWhitespace) =>
        val res = jml.SwitchStatement(visitExpression(condition),
                                      cases.map(visitSwitchCaseBlock),
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitSwitchCaseBlock(node: ir.SwitchCaseBlock): jml.SwitchCaseBlock = {
    node match {
      case ir.SwitchCaseBlock(labels, statements, srcLocation, precedingWhitespace) =>
        val res = jml.SwitchCaseBlock(labels.map(visitSwitchLabel),
                                      statements.map(visitStatement),
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitSwitchLabel(node: ir.SwitchLabel): jml.SwitchLabel = {
    node match {
      case ir.SwitchLabel(expression, srcLocation, precedingWhitespace) =>
        val irExpr = expression.map(visitExpression)
        val res = jml.SwitchLabel(irExpr,
                                  srcLocation,
                                  precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitSynchronizedStatement(node: ir.SynchronizedStatement): jml.SynchronizedStatement = {
    node match {
      case ir.SynchronizedStatement(condition, body, srcLocation, precedingWhitespace) =>
        val res = jml.SynchronizedStatement(visitExpression(condition),
                                            visitStatement(body),
                                            srcLocation,
                                            precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitReturnStatement(node: ir.ReturnStatement): jml.ReturnStatement = {
    node match {
      case ir.ReturnStatement(expression, srcLocation, precedingWhitespace) =>
        val irExpr = expression.map(visitExpression)
        val res = jml.ReturnStatement(irExpr,
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitThrowStatement(node: ir.ThrowStatement): jml.ThrowStatement = {
    node match {
      case ir.ThrowStatement(expression, srcLocation, precedingWhitespace) =>
        val res = jml.ThrowStatement(visitExpression(expression),
                                    srcLocation,
                                    precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitBreakStatement(node: ir.BreakStatement): jml.BreakStatement = {
    node match {
      case ir.BreakStatement(label, srcLocation, precedingWhitespace) =>
        val irLabel = label.map(visitStringNode)
        val res = jml.BreakStatement(irLabel,
                                    srcLocation,
                                    precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitContinueStatement(node: ir.ContinueStatement): jml.ContinueStatement = {
    node match {
      case ir.ContinueStatement(label, srcLocation, precedingWhitespace) =>
        val irLabel = label.map(visitStringNode)
        val res = jml.ContinueStatement(irLabel,
                                        srcLocation,
                                        precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitEmptyStatement(node: ir.EmptyStatement): jml.EmptyStatement = {
    val res = new jml.EmptyStatement(node.srcLocation)
    res.copyWhitespace(node)
    res
  }
  def visitExpressionStatement(node: ir.ExpressionStatement): jml.ExpressionStatement = {
    node match {
      case ir.ExpressionStatement(expression, srcLocation, precedingWhitespace) =>
        val res = jml.ExpressionStatement(visitExpression(expression),
                                          srcLocation,
                                          precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitLabeledStatement(node: ir.LabeledStatement): jml.LabeledStatement = {
    node match {
      case ir.LabeledStatement(label, statement, srcLocation, precedingWhitespace) =>
        val res = jml.LabeledStatement(visitLabel(label),
                                      visitStatement(statement),
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitLabel(node: ir.Label): jml.Label = {
    node match {
      case ir.Label(label, srcLocation, precedingWhitespace) =>
        val res = jml.Label(visitStringNode(label),
                            srcLocation,
                            precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitArguments(node: ir.Arguments): jml.Arguments = {
    node match {
      case ir.Arguments(arguments, srcLocation, precedingWhitespace) =>
        val res = jml.Arguments(arguments.map(visitExpression),
                                srcLocation,
                                precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }


  def visitExpression(node: ir.Expression): jml.Expression = {
    node match {
      case expr: ir.NestedInvocationExpression => visitNestedInvocationExpression(expr)
      case expr: ir.BinaryExpression => visitBinaryExpression(expr)
      case expr: ir.InstanceofExpression => visitInstanceofExpression(expr)
      case expr: ir.ArrayIndexExpression => visitArrayIndexExpression(expr)
      case expr: ir.MethodCallExpression => visitMethodCallExpression(expr)
      case expr: ir.NewObjectExpression => visitNewObjectExpression(expr)
      case expr: ir.NewArrayExpression => visitNewArrayExpression(expr)
      case expr: ir.CastExpression => visitCastExpression(expr)
      case expr: ir.PostfixExpression => visitPostfixExpression(expr)
      case expr: ir.PrefixExpression => visitPrefixExpression(expr)
      case expr: ir.TernaryExpression => visitTernaryExpression(expr)
      case expr: ir.LambdaExpression => visitLambdaExpression(expr)
      case expr: ir.MethodReferenceExpression => visitMethodReferenceExpression(expr)
      case expr: ir.ParExpression => visitParExpression(expr)
      case expr: ir.Literal => visitLiteral(expr)
      case expr: ir.ClassName => visitClassName(expr)
      case expr: ir.GenericInvocationExpression => visitGenericInvocationExpression(expr)
      case expr: ir.ResultExpression => visitResultExpression(expr)
      case expr: ir.OldExpression => visitOldExpression(expr)
      case expr: ir.PreExpression => visitPreExpression(expr)
      case expr: ir.NotAssignedExpression => visitNotAssignedExpression(expr)
      case expr: ir.NotModifiedExpression => visitNotModifiedExpression(expr)
      case expr: ir.OnlyAccessedExpression => visitOnlyAccessedExpression(expr)
      case expr: ir.OnlyAssignedExpression => visitOnlyAssignedExpression(expr)
      case expr: ir.FreshExpression => visitFreshExpression(expr)
      case expr: ir.NonNullElementsExpression => visitNonNullElementsExpression(expr)
      case expr: ir.TypeOfExpression => visitTypeOfExpression(expr)
      case expr: ir.TypeExpression => visitTypeExpression(expr)
      case expr: ir.ForallExpression => visitForallExpression(expr)
      case expr: ir.ExistsExpression => visitExistsExpression(expr)
      case expr: ir.MaxQuantifiedExpression => visitMaxQuantifiedExpression(expr)
      case expr: ir.MinExpression => visitMinExpression(expr)
      case expr: ir.NumOfExpression => visitNumOfExpression(expr)
      case expr: ir.ProductExpression => visitProductExpression(expr)
      case expr: ir.SumExpression => visitSumExpression(expr)
      case expr: ir.StringNode => visitStringNode(expr)
      case expr: ir.SetComprehension => visitSetComprehension(expr)
      case expr: ir.UnfoldingExpression => visitUnfoldingExpression(expr)
      case expr: ir.ValuedPermissionExpression => visitValuedPermissionExpression(expr)
      case expr: ir.PermissionExpression => throw new PermissionExpressionException(expr)
      case expr: ir.CurrentPermissionExpression => throw new CurrentPermissionValueException()
      case expr: ir.ForallWithPermissionsExpression => visitForallWithPermissionsExpression(expr)
      case expr: ir.ScaledPredicateExpression => visitScaledPredicateExpression(expr)
      case _ => ???
    }
    //todo: more?

  }

  def visitNonTranslatedStatement(node: ir.NonTranslatedStatement): jml.NonTranslatedStatement = {
    node match {
      case ir.NonTranslatedStatement(text, srcLocation, precedingWhitespace) =>
        val res = jml.NonTranslatedStatement(text, srcLocation, precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitNestedInvocationExpression(node: ir.NestedInvocationExpression): jml.NestedInvocationExpression = {
    node match {
      case ir.NestedInvocationExpression(left, right, srcLocation, precedingWhitespace) =>
        val res = jml.NestedInvocationExpression(visitExpression(left),
                                                visitExpression(right),
                                                srcLocation,
                                                precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitBinaryExpression(node: ir.BinaryExpression): jml.Expression = {
    node match {
      case ir.BinaryExpression(left, operator, right, srcLocation, precedingWhitespace) =>
        var leftJML: jml.Expression = null
        var rightJML: jml.Expression = null
        try
          leftJML = visitExpression(left)
        catch {
          case p: PermissionExpressionException =>
            collectedPermissions.append(p.expr)
          case p: MultiplePermissionsException =>
        }
        try
          rightJML = visitExpression(right)
        catch {
          case p: PermissionExpressionException =>
            collectedPermissions.append(p.expr)
          case p: MultiplePermissionsException =>
        }
        if(leftJML == null && rightJML != null) {
          val leftText = left.toString.replaceAll("""\s+""", " ")
          System.err.println(s"[info] left side of binary expression (\"${leftText}\") was a permission, and was dropped${left.srcLocation}")
          rightJML
        } else if(leftJML != null && rightJML == null && !operator.isInstanceOf[ir.ImpliesOperator]) {
          val rightText = right.toString.replaceAll("""\s+""", " ")
          System.err.println(s"[info] right side of binary expression (\"${rightText}\") was a permission, and was dropped${right.srcLocation}")
          leftJML
        } else if(leftJML != null && rightJML != null) {
          val op = visitStringNode(operator)
          val res = jml.BinaryExpression(leftJML,
                                        op,
                                        rightJML,
                                        srcLocation,
                                        precedingWhitespace)
          res.copyWhitespace(node)
          res
        } else /* both leftJML and rightJML are permissions */ {
          throw new MultiplePermissionsException()
        }
    }
  }

  def visitInstanceofExpression(node: ir.InstanceofExpression): jml.InstanceofExpression = {
    node match {
      case ir.InstanceofExpression(left, right, srcLocation, precedingWhitespace) =>
        val res = jml.InstanceofExpression(visitExpression(left),
          visitTypeIndication(right),
          srcLocation)
        res.copyWhitespace(node)
        res
    }
  }

  def visitArrayIndexExpression(node: ir.ArrayIndexExpression): jml.ArrayIndexExpression = {
    node match {
      case ir.ArrayIndexExpression(arrayExpression, indexExpression, srcLocation, precedingWhitespace) =>
        val res = jml.ArrayIndexExpression(visitExpression(arrayExpression),
                                          visitIndexExpression(indexExpression),
                                          srcLocation,
                                          precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitIndexExpression(node: ir.IndexExpression): jml.IndexExpression = {
    node match {
      case idx: ir.IndexAllExpression => visitIndexAllExpression(idx)
      case idx: ir.RangeExpression => visitRangeExpression(idx)
      case idx: ir.Expression => visitExpression(idx)
      case _ => ???
    }
  }

  def visitIndexAllExpression(node: ir.IndexAllExpression): jml.IndexAllExpression = {
    node match {
      case ir.IndexAllExpression(srcLocation, precedingWhitespace) =>
        val res = jml.IndexAllExpression(srcLocation, precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitRangeExpression(node: ir.RangeExpression): jml.RangeExpression = {
    node match {
      case ir.RangeExpression(start, end, srcLocation, precedingWhitespace) =>
        val res = jml.RangeExpression(visitExpression(start),
                                      visitExpression(end),
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitMethodCallExpression(node: ir.MethodCallExpression): jml.MethodCallExpression = {
    node match {
      case ir.MethodCallExpression(name, arguments, srcLocation, precedingWhitespace) =>
        val res = jml.MethodCallExpression(visitStringNode(name),
                                          visitArguments(arguments),
                                          srcLocation,
                                          precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitNewObjectExpression(node: ir.NewObjectExpression): jml.NewObjectExpression = {
    node match {
      case ir.NewObjectExpression(typeArguments, typeIndication, arguments, body, srcLocation, precedingWhitespace) =>
        val typeArgs = typeArguments.map(visitTypeArgumentClause)
        val irBody = body.map(visitBody)
        val res = jml.NewObjectExpression(typeArgs,
                                          visitTypeIndication(typeIndication),
                                          visitArguments(arguments),
                                          irBody,
                                          srcLocation,
                                          precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitNewArrayExpression(node: ir.NewArrayExpression): jml.NewArrayExpression = {
    node match {
      case ir.NewArrayExpression(typeIndication, dimensions, initializer, srcLocation, precedingWhitespace) =>
        val irInit = initializer.map(visitArrayInitializer)
        val res = jml.NewArrayExpression(visitTypeIndication(typeIndication),
                                        dimensions.map(visitDimension),
                                        irInit,
                                        srcLocation,
                                        precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitDimension(node: ir.Dimension): jml.Dimension = {
    node match {
      case ir.Dimension(sizeExpression, srcLocation, precedingWhitespace) =>
        val irSize = sizeExpression.map(visitExpression)
        val res = jml.Dimension(irSize,
                                srcLocation,
                                precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitCastExpression(node: ir.CastExpression): jml.CastExpression = {
    node match {
      case ir.CastExpression(annotations, typeIndications, expression, srcLocation, precedingWhitespace) =>
        val res = jml.CastExpression(annotations.map(visitStringAnnotation),
                                    typeIndications.map(visitTypeIndication),
                                    visitExpression(expression),
                                    srcLocation,
                                    precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitPostfixExpression(node: ir.PostfixExpression): jml.PostfixExpression = {
    node match {
      case ir.PostfixExpression(expression, postfix, srcLocation, precedingWhitespace) =>
        val res = jml.PostfixExpression(visitExpression(expression),
                                        visitStringNode(postfix),
                                        srcLocation,
                                        precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitPrefixExpression(node: ir.PrefixExpression): jml.PrefixExpression = {
    node match {
      case ir.PrefixExpression(expression, prefix, srcLocation, precedingWhitespace) =>
        val res = jml.PrefixExpression(visitExpression(expression),
                                      visitStringNode(prefix),
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitTernaryExpression(node: ir.TernaryExpression): jml.TernaryExpression = {
    node match {
      case ir.TernaryExpression(condition, thenExpression, elseExpression, srcLocation, precedingWhitespace) =>
        val res = jml.TernaryExpression(visitExpression(condition),
                                        visitExpression(thenExpression),
                                        visitExpression(elseExpression),
                                        srcLocation,
                                        precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitLambdaExpression(node: ir.LambdaExpression): jml.LambdaExpression = {
    node match {
      case ir.LambdaExpression(hasParentheses, parameters, body, srcLocation, precedingWhitespace) =>
        val res = jml.LambdaExpression(hasParentheses,
                                      parameters.map(visitParameter),
                                      visitLambdaBody(body),
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitLambdaBody(node: ir.LambdaBody): jml.LambdaBody = {
    node match {
      case body: ir.Expression => visitExpression(body)
      case _ => ???
    }
  }

  def visitMethodReferenceExpression(node: ir.MethodReferenceExpression): jml.MethodReferenceExpression = {
    node match {
      case ir.MethodReferenceExpression(left, typeArgument, right, srcLocation, precedingWhitespace) =>
        val irArg = typeArgument.map(visitTypeArgumentClause)
        val res = jml.MethodReferenceExpression(visitExpression(left),
                                                irArg,
                                                visitExpression(right),
                                                srcLocation,
                                                precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitParExpression(node: ir.ParExpression): jml.ParExpression = {
    node match {
      case ir.ParExpression(expression, srcLocation, precedingWhitespace) =>
        val res = jml.ParExpression(visitExpression(expression),
                                    srcLocation,
                                    precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitLiteral(node: ir.Literal): jml.Literal = {
    node match {
      case ir.Literal(value, srcLocation, precedingWhitespace) =>
        val res = jml.Literal(visitStringNode(value),
                              srcLocation,
                              precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitClassName(node: ir.ClassName): jml.ClassName = {
    node match {
      case ir.ClassName(typeIndication, srcLocation, precedingWhitespace) =>
        val res = jml.ClassName(visitTypeIndication(typeIndication),
                                srcLocation,
                                precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitGenericInvocationExpression(node: ir.GenericInvocationExpression): jml.GenericInvocationExpression = {
    node match {
      case ir.GenericInvocationExpression(typeArgumentClause, name, arguments, srcLocation, precedingWhitespace) =>
        val irArgs = arguments.map(visitArguments)
        val res = jml.GenericInvocationExpression(visitTypeArgumentClause(typeArgumentClause),
                                                  visitStringNode(name),
                                                  irArgs,
                                                  srcLocation,
                                                  precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitJmlModifierBlock(node: ir.JmlModifierBlock): jml.JmlModifierBlock = {
    node match {
      case ir.JmlModifierBlock(startSpec, modifiers, endSpec, srcLocation, precedingWhitespace) =>
        val irStart = startSpec.map(visitStringNode).map(transformStartSpec)
        val irEnd = endSpec.map(visitStringNode)
        val oldWrapInSpec = wrapInSpec
        wrapInSpec = false
        val res = jml.JmlModifierBlock(irStart,
                                      modifiers.map(visitModifier),
                                      irEnd,
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        wrapInSpec = oldWrapInSpec
        res
    }
  }

  def visitPureModifier(node: ir.PureModifier): jml.PureModifier = {
    node match {
      case ir.PureModifier(srcLocation, precedingWhitespace) =>
        val res = jml.PureModifier(srcLocation, precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitNullableModifier(node: ir.NullableModifier): jml.NullableModifier = {
    node match {
      case ir.NullableModifier(srcLocation, precedingWhitespace) =>
        val res = jml.NullableModifier(srcLocation, precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitNonnullModifier(node: ir.NonnullModifier): jml.NonnullModifier = {
    node match {
      case ir.NonnullModifier(srcLocation, precedingWhitespace) =>
        val res = jml.NonnullModifier(srcLocation, precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitGhostModifier(node: ir.GhostModifier): jml.GhostModifier = {
    node match {
      case ir.GhostModifier(srcLocation, precedingWhitespace) =>
        val res = jml.GhostModifier(srcLocation, precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitSpecMethodModifier(node: ir.SpecMethodModifier): jml.SpecMethodModifier = {
    node match {
      case ir.SpecMethodModifier(srcLocation, precedingWhitespace) =>
        val res = jml.SpecMethodModifier(srcLocation, precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitHelperModifier(node: ir.HelperModifier): jml.HelperModifier = {
    node match {
      case ir.HelperModifier(srcLocation, precedingWhitespace) =>
        val res = jml.HelperModifier(srcLocation, precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitMonitoredModifier(node: ir.MonitoredModifier): jml.MonitoredModifier = {
    node match {
      case ir.MonitoredModifier(srcLocation, precedingWhitespace) =>
        val res = jml.MonitoredModifier(srcLocation, precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitModelModifier(node: ir.ModelModifier): jml.ModelModifier = {
    node match {
      case ir.ModelModifier(srcLocation, precedingWhitespace) =>
        val res = jml.ModelModifier(srcLocation, precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitInstanceModifier(node: ir.InstanceModifier): jml.InstanceModifier = {
    node match {
      case ir.InstanceModifier(srcLocation, precedingWhitespace) =>
        val res = jml.InstanceModifier(srcLocation, precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitContract(node: ir.Contract): jml.Contract = {
    node match {
      case ir.Contract(clauses, srcLocation, precedingWhitespace) =>
        collectedPermissions = ListBuffer()
        val res = jml.Contract(clauses.map(visitContractClause),
                              srcLocation,
                              precedingWhitespace)
        res.startSpec = node.startSpec.map(visitStringNode).map(transformStartSpec)
        res.endSpec = node.endSpec.map(visitStringNode)
        if(collectedPermissions.nonEmpty) {
          val modifies: ListBuffer[jml.AssignableClause] = ListBuffer()
          for(perm <- collectedPermissions) {
            val target = visitExpression(perm.target)
            target.precedingWhitespace = target.precedingWhitespace + " "
            val ass = jml.AssignableClause(List(target), perm.srcLocation)
            if(res.endSpec.isEmpty && wrapInSpec) {
              ass.assignableKeyword.precedingWhitespace = " "
              ass.startSpec = Some(jml.StringNode("//@", perm.srcLocation))
              if(modifies.isEmpty) {
                ass.startSpec.get.precedingWhitespace = System.lineSeparator() + "    "
              } else {
                ass.startSpec.get.precedingWhitespace = "    "
              }
              ass.endSpec = Some(jml.StringNode(System.lineSeparator(), perm.srcLocation))
            } else {
              ass.precedingWhitespace = System.lineSeparator() + "      "
            }
            modifies.append(ass)
          }
          res.clauses = res.clauses.appendedAll(modifies)
        }
        res.copyWhitespace(node)
        res
    }
  }

  def visitSpecCase(node: ir.SpecCase): jml.SpecCase = {
    collectedPermissions = ListBuffer()
    val res = node match {
      case spec: ir.NormalBehavior => visitNormalBehavior(spec)
      case spec: ir.ExceptionalBehavior => visitExceptionalBehavior(spec)
      case spec: ir.Behavior => visitBehavior(spec)
      case spec: ir.SpecCase => visitLightSpecCase(spec)
      case _ => ???
    }
    if(collectedPermissions.nonEmpty) {
      val modifies: ListBuffer[jml.AssignableClause] = ListBuffer()
      for(perm <- collectedPermissions) {
        val target = visitExpression(perm.target)
        target.precedingWhitespace = target.precedingWhitespace + " "
        val ass = jml.AssignableClause(List(target), perm.srcLocation, System.lineSeparator() + "      ")
        modifies.append(ass)
      }
      res.clauses = res.clauses.appendedAll(modifies)
    }
    res
  }

  def visitLightSpecCase(node: ir.SpecCase): jml.SpecCase = {
    node match {
      case ir.SpecCase(startSpec, hasAlso, clauses, endSpec, srcLocation, precedingWhitespace) =>
        val irStart = startSpec.map(visitStringNode).map(transformStartSpec)
        val irEnd = endSpec.map(visitStringNode)
        val res = jml.SpecCase(irStart,
                                    hasAlso,
                                    clauses.map(visitContractClause),
                                    irEnd,
                                    srcLocation,
                                    precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitLoopSpecOrLabel(node: ir.LoopSpecOrLabel): jml.LoopSpecOrLabel = {
    node match {
      case l: ir.Label => visitLabel(l)
      case spec: ir.LoopSpecification => visitLoopSpecification(spec)
    }
  }

  def visitLoopSpecification(node: ir.LoopSpecification): jml.LoopSpecification = {
    node match {
      case ir.LoopSpecification(startSpec, clauses, endSpec, srcLocation, precedingWhitespace) =>
        collectedPermissions = ListBuffer()
        var jmlClauses = clauses.map(visitLoopClause)
        val jmlStart = startSpec.map(visitStringNode).map(transformStartSpec)
        val jmlEnd = endSpec.map(visitStringNode)
        if(collectedPermissions.nonEmpty) {
          val modifies: ListBuffer[jml.LoopModifiesClause] = ListBuffer()
          for(perm <- collectedPermissions) {
            val target = visitExpression(perm.target)
            target.precedingWhitespace = target.precedingWhitespace + " "
            val ass = jml.LoopModifiesClause(List(target), perm.srcLocation, System.lineSeparator() + "      ")
            if(wrapInSpec && !jmlEnd.getOrElse("").toString.contains("*/")) {
              if(jmlStart.isDefined) {
                if(jmlStart.get.text.contains("//@")){
                  jmlStart.get.text = jmlStart.get.text.replace("//@", "/*@")
                } else if (jmlStart.get.text.startsWith("@") && jmlStart.get.precedingWhitespace.endsWith("//")) {
                  jmlStart.get.precedingWhitespace = jmlStart.get.precedingWhitespace.take(jmlStart.get.precedingWhitespace.length-1) + '*'
                }
              }
              if(jmlEnd.isDefined) {
                jmlEnd.get.trailingWhitespace = jmlEnd.get.text + jmlEnd.get.trailingWhitespace
                jmlEnd.get.text = "*/"
              }
            }
            modifies.append(ass)
          }
          jmlClauses = jmlClauses.appendedAll(modifies)
        }
        val res = jml.LoopSpecification(jmlStart, jmlClauses, jmlEnd, srcLocation, precedingWhitespace)
        res
    }
  }

  def visitContractClause(node: ir.ContractClause): jml.ContractClause = {
    try
      val res = node match {
        case clause: ir.AssignableClause => visitAssignableClause(clause)
        case clause: ir.AccessibleClause => visitAccessibleClause(clause)
        case clause: ir.ForallVariableDeclaration => visitForallVariableDeclaration(clause)
        case clause: ir.OldVariableDeclaration => visitOldVariableDeclaration(clause)
        case clause: ir.Precondition => visitPrecondition(clause)
        case clause: ir.Postcondition => visitPostcondition(clause)
        case clause: ir.DivergeClause => visitDivergeClause(clause)
        case clause: ir.MeasuredByClause => visitMeasuredByClause(clause)
        case clause: ir.ExceptionalPostcondition => visitExceptionalPostcondition(clause)
        case clause: ir.SignalsOnlyClause => visitSignalsOnlyClause(clause)
        case clause: ir.NonTranslatedContractClause => visitNonTranslatedContractClause(clause)
        case clause: ir.SpecCase => visitSpecCase(clause)
        case _ => ???
      }
      res.startSpec = node.startSpec.map(visitStringNode).map(transformStartSpec)
      res.endSpec = node.endSpec.map(visitStringNode)
      res
    catch {
      case e: NotSupportedException =>
        val text = Helper.commentUnknownNode(node, s"${e.construct} not supported, whole ContractClause commented out")
        jml.NonTranslatedContractClause(text, node.srcLocation)
    }
  }

  def visitAssignableClause(node: ir.AssignableClause): jml.AssignableClause = {
    node match {
      case ir.AssignableClause(expressions, srcLocation, precedingWhitespace) =>
        val res = jml.AssignableClause(expressions.map(visitExpression),
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitLoopModifiesClause(node: ir.LoopModifiesClause): jml.LoopModifiesClause = {
    node match {
      case ir.LoopModifiesClause(expressions, srcLocation, precedingWhitespace) =>
        val res = jml.LoopModifiesClause(expressions.map(visitExpression),
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitAccessibleClause(node: ir.AccessibleClause): jml.AccessibleClause = {
    node match {
      case ir.AccessibleClause(expressions, srcLocation, precedingWhitespace) =>
        val res = jml.AccessibleClause(expressions.map(visitExpression),
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitForallVariableDeclaration(node: ir.ForallVariableDeclaration): jml.ForallVariableDeclaration = {
    node match {
      case ir.ForallVariableDeclaration(declaration, srcLocation, precedingWhitespace) =>
        val res = jml.ForallVariableDeclaration(visitQuantifiedVarDeclaration(declaration),
                                                srcLocation,
                                                precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitOldVariableDeclaration(node: ir.OldVariableDeclaration): jml.OldVariableDeclaration = {
    node match {
      case ir.OldVariableDeclaration(modifiers, typeIndication, declarators, srcLocation, precedingWhitespace) =>
        val res = jml.OldVariableDeclaration(modifiers.map(visitModifier),
                                            visitTypeIndication(typeIndication),
                                            declarators.map(visitVariableDeclarator),
                                            srcLocation,
                                            precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitNonTranslatedContractClause(node: ir.NonTranslatedContractClause): jml.NonTranslatedContractClause = {
    node match {
      case ir.NonTranslatedContractClause(text, srcLocation, precedingWhitespace) =>
        val res = jml.NonTranslatedContractClause(text,
          srcLocation,
          precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitNonTranslatedLoopClause(node: ir.NonTranslatedLoopClause): jml.NonTranslatedLoopClause = {
    node match {
      case ir.NonTranslatedLoopClause(text, srcLocation, precedingWhitespace) =>
        val res = jml.NonTranslatedLoopClause(text,
                                              srcLocation,
                                              precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitPrecondition(node: ir.Precondition): jml.ContractClause = {
    node match {
      case ir.Precondition(expression, srcLocation, precedingWhitespace) =>
        try
          val res = jml.Precondition(visitExpression(expression),
                                    srcLocation,
                                    precedingWhitespace)
          res.copyWhitespace(node)
          res
        catch {
          case p: PermissionExpressionException =>
            collectedPermissions.append(p.expr)
            val err_text = "cannot express permission amount exactly. commented out"
            val text = Helper.commentNode(node, node.requiresKeyword, err_text)
            val res = jml.NonTranslatedContractClause(text, node.srcLocation)
            res.startSpec = node.startSpec.map(visitStringNode).map(transformStartSpec)
            res
          case p: MultiplePermissionsException =>
            val err_text = "cannot express permission amount exactly. commented out"
            val text = Helper.commentNode(node, node.requiresKeyword, err_text)
            val res = jml.NonTranslatedContractClause(text, node.srcLocation)
            res.startSpec = node.startSpec.map(visitStringNode).map(transformStartSpec)
            res
        }
      }
  }

  def visitPostcondition(node: ir.Postcondition): jml.ContractClause = {
    node match {
      case ir.Postcondition(expression, srcLocation, precedingWhitespace) =>
        val collectedBefore = collectedPermissions.toList
        try
          val res = jml.Postcondition(visitExpression(expression),
                                      srcLocation,
                                      precedingWhitespace)
          res.copyWhitespace(node)
          collectedPermissions = ListBuffer.from(collectedBefore)
          res
        catch {
        case p: PermissionExpressionException =>
          val err_text = "cannot express permissions in post-conditions. commented out"
          val text = Helper.commentNode(node, node.ensuresKeyword, err_text)
          collectedPermissions = ListBuffer.from(collectedBefore)
          jml.NonTranslatedContractClause(text, node.srcLocation)
        case p: MultiplePermissionsException =>
          val err_text = "cannot express permissions in post-conditions. commented out"
          val text = Helper.commentNode(node, node.ensuresKeyword, err_text)
          collectedPermissions = ListBuffer.from(collectedBefore)
          jml.NonTranslatedContractClause(text, node.srcLocation)
        }

    }
  }

  def visitDivergeClause(node: ir.DivergeClause): jml.DivergeClause = {
    node match {
      case ir.DivergeClause(expression, srcLocation, precedingWhitespace) =>
        val res = jml.DivergeClause(visitExpression(expression),
                                    srcLocation,
                                    precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitLoopClause(node: ir.LoopClause): jml.LoopClause = {
    try
      node match {
        case clause: ir.LoopModifiesClause => visitLoopModifiesClause(clause)
        case clause: ir.LoopInvariant => visitLoopInvariant(clause)
        case clause: ir.DecreasesClause => visitDecreasesClause(clause)
        case clause: ir.NonTranslatedLoopClause => visitNonTranslatedLoopClause(clause)
        case _ => ???
      }
    catch {
      case e: NotSupportedException =>
        val text = Helper.commentUnknownNode(node, s"${e.construct} not supported, whole LoopClause commented out")
        jml.NonTranslatedLoopClause(text, node.srcLocation)
    }
  }

  def visitLoopInvariant(node: ir.LoopInvariant): jml.LoopClause = {
    try
      node match {
        case ir.LoopInvariant(expression, srcLocation, precedingWhitespace) =>
          val res = jml.LoopInvariant(visitExpression(expression),
                                      srcLocation,
                                      precedingWhitespace)
          res.copyWhitespace(node)
          res
      }
    catch {
      case p: PermissionExpressionException =>
        val err_text = "cannot express permission amount exactly. commented out"
        collectedPermissions.append(p.expr)
        val text = Helper.commentNode(node, node.invariantKeyword, err_text)
        val res = jml.NonTranslatedLoopClause(text, node.srcLocation)
        res
      case p: MultiplePermissionsException =>
        val err_text = "cannot express permission amount exactly. commented out"
        val text = Helper.commentNode(node, node.invariantKeyword, err_text)
        val res = jml.NonTranslatedLoopClause(text, node.srcLocation)
        res
    }

  }

  def visitDecreasesClause(node: ir.DecreasesClause): jml.DecreasesClause = {
    node match {
      case ir.DecreasesClause(expression, srcLocation, precedingWhitespace) =>
        val res = jml.DecreasesClause(visitExpression(expression),
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitMeasuredByClause(node: ir.MeasuredByClause): jml.MeasuredByClause = {
    node match {
      case ir.MeasuredByClause(expression, predicate, srcLocation, precedingWhitespace) =>
        val irPred = predicate.map(visitExpression)
        val res = jml.MeasuredByClause(visitExpression(expression),
                                      irPred,
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitExceptionalPostcondition(node: ir.ExceptionalPostcondition): jml.ExceptionalPostcondition = {
    node match {
      case ir.ExceptionalPostcondition(exceptionType, exceptionName, expression, srcLocation, precedingWhitespace) =>
        val res = jml.ExceptionalPostcondition(visitTypeIndication(exceptionType),
                                              exceptionName.map(visitStringNode),
                                              expression.map(visitExpression),
                                              srcLocation,
                                              precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitSignalsOnlyClause(node: ir.SignalsOnlyClause): jml.SignalsOnlyClause = {
    node match {
      case ir.SignalsOnlyClause(exceptionTypes, srcLocation, precedingWhitespace) =>
        val res = jml.SignalsOnlyClause(exceptionTypes.map(visitTypeIndication),
                                        srcLocation,
                                        precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitBehavior(node: ir.Behavior): jml.Behavior = {
    node match {
      case ir.Behavior(startSpec, hasAlso, privacy, clauses, endSpec, srcLocation, precedingWhitespace) =>
        val irStart = startSpec.map(visitStringNode).map(transformStartSpec)
        val irPrivacy = privacy.map(visitStringNode)
        val irEnd = endSpec.map(visitStringNode)
        val res = jml.Behavior(irStart,
                              hasAlso,
                              irPrivacy,
                              clauses.map(visitContractClause),
                              irEnd,
                              srcLocation,
                              precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitNormalBehavior(node: ir.NormalBehavior): jml.NormalBehavior = {
    node match {
      case ir.NormalBehavior(startSpec, hasAlso, privacy, clauses, endSpec, srcLocation, precedingWhitespace) =>
        val irStart = startSpec.map(visitStringNode).map(transformStartSpec)
        val irPrivacy = privacy.map(visitStringNode)
        val irEnd = endSpec.map(visitStringNode)
        val res = jml.NormalBehavior(irStart,
                                    hasAlso,
                                    irPrivacy,
                                    clauses.map(visitContractClause),
                                    irEnd,
                                    srcLocation,
                                    precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitExceptionalBehavior(node: ir.ExceptionalBehavior): jml.ExceptionalBehavior = {
    node match {
      case ir.ExceptionalBehavior(startSpec, hasAlso, privacy, clauses, endSpec, srcLocation, precedingWhitespace) =>
        val irStart = startSpec.map(visitStringNode).map(transformStartSpec)
        val irPrivacy = privacy.map(visitStringNode)
        val irEnd = endSpec.map(visitStringNode)
        val res = jml.ExceptionalBehavior(irStart,
                                          hasAlso,
                                          irPrivacy,
                                          clauses.map(visitContractClause),
                                          irEnd,
                                          srcLocation,
                                          precedingWhitespace)
        res.copyWhitespace(node)
        res
      }
    }

  def visitAssertStatement(node: ir.AssertStatement): jml.AssertStatement = {
    node match {
      case ir.AssertStatement(content, message, srcLocation, precedingWhitespace) =>
        val irMsg = message.map(visitExpression)
        val res = jml.AssertStatement(visitExpression(content),
                                      irMsg,
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitAssumeStatement(node: ir.AssumeStatement): jml.AssumeStatement = {
    node match {
      case ir.AssumeStatement(content, message, srcLocation, precedingWhitespace) =>
        val irMsg = message.map(visitExpression)
        val res = jml.AssumeStatement(visitExpression(content),
                                      irMsg,
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitUnreachableStatement(node: ir.UnreachableStatement): jml.UnreachableStatement = {
    node match {
      case ir.UnreachableStatement(srcLocation, precedingWhitespace) =>
        val res = jml.UnreachableStatement(srcLocation, precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitSetStatement(node: ir.SetStatement): jml.SetStatement = {
    node match {
      case ir.SetStatement(assignment, srcLocation, precedingWhitespace) =>
        val res = jml.SetStatement(visitExpression(assignment),
                                  srcLocation,
                                  precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitGhostStatement(node: ir.GhostStatement): jml.Statement = {
    node match {
      case ir.GhostStatement(statement, srcLocation, precedingWhitespace) =>
        try
          statement match {
            case local: ir.LocalDeclaration =>
              val res = visitLocalDeclaration(local)
              val ghost = visitGhostModifier(node.ghostKeyword)
              res.modifiers = res.modifiers.prepended(ghost)
              res
            case expr: ir.ExpressionStatement =>
              expr.expression match {
                case ir.BinaryExpression(left, operator, right, bSrcLocation, bPrecedingWhitespace) =>
                  if(List("=", "+=", "-=", "*=", "/=", "&=", "|=", "^=", ">>=", ">>>=", "<<=", "%=").contains(operator.text)) {
                    val res = jml.SetStatement(visitExpression(expr.expression), srcLocation)
                    res.copyWhitespace(expr)
                    res.setKeyword.copyWhitespace(node.ghostKeyword)
                    res.precedingWhitespace = node.precedingWhitespace + res.precedingWhitespace
                    res.trailingWhitespace += node.trailingWhitespace
                    res
                  } else {
                    throw new NotSupportedException("general ghost statement", srcLocation)
                  }
                case _ =>
                  throw new NotSupportedException("general ghost statement", srcLocation)
              }
            case _ =>
              throw new NotSupportedException("general ghost statement", srcLocation)
          }
        catch {
          case e: NotSupportedException =>
            val text = Helper.commentNode(node, node.ghostKeyword, e.construct + " not supported, commented out.")
            System.err.println(e.getMessage)
            val res = jml.NonTranslatedStatement(text, srcLocation)
            res.copyWhitespace(node)
            res
        }
    }
  }

  def visitJmlStatementBlock(node: ir.JmlStatementBlock): jml.JmlStatementBlock = {
    node match {
      case ir.JmlStatementBlock(startSpec, statements, endSpec, srcLocation, precedingWhitespace) =>
        val irStart = startSpec.map(visitStringNode).map(transformStartSpec)
        val irEnd = endSpec.map(visitStringNode)
        val oldWrapInSpec = wrapInSpec
        wrapInSpec = false
        val res = jml.JmlStatementBlock(irStart,
                                        statements.map(visitStatement),
                                        irEnd,
                                        srcLocation,
                                        precedingWhitespace)
        res.copyWhitespace(node)
        wrapInSpec = oldWrapInSpec
        res
    }
  }

  def visitResultExpression(node: ir.ResultExpression): jml.ResultExpression = {
    node match {
      case ir.ResultExpression(src, precedingWhitespace) =>
        val res = jml.ResultExpression(src, precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitOldExpression(node: ir.OldExpression): jml.OldExpression = {
    node match {
      case ir.OldExpression(expression, label, srcLocation, precedingWhitespace) =>
        val irLabel = label.map(visitStringNode)
        val res = jml.OldExpression(visitExpression(expression),
                                    irLabel,
                                    srcLocation,
                                    precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitPreExpression(node: ir.PreExpression): jml.PreExpression = {
    node match {
      case ir.PreExpression(expression, srcLocation, precedingWhitespace) =>
        val res = jml.PreExpression(visitExpression(expression),
                                    srcLocation,
                                    precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitNotAssignedExpression(node: ir.NotAssignedExpression): jml.NotAssignedExpression = {
    node match {
      case ir.NotAssignedExpression(storeRefs, srcLocation, precedingWhitespace) =>
        val res = jml.NotAssignedExpression(storeRefs.map(visitExpression),
                                            srcLocation,
                                            precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitNotModifiedExpression(node: ir.NotModifiedExpression): jml.NotModifiedExpression = {
    node match {
      case ir.NotModifiedExpression(storeRefs, srcLocation, precedingWhitespace) =>
        val res = jml.NotModifiedExpression(storeRefs.map(visitExpression),
                                            srcLocation,
                                            precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitOnlyAccessedExpression(node: ir.OnlyAccessedExpression): jml.OnlyAccessedExpression = {
    node match {
      case ir.OnlyAccessedExpression(storeRefs, srcLocation, precedingWhitespace) =>
        val res = jml.OnlyAccessedExpression(storeRefs.map(visitExpression),
                                            srcLocation,
                                            precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitOnlyAssignedExpression(node: ir.OnlyAssignedExpression): jml.OnlyAssignedExpression = {
    node match {
      case ir.OnlyAssignedExpression(storeRefs, srcLocation, precedingWhitespace) =>
        val res = jml.OnlyAssignedExpression(storeRefs.map(visitExpression),
                                            srcLocation,
                                            precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitFreshExpression(node: ir.FreshExpression): jml.FreshExpression = {
    node match {
      case ir.FreshExpression(expressions, srcLocation, precedingWhitespace) =>
        val res = jml.FreshExpression(expressions.map(visitExpression),
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitNonNullElementsExpression(node: ir.NonNullElementsExpression): jml.NonNullElementsExpression = {
    node match {
      case ir.NonNullElementsExpression(expression, srcLocation, precedingWhitespace) =>
        val res = jml.NonNullElementsExpression(visitExpression(expression),
                                                srcLocation,
                                                precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitTypeOfExpression(node: ir.TypeOfExpression): jml.TypeOfExpression = {
    node match {
      case ir.TypeOfExpression(expression, srcLocation, precedingWhitespace) =>
        val res = jml.TypeOfExpression(visitExpression(expression),
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitTypeExpression(node: ir.TypeExpression): jml.TypeExpression = {
    node match {
      case ir.TypeExpression(indication, srcLocation, precedingWhitespace) =>
        val res = jml.TypeExpression(visitTypeIndication(indication),
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }


  def visitForallExpression(node: ir.ForallExpression): jml.ForallExpression = {
    node match {
      case ir.ForallExpression(declarations, guard, expression, srcLocation, precedingWhitespace) =>
        def nestedQuantifier(declarations: List[ir.QuantifiedVarDeclaration]): jml.Expression = {
          if(declarations.isEmpty) {
            visitExpression(expression)
          } else {
            jml.ForallExpression(visitQuantifiedVarDeclaration(declarations.head),
                                None,
                                nestedQuantifier(declarations.tail),
                                srcLocation,
                                " ")
          }
        }

        val res = jml.ForallExpression(visitQuantifiedVarDeclaration(declarations.head),
                                      guard.map(visitExpression),
                                      nestedQuantifier(declarations.tail),
                                      srcLocation,
                                      " ")
        res.copyWhitespace(node)
        res
    }
  }
  def visitExistsExpression(node: ir.ExistsExpression): jml.ExistsExpression = {
    node match {
      case ir.ExistsExpression(declarations, guard, expression, srcLocation, precedingWhitespace) =>
        def nestedQuantifier(declarations: List[ir.QuantifiedVarDeclaration]): jml.Expression = {
          if(declarations.isEmpty) {
            visitExpression(expression)
          } else {
            jml.ExistsExpression(visitQuantifiedVarDeclaration(declarations.head),
              None,
              nestedQuantifier(declarations.tail),
              srcLocation,
              " ")
          }
        }

        val res = jml.ExistsExpression(visitQuantifiedVarDeclaration(declarations.head),
                                      guard.map(visitExpression),
                                      nestedQuantifier(declarations.tail),
                                      srcLocation,
                                      " ")
        res.copyWhitespace(node)
        res
    }
  }
  def visitMaxQuantifiedExpression(node: ir.MaxQuantifiedExpression): jml.MaxQuantifiedExpression = {
    node match {
      case ir.MaxQuantifiedExpression(declarations, guard, expression, srcLocation, precedingWhitespace) =>
        def nestedQuantifier(declarations: List[ir.QuantifiedVarDeclaration]): jml.Expression = {
          if(declarations.isEmpty) {
            visitExpression(expression)
          } else {
            jml.MaxQuantifiedExpression(visitQuantifiedVarDeclaration(declarations.head),
                                        None,
                                        nestedQuantifier(declarations.tail),
                                        srcLocation,
                                        " ")
          }
        }

        val res = jml.MaxQuantifiedExpression(visitQuantifiedVarDeclaration(declarations.head),
                                              guard.map(visitExpression),
                                              nestedQuantifier(declarations.tail),
                                              srcLocation,
                                              " ")
        res.copyWhitespace(node)
        res
    }
  }
  def visitMinExpression(node: ir.MinExpression): jml.MinExpression = {
    node match {
      case ir.MinExpression(declarations, guard, expression, srcLocation, precedingWhitespace) =>
        def nestedQuantifier(declarations: List[ir.QuantifiedVarDeclaration]): jml.Expression = {
          if(declarations.isEmpty) {
            visitExpression(expression)
          } else {
            jml.MinExpression(visitQuantifiedVarDeclaration(declarations.head),
              None,
              nestedQuantifier(declarations.tail),
              srcLocation,
              " ")
          }
        }

        val res = jml.MinExpression(visitQuantifiedVarDeclaration(declarations.head),
          guard.map(visitExpression),
          nestedQuantifier(declarations.tail),
          srcLocation,
          " ")
        res.copyWhitespace(node)
        res
    }
  }
  def visitNumOfExpression(node: ir.NumOfExpression): jml.NumOfExpression = {
    node match {
      case ir.NumOfExpression(declarations, guard, expression, srcLocation, precedingWhitespace) =>
        def nestedQuantifier(declarations: List[ir.QuantifiedVarDeclaration]): jml.Expression = {
          if(declarations.isEmpty) {
            visitExpression(expression)
          } else {
            jml.NumOfExpression(visitQuantifiedVarDeclaration(declarations.head),
              None,
              nestedQuantifier(declarations.tail),
              srcLocation,
              " ")
          }
        }

        val res = jml.NumOfExpression(visitQuantifiedVarDeclaration(declarations.head),
          guard.map(visitExpression),
          nestedQuantifier(declarations.tail),
          srcLocation,
          " ")
        res.copyWhitespace(node)
        res
    }
  }
  def visitProductExpression(node: ir.ProductExpression): jml.ProductExpression = {
    node match {
      case ir.ProductExpression(declarations, guard, expression, srcLocation, precedingWhitespace) =>
        def nestedQuantifier(declarations: List[ir.QuantifiedVarDeclaration]): jml.Expression = {
          if(declarations.isEmpty) {
            visitExpression(expression)
          } else {
            jml.ProductExpression(visitQuantifiedVarDeclaration(declarations.head),
              None,
              nestedQuantifier(declarations.tail),
              srcLocation,
              " ")
          }
        }

        val res = jml.ProductExpression(visitQuantifiedVarDeclaration(declarations.head),
          guard.map(visitExpression),
          nestedQuantifier(declarations.tail),
          srcLocation,
          " ")
        res.copyWhitespace(node)
        res
    }
  }
  def visitSumExpression(node: ir.SumExpression): jml.SumExpression = {
    node match {
      case ir.SumExpression(declarations, guard, expression, srcLocation, precedingWhitespace) =>
        def nestedQuantifier(declarations: List[ir.QuantifiedVarDeclaration]): jml.Expression = {
          if(declarations.isEmpty) {
            visitExpression(expression)
          } else {
            jml.SumExpression(visitQuantifiedVarDeclaration(declarations.head),
              None,
              nestedQuantifier(declarations.tail),
              srcLocation,
              " ")
          }
        }

        val res = jml.SumExpression(visitQuantifiedVarDeclaration(declarations.head),
          guard.map(visitExpression),
          nestedQuantifier(declarations.tail),
          srcLocation,
          " ")
        res.copyWhitespace(node)
        res
    }
  }


  def visitQuantifiedVarDeclaration(node: ir.QuantifiedVarDeclaration): jml.QuantifiedVarDeclaration = {
    node match {
      case decl: ir.NonNullQuantifiedVarDeclaration =>
        visitNonNullQuantifiedVarDeclaration(decl)
      case decl: ir.NullableQuantifiedVarDeclaration =>
        visitNullableQuantifiedVarDeclaration(decl)
      case ir.QuantifiedVarDeclaration(declType, declarators, srcLocation, precedingWhitespace) =>
        val res = jml.QuantifiedVarDeclaration(visitTypeIndication(declType),
                                              declarators.map(visitStringNode),
                                              srcLocation,
                                              precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitNonNullQuantifiedVarDeclaration(node: ir.NonNullQuantifiedVarDeclaration): jml.NonNullQuantifiedVarDeclaration = {
    node match {
      case ir.NonNullQuantifiedVarDeclaration(declType, declarators, srcLocation, precedingWhitespace) =>
        val res = jml.NonNullQuantifiedVarDeclaration(visitTypeIndication(declType),
                                                      declarators.map(visitStringNode),
                                                      srcLocation,
                                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitNullableQuantifiedVarDeclaration(node: ir.NullableQuantifiedVarDeclaration): jml.NullableQuantifiedVarDeclaration = {
    node match {
      case ir.NullableQuantifiedVarDeclaration(declType, declarators, srcLocation, precedingWhitespace) =>
        val res = jml.NullableQuantifiedVarDeclaration(visitTypeIndication(declType),
                                                      declarators.map(visitStringNode),
                                                      srcLocation,
                                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitSetComprehension(node: ir.SetComprehension): jml.SetComprehension = {
    node match {
      case ir.SetComprehension(typeIndication, declaration, containsExpression, filterExpression, srcLocation, precedingWhitespace) =>
        val res = jml.SetComprehension(visitTypeIndication(typeIndication),
                                      visitQuantifiedVarDeclaration(declaration),
                                      visitExpression(containsExpression),
                                      visitExpression(filterExpression),
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitUnfoldingExpression(node: ir.UnfoldingExpression): jml.Expression = {
    node match {
      case ir.UnfoldingExpression(predicate, expression, srcLocation, precedingWhitespace) =>
        val res = visitExpression(expression)
        res.precedingWhitespace = node.precedingWhitespace
                                + node.unfoldingKeyword.precedingWhitespace + node.unfoldingKeyword.trailingWhitespace
                                + node.predicate.precedingWhitespace + node.predicate.trailingWhitespace
                                + node.inKeyword.precedingWhitespace + node.inKeyword.trailingWhitespace
                                + res.precedingWhitespace
        System.err.println(s"[Warning] Unfolding expression '${node.toString}' (${node.srcLocation}) simplified to inner expression only.")
        res
    }
  }

  def visitValuedPermissionExpression(node: ir.ValuedPermissionExpression): jml.Expression = {
    node match {
      case ir.ValuedPermissionExpression(target, value, amount, srcLocation, precedingWhitespace) =>
        collectedPermissions.append(node)
        val res = jml.BinaryExpression(visitExpression(target), jml.StringNode("==", srcLocation),
                                      visitExpression(value), srcLocation)
        res.copyWhitespace(node)
        res
    }
  }

  def visitForallWithPermissionsExpression(node: ir.ForallWithPermissionsExpression): jml.Expression = {
    node match {
      case ir.ForallWithPermissionsExpression(declaration, guard, expression, srcLocation, precedingWhitespace) =>
        val oldCollected = collectedPermissions.toList
        try
          def nestedQuantifier(declarations: List[ir.QuantifiedVarDeclaration]): jml.Expression = {
            if(declarations.isEmpty) {
              visitExpression(expression)
            } else {
              jml.ForallExpression(visitQuantifiedVarDeclaration(declarations.head),
                                    None,
                                    nestedQuantifier(declarations.tail),
                                    srcLocation,
                                    " ")
            }
          }

          val res = jml.ForallExpression(visitQuantifiedVarDeclaration(declaration.head),
                                        guard.map(visitExpression),
                                        nestedQuantifier(declaration.tail),
                                        srcLocation,
                                        " ")
          res.copyWhitespace(node)
          val newCollected = collectedPermissions.toList
          collectedPermissions = ListBuffer.from(oldCollected)
          val tooComplex: ListBuffer[ir.PermissionExpression] = ListBuffer()
          for(perm <- newCollected.diff(oldCollected)) {
            tooComplex.appendAll(rewriteArrayPermission(srcLocation, perm))
          }
          if(tooComplex.nonEmpty) {
            System.err.println(s"The permission expressions ${tooComplex.map(x => "\"" + x.toString + "\"").mkString(", ")} "
                               + s"cannot be used for modifiable clauses and are dropped${srcLocation.toString}")
          }
          res
        catch {
          case e: PermissionExpressionException =>
            collectedPermissions = ListBuffer.from(oldCollected)
            val tooComplex = rewriteArrayPermission(srcLocation, e.expr)
            if(tooComplex.nonEmpty) {
              System.err.println(s"The permission expressions ${tooComplex.map(x => "\"" + x.toString + "\"").mkString(", ")} "
                                + s"cannot be used for modifiable clauses and are dropped${srcLocation.toString}")
            }
            throw new MultiplePermissionsException()
          case e: MultiplePermissionsException =>
            val newCollected = collectedPermissions.toList
            collectedPermissions = ListBuffer.from(oldCollected)
            val tooComplex: ListBuffer[ir.PermissionExpression] = ListBuffer()
            for(perm <- newCollected.diff(oldCollected)) {
              tooComplex.appendAll(rewriteArrayPermission(srcLocation, perm))
            }
            if(tooComplex.nonEmpty) {
              System.err.println(s"The permission expressions ${tooComplex.map(x => "\"" + x.toString + "\"").mkString(", ")} "
                                + s"cannot be used for modifiable clauses and are dropped${srcLocation.toString}")
            }
            throw new MultiplePermissionsException()
        }
    }
  }

  private def rewriteArrayPermission(srcLocation: SourceLocation, perm: ir.PermissionExpression): List[ir.PermissionExpression] = {
    val tooComplex: ListBuffer[ir.PermissionExpression] = ListBuffer()
    perm.target match {
      case arr: ir.ArrayIndexExpression =>
        def checkArrayExpression(expr: ir.Expression): Boolean = {
          expr match {
            case _: ir.ResultExpression => true
            case _: ir.StringNode => true
            case _: ir.Literal => true
            case p: ir.ParExpression => checkArrayExpression(p.expression)
            case n: ir.NestedInvocationExpression =>
              checkArrayExpression(n.left) && checkArrayExpression(n.right)
            case _ => false
          }
        }

        if (checkArrayExpression(arr.arrayExpression)) {
          val length = ir.NestedInvocationExpression(arr.arrayExpression,
                                                    ir.StringNode("length", srcLocation),
                                                    srcLocation)
          val range = ir.RangeExpression(ir.StringNode("0", srcLocation),
            length,
            srcLocation)
          val newArr = ir.ArrayIndexExpression(arr.arrayExpression,
                                                range,
                                                srcLocation, " ")
          collectedPermissions.append(new ir.PermissionExpression(newArr, perm.amount, perm.srcLocation))
        } else {
          tooComplex.append(perm)
        }
      case _ =>
        tooComplex.append(perm)
    }
    tooComplex.toList
  }

  def visitScaledPredicateExpression(node: ir.ScaledPredicateExpression): jml.Expression = {
    node match {
      case ir.ScaledPredicateExpression(amount, predicate, srcLocation, precedingWhitespace) =>
        val res = visitExpression(predicate)
        res.copyWhitespace(predicate)
        res.precedingWhitespace = node.precedingWhitespace
                    + node.openBracket.precedingWhitespace + node.openBracket.trailingWhitespace
                    + amount.precedingWhitespace + amount.trailingWhitespace
                    + node.closeBracket.precedingWhitespace + node.closeBracket.trailingWhitespace
                    + res.precedingWhitespace
        System.err.println(s"[Warning] Expression with permission scaling(${node.srcLocation}) not supported in "
                          + s"OpenJML, scaling amount dropped.")
        res
    }
  }

  def visitSpecDeclarationBlockOnTypeLevel(node: ir.SpecDeclarationBlock): jml.SpecDeclarationBlock = {
    val start = node.startSpec.map(visitStringNode).map(transformStartSpec)
    val decls = node.declarations.map(visitSpecDeclaration)
    val end = node.endSpec.map(visitStringNode)
    jml.SpecDeclarationBlock(start, decls, end, node.srcLocation, node.precedingWhitespace)
  }

  def visitSpecDeclarationBlock(node: ir.SpecDeclarationBlock): jml.SpecDeclarationBlock = {
    node match {
      case ir.SpecDeclarationBlock(startSpec, declarations, endSpec, srcLocation, precedingWhitespace) =>
        val oldWrapInSpec = wrapInSpec
        wrapInSpec = false
        val res = jml.SpecDeclarationBlock(startSpec.map(visitStringNode).map(transformStartSpec),
                                          declarations.map(visitSpecDeclaration),
                                          endSpec.map(visitStringNode),
                                          srcLocation,
                                          precedingWhitespace)
        res.copyWhitespace(node)
        wrapInSpec = oldWrapInSpec
        res
    }
  }

  def visitSpecDeclaration(node: ir.SpecDeclaration): jml.SpecDeclaration = {
    try
      node match {
        case decl: ir.InvariantDeclaration => visitInvariantDeclaration(decl)
        case decl: ir.InitiallyDeclaration => visitInitiallyDeclaration(decl)
        case decl: ir.ReadableDeclaration => visitReadableDeclaration(decl)
        case decl: ir.WritableDeclaration => visitWritableDeclaration(decl)
        case decl: ir.MonitorsForDeclaration => visitMonitorsForDeclaration(decl)
        case decl: ir.AxiomDeclaration => visitAxiomDeclaration(decl)
        case decl: ir.SpecMethodDeclaration => visitSpecMethodDeclaration(decl)
        case decl: ir.GhostDeclaration => visitGhostDeclaration(decl)
        case decl: ir.ModelDeclaration => visitModelDeclaration(decl)
        case decl: ir.NonTranslatedSpecDeclaration => visitNonTranslatedSpecDeclaration(decl)
        case decl: ir.PredicateDeclaration => visitPredicateDeclaration(decl)
        case d =>
          println(s"Missing implementation for SpecDeclaration $d")
          null
      }
    catch {
      case e: NotSupportedException =>
        val text = Helper.commentUnknownNode(node, s"${e.construct} not supported, whole SpecDeclaration commented out")
        jml.NonTranslatedSpecDeclaration(text, node.srcLocation)
    }
  }

  def visitPredicateDeclaration(node: ir.PredicateDeclaration): jml.SpecDeclaration = {
    node match {
      case ir.PredicateDeclaration(name, tType, params, expr, srcLocation, precedingWhitespace) =>
        val ttype = jml.StringTypeIndication("boolean", Nil, "boolean", tType.srcLocation)
        ttype.copyWhitespace(tType)
        val body = jml.EmptyStatement(srcLocation)
        val model = jml.ModelModifier(srcLocation, tType.precedingWhitespace)
        val pure = jml.PureModifier(srcLocation, " ")
        ttype.precedingWhitespace = " "
        val res = jml.SpecMethodDeclaration(visitStringNode(name), Option(ttype), params.map(visitParameter), body, srcLocation)
        res.modifiers = node.modifiers.map(visitModifier).prepended(pure).prepended(model)
        res.contractBefore = node.contractBefore.map(visitContract)
        res.copyWhitespace(node)
        if (expr.nonEmpty) {
          try
            // turn predicate body into postcondition
            val result = jml.ParExpression(visitExpression(expr.get), srcLocation, " ")
            val postExpr = jml.BinaryExpression(jml.ResultExpression(srcLocation, " "),
              jml.StringNode("==", srcLocation, " "),
              result,
              srcLocation)
            val clause = jml.Postcondition(postExpr, srcLocation)
            clause.precedingWhitespace = node.precedingWhitespace
            clause.ensuresKeyword.precedingWhitespace = node.tType.precedingWhitespace
            val contract = jml.Contract(List(clause), srcLocation)
            res.contractBefore = Option(contract)
          catch {
            // if body only contains Permissions/predicates: no postcondition needed
            case e: MultiplePermissionsException =>
            case e: PermissionExpressionException =>
          }
        }
        res
    }
  }

  def visitNonTranslatedSpecDeclaration(decl: ir.NonTranslatedSpecDeclaration): jml.SpecDeclaration = {
    decl match {
      case ir.NonTranslatedSpecDeclaration(text, srcLocation, precedingWhitespace) =>
        jml.NonTranslatedSpecDeclaration(text, srcLocation, precedingWhitespace)
    }
  }
  def visitInvariantDeclaration(node: ir.InvariantDeclaration): jml.InvariantDeclaration = {
    node match {
      case ir.InvariantDeclaration(expression, srcLocation, precedingWhitespace) =>
        val res = jml.InvariantDeclaration(visitExpression(expression),
                                          srcLocation,
                                          precedingWhitespace)
        res.modifiers = node.modifiers.map(visitModifier)
        res.copyWhitespace(node)
        res
    }
  }
  def visitInitiallyDeclaration(node: ir.InitiallyDeclaration): jml.InitiallyDeclaration = {
    node match {
      case ir.InitiallyDeclaration(expression, srcLocation, precedingWhitespace) =>
        val res = jml.InitiallyDeclaration(visitExpression(expression),
                                          srcLocation,
                                          precedingWhitespace)
        res.modifiers = node.modifiers.map(visitModifier)
        res.copyWhitespace(node)
        res
    }
  }
  def visitReadableDeclaration(node: ir.ReadableDeclaration): jml.ReadableDeclaration = {
    node match {
      case ir.ReadableDeclaration(identifier, expression, srcLocation, precedingWhitespace) =>
        val res = jml.ReadableDeclaration(visitStringNode(identifier),
                                          visitExpression(expression),
                                          srcLocation,
                                          precedingWhitespace)
        res.modifiers = node.modifiers.map(visitModifier)
        res.copyWhitespace(node)
        res
    }
  }
  def visitWritableDeclaration(node: ir.WritableDeclaration): jml.WritableDeclaration = {
    node match {
      case ir.WritableDeclaration(identifier, expression, srcLocation, precedingWhitespace) =>
        val res = jml.WritableDeclaration(visitStringNode(identifier),
                                          visitExpression(expression),
                                          srcLocation,
                                          precedingWhitespace)
        res.modifiers = node.modifiers.map(visitModifier)
        res.copyWhitespace(node)
        res
    }
  }
  def visitMonitorsForDeclaration(node: ir.MonitorsForDeclaration): jml.MonitorsForDeclaration = {
    node match {
      case ir.MonitorsForDeclaration(identifier, expressions, srcLocation, precedingWhitespace) =>
        val res = jml.MonitorsForDeclaration(visitStringNode(identifier),
                                            expressions.map(visitExpression),
                                            srcLocation,
                                            precedingWhitespace)
        res.modifiers = node.modifiers.map(visitModifier)
        res.copyWhitespace(node)
        res
    }
  }
  def visitAxiomDeclaration(node: ir.AxiomDeclaration): jml.AxiomDeclaration = {
    node match {
      case ir.AxiomDeclaration(expression, srcLocation, precedingWhitespace) =>
        val res = jml.AxiomDeclaration(visitExpression(expression),
          srcLocation,
          precedingWhitespace)
        res.modifiers = node.modifiers.map(visitModifier)
        res.copyWhitespace(node)
        res
    }
  }

  def visitSpecMethodDeclaration(node: ir.SpecMethodDeclaration): jml.SpecMethodDeclaration = {
    node match {
      case ir.SpecMethodDeclaration(name, returnType, parameters, contractWithin,
                                    body, srcLocation, precedingWhitespace) =>
        var contractBefore: Option[jml.Contract] = None
        node.contractBefore match {
          case None =>
            contractBefore = contractWithin.map(visitContract)
          case Some(before) =>
            contractBefore = node.contractBefore.map(visitContract)
            if(contractWithin.isEmpty) {
              // merge contractBefore and contractWithin
              val within = visitContract(contractWithin.get)
              if(within.clauses.nonEmpty) {
                within.clauses.head.precedingWhitespace = before.endSpec.getOrElse("").toString 
                                                        + before.trailingWhitespace 
                                                        + within.precedingWhitespace 
                                                        + within.startSpec.getOrElse("").toString
                before.endSpec = within.endSpec
              } else {
                before.trailingWhitespace += within.toString
              }
              before.clauses = before.clauses.appendedAll(within.clauses)
            }
        }
        val irReturn = returnType.map(visitTypeIndication)
        val res = jml.SpecMethodDeclaration(visitStringNode(name),
                                            irReturn,
                                            parameters.map(visitParameter),
                                            visitBlockOrSemicolon(body),
                                            srcLocation,
                                            precedingWhitespace)
        res.contractBefore = node.contractBefore.map(visitContract)
        res.modifiers = node.modifiers.map(visitModifier)
        res.copyWhitespace(node)
        res
    }
  }

}
