package languages.OpenJML

import languages.OpenJML.syntaxTree.*
import org.antlr.v4.runtime.{CommonTokenStream, ParserRuleContext, Token}
import org.antlr.v4.runtime.tree.TerminalNode
import generated.openJML.{JavaParser, JavaParserBaseVisitor}
import generic.{Helper, NotSupportedException, SourceLocation}

import scala.collection.mutable.ListBuffer
import scala.jdk.CollectionConverters.*

class OpenJMLParserVisitor(val tokens: CommonTokenStream, val filename: String)
      extends JavaParserBaseVisitor[OpenJMLTreeNode] {

  val printer: ParserPrintVisitor = new ParserPrintVisitor(tokens)

  def getWhitespaceLeft(ctx: ParserRuleContext): String = {
    val hiddenTokens = tokens.getHiddenTokensToLeft(ctx.start.getTokenIndex)
    if (hiddenTokens == null) "" else hiddenTokens.asScala.map(wsToString).mkString
  }

  def wsToString(token: Token): String = {
    if (token.getType == JavaParser.ExtraAt) {
      token.getText.replace('@', ' ')
    } else {
      token.getText
    }
  }

  def getWhitespaceRight(ctx: ParserRuleContext): String = {
    val hiddenTokens = tokens.getHiddenTokensToRight(ctx.stop.getTokenIndex)
    if (hiddenTokens == null) "" else hiddenTokens.asScala.map(_.getText).mkString
  }

  def getWhitespaceLeft(node: TerminalNode): String = {
    val hiddenTokens = tokens.getHiddenTokensToLeft(node.getSymbol.getTokenIndex)
    if (hiddenTokens == null) "" else hiddenTokens.asScala.map(_.getText).mkString
  }

  def getWhitespaceRight(node: TerminalNode): String = {
    val hiddenTokens = tokens.getHiddenTokensToRight(node.getSymbol.getTokenIndex)
    if (hiddenTokens == null) "" else hiddenTokens.asScala.map(_.getText).mkString
  }

  def getWhitespaceLeft(token: Token): String = {
    val hiddenTokens = tokens.getHiddenTokensToLeft(token.getTokenIndex)
    if (hiddenTokens == null) "" else hiddenTokens.asScala.map(_.getText).mkString
  }

  def addWhitespaceRight[T <: OpenJMLTreeNode](node: T, ctx: ParserRuleContext): T = {
    node.trailingWhitespace = getWhitespaceRight(ctx)
    node
  }

  def addWhitespaceRight[T <: OpenJMLTreeNode](node: T, terminal: TerminalNode): T = {
    node.trailingWhitespace = getWhitespaceRight(terminal)
    node
  }

  def getSourceLocation(ctx: ParserRuleContext) : SourceLocation = {
    SourceLocation(filename, ctx.start.getLine, ctx.stop.getLine)
  }

  def getSourceLocation(node: TerminalNode) : SourceLocation = {
    SourceLocation(filename, node.getSymbol.getLine)
  }

  def getSourceLocation(token: Token) : SourceLocation = {
    SourceLocation(filename, token.getLine)
  }

  def visitTextNode(ctx: ParserRuleContext): StringNode = {
    val src = getSourceLocation(ctx)
    StringNode(ctx.getText, src, getWhitespaceLeft(ctx))
  }

  def visitTextNodeWithWS(ctx: ParserRuleContext): StringNode = {
    val src = getSourceLocation(ctx)
    StringNode(ctx.getText, src, getWhitespaceLeft(ctx), getWhitespaceRight(ctx))
  }

  override def visitJavaIdentifier(ctx: JavaParser.JavaIdentifierContext): StringNode = {
    visitTextNode(ctx)
  }
  def visitJavaIdentifierWithWS(ctx: JavaParser.JavaIdentifierContext): StringNode = {
    visitTextNodeWithWS(ctx)
  }

  def printNode(ctx: ParserRuleContext): StringNode = {
    val src = getSourceLocation(ctx)
    val text = printer.visitSkipFirstWS(ctx)
    StringNode(text, src, getWhitespaceLeft(ctx))
  }

  def printNodeWithWS(ctx: ParserRuleContext): StringNode = {
    val src = getSourceLocation(ctx)
    val text = printer.visitSkipFirstWS(ctx)
    StringNode(text, src, getWhitespaceLeft(ctx), getWhitespaceRight(ctx))
  }

  override def visitTypeType(ctx: JavaParser.TypeTypeContext): StringTypeIndication = {
    val text = printer.visitSkipFirstWS(ctx)
    new StringTypeIndication(text, Nil, ctx.getText, getSourceLocation(ctx), getWhitespaceLeft(ctx))
  }
  def visitTypeTypeWithWS(ctx: JavaParser.TypeTypeContext): StringTypeIndication = {
    val text = printer.visitSkipFirstWS(ctx)
    new StringTypeIndication(text, Nil, ctx.getText, getSourceLocation(ctx),
                             getWhitespaceLeft(ctx), getWhitespaceRight(ctx))
  }

  def visitTypeListAsList(ctx: JavaParser.TypeListContext): List[TypeIndication] = {
    val types = ctx.typeType().asScala.map(visitTypeTypeWithWS).toList
    types.last.trailingWhitespace = ""
    types
  }

  override def visitCompilationUnit(ctx: JavaParser.CompilationUnitContext): CompilationUnit = {
    val src = SourceLocation(filename, ctx.start.getLine, ctx.stop.getLine)
    var pkgDecl: Option[PackageDeclaration] = None
    pkgDecl = Option(ctx.packageDeclaration()).map(visitPackageDeclaration)
    val imports = ctx.importDeclaration().asScala.map(visitImportDeclaration).toList
    val declarations = ctx.typeDeclaration().asScala.map(visitTypeDeclaration).toList
    val res = CompilationUnit(pkgDecl, imports, declarations, src)
    res.trailingWhitespace = getWhitespaceLeft(ctx.EOF())
    res
  }

  override def visitPackageDeclaration(ctx: JavaParser.PackageDeclarationContext): PackageDeclaration = {
    val src = getSourceLocation(ctx)
    val name = visitQualifiedName(ctx.qualifiedName())
    val annotations = ctx.annotation().asScala.map(visitAnnotation).toList
    val pkg = PackageDeclaration(name, annotations, src)
    pkg.packageKeyword.precedingWhitespace = getWhitespaceLeft(ctx.PACKAGE())
    pkg.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
    pkg
  }

  override def visitImportDeclaration(ctx: JavaParser.ImportDeclarationContext): ImportDeclaration = {
    val name = visitQualifiedName(ctx.qualifiedName())
    if (ctx.DOT() != null) {
      name.text += getWhitespaceLeft(ctx.DOT()) + ctx.DOT().getText
      name.text += getWhitespaceLeft(ctx.MUL()) + ctx.MUL().getText
    }
    var modifiers: List[Modifier] = Nil
    if (ctx.STATIC() != null) {
      val static = StaticModifier(getSourceLocation(ctx.STATIC()), getWhitespaceLeft(ctx.STATIC()))
      modifiers = modifiers.appended(static)
    }
    val imp = ImportDeclaration(name, getSourceLocation(ctx))
    imp.modifiers = modifiers
    imp.importKeyword.precedingWhitespace = getWhitespaceLeft(ctx.IMPORT())
    imp.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
    imp
  }

  override def visitTypeDeclaration(ctx: JavaParser.TypeDeclarationContext): TypeDeclaration  = {
    var res: TypeDeclaration = null
    if (ctx.SEMI() != null) {
      res = new EmptyStatement(getSourceLocation(ctx), getWhitespaceLeft(ctx.SEMI()))
    } else {
      val modifiers = ctx.classOrInterfaceModifier().asScala.map(visitClassOrInterfaceModifier).toList
      if (ctx.classDeclaration() != null) {
        val cls = visitClassDeclaration(ctx.classDeclaration())
        cls.modifiers = modifiers
        cls.srcLocation = getSourceLocation(ctx)
        res = cls
      } else if (ctx.enumDeclaration() != null) {
        val enumDecl = visitEnumDeclaration(ctx.enumDeclaration())
        enumDecl.modifiers = modifiers
        enumDecl.srcLocation = getSourceLocation(ctx)
        res = enumDecl
      } else if (ctx.interfaceDeclaration() != null) {
        val interface = visitInterfaceDeclaration(ctx.interfaceDeclaration())
        interface.modifiers = modifiers
        interface.srcLocation = getSourceLocation(ctx)
        res = interface
      } else if (ctx.annotationTypeDeclaration() != null) {
        val annotationType = visitAnnotationTypeDeclaration(ctx.annotationTypeDeclaration())
        annotationType.modifiers = modifiers
        annotationType.srcLocation = getSourceLocation(ctx)
        res = annotationType
      } else {
        println(s"Unknown type declaration: ${ctx.getText}")
      }

    }
    res
  }

  override def visitModifier(ctx: JavaParser.ModifierContext): Modifier = {
    ctx.altNum match {
      case 1 => visitClassOrInterfaceModifier(ctx.classOrInterfaceModifier())
      case _ =>
        new ModifierStringNode(ctx.getText, getSourceLocation(ctx), getWhitespaceLeft(ctx))
    }
  }

  override def visitClassOrInterfaceModifier(ctx: JavaParser.ClassOrInterfaceModifierContext): Modifier = {
    if (ctx.annotation() != null) {
      visitAnnotation(ctx.annotation())
    } else if(ctx.valEmbedModifiers() != null) {
      visitValEmbedModifiers(ctx.valEmbedModifiers())
    } else if(ctx.STATIC() != null) {
      StaticModifier(getSourceLocation(ctx), getWhitespaceLeft(ctx))
    } else {
      new ModifierStringNode(ctx.getText, getSourceLocation(ctx), getWhitespaceLeft(ctx))
    }
  }

  override def visitVariableModifier(ctx: JavaParser.VariableModifierContext): Modifier = {
    if (ctx.annotation() != null) {
      visitAnnotation(ctx.annotation())
    } else if (ctx.valEmbedModifiers() != null) {
      visitValEmbedModifiers(ctx.valEmbedModifiers())
    } else {
      new ModifierStringNode(ctx.getText, getSourceLocation(ctx), getWhitespaceLeft(ctx))
    }
  }

  override def visitClassDeclaration(ctx: JavaParser.ClassDeclarationContext): ClassDeclaration = {
    val name = visitJavaIdentifier(ctx.javaIdentifier())
    var sup: Option[SuperclassClause] = None
    if (ctx.EXTENDS() != null) {
      val ext = SuperclassClause(visitTypeType(ctx.typeType()), getSourceLocation(ctx.typeType()))
      ext.extendsKeyword.precedingWhitespace = getWhitespaceLeft(ctx.EXTENDS())
      sup = Option(ext)
    }
    var interfaces: Option[InterfaceClause] = None
    if (ctx.IMPLEMENTS() != null) {
      val int = InterfaceClause(visitTypeListAsList(ctx.typeList()), getSourceLocation(ctx.typeList()))
      int.implementsKeyword.precedingWhitespace = getWhitespaceLeft(ctx.IMPLEMENTS())
      interfaces = Option(int)
    }
    val typeParameterClause = if (ctx.typeParameters() != null)
                                Option(visitTypeParameters(ctx.typeParameters()))
                              else None
    val body = visitClassBody(ctx.classBody())
    val res = ClassDeclaration(name, typeParameterClause, sup, interfaces, body, getSourceLocation(ctx))
    res.classKeyword.precedingWhitespace = getWhitespaceLeft(ctx.CLASS())
    res
  }

  override def visitTypeParameters(ctx: JavaParser.TypeParametersContext): TypeParameterClause = {
    val parameters = ctx.typeParameter().asScala
                      .map(x=>addWhitespaceRight(visitTypeParameter(x), x)).toList
    val clause = TypeParameterClause(parameters, getSourceLocation(ctx))
    clause.openAngle.precedingWhitespace = getWhitespaceLeft(ctx.LT())
    clause.closeAngle.precedingWhitespace = getWhitespaceLeft(ctx.GT())
    // WS before closeAngle already covered by trailing WS of last typeParameter
    parameters.last.trailingWhitespace = ""
    clause
  }

  override def visitTypeParameter(ctx: JavaParser.TypeParameterContext): TypeParameter = {
    val name = visitJavaIdentifier(ctx.javaIdentifier())
    val typeName = new StringTypeIndication(name.text, Nil, ctx.javaIdentifier().getText, name.srcLocation)
    var annotations: List[StringAnnotation] = Nil
    var superclass: Option[GenericsSuperclassClause] = None
    if (ctx.EXTENDS() != null) {
      val sup = GenericsSuperclassClause(visitTypeBoundAsList(ctx.typeBound()),
                                         getSourceLocation(ctx.typeBound()))
      sup.extendsKeyword.precedingWhitespace = getWhitespaceLeft(ctx.EXTENDS())
      superclass = Option(sup)
      // split list of annotations depending on whether before "javaIdentifier" or before "typeBound"
      val typeAnnotations = ctx.annotation().asScala
                              .filter(_.start.getTokenIndex < ctx.EXTENDS().getSymbol.getTokenIndex)
      annotations = typeAnnotations.map(visitAnnotation).toList
      val boundAnnotations = ctx.annotation().asScala
                              .filter(_.start.getTokenIndex > ctx.EXTENDS().getSymbol.getTokenIndex)
      // add annotations before "typeBound" to first "typeType"
      sup.superclasses.head.annotations = sup.superclasses.head.annotations
                                            .prependedAll(boundAnnotations.map(visitAnnotation).toList)
    } else {
      annotations = ctx.annotation().asScala.map(visitAnnotation).toList
    }
    val param = TypeParameter(annotations, typeName, superclass, getSourceLocation(ctx))
    param
  }

  def visitTypeBoundAsList(ctx: JavaParser.TypeBoundContext): List[TypeIndication] = {
    val types = ctx.typeType().asScala.map(visitTypeTypeWithWS).toList
    types.last.trailingWhitespace = ""
    types
  }

  override def visitEnumDeclaration(ctx: JavaParser.EnumDeclarationContext): EnumDeclaration = {
    val name = visitJavaIdentifier(ctx.javaIdentifier())
    var interfaces: Option[InterfaceClause] = None
    if (ctx.IMPLEMENTS() != null) {
      val int = InterfaceClause(visitTypeListAsList(ctx.typeList()), getSourceLocation(ctx.typeList()))
      int.implementsKeyword.precedingWhitespace = getWhitespaceLeft(ctx.IMPLEMENTS())
      interfaces = Option(int)
    }
    var constants: List[EnumConstant] = Nil
    if (ctx.enumConstants() != null) {
      constants = ctx.enumConstants().enumConstant().asScala
                      .map(x=>addWhitespaceRight(visitEnumConstant(x),x)).toList
    }
    var comma: Option[StringNode] = None
    if (ctx.COMMA() != null) {
      comma = Option(StringNode(",", getSourceLocation(ctx.COMMA())))
      comma.get.precedingWhitespace = getWhitespaceLeft(ctx.COMMA())
    }
    var decls: List[MemberDeclaration] = Nil
    var semi: Option[StringNode] = None
    if (ctx.enumBodyDeclarations() != null) {
      semi = Option(StringNode(";", getSourceLocation(ctx.enumBodyDeclarations().SEMI())))
      semi.get.precedingWhitespace = getWhitespaceLeft(ctx.enumBodyDeclarations().SEMI())
      decls = ctx.enumBodyDeclarations().classBodyDeclaration().asScala
                .map(visitClassBodyDeclaration).toList
    }
    val res = EnumDeclaration(name, interfaces, constants, decls, comma, semi, getSourceLocation(ctx))
    res.enumKeyword.precedingWhitespace = getWhitespaceLeft(ctx.ENUM())
    res.openBrace.precedingWhitespace = getWhitespaceLeft(ctx.LBRACE())
    res.closeBrace.precedingWhitespace = getWhitespaceLeft(ctx.RBRACE())
    res
  }

  override def visitEnumConstant(ctx: JavaParser.EnumConstantContext): EnumConstant = {
    val annotations = ctx.annotation().asScala.map(visitAnnotation).toList
    val name = visitJavaIdentifier(ctx.javaIdentifier())
    val args = if(ctx.arguments() != null) Option(visitArguments(ctx.arguments())) else None
    val body = if(ctx.classBody() != null) Option(visitClassBody(ctx.classBody())) else None
    val constant = EnumConstant(name, annotations, args, body, getSourceLocation(ctx))
    constant
  }

  override def visitInterfaceDeclaration(ctx: JavaParser.InterfaceDeclarationContext): InterfaceDeclaration = {
    val name = visitJavaIdentifier(ctx.javaIdentifier())
    var sup: Option[MultiSuperclassClause] = None
    if (ctx.EXTENDS() != null) {
      val ext = MultiSuperclassClause(visitTypeListAsList(ctx.typeList()),
                                      getSourceLocation(ctx.typeList()))
      ext.extendsKeyword.precedingWhitespace = getWhitespaceLeft(ctx.EXTENDS())
      sup = Option(ext)
    }
    val typeParameterClause = if (ctx.typeParameters() != null)
                                Option(visitTypeParameters(ctx.typeParameters()))
                              else None
    val body = visitInterfaceBody(ctx.interfaceBody())
    val res = InterfaceDeclaration(name, typeParameterClause, sup, body, getSourceLocation(ctx))
    res.interfaceKeyword.precedingWhitespace = getWhitespaceLeft(ctx.INTERFACE())
    res
  }

  override def visitClassBody(ctx: JavaParser.ClassBodyContext): Body = {
    val decls = ctx.classBodyDeclaration().asScala.map(visitClassBodyDeclaration).toList
    val body = Body(decls, getSourceLocation(ctx))
    body.openBrace.precedingWhitespace = getWhitespaceLeft(ctx.LBRACE())
    body.closeBrace.precedingWhitespace = getWhitespaceLeft(ctx.RBRACE())
    body
  }

  override def visitInterfaceBody(ctx: JavaParser.InterfaceBodyContext): Body = {
    val decls = ctx.interfaceBodyDeclaration().asScala.map(visitInterfaceBodyDeclaration).toList
    val body = Body(decls, getSourceLocation(ctx))
    body.openBrace.precedingWhitespace = getWhitespaceLeft(ctx.LBRACE())
    body.closeBrace.precedingWhitespace = getWhitespaceLeft(ctx.RBRACE())
    body
  }

  override def visitClassBodyDeclaration(ctx: JavaParser.ClassBodyDeclarationContext): MemberDeclaration = {
    var decl: MemberDeclaration = null
    ctx.altNum match {
      case 1 =>
        decl = new EmptyStatement(getSourceLocation(ctx), getWhitespaceLeft(ctx.SEMI()))
      case 2 =>
        decl = visitBlock(ctx.block())
        if (ctx.STATIC() != null) {
          val static = StaticModifier(getSourceLocation(ctx.STATIC()), getWhitespaceLeft(ctx.STATIC()))
          decl.modifiers = decl.modifiers.prepended(static)
        }
      case 3 =>
        decl = visitMemberDeclaration(ctx.memberDeclaration())
        decl.contractBefore = if (ctx.valEmbedContract() != null) {
                                Option(visitValEmbedContract(ctx.valEmbedContract()))
                              } else if (ctx.valEmbedContractStart() != null) {
                                Option(visitValEmbedContractStart(ctx.valEmbedContractStart()))
                              } else { None }
        decl.modifiers = decl.modifiers.prependedAll(ctx.modifier().asScala.map(visitModifier))
        if (ctx.valEmbedModifiersEnd() != null) {
          decl.modifiers = decl.modifiers.prepended(visitValEmbedModifiersEnd(ctx.valEmbedModifiersEnd()))
        }
      case 4 =>
        decl = visitValEmbedDeclarationBlock(ctx.valEmbedDeclarationBlock())
      case 5 =>
        decl = visitValDeclaration(ctx.valDeclaration())
      case x =>
        println(s"Error: unexpected AltNum in ClassBodyDeclaration: $x")
    }
    decl.srcLocation = getSourceLocation(ctx)
    decl
  }

  override def visitMemberDeclaration(ctx: JavaParser.MemberDeclarationContext): MemberDeclaration = {
    ctx.altNum match {
      case 1 => visitMethodDeclaration(ctx.methodDeclaration())
      case 2 => visitGenericMethodDeclaration(ctx.genericMethodDeclaration())
      case 3 => visitFieldDeclaration(ctx.fieldDeclaration())
      case 4 => visitConstructorDeclaration(ctx.constructorDeclaration())
      case 5 => visitGenericConstructorDeclaration(ctx.genericConstructorDeclaration())
      case 6 => visitInterfaceDeclaration(ctx.interfaceDeclaration())
      case 7 => visitAnnotationTypeDeclaration(ctx.annotationTypeDeclaration())
      case 8 => visitClassDeclaration(ctx.classDeclaration())
      case 9 => visitEnumDeclaration(ctx.enumDeclaration())
      case x =>
        println(s"Error: unexpected AltNum in MemberDeclaration: $x")
        null
    }
  }

  def setMethodDeclarationFields(ctx: JavaParser.MethodDeclarationContext, decl: MethodDeclaration): Unit = {
    decl.name = visitJavaIdentifier(ctx.javaIdentifier())
    decl.returnType = Option(visitTypeTypeOrVoid(ctx.typeTypeOrVoid()))
    if (ctx.formalParameters().formalParameterList() != null) {
      decl.parameters = visitFormalParameterListAsList(ctx.formalParameters().formalParameterList())
    }
    var dimensions = ""
    for (x <- 0 until ctx.LBRACK().size()) {
      dimensions += getWhitespaceLeft(ctx.LBRACK(x)) + ctx.LBRACK(x).getText
      dimensions += getWhitespaceLeft(ctx.RBRACK(x)) + ctx.RBRACK(x).getText
    }
    if (dimensions != "") {
      decl.dimensions =  Option(StringNode(dimensions, getSourceLocation(ctx.LBRACK(0)))) // todo: better src?
    }
    if (ctx.THROWS() != null) {
      decl.throws = Option(ThrowsClause(visitQualifiedNameListAsList(ctx.qualifiedNameList()),
                                        getSourceLocation(ctx.qualifiedNameList())))
      decl.throws.get.throwsKeyword.precedingWhitespace = getWhitespaceLeft(ctx.THROWS())
    }
    decl.body = visitMethodBody(ctx.methodBody())
    decl.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.formalParameters().LPAREN())
    decl.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.formalParameters().RPAREN())
  }

  override def visitMethodDeclaration(ctx: JavaParser.MethodDeclarationContext): MethodDeclaration = {
    val decl = new MethodDeclaration(null, null, Nil, None, None, None, None,
                                     null, getSourceLocation(ctx))
    setMethodDeclarationFields(ctx, decl)
    decl
  }

  override def visitMethodBody(ctx: JavaParser.MethodBodyContext): BlockOrSemicolon = {
    if (ctx.SEMI() != null) {
      new EmptyStatement(getSourceLocation(ctx), getWhitespaceLeft(ctx.SEMI()))
    } else {
      visitBlock(ctx.block())
    }
  }

  override def visitTypeTypeOrVoid(ctx: JavaParser.TypeTypeOrVoidContext): TypeIndication = {
    if (ctx.typeType() != null) {
      visitTypeType(ctx.typeType())
    } else {
      new StringTypeIndication(ctx.VOID().getText, Nil, ctx.VOID().getText,
                               getSourceLocation(ctx), getWhitespaceLeft(ctx.VOID()))
    }
  }

  override def visitGenericMethodDeclaration(ctx: JavaParser.GenericMethodDeclarationContext):
        GenericMethodDeclaration = {
    val typeParameters = visitTypeParameters(ctx.typeParameters())
    val decl = GenericMethodDeclaration(null, typeParameters, null, Nil, None, None, None, None,
                                        null, getSourceLocation(ctx))
    setMethodDeclarationFields(ctx.methodDeclaration(), decl)
    decl
  }

  def setConstructorDeclarationFields(ctx: JavaParser.ConstructorDeclarationContext,
                                      decl: MethodDeclaration): Unit = {
    decl.name = visitJavaIdentifier(ctx.javaIdentifier())
    if (ctx.formalParameters().formalParameterList() != null) {
      decl.parameters = visitFormalParameterListAsList(ctx.formalParameters().formalParameterList())
    }
    if (ctx.THROWS() != null) {
      decl.throws = Option(ThrowsClause(visitQualifiedNameListAsList(ctx.qualifiedNameList()),
                                        getSourceLocation(ctx.qualifiedNameList())))
      decl.throws.get.throwsKeyword.precedingWhitespace = getWhitespaceLeft(ctx.THROWS())
    }
    decl.body = visitBlock(ctx.constructorBody)
    decl.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.formalParameters().LPAREN())
    decl.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.formalParameters().RPAREN())
  }
  override def visitGenericConstructorDeclaration(ctx: JavaParser.GenericConstructorDeclarationContext):
                      GenericMethodDeclaration = {
    val typeParameters = visitTypeParameters(ctx.typeParameters())
    val decl = GenericMethodDeclaration(null, typeParameters, None, Nil, None, None, None, None,
                                        null, getSourceLocation(ctx))
    setConstructorDeclarationFields(ctx.constructorDeclaration(), decl)
    decl
  }
  override def visitConstructorDeclaration(ctx: JavaParser.ConstructorDeclarationContext):
              MethodDeclaration = {
    val decl = new MethodDeclaration(null, None, Nil, None, None, None, None,
                                     null, getSourceLocation(ctx))
    setConstructorDeclarationFields(ctx, decl)
    decl
  }

  override def visitFieldDeclaration(ctx: JavaParser.FieldDeclarationContext): FieldDeclaration = {
    val fieldType = visitTypeType(ctx.typeType())
    val declarators = visitVariableDeclaratorsAsList(ctx.variableDeclarators())
    val field = FieldDeclaration(fieldType, declarators, getSourceLocation(ctx))
    field.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
    field
  }

  override def visitInterfaceBodyDeclaration(ctx: JavaParser.InterfaceBodyDeclarationContext):
                MemberDeclaration = {
    ctx.altNum match {
      case 1 =>
        val decl = visitInterfaceMemberDeclaration(ctx.interfaceMemberDeclaration())
        decl.contractBefore = if (ctx.valEmbedContract() != null) {
                                Option(visitValEmbedContract(ctx.valEmbedContract()))
                              } else if (ctx.valEmbedContractStart() != null) {
                                Option(visitValEmbedContractStart(ctx.valEmbedContractStart()))
                              } else { None }
        decl.modifiers = decl.modifiers.prependedAll(ctx.modifier().asScala.map(visitModifier))
        if (ctx.valEmbedModifiersEnd() != null) {
          decl.modifiers = decl.modifiers.prepended(visitValEmbedModifiersEnd(ctx.valEmbedModifiersEnd()))
        }
        decl.srcLocation = getSourceLocation(ctx)
        decl
      case 2 =>
        visitValEmbedDeclarationBlock(ctx.valEmbedDeclarationBlock())
      case 3 =>
        new EmptyStatement(getSourceLocation(ctx), getWhitespaceLeft(ctx.SEMI()))
      case 4 =>
        visitValDeclaration(ctx.valDeclaration())
      case x =>
        println(s"Error: unexpected AltNum in InterfaceBodyDeclaration: $x")
        null
    }
  }

  override def visitInterfaceMemberDeclaration(ctx: JavaParser.InterfaceMemberDeclarationContext):
                  MemberDeclaration = {
    ctx.altNum match {
      case 1 => visitConstDeclaration(ctx.constDeclaration())
      case 2 => visitInterfaceMethodDeclaration(ctx.interfaceMethodDeclaration())
      case 3 => visitGenericInterfaceMethodDeclaration(ctx.genericInterfaceMethodDeclaration())
      case 4 => visitInterfaceDeclaration(ctx.interfaceDeclaration())
      case 5 => visitAnnotationTypeDeclaration(ctx.annotationTypeDeclaration())
      case 6 => visitClassDeclaration(ctx.classDeclaration())
      case 7 => visitEnumDeclaration(ctx.enumDeclaration())
      case x =>
        println(s"Error: unexpected AltNum in InterfaceMemberDeclaration: $x")
        null
    }
  }

  override def visitConstDeclaration(ctx: JavaParser.ConstDeclarationContext): FieldDeclaration = {
    val fieldType = visitTypeType(ctx.typeType())
    val decls = ctx.constantDeclarator().asScala
                  .map(x=>addWhitespaceRight(visitConstantDeclarator(x),x)).toList
    decls.last.trailingWhitespace = ""
    val field = FieldDeclaration(fieldType, decls, getSourceLocation(ctx))
    field.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
    field
  }

  override def visitConstantDeclarator(ctx: JavaParser.ConstantDeclaratorContext): VariableDeclarator = {
    val name = visitJavaIdentifier(ctx.javaIdentifier())
    for (i <- 0 until ctx.LBRACK().size()) {
      name.text += getWhitespaceLeft(ctx.LBRACK(i)) + ctx.LBRACK(i).getText
      name.text += getWhitespaceLeft(ctx.RBRACK(i)) + ctx.RBRACK(i).getText
    }
    val initializer = Option(visitVariableInitializer(ctx.variableInitializer()))
    val decl = VariableDeclarator(name, initializer, getSourceLocation(ctx))
    decl.equalSign.precedingWhitespace = getWhitespaceLeft(ctx.ASSIGN())
    decl
  }

  def setInterfaceMethodDeclarationFields(ctx: JavaParser.InterfaceMethodDeclarationContext,
                                          decl: MethodDeclaration): Unit = {
    decl.name = visitJavaIdentifier(ctx.javaIdentifier())
    decl.returnType = None
    if (ctx.typeParameters() != null) {
      val typeParameters = visitTypeParameters(ctx.typeParameters())
      val annotations = ctx.annotation().asScala.map(visitAnnotation).toList
      val text = printer.visitSkipFirstWS(ctx.typeTypeOrVoid())
      val typeIndication = new StringGenericTypeIndication(text, typeParameters, annotations,
                                                           ctx.getText, getSourceLocation(ctx),
                                                           getWhitespaceLeft(ctx))
      decl.returnType = Option(typeIndication)
    } else {
      decl.returnType = Option(visitTypeTypeOrVoid(ctx.typeTypeOrVoid()))
    }
    decl.parameters = if (ctx.formalParameters().formalParameterList() != null)
                      visitFormalParameterListAsList(ctx.formalParameters().formalParameterList())
                    else Nil
    var dimensions = ""
    for (x <- 0 until ctx.LBRACK().size()) {
      dimensions += getWhitespaceLeft(ctx.LBRACK(x)) + ctx.LBRACK(x).getText
      dimensions += getWhitespaceLeft(ctx.RBRACK(x)) + ctx.RBRACK(x).getText
    }
    decl.dimensions = if (dimensions=="") None
                      else Option(StringNode(dimensions, getSourceLocation(ctx.LBRACK(0)))) // todo: better src?
    decl.throws = None
    if (ctx.THROWS() != null) {
      decl.throws = Option(ThrowsClause(visitQualifiedNameListAsList(ctx.qualifiedNameList()),
                                        getSourceLocation(ctx.qualifiedNameList())))
      decl.throws.get.throwsKeyword.precedingWhitespace = getWhitespaceLeft(ctx.THROWS())
    }
    decl.body = visitMethodBody(ctx.methodBody())
    decl.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.formalParameters().LPAREN())
    decl.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.formalParameters().RPAREN())
  }

  override def visitInterfaceMethodDeclaration(ctx: JavaParser.InterfaceMethodDeclarationContext):
            MethodDeclaration = {
    val decl = new MethodDeclaration(null, null, Nil, None, None, None, None,
                                     null, getSourceLocation(ctx))
    setInterfaceMethodDeclarationFields(ctx, decl)
    decl
  }

  override def visitInterfaceMethodModifier(ctx: JavaParser.InterfaceMethodModifierContext): Modifier = {
    ctx.altNum match {
      case 1 => visitAnnotation(ctx.annotation())
      case 5 => StaticModifier(getSourceLocation(ctx), getWhitespaceLeft(ctx))
      case _ =>
        new ModifierStringNode(ctx.getText, getSourceLocation(ctx), getWhitespaceLeft(ctx))
    }
  }

  override def visitGenericInterfaceMethodDeclaration(ctx: JavaParser.GenericInterfaceMethodDeclarationContext):
                GenericMethodDeclaration = {
    val typeParameters = visitTypeParameters(ctx.typeParameters())
    val decl = GenericMethodDeclaration(null, typeParameters, null, Nil, None, None, None, None,
                                        null, getSourceLocation(ctx))
    setInterfaceMethodDeclarationFields(ctx.interfaceMethodDeclaration(), decl)
    decl
  }

  def visitVariableDeclaratorsAsList(ctx: JavaParser.VariableDeclaratorsContext): List[VariableDeclarator] = {
    val decls = ctx.variableDeclarator().asScala
                  .map(x=>addWhitespaceRight(visitVariableDeclarator(x),x)).toList
    decls.last.trailingWhitespace = ""
    decls
  }

  override def visitVariableDeclarator(ctx: JavaParser.VariableDeclaratorContext): VariableDeclarator = {
    val name = visitVariableDeclaratorId(ctx.variableDeclaratorId())
    val initializer = if (ctx.ASSIGN() == null) None
                      else Option(visitVariableInitializer(ctx.variableInitializer()))
    val decl = VariableDeclarator(name, initializer, getSourceLocation(ctx))
    if (ctx.ASSIGN() != null) {
      decl.equalSign.precedingWhitespace = getWhitespaceLeft(ctx.ASSIGN())
    }
    decl
  }

  override def visitVariableDeclaratorId(ctx: JavaParser.VariableDeclaratorIdContext): StringNode = {
    printNode(ctx)
  }

  override def visitVariableInitializer(ctx: JavaParser.VariableInitializerContext): Initializer = {
    ctx.altNum match {
      case 1 => visitArrayInitializer(ctx.arrayInitializer())
      case 2 => visitExpression(ctx.expression())
      case x =>
        println(s"Error: unexpected AltNum in VariableInitializer: $x")
        null
    }
  }

  override def visitArrayInitializer(ctx: JavaParser.ArrayInitializerContext): ArrayInitializer = {
    val initializers = ctx.variableInitializer().asScala
                        .map(x=>addWhitespaceRight(visitVariableInitializer(x),x)).toList
    var comma: Option[StringNode] = None
    if (ctx.variableInitializer().size() == ctx.COMMA().size() && ctx.COMMA().size() > 0) {
      val trailingComma: StringNode = StringNode(",", getSourceLocation(ctx.COMMA().asScala.last))
      trailingComma.precedingWhitespace = getWhitespaceLeft(ctx.COMMA().asScala.last)
      comma = Option(trailingComma)
    }
    val ini = ArrayInitializer(initializers, comma, getSourceLocation(ctx))
    ini.openBrace.precedingWhitespace = getWhitespaceLeft(ctx.LBRACE())
    ini.closeBrace.precedingWhitespace = getWhitespaceLeft(ctx.RBRACE())
    ini
  }

  override def visitClassOrInterfaceType(ctx: JavaParser.ClassOrInterfaceTypeContext): TypeIndication = {
    val text = printer.visitSkipFirstWS(ctx)
    new StringTypeIndication(text, Nil, ctx.getText, getSourceLocation(ctx), getWhitespaceLeft(ctx))
  }

  override def visitTypeArgument(ctx: JavaParser.TypeArgumentContext): TypeArgument = {
    ctx.altNum match {
      case 1 =>
        visitTypeType(ctx.typeType())
      case 2 =>
        val annotations = ctx.annotation().asScala.map(visitAnnotation).toList
        var superType: Option[TypeIndication] = None
        var extendsType: Option[TypeIndication] = None
        if (ctx.EXTENDS() != null) {
      extendsType = Some(visitTypeType(ctx.typeType()))
        }
        if (ctx.SUPER() != null) {
          superType = Some(visitTypeType(ctx.typeType()))
        }
        val arg = BoundedTypeArgument(annotations, superType, extendsType, getSourceLocation(ctx))
        arg.questionMark.precedingWhitespace = getWhitespaceLeft(ctx.QUESTION())
        if (ctx.EXTENDS() != null) {
          arg.extendsKeyword.precedingWhitespace = getWhitespaceLeft(ctx.EXTENDS())
        }
        if (ctx.SUPER() != null) {
          arg.superKeyword.precedingWhitespace = getWhitespaceLeft(ctx.EXTENDS())
        }
        arg
    }
  }

  def visitQualifiedNameListAsList(ctx: JavaParser.QualifiedNameListContext): List[StringNode] = {
    val names = ctx.qualifiedName().asScala.map(visitQualifiedNameWithWS).toList
    names.last.trailingWhitespace = ""
    names
  }

  def visitFormalParameterListAsList(ctx: JavaParser.FormalParameterListContext): List[Parameter] = {
    var params:List[Parameter] = Nil
    if (ctx.formalParameter() != null) {
      params = ctx.formalParameter().asScala
                  .map(x=>addWhitespaceRight(visitFormalParameter(x),x)).toList
    }
    if (ctx.lastFormalParameter() != null) {
      params.appended(visitLastFormalParameter(ctx.lastFormalParameter()))
    }
    params
  }

  override def visitFormalParameter(ctx: JavaParser.FormalParameterContext): Parameter = {
    val modifiers = ctx.variableModifier().asScala.map(visitVariableModifier).toList
    val paramType = visitTypeType(ctx.typeType())
    val name = visitVariableDeclaratorId(ctx.variableDeclaratorId())
    val param = Parameter(paramType, name, getSourceLocation(ctx))
    param.modifiers = modifiers
    param
  }

  override def visitLastFormalParameter(ctx: JavaParser.LastFormalParameterContext): VarargsParameter = {
    val modifiers = ctx.variableModifier().asScala.map(visitVariableModifier).toList
    val paramType = visitTypeType(ctx.typeType())
    val annotations = ctx.annotation().asScala.map(visitAnnotation).toList
    val name = visitVariableDeclaratorId(ctx.variableDeclaratorId())
    val param = VarargsParameter(paramType, annotations, name, getSourceLocation(ctx))
    param.modifiers = modifiers
    param.dots.precedingWhitespace = getWhitespaceLeft(ctx.ELLIPSIS())
    param
  }

  override def visitQualifiedName(ctx: JavaParser.QualifiedNameContext): StringNode = printNode(ctx)
  def visitQualifiedNameWithWS(ctx: JavaParser.QualifiedNameContext): StringNode = printNodeWithWS(ctx)
  def visitQualifiedNameWithWSAsTypeIndication(ctx: JavaParser.QualifiedNameContext): TypeIndication = {
    val text = printer.visitSkipFirstWS(ctx)
    new StringTypeIndication(text, Nil, ctx.getText, getSourceLocation(ctx),
                            getWhitespaceLeft(ctx), getWhitespaceRight(ctx))
  }


  override def visitLiteral(ctx: JavaParser.LiteralContext): Literal = {
    Literal(printNode(ctx), getSourceLocation(ctx))
  }

  // TODO: how much of the annotation-related ones do we need?
  override def visitAnnotation(ctx: JavaParser.AnnotationContext): StringAnnotation = {
    val src = getSourceLocation(ctx)
    val text = printer.visit(ctx)
    StringAnnotation(text, src)
  }

  override def visitAnnotationTypeDeclaration(ctx: JavaParser.AnnotationTypeDeclarationContext):
            AnnotationTypeDeclaration = {
    // todo
    super.visitAnnotationTypeDeclaration(ctx).asInstanceOf[AnnotationTypeDeclaration]
  }


  override def visitBlock(ctx: JavaParser.BlockContext): Block = {
    val statements = ctx.blockStatement().asScala.map(visitBlockStatement).toList
    val block = Block(None, statements, getSourceLocation(ctx))
    block.openBrace.precedingWhitespace = getWhitespaceLeft(ctx.LBRACE())
    block.closeBrace.precedingWhitespace = getWhitespaceLeft(ctx.RBRACE())
    block
  }

  override def visitBlockStatement(ctx: JavaParser.BlockStatementContext): Statement = {
    ctx.altNum match {
      case 1 =>
        val decl = visitLocalVariableDeclaration(ctx.localVariableDeclaration())
        decl.isStatement = true
        decl.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
        decl
      case 2 =>
        visitStatement(ctx.statement())
      case 3 =>
        visitLocalTypeDeclaration(ctx.localTypeDeclaration())
      case 4 =>
        visitValEmbedStatementBlock(ctx.valEmbedStatementBlock())
      case x =>
        println(s"Error: unexpected AltNum in BlockStatement: $x")
        null
    }
  }

  override def visitLocalVariableDeclaration(ctx: JavaParser.LocalVariableDeclarationContext):
            LocalDeclaration = {
    val modifiers = ctx.variableModifier().asScala.map(visitVariableModifier).toList
    val typeIndication = visitTypeType(ctx.typeType())
    val declarators = visitVariableDeclaratorsAsList(ctx.variableDeclarators())
    val decl = LocalDeclaration(modifiers, typeIndication, declarators, isStatement = false, getSourceLocation(ctx))
    decl
  }

  override def visitLocalTypeDeclaration(ctx: JavaParser.LocalTypeDeclarationContext): Statement = {
    if (ctx.classDeclaration() != null) {
      val decl = visitClassDeclaration(ctx.classDeclaration())
      decl.modifiers = decl.modifiers.prependedAll(ctx.classOrInterfaceModifier().asScala
                                                    .map(visitClassOrInterfaceModifier))
      decl
    } else if (ctx.interfaceDeclaration() != null){
      val decl = visitInterfaceDeclaration(ctx.interfaceDeclaration())
      decl.modifiers = decl.modifiers.prependedAll(ctx.classOrInterfaceModifier().asScala
                                                      .map(visitClassOrInterfaceModifier))
      decl
    } else {
      new EmptyStatement(getSourceLocation(ctx), getWhitespaceLeft(ctx.SEMI()))
    }
  }

  override def visitStatement(ctx: JavaParser.StatementContext): Statement = {
    ctx.altNum match {
      case 1 =>
        visitBlock(ctx.block())
      case 2 =>
        val content = visitExpression(ctx.expression(0))
        val message = if (ctx.COLON() != null) Option(visitExpression(ctx.expression(1)))
                      else None
        val assertion = JavaAssertion(content, message, getSourceLocation(ctx))
        assertion.assertKeyword.precedingWhitespace = getWhitespaceLeft(ctx.ASSERT())
        if (ctx.COLON() != null) {
          assertion.colon.precedingWhitespace = getWhitespaceLeft(ctx.COLON())
        }
        assertion.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
        assertion
      case 3 =>
        val cond = visitExpression(ctx.parExpression().expression())
        val thenStatement = visitStatement(ctx.statement(0))
        val elseStatement = if (ctx.ELSE() != null) Option(visitStatement(ctx.statement(1)))
                            else None
        val ifStatement = IfStatement(cond, thenStatement, elseStatement, getSourceLocation(ctx))
        ifStatement.ifKeyword.precedingWhitespace = getWhitespaceLeft(ctx.IF())
        ifStatement.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.parExpression().LPAREN())
        ifStatement.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.parExpression().RPAREN())
        if (ctx.ELSE() != null) {
          ifStatement.elseKeyword.precedingWhitespace = getWhitespaceLeft(ctx.ELSE())
        }
        ifStatement
      case 4 =>
        val contract = if (ctx.valEmbedLoopSpec() == null) Nil
                        else visitValEmbedLoopSpecAsList(ctx.valEmbedLoopSpec())
        val control = visitForControl(ctx.forControl())
        val body = visitStatement(ctx.statement(0))
        val loop = ForLoop(contract, control, Nil, body, getSourceLocation(ctx))
        loop.forKeyword.precedingWhitespace = getWhitespaceLeft(ctx.FOR())
        loop.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.LPAREN())
        loop.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.RPAREN())
        loop
      case 5 =>
        val contract = if (ctx.valEmbedLoopSpec() == null) Nil
                        else visitValEmbedLoopSpecAsList(ctx.valEmbedLoopSpec())
        val cond = visitExpression(ctx.parExpression().expression())
        val body = visitStatement(ctx.statement(0))
        val loop = WhileLoop(contract, cond, Nil, body, getSourceLocation(ctx))
        loop.whileKeyword.precedingWhitespace = getWhitespaceLeft(ctx.WHILE())
        loop.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.parExpression().LPAREN())
        loop.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.parExpression().RPAREN())
        loop
      case 6 =>
        val contract = if (ctx.valEmbedLoopSpec() == null) Nil
                       else visitValEmbedLoopSpecAsList(ctx.valEmbedLoopSpec())
        val cond = visitExpression(ctx.parExpression().expression())
        val body = visitStatement(ctx.statement(0))
        val loop = DoLoop(contract, body, cond, getSourceLocation(ctx))
        loop.doKeyword.precedingWhitespace = getWhitespaceLeft(ctx.DO())
        loop.whileKeyword.precedingWhitespace = getWhitespaceLeft(ctx.WHILE())
        loop.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.parExpression().LPAREN())
        loop.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.parExpression().RPAREN())
        loop.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
        loop
      case 7 =>
        val block = visitBlock(ctx.block())
        val catches = ctx.catchClause().asScala.map(visitCatchClause).toList
        val finallyBlock = if(ctx.finallyBlock() == null) None
                           else Option(visitBlock(ctx.finallyBlock().block()))
        val tryBlock = TryBlock(None, block, catches, finallyBlock, getSourceLocation(ctx))
        tryBlock.tryKeyword.precedingWhitespace = getWhitespaceLeft(ctx.TRY())
        if (ctx.finallyBlock() != null) {
          tryBlock.tryKeyword.precedingWhitespace = getWhitespaceLeft(ctx.finallyBlock().FINALLY())
        }
        tryBlock
      case 8 =>
        val resource = visitResourceSpecification(ctx.resourceSpecification())
        val block = visitBlock(ctx.block())
        val catches = ctx.catchClause().asScala.map(visitCatchClause).toList
        val finallyBlock = if(ctx.finallyBlock() != null) Option(visitBlock(ctx.finallyBlock().block()))
                           else None
        val tryBlock = TryBlock(Option(resource), block, catches, finallyBlock, getSourceLocation(ctx))
        tryBlock.tryKeyword.precedingWhitespace = getWhitespaceLeft(ctx.TRY())
        if (ctx.finallyBlock() != null) {
          tryBlock.finallyKeyword.precedingWhitespace = getWhitespaceLeft(ctx.finallyBlock().FINALLY())
        }
        tryBlock
      case 9 =>
        val cond = visitExpression(ctx.parExpression().expression())
        var blocks = ctx.switchBlockStatementGroup().asScala
                        .map(visitSwitchBlockStatementGroup).toList
        if (!ctx.switchLabel().isEmpty) {
          val lastBlock = SwitchCaseBlock(ctx.switchLabel().asScala.map(visitSwitchLabel).toList, Nil,
                                          getSourceLocation(ctx.switchLabel(0)))
          blocks = blocks.appended(lastBlock)
        }
        val switchStatement = SwitchStatement(cond, blocks, getSourceLocation(ctx))
        switchStatement.switchKeyword.precedingWhitespace = getWhitespaceLeft(ctx.SWITCH())
        switchStatement.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.parExpression().LPAREN())
        switchStatement.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.parExpression().RPAREN())
        switchStatement.openBrace.precedingWhitespace = getWhitespaceLeft(ctx.LBRACE())
        switchStatement.closeBrace.precedingWhitespace = getWhitespaceLeft(ctx.RBRACE())
        switchStatement
      case 10 =>
        val cond = visitExpression(ctx.parExpression().expression())
        val body = visitBlock(ctx.block())
        val synchronizedStatement = SynchronizedStatement(cond, body, getSourceLocation(ctx))
        synchronizedStatement.synchronizedKeyword.precedingWhitespace = getWhitespaceLeft(ctx.SYNCHRONIZED())
        synchronizedStatement.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.parExpression().LPAREN())
        synchronizedStatement.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.parExpression().RPAREN())
        synchronizedStatement
      case 11 =>
        val res = if (!ctx.expression().isEmpty) Option(visitExpression(ctx.expression(0))) else None
        val returnStatement = ReturnStatement(res, getSourceLocation(ctx))
        returnStatement.returnKeyword.precedingWhitespace = getWhitespaceLeft(ctx.RETURN())
        returnStatement.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
        returnStatement
      case 12 =>
        val res = visitExpression(ctx.expression(0))
        val throwStatement = ThrowStatement(res, getSourceLocation(ctx))
        throwStatement.throwKeyword.precedingWhitespace = getWhitespaceLeft(ctx.THROW())
        throwStatement.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
        throwStatement
      case 13 =>
        val res = if(ctx.javaIdentifier() != null) Option(visitJavaIdentifier(ctx.javaIdentifier()))
                  else None
        val break = BreakStatement(res, getSourceLocation(ctx))
        break.breakKeyword.precedingWhitespace = getWhitespaceLeft(ctx.BREAK())
        break.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
        break
      case 14 =>
        val res = if(ctx.javaIdentifier() != null) Option(visitJavaIdentifier(ctx.javaIdentifier()))
                  else None
        val cont = ContinueStatement(res, getSourceLocation(ctx))
        cont.continueKeyword.precedingWhitespace = getWhitespaceLeft(ctx.CONTINUE())
        cont.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
        cont
      case 15 =>
        new EmptyStatement(getSourceLocation(ctx), getWhitespaceLeft(ctx.SEMI()))
      case 16 =>
        val expr = ExpressionStatement(visitExpression(ctx.expression(0)), getSourceLocation(ctx))
        expr.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
        expr
      case 17 =>
        val label = visitJavaIdentifier(ctx.javaIdentifier())
        val lbl = Label(label, getSourceLocation(ctx.javaIdentifier()))
        lbl.colon.precedingWhitespace = getWhitespaceLeft(ctx.COLON())
        val stmt = visitStatement(ctx.statement(0))
        LabeledStatement(lbl, stmt, getSourceLocation(ctx))
      case 18 =>
        visitValStatement(ctx.valStatement())
      case x =>
        println(s"Error: unexpected AltNum in Statement: $x")
        null
    }
  }

  override def visitCatchClause(ctx: JavaParser.CatchClauseContext): CatchClause = {
    val modifiers = ctx.variableModifier().asScala.map(visitVariableModifier).toList
    val types = visitCatchTypeAsList(ctx.catchType())
    val name = visitJavaIdentifier(ctx.javaIdentifier())
    val block = visitBlock(ctx.block())
    val catchClause = CatchClause(types, name, block, getSourceLocation(ctx))
    catchClause.catchKeyword.precedingWhitespace = getWhitespaceLeft(ctx.CATCH())
    catchClause.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.LPAREN())
    catchClause.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.RPAREN())
    catchClause.modifiers = modifiers
    catchClause
  }

  def visitCatchTypeAsList(ctx: JavaParser.CatchTypeContext): List[TypeIndication] = {
    ctx.qualifiedName().asScala.map(visitQualifiedNameWithWSAsTypeIndication).toList
  }

  override def visitResourceSpecification(ctx: JavaParser.ResourceSpecificationContext):
          TryResourceSpecification = {
    val semi = if (ctx.SEMI() != null)
                  Option(StringNode(ctx.SEMI().getText,
                         getSourceLocation(ctx.SEMI()), getWhitespaceLeft(ctx.SEMI())))
               else None
    val resources = ctx.resources().resource().asScala
                        .map(x=>addWhitespaceRight(visitResource(x),x)).toList
    val res = TryResourceSpecification(resources, semi, getSourceLocation(ctx))
    res.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.LPAREN())
    res.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.RPAREN())
    res
  }

  override def visitResource(ctx: JavaParser.ResourceContext): TryResource = {
    val modifiers = ctx.variableModifier().asScala.map(visitVariableModifier).toList
    val rType = visitClassOrInterfaceType(ctx.classOrInterfaceType())
    val variable = visitVariableDeclaratorId(ctx.variableDeclaratorId())
    val expr = visitExpression(ctx.expression())
    val res = TryResource(modifiers, rType, variable, expr, getSourceLocation(ctx))
    res.equalsSign. precedingWhitespace = getWhitespaceLeft(ctx.ASSIGN())
    res
  }

  override def visitSwitchBlockStatementGroup(ctx: JavaParser.SwitchBlockStatementGroupContext):
                SwitchCaseBlock = {
    SwitchCaseBlock(ctx.switchLabel().asScala.map(visitSwitchLabel).toList,
                    ctx.blockStatement().asScala.map(visitBlockStatement).toList,
                    getSourceLocation(ctx))
  }

  override def visitSwitchLabel(ctx: JavaParser.SwitchLabelContext): SwitchLabel = {
    if (ctx.CASE() != null) {
      var cond: Expression = null
      if(ctx.expression() != null) {
        cond = visitExpression(ctx.expression())
      } else {
        cond = visitJavaIdentifier(ctx.javaIdentifier())
      }
      val label = SwitchLabel(Option(cond), getSourceLocation(ctx))
      label.caseKeyword.precedingWhitespace = getWhitespaceLeft(ctx.CASE())
      label.colon.precedingWhitespace = getWhitespaceLeft(ctx.COLON())
      label
    } else {
      val label = SwitchLabel(None, getSourceLocation(ctx))
      label.defaultKeyword.precedingWhitespace = getWhitespaceLeft(ctx.DEFAULT())
      label.colon.precedingWhitespace = getWhitespaceLeft(ctx.COLON())
      label
    }
  }

  override def visitForControl(ctx: JavaParser.ForControlContext): ForLoopControl = {
    if (ctx.enhancedForControl() != null) {
      visitEnhancedForControl(ctx.enhancedForControl())
    } else {
      val init = if(ctx.forInit()!=null) Option(visitForInit(ctx.forInit())) else None
      val cond = if(ctx.expression()!=null) Option(visitExpression(ctx.expression())) else None
      val update = if(ctx.expressionList()==null) Nil
                   else ctx.expressionList().expression().asScala
                          .map(x=>addWhitespaceRight(visitExpression(x),x)).toList
      val ctl = RegularForLoopControl(init, cond, update, getSourceLocation(ctx))
      ctl.firstSemicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI(0))
      ctl.secondSemicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI(1))
      ctl
    }
  }

  override def visitForInit(ctx: JavaParser.ForInitContext): ForInit = {
    if (ctx.localVariableDeclaration() != null) {
      visitLocalVariableDeclaration(ctx.localVariableDeclaration())
    } else {
      val exprs = ctx.expressionList().expression().asScala
                    .map(x=>addWhitespaceRight(visitExpression(x),x)).toList
      ExpressionListForInit(exprs, getSourceLocation(ctx))
    }
  }

  override def visitEnhancedForControl(ctx: JavaParser.EnhancedForControlContext): ForEachLoopControl = {
    val modifiers = ctx.variableModifier().asScala.map(visitVariableModifier).toList
    val cType = visitTypeType(ctx.typeType())
    val variable = visitVariableDeclaratorId(ctx.variableDeclaratorId())
    val control = visitExpression(ctx.expression())
    val ctl = ForEachLoopControl(cType, variable, control, getSourceLocation(ctx))
    ctl.modifiers = modifiers
    ctl.colon.precedingWhitespace = getWhitespaceLeft(ctx.COLON())
    ctl
  }


  override def visitParExpression(ctx: JavaParser.ParExpressionContext): Expression = {
    val parEx = ParExpression(visitExpression(ctx.expression()), getSourceLocation(ctx))
    parEx.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.LPAREN())
    parEx.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.RPAREN())
    parEx
  }

  override def visitMethodCall(ctx: JavaParser.MethodCallContext): MethodCallExpression = {
    var name: StringNode = null
    if (ctx.javaIdentifier() != null) {
      name = visitJavaIdentifier(ctx.javaIdentifier())
    } else if (ctx.THIS() != null) {
      name = StringNode(ctx.THIS().getText, getSourceLocation(ctx.THIS()), getWhitespaceLeft(ctx.THIS()))
    } else {
      name = StringNode(ctx.SUPER().getText, getSourceLocation(ctx.SUPER()), getWhitespaceLeft(ctx.SUPER()))
    }
    val args: List[Expression] = if(ctx.expressionList() == null) List()
                                 else ctx.expressionList().expression().asScala
                                      .map(x=>addWhitespaceRight(visitExpression(x),x)).toList
    val argsrc = if(ctx.expressionList() == null) getSourceLocation(ctx) else getSourceLocation(ctx.expressionList())
    val arguments = Arguments(args, argsrc)
    arguments.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.LPAREN())
    arguments.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.RPAREN())
    MethodCallExpression(name, arguments, getSourceLocation(ctx))
  }

  def visitExpression(ctx: JavaParser.ExpressionContext): Expression = {
    visit(ctx).asInstanceOf[Expression]
  }

  override def visitPrimaryExpression(ctx: JavaParser.PrimaryExpressionContext): Expression
        = visitPrimary(ctx.primary())

  override def visitValPrimaryExpression(ctx: JavaParser.ValPrimaryExpressionContext): Expression
        = visitValPrimary(ctx.valPrimary())

  override def visitBinaryExpression(ctx: JavaParser.BinaryExpressionContext): Expression = {
    var left = visitExpression(ctx.expression(0))
    var right = visitExpression(ctx.expression(1))
    var op: StringNode = null
    if (ctx.mulOp() != null) {
      op = StringNode(ctx.mulOp().getText, getSourceLocation(ctx.mulOp()), getWhitespaceLeft(ctx.mulOp()))
    } else if (ctx.bop != null) {
      op = StringNode(ctx.bop.getText, getSourceLocation(ctx.bop), getWhitespaceLeft(ctx.bop))
    } else if (ctx.andOp() != null) {
      op = StringNode(ctx.andOp().getText, getSourceLocation(ctx.andOp()), getWhitespaceLeft(ctx.andOp()))
    } else if (ctx.impOp() != null) {
      op = ImpliesOperator(getSourceLocation(ctx.impOp()), getWhitespaceLeft(ctx.impOp()))
      if(ctx.impOp().valImpOp().BACK_IMPLIES() != null){
        val tmp = left
        left = right
        right = left
      }
    } else if (ctx.eqOp() != null) {
      op = StringNode(ctx.eqOp().getText, getSourceLocation(ctx.eqOp()), getWhitespaceLeft(ctx.eqOp()))
    } else{
      var text = ""
      if (ctx.GT().size() > 0){
        for (x <- ctx.GT().asScala) {
          text += x.getText
        }
        op = StringNode(text, getSourceLocation(ctx.GT(0)), getWhitespaceLeft(ctx.GT(0)))
      } else {
        for (x <- ctx.LT().asScala) {
          text += x.getText
        }
        op = StringNode(text, getSourceLocation(ctx.LT(0)), getWhitespaceLeft(ctx.LT(0)))
      }
    }
    BinaryExpression(left, op, right, getSourceLocation(ctx))
  }

  override def visitInstanceofExpression(ctx: JavaParser.InstanceofExpressionContext): Expression = {
    val left = visitExpression(ctx.expression())
    val right = visitTypeType(ctx.typeType())
    val res = InstanceofExpression(left, right, getSourceLocation(ctx))
    res.instanceofKeyword.precedingWhitespace = getWhitespaceLeft(ctx.INSTANCEOF())
    res
  }

  override def visitNestedInvocationExpression(ctx: JavaParser.NestedInvocationExpressionContext):
          Expression = {
    val left = visitExpression(ctx.expression())
    var right: Expression = null
    if (ctx.javaIdentifier() != null) {
      right = visitJavaIdentifier(ctx.javaIdentifier())
    } else if (ctx.methodCall() != null) {
      right = visitMethodCall(ctx.methodCall())
    } else if (ctx.THIS() != null) {
      right = StringNode(ctx.THIS().getText, getSourceLocation(ctx.THIS()),
                          getWhitespaceLeft(ctx.THIS()))
    } else if (ctx.NEW() != null) {
      val creator = visitInnerCreator(ctx.innerCreator())
      creator.newKeyword.precedingWhitespace = getWhitespaceLeft(ctx.NEW())
      if (ctx.nonWildcardTypeArguments() != null) {
        creator.typeArguments = Some(visitNonWildcardTypeArguments(ctx.nonWildcardTypeArguments()))
      }
      right = creator
    } else if (ctx.SUPER() != null) {
      if (ctx.superSuffix().javaIdentifier() != null) {
        var text = ctx.SUPER().getText
        text += getWhitespaceLeft(ctx.superSuffix().DOT())
        text += ctx.superSuffix().DOT().getText
        text += getWhitespaceLeft(ctx.superSuffix().javaIdentifier())
        text += ctx.superSuffix().javaIdentifier().getText
        if (ctx.superSuffix().arguments()!=null) {
          val name = StringNode(text, getSourceLocation(ctx), getWhitespaceLeft(ctx.SUPER()))
          val args = visitArguments(ctx.superSuffix().arguments())
          right = MethodCallExpression(name, args, getSourceLocation(ctx))
        } else {
          right = StringNode(text, getSourceLocation(ctx), getWhitespaceLeft(ctx.SUPER()))
        }
      } else {
        val name = StringNode(ctx.SUPER().getText, getSourceLocation(ctx.SUPER()),
                              getWhitespaceLeft(ctx.SUPER()))
        val arguments = visitArguments(ctx.superSuffix().arguments())
        right = MethodCallExpression(name, arguments, getSourceLocation(ctx))
      }
    } else {
      right = visitExplicitGenericInvocation(ctx.explicitGenericInvocation())
    }
    val expr = NestedInvocationExpression(left, right, getSourceLocation(ctx))
    expr.dot.precedingWhitespace = getWhitespaceLeft(ctx.DOT())
    expr
  }

  override def visitArrayIndexExpression(ctx: JavaParser.ArrayIndexExpressionContext): Expression = {
    val arrExpr = visitExpression(ctx.expression())
    var index: IndexExpression = null
    if (ctx.indexExpression().expression() != null) {
      index = visitExpression(ctx.indexExpression().expression())
    } else {
      val idxExprCtx = ctx.indexExpression().valIndexExpression()
      if (idxExprCtx.MUL() != null) {
        val idx = IndexAllExpression(getSourceLocation(ctx))
        idx.star.precedingWhitespace = getWhitespaceLeft(idxExprCtx.MUL())
        index = idx
      } else {
        val start = visitExpression(idxExprCtx.langExpr(0).expression())
        val end = visitExpression(idxExprCtx.langExpr(1).expression())
        val range = RangeExpression(start, end, getSourceLocation(ctx))
        range.dots.precedingWhitespace = getWhitespaceLeft(idxExprCtx.RANGE_TO())
        index = range
      }
    }
    val arr = ArrayIndexExpression(arrExpr, index, getSourceLocation(ctx))
    arr.openBracket.precedingWhitespace = getWhitespaceLeft(ctx.LBRACK())
    arr.closeBracket.precedingWhitespace = getWhitespaceLeft(ctx.RBRACK())
    arr
  }

  override def visitMethodCallExpression(ctx: JavaParser.MethodCallExpressionContext): Expression =
    visitMethodCall(ctx.methodCall())

  override def visitNewExpression(ctx: JavaParser.NewExpressionContext): Expression = {
    val newExpr = visitCreator(ctx.creator())
    newExpr.newKeyword.precedingWhitespace = getWhitespaceLeft(ctx.NEW())
    newExpr
  }

  override def visitCastExpression(ctx: JavaParser.CastExpressionContext): Expression = {
    val annotations = ctx.annotation().asScala.map(visitAnnotation).toList
    val types = ctx.typeType().asScala.map(visitTypeTypeWithWS).toList
    val expr = visitExpression(ctx.expression())
    val cast = CastExpression(annotations, types, expr, getSourceLocation(ctx))
    cast.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.LPAREN())
    cast.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.RPAREN())
    cast
  }

  override def visitPostfixExpression(ctx: JavaParser.PostfixExpressionContext): Expression = {
    val expr = visitExpression(ctx.expression())
    val op = StringNode(ctx.postfix.getText, getSourceLocation(ctx.postfix), getWhitespaceLeft(ctx.postfix))
    PostfixExpression(expr, op, getSourceLocation(ctx))
  }

  override def visitPrefixExpression(ctx: JavaParser.PrefixExpressionContext): Expression = {
    val expr = visitExpression(ctx.expression())
    val op = StringNode(ctx.prefix.getText, getSourceLocation(ctx.prefix), getWhitespaceLeft(ctx.prefix))
    PrefixExpression(expr, op, getSourceLocation(ctx))
  }

  override def visitTernaryExpression(ctx: JavaParser.TernaryExpressionContext): Expression = {
    val cond = visitExpression(ctx.expression(0))
    val thenExpr = visitExpression(ctx.expression(1))
    val elseExpr = visitExpression(ctx.expression(2))
    val tern = TernaryExpression(cond, thenExpr, elseExpr, getSourceLocation(ctx))
    tern.questionmark.precedingWhitespace = getWhitespaceLeft(ctx.QUESTION())
    tern.colon.precedingWhitespace = getWhitespaceLeft(ctx.COLON())
    tern
  }

  override def visitLamExpression(ctx: JavaParser.LamExpressionContext): Expression =
    visitLambdaExpression(ctx.lambdaExpression())

  override def visitMethodRefExpression(ctx: JavaParser.MethodRefExpressionContext): Expression = {
    if (ctx.expression() != null) {
      val left = visitExpression(ctx.expression())
      var typeArgumentClause: Option[TypeArgumentClause] = None
      typeArgumentClause = Option(ctx.typeArguments()).map(visitTypeArguments)
      val right = visitJavaIdentifier(ctx.javaIdentifier())
      val expr = MethodReferenceExpression(left, typeArgumentClause, right, getSourceLocation(ctx))
      expr.doubleColon.precedingWhitespace = getWhitespaceLeft(ctx.COLONCOLON())
      expr
    } else if (ctx.typeType() != null) {
      val left = visitTypeType(ctx.typeType())
      var typeArgumentClause: Option[TypeArgumentClause] = None
      typeArgumentClause = Option(ctx.typeArguments()).map(visitTypeArguments)
      val right = if (ctx.javaIdentifier()!=null) visitJavaIdentifier(ctx.javaIdentifier())
      else StringNode(ctx.NEW().getText, getSourceLocation(ctx.NEW()), getWhitespaceLeft(ctx.NEW()))
      val expr = MethodReferenceExpression(left, typeArgumentClause, right, getSourceLocation(ctx))
      expr.doubleColon.precedingWhitespace = getWhitespaceLeft(ctx.COLONCOLON())
      expr
    } else {
      val text = printer.visitSkipFirstWS(ctx.classType())
      val left = new StringTypeIndication(text, Nil, ctx.classType().getText,
                                          getSourceLocation(ctx.classType()), getWhitespaceLeft(ctx))
      var typeArgumentClause: Option[TypeArgumentClause] = None
      typeArgumentClause = Option(ctx.typeArguments()).map(visitTypeArguments)
      val right = StringNode(ctx.NEW().getText, getSourceLocation(ctx.NEW()), getWhitespaceLeft(ctx.NEW()))
      val expr = MethodReferenceExpression(left, typeArgumentClause, right, getSourceLocation(ctx))
      expr.doubleColon.precedingWhitespace = getWhitespaceLeft(ctx.COLONCOLON())
      expr
    }
  }

  override def visitLambdaExpression(ctx: JavaParser.LambdaExpressionContext): Expression = {
    val params = visitLambdaParametersAsList(ctx.lambdaParameters())
    val hasParentheses = (ctx.lambdaParameters().LPAREN() != null)
    val body = visitLambdaBody(ctx.lambdaBody())
    val expr = LambdaExpression(hasParentheses, params, body, getSourceLocation(ctx))
    if (hasParentheses) {
      expr.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.lambdaParameters().LPAREN())
      expr.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.lambdaParameters().RPAREN())
    }
    expr.arrow.precedingWhitespace = getWhitespaceLeft(ctx.ARROW())
    expr
  }

  def visitLambdaParametersAsList(ctx: JavaParser.LambdaParametersContext): List[Parameter] = {
    ctx.altNum match {
      case 1 =>
        val name = visitJavaIdentifier(ctx.javaIdentifier(0))
        val noType = new NoType(getSourceLocation(ctx))
        List(Parameter(noType, name, getSourceLocation(ctx)))
      case 2 =>
        if (ctx.formalParameterList() != null) {
          visitFormalParameterListAsList(ctx.formalParameterList())
        } else {
          Nil
        }
      case 3 =>
        var list: List[Parameter] = Nil
        val noType = new NoType(getSourceLocation(ctx))
        for (ident <- ctx.javaIdentifier().asScala) {
          val name = visitJavaIdentifier(ident)
          val param = Parameter(noType, name, getSourceLocation(ident))
          param.trailingWhitespace = getWhitespaceRight(ident)
          list = list.appended(param)
        }
        list.last.trailingWhitespace = ""
        list
    }
  }

  override def visitLambdaBody(ctx: JavaParser.LambdaBodyContext): LambdaBody = {
    ctx.altNum match {
      case 1 =>
        visitExpression(ctx.expression())
      case 2 =>
        visitBlock(ctx.block())
    }
  }

  override def visitPrimary(ctx: JavaParser.PrimaryContext): Expression = {
    ctx.altNum match {
      case 1 =>
        val expr = ParExpression(visitExpression(ctx.expression()), getSourceLocation(ctx))
        expr.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.LPAREN())
        expr.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.RPAREN())
        expr
      case 2 =>
        val name = StringNode(ctx.THIS().getText, getSourceLocation(ctx))
        Literal(name, getSourceLocation(ctx), getWhitespaceLeft(ctx.THIS()))
      case 3 =>
        val name = StringNode(ctx.SUPER().getText, getSourceLocation(ctx))
        Literal(name, getSourceLocation(ctx), getWhitespaceLeft(ctx.SUPER()))
      case 4 =>
        val name = StringNode(ctx.literal().getText, getSourceLocation(ctx))
        Literal(name, getSourceLocation(ctx), getWhitespaceLeft(ctx.literal()))
      case 5 =>
        visitJavaIdentifier(ctx.javaIdentifier())
      case 6 =>
        val typeInd = visitTypeTypeOrVoid(ctx.typeTypeOrVoid())
        val expr = ClassName(typeInd, getSourceLocation(ctx))
        expr.dot.precedingWhitespace = getWhitespaceLeft(ctx.DOT())
        expr.classKeyword.precedingWhitespace = getWhitespaceLeft(ctx.CLASS())
        expr
      case 7 =>
        val typeParam = visitNonWildcardTypeArguments(ctx.nonWildcardTypeArguments())
        if (ctx.explicitGenericInvocationSuffix() != null) {
          if (ctx.explicitGenericInvocationSuffix().SUPER() != null) {
            if (ctx.explicitGenericInvocationSuffix().superSuffix().javaIdentifier() != null) {
              var text = ctx.explicitGenericInvocationSuffix().SUPER().getText
              text += getWhitespaceLeft(ctx.explicitGenericInvocationSuffix().superSuffix().DOT())
              text += ctx.explicitGenericInvocationSuffix().superSuffix().DOT().getText
              text += getWhitespaceLeft(ctx.explicitGenericInvocationSuffix().superSuffix().javaIdentifier())
              text += ctx.explicitGenericInvocationSuffix().superSuffix().javaIdentifier().getText
              val name = StringNode(text, getSourceLocation(ctx), getWhitespaceLeft(ctx))
              val arguments = if (ctx.explicitGenericInvocationSuffix().superSuffix().arguments()!=null)
                                Some(visitArguments(ctx.explicitGenericInvocationSuffix()
                                      .superSuffix().arguments()))
                              else None
              GenericInvocationExpression(typeParam, name, arguments, getSourceLocation(ctx))

            } else {
              val sup = ctx.explicitGenericInvocationSuffix().SUPER()
              val name = StringNode(sup.getText, getSourceLocation(sup), getWhitespaceLeft(sup))
              val arguments = visitArguments(ctx.explicitGenericInvocationSuffix().superSuffix().arguments())
              GenericInvocationExpression(typeParam, name, Some(arguments), getSourceLocation(ctx))
            }
          } else {
            val name = visitJavaIdentifier(ctx.explicitGenericInvocationSuffix().javaIdentifier())
            val arguments = visitArguments(ctx.explicitGenericInvocationSuffix().arguments())
            GenericInvocationExpression(typeParam, name, Some(arguments), getSourceLocation(ctx))
          }
        } else {
          val name = StringNode(ctx.THIS().getText, getSourceLocation(ctx.THIS()),
                                getWhitespaceLeft(ctx.THIS()))
          val arguments = visitArguments(ctx.arguments())
          GenericInvocationExpression(typeParam, name, Some(arguments), getSourceLocation(ctx))
        }
    }
  }

  override def visitCreator(ctx: JavaParser.CreatorContext): NewExpression = {
    val typeIndication = visitCreatedName(ctx.createdName())
    if (ctx.arrayCreatorRest() != null) {
      var dimensions: List[Dimension] = Nil
      for (i <- 0 until ctx.arrayCreatorRest().LBRACK().size()) {
        var size: Option[Expression] = None
        if (i < ctx.arrayCreatorRest().expression().size()) {
          size = Some(visitExpression(ctx.arrayCreatorRest().expression(i)))
        }
        val dim = Dimension(size, getSourceLocation(ctx.arrayCreatorRest().LBRACK(i)))
        dim.openBracket.precedingWhitespace = getWhitespaceLeft(ctx.arrayCreatorRest().LBRACK(i))
        dim.closeBracket.precedingWhitespace = getWhitespaceLeft(ctx.arrayCreatorRest().RBRACK(i))
        dimensions = dimensions.appended(dim)
      }
      val initializer = if(ctx.arrayCreatorRest().arrayInitializer() != null)
                          Some(visitArrayInitializer(ctx.arrayCreatorRest().arrayInitializer()))
                        else None
      NewArrayExpression(typeIndication, dimensions, initializer, getSourceLocation(ctx))
    } else if (ctx.setComprehension() != null) {
      val set = visitSetComprehension(ctx.setComprehension())
      set.typeIndication = typeIndication
      set
    } else {
      val typeArgumentClause = Option(ctx.nonWildcardTypeArguments()).map(visitNonWildcardTypeArguments)
      val args = visitArguments(ctx.classCreatorRest().arguments())
      val body = Option(ctx.classCreatorRest().classBody()).map(visitClassBody)
      NewObjectExpression(typeArgumentClause, typeIndication, args, body, getSourceLocation(ctx))
    }
  }

  override def visitCreatedName(ctx: JavaParser.CreatedNameContext): TypeIndication = {
    val text = printer.visitSkipFirstWS(ctx)
    new StringTypeIndication(text, Nil, ctx.getText, getSourceLocation(ctx), getWhitespaceLeft(ctx))
  }

  override def visitInnerCreator(ctx: JavaParser.InnerCreatorContext): NewObjectExpression = {
    var fulltext = printer.visit(ctx.javaIdentifier())
    var text = ctx.javaIdentifier().getText
    if (ctx.nonWildcardTypeArgumentsOrDiamond() != null) {
      fulltext += printer.visit(ctx.nonWildcardTypeArgumentsOrDiamond())
      text += ctx.nonWildcardTypeArgumentsOrDiamond().getText
    }
    val typeIndication = new StringTypeIndication(fulltext, Nil, text,
                                                  getSourceLocation(ctx.javaIdentifier()))
    val args = visitArguments(ctx.classCreatorRest().arguments())
    val body = Option(ctx.classCreatorRest().classBody()).map(visitClassBody)
    NewObjectExpression(None, typeIndication, args, body, getSourceLocation(ctx))
  }

  override def visitExplicitGenericInvocation(ctx: JavaParser.ExplicitGenericInvocationContext):
                  GenericInvocationExpression = {
    val typeParam = visitNonWildcardTypeArguments(ctx.nonWildcardTypeArguments())
    if (ctx.explicitGenericInvocationSuffix().SUPER() != null) {
      if (ctx.explicitGenericInvocationSuffix().superSuffix().javaIdentifier() != null) {
        var text = ctx.explicitGenericInvocationSuffix().SUPER().getText
        text += getWhitespaceLeft(ctx.explicitGenericInvocationSuffix().superSuffix().DOT())
        text += ctx.explicitGenericInvocationSuffix().superSuffix().DOT().getText
        text += getWhitespaceLeft(ctx.explicitGenericInvocationSuffix().superSuffix().javaIdentifier())
        text += ctx.explicitGenericInvocationSuffix().superSuffix().javaIdentifier().getText
        val name = StringNode(text, getSourceLocation(ctx), getWhitespaceLeft(ctx))
        val arguments = if (ctx.explicitGenericInvocationSuffix().superSuffix().arguments()!=null)
          Some(visitArguments(ctx.explicitGenericInvocationSuffix().superSuffix().arguments()))
        else None
        GenericInvocationExpression(typeParam, name, arguments, getSourceLocation(ctx))
      } else {
        val sup = ctx.explicitGenericInvocationSuffix().SUPER()
        val name = StringNode(sup.getText, getSourceLocation(sup), getWhitespaceLeft(sup))
        val arguments = visitArguments(ctx.explicitGenericInvocationSuffix().superSuffix().arguments())
        GenericInvocationExpression(typeParam, name, Some(arguments), getSourceLocation(ctx))
      }
    } else {
      val name = visitJavaIdentifier(ctx.explicitGenericInvocationSuffix().javaIdentifier())
      val arguments = visitArguments(ctx.explicitGenericInvocationSuffix().arguments())
      GenericInvocationExpression(typeParam, name, Some(arguments), getSourceLocation(ctx))
    }
  }

  override def visitNonWildcardTypeArguments(ctx: JavaParser.NonWildcardTypeArgumentsContext):
                TypeArgumentClause = {
    val parameters = ctx.typeList().typeType().asScala
    val types = parameters.map(visitTypeTypeWithWS).toList
    val clause = TypeArgumentClause(types, getSourceLocation(ctx))
    clause.openAngle.precedingWhitespace = getWhitespaceLeft(ctx.LT())
    clause.closeAngle.precedingWhitespace = getWhitespaceLeft(ctx.GT())
    // WS before closeAngle already covered by trailing WS of last typeParameter
    types.last.trailingWhitespace = ""
    clause
  }

  override def visitTypeArguments(ctx: JavaParser.TypeArgumentsContext): TypeArgumentClause = {
    val args = ctx.typeArgument().asScala.map(x=>addWhitespaceRight(visitTypeArgument(x),x)).toList
    val clause = TypeArgumentClause(args, getSourceLocation(ctx))
    clause.openAngle.precedingWhitespace = getWhitespaceLeft(ctx.LT())
    clause.closeAngle.precedingWhitespace = getWhitespaceLeft(ctx.GT())
    clause
  }

  override def visitArguments(ctx: JavaParser.ArgumentsContext): Arguments = {
    val args = if (ctx.expressionList() == null) Nil
               else ctx.expressionList().expression().asScala
                      .map(x=>addWhitespaceRight(visitExpression(x),x)).toList
    val arguments = Arguments(args, getSourceLocation(ctx))
    arguments.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.LPAREN())
    arguments.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.RPAREN())
    arguments
  }


  override def visitLangStatement(ctx: JavaParser.LangStatementContext): Statement = {
    visitBlockStatement(ctx.blockStatement())
  }

  override def visitValEmbedDeclarationBlock(ctx: JavaParser.ValEmbedDeclarationBlockContext):
                SpecDeclarationBlock = {
    val decls = ctx.valDeclaration().asScala.map(visitValDeclaration).toList
    val startSpec = StringNode(ctx.startSpec().getText,
                               getSourceLocation(ctx.startSpec()), getWhitespaceLeft(ctx.startSpec()))
    val endSpec = StringNode(ctx.endSpec().getText,
                             getSourceLocation(ctx.endSpec()), getWhitespaceLeft(ctx.endSpec()))
    SpecDeclarationBlock(Some(startSpec), decls, Some(endSpec), getSourceLocation(ctx))
  }

  override def visitValEmbedStatementBlock(ctx: JavaParser.ValEmbedStatementBlockContext): JmlStatementBlock = {
    val statements = ctx.valOrLangStatement().asScala.map(visitValOrLangStatement).toList
    var startSpec: Option[StringNode] = None
    var endSpec: Option[StringNode] = None
    if (ctx.startSpec() != null) {
      startSpec = Some(StringNode(ctx.startSpec().getText,
                                  getSourceLocation(ctx.startSpec()), getWhitespaceLeft(ctx.startSpec())))
      endSpec = Some(StringNode(ctx.endSpec().getText,
                                getSourceLocation(ctx.endSpec()), getWhitespaceLeft(ctx.endSpec())))
    }
    JmlStatementBlock(startSpec, statements, endSpec, getSourceLocation(ctx))
  }

  override def visitValOrLangStatement(ctx: JavaParser.ValOrLangStatementContext): Statement = {
    if(ctx.valStatement != null){
      visitValStatement(ctx.valStatement)
    } else {
      visitLangStatement(ctx.langStatement)
    }
  }

  override def visitValModifier(ctx: JavaParser.ValModifierContext): Modifier = {
    if (ctx.langModifier() != null) {
      visitClassOrInterfaceModifier(ctx.langModifier().classOrInterfaceModifier())
    }
    else {
      visitValSpecModifier(ctx.valSpecModifier())
    }
  }

  override def visitValSpecModifier(ctx: JavaParser.ValSpecModifierContext): Modifier = {
    if (ctx.VAL_PURE() != null){
      PureModifier(getSourceLocation(ctx), getWhitespaceLeft(ctx))
    }
    else if (ctx.NULLABLE() != null){
      NullableModifier(getSourceLocation(ctx), getWhitespaceLeft(ctx))
    }
    else if (ctx.NULLABLE_BY_DEFAULT() != null){
      NullableByDefaultModifier(getSourceLocation(ctx), getWhitespaceLeft(ctx))
    }
    else if (ctx.NON_NULL() != null){
      NonnullModifier(getSourceLocation(ctx), getWhitespaceLeft(ctx))
    }
    else if (ctx.VAL_GHOST() != null){
      GhostModifier(getSourceLocation(ctx), getWhitespaceLeft(ctx))
    }
    else if (ctx.HELPER() != null){
      HelperModifier(getSourceLocation(ctx), getWhitespaceLeft(ctx))
    }
    else if (ctx.MONITORED() != null){
      MonitoredModifier(getSourceLocation(ctx), getWhitespaceLeft(ctx))
    }
    else if (ctx.MODEL() != null){
      ModelModifier(getSourceLocation(ctx), getWhitespaceLeft(ctx))
    }
    else if (ctx.INSTANCE() != null){
      InstanceModifier(getSourceLocation(ctx), getWhitespaceLeft(ctx))
    }
    else if (ctx.SPEC_PUBLIC() != null){
      new SpecPublicModifier(getSourceLocation(ctx), getWhitespaceLeft(ctx))
    }
    else if (ctx.SPEC_PROTECTED() != null){
      new SpecProtectedModifier(getSourceLocation(ctx), getWhitespaceLeft(ctx))
    }
    else if (ctx.mathModifier() != null) {
      System.err.println("Math modifiers not supported, dropped.")
      ModifierStringNode("", getSourceLocation(ctx), getWhitespaceLeft(ctx.mathModifier()))
    }
    else {
      // todo
      null
    }
  }

  override def visitValEmbedContract(ctx: JavaParser.ValEmbedContractContext): Contract = {
    val specCases = ctx.valEmbedContractBlock().asScala.flatMap(visitValEmbedContractBlockAsList)
    Contract(groupSpecCases(specCases.toList), getSourceLocation(ctx))
  }

  def visitValEmbedContractBlockAsList(ctx: JavaParser.ValEmbedContractBlockContext): List[SpecCase] = {
    if(ctx.valLinewiseContractBlock() != null) {
      visitValLinewiseContractBlockAsList(ctx.valLinewiseContractBlock())
    } else {
      val cases = ctx.valSpecCase().asScala.map(visitValSpecCase)
      var offset = 1
      if (ctx.VAL_ALSO().size() == ctx.valSpecCase().size()) {
        offset = 0
        cases.head.hasAlso = true
        cases.head.alsoKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_ALSO().asScala.head)
      }
      for (i <- 1 until cases.size) {
        cases(i).hasAlso = true
        cases(i).alsoKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_ALSO(i - offset))
      }
      if (ctx.startSpec() != null) {
        cases.head.startSpec = Some(StringNode(ctx.startSpec().getText,
                                              getSourceLocation(ctx.startSpec()),
                                              getWhitespaceLeft(ctx.startSpec())))
        cases.last.endSpec = Some(StringNode(ctx.endSpec().getText,
                                            getSourceLocation(ctx.endSpec()),
                                            getWhitespaceLeft(ctx.endSpec())))
      }
      cases.toList
    }
  }

  def visitValLinewiseContractBlockAsList(ctx: JavaParser.ValLinewiseContractBlockContext): List[SpecCase] = {
    val res = ctx.valLinewiseAdditionalSpecCase().asScala.map(visitValLinewiseAdditionalSpecCase)
    if(ctx.valLinewiseSpecCase() != null) {
      val head = visitValLinewiseSpecCase(ctx.valLinewiseSpecCase())
      if(ctx.valEmbedAlso() != null) {
        head.hasAlso = true
        head.alsoKeyword.precedingWhitespace = getWhitespaceLeft(ctx.valEmbedAlso().startSpec())
                                              + ctx.valEmbedAlso().startSpec().getText
                                              + getWhitespaceLeft(ctx.valEmbedAlso().VAL_ALSO())
        head.alsoKeyword.trailingWhitespace = getWhitespaceLeft(ctx.valEmbedAlso().endSpec())
                                              + ctx.valEmbedAlso().endSpec().getText
      }
      res.prepended(head).toList
    } else {
      val head = visitValLinewiseAlsoBehaviorClause(ctx.valLinewiseAlsoBehaviorClause())
      res.prepended(head).toList
    }
  }

  def groupSpecCases(specCases: List[SpecCase]): List[SpecCase] = {
    val collected: ListBuffer[List[SpecCase]] = ListBuffer()
    var current: ListBuffer[SpecCase] = ListBuffer()
    for(spec:SpecCase <- specCases) {
      if(spec.isInstanceOf[Behavior] || spec.hasAlso) {
        // skip empty group (in case "also"/"behavior" is the first item encountered, then the first group is empty)
        if(current.nonEmpty){
          collected.append(current.toList)
        }
        current = ListBuffer(spec)
      } else {
        current.append(spec)
      }
    }
    collected.append(current.toList)

    current = ListBuffer()
    for(group: List[SpecCase] <- collected) {
      if(group.size > 1) {
        ???
      }
      val head: SpecCase = group.head
      var preceding = ""
        head match {
          case b: Behavior =>
            if (b.privacy.isDefined) {
              b.privacy.get.precedingWhitespace = preceding + b.privacy.get.precedingWhitespace
            } else {
              b.behaviorKeyword.precedingWhitespace = preceding + b.behaviorKeyword.precedingWhitespace
            }
            preceding = ""
          case _ =>
        }
        if(head.clauses.nonEmpty) {
          head.clauses.head.precedingWhitespace = preceding + head.clauses.head.precedingWhitespace
          preceding = ""
        }
//      preceding += head.endSpec.getOrElse("").toString + head.trailingWhitespace
      for (spec: SpecCase <- group.tail) {
        preceding += spec.precedingWhitespace
        if (spec.clauses.nonEmpty) {
          spec.clauses.head.startSpec = spec.startSpec
          spec.clauses.head.precedingWhitespace = preceding + spec.clauses.head.precedingWhitespace
          preceding = ""
          spec.clauses.last.endSpec = spec.endSpec
        } else {
          preceding += spec.startSpec.getOrElse("").toString
          preceding += spec.endSpec.getOrElse("").toString
        }
        preceding += spec.trailingWhitespace
        head.clauses = head.clauses.appendedAll(spec.clauses)
      }

//      head.endSpec = None
//      head.trailingWhitespace = preceding
      current.append(head)
    }
    current.toList
  }

  override def visitValEmbedContractStart(ctx: JavaParser.ValEmbedContractStartContext): Contract = {
    val specCases = ctx.valEmbedContractBlock().asScala.flatMap(visitValEmbedContractBlockAsList)
    specCases.head.startSpec = Some(StringNode(ctx.startSpec().getText,
                                    getSourceLocation(ctx.startSpec()), getWhitespaceLeft(ctx.startSpec())))

    Contract(groupSpecCases(specCases.toList), getSourceLocation(ctx))
  }



  override def visitValSpecCase(ctx: JavaParser.ValSpecCaseContext): SpecCase = {
    if(ctx.valBehaviorClause()!=null) {
      visitValBehaviorClause(ctx.valBehaviorClause())
    } else {
      val clauses = ctx.valContractClause().asScala.map(visitValContractClause).toList
      SpecCase(None, hasAlso = false, clauses, None, getSourceLocation(ctx))
    }
  }

  override def visitValLinewiseSpecCase(ctx: JavaParser.ValLinewiseSpecCaseContext): SpecCase = {
    if(ctx.valLinewiseBehaviorClause()!=null) {
      visitValLinewiseBehaviorClause(ctx.valLinewiseBehaviorClause())
    } else {
      val clauses: ListBuffer[ContractClause] = ListBuffer()
      for(i <- ctx.startSpec().asScala.indices) {
        val clause = visitValContractClause(ctx.valContractClause(i))
        clause.startSpec = Some(StringNode(ctx.startSpec(i).getText,
                                            getSourceLocation(ctx.startSpec(i)), 
                                            getWhitespaceLeft(ctx.startSpec(i))))
        clause.endSpec = Some(StringNode(ctx.endSpec(i).getText,
                                          getSourceLocation(ctx.endSpec(i)), 
                                          getWhitespaceLeft(ctx.endSpec(i))))
        clauses.append(clause)
      }
      SpecCase(None, hasAlso = false, clauses.toList, None, getSourceLocation(ctx))
    }
  }

  override def visitValLinewiseAdditionalSpecCase(ctx: JavaParser.ValLinewiseAdditionalSpecCaseContext): SpecCase = {
    if (ctx.valLinewiseAlsoBehaviorClause() != null) {
      visitValLinewiseAlsoBehaviorClause(ctx.valLinewiseAlsoBehaviorClause())
    } else {
      val specCase = visitValLinewiseSpecCase(ctx.valLinewiseSpecCase())
      specCase.hasAlso = true
      specCase.alsoKeyword.precedingWhitespace = getWhitespaceLeft(ctx.valEmbedAlso().startSpec())
                                                + ctx.valEmbedAlso().startSpec().getText 
                                                + getWhitespaceLeft(ctx.valEmbedAlso().VAL_ALSO())
      specCase.alsoKeyword.trailingWhitespace = getWhitespaceLeft(ctx.valEmbedAlso().endSpec())
                                                + ctx.valEmbedAlso().endSpec().getText 
      specCase
    }
  }
  override def visitPrivacy(ctx: JavaParser.PrivacyContext): StringNode = {
    StringNode(ctx.getText, getSourceLocation(ctx), getWhitespaceLeft(ctx))
  }

  override def visitValBehaviorClause(ctx: JavaParser.ValBehaviorClauseContext): Behavior = {
    var privacy: Option[StringNode] = None
    privacy = Option(ctx.privacy()).map(visitPrivacy)
    val clauses = ctx.valContractClause().asScala.map(visitValContractClause).toList
    if (ctx.VAL_BEHAVIOR() != null) {
      val behavior = new Behavior(None, false, privacy, clauses, None, getSourceLocation(ctx))
      behavior.behaviorKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_BEHAVIOR())
      behavior
    } else if (ctx.VAL_BEHAVIOUR() != null) {
      val behavior = new Behavior(None, false, privacy, clauses, None, getSourceLocation(ctx))
      behavior.behaviorKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_BEHAVIOUR())
      behavior
    } else if (ctx.VAL_NORMAL_BEHAVIOR() != null) {
      val behavior = new NormalBehavior(None, false, privacy, clauses, None, getSourceLocation(ctx))
      behavior.behaviorKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_NORMAL_BEHAVIOR())
      behavior
    } else if (ctx.VAL_NORMAL_BEHAVIOUR() != null) {
      val behavior = new NormalBehavior(None, false, privacy, clauses, None, getSourceLocation(ctx))
      behavior.behaviorKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_NORMAL_BEHAVIOUR())
      behavior
    } else if (ctx.VAL_EXCEPTIONAL_BEHAVIOR() != null) {
      val behavior = new ExceptionalBehavior(None, false, privacy, clauses, None, getSourceLocation(ctx))
      behavior.behaviorKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_EXCEPTIONAL_BEHAVIOR())
      behavior
    } else if (ctx.VAL_EXCEPTIONAL_BEHAVIOUR() != null) {
      val behavior = new ExceptionalBehavior(None, false, privacy, clauses, None, getSourceLocation(ctx))
      behavior.behaviorKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_EXCEPTIONAL_BEHAVIOUR())
      behavior
    } else {
      ???
    }
  }
  override def visitValLinewiseBehaviorClause(ctx: JavaParser.ValLinewiseBehaviorClauseContext): Behavior = {
    var privacy: Option[StringNode] = None
    if (ctx.privacy() != null) {
      privacy = Some(visitPrivacy(ctx.privacy()))
      privacy.get.precedingWhitespace = getWhitespaceLeft(ctx.startSpec(0)) + ctx.startSpec(0).getText 
                                    + privacy.get.precedingWhitespace
    }
    val clauses: ListBuffer[ContractClause] = ListBuffer()
    for(i <- ctx.valContractClause().asScala.indices) {
      val clause = visitValContractClause(ctx.valContractClause(i))
      clause.startSpec = Some(StringNode(ctx.startSpec(i+1).getText,
                                        getSourceLocation(ctx.startSpec(i+1)),
                                        getWhitespaceLeft(ctx.startSpec(i+1))))
      clause.endSpec = Some(StringNode(ctx.endSpec(i+1).getText,
                                        getSourceLocation(ctx.endSpec(i+1)),
                                        getWhitespaceLeft(ctx.endSpec(i+1))))
      clauses.append(clause)
    }
    var behavior: Behavior = null
    if (ctx.VAL_BEHAVIOR() != null) {
      behavior = new Behavior(None, false, privacy, clauses.toList, None, getSourceLocation(ctx))
      behavior.behaviorKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_BEHAVIOR())
    } else if (ctx.VAL_BEHAVIOUR() != null) {
      behavior = new Behavior(None, false, privacy, clauses.toList, None, getSourceLocation(ctx))
      behavior.behaviorKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_BEHAVIOUR())
    } else if (ctx.VAL_NORMAL_BEHAVIOR() != null) {
      behavior = new NormalBehavior(None, false, privacy, clauses.toList, None, getSourceLocation(ctx))
      behavior.behaviorKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_NORMAL_BEHAVIOR())
    } else if (ctx.VAL_NORMAL_BEHAVIOUR() != null) {
      behavior = new NormalBehavior(None, false, privacy, clauses.toList, None, getSourceLocation(ctx))
      behavior.behaviorKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_NORMAL_BEHAVIOUR())
    } else if (ctx.VAL_EXCEPTIONAL_BEHAVIOR() != null) {
      behavior = new ExceptionalBehavior(None, false, privacy, clauses.toList, None, getSourceLocation(ctx))
      behavior.behaviorKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_EXCEPTIONAL_BEHAVIOR())
    } else if (ctx.VAL_EXCEPTIONAL_BEHAVIOUR() != null) {
      behavior = new ExceptionalBehavior(None, false, privacy, clauses.toList, None, getSourceLocation(ctx))
      behavior.behaviorKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_EXCEPTIONAL_BEHAVIOUR())
    } else {
      ???
    }
    behavior.behaviorKeyword.trailingWhitespace = getWhitespaceLeft(ctx.endSpec(0)) + ctx.endSpec(0).getText
    if(ctx.privacy() == null) {
      behavior.behaviorKeyword.precedingWhitespace = getWhitespaceLeft(ctx.startSpec(0)) + ctx.startSpec(0).getText
    }
    behavior
  }
  override def visitValLinewiseAlsoBehaviorClause(ctx: JavaParser.ValLinewiseAlsoBehaviorClauseContext): Behavior = {
    var privacy: Option[StringNode] = None
    privacy = Option(ctx.privacy()).map(visitPrivacy)
    val clauses: ListBuffer[ContractClause] = ListBuffer()
    for(i <- ctx.valContractClause().asScala.indices) {
      val clause = visitValContractClause(ctx.valContractClause(i))
      clause.startSpec = Some(StringNode(ctx.startSpec(i+1).getText,
                                        getSourceLocation(ctx.startSpec(i+1)),
                                        getWhitespaceLeft(ctx.startSpec(i+1))))
      clause.endSpec = Some(StringNode(ctx.endSpec(i+1).getText,
                                        getSourceLocation(ctx.endSpec(i+1)),
                                        getWhitespaceLeft(ctx.endSpec(i+1))))
      clauses.append(clause)
    }
    var behavior: Behavior = null
    if (ctx.VAL_BEHAVIOR() != null) {
      behavior = new Behavior(None, false, privacy, clauses.toList, None, getSourceLocation(ctx))
      behavior.behaviorKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_BEHAVIOR())
    } else if (ctx.VAL_BEHAVIOUR() != null) {
      behavior = new Behavior(None, false, privacy, clauses.toList, None, getSourceLocation(ctx))
      behavior.behaviorKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_BEHAVIOUR())
    } else if (ctx.VAL_NORMAL_BEHAVIOR() != null) {
      behavior = new NormalBehavior(None, false, privacy, clauses.toList, None, getSourceLocation(ctx))
      behavior.behaviorKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_NORMAL_BEHAVIOR())
    } else if (ctx.VAL_NORMAL_BEHAVIOUR() != null) {
      behavior = new NormalBehavior(None, false, privacy, clauses.toList, None, getSourceLocation(ctx))
      behavior.behaviorKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_NORMAL_BEHAVIOUR())
    } else if (ctx.VAL_EXCEPTIONAL_BEHAVIOR() != null) {
      behavior = new ExceptionalBehavior(None, false, privacy, clauses.toList, None, getSourceLocation(ctx))
      behavior.behaviorKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_EXCEPTIONAL_BEHAVIOR())
    } else if (ctx.VAL_EXCEPTIONAL_BEHAVIOUR() != null) {
      behavior = new ExceptionalBehavior(None, false, privacy, clauses.toList, None, getSourceLocation(ctx))
      behavior.behaviorKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_EXCEPTIONAL_BEHAVIOUR())
    } else {
      ???
    }
    behavior.hasAlso = true
    behavior.alsoKeyword.precedingWhitespace = getWhitespaceLeft(ctx.startSpec(0)) + ctx.startSpec(0).getText
                                                + getWhitespaceLeft(ctx.VAL_ALSO())
    behavior.behaviorKeyword.trailingWhitespace = getWhitespaceLeft(ctx.endSpec(0)) + ctx.endSpec(0).getText
    behavior
  }

  override def visitValContractClause(ctx: JavaParser.ValContractClauseContext): ContractClause = {
    ctx.altNum match {
      case 1 =>
        val expr = visitStoreRefListAsList(ctx.storeRefList)
        val clause = AssignableClause(expr, getSourceLocation(ctx))
        clause.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
        if (ctx.VAL_MODIFIES() != null) {
          clause.assignableKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_MODIFIES())
        } else if (ctx.VAL_MODIFIABLE() != null) {
          clause.assignableKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_MODIFIABLE())
        } else if (ctx.VAL_ASSIGNABLE() != null) {
          clause.assignableKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_ASSIGNABLE())
        } else if (ctx.VAL_ASSIGNS() != null) {
          clause.assignableKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_ASSIGNS())
        }
        clause
      case 2 =>
        val expr = visitStoreRefListAsList(ctx.storeRefList)
        val clause = AccessibleClause(expr, getSourceLocation(ctx))
        clause.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
        clause.accessibleKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_ACCESSIBLE())
        clause
      case 3 =>
        val typeIndication = visitTypeSpec(ctx.typeSpec())
        val name = visitQuantifiedVarDeclarator(ctx.quantifiedVarDeclarator())
        var declarator: QuantifiedVarDeclaration = null
        if (ctx.NULLABLE() != null) {
          val decl = new NullableQuantifiedVarDeclaration(typeIndication, List(name), getSourceLocation(ctx))
          decl.nullableKeyword.precedingWhitespace = getWhitespaceLeft(ctx.NULLABLE())
          declarator = decl
        } else if (ctx.NON_NULL() != null) {
          val decl = new NonNullQuantifiedVarDeclaration(typeIndication, List(name), getSourceLocation(ctx))
          decl.nonNullKeyword.precedingWhitespace = getWhitespaceLeft(ctx.NON_NULL())
          declarator = decl
        } else {
          declarator = new QuantifiedVarDeclaration(typeIndication, List(name), getSourceLocation(ctx))
        }
        val forall = ForallVariableDeclaration(declarator, getSourceLocation(ctx))
        forall.forallKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_FORALL())
        forall.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
        forall
      case 4 =>
        var modifiers: List[Modifier] = Nil
        if (ctx.NULLABLE() != null) {
          val mod = new ModifierStringNode(ctx.NULLABLE().getText,
                                           getSourceLocation(ctx.NULLABLE()),
                                           getWhitespaceLeft(ctx.NULLABLE()))
          modifiers = modifiers.appended(mod)
        } else if (ctx.NON_NULL() != null) {
          val mod = new ModifierStringNode(ctx.NON_NULL().getText,
                                           getSourceLocation(ctx.NON_NULL()),
                                           getWhitespaceLeft(ctx.NON_NULL()))
          modifiers = modifiers.appended(mod)
        }
        val typeIndication = visitTypeSpec(ctx.typeSpec())
        val declarators = ctx.specVariableDeclarator().asScala
                            .map(x=>addWhitespaceRight(visitSpecVariableDeclarator(x),x)).toList
        val old = OldVariableDeclaration(modifiers, typeIndication, declarators, getSourceLocation(ctx))
        old.oldKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_OLD())
        old.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
        old
      case 5 =>
        val expr = visitExpression(ctx.langExpr(0).expression())
        val clause = Precondition(expr, getSourceLocation(ctx))
        clause.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
        if (ctx.VAL_PRE() != null) {
          clause.requiresKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_PRE())
        }  else if (ctx.VAL_REQUIRES() != null) {
          clause.requiresKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_REQUIRES())
        }
        clause
      case 6 =>
        val expr = visitExpression(ctx.langExpr(0).expression())
        val clause = Postcondition(expr, getSourceLocation(ctx))
        clause.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
        if (ctx.VAL_POST() != null) {
          clause.ensuresKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_POST())
        }  else if (ctx.VAL_ENSURES() != null) {
          clause.ensuresKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_ENSURES())
        }
        clause
      case 7 =>
        val expr = visitExpression(ctx.langExpr(0).expression())
        val clause = DivergeClause(expr, getSourceLocation(ctx))
        clause.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
        clause.divergesKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_DIVERGES())
        clause
      case 8 =>
        val exType = visitTypeType(ctx.langType(0).typeType())
        val exId = Option(ctx.langId()).map(x => visitJavaIdentifier(x.javaIdentifier()))
        val expr = Option(ctx.langExpr(0)).map(x => visitExpression(x.expression()))
        val clause = ExceptionalPostcondition(exType, exId, expr, getSourceLocation(ctx))
        clause.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
        clause.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.LPAREN())
        clause.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.RPAREN())
        if (ctx.VAL_SIGNALS() != null) {
          clause.signalsKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_SIGNALS())
        }  else if (ctx.VAL_EXSURES() != null) {
          clause.signalsKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_EXSURES())
        }
        clause
      case 9 =>
        var extypes: List[TypeIndication] = Nil
        if (ctx.langType().size > 0) {
          extypes = ctx.langType().asScala
                      .map(x => addWhitespaceRight(visitTypeType(x.typeType),x)).toList
        }
        val clause = SignalsOnlyClause(extypes, getSourceLocation(ctx))
        clause.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
        clause.signalsOnlyKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_SIGNALS_ONLY())
        if (ctx.NOTHING() != null) {
          clause.nothingKeyword.precedingWhitespace = getWhitespaceLeft(ctx.NOTHING())
        }
        clause
      case 10 =>
        val expr = visitExpression(ctx.langExpr(0).expression())
        var predicate: Option[Expression] = None
        if(ctx.IF() != null) {
          predicate = Some(visitExpression(ctx.langExpr(1).expression()))
        }
        val clause = MeasuredByClause(expr, predicate, getSourceLocation(ctx))
        clause.measuredByKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_MEASURED_BY())
        if(predicate.isDefined) {
          clause.ifKeyword.precedingWhitespace = getWhitespaceLeft(ctx.IF())
        }
        clause.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
        clause
    }
  }

  def visitValEmbedLoopSpecAsList(ctx: JavaParser.ValEmbedLoopSpecContext): List[LoopSpecification] = {
    ctx.valEmbedLoopSpecBlock().asScala.map(visitValEmbedLoopSpecBlock).toList
  }

  override def visitValEmbedLoopSpecBlock(ctx: JavaParser.ValEmbedLoopSpecBlockContext): LoopSpecification = {
    val clauses = ctx.valLoopClause().asScala.map(visitValLoopClause).toList
    val openSpec = Option(ctx.startSpec()).map(x=>StringNode(x.getText, getSourceLocation(x), getWhitespaceLeft(x)))
    val closeSpec = Option(ctx.endSpec()).map(x=>StringNode(x.getText, getSourceLocation(x), getWhitespaceLeft(x)))
    LoopSpecification(openSpec, clauses, closeSpec, getSourceLocation(ctx))
  }

  override def visitValLoopClause(ctx: JavaParser.ValLoopClauseContext): LoopClause = {
    ctx.altNum match {
      case 1 =>
        val expr = visitExpression(ctx.langExpr.expression())
        val clause = LoopInvariant(expr, getSourceLocation(ctx))
        clause.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
        if (ctx.VAL_LOOP_INVARIANT() != null) {
          clause.invariantKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_LOOP_INVARIANT())
        } else if (ctx.VAL_MAINTAINING() != null) {
          clause.invariantKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_MAINTAINING())
        }
        clause
      case 2 =>
        val expr = visitExpression(ctx.langExpr.expression())
        val clause = DecreasesClause(expr, getSourceLocation(ctx))
        clause.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
        if (ctx.VAL_DECREASES() != null) {
          clause.decreasesKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_DECREASES())
        } else if (ctx.VAL_DECREASING() != null) {
          clause.decreasesKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_DECREASING())
        } else if (ctx.VAL_LOOP_DECREASES() != null) {
          clause.decreasesKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_LOOP_DECREASES())
        }
        clause
      case 3 =>
        val expr = ctx.valExpressionList().langExpr().asScala
          .map(x => addWhitespaceRight(visitExpression(x.expression()), x)).toList
        val clause = LoopModifiesClause(expr, getSourceLocation(ctx))
        clause.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
        if(ctx.VAL_LOOP_MODIFIES() != null) {
          clause.loopModifiesKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_LOOP_MODIFIES())
        } else {
          clause.loopModifiesKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_LOOP_ASSIGNS())
        }
        clause
    }
  }

  override def visitTypeSpec(ctx: JavaParser.TypeSpecContext): TypeIndication = {
    var typeInd:TypeIndication = null
    if (ctx.TYPE_TYPE() != null) {
      val text = printer.visitSkipFirstWS(ctx)
      typeInd = new StringTypeIndication(text, Nil, ctx.getText,
                                        getSourceLocation(ctx), getWhitespaceLeft(ctx))
    } else {
      typeInd = visitTypeType(ctx.langType().typeType())
    }
    if (ctx.ownershipModifier().size() > 0) {
      typeInd.modifiers = ctx.ownershipModifier().asScala.map(visitOwnershipModifier).toList
    }
    typeInd
  }

  override def visitOwnershipModifier(ctx: JavaParser.OwnershipModifierContext): Modifier = {
    new ModifierStringNode(ctx.getText, getSourceLocation(ctx), getWhitespaceLeft(ctx))
  }

  override def visitValEmbedModifiers(ctx: JavaParser.ValEmbedModifiersContext): Modifier = {
    var modifiers: List[Modifier] = Nil
    if(ctx.valModifier()!=null) {
      modifiers = modifiers.appendedAll(ctx.valModifier().asScala.map(visitValModifier))
    }
    if(ctx.valSpecModifier() != null) {
      modifiers = modifiers.appendedAll(ctx.valSpecModifier().asScala.map(visitValSpecModifier))
    }
    var startSpec: Option[StringNode] = None
    var endSpec: Option[StringNode] = None
    if (ctx.startSpec() != null) {
      startSpec = Some(StringNode(ctx.startSpec().getText, getSourceLocation(ctx.startSpec()),
                                   getWhitespaceLeft(ctx.startSpec())))
    }
    if (ctx.endSpec() != null) {
      endSpec = Some(StringNode(ctx.endSpec().getText, getSourceLocation(ctx.endSpec()),
                                getWhitespaceLeft(ctx.endSpec())))
    }
    JmlModifierBlock(startSpec, modifiers, endSpec, getSourceLocation(ctx))
  }

  override def visitValEmbedModifiersEnd(ctx: JavaParser.ValEmbedModifiersEndContext): Modifier = {
    val modifiers = ctx.valModifier().asScala.map(visitValModifier).toList
    val endSpec = Some(StringNode(ctx.endSpec().getText,
                                  getSourceLocation(ctx.endSpec()), getWhitespaceLeft(ctx.endSpec())))
    JmlModifierBlock(None, modifiers, endSpec, getSourceLocation(ctx))
  }

  override def visitValStatement(ctx: JavaParser.ValStatementContext): Statement = {
    ctx.altNum match {
      case 1 =>
        val content = visitExpression(ctx.langExpr(0).expression())
        var msg: Option[Expression] = None
        if (ctx.COLON() != null) {
          msg = Some(visitExpression(ctx.langExpr(1).expression()))
        }
        val stmt = AssertStatement(content, msg, getSourceLocation(ctx))
        stmt.assertKeyword.precedingWhitespace = getWhitespaceLeft(ctx.ASSERT())
        stmt.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
        if (ctx.COLON() != null) {
          stmt.colon.precedingWhitespace = getWhitespaceLeft(ctx.COLON())
        }
        stmt
      case 2 =>
        val content = visitExpression(ctx.langExpr(0).expression())
        var msg: Option[Expression] = None
        if (ctx.COLON() != null) {
          msg = Some(visitExpression(ctx.langExpr(1).expression()))
        }
        val stmt = AssumeStatement(content, msg, getSourceLocation(ctx))
        stmt.assumeKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_ASSUME())
        stmt.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
        if (ctx.COLON() != null) {
          stmt.colon.precedingWhitespace = getWhitespaceLeft(ctx.COLON())
        }
        stmt
      case 3 =>
        val assignment = visitExpression(ctx.langExpr(0).expression())
        val stmt = SetStatement(assignment, getSourceLocation(ctx))
        stmt.setKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_SET())
        stmt.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
        stmt
      case 4 =>
        val stmt = UnreachableStatement(getSourceLocation(ctx))
        stmt.unreachableKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_UNREACHABLE())
        stmt
      case 5 =>
        visitValBlock(ctx.valBlock())
    }
  }

  override def visitValBlock(ctx: JavaParser.ValBlockContext): Statement = {
    val statements = ctx.valOrLangStatement().asScala.map(visitValOrLangStatement).toList
    val block = Block(None, statements, getSourceLocation(ctx))
    block.openBrace.precedingWhitespace = getWhitespaceLeft(ctx.LBRACE())
    block.closeBrace.precedingWhitespace = getWhitespaceLeft(ctx.RBRACE())
    block
  }

  override def visitValPrimary(ctx: JavaParser.ValPrimaryContext): Expression = {
    ctx.altNum match {
      case 1 =>
        ResultExpression(getSourceLocation(ctx.RESULT()), getWhitespaceLeft(ctx.RESULT()))
      case 2 =>
        val expr = visitExpression(ctx.langExpr().expression())
        if (ctx.OLD() != null) {
          var label: Option[StringNode] = None
          if(ctx.langId().size() > 0) {
            label = Some(visitJavaIdentifier(ctx.langId(0).javaIdentifier()))
          }
          val old = OldExpression(expr, label, getSourceLocation(ctx))
          old.oldKeyword.precedingWhitespace = getWhitespaceLeft(ctx.OLD())
          old.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.LPAREN())
          old.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.RPAREN())
          if (ctx.COMMA().size() > 0) {
            old.comma.precedingWhitespace = getWhitespaceLeft(ctx.COMMA(0))
          }
          old
        } else {
          val pre = PreExpression(expr, getSourceLocation(ctx))
          pre.keyword.precedingWhitespace = getWhitespaceLeft(ctx.OLD())
          pre.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.LPAREN())
          pre.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.RPAREN())
          pre
        }
      case 3 =>
        val storeRefs = visitStoreRefListAsList(ctx.storeRefList())
        val expr = NotAssignedExpression(storeRefs, getSourceLocation(ctx))
        expr.keyword.precedingWhitespace = getWhitespaceLeft(ctx.NOT_ASSIGNED())
        expr.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.LPAREN())
        expr.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.RPAREN())
        expr
      case 4 =>
        val storeRefs = visitStoreRefListAsList(ctx.storeRefList())
        val expr = NotModifiedExpression(storeRefs, getSourceLocation(ctx))
        expr.keyword.precedingWhitespace = getWhitespaceLeft(ctx.NOT_MODIFIED())
        expr.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.LPAREN())
        expr.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.RPAREN())
        expr
      case 5 =>
        val storeRefs = visitStoreRefListAsList(ctx.storeRefList())
        val expr = OnlyAccessedExpression(storeRefs, getSourceLocation(ctx))
        expr.keyword.precedingWhitespace = getWhitespaceLeft(ctx.ONLY_ACCESSED())
        expr.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.LPAREN())
        expr.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.RPAREN())
        expr
      case 6 =>
        val storeRefs = visitStoreRefListAsList(ctx.storeRefList())
        val expr = OnlyAssignedExpression(storeRefs, getSourceLocation(ctx))
        expr.keyword.precedingWhitespace = getWhitespaceLeft(ctx.ONLY_ASSIGNED())
        expr.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.LPAREN())
        expr.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.RPAREN())
        expr
      case 9 =>
        val expressions = ctx.expressionList().expression().asScala.map(visitExpression).toList
        val expr = FreshExpression(expressions, getSourceLocation(ctx))
        expr.freshKeyword.precedingWhitespace = getWhitespaceLeft(ctx.FRESH())
        expr.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.LPAREN())
        expr.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.RPAREN())
        expr
      case 14 =>
        val langExpr = visitExpression(ctx.langExpr().expression())
        val expr = NonNullElementsExpression(langExpr, getSourceLocation(ctx))
        expr.keyword.precedingWhitespace = getWhitespaceLeft(ctx.NONNULLELEMENTS())
        expr.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.LPAREN())
        expr.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.RPAREN())
        expr
      case 15 =>
        val langExpr = visitExpression(ctx.langExpr().expression())
        val expr = TypeOfExpression(langExpr, getSourceLocation(ctx))
        expr.keyword.precedingWhitespace = getWhitespaceLeft(ctx.TYPEOF())
        expr.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.LPAREN())
        expr.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.RPAREN())
        expr

      case 17 =>
        val typeInd = visitTypeType(ctx.langType().typeType())
        val typeExpr = TypeExpression(typeInd, getSourceLocation(ctx))
        typeExpr.typeKeyword.precedingWhitespace = getWhitespaceLeft(ctx.TYPE())
        typeExpr.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.LPAREN())
        typeExpr.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.RPAREN())
        typeExpr
      case 24 =>
        val quantExpr = visitSpecQuantifiedExpression(ctx.specQuantifiedExpression())
        quantExpr.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.LPAREN())
        quantExpr.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.RPAREN())
        quantExpr
      case 25 =>
        val quantExpr =visitSpecQuantifiedExpression(ctx.specQuantifiedExpression())
        quantExpr.openParenthesis.precedingWhitespace = quantExpr.quantifier.precedingWhitespace
        quantExpr.quantifier.precedingWhitespace = ""
        quantExpr
      case 26 =>
        StringNode(ctx.COUNT().getText, getSourceLocation(ctx.COUNT()), getWhitespaceLeft(ctx.COUNT()))
      case 27 =>
        StringNode(ctx.NOTHING().getText, getSourceLocation(ctx.NOTHING()), getWhitespaceLeft(ctx.NOTHING()))
      case 28 =>
        InformalDescription(printer.visitSkipFirstWS(ctx), getSourceLocation(ctx), getWhitespaceLeft(ctx))
      case _ =>
        StringNode("not yet implemented", null, " ")
    }
  }

  override def visitSpecQuantifiedExpression(ctx: JavaParser.SpecQuantifiedExpressionContext):
                QuantifiedExpression = {
    val decl = visitQuantifiedVarDecls(ctx.quantifiedVarDecls())
    var guard: Option[Expression] = None
    var expr_idx: Int = 0
    if (ctx.langExpr().size() > 1) {
      guard = Some(visitExpression(ctx.langExpr(0).expression()))
      expr_idx = 1
    }
    val expr = visitExpression(ctx.langExpr(expr_idx).expression())
    var quantExpr: QuantifiedExpression = null
    if(ctx.quantifier().FORALL() != null) {
      quantExpr = ForallExpression(decl, guard, expr, getSourceLocation(ctx))
    } else if(ctx.quantifier().EXISTS() != null) {
      quantExpr = ExistsExpression(decl, guard, expr, getSourceLocation(ctx))
    } else if(ctx.quantifier().MAX() != null) {
      quantExpr = MaxQuantifiedExpression(decl, guard, expr, getSourceLocation(ctx))
    } else if(ctx.quantifier().MIN() != null) {
      quantExpr = MinExpression(decl, guard, expr, getSourceLocation(ctx))
    } else if(ctx.quantifier().NUM_OF() != null) {
      quantExpr = NumOfExpression(decl, guard, expr, getSourceLocation(ctx))
    } else if(ctx.quantifier().PRODUCT() != null) {
      quantExpr = ProductExpression(decl, guard, expr, getSourceLocation(ctx))
    } else if(ctx.quantifier().SUM() != null) {
      quantExpr = SumExpression(decl, guard, expr, getSourceLocation(ctx))
    }
    quantExpr.quantifier.precedingWhitespace = getWhitespaceLeft(ctx.quantifier())
    quantExpr.firstSemicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI(0))
    if (guard.nonEmpty) {
      quantExpr.secondSemicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI(1))
    }
    quantExpr
  }

  override def visitQuantifiedVarDecls(ctx: JavaParser.QuantifiedVarDeclsContext): QuantifiedVarDeclaration = {
    val declType = visitTypeType(ctx.langType().typeType())
    val decls = ctx.quantifiedVarDeclarator().asScala
                  .map(x=>addWhitespaceRight(visitQuantifiedVarDeclarator(x),x)).toList
    if (ctx.NULLABLE() != null) {
      new NullableQuantifiedVarDeclaration(declType, decls, getSourceLocation(ctx))
    } else if (ctx.NON_NULL() != null) {
      new NonNullQuantifiedVarDeclaration(declType, decls, getSourceLocation(ctx))
    } else {
      new QuantifiedVarDeclaration(declType, decls, getSourceLocation(ctx))
    }
  }

  override def visitQuantifiedVarDeclarator(ctx: JavaParser.QuantifiedVarDeclaratorContext): StringNode = {
    printNode(ctx)
  }

  override def visitSpecVariableDeclarator(ctx: JavaParser.SpecVariableDeclaratorContext):
                VariableDeclarator = {
    val name = visitJavaIdentifier(ctx.langId().javaIdentifier())
    for(i <- 0 until ctx.LBRACK().size()) {
      name.text = name.text.appendedAll(getWhitespaceLeft(ctx.LBRACK(i)))
      name.text = name.text.appendedAll(ctx.LBRACK(i).getText)
      name.text = name.text.appendedAll(getWhitespaceLeft(ctx.RBRACK(i)))
      name.text = name.text.appendedAll(ctx.RBRACK(i).getText)
    }
    var initializer: Option[Initializer] = None
    if (ctx.ASSIGN() != null) {
      initializer = Some(visitSpecInitializer(ctx.specInitializer()))
    }
    val decl = VariableDeclarator(name, initializer, getSourceLocation(ctx))
    decl.equalSign.precedingWhitespace = getWhitespaceLeft(ctx.ASSIGN())
    decl
  }

  override def visitSpecInitializer(ctx: JavaParser.SpecInitializerContext): Initializer = {
    if (ctx.expression() != null) {
      visitExpression(ctx.expression())
    } else {
      visitSpecArrayInitializer(ctx.specArrayInitializer())
    }
  }

  override def visitSpecArrayInitializer(ctx: JavaParser.SpecArrayInitializerContext): ArrayInitializer = {
    val initializers = ctx.specInitializer().asScala
                        .map(x=>addWhitespaceRight(visitSpecInitializer(x),x)).toList
    var comma: Option[StringNode] = None
    if (ctx.COMMA().size() >= ctx.specInitializer().size()) {
      val trailingComma = ctx.COMMA().asScala.last
      comma = Some(StringNode(trailingComma.getText, getSourceLocation(trailingComma),
                              getWhitespaceLeft(trailingComma)))
    }
    val arr = ArrayInitializer(initializers, comma, getSourceLocation(ctx))
    arr.openBrace.precedingWhitespace = getWhitespaceLeft(ctx.LBRACE())
    arr.closeBrace.precedingWhitespace = getWhitespaceLeft(ctx.RBRACE())
    arr
  }

  def visitStoreRefListAsList(ctx: JavaParser.StoreRefListContext): List[Expression] = {
    if (ctx.storeRefKeyword() != null) {
      List(printNode(ctx.storeRefKeyword()))
    } else {
      ctx.storeRef().asScala.map(visitStoreRef).toList
    }
  }

  override def visitStoreRef(ctx: JavaParser.StoreRefContext): Expression = {
    if(ctx.storeRefExpression() != null) {
      def visitStoreRefExpression(name: JavaParser.StoreRefNameContext,
                                  suffixes: List[JavaParser.StoreRefNameSuffixContext]): Expression = {
        if(suffixes.isEmpty) {
          printNode(name)
        } else {
          val last = suffixes.last
          if(last.specArrayRefExpr() != null) {
            val res = ArrayIndexExpression(visitStoreRefExpression(name, suffixes.dropRight(1)),
                                            visitSpecArrayRefExpr(last.specArrayRefExpr()),
                                            getSourceLocation(ctx))
            res.openBracket.precedingWhitespace = getWhitespaceLeft(last.LBRACK())
            res.closeBracket.precedingWhitespace = getWhitespaceLeft(last.RBRACK())
            res
          } else if(last.MUL() != null) {
            val res = NestedInvocationExpression(visitStoreRefExpression(name, suffixes.dropRight(1)),
                                                IndexAllExpression(getSourceLocation(last.MUL()), getWhitespaceLeft(last.MUL())),
                                                getSourceLocation(ctx))
            res.dot.precedingWhitespace = getWhitespaceLeft(last.DOT())
            res
          } else if(last.langId() != null) {
            val res = NestedInvocationExpression(visitStoreRefExpression(name, suffixes.dropRight(1)),
              visitJavaIdentifier(last.langId().javaIdentifier()),
              getSourceLocation(ctx))
            res.dot.precedingWhitespace = getWhitespaceLeft(last.DOT())
            res
          } else {
            val res = NestedInvocationExpression(visitStoreRefExpression(name, suffixes.dropRight(1)),
              StringNode(last.THIS().getText, getSourceLocation(last.THIS()), getWhitespaceLeft(last.THIS())),
              getSourceLocation(ctx))
            res.dot.precedingWhitespace = getWhitespaceLeft(last.DOT())
            res
          }
        }
      }

      visitStoreRefExpression(ctx.storeRefExpression().storeRefName(),
                              ctx.storeRefExpression().storeRefNameSuffix().asScala.toList)
    } else {
      visitInformalDescription(ctx.informalDescription())
    }
  }

  override def visitSpecArrayRefExpr(ctx: JavaParser.SpecArrayRefExprContext): IndexExpression = {
    if(ctx.RANGE_TO() != null) {
      val res = RangeExpression(visitExpression(ctx.expression(0)), visitExpression(ctx.expression(1)), getSourceLocation(ctx))
      res.dots.precedingWhitespace = getWhitespaceLeft(ctx.RANGE_TO())
      res
    } else if (ctx.FLOAT_LITERAL().asScala.nonEmpty) {
      if(ctx.FLOAT_LITERAL(0).getText.endsWith(".") && ctx.FLOAT_LITERAL(1).getText.startsWith(".")) {
        val lowerS = StringNode(ctx.FLOAT_LITERAL(0).getText.dropRight(1), getSourceLocation(ctx.FLOAT_LITERAL(0)))
        val lower = Literal(lowerS, lowerS.srcLocation, getWhitespaceLeft(ctx.FLOAT_LITERAL(0)))
        val upperS = StringNode(ctx.FLOAT_LITERAL(1).getText.drop(1), getSourceLocation(ctx.FLOAT_LITERAL(1)))
        val upper = Literal(upperS, upperS.srcLocation)
        val res = RangeExpression(lower, upper, getSourceLocation(ctx))
        res
      } else {
        System.err.println(s"Array range '${printer.visit(ctx)}' is invalid syntax, aborting")
        throw new IllegalArgumentException()
      }
    } else if(ctx.MUL() != null) {
      IndexAllExpression(getSourceLocation(ctx.MUL()), getWhitespaceLeft(ctx.MUL()))
    } else {
      visitExpression(ctx.expression(0))
    }
  }

  override def visitInformalDescription(ctx: JavaParser.InformalDescriptionContext): InformalDescription = {
    InformalDescription(printer.visitSkipFirstWS(ctx), getSourceLocation(ctx), getWhitespaceLeft(ctx))
  }

  override def visitSetComprehension(ctx: JavaParser.SetComprehensionContext): SetComprehension = {
    val typeInd = visitTypeType(ctx.langType().typeType())
    val declarator = visitQuantifiedVarDeclarator(ctx.quantifiedVarDeclarator())
    var declaration: QuantifiedVarDeclaration = null
    if (ctx.NON_NULL() != null) {
      val decl = new NonNullQuantifiedVarDeclaration(typeInd, List(declarator), getSourceLocation(ctx))
      decl.nonNullKeyword.precedingWhitespace = getWhitespaceLeft(ctx.NON_NULL())
      declaration = decl
    } else if (ctx.NULLABLE() != null) {
      val decl = new NullableQuantifiedVarDeclaration(typeInd, List(declarator), getSourceLocation(ctx))
      decl.nullableKeyword.precedingWhitespace = getWhitespaceLeft(ctx.NON_NULL())
      declaration = decl
    } else {
      declaration = new QuantifiedVarDeclaration(typeInd, List(declarator), getSourceLocation(ctx))
    }
    val containsExpr = visitExpression(ctx.expression(0))
    val filterExpr = visitExpression(ctx.expression(1))
    // typeIndication of SetComprehension is set when creating newExpression
    val set = SetComprehension(null, declaration, containsExpr, filterExpr, getSourceLocation(ctx))
    set.openBrace.precedingWhitespace = getWhitespaceLeft(ctx.LBRACE())
    set.bar.precedingWhitespace = getWhitespaceLeft(ctx.BITOR())
    set.and.precedingWhitespace = getWhitespaceLeft(ctx.AND())
    set.closeBrace.precedingWhitespace = getWhitespaceLeft(ctx.RBRACE())
    set
  }

  override def visitValDeclaration(ctx: JavaParser.ValDeclarationContext): SpecDeclaration = {
    ctx.altNum match {
      case 1 =>
        val modifiers = ctx.valModifier().asScala.map(visitValModifier).toList
        val expr = visitExpression(ctx.langExpr().expression())
        val decl = InvariantDeclaration(expr, getSourceLocation(ctx))
        decl.modifiers = modifiers
        decl.invariantKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_INVARIANT())
        decl.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
        decl
      case 2 =>
        val modifiers = ctx.valModifier().asScala.map(visitValModifier).toList
        val expr = visitExpression(ctx.langExpr().expression())
        val decl = InitiallyDeclaration(expr, getSourceLocation(ctx))
        decl.modifiers = modifiers
        decl.initiallyKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_INITIALLY())
        decl.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
        decl
      case 3 =>
        val modifiers = ctx.valModifier().asScala.map(visitValModifier).toList
        val expr = visitExpression(ctx.langExpr().expression())
        val ident = visitJavaIdentifier(ctx.langId().javaIdentifier())
        val decl = ReadableDeclaration(ident, expr, getSourceLocation(ctx))
        decl.modifiers = modifiers
        decl.readableKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_READABLE())
        decl.ifKeyword.precedingWhitespace = getWhitespaceLeft(ctx.IF())
        decl.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
        decl
      case 4 =>
        val modifiers = ctx.valModifier().asScala.map(visitValModifier).toList
        val expr = visitExpression(ctx.langExpr().expression())
        val ident = visitJavaIdentifier(ctx.langId().javaIdentifier())
        val decl = WritableDeclaration(ident, expr, getSourceLocation(ctx))
        decl.modifiers = modifiers
        decl.writable.precedingWhitespace = getWhitespaceLeft(ctx.VAL_WRITABLE())
        decl.ifKeyword.precedingWhitespace = getWhitespaceLeft(ctx.IF())
        decl.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
        decl
      case 5 =>
        val modifiers = ctx.valModifier().asScala.map(visitValModifier).toList
        val expr = ctx.valExpressionList().langExpr().asScala
                    .map(x => addWhitespaceRight(visitExpression(x.expression()),x)).toList
        val ident = visitJavaIdentifier(ctx.langId().javaIdentifier())
        val decl = MonitorsForDeclaration(ident, expr, getSourceLocation(ctx))
        decl.modifiers = modifiers
        decl.monitorsForKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_MONITORS_FOR())
        decl.equals.precedingWhitespace = getWhitespaceLeft(ctx.ASSIGN())
        decl.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
        decl
      case 6 =>
        val expr = visitExpression(ctx.langExpr().expression())
        val decl = AxiomDeclaration(expr, getSourceLocation(ctx))
        decl.axiomKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_INVARIANT())
        decl.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
        decl
      case 7 =>
        val specs = ctx.valSpecCase().asScala.map(visitValSpecCase)
        var offset = 1
        if (ctx.VAL_ALSO().size() == ctx.valSpecCase().size()) {
          offset = 0
          specs.head.hasAlso = true
          specs.head.alsoKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_ALSO().asScala.head)
        }
        for (i <- 1 until specs.size) {
          specs(i).hasAlso = true
          specs(i).alsoKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_ALSO(i - offset))
        }
        val contract = Contract(specs.toList, getSourceLocation(ctx))
        var modifiers = ctx.valModifier().asScala.map(visitValModifier)
        modifiers = modifiers.map {
          case mod: ModelModifier =>
            val res = SpecMethodModifier(mod.srcLocation, mod.precedingWhitespace)
            res.trailingWhitespace = mod.trailingWhitespace
            res
          case x => x
        }
        val pure = PureModifier(getSourceLocation(ctx.VAL_PURE()),getWhitespaceLeft(ctx.VAL_PURE()))
        modifiers.append(pure)
        val retType = visitTypeType(ctx.langType().typeType())
        val name = visitJavaIdentifier(ctx.langId().javaIdentifier())
        val body = visitMethodBody(ctx.methodBody())
        var parameters: List[Parameter] = Nil
        if (ctx.formalParameters().formalParameterList() != null) {
          parameters = visitFormalParameterListAsList(ctx.formalParameters().formalParameterList())
        }
        val res = SpecMethodDeclaration(name, Some(retType), parameters, body, getSourceLocation(ctx))
        res.contractBefore = Some(contract)
        res.modifiers = modifiers.toList
        if (ctx.formalParameters().formalParameterList() != null) {
          res.parameters = visitFormalParameterListAsList(ctx.formalParameters().formalParameterList())
        }
        res.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.formalParameters().LPAREN())
        res.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.formalParameters().RPAREN())
        res
      case 8 =>
        def findModifiers(modifiers: List[Modifier]): List[Modifier] = {
          def checkModifier(modifier: Modifier): List[Modifier] = {
            modifier match {
              case _: GhostModifier => List(modifier)
              case _: ModelModifier => List(modifier)
              case j: JmlModifierBlock => findModifiers(j.modifiers)
              case _ => List()
            }
          }
          modifiers.flatMap(checkModifier)
        }

        val member = visitMemberDeclaration(ctx.memberDeclaration())
        val modifiers = ctx.valModifier().asScala.map(visitValModifier)
        val ghostsAndModels = findModifiers(modifiers.toList)
        val ghosts = ghostsAndModels.filter(x => x.isInstanceOf[GhostModifier])
        if(ghosts.nonEmpty) {
          val ghost = ghosts.head.asInstanceOf[GhostModifier]
          if(ghost != modifiers.head) {
            ghost.precedingWhitespace = modifiers.head.precedingWhitespace + ghost.precedingWhitespace
            modifiers.head.precedingWhitespace = " "
          }
          member.modifiers = modifiers.filter(x => x != ghost).toList
          GhostDeclaration(ghost, member, getSourceLocation(ctx))
        } else {
          val models = ghostsAndModels.filter(x => x.isInstanceOf[ModelModifier])
          if(models.nonEmpty) {
            var model = models.head.asInstanceOf[ModelModifier]
            if(model != modifiers.head) {
              model.precedingWhitespace = modifiers.head.precedingWhitespace + model.precedingWhitespace
              modifiers.head.precedingWhitespace = " "
            }
            member.modifiers = modifiers.filter(x => x != model).toList
            model = member match {
              case specMeth: MethodDeclaration =>
                val specMod = SpecMethodModifier(model.srcLocation, model.precedingWhitespace)
                specMod.trailingWhitespace = model.trailingWhitespace
                specMod
              case _ => model
            }
            ModelDeclaration(model, member, getSourceLocation(ctx))
          } else {
            System.err.println(s"[Warning] Only model and ghost declarations supported${getSourceLocation(ctx).toString}, declaration commented")
            val text = printNode(ctx)
            NonTranslatedSpecDeclaration(Helper.commentUnknownNode(text, ""), getSourceLocation(ctx))
          }
        }
    }
  }
}
