package languages.OpenJML

import generic.{Helper, NotSupportedException}
import intermediateRepresentation.syntaxTree as ir
import languages.OpenJML.syntaxTree as jml


class ToIrVisitor {

  // indicates whether new modifiers, etc need to be wrapped in specifications
  // or if they already are (e.g. inside ghost declarations)
  var wrapInSpec = true

  // indicates whether the class currently being visited has nullable_by_default
  var isNullableDefault = false

  // types that do not need to get a "non_null" modifier because they are never null
  val nonNullableTypes = List("void", "boolean", "byte", "char", "short", "int", "long", "float", "double")

  def hasNullableModifier(modifiers: List[jml.Modifier]): Boolean = {
    def checkMod(modifier: jml.Modifier): Boolean = {
      modifier match {
        case _: jml.NullableModifier => true
        case m: jml.JmlModifierBlock => hasNullableModifier(m.modifiers)
        case _ => false
      }
    }

    modifiers.exists(checkMod)
  }
  def hasNullableByDefaultModifier(modifiers: List[jml.Modifier]): Boolean = {
    def checkMod(modifier: jml.Modifier): Boolean = {
      modifier match {
        case _: jml.NullableByDefaultModifier => true
        case m: jml.JmlModifierBlock => hasNullableByDefaultModifier(m.modifiers)
        case _ => false
      }
    }

    modifiers.exists(checkMod)
  }
  def hasNonnullModifier(modifiers: List[jml.Modifier]): Boolean = {
    def checkMod(modifier: jml.Modifier): Boolean = {
      modifier match {
        case _: jml.NonnullModifier => true
        case m: jml.JmlModifierBlock => hasNonnullModifier(m.modifiers)
        case _ => false
      }
    }

    modifiers.exists(checkMod)
  }


  def visitStringNode(node: jml.StringNode): ir.StringNode = {
    node match {
      case imp: jml.ImpliesOperator => visitImpliesOperator(imp)
      case jml.StringNode(text, src, pred, trail) =>
        ir.StringNode(text, src, pred, trail)
    }
  }
  def visitImpliesOperator(node: jml.ImpliesOperator): ir.ImpliesOperator = {
    node match {
      case jml.ImpliesOperator(srcLocation, pred) =>
        val res = new ir.ImpliesOperator(srcLocation)
        res.copyWhitespace(node)
        res
    }
  }
  def visitModifierStringNode(node: jml.ModifierStringNode): ir.ModifierStringNode = {
    node match {
      case jml.ModifierStringNode(text, src, pred) =>
        val res = ir.ModifierStringNode(text, src, pred)
        res.copyWhitespace(node)
        res
    }
  }

  def visitCompilationUnit(node: jml.CompilationUnit): ir.CompilationUnit = {
    node match {
      case nt: jml.NonTranslatedCompilationUnit =>
        visitNonTranslatedCompilationUnit(nt)
      case jml.CompilationUnit(pkg, imports, declarations, srcLocation, precedingWhitespace) =>
        try
          var irPkg: Option[ir.PackageDeclaration] = None
          if (pkg.isDefined) {
            irPkg = Some(visitPackageDeclaration(pkg.get))
          }
          val res = ir.CompilationUnit(irPkg,
                                        imports.map(visitImportDeclaration),
                                        declarations.map(visitTypeDeclaration),
                                        srcLocation,
                                        precedingWhitespace)
          res.copyWhitespace(node)
          res
        catch {
          case e: NotSupportedException =>
            val text = Helper.commentUnknownNode(node, s"${e.construct} not supported, whole CompilationUnit commented out")
            ir.NonTranslatedCompilationUnit(text, node.srcLocation)
        }
    }
  }

  def visitNonTranslatedCompilationUnit(node: jml.NonTranslatedCompilationUnit): ir.NonTranslatedCompilationUnit = {
    node match {
      case jml.NonTranslatedCompilationUnit(text, srcLocation, precedingWhitespace) =>
        val res = ir.NonTranslatedCompilationUnit(text, srcLocation, precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitPackageDeclaration(node: jml.PackageDeclaration): ir.PackageDeclaration = {
    node match {
      case jml.PackageDeclaration(name, annotations, srcLocation, precedingWhitespace) =>
        val res = ir.PackageDeclaration(visitStringNode(name),
                                        annotations.map(visitStringAnnotation),
                                        srcLocation,
                                        precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitImportDeclaration(node: jml.ImportDeclaration): ir.ImportDeclaration = {
    node match {
      case jml.ImportDeclaration(name, srcLocation, precedingWhitespace) =>
        val res = ir.ImportDeclaration(visitStringNode(name),
                              srcLocation,
                              precedingWhitespace)
        res.modifiers = node.modifiers.map(visitModifier)
        res.copyWhitespace(node)
        res
    }
  }

  def visitTypeDeclaration(node: jml.TypeDeclaration): ir.TypeDeclaration = {
    try
      node match {
        case decl: jml.AnnotationTypeDeclaration => visitAnnotationTypeDeclaration(decl)
        case decl: jml.ClassDeclaration => visitClassDeclaration(decl)
        case decl: jml.EnumDeclaration => visitEnumDeclaration(decl)
        case decl: jml.InterfaceDeclaration => visitInterfaceDeclaration(decl)
        case decl: jml.EmptyStatement => visitEmptyStatement(decl)
        case decl: jml.NonTranslatedTypeDeclaration => visitNonTranslatedTypeDeclaration(decl)
        case _ => ???
      }
    catch {
      case e: NotSupportedException =>
        val text = Helper.commentUnknownNode(node, s"${e.construct} not supported, whole TypeDeclaration commented out")
        ir.NonTranslatedTypeDeclaration(text, node.srcLocation)
    }
  }

  def visitNonTranslatedTypeDeclaration(node: jml.NonTranslatedTypeDeclaration): ir.NonTranslatedTypeDeclaration = {
    node match {
      case jml.NonTranslatedTypeDeclaration(text, srcLocation, precedingWhitespace) =>
        val res = ir.NonTranslatedTypeDeclaration(text, srcLocation, precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitGhostDeclaration(node: jml.GhostDeclaration): ir.GhostDeclaration = {
    node match {
      case jml.GhostDeclaration(ghost, decl, srcLocation, precedingWhitespace) =>
        val res = ir.GhostDeclaration(visitGhostModifier(ghost),
          visitMemberDeclaration(decl),
          srcLocation,
          precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitModelDeclaration(node: jml.ModelDeclaration): ir.ModelDeclaration = {
    node match {
      case jml.ModelDeclaration(model, decl, srcLocation, precedingWhitespace) =>
        val res = ir.ModelDeclaration(visitModelModifier(model),
          visitMemberDeclaration(decl),
          srcLocation,
          precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitModifier(node: jml.Modifier): ir.Modifier = {
    node match {
      case mod: jml.StaticModifier => visitStaticModifier(mod)
      case mod: jml.Annotation => visitAnnotation(mod)
      case mod: jml.PureModifier => visitPureModifier(mod)
      case mod: jml.NullableModifier => visitNullableModifier(mod)
      case mod: jml.NullableByDefaultModifier => visitNullableByDefaultModifier(mod)
      case mod: jml.NonnullModifier => visitNonnullModifier(mod)
      case mod: jml.GhostModifier => visitGhostModifier(mod)
      case mod: jml.HelperModifier => visitHelperModifier(mod)
      case mod: jml.MonitoredModifier => visitMonitoredModifier(mod)
      case mod: jml.ModelModifier => visitModelModifier(mod)
      case mod: jml.SpecMethodModifier => visitSpecMethodModifier(mod)
      case mod: jml.InstanceModifier => visitInstanceModifier(mod)
      case mod: jml.SpecPublicModifier => visitSpecPublicModifier(mod)
      case mod: jml.SpecProtectedModifier => visitSpecProtectedModifier(mod)
      case mod: jml.JmlModifierBlock => visitJmlModifierBlock(mod)
      case mod: jml.ModifierStringNode => visitModifierStringNode(mod)
      case _ => ???
    }
    // todo: more?
  }
  def visitStaticModifier(node: jml.StaticModifier): ir.StaticModifier = {
    node match {
      case jml.StaticModifier(srcLocation, precedingWhitespace) =>
        val res = ir.StaticModifier(srcLocation, precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitGenericsSuperclassClause(node: jml.GenericsSuperclassClause): ir.GenericsSuperclassClause = {
    node match {
      case jml.GenericsSuperclassClause(superclasses, srcLocation, precedingWhitespace) =>
        val res = ir.GenericsSuperclassClause(superclasses.map(visitTypeIndication),
                                              srcLocation,
                                              precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitSuperclassClause(node: jml.SuperclassClause): ir.SuperclassClause = {
    node match {
      case jml.SuperclassClause(superclass, srcLocation, precedingWhitespace) =>
        val res = ir.SuperclassClause(visitTypeIndication(superclass),
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitMultiSuperclassClause(node: jml.MultiSuperclassClause): ir.MultiSuperclassClause = {
    node match {
      case jml.MultiSuperclassClause(superclasses, srcLocation, precedingWhitespace) =>
        val res = ir.MultiSuperclassClause(superclasses.map(visitTypeIndication),
                                            srcLocation,
                                            precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitInterfaceClause(node: jml.InterfaceClause): ir.InterfaceClause = {
    node match {
      case jml.InterfaceClause(interfaces, srcLocation, precedingWhitespace) =>
        val res = ir.InterfaceClause(interfaces.map(visitTypeIndication),
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitClassDeclaration(node: jml.ClassDeclaration): ir.ClassDeclaration = {
    node match {
      case jml.ClassDeclaration(name, typeParameters, superclass, interfaces, body, srcLocation, precedingWhitespace) =>
        val irTypeParameters = typeParameters.map(visitTypeParameterClause)
        val irSuperclass = superclass.map(visitSuperclassClause)
        val irInterfaces = interfaces.map(visitInterfaceClause)
        val oldNullable = isNullableDefault
        isNullableDefault = hasNullableByDefaultModifier(node.modifiers) || isNullableDefault
        val res = ir.ClassDeclaration(visitStringNode(name),
                                      irTypeParameters,
                                      irSuperclass,
                                      irInterfaces,
                                      visitBody(body),
                                      srcLocation,
                                      precedingWhitespace)
        res.modifiers = node.modifiers.map(visitModifier)
        isNullableDefault = oldNullable
        res.copyWhitespace(node)
        res
    }
  }

  def visitTypeParameterClause(node: jml.TypeParameterClause): ir.TypeParameterClause = {
    node match {
      case jml.TypeParameterClause(parameters, srcLocation, precedingWhitespace) =>
        val res = ir.TypeParameterClause(parameters.map(visitTypeParameter),
                                          srcLocation,
                                          precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitTypeParameter(node: jml.TypeParameter): ir.TypeParameter = {
    node match {
      case jml.TypeParameter(annotations, name, superclass, srcLocation, precedingWhitespace) =>
        val irSuperclass = superclass.map(visitGenericsSuperclassClause)
        val res = ir.TypeParameter(annotations.map(visitStringAnnotation),
                                    visitTypeIndication(name),
                                    irSuperclass,
                                    srcLocation,
                                    precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitEnumDeclaration(node: jml.EnumDeclaration): ir.EnumDeclaration = {
    node match {
      case jml.EnumDeclaration(name, interfaces, constants, declarations, comma, semicolon, srcLocation, precedingWhitespace) =>
        val irInterfaces = interfaces.map(visitInterfaceClause)
        val irComma = comma.map(visitStringNode)
        val irSemicolon = semicolon.map(visitStringNode)
        val res = ir.EnumDeclaration(visitStringNode(name),
                                      irInterfaces,
                                      constants.map(visitEnumConstant),
                                      declarations.map(visitMemberDeclaration),
                                      irComma,
                                      irSemicolon,
                                      srcLocation,
                                      precedingWhitespace)
        res.modifiers = node.modifiers.map(visitModifier)
        res.copyWhitespace(node)
        res
    }
  }
  def visitEnumConstant(node: jml.EnumConstant): ir.EnumConstant = {
    node match {
      case jml.EnumConstant(name, annotations, arguments, body, srcLocation, precedingWhitespace) =>
        val irArgs = arguments.map(visitArguments)
        val irBody = body.map(visitBody)
        val res = ir.EnumConstant(visitStringNode(name),
                                  annotations.map(visitStringAnnotation),
                                  irArgs,
                                  irBody,
                                  srcLocation,
                                  precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitInterfaceDeclaration(node: jml.InterfaceDeclaration): ir.InterfaceDeclaration = {
    node match {
      case jml.InterfaceDeclaration(name, typeParameter, superInterfaces, body, srcLocation, precedingWhitespace) =>
        val irTypeParameters = typeParameter.map(visitTypeParameterClause)
        val irSuperclass = superInterfaces.map(visitMultiSuperclassClause)
        val oldNullable = isNullableDefault
        isNullableDefault = hasNullableByDefaultModifier(node.modifiers)
        val res = ir.InterfaceDeclaration(visitStringNode(name),
                                          irTypeParameters,
                                          irSuperclass,
                                          visitBody(body),
                                          srcLocation,
                                          precedingWhitespace)
        res.modifiers = node.modifiers.map(visitModifier)
        res.copyWhitespace(node)
        isNullableDefault = oldNullable
        res
    }
  }

  def visitBody(node: jml.Body): ir.Body = {
    node match {
      case jml.Body(declarations, srcLocation, precedingWhitespace) =>
        val res = ir.Body(declarations.map(visitMemberDeclaration),
                          srcLocation,
                          precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitMemberDeclaration(node: jml.MemberDeclaration): ir.MemberDeclaration = {
    try
      node match {
        case member: jml.ClassDeclaration => visitClassDeclaration(member)
        case member: jml.EnumDeclaration => visitEnumDeclaration(member)
        case member: jml.InterfaceDeclaration => visitInterfaceDeclaration(member)
        case member: jml.GenericMethodDeclaration => visitGenericMethodDeclaration(member)
        case member: jml.MethodDeclaration => visitMethodDeclaration(member)
        case member: jml.BlockOrSemicolon => visitBlockOrSemicolon(member)
        case member: jml.FieldDeclaration => visitFieldDeclaration(member)
        case member: jml.AnnotationTypeDeclaration => visitAnnotationTypeDeclaration(member)
        case member: jml.SpecDeclaration => visitSpecDeclaration(member)
        case member: jml.SpecDeclarationBlock => visitSpecDeclarationBlock(member)
        case member: jml.NonTranslatedMemberDeclaration => visitNonTranslatedMemberDeclaration(member)
        case _ => ???
      }
    catch {
      case e: NotSupportedException =>
        val text = Helper.commentUnknownNode(node, s"${e.construct} not supported, whole MemberDeclaration commented out")
        ir.NonTranslatedMemberDeclaration(text, node.srcLocation)
    }
  }
  def visitNonTranslatedMemberDeclaration(node: jml.NonTranslatedMemberDeclaration): ir.NonTranslatedMemberDeclaration = {
    node match {
      case jml.NonTranslatedMemberDeclaration(text, srcLocation, precedingWhitespace) =>
        val res = ir.NonTranslatedMemberDeclaration(text, srcLocation, precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitMethodDeclaration(node: jml.MethodDeclaration): ir.MethodDeclaration = {
    node match {
      case jml.MethodDeclaration(name, returnType, parameters, dimensions, throws, default, contractWithin,
                                 body, srcLocation, precedingWhitespace) =>
        val irReturn = returnType.map(visitTypeIndication)
        val irDims = dimensions.map(visitStringNode)
        val irThrows = throws.map(visitThrowsClause)
        val irDefault = default.map(visitDefaultClause)
        val irContractWithin = contractWithin.map(visitContract)
        val res = new ir.MethodDeclaration(visitStringNode(name),
                                            irReturn,
                                            parameters.map(visitParameter),
                                            irDims,
                                            irThrows,
                                            irDefault,
                                            irContractWithin,
                                            visitBlockOrSemicolon(body),
                                            srcLocation,
                                            precedingWhitespace)
        res.contractBefore = if(node.contractBefore.isEmpty) None else Some(visitContract(node.contractBefore.get))
        res.modifiers = node.modifiers.map(visitModifier)
        if(!isNullableDefault && !hasNullableModifier(node.modifiers) && !hasNonnullModifier(node.modifiers)) {
          if(returnType.isDefined && !nonNullableTypes.contains(returnType.get.typeAsString)){
            var mod: ir.Modifier = ir.NonnullModifier(srcLocation, " ")
            if(wrapInSpec) {
              mod = ir.JmlModifierBlock(Some(ir.StringNode("/*@", srcLocation, " ")),
                                        List(mod),
                                        Some(ir.StringNode("*/", srcLocation, " ")),
                                        srcLocation)
            }
            if(res.modifiers.isEmpty && returnType.isDefined) {
              mod.precedingWhitespace = res.returnType.get.precedingWhitespace
              res.returnType.get.precedingWhitespace = " "
            }
            res.modifiers = res.modifiers.appended(mod)
          }
        }
        res.copyWhitespace(node)
        res
    }
  }


  def visitGenericMethodDeclaration(node: jml.GenericMethodDeclaration): ir.GenericMethodDeclaration = {
    node match {
      case jml.GenericMethodDeclaration(pName, typeParameters, pReturnType,
                                        pParameters, pDimensions, pThrows, pDefault, pContractWithin,
                                        pBody, pSrcLocation, pPrecedingWhitespace) =>
        val irReturn = pReturnType.map(visitTypeIndication)
        val irDims = pDimensions.map(visitStringNode)
        val irThrows = pThrows.map(visitThrowsClause)
        val irDefault = pDefault.map(visitDefaultClause)
        val irContractWithin = pContractWithin.map(visitContract)
        val res = ir.GenericMethodDeclaration(visitStringNode(pName),
                                              visitTypeParameterClause(typeParameters),
                                              irReturn,
                                              pParameters.map(visitParameter),
                                              irDims,
                                              irThrows,
                                              irDefault,
                                              irContractWithin,
                                              visitBlockOrSemicolon(pBody),
                                              pSrcLocation,
                                              pPrecedingWhitespace)
//        res.contractBefore = if(node.contractBefore.isEmpty) None else Some(visitContract(node.contractBefore.get))
        res.modifiers = node.modifiers.map(visitModifier)
        if(!isNullableDefault && !hasNullableModifier(node.modifiers) && !hasNonnullModifier(node.modifiers)) {
          if(pReturnType.isDefined && !nonNullableTypes.contains(pReturnType.get.typeAsString)){
            var mod: ir.Modifier = ir.NonnullModifier(pSrcLocation, " ")
            if(wrapInSpec) {
              mod = ir.JmlModifierBlock(Some(ir.StringNode("/*@", pSrcLocation, " ")),
                List(mod),
                Some(ir.StringNode("*/", pSrcLocation, " ")),
                pSrcLocation)
            }
            if(res.modifiers.isEmpty && pReturnType.isDefined) {
              mod.precedingWhitespace = res.pReturnType.get.precedingWhitespace
              res.pReturnType.get.precedingWhitespace = " "
            }
            res.modifiers = res.modifiers.appended(mod)
          }
        }
        res.copyWhitespace(node)
        res
    }
  }

  def visitThrowsClause(node: jml.ThrowsClause): ir.ThrowsClause = {
    node match {
      case jml.ThrowsClause(exceptions, srcLocation, precedingWhitespace) =>
        val res = ir.ThrowsClause(exceptions.map(visitStringNode),
                                  srcLocation,
                                  precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitTypeIndication(node: jml.TypeIndication): ir.TypeIndication = {
    node match {
      case indication: jml.NoType => visitNoType(indication)
      case indication: jml.StringGenericTypeIndication => visitStringGenericTypeIndication(indication)
      case indication: jml.StringTypeIndication => visitStringTypeIndication(indication)
      case _ => ???
    }
  }
  def visitStringTypeIndication(node: jml.StringTypeIndication): ir.StringTypeIndication = {
    node match {
      case jml.StringTypeIndication(fullTypeString, annotations, typeAsString,
                                    pSrcLocation, pPrecedingWhitespace, pTrailingWhitespace) =>
        val res = new ir.StringTypeIndication(fullTypeString,
                                              annotations.map(visitAnnotation),
                                              typeAsString,
                                              pSrcLocation,
                                              pPrecedingWhitespace,
                                              pTrailingWhitespace)
        res.modifiers = node.modifiers.map(visitModifier)
        res.copyWhitespace(node)
        res
    }
  }

  def visitStringGenericTypeIndication(node: jml.StringGenericTypeIndication): ir.StringGenericTypeIndication = {
    node match {
      case jml.StringGenericTypeIndication(fullTypeString, typeParameters, annotations, typeAsString,
                                            pSrcLocation, pPrecedingWhitespace) =>
        val res = ir.StringGenericTypeIndication(fullTypeString,
                                                  visitTypeParameterClause(typeParameters),
                                                  annotations.map(visitAnnotation),
                                                  typeAsString,
                                                  pSrcLocation,
                                                  pPrecedingWhitespace)
        res.modifiers = node.modifiers.map(visitModifier)
        res.copyWhitespace(node)
        res
    }
  }
  def visitNoType(node: jml.NoType): ir.NoType = {
    node match {
      case jml.NoType(srcLocation) =>
        val res = new ir.NoType(srcLocation)
        res.copyWhitespace(node)
        res
    }
  }

  def visitBlockOrSemicolon(node: jml.BlockOrSemicolon): ir.BlockOrSemicolon = {
    node match {
      case block: jml.Block => visitBlock(block)
      case semi: jml.EmptyStatement => visitEmptyStatement(semi)
      case _ => ???
    }
  }

  def visitFieldDeclaration(node: jml.FieldDeclaration): ir.FieldDeclaration = {
    node match {
      case jml.FieldDeclaration(fieldType, declarators, srcLocation, precedingWhitespace) =>
        val res = ir.FieldDeclaration(visitTypeIndication(fieldType),
                                      declarators.map(visitVariableDeclarator),
                                      srcLocation,
                                      precedingWhitespace)
        res.modifiers = node.modifiers.map(visitModifier)
        if(!isNullableDefault && !hasNullableModifier(node.modifiers) && !hasNonnullModifier(node.modifiers)) {
          if(!nonNullableTypes.contains(fieldType.typeAsString)){
            var mod: ir.Modifier = ir.NonnullModifier(srcLocation, " ")
            if(wrapInSpec) {
              mod = ir.JmlModifierBlock(Some(ir.StringNode("/*@", srcLocation, " ")),
                List(mod),
                Some(ir.StringNode("*/", srcLocation, " ")),
                srcLocation)
            }
            if(res.modifiers.isEmpty) {
              mod.precedingWhitespace = res.fieldType.precedingWhitespace
              res.fieldType.precedingWhitespace = " "
            }
            res.modifiers = res.modifiers.appended(mod)
          }
        }
        res.copyWhitespace(node)
        res
    }
  }

  def visitVariableDeclarator(node: jml.VariableDeclarator): ir.VariableDeclarator = {
    node match {
      case jml.VariableDeclarator(name, initializer, srcLocation, precedingWhitespace) =>
        val irInit = initializer.map(visitInitializer)
        val res = ir.VariableDeclarator(visitStringNode(name),
                                        irInit,
                                        srcLocation,
                                        precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitInitializer(node: jml.Initializer): ir.Initializer = {
    node match {
      case init: jml.ArrayInitializer => visitArrayInitializer(init)
      case init: jml.Expression => visitExpression(init)
      case _ => ???
    }
  }
  def visitArrayInitializer(node: jml.ArrayInitializer): ir.ArrayInitializer = {
    node match {
      case jml.ArrayInitializer(initializers, trailingComma, srcLocation, precedingWhitespace) =>
        val irComma = trailingComma.map(visitStringNode)
        val res = ir.ArrayInitializer(initializers.map(visitInitializer),
                                      irComma,
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitTypeArgumentClause(node: jml.TypeArgumentClause): ir.TypeArgumentClause = {
    node match {
      case jml.TypeArgumentClause(parameters, srcLocation, precedingWhitespace) =>
        val res = ir.TypeArgumentClause(parameters.map(visitTypeArgument),
                                        srcLocation,
                                        precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitTypeArgument(node: jml.TypeArgument): ir.TypeArgument = {
    node match {
      case arg: jml.BoundedTypeArgument => visitBoundedTypeArgument(arg)
      case arg: jml.TypeIndication => visitTypeIndication(arg)
      case _ => ???
    }
  }
  def visitBoundedTypeArgument(node: jml.BoundedTypeArgument): ir.BoundedTypeArgument = {
    node match {
      case jml.BoundedTypeArgument(annotations, typeSuper, typeExtends, srcLocation, precedingWhitespace) =>
        val irSuper = typeSuper.map(visitTypeIndication)
        val irExtends = typeExtends.map(visitTypeIndication)
        val res = ir.BoundedTypeArgument(annotations.map(visitAnnotation),
                                          irSuper,
                                          irExtends,
                                          srcLocation,
                                          precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitParameter(node: jml.Parameter): ir.Parameter = {
    node match {
      case varargs: jml.VarargsParameter =>
        visitVarargsParameter(varargs)
      case jml.Parameter(parameterType, name, srcLocation, precedingWhitespace) =>
        val res = ir.Parameter(visitTypeIndication(parameterType),
                                visitStringNode(name),
                                srcLocation,
                                precedingWhitespace)
        res.modifiers = node.modifiers.map(visitModifier)
        if(!isNullableDefault && !hasNullableModifier(node.modifiers) && !hasNonnullModifier(node.modifiers)) {
          if(!nonNullableTypes.contains(parameterType.typeAsString)){
            var mod: ir.Modifier = ir.NonnullModifier(srcLocation, " ")
            if(wrapInSpec) {
              mod = ir.JmlModifierBlock(Some(ir.StringNode("/*@", srcLocation, " ")),
                List(mod),
                Some(ir.StringNode("*/", srcLocation, " ")),
                srcLocation)
            }
            res.modifiers = res.modifiers.appended(mod)
          }
        }
        res.copyWhitespace(node)
        res
    }
  }
  def visitVarargsParameter(node: jml.VarargsParameter): ir.VarargsParameter = {
    node match {
      case jml.VarargsParameter(parameterType, annotations, name, srcLocation, precedingWhitespace) =>
        val res = ir.VarargsParameter(visitTypeIndication(parameterType),
                                      annotations.map(visitAnnotation),
                                      visitStringNode(name),
                                      srcLocation,
                                      precedingWhitespace)
        res.modifiers = node.modifiers.map(visitModifier)
        if(!isNullableDefault && !hasNullableModifier(node.modifiers) && !hasNonnullModifier(node.modifiers)) {
          if(!nonNullableTypes.contains(parameterType.typeAsString)){
            var mod: ir.Modifier = ir.NonnullModifier(srcLocation, " ")
            if(wrapInSpec) {
              mod = ir.JmlModifierBlock(Some(ir.StringNode("/*@", srcLocation, " ")),
                List(mod),
                Some(ir.StringNode("*/", srcLocation, " ")),
                srcLocation)
            }
            res.modifiers = res.modifiers.appended(mod)
          }
        }
        res.copyWhitespace(node)
        res
    }
  }

  def visitAnnotation(node: jml.Annotation): ir.Annotation = {
    node match {
      case ann: jml.StringAnnotation => visitStringAnnotation(ann)
      case _ => ???
    }
  }

  def visitStringAnnotation(node: jml.StringAnnotation): ir.StringAnnotation = {
    node match {
      case jml.StringAnnotation(text, srcLocation, precedingWhitespace) =>
        val res = ir.StringAnnotation(text,
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitAnnotationTypeDeclaration(node: jml.AnnotationTypeDeclaration): ir.AnnotationTypeDeclaration = {
    node match {
      case jml.AnnotationTypeDeclaration(name, body, srcLocation, precedingWhitespace) =>
        val res = ir.AnnotationTypeDeclaration(visitStringNode(name),
                                                visitBody(body),
                                                srcLocation,
                                                precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitDefaultClause(node: jml.DefaultClause): ir.DefaultClause = {
    node match {
      case jml.DefaultClause(value, srcLocation, precedingWhitespace) =>
        val res = ir.DefaultClause(visitStringNode(value),
                                    srcLocation,
                                    precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitStatement(node: jml.Statement): ir.Statement = {
    try
      node match {
        case local: jml.LocalDeclaration => visitLocalDeclaration(local)
        case block: jml.Block => visitBlock(block)
        case assert: jml.JavaAssertion => visitJavaAssertion(assert)
        case ifStmt: jml.IfStatement => visitIfStatement(ifStmt)
        case forLoop: jml.ForLoop => visitForLoop(forLoop)
        case whileLoop: jml.WhileLoop => visitWhileLoop(whileLoop)
        case doLoop: jml.DoLoop => visitDoLoop(doLoop)
        case tryBlock: jml.TryBlock => visitTryBlock(tryBlock)
        case switch: jml.SwitchStatement => visitSwitchStatement(switch)
        case sync: jml.SynchronizedStatement => visitSynchronizedStatement(sync)
        case ret: jml.ReturnStatement => visitReturnStatement(ret)
        case throwStmt: jml.ThrowStatement => visitThrowStatement(throwStmt)
        case break: jml.BreakStatement => visitBreakStatement(break)
        case continue: jml.ContinueStatement => visitContinueStatement(continue)
        case empty: jml.EmptyStatement => visitEmptyStatement(empty)
        case expr: jml.ExpressionStatement => visitExpressionStatement(expr)
        case labeled: jml.LabeledStatement => visitLabeledStatement(labeled)
        case ghost: jml.GhostStatement => visitGhostStatement(ghost)
        case assert: jml.AssertStatement => visitAssertStatement(assert)
        case assume: jml.AssumeStatement => visitAssumeStatement(assume)
        case set: jml.SetStatement => visitSetStatement(set)
        case unreach: jml.UnreachableStatement => visitUnreachableStatement(unreach)
        case jmlBlock: jml.JmlStatementBlock => visitJmlStatementBlock(jmlBlock)
        case nontrans: jml.NonTranslatedStatement => visitNonTranslatedStatement(nontrans)
        case _ => ???
        // todo: ClassDecl, etc
      }
    catch {
      case e: NotSupportedException =>
        val text = Helper.commentUnknownNode(node, s"${e.construct} not supported, whole Statement commented out")
        ir.NonTranslatedStatement(text, node.srcLocation)
    }
  }

  def visitBlock(node: jml.Block): ir.Block = {
    node match {
      case jml.Block(static, statements, srcLocation, precedingWhitespace) =>
        val irStatic = static.map(visitModifier)
        val res = ir.Block(irStatic,
                            statements.map(visitStatement),
                            srcLocation,
                            precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitLocalDeclaration(node: jml.LocalDeclaration): ir.LocalDeclaration = {
    node match {
      case jml.LocalDeclaration(modifiers, declType, declarators, isStatement, srcLocation, precedingWhitespace) =>
        val res = ir.LocalDeclaration(modifiers.map(visitModifier),
                                      visitTypeIndication(declType),
                                      declarators.map(visitVariableDeclarator),
                                      isStatement,
                                      srcLocation,
                                      precedingWhitespace)
        if(!isNullableDefault && !hasNullableModifier(node.modifiers) && !hasNonnullModifier(node.modifiers)) {
          if(!nonNullableTypes.contains(declType.typeAsString)){
            var mod: ir.Modifier = ir.NonnullModifier(srcLocation, " ")
            if(wrapInSpec) {
              mod = ir.JmlModifierBlock(Some(ir.StringNode("/*@", srcLocation, " ")),
                List(mod),
                Some(ir.StringNode("*/", srcLocation, " ")),
                srcLocation)
            }
            if(res.modifiers.isEmpty) {
              mod.precedingWhitespace = res.declType.precedingWhitespace
              res.declType.precedingWhitespace = " "
            }
            res.modifiers = res.modifiers.appended(mod)
          }
        }
        res.copyWhitespace(node)
        res
    }
  }
  def visitJavaAssertion(node: jml.JavaAssertion): ir.JavaAssertion = {
    node match {
      case jml.JavaAssertion(content, message, srcLocation, precedingWhitespace) =>
        val irMsg = message.map(visitExpression)
        val res = ir.JavaAssertion(visitExpression(content),
                                    irMsg,
                                    srcLocation,
                                    precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitIfStatement(node: jml.IfStatement): ir.IfStatement = {
    node match {
      case jml.IfStatement(condition, thenStatement, elseStatement, srcLocation, precedingWhitespace) =>
        val irElse = elseStatement.map(visitStatement)
        val res = ir.IfStatement(visitExpression(condition),
                                  visitStatement(thenStatement),
                                  irElse,
                                  srcLocation,
                                  precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitForLoop(node: jml.ForLoop): ir.ForLoop = {
    node match {
      case jml.ForLoop(contractBefore, control, contractWithin, body, srcLocation, precedingWhitespace) =>
        val irContractBefore = if (contractBefore.isEmpty) Nil else contractBefore.map(visitLoopSpecOrLabel)
        val irContractWithin = if (contractWithin.isEmpty) Nil else contractWithin.map(visitLoopSpecOrLabel)
        val res = ir.ForLoop(irContractBefore,
                              visitForLoopControl(control),
                              irContractWithin,
                              visitStatement(body),
                              srcLocation,
                              precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitForLoopControl(node: jml.ForLoopControl): ir.ForLoopControl = {
    node match {
      case init: jml.ForEachLoopControl => visitForEachLoopControl(init)
      case init: jml.RegularForLoopControl => visitRegularForLoopControl(init)
    }
  }
  def visitForInit(node: jml.ForInit): ir.ForInit = {
    node match {
      case init: jml.ExpressionListForInit =>
        visitExpressionListForInit(init)
      case init: jml.LocalDeclaration =>
        visitLocalDeclaration(init)
      case _ =>
        ???
    }
  }
  def visitExpressionListForInit(node: jml.ExpressionListForInit): ir.ExpressionListForInit = {
    node match {
      case jml.ExpressionListForInit(expressions, srcLocation, precedingWhitespace) =>
        val res = ir.ExpressionListForInit(expressions.map(visitExpression),
                                          srcLocation,
                                          precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitRegularForLoopControl(node: jml.RegularForLoopControl): ir.RegularForLoopControl = {
    node match {
      case jml.RegularForLoopControl(init, condition, update, srcLocation, precedingWhitespace) =>
        val irInit = init.map(visitForInit)
        val irCondition = condition.map(visitExpression)
        val res = ir.RegularForLoopControl(irInit,
                                            irCondition,
                                            update.map(visitExpression),
                                            srcLocation,
                                            precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitForEachLoopControl(node: jml.ForEachLoopControl): ir.ForEachLoopControl = {
    node match {
      case jml.ForEachLoopControl(controlType, variable, control, srcLocation, precedingWhitespace) =>
        val res = ir.ForEachLoopControl(visitTypeIndication(controlType),
                                        visitStringNode(variable),
                                        visitExpression(control),
                                        srcLocation,
                                        precedingWhitespace)
        res.modifiers = node.modifiers.map(visitModifier)
        res.copyWhitespace(node)
        res
    }
  }

  def visitWhileLoop(node: jml.WhileLoop): ir.WhileLoop = {
    node match {
      case jml.WhileLoop(contractBefore, condition, contractWithin, body, srcLocation, precedingWhitespace) =>
        val irContractBefore = if (contractBefore.isEmpty) Nil else contractBefore.map(visitLoopSpecOrLabel)
        val irContractWithin = if (contractWithin.isEmpty) Nil else contractWithin.map(visitLoopSpecOrLabel)
        val res = ir.WhileLoop(irContractBefore,
                                visitExpression(condition),
                                irContractWithin,
                                visitStatement(body),
                                srcLocation,
                                precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitDoLoop(node: jml.DoLoop): ir.DoLoop = {
    node match {
      case jml.DoLoop(contractBefore, body, condition, srcLocation, precedingWhitespace) =>
        val irContractBefore = if (contractBefore.isEmpty) Nil else contractBefore.map(visitLoopSpecOrLabel)
        val res = ir.DoLoop(irContractBefore,
                            visitStatement(body),
                            visitExpression(condition),
                            srcLocation,
                            precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitTryBlock(node: jml.TryBlock): ir.TryBlock = {
    node match {
      case jml.TryBlock(resourceSpecification, tryBlock, catchClauses, finallyBlock, srcLocation, precedingWhitespace) =>
        val irResSpec = if (resourceSpecification.isEmpty) None
                        else Some(visitTryResourceSpecification(resourceSpecification.get))
        val irFinally = finallyBlock.map(visitBlock)
        val res = ir.TryBlock(irResSpec,
                              visitBlock(tryBlock),
                              catchClauses.map(visitCatchClause),
                              irFinally,
                              srcLocation,
                              precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitCatchClause(node: jml.CatchClause): ir.CatchClause = {
    node match {
      case jml.CatchClause(types, name, block, srcLocation, precedingWhitespace) =>
        val res = ir.CatchClause(types.map(visitTypeIndication),
                                  visitStringNode(name),
                                  visitBlock(block),
                                  srcLocation,
                                  precedingWhitespace)
        res.modifiers = node.modifiers.map(visitModifier)
        res.copyWhitespace(node)
        res
    }
  }
  def visitTryResourceSpecification(node: jml.TryResourceSpecification): ir.TryResourceSpecification = {
    node match {
      case jml.TryResourceSpecification(resources, trailingSemicolon, srcLocation, precedingWhitespace) =>
        val irSemi = trailingSemicolon.map(visitStringNode)
        val res = ir.TryResourceSpecification(resources.map(visitTryResource),
                                              irSemi,
                                              srcLocation,
                                              precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitTryResource(node: jml.TryResource): ir.TryResource = {
    node match {
      case jml.TryResource(modifiers, resourceType, variable, expression, srcLocation, precedingWhitespace) =>
        val res = ir.TryResource(modifiers.map(visitModifier),
                                  visitTypeIndication(resourceType),
                                  visitStringNode(variable),
                                  visitExpression(expression),
                                  srcLocation,
                                  precedingWhitespace)
        res.modifiers = node.modifiers.map(visitModifier)
        res.copyWhitespace(node)
        res
    }
  }

  def visitSwitchStatement(node: jml.SwitchStatement): ir.SwitchStatement = {
    node match {
      case jml.SwitchStatement(condition, cases, srcLocation, precedingWhitespace) =>
        val res = ir.SwitchStatement(visitExpression(condition),
                                      cases.map(visitSwitchCaseBlock),
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitSwitchCaseBlock(node: jml.SwitchCaseBlock): ir.SwitchCaseBlock = {
    node match {
      case jml.SwitchCaseBlock(labels, statements, srcLocation, precedingWhitespace) =>
        val res = ir.SwitchCaseBlock(labels.map(visitSwitchLabel),
                                      statements.map(visitStatement),
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitSwitchLabel(node: jml.SwitchLabel): ir.SwitchLabel = {
    node match {
      case jml.SwitchLabel(expression, srcLocation, precedingWhitespace) =>
        val irExpr = expression.map(visitExpression)
        val res = ir.SwitchLabel(irExpr,
                                  srcLocation,
                                  precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitSynchronizedStatement(node: jml.SynchronizedStatement): ir.SynchronizedStatement = {
    node match {
      case jml.SynchronizedStatement(condition, body, srcLocation, precedingWhitespace) =>
        val res = ir.SynchronizedStatement(visitExpression(condition),
                                            visitStatement(body),
                                            srcLocation,
                                            precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitReturnStatement(node: jml.ReturnStatement): ir.ReturnStatement = {
    node match {
      case jml.ReturnStatement(expression, srcLocation, precedingWhitespace) =>
        val irExpr = expression.map(visitExpression)
        val res = ir.ReturnStatement(irExpr,
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitThrowStatement(node: jml.ThrowStatement): ir.ThrowStatement = {
    node match {
      case jml.ThrowStatement(expression, srcLocation, precedingWhitespace) =>
        val res = ir.ThrowStatement(visitExpression(expression),
                                    srcLocation,
                                    precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitBreakStatement(node: jml.BreakStatement): ir.BreakStatement = {
    node match {
      case jml.BreakStatement(label, srcLocation, precedingWhitespace) =>
        val irLabel = label.map(visitStringNode)
        val res = ir.BreakStatement(irLabel,
                                    srcLocation,
                                    precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitContinueStatement(node: jml.ContinueStatement): ir.ContinueStatement = {
    node match {
      case jml.ContinueStatement(label, srcLocation, precedingWhitespace) =>
        val irLabel = label.map(visitStringNode)
        val res = ir.ContinueStatement(irLabel,
                                        srcLocation,
                                        precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitEmptyStatement(node: jml.EmptyStatement): ir.EmptyStatement = {
    val res = new ir.EmptyStatement(node.srcLocation)
    res.copyWhitespace(node)
    res
  }
  def visitExpressionStatement(node: jml.ExpressionStatement): ir.ExpressionStatement = {
    node match {
      case jml.ExpressionStatement(expression, srcLocation, precedingWhitespace) =>
        val res = ir.ExpressionStatement(visitExpression(expression),
                                          srcLocation,
                                          precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitLabeledStatement(node: jml.LabeledStatement): ir.LabeledStatement = {
    node match {
      case jml.LabeledStatement(label, statement, srcLocation, precedingWhitespace) =>
        val res = ir.LabeledStatement(visitLabel(label),
                                      visitStatement(statement),
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitLabel(node: jml.Label): ir.Label = {
    node match {
      case jml.Label(label, srcLocation, precedingWhitespace) =>
        val res = ir.Label(visitStringNode(label),
                            srcLocation,
                            precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitArguments(node: jml.Arguments): ir.Arguments = {
    node match {
      case jml.Arguments(arguments, srcLocation, precedingWhitespace) =>
        val res = ir.Arguments(arguments.map(visitExpression),
                                srcLocation,
                                precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }


  def visitExpression(node: jml.Expression): ir.Expression = {
    node match {
      case expr: jml.NestedInvocationExpression => visitNestedInvocationExpression(expr)
      case expr: jml.BinaryExpression => visitBinaryExpression(expr)
      case expr: jml.InstanceofExpression => visitInstanceofExpression(expr)
      case expr: jml.ArrayIndexExpression => visitArrayIndexExpression(expr)
      case expr: jml.MethodCallExpression => visitMethodCallExpression(expr)
      case expr: jml.NewObjectExpression => visitNewObjectExpression(expr)
      case expr: jml.NewArrayExpression => visitNewArrayExpression(expr)
      case expr: jml.CastExpression => visitCastExpression(expr)
      case expr: jml.PostfixExpression => visitPostfixExpression(expr)
      case expr: jml.PrefixExpression => visitPrefixExpression(expr)
      case expr: jml.TernaryExpression => visitTernaryExpression(expr)
      case expr: jml.LambdaExpression => visitLambdaExpression(expr)
      case expr: jml.MethodReferenceExpression => visitMethodReferenceExpression(expr)
      case expr: jml.ParExpression => visitParExpression(expr)
      case expr: jml.Literal => visitLiteral(expr)
      case expr: jml.ClassName => visitClassName(expr)
      case expr: jml.GenericInvocationExpression => visitGenericInvocationExpression(expr)
      case expr: jml.ResultExpression => visitResultExpression(expr)
      case expr: jml.OldExpression => visitOldExpression(expr)
      case expr: jml.PreExpression => visitPreExpression(expr)
      case expr: jml.NotAssignedExpression => visitNotAssignedExpression(expr)
      case expr: jml.NotModifiedExpression => visitNotModifiedExpression(expr)
      case expr: jml.OnlyAccessedExpression => visitOnlyAccessedExpression(expr)
      case expr: jml.OnlyAssignedExpression => visitOnlyAssignedExpression(expr)
      case expr: jml.FreshExpression => visitFreshExpression(expr)
      case expr: jml.NonNullElementsExpression => visitNonNullElementsExpression(expr)
      case expr: jml.TypeOfExpression => visitTypeOfExpression(expr)
      case expr: jml.TypeExpression => visitTypeExpression(expr)
      case expr: jml.ForallExpression => visitForallExpression(expr)
      case expr: jml.ExistsExpression => visitExistsExpression(expr)
      case expr: jml.MaxQuantifiedExpression => visitMaxQuantifiedExpression(expr)
      case expr: jml.MinExpression => visitMinExpression(expr)
      case expr: jml.NumOfExpression => visitNumOfExpression(expr)
      case expr: jml.ProductExpression => visitProductExpression(expr)
      case expr: jml.SumExpression => visitSumExpression(expr)
      case expr: jml.StringNode => visitStringNode(expr)
      case expr: jml.SetComprehension => visitSetComprehension(expr)
      case expr: jml.InformalDescription => throw new NotSupportedException("informal description", expr.srcLocation)
      case _ => ???
    }
    //todo: more?

  }

  def visitNestedInvocationExpression(node: jml.NestedInvocationExpression): ir.NestedInvocationExpression = {
    node match {
      case jml.NestedInvocationExpression(left, right, srcLocation, precedingWhitespace) =>
        val res = ir.NestedInvocationExpression(visitExpression(left),
                                                visitExpression(right),
                                                srcLocation,
                                                precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitBinaryExpression(node: jml.BinaryExpression): ir.BinaryExpression = {
    node match {
      case jml.BinaryExpression(left, operator, right, srcLocation, precedingWhitespace) =>
        var irLeft = visitExpression(left)
        var irRight = visitExpression(right)
        val op = visitStringNode(operator)
        if(operator.text == "<==>" || operator.text == "<=!=>") {
          val pleft = ir.ParExpression(irLeft, irLeft.srcLocation)
          pleft.copyWhitespace(irLeft)
          irLeft.precedingWhitespace = ""
          irLeft.trailingWhitespace = ""
          irLeft = pleft
          val pright = ir.ParExpression(irRight, irRight.srcLocation)
          pright.copyWhitespace(irRight)
          irRight.precedingWhitespace = ""
          irRight.trailingWhitespace = ""
          irRight = pright
          if(op.text == "<==>") {
            op.text = "=="
          } else {
            op.text = "!="
          }
        }
        val res = ir.BinaryExpression(irLeft,
                                      op,
                                      irRight,
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitInstanceofExpression(node: jml.InstanceofExpression): ir.InstanceofExpression = {
    node match {
      case jml.InstanceofExpression(left, right, srcLocation, precedingWhitespace) =>
        val res = ir.InstanceofExpression(visitExpression(left),
          visitTypeIndication(right),
          srcLocation)
        res.copyWhitespace(node)
        res
    }
  }

  def visitArrayIndexExpression(node: jml.ArrayIndexExpression): ir.ArrayIndexExpression = {
    node match {
      case jml.ArrayIndexExpression(arrayExpression, indexExpression, srcLocation, precedingWhitespace) =>
        val res = ir.ArrayIndexExpression(visitExpression(arrayExpression),
                                          visitIndexExpression(indexExpression),
                                          srcLocation,
                                          precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitIndexExpression(node: jml.IndexExpression): ir.IndexExpression = {
    node match {
      case idx: jml.IndexAllExpression => visitIndexAllExpression(idx)
      case idx: jml.RangeExpression => visitRangeExpression(idx)
      case idx: jml.Expression => visitExpression(idx)
      case _ => ???
    }
  }

  def visitIndexAllExpression(node: jml.IndexAllExpression): ir.IndexAllExpression = {
    node match {
      case jml.IndexAllExpression(srcLocation, precedingWhitespace) =>
        val res = ir.IndexAllExpression(srcLocation, precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitRangeExpression(node: jml.RangeExpression): ir.RangeExpression = {
    node match {
      case jml.RangeExpression(start, end, srcLocation, precedingWhitespace) =>
        val res = ir.RangeExpression(visitExpression(start),
                                      visitExpression(end),
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitMethodCallExpression(node: jml.MethodCallExpression): ir.MethodCallExpression = {
    node match {
      case jml.MethodCallExpression(name, arguments, srcLocation, precedingWhitespace) =>
        val res = ir.MethodCallExpression(visitStringNode(name),
                                          visitArguments(arguments),
                                          srcLocation,
                                          precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitNewObjectExpression(node: jml.NewObjectExpression): ir.NewObjectExpression = {
    node match {
      case jml.NewObjectExpression(typeArguments, typeIndication, arguments, body, srcLocation, precedingWhitespace) =>
        val typeArgs = typeArguments.map(visitTypeArgumentClause)
        val irBody = body.map(visitBody)
        val res = ir.NewObjectExpression(typeArgs,
                                          visitTypeIndication(typeIndication),
                                          visitArguments(arguments),
                                          irBody,
                                          srcLocation,
                                          precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitNewArrayExpression(node: jml.NewArrayExpression): ir.NewArrayExpression = {
    node match {
      case jml.NewArrayExpression(typeIndication, dimensions, initializer, srcLocation, precedingWhitespace) =>
        val irInit = initializer.map(visitArrayInitializer)
        val res = ir.NewArrayExpression(visitTypeIndication(typeIndication),
                                        dimensions.map(visitDimension),
                                        irInit,
                                        srcLocation,
                                        precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitDimension(node: jml.Dimension): ir.Dimension = {
    node match {
      case jml.Dimension(sizeExpression, srcLocation, precedingWhitespace) =>
        val irSize = sizeExpression.map(visitExpression)
        val res = ir.Dimension(irSize,
                                srcLocation,
                                precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitCastExpression(node: jml.CastExpression): ir.CastExpression = {
    node match {
      case jml.CastExpression(annotations, typeIndications, expression, srcLocation, precedingWhitespace) =>
        val res = ir.CastExpression(annotations.map(visitStringAnnotation),
                                    typeIndications.map(visitTypeIndication),
                                    visitExpression(expression),
                                    srcLocation,
                                    precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitPostfixExpression(node: jml.PostfixExpression): ir.PostfixExpression = {
    node match {
      case jml.PostfixExpression(expression, postfix, srcLocation, precedingWhitespace) =>
        val res = ir.PostfixExpression(visitExpression(expression),
                                        visitStringNode(postfix),
                                        srcLocation,
                                        precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitPrefixExpression(node: jml.PrefixExpression): ir.PrefixExpression = {
    node match {
      case jml.PrefixExpression(expression, prefix, srcLocation, precedingWhitespace) =>
        val res = ir.PrefixExpression(visitExpression(expression),
                                      visitStringNode(prefix),
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitTernaryExpression(node: jml.TernaryExpression): ir.TernaryExpression = {
    node match {
      case jml.TernaryExpression(condition, thenExpression, elseExpression, srcLocation, precedingWhitespace) =>
        val res = ir.TernaryExpression(visitExpression(condition),
                                        visitExpression(thenExpression),
                                        visitExpression(elseExpression),
                                        srcLocation,
                                        precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitLambdaExpression(node: jml.LambdaExpression): ir.LambdaExpression = {
    node match {
      case jml.LambdaExpression(hasParentheses, parameters, body, srcLocation, precedingWhitespace) =>
        val res = ir.LambdaExpression(hasParentheses,
                                      parameters.map(visitParameter),
                                      visitLambdaBody(body),
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitLambdaBody(node: jml.LambdaBody): ir.LambdaBody = {
    node match {
      case body: jml.Expression => visitExpression(body)
      case _ => ???
    }
  }

  def visitMethodReferenceExpression(node: jml.MethodReferenceExpression): ir.MethodReferenceExpression = {
    node match {
      case jml.MethodReferenceExpression(left, typeArgument, right, srcLocation, precedingWhitespace) =>
        val irArg = typeArgument.map(visitTypeArgumentClause)
        val res = ir.MethodReferenceExpression(visitExpression(left),
                                                irArg,
                                                visitExpression(right),
                                                srcLocation,
                                                precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitParExpression(node: jml.ParExpression): ir.ParExpression = {
    node match {
      case jml.ParExpression(expression, srcLocation, precedingWhitespace) =>
        val res = ir.ParExpression(visitExpression(expression),
                                    srcLocation,
                                    precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitLiteral(node: jml.Literal): ir.Literal = {
    node match {
      case jml.Literal(value, srcLocation, precedingWhitespace) =>
        val res = ir.Literal(visitStringNode(value),
                              srcLocation,
                              precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitClassName(node: jml.ClassName): ir.ClassName = {
    node match {
      case jml.ClassName(typeIndication, srcLocation, precedingWhitespace) =>
        val res = ir.ClassName(visitTypeIndication(typeIndication),
                                srcLocation,
                                precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitGenericInvocationExpression(node: jml.GenericInvocationExpression): ir.GenericInvocationExpression = {
    node match {
      case jml.GenericInvocationExpression(typeArgumentClause, name, arguments, srcLocation, precedingWhitespace) =>
        val irArgs = arguments.map(visitArguments)
        val res = ir.GenericInvocationExpression(visitTypeArgumentClause(typeArgumentClause),
                                                  visitStringNode(name),
                                                  irArgs,
                                                  srcLocation,
                                                  precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitJmlModifierBlock(node: jml.JmlModifierBlock): ir.JmlModifierBlock = {
    node match {
      case jml.JmlModifierBlock(startSpec, modifiers, endSpec, srcLocation, precedingWhitespace) =>
        val irStart = startSpec.map(visitStringNode)
        val irEnd = endSpec.map(visitStringNode)
        val oldWrapInSpec = wrapInSpec
        wrapInSpec = false
        val res = ir.JmlModifierBlock(irStart,
                                      modifiers.map(visitModifier),
                                      irEnd,
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        wrapInSpec = oldWrapInSpec
        res
    }
  }

  def visitPureModifier(node: jml.PureModifier): ir.PureModifier = {
    node match {
      case jml.PureModifier(srcLocation, precedingWhitespace) =>
        val res = ir.PureModifier(srcLocation, precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitNullableModifier(node: jml.NullableModifier): ir.NullableModifier = {
    node match {
      case jml.NullableModifier(srcLocation, precedingWhitespace) =>
        val res = ir.NullableModifier(srcLocation, precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitNullableByDefaultModifier(node: jml.NullableByDefaultModifier): ir.ModifierStringNode = {
    node match {
      case jml.NullableByDefaultModifier(srcLocation, precedingWhitespace) =>
        val res = ir.ModifierStringNode("", srcLocation, precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitNonnullModifier(node: jml.NonnullModifier): ir.NonnullModifier = {
    node match {
      case jml.NonnullModifier(srcLocation, precedingWhitespace) =>
        val res = ir.NonnullModifier(srcLocation, precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitGhostModifier(node: jml.GhostModifier): ir.GhostModifier = {
    node match {
      case jml.GhostModifier(srcLocation, precedingWhitespace) =>
        val res = ir.GhostModifier(srcLocation, precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitHelperModifier(node: jml.HelperModifier): ir.HelperModifier = {
    node match {
      case jml.HelperModifier(srcLocation, precedingWhitespace) =>
        val res = ir.HelperModifier(srcLocation, precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitMonitoredModifier(node: jml.MonitoredModifier): ir.MonitoredModifier = {
    node match {
      case jml.MonitoredModifier(srcLocation, precedingWhitespace) =>
        val res = ir.MonitoredModifier(srcLocation, precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitModelModifier(node: jml.ModelModifier): ir.ModelModifier = {
    node match {
      case jml.ModelModifier(srcLocation, precedingWhitespace) =>
        val res = ir.ModelModifier(srcLocation, precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitSpecMethodModifier(node: jml.SpecMethodModifier): ir.SpecMethodModifier = {
    node match {
      case jml.SpecMethodModifier(srcLocation, precedingWhitespace) =>
        val res = ir.SpecMethodModifier(srcLocation, precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitInstanceModifier(node: jml.InstanceModifier): ir.InstanceModifier = {
    node match {
      case jml.InstanceModifier(srcLocation, precedingWhitespace) =>
        val res = ir.InstanceModifier(srcLocation)
        res.copyWhitespace(node)
        res
    }
  }

  def visitSpecPublicModifier(node: jml.SpecPublicModifier): ir.ModifierStringNode = {
    node match {
      case jml.SpecPublicModifier(srcLocation, precedingWhitespace) =>
        System.err.println(s"[warning] spec_public modifier not supported ${srcLocation.toString}, dropped.")
        val res = ir.ModifierStringNode("", srcLocation)
        res.copyWhitespace(node)
        res
    }
  }
  def visitSpecProtectedModifier(node: jml.SpecProtectedModifier): ir.ModifierStringNode = {
    node match {
      case jml.SpecProtectedModifier(srcLocation, precedingWhitespace) =>
        System.err.println(s"[warning] spec_protected modifier not supported ${srcLocation.toString}, dropped.")
        val res = ir.ModifierStringNode("", srcLocation)
        res.copyWhitespace(node)
        res
    }
  }

  def visitContract(node: jml.Contract): ir.Contract = {
    node match {
      case jml.Contract(clauses, srcLocation, precedingWhitespace) =>
        val res = ir.Contract(clauses.map(visitContractClause),
                              srcLocation,
                              precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitSpecCase(node: jml.SpecCase): ir.SpecCase = {
    node match {
      case spec: jml.NormalBehavior => visitNormalBehavior(spec)
      case spec: jml.ExceptionalBehavior => visitExceptionalBehavior(spec)
      case spec: jml.Behavior => visitBehavior(spec)
      case spec: jml.SpecCase => visitLightSpecCase(spec)
      case _ => ???
    }
  }

  def visitLightSpecCase(node: jml.SpecCase): ir.SpecCase = {
    node match {
      case jml.SpecCase(startSpec, hasAlso, clauses, endSpec, srcLocation, precedingWhitespace) =>
        val irStart = startSpec.map(visitStringNode)
        val irEnd = endSpec.map(visitStringNode)
        val res = ir.SpecCase(irStart,
                                    hasAlso,
                                    clauses.map(visitContractClause),
                                    irEnd,
                                    srcLocation,
                                    precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitLoopSpecOrLabel(node: jml.LoopSpecOrLabel): ir.LoopSpecOrLabel = {
    node match {
      case l: jml.Label => visitLabel(l)
      case spec: jml.LoopSpecification => visitLoopSpecification(spec)
    }
  }

  def visitLoopSpecification(node: jml.LoopSpecification): ir.LoopSpecification = {
    node match {
      case jml.LoopSpecification(startSpec, clauses, endSpec, srcLocation, precedingWhitespace) =>
        val irClauses = clauses.map(visitLoopClause)
        val irStart = startSpec.map(visitStringNode)
        val irEnd = endSpec.map(visitStringNode)
        val res = ir.LoopSpecification(irStart, irClauses, irEnd, srcLocation, precedingWhitespace)
        res
    }
  }

  def visitContractClause(node: jml.ContractClause): ir.ContractClause = {
    try
      val res = node match {
        case clause: jml.AssignableClause => visitAssignableClause(clause)
        case clause: jml.AccessibleClause => visitAccessibleClause(clause)
        case clause: jml.ForallVariableDeclaration => visitForallVariableDeclaration(clause)
        case clause: jml.OldVariableDeclaration => visitOldVariableDeclaration(clause)
        case clause: jml.Precondition => visitPrecondition(clause)
        case clause: jml.Postcondition => visitPostcondition(clause)
        case clause: jml.DivergeClause => visitDivergeClause(clause)
        case clause: jml.MeasuredByClause => visitMeasuredByClause(clause)
        case clause: jml.ExceptionalPostcondition => visitExceptionalPostcondition(clause)
        case clause: jml.SignalsOnlyClause => visitSignalsOnlyClause(clause)
        case clause: jml.NonTranslatedContractClause => visitNonTranslatedContractClause(clause)
        case clause: jml.SpecCase => visitSpecCase(clause)
        case _ => ???
      }
      res.startSpec = node.startSpec.map(visitStringNode)
      res.endSpec = node.endSpec.map(visitStringNode)
      res
    catch {
      case e: NotSupportedException =>
        val text = Helper.commentUnknownNode(node, s"${e.construct} not supported, whole ContractClause commented out")
        ir.NonTranslatedContractClause(text, node.srcLocation)
    }
  }

  def visitAssignableClause(node: jml.AssignableClause): ir.AssignableClause = {
    node match {
      case jml.AssignableClause(expressions, srcLocation, precedingWhitespace) =>
        val res = ir.AssignableClause(expressions.map(visitExpression),
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitAccessibleClause(node: jml.AccessibleClause): ir.AccessibleClause = {
    node match {
      case jml.AccessibleClause(expressions, srcLocation, precedingWhitespace) =>
        val res = ir.AccessibleClause(expressions.map(visitExpression),
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitForallVariableDeclaration(node: jml.ForallVariableDeclaration): ir.ForallVariableDeclaration = {
    node match {
      case jml.ForallVariableDeclaration(declaration, srcLocation, precedingWhitespace) =>
        val res = ir.ForallVariableDeclaration(visitQuantifiedVarDeclaration(declaration),
                                                srcLocation,
                                                precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitOldVariableDeclaration(node: jml.OldVariableDeclaration): ir.OldVariableDeclaration = {
    node match {
      case jml.OldVariableDeclaration(modifiers, typeIndication, declarators, srcLocation, precedingWhitespace) =>
        val res = ir.OldVariableDeclaration(modifiers.map(visitModifier),
                                            visitTypeIndication(typeIndication),
                                            declarators.map(visitVariableDeclarator),
                                            srcLocation,
                                            precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitNonTranslatedContractClause(node: jml.NonTranslatedContractClause): ir.NonTranslatedContractClause = {
    node match {
      case jml.NonTranslatedContractClause(text, srcLocation, precedingWhitespace) =>
        val res = ir.NonTranslatedContractClause(text,
          srcLocation,
          precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitPrecondition(node: jml.Precondition): ir.Precondition = {
    node match {
      case jml.Precondition(expression, srcLocation, precedingWhitespace) =>
        val res = ir.Precondition(visitExpression(expression),
                                  srcLocation,
                                  precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitPostcondition(node: jml.Postcondition): ir.Postcondition = {
    node match {
      case jml.Postcondition(expression, srcLocation, precedingWhitespace) =>
        val res = ir.Postcondition(visitExpression(expression),
                                    srcLocation,
                                    precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitDivergeClause(node: jml.DivergeClause): ir.DivergeClause = {
    node match {
      case jml.DivergeClause(expression, srcLocation, precedingWhitespace) =>
        val res = ir.DivergeClause(visitExpression(expression),
                                    srcLocation,
                                    precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitLoopClause(node: jml.LoopClause): ir.LoopClause = {
    try
      node match {
        case clause: jml.LoopModifiesClause => visitLoopModifiesClause(clause)
        case clause: jml.LoopInvariant => visitLoopInvariant(clause)
        case clause: jml.DecreasesClause => visitDecreasesClause(clause)
        case clause: jml.NonTranslatedLoopClause => visitNonTranslatedLoopClause(clause)
        case _ => ???
      }
    catch {
      case e: NotSupportedException =>
        val text = Helper.commentUnknownNode(node, s"${e.construct} not supported, whole LoopClause commented out")
        ir.NonTranslatedLoopClause(text, node.srcLocation)
    }
  }

  def visitLoopModifiesClause(node: jml.LoopModifiesClause): ir.LoopModifiesClause = {
    node match {
      case jml.LoopModifiesClause(expressions, srcLocation, precedingWhitespace) =>
        val res = ir.LoopModifiesClause(expressions.map(visitExpression),
                                    srcLocation,
                                    precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitLoopInvariant(node: jml.LoopInvariant): ir.LoopInvariant = {
    node match {
      case jml.LoopInvariant(expression, srcLocation, precedingWhitespace) =>
        val res = ir.LoopInvariant(visitExpression(expression),
                                    srcLocation,
                                    precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitDecreasesClause(node: jml.DecreasesClause): ir.DecreasesClause = {
    node match {
      case jml.DecreasesClause(expression, srcLocation, precedingWhitespace) =>
        val res = ir.DecreasesClause(visitExpression(expression),
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitNonTranslatedLoopClause(node: jml.NonTranslatedLoopClause): ir.NonTranslatedLoopClause = {
    node match {
      case jml.NonTranslatedLoopClause(text, srcLocation, precedingWhitespace) =>
        val res = ir.NonTranslatedLoopClause(text,
          srcLocation,
          precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitMeasuredByClause(node: jml.MeasuredByClause): ir.MeasuredByClause = {
    node match {
      case jml.MeasuredByClause(expression, predicate, srcLocation, precedingWhitespace) =>
        val irPred = predicate.map(visitExpression)
        val res = ir.MeasuredByClause(visitExpression(expression),
                                      irPred,
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitExceptionalPostcondition(node: jml.ExceptionalPostcondition): ir.ExceptionalPostcondition = {
    node match {
      case jml.ExceptionalPostcondition(exceptionType, exceptionName, expression, srcLocation, precedingWhitespace) =>
        val res = ir.ExceptionalPostcondition(visitTypeIndication(exceptionType),
                                              exceptionName.map(visitStringNode),
                                              expression.map(visitExpression),
                                              srcLocation,
                                              precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitSignalsOnlyClause(node: jml.SignalsOnlyClause): ir.SignalsOnlyClause = {
    node match {
      case jml.SignalsOnlyClause(exceptionTypes, srcLocation, precedingWhitespace) =>
        val res = ir.SignalsOnlyClause(exceptionTypes.map(visitTypeIndication),
                                        srcLocation,
                                        precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitBehavior(node: jml.Behavior): ir.Behavior = {
    node match {
      case jml.Behavior(startSpec, hasAlso, privacy, clauses, endSpec, srcLocation, precedingWhitespace) =>
        val irStart = startSpec.map(visitStringNode)
        val irPrivacy = privacy.map(visitStringNode)
        val irEnd = endSpec.map(visitStringNode)
        val res = ir.Behavior(irStart,
                              hasAlso,
                              irPrivacy,
                              clauses.map(visitContractClause),
                              irEnd,
                              srcLocation,
                              precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitNormalBehavior(node: jml.NormalBehavior): ir.NormalBehavior = {
    node match {
      case jml.NormalBehavior(startSpec, hasAlso, privacy, clauses, endSpec, srcLocation, precedingWhitespace) =>
        val irStart = startSpec.map(visitStringNode)
        val irPrivacy = privacy.map(visitStringNode)
        val irEnd = endSpec.map(visitStringNode)
        val res = ir.NormalBehavior(irStart,
                                    hasAlso,
                                    irPrivacy,
                                    clauses.map(visitContractClause),
                                    irEnd,
                                    srcLocation,
                                    precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitExceptionalBehavior(node: jml.ExceptionalBehavior): ir.ExceptionalBehavior = {
    node match {
      case jml.ExceptionalBehavior(startSpec, hasAlso, privacy, clauses, endSpec, srcLocation, precedingWhitespace) =>
        val irStart = startSpec.map(visitStringNode)
        val irPrivacy = privacy.map(visitStringNode)
        val irEnd = endSpec.map(visitStringNode)
        val res = ir.ExceptionalBehavior(irStart,
                                          hasAlso,
                                          irPrivacy,
                                          clauses.map(visitContractClause),
                                          irEnd,
                                          srcLocation,
                                          precedingWhitespace)
        res.copyWhitespace(node)
        res
      }
    }

  def visitGhostStatement(node: jml.GhostStatement): ir.GhostStatement = {
    node match {
      case jml.GhostStatement(statement, srcLocation, precedingWhitespace) =>
        val res = ir.GhostStatement(visitStatement(statement), srcLocation)
        res.copyWhitespace(node)
        res
    }
  }

  def visitAssertStatement(node: jml.AssertStatement): ir.AssertStatement = {
    node match {
      case jml.AssertStatement(content, message, srcLocation, precedingWhitespace) =>
        val irMsg = message.map(visitExpression)
        val res = ir.AssertStatement(visitExpression(content),
                                      irMsg,
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitAssumeStatement(node: jml.AssumeStatement): ir.AssumeStatement = {
    node match {
      case jml.AssumeStatement(content, message, srcLocation, precedingWhitespace) =>
        val irMsg = message.map(visitExpression)
        val res = ir.AssumeStatement(visitExpression(content),
                                      irMsg,
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitUnreachableStatement(node: jml.UnreachableStatement): ir.UnreachableStatement = {
    node match {
      case jml.UnreachableStatement(srcLocation, precedingWhitespace) =>
        val res = ir.UnreachableStatement(srcLocation, precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitSetStatement(node: jml.SetStatement): ir.SetStatement = {
    node match {
      case jml.SetStatement(assignment, srcLocation, precedingWhitespace) =>
        val res = ir.SetStatement(visitExpression(assignment),
                                  srcLocation,
                                  precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitJmlStatementBlock(node: jml.JmlStatementBlock): ir.JmlStatementBlock = {
    node match {
      case jml.JmlStatementBlock(startSpec, statements, endSpec, srcLocation, precedingWhitespace) =>
        val irStart = startSpec.map(visitStringNode)
        val irEnd = endSpec.map(visitStringNode)
        val oldWrapInSpec = wrapInSpec
        wrapInSpec = false
        val res = ir.JmlStatementBlock(irStart,
                                        statements.map(visitStatement),
                                        irEnd,
                                        srcLocation,
                                        precedingWhitespace)
        res.copyWhitespace(node)
        wrapInSpec = oldWrapInSpec
        res
    }
  }

  def visitNonTranslatedStatement(node: jml.NonTranslatedStatement): ir.NonTranslatedStatement = {
    node match {
      case jml.NonTranslatedStatement(text, srcLocation, precedingWhitespace) =>
        val res = ir.NonTranslatedStatement(text,
          srcLocation,
          precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitResultExpression(node: jml.ResultExpression): ir.ResultExpression = {
    node match {
      case jml.ResultExpression(src, precedingWhitespace) =>
        val res = ir.ResultExpression(src, precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitOldExpression(node: jml.OldExpression): ir.OldExpression = {
    node match {
      case jml.OldExpression(expression, label, srcLocation, precedingWhitespace) =>
        val irLabel = label.map(visitStringNode)
        val res = ir.OldExpression(visitExpression(expression),
                                    irLabel,
                                    srcLocation,
                                    precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitPreExpression(node: jml.PreExpression): ir.PreExpression = {
    node match {
      case jml.PreExpression(expression, srcLocation, precedingWhitespace) =>
        val res = ir.PreExpression(visitExpression(expression),
                                    srcLocation,
                                    precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitNotAssignedExpression(node: jml.NotAssignedExpression): ir.NotAssignedExpression = {
    node match {
      case jml.NotAssignedExpression(storeRefs, srcLocation, precedingWhitespace) =>
        val res = ir.NotAssignedExpression(storeRefs.map(visitExpression),
                                            srcLocation,
                                            precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitNotModifiedExpression(node: jml.NotModifiedExpression): ir.NotModifiedExpression = {
    node match {
      case jml.NotModifiedExpression(storeRefs, srcLocation, precedingWhitespace) =>
        val res = ir.NotModifiedExpression(storeRefs.map(visitExpression),
                                            srcLocation,
                                            precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitOnlyAccessedExpression(node: jml.OnlyAccessedExpression): ir.OnlyAccessedExpression = {
    node match {
      case jml.OnlyAccessedExpression(storeRefs, srcLocation, precedingWhitespace) =>
        val res = ir.OnlyAccessedExpression(storeRefs.map(visitExpression),
                                            srcLocation,
                                            precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitOnlyAssignedExpression(node: jml.OnlyAssignedExpression): ir.OnlyAssignedExpression = {
    node match {
      case jml.OnlyAssignedExpression(storeRefs, srcLocation, precedingWhitespace) =>
        val res = ir.OnlyAssignedExpression(storeRefs.map(visitExpression),
                                            srcLocation,
                                            precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitFreshExpression(node: jml.FreshExpression): ir.FreshExpression = {
    node match {
      case jml.FreshExpression(expressions, srcLocation, precedingWhitespace) =>
        val res = ir.FreshExpression(expressions.map(visitExpression),
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitNonNullElementsExpression(node: jml.NonNullElementsExpression): ir.NonNullElementsExpression = {
    node match {
      case jml.NonNullElementsExpression(expression, srcLocation, precedingWhitespace) =>
        val res = ir.NonNullElementsExpression(visitExpression(expression),
                                                srcLocation,
                                                precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitTypeOfExpression(node: jml.TypeOfExpression): ir.TypeOfExpression = {
    node match {
      case jml.TypeOfExpression(expression, srcLocation, precedingWhitespace) =>
        val res = ir.TypeOfExpression(visitExpression(expression),
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitTypeExpression(node: jml.TypeExpression): ir.TypeExpression = {
    node match {
      case jml.TypeExpression(indication, srcLocation, precedingWhitespace) =>
        val res = ir.TypeExpression(visitTypeIndication(indication),
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }


  def visitForallExpression(node: jml.ForallExpression): ir.ForallExpression = {
    node match {
      case jml.ForallExpression(declaration, guard, expression, srcLocation, precedingWhitespace) =>
        val irGuard = guard.map(visitExpression)
        val res = ir.ForallExpression(List(visitQuantifiedVarDeclaration(declaration)),
                                      irGuard,
                                      visitExpression(expression),
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitExistsExpression(node: jml.ExistsExpression): ir.ExistsExpression = {
    node match {
      case jml.ExistsExpression(declaration, guard, expression, srcLocation, precedingWhitespace) =>
        val irGuard = guard.map(visitExpression)
        val res = ir.ExistsExpression(List(visitQuantifiedVarDeclaration(declaration)),
                                      irGuard,
                                      visitExpression(expression),
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitMaxQuantifiedExpression(node: jml.MaxQuantifiedExpression): ir.MaxQuantifiedExpression = {
    node match {
      case jml.MaxQuantifiedExpression(declaration, guard, expression, srcLocation, precedingWhitespace) =>
        val irGuard = guard.map(visitExpression)
        val res = ir.MaxQuantifiedExpression(List(visitQuantifiedVarDeclaration(declaration)),
                                              irGuard,
                                              visitExpression(expression),
                                              srcLocation,
                                              precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitMinExpression(node: jml.MinExpression): ir.MinExpression = {
    node match {
      case jml.MinExpression(declaration, guard, expression, srcLocation, precedingWhitespace) =>
        val irGuard = guard.map(visitExpression)
        val res = ir.MinExpression(List(visitQuantifiedVarDeclaration(declaration)),
                                    irGuard,
                                    visitExpression(expression),
                                    srcLocation,
                                    precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitNumOfExpression(node: jml.NumOfExpression): ir.NumOfExpression = {
    node match {
      case jml.NumOfExpression(declaration, guard, expression, srcLocation, precedingWhitespace) =>
        val irGuard = guard.map(visitExpression)
        val res = ir.NumOfExpression(List(visitQuantifiedVarDeclaration(declaration)),
                                      irGuard,
                                      visitExpression(expression),
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitProductExpression(node: jml.ProductExpression): ir.ProductExpression = {
    node match {
      case jml.ProductExpression(declaration, guard, expression, srcLocation, precedingWhitespace) =>
        val irGuard = guard.map(visitExpression)
        val res = ir.ProductExpression(List(visitQuantifiedVarDeclaration(declaration)),
                                        irGuard,
                                        visitExpression(expression),
                                        srcLocation,
                                        precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitSumExpression(node: jml.SumExpression): ir.SumExpression = {
    node match {
      case jml.SumExpression(declaration, guard, expression, srcLocation, precedingWhitespace) =>
        val irGuard = guard.map(visitExpression)
        val res = ir.SumExpression(List(visitQuantifiedVarDeclaration(declaration)),
                                    irGuard,
                                    visitExpression(expression),
                                    srcLocation,
                                    precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }


  def visitQuantifiedVarDeclaration(node: jml.QuantifiedVarDeclaration): ir.QuantifiedVarDeclaration = {
    node match {
      case decl: jml.NonNullQuantifiedVarDeclaration =>
        visitNonNullQuantifiedVarDeclaration(decl)
      case decl: jml.NullableQuantifiedVarDeclaration =>
        visitNullableQuantifiedVarDeclaration(decl)
      case jml.QuantifiedVarDeclaration(declType, declarators, srcLocation, precedingWhitespace) =>
        val res = ir.QuantifiedVarDeclaration(visitTypeIndication(declType),
                                              declarators.map(visitStringNode),
                                              srcLocation,
                                              precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitNonNullQuantifiedVarDeclaration(node: jml.NonNullQuantifiedVarDeclaration): ir.NonNullQuantifiedVarDeclaration = {
    node match {
      case jml.NonNullQuantifiedVarDeclaration(declType, declarators, srcLocation, precedingWhitespace) =>
        val res = ir.NonNullQuantifiedVarDeclaration(visitTypeIndication(declType),
                                                      declarators.map(visitStringNode),
                                                      srcLocation,
                                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitNullableQuantifiedVarDeclaration(node: jml.NullableQuantifiedVarDeclaration): ir.NullableQuantifiedVarDeclaration = {
    node match {
      case jml.NullableQuantifiedVarDeclaration(declType, declarators, srcLocation, precedingWhitespace) =>
        val res = ir.NullableQuantifiedVarDeclaration(visitTypeIndication(declType),
                                                      declarators.map(visitStringNode),
                                                      srcLocation,
                                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitSetComprehension(node: jml.SetComprehension): ir.SetComprehension = {
    node match {
      case jml.SetComprehension(typeIndication, declaration, containsExpression, filterExpression, srcLocation, precedingWhitespace) =>
        val res = ir.SetComprehension(visitTypeIndication(typeIndication),
                                      visitQuantifiedVarDeclaration(declaration),
                                      visitExpression(containsExpression),
                                      visitExpression(filterExpression),
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitSpecDeclarationBlock(node: jml.SpecDeclarationBlock): ir.SpecDeclarationBlock = {
    node match {
      case jml.SpecDeclarationBlock(startSpec, declarations, endSpec, srcLocation, precedingWhitespace) =>
        val oldWrapInSpec = wrapInSpec
        wrapInSpec = false
        val res = ir.SpecDeclarationBlock(startSpec.map(visitStringNode),
                                          declarations.map(visitSpecDeclaration),
                                          endSpec.map(visitStringNode),
                                          srcLocation,
                                          precedingWhitespace)
        res.copyWhitespace(node)
        wrapInSpec = oldWrapInSpec
        res
    }
  }

  def visitSpecDeclaration(node: jml.SpecDeclaration): ir.SpecDeclaration = {
    try
      node match {
        case decl: jml.InvariantDeclaration => visitInvariantDeclaration(decl)
        case decl: jml.InitiallyDeclaration => visitInitiallyDeclaration(decl)
        case decl: jml.ReadableDeclaration => visitReadableDeclaration(decl)
        case decl: jml.WritableDeclaration => visitWritableDeclaration(decl)
        case decl: jml.MonitorsForDeclaration => visitMonitorsForDeclaration(decl)
        case decl: jml.AxiomDeclaration => visitAxiomDeclaration(decl)
        case decl: jml.SpecMethodDeclaration => visitSpecMethodDeclaration(decl)
        case decl: jml.GhostDeclaration => visitGhostDeclaration(decl)
        case decl: jml.ModelDeclaration => visitModelDeclaration(decl)
        case decl: jml.NonTranslatedSpecDeclaration => visitNonTranslatedSpecDeclaration(decl)
        case _ => ???
      }
    catch {
      case e: NotSupportedException =>
        val text = Helper.commentUnknownNode(node, s"${e.construct} not supported, whole SpecDeclaration commented out")
        ir.NonTranslatedSpecDeclaration(text, node.srcLocation)
    }
  }

  def visitNonTranslatedSpecDeclaration(decl: jml.NonTranslatedSpecDeclaration): ir.SpecDeclaration = {
    decl match {
      case jml.NonTranslatedSpecDeclaration(text, srcLocation, precedingWhitespace) =>
        ir.NonTranslatedSpecDeclaration(text, srcLocation, precedingWhitespace)
    }
  }
  def visitInvariantDeclaration(node: jml.InvariantDeclaration): ir.InvariantDeclaration = {
    node match {
      case jml.InvariantDeclaration(expression, srcLocation, precedingWhitespace) =>
        val res = ir.InvariantDeclaration(visitExpression(expression),
                                          srcLocation,
                                          precedingWhitespace)
        res.modifiers = node.modifiers.map(visitModifier)
        res.copyWhitespace(node)
        res
    }
  }
  def visitInitiallyDeclaration(node: jml.InitiallyDeclaration): ir.InitiallyDeclaration = {
    node match {
      case jml.InitiallyDeclaration(expression, srcLocation, precedingWhitespace) =>
        val res = ir.InitiallyDeclaration(visitExpression(expression),
                                          srcLocation,
                                          precedingWhitespace)
        res.modifiers = node.modifiers.map(visitModifier)
        res.copyWhitespace(node)
        res
    }
  }
  def visitReadableDeclaration(node: jml.ReadableDeclaration): ir.ReadableDeclaration = {
    node match {
      case jml.ReadableDeclaration(identifier, expression, srcLocation, precedingWhitespace) =>
        val res = ir.ReadableDeclaration(visitStringNode(identifier),
                                          visitExpression(expression),
                                          srcLocation,
                                          precedingWhitespace)
        res.modifiers = node.modifiers.map(visitModifier)
        res.copyWhitespace(node)
        res
    }
  }
  def visitWritableDeclaration(node: jml.WritableDeclaration): ir.WritableDeclaration = {
    node match {
      case jml.WritableDeclaration(identifier, expression, srcLocation, precedingWhitespace) =>
        val res = ir.WritableDeclaration(visitStringNode(identifier),
                                          visitExpression(expression),
                                          srcLocation,
                                          precedingWhitespace)
        res.modifiers = node.modifiers.map(visitModifier)
        res.copyWhitespace(node)
        res
    }
  }
  def visitMonitorsForDeclaration(node: jml.MonitorsForDeclaration): ir.MonitorsForDeclaration = {
    node match {
      case jml.MonitorsForDeclaration(identifier, expressions, srcLocation, precedingWhitespace) =>
        val res = ir.MonitorsForDeclaration(visitStringNode(identifier),
                                            expressions.map(visitExpression),
                                            srcLocation,
                                            precedingWhitespace)
        res.modifiers = node.modifiers.map(visitModifier)
        res.copyWhitespace(node)
        res
    }
  }
  def visitAxiomDeclaration(node: jml.AxiomDeclaration): ir.AxiomDeclaration = {
    node match {
      case jml.AxiomDeclaration(expression, srcLocation, precedingWhitespace) =>
        val res = ir.AxiomDeclaration(visitExpression(expression),
          srcLocation,
          precedingWhitespace)
        res.modifiers = node.modifiers.map(visitModifier)
        res.copyWhitespace(node)
        res
    }
  }

  def visitSpecMethodDeclaration(node: jml.SpecMethodDeclaration): ir.SpecMethodDeclaration = {
    node match {
      case jml.SpecMethodDeclaration(name, returnType, parameters,
                                     body, srcLocation, precedingWhitespace) =>
        val irReturn = returnType.map(visitTypeIndication)
        val res = ir.SpecMethodDeclaration(visitStringNode(name),
                                          irReturn,
                                          parameters.map(visitParameter),
                                          None,
                                          visitBlockOrSemicolon(body),
                                          srcLocation,
                                          precedingWhitespace)
        res.contractBefore = node.contractBefore.map(visitContract)
        res.modifiers = node.modifiers.map(visitModifier)
        res.copyWhitespace(node)
        res
    }
  }

}
