package languages.VerCors

import languages.VerCors.syntaxTree.*

import scala.collection.mutable.ListBuffer

class DesugarVisitor {
  def visitCompilationUnit(node: CompilationUnit): Unit = {
    node.declarations.foreach(visitTypeDeclaration)
  }

  def visitTypeDeclaration(node: TypeDeclaration): Unit = {
    node match {
      case classD: ClassDeclaration => visitClassDeclaration(classD)
      case inter: InterfaceDeclaration => visitInterfaceDeclaration(inter)
      case _ =>
    }
  }

  def visitClassDeclaration(declaration: ClassDeclaration):Unit = {
    val meths = declaration.body.declarations
                  .filter(x => x.isInstanceOf[MethodDeclaration])
                  .map(x => x.asInstanceOf[MethodDeclaration])
    val decls: ListBuffer[GhostDeclaration] = ListBuffer()
    for (method <- meths) {
      // cannot have ghost I/O on constructor, only on "real" methods
      if(method.returnType.isDefined) {
        val ghosts = extractGhostParameters(method)
        for (g <- ghosts) {
          val name = StringNode(getGhostVarName(method, g.name), g.srcLocation, g.name.precedingWhitespace)
          val decl = FieldDeclaration(g.argType,
            List(VariableDeclarator(name, None, g.srcLocation)),
            g.srcLocation,
            g.precedingWhitespace)
          val asGhost = GhostDeclaration(GhostModifier(g.srcLocation), decl, g.srcLocation, System.lineSeparator() + "        ")
          decls.append(asGhost)
        }
      }
    }
    if (decls.nonEmpty) {
      val start = Some(StringNode("/*@", declaration.srcLocation, System.lineSeparator() + "    "))
      val end = Some(StringNode("*/", declaration.srcLocation, System.lineSeparator() + "     "))
      val block = SpecDeclarationBlock(start, decls.toList, end, declaration.srcLocation)
      declaration.body.declarations = declaration.body.declarations.prepended(block)
    }
    declaration.body.declarations.foreach(visitMemberDeclaration)
  }
  def visitInterfaceDeclaration(declaration: InterfaceDeclaration): Unit = {
    declaration.body.declarations.foreach(visitMemberDeclaration)
  }

  def visitMemberDeclaration(node: MemberDeclaration): Unit = {
    node match {
      case member: ClassDeclaration => visitClassDeclaration(member)
      case member: InterfaceDeclaration => visitInterfaceDeclaration(member)
      case member: MethodDeclaration => visitMethodDeclaration(member)
      case member: FieldDeclaration => visitFieldDeclaration(member)
      case member: GhostDeclaration => visitMemberDeclaration(member.decl)
      case member: SpecDeclarationBlock => member.declarations.foreach(visitMemberDeclaration)
      case member: PredicateDeclaration =>
      case _ => // todo: any other? e.g. nested classes, generic methods
    }
  }

  def visitMethodDeclaration(declaration: MethodDeclaration): Unit = {
//    declaration.contractBefore.foreach(visitContract)
//    declaration.contractWithin.foreach(visitContract)
    declaration.body match {
//      case block: Block => visitBlock(block)
      case _ =>
    }

    val invars = desugarContext(declaration)
    if(invars.nonEmpty) {
      declaration.body match {
        case body: Block => addLoopInvariants(invars, body)
        case _ =>
      }
    }

    val ghosts: ListBuffer[GhostIO] = extractGhostParameters(declaration)
    if(declaration.returnType.isDefined) {
      makeGhostLocalDeclarations(declaration, ghosts)
      // todo: global declaration and renaming of variable in pre/postconditions
    } else if(ghosts.nonEmpty) {
      System.err.println(s"Translation of ghost parameter on Constructor not supported${ghosts.head.srcLocation.toString}")
    }
  }

  private def makeGhostLocalDeclarations(declaration: MethodDeclaration, ghosts: ListBuffer[GhostIO]): Unit = {
    val statements: ListBuffer[Statement] = ListBuffer()
    for (ghost <- ghosts) {
      declaration.body match {
        case body: Block =>
          var declarator: VariableDeclarator = null
          ghost match {
            case _: GhostArgument =>
              val init = StringNode(getGhostVarName(declaration, ghost.name), ghost.srcLocation)
              declarator = VariableDeclarator(ghost.name, Some(init), ghost.srcLocation)
            case _: GhostReturn =>
              declarator = VariableDeclarator(ghost.name, None, ghost.srcLocation)
            case _ =>
          }
          val assignment = LocalDeclaration(Nil, ghost.argType, List(declarator), true, ghost.srcLocation, " ")
          val start = Some(StringNode("//@", ghost.srcLocation, System.lineSeparator() + "        "))
          val ghostAssign = JmlStatementBlock(start, List(GhostStatement(assignment, ghost.srcLocation, " ")), None, ghost.srcLocation)
          statements.append(ghostAssign)
        case _ =>
      }
    }
    if (statements.nonEmpty) {
      val oldBlock = declaration.body.asInstanceOf[Block]
      val block = Block(oldBlock.static, oldBlock.statements.prependedAll(statements), oldBlock.srcLocation)
      block.copyWhitespace(oldBlock)
      declaration.body = block
    }
  }

  private def extractGhostParameters(declaration: MethodDeclaration): ListBuffer[GhostIO] = {
    val ghosts: ListBuffer[GhostIO] = ListBuffer()
    declaration.contractBefore match {
      case Some(contract) =>
        ghosts.appendAll(contract.clauses.filter(x => x.isInstanceOf[GhostIO]).map(x => x.asInstanceOf[GhostIO]))
      case _ =>
    }
    declaration.contractWithin match {
      case Some(contract) =>
        ghosts.appendAll(contract.clauses.filter(x => x.isInstanceOf[GhostIO]).map(x => x.asInstanceOf[GhostIO]))
      case _ =>
    }
    ghosts
  }

  private def getGhostVarName(methodDeclaration: MethodDeclaration, stringNode: StringNode): String = {
    methodDeclaration.name.text +"_"+ stringNode.text
  }

  def visitFieldDeclaration(declaration: FieldDeclaration): Unit = {
    // declaration.declarators.foreach(x => x.initializer.foreach(visitInitializer))
  }

  def visitInitializer(node: Initializer): Unit = {
    node match {
//      case init: Expression => visitExpression(init)
      case _ => ???
    }
  }




  def desugarContext(node: MethodDeclaration): List[LoopInvariant] = {
    var invars: List[LoopInvariant] = Nil
    node.contractBefore match {
      case Some(contract) =>
        invars = extractInvariants(contract.clauses)
        // reorder: pull context_everywhere to the front
        val ce = contract.clauses.filter(x => x.isInstanceOf[ContextEverywhere])
        val clauses = ce.appendedAll(contract.clauses.filterNot(x => x.isInstanceOf[ContextEverywhere]))
        // now we can desugar
        contract.clauses = clauses.flatMap(desugarContext)
      case _ =>
    }
    node.contractWithin match {
      case Some(contract) =>
        invars = invars.appendedAll(extractInvariants(contract.clauses))
        // reorder: pull context_everywhere to the front
        val ce = contract.clauses.filter(x => x.isInstanceOf[ContextEverywhere])
        val clauses = ce.appendedAll(contract.clauses.filterNot(x => x.isInstanceOf[ContextEverywhere]))
        // now we can desugar
        contract.clauses = clauses.flatMap(desugarContext)
      case _ =>
    }
    invars
  }

  def extractInvariants(clauses: List[ContractClause]): List[LoopInvariant] = {
    val ce = clauses.filter(x => x.isInstanceOf[ContextEverywhere]).map(x => x.asInstanceOf[ContextEverywhere])
    ce.map(x => LoopInvariant(x.expression, x.srcLocation, System.lineSeparator() + "      "))
  }

  def desugarContext(node: ContractClause): List[ContractClause] = {
    node match {
      case context: Context =>
        val pre = Precondition(context.expression, context.srcLocation, context.precedingWhitespace)
        pre.requiresKeyword.precedingWhitespace = context.contextKeyword.precedingWhitespace
        pre.semicolon.precedingWhitespace = context.semicolon.precedingWhitespace
        pre.startSpec = context.startSpec
        val post = Postcondition(context.expression, context.srcLocation, context.precedingWhitespace)
        post.ensuresKeyword.precedingWhitespace = context.contextKeyword.precedingWhitespace
        post.semicolon.precedingWhitespace = context.semicolon.precedingWhitespace
        post.trailingWhitespace = context.trailingWhitespace
        post.endSpec = context.endSpec
        List(pre, post)
      case context: ContextEverywhere =>
        val pre = Precondition(context.expression, context.srcLocation, context.precedingWhitespace)
        pre.requiresKeyword.precedingWhitespace = context.contextEverywhereKeyword.precedingWhitespace
        pre.semicolon.precedingWhitespace = context.semicolon.precedingWhitespace
        pre.startSpec = context.startSpec
        val post = Postcondition(context.expression, context.srcLocation, context.precedingWhitespace)
        post.ensuresKeyword.precedingWhitespace = context.contextEverywhereKeyword.precedingWhitespace
        post.semicolon.precedingWhitespace = context.semicolon.precedingWhitespace
        post.trailingWhitespace = context.trailingWhitespace
        post.endSpec = context.endSpec
        List(pre, post)
      case _ => List(node)
    }
  }

  def addLoopInvariants(invars: List[LoopInvariant], stmt: Statement): Unit = {
    stmt match {
      case loop: Loop => addLoopInvariants(invars, loop)
      case cond: IfStatement =>
        addLoopInvariants(invars, cond.thenStatement)
        if(cond.elseStatement.isDefined){
          addLoopInvariants(invars, cond.elseStatement.get)
        }
      case block: Block =>
        block.statements.foreach(x => addLoopInvariants(invars, x))
      case ghost: GhostStatement => addLoopInvariants(invars, ghost.statement)
      case jmlBlock: JmlStatementBlock =>
        jmlBlock.statements.foreach(x => addLoopInvariants(invars, x))
      case switch: SwitchStatement =>
        switch.cases.foreach(x => x.statements.foreach(y => addLoopInvariants(invars, y)))
      case sync: SynchronizedStatement =>
        addLoopInvariants(invars, sync.body)
      case labeled: LabeledStatement =>
        addLoopInvariants(invars, labeled.statement)
      case tryBlock: TryBlock =>
        addLoopInvariants(invars, tryBlock.tryBlock)
        tryBlock.catchClauses.foreach(x => addLoopInvariants(invars, x.block))
        if(tryBlock.finallyBlock.isDefined) {
          addLoopInvariants(invars, tryBlock.finallyBlock.get)
        }
      case _ =>
    }
  }

  def addLoopInvariants(invars: List[LoopInvariant], loop: Loop): Unit = {
    val spec = LoopSpecification(Some(StringNode("/*@", invars.head.srcLocation)),
                                invars,
                                Some(StringNode("*/", invars.head.srcLocation)),
                                invars.head.srcLocation)
    loop match {
      case l: LoopWithContractWithin => 
        if(l.contractWithin.nonEmpty) {
          l.contractWithin = l.contractWithin.prepended(spec)
        }
        else {
          loop.contractBefore = loop.contractBefore.prepended(spec)
        }
      case _ =>
        loop.contractBefore = loop.contractBefore.prepended(spec)
    }
    addLoopInvariants(invars, loop.body)
  }

  def transformLoopContract(node: LoopSpecification): Unit = {
    for(clause <- node.clauses) {
      clause match {
        case pre: Precondition =>
          // todo: turn "requires expr(i);" into "loop_invariant (\forall* int j; i<= j && j<loop_length; expr(j));"
        case post: Postcondition =>
          // todo: turn "ensures expr(i);" into "loop_invariant (\forall* int j; 0<= j && j<i; expr(j));"
        case _ =>
      }
    }
  }

  def desugarIndexAllExpression(node: PermissionPredicate): Expression = {
    node.target match {
      case idx: ArrayIndexExpression =>
        if(idx.indexExpression.isInstanceOf[IndexAllExpression]) {
          val arrayRef = idx.arrayExpression
          val intType = StringTypeIndication("int", Nil, "int", node.srcLocation, " ")
          val variable = StringNode("i", node.srcLocation, " ")
          val decl = QuantifiedVarDeclaration(intType, List(variable.copy()), node.srcLocation)
          val lowerBound = BinaryExpression(StringNode("0",node.srcLocation), StringNode("<=",node.srcLocation),variable.copy(), node.srcLocation)
          val length = BinaryExpression(arrayRef, StringNode(".", node.srcLocation),StringNode("length", node.srcLocation), node.srcLocation)
          val upperBound = BinaryExpression(variable.copy(), StringNode("<",node.srcLocation),length, node.srcLocation)
          val guard = Some(BinaryExpression(lowerBound, StringNode("&&", node.srcLocation),upperBound,node.srcLocation))
          val target = ArrayIndexExpression(idx.arrayExpression, variable.copy(), idx.srcLocation, idx.precedingWhitespace)
          target.trailingWhitespace = idx.trailingWhitespace
          val perm = PermissionPredicate(target,node.amount, node.srcLocation, node.precedingWhitespace)
          perm.trailingWhitespace = node.trailingWhitespace
          ForallStarExpression(List(decl), guard, perm, node.srcLocation, " ")
        } else {
          node
        }
      case _ => node
    }
  }
}
