package languages.VerCors

import intermediateRepresentation.syntaxTree as ir
import languages.VerCors.syntaxTree as vercors
import generic.{Helper, NotSupportedException, SourceLocation}

import scala.collection.mutable.ListBuffer

class FromIRVisitor {

  // indicates whether new modifiers, etc need to be wrapped in specifications
  // or if they already are (e.g. inside ghost declarations)
  var wrapInSpec = true

  def visitStringNode(node: ir.StringNode): vercors.StringNode = {
    node match {
      case imp: ir.ImpliesOperator => visitImpliesOperator(imp)
      case sep: ir.SeparatingConjunctOperator => visitSeparatingConjunctOperator(sep)
      case ir.StringNode(text, src, pred, trail) =>
        vercors.StringNode(text, src, pred, trail)
    }
  }
  def visitImpliesOperator(node: ir.ImpliesOperator): vercors.ImpliesOperator = {
    node match {
      case ir.ImpliesOperator(srcLocation, pred) =>
        val res = new vercors.ImpliesOperator(srcLocation)
        res.copyWhitespace(node)
        res
    }
  }
  def visitSeparatingConjunctOperator(node: ir.SeparatingConjunctOperator): vercors.SeparatingConjunctOperator = {
    node match {
      case ir.SeparatingConjunctOperator(text, srcLocation, pred) =>
        val res = new vercors.SeparatingConjunctOperator(srcLocation)
        res.copyWhitespace(node)
        res
    }
  }
  def visitModifierStringNode(node: ir.ModifierStringNode): vercors.ModifierStringNode = {
    node match {
      case ir.ModifierStringNode(text, src, pred) =>
        val res = vercors.ModifierStringNode(text, src, pred)
        res.copyWhitespace(node)
        res
    }
  }

  private def reorderModifiers(modifiers: List[vercors.Modifier]): List[vercors.Modifier] = {
    def findGhost(m: generic.syntaxTree.Modifier): Boolean
        = (m.isInstanceOf[vercors.ModifierStringNode]
            && m.asInstanceOf[vercors.ModifierStringNode].text == "ghost")
          || (m.isInstanceOf[vercors.GhostModifier])

    def reorderRecursive(modifier: vercors.Modifier): vercors.Modifier =
      modifier match {
        case asBlock: vercors.JmlModifierBlock =>
          vercors.JmlModifierBlock(asBlock.startSpec, reorderModifiers(asBlock.modifiers), asBlock.endSpec,
            asBlock.srcLocation, asBlock.precedingWhitespace)
        case _ =>
          modifier
      }

    if (Helper.findModifier(modifiers, findGhost)) {
      modifiers.map(reorderRecursive).filter(findGhost).appendedAll(modifiers.filterNot(findGhost))
    } else {
      modifiers
    }
  }




  def visitCompilationUnit(node: ir.CompilationUnit): vercors.CompilationUnit = {
    try
      node match {
        case nt: ir.NonTranslatedCompilationUnit =>
          visitNonTranslatedCompilationUnit(nt)
        case ir.CompilationUnit(pkg, imports, declarations, srcLocation, precedingWhitespace) =>
          var irPkg: Option[vercors.PackageDeclaration] = None
          if (pkg.isDefined) {
            irPkg = Some(visitPackageDeclaration(pkg.get))
          }
          val res = vercors.CompilationUnit(irPkg,
                                        imports.map(visitImportDeclaration),
                                        declarations.map(visitTypeDeclaration),
                                        srcLocation,
                                        precedingWhitespace)
          res.copyWhitespace(node)
          res
      }
    catch {
      case e: NotSupportedException =>
        val text = Helper.commentUnknownNode(node, s"${e.construct} not supported, whole CompilationUnit commented out")
        vercors.NonTranslatedCompilationUnit(text, node.srcLocation)
    }
  }

  def visitNonTranslatedCompilationUnit(node: ir.NonTranslatedCompilationUnit): vercors.NonTranslatedCompilationUnit = {
    node match {
      case ir.NonTranslatedCompilationUnit(text, srcLocation, precedingWhitespace) =>
        val res = vercors.NonTranslatedCompilationUnit(text, srcLocation, precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitPackageDeclaration(node: ir.PackageDeclaration): vercors.PackageDeclaration = {
    node match {
      case ir.PackageDeclaration(name, annotations, srcLocation, precedingWhitespace) =>
        val res = vercors.PackageDeclaration(visitStringNode(name),
                                        annotations.map(visitStringAnnotation),
                                        srcLocation,
                                        precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitImportDeclaration(node: ir.ImportDeclaration): vercors.ImportDeclaration = {
    node match {
      case ir.ImportDeclaration(name, srcLocation, precedingWhitespace) =>
        val res = vercors.ImportDeclaration(visitStringNode(name),
                              srcLocation,
                              precedingWhitespace)
        res.modifiers = node.modifiers.map(visitModifier)
        res.copyWhitespace(node)
        res
    }
  }

  def visitTypeDeclaration(node: ir.TypeDeclaration): vercors.TypeDeclaration = {
    try
      node match {
        case decl: ir.AnnotationTypeDeclaration => visitAnnotationTypeDeclaration(decl)
        case decl: ir.ClassDeclaration => visitClassDeclaration(decl)
        case decl: ir.EnumDeclaration => visitEnumDeclaration(decl)
        case decl: ir.InterfaceDeclaration => visitInterfaceDeclaration(decl)
        case decl: ir.EmptyStatement => visitEmptyStatement(decl)
        case decl: ir.NonTranslatedTypeDeclaration => visitNonTranslatedTypeDeclaration(decl)
        case _ => ???
      }
    catch {
      case e: NotSupportedException =>
        val text = Helper.commentUnknownNode(node, s"${e.construct} not supported, whole TypeDeclaration commented out")
        vercors.NonTranslatedTypeDeclaration(text, node.srcLocation)
    }
  }

  def visitNonTranslatedTypeDeclaration(node: ir.NonTranslatedTypeDeclaration): vercors.NonTranslatedTypeDeclaration = {
    node match {
      case ir.NonTranslatedTypeDeclaration(text, srcLocation, precedingWhitespace) =>
        val res = vercors.NonTranslatedTypeDeclaration(text, srcLocation, precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitGhostDeclaration(node: ir.GhostDeclaration): vercors.GhostDeclaration = {
    node match {
      case ir.GhostDeclaration(ghost, decl, srcLocation, precedingWhitespace) =>
        val res = vercors.GhostDeclaration(visitGhostModifier(ghost),
          visitMemberDeclaration(decl),
          srcLocation,
          precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitModelDeclaration(node: ir.ModelDeclaration): vercors.GhostDeclaration = {
    node match {
      case ir.ModelDeclaration(model, decl, srcLocation, precedingWhitespace) =>
        val res = vercors.GhostDeclaration(visitModelModifier(model),
          visitMemberDeclaration(decl),
          srcLocation,
          precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitModifier(node: ir.Modifier): vercors.Modifier = {
    node match {
      case mod: ir.StaticModifier => visitStaticModifier(mod)
      case mod: ir.Annotation => visitAnnotation(mod)
      case mod: ir.ModifierStringNode => visitModifierStringNode(mod)
      case mod: ir.PureModifier => visitPureModifier(mod)
      case mod: ir.NullableModifier => visitNullableModifier(mod)
      case mod: ir.NonnullModifier => visitNonnullModifier(mod)
      case mod: ir.GhostModifier => visitGhostModifier(mod)
      case mod: ir.HelperModifier => visitHelperModifier(mod)
      case mod: ir.MonitoredModifier => visitMonitoredModifier(mod)
      case mod: ir.JmlModifierBlock => visitJmlModifierBlock(mod)
      case mod: ir.ModelModifier => visitModelModifier(mod)
      case mod: ir.InstanceModifier => visitInstanceModifier(mod)
      case mod: ir.SpecMethodModifier => visitSpecMethodModifier(mod)
      case _ => ???
    }
    // todo: more?
  }
  def visitStaticModifier(node: ir.StaticModifier): vercors.StaticModifier = {
    node match {
      case ir.StaticModifier(srcLocation, precedingWhitespace) =>
        val res = vercors.StaticModifier(srcLocation, precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitVariableModifier(node: ir.Modifier): vercors.Modifier = {
    node match {
      case mod: ir.Annotation => visitAnnotation(mod)
      case mod: ir.ModifierStringNode => visitModifierStringNode(mod)
      case mod: ir.JmlModifierBlock => visitJmlModifierBlockForVariable(mod)
      case _ => 
        System.err.println(s"[Warning] The modifier \"$node\" is not supported by VerCors for usage at ${node.srcLocation}")
        new vercors.ModifierStringNode("", node.srcLocation)
    }
    // todo: more?
  }
  def visitModifierOfParam(node: ir.Modifier): vercors.Modifier = {
    node match {
      case mod: ir.Annotation => visitAnnotation(mod)
      case mod: ir.ModifierStringNode => visitModifierStringNode(mod)
      case mod: ir.JmlModifierBlock => visitJmlModifierBlockForParameter(mod)
      case _ =>
        System.err.println(s"[Warning] The modifier \"$node\" is not supported by VerCors for usage at ${node.srcLocation}")
        new vercors.ModifierStringNode("", node.srcLocation)
    }
    // todo: more?
  }

  def visitGenericsSuperclassClause(node: ir.GenericsSuperclassClause): vercors.GenericsSuperclassClause = {
    node match {
      case ir.GenericsSuperclassClause(superclasses, srcLocation, precedingWhitespace) =>
        val res = vercors.GenericsSuperclassClause(superclasses.map(visitTypeIndication),
                                              srcLocation,
                                              precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitSuperclassClause(node: ir.SuperclassClause): vercors.SuperclassClause = {
    node match {
      case ir.SuperclassClause(superclass, srcLocation, precedingWhitespace) =>
        val res = vercors.SuperclassClause(visitTypeIndication(superclass),
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitMultiSuperclassClause(node: ir.MultiSuperclassClause): vercors.MultiSuperclassClause = {
    node match {
      case ir.MultiSuperclassClause(superclasses, srcLocation, precedingWhitespace) =>
        val res = vercors.MultiSuperclassClause(superclasses.map(visitTypeIndication),
                                            srcLocation,
                                            precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitInterfaceClause(node: ir.InterfaceClause): vercors.InterfaceClause = {
    node match {
      case ir.InterfaceClause(interfaces, srcLocation, precedingWhitespace) =>
        val res = vercors.InterfaceClause(interfaces.map(visitTypeIndication),
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitClassDeclaration(node: ir.ClassDeclaration): vercors.ClassDeclaration = {
    def checkStatic(modifiers: List[ir.Modifier]): Boolean
          = Helper.findModifier(modifiers, x=>x.isInstanceOf[ir.StaticModifier])

    val invars = getInvariants(node).reverse
    val stat = invars.filter(x => checkStatic(x.modifiers))
    val inst = invars.filter(x => !checkStatic(x.modifiers))
    node.body.declarations.foreach(decl => addInstanceInvariantClauses(decl, inst))
    node.body.declarations.foreach(decl => addStaticInvariantClauses(decl, stat))
    val init_clauses = getInitiallys(node).map(transformInitially)
    node.body.declarations.foreach(decl => addInitiallyClauses(decl, init_clauses))

    node match {
      case ir.ClassDeclaration(name, typeParameters, superclass, interfaces, body, srcLocation, precedingWhitespace) =>
        val irTypeParameters = typeParameters.map(visitTypeParameterClause)
        val irSuperclass = superclass.map(visitSuperclassClause)
        val irInterfaces = interfaces.map(visitInterfaceClause)
        val res = vercors.ClassDeclaration(visitStringNode(name),
                                      irTypeParameters,
                                      irSuperclass,
                                      irInterfaces,
                                      visitBody(body),
                                      srcLocation,
                                      precedingWhitespace)
        res.modifiers = reorderModifiers(node.modifiers.map(visitModifier))
        res.copyWhitespace(node)
        res
    }
  }

  def getInvariants(node: ir.ClassDeclaration): List[ir.InvariantDeclaration] = {
    def getInvar(decl: ir.MemberDeclaration): List[ir.InvariantDeclaration] = {
      decl match {
        case i: ir.InvariantDeclaration => List(i)
        case spec: ir.SpecDeclarationBlock => spec.declarations.flatMap(getInvar)
        case _ => List()
      }
    }

    node.body.declarations.flatMap(getInvar)
  }

  def getInvariants(node: ir.InterfaceDeclaration): List[ir.InvariantDeclaration] = {
    def getInvar(decl: ir.MemberDeclaration): List[ir.InvariantDeclaration] = {
      decl match {
        case i: ir.InvariantDeclaration => List(i)
        case spec: ir.SpecDeclarationBlock => spec.declarations.flatMap(getInvar)
        case _ => List()
      }
    }

    node.body.declarations.flatMap(getInvar)
  }

  def makePreconditionFromInvariant(invar: ir.InvariantDeclaration): ir.ContractClause = {
    val pre = ir.Precondition(invar.expression, invar.srcLocation, invar.precedingWhitespace)
    pre.requiresKeyword.precedingWhitespace = invar.invariantKeyword.precedingWhitespace
    pre
  }
  def makePostconditionFromInvariant(invar: ir.InvariantDeclaration): ir.ContractClause = {
    val post = ir.Postcondition(invar.expression, invar.srcLocation, invar.precedingWhitespace)
    post.ensuresKeyword.precedingWhitespace = invar.invariantKeyword.precedingWhitespace
    post
  }

  def addStaticInvariantClauses(decl: ir.MemberDeclaration, invars: List[ir.InvariantDeclaration]): Unit = {
    decl match {
      case spec: ir.SpecDeclarationBlock => spec.declarations.foreach(x => addStaticInvariantClauses(x, invars))
      case meth: ir.MethodDeclaration =>
        if(!Helper.findModifier(meth.modifiers, x=>x.isInstanceOf[ir.HelperModifier])) {
          if(meth.contractBefore.isEmpty && meth.contractWithin.isEmpty) {
            meth.contractBefore = Some(ir.Contract(Nil, meth.srcLocation))
          }
          val contract = if(meth.contractBefore.isDefined) meth.contractBefore.get else meth.contractWithin.get
          invars.map(makePostconditionFromInvariant).foreach(x => contract.addClauseBeforeContract[ir.IrTreeNode](x, wrapInSpec, ir.StringNode.apply))
          invars.map(makePreconditionFromInvariant).foreach(x => contract.addClauseBeforeContract[ir.IrTreeNode](x, wrapInSpec, ir.StringNode.apply))
        }
      case _ =>
    }
  }
  def addInstanceInvariantClauses(decl: ir.MemberDeclaration, invars: List[ir.InvariantDeclaration]): Unit = {
    decl match {
      case spec: ir.SpecDeclarationBlock => spec.declarations.foreach(x => addInstanceInvariantClauses(x, invars))
      case meth: ir.MethodDeclaration =>
        if(!Helper.findModifier(meth.modifiers, x=>x.isInstanceOf[ir.HelperModifier]||x.isInstanceOf[ir.StaticModifier])) {
          if(meth.contractBefore.isEmpty && meth.contractWithin.isEmpty) {
            meth.contractBefore = Some(ir.Contract(Nil, meth.srcLocation))
          }
          val contract = if(meth.contractBefore.isDefined) meth.contractBefore.get else meth.contractWithin.get
          if(meth.returnType.isEmpty) {
            invars.map(makePostconditionFromInvariant).foreach(x => contract.addClauseBeforeContract[ir.IrTreeNode](x, wrapInSpec, ir.StringNode.apply))
          } else if(meth.name.text == "finalize") {
            invars.map(makePreconditionFromInvariant).foreach(x => contract.addClauseBeforeContract[ir.IrTreeNode](x, wrapInSpec, ir.StringNode.apply))
          } else {
            invars.map(makePostconditionFromInvariant).foreach(x => contract.addClauseBeforeContract[ir.IrTreeNode](x, wrapInSpec, ir.StringNode.apply))
            invars.map(makePreconditionFromInvariant).foreach(x => contract.addClauseBeforeContract[ir.IrTreeNode](x, wrapInSpec, ir.StringNode.apply))
          }
        }
      case _ =>
    }
  }

  def getInitiallys(node: ir.ClassDeclaration): List[ir.InitiallyDeclaration] = {
    def getInits(decl: ir.MemberDeclaration): List[ir.InitiallyDeclaration] = {
      decl match {
        case i: ir.InitiallyDeclaration => List(i)
        case spec: ir.SpecDeclarationBlock => spec.declarations.flatMap(getInits)
        case _ => List()
      }
    }

    node.body.declarations.flatMap(getInits)
  }

  def transformInitially(init: ir.InitiallyDeclaration): ir.ContractClause = {
    val post = ir.Postcondition(init.expression, init.srcLocation, init.precedingWhitespace)
    post.ensuresKeyword.precedingWhitespace = init.initiallyKeyword.precedingWhitespace
    post
  }

  def addInitiallyClauses(decl: ir.MemberDeclaration, clauses: List[ir.ContractClause]): Unit = {
    def findHelper(mod: ir.Modifier): Boolean = {
      mod match {
        case _: ir.HelperModifier => true
        case spec: ir.JmlModifierBlock => spec.modifiers.exists(findHelper)
        case _ => false
      }
    }

    decl match {
      case spec: ir.SpecDeclarationBlock => spec.declarations.foreach(x => addInitiallyClauses(x, clauses))
      case meth: ir.MethodDeclaration =>
        if(!meth.modifiers.exists(findHelper)) {
          if(meth.contractBefore.isEmpty && meth.contractWithin.isEmpty) {
            meth.contractBefore = Some(ir.Contract(Nil, meth.srcLocation))
          }
          val contract = if(meth.contractBefore.isDefined) meth.contractBefore.get else meth.contractWithin.get
          if(meth.returnType.isEmpty) {
            clauses.foreach(x => contract.addClauseBeforeContract[ir.IrTreeNode](x, wrapInSpec, ir.StringNode.apply))
          }
        }
      case _ =>
    }
  }

  def visitTypeParameterClause(node: ir.TypeParameterClause): vercors.TypeParameterClause = {
    node match {
      case ir.TypeParameterClause(parameters, srcLocation, precedingWhitespace) =>
        val res = vercors.TypeParameterClause(parameters.map(visitTypeParameter),
                                          srcLocation,
                                          precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitTypeParameter(node: ir.TypeParameter): vercors.TypeParameter = {
    node match {
      case ir.TypeParameter(annotations, name, superclass, srcLocation, precedingWhitespace) =>
        val irSuperclass = superclass.map(visitGenericsSuperclassClause)
        val res = vercors.TypeParameter(annotations.map(visitStringAnnotation),
                                    visitTypeIndication(name),
                                    irSuperclass,
                                    srcLocation,
                                    precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitEnumDeclaration(node: ir.EnumDeclaration): vercors.EnumDeclaration = {
    node match {
      case ir.EnumDeclaration(name, interfaces, constants, declarations, comma, semicolon, srcLocation, precedingWhitespace) =>
        val irInterfaces = interfaces.map(visitInterfaceClause)
        val irComma = comma.map(visitStringNode)
        val irSemicolon = semicolon.map(visitStringNode)
        val res = vercors.EnumDeclaration(visitStringNode(name),
                                      irInterfaces,
                                      constants.map(visitEnumConstant),
                                      declarations.map(visitMemberDeclaration),
                                      irComma,
                                      irSemicolon,
                                      srcLocation,
                                      precedingWhitespace)
        res.modifiers = reorderModifiers(node.modifiers.map(visitModifier))
        res.copyWhitespace(node)
        res
    }
  }
  def visitEnumConstant(node: ir.EnumConstant): vercors.EnumConstant = {
    node match {
      case ir.EnumConstant(name, annotations, arguments, body, srcLocation, precedingWhitespace) =>
        val irArgs = arguments.map(visitArguments)
        val irBody = body.map(visitBody)
        val res = vercors.EnumConstant(visitStringNode(name),
                                  annotations.map(visitStringAnnotation),
                                  irArgs,
                                  irBody,
                                  srcLocation,
                                  precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitInterfaceDeclaration(node: ir.InterfaceDeclaration): vercors.InterfaceDeclaration = {
    node match {
      case ir.InterfaceDeclaration(name, typeParameter, superInterfaces, body, srcLocation, precedingWhitespace) =>
        def checkInstance(modifiers: List[ir.Modifier]): Boolean
        = Helper.findModifier(modifiers, x=>x.isInstanceOf[ir.InstanceModifier])

        val invars = getInvariants(node).reverse
        val stat = invars.filter(x => !checkInstance(x.modifiers))
        val inst = invars.filter(x => checkInstance(x.modifiers))
        node.body.declarations.foreach(decl => addInstanceInvariantClauses(decl, inst))
        node.body.declarations.foreach(decl => addStaticInvariantClauses(decl, stat))

        val irTypeParameters = typeParameter.map(visitTypeParameterClause)
        val irSuperclass = superInterfaces.map(visitMultiSuperclassClause)
        val res = vercors.InterfaceDeclaration(visitStringNode(name),
                                          irTypeParameters,
                                          irSuperclass,
                                          visitBody(body),
                                          srcLocation,
                                          precedingWhitespace)
        res.modifiers = node.modifiers.map(visitModifier)
        res.copyWhitespace(node)
        res
    }
  }

  def visitBody(node: ir.Body): vercors.Body = {
    node match {
      case ir.Body(declarations, srcLocation, precedingWhitespace) =>
        val res = vercors.Body(declarations.map(visitMemberDeclaration),
                          srcLocation,
                          precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitMemberDeclaration(node: ir.MemberDeclaration): vercors.MemberDeclaration = {
    try
      node match {
        case member: ir.ClassDeclaration => visitClassDeclaration(member)
        case member: ir.EnumDeclaration => visitEnumDeclaration(member)
        case member: ir.InterfaceDeclaration => visitInterfaceDeclaration(member)
        case member: ir.GenericMethodDeclaration => visitGenericMethodDeclaration(member)
        case member: ir.MethodDeclaration => visitMethodDeclaration(member)
        case member: ir.BlockOrSemicolon => visitBlockOrSemicolon(member)
        case member: ir.FieldDeclaration => visitFieldDeclaration(member)
        case member: ir.AnnotationTypeDeclaration => visitAnnotationTypeDeclaration(member)
        case member: ir.SpecDeclaration => visitSpecDeclaration(member)
        case member: ir.SpecDeclarationBlock => visitSpecDeclarationBlock(member)
        case member: ir.NonTranslatedMemberDeclaration => visitNonTranslatedMemberDeclaration(member)
        case _ => ???
      }
    catch {
      case e: NotSupportedException =>
        val text = Helper.commentUnknownNode(node, s"${e.construct} not supported, whole MemberDeclaration commented out")
        vercors.NonTranslatedMemberDeclaration(text, node.srcLocation)
    }
  }
  def visitNonTranslatedMemberDeclaration(node: ir.NonTranslatedMemberDeclaration): vercors.NonTranslatedMemberDeclaration = {
    node match {
      case ir.NonTranslatedMemberDeclaration(text, srcLocation, precedingWhitespace) =>
        val res = vercors.NonTranslatedMemberDeclaration(text, srcLocation, precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def checkForNonNullModifier(modifiers: List[ir.Modifier]): Boolean = {
    def checkModifier(m: ir.Modifier): Boolean = {
      m match {
        case _: ir.NonnullModifier => true
        case b: ir.JmlModifierBlock => checkForNonNullModifier(b.modifiers)
        case _ => false
      }
    }

    modifiers.exists(checkModifier)
  }

  def visitMethodDeclaration(node: ir.MethodDeclaration): vercors.MethodDeclaration = {
    node match {
      case ir.MethodDeclaration(name, returnType, parameters, dimensions, throws, default, contractWithin,
                                 body, srcLocation, precedingWhitespace) =>
        val irReturn = returnType.map(visitTypeIndication)
        val irDims = dimensions.map(visitStringNode)
        val irThrows = throws.map(visitThrowsClause)
        val irDefault = default.map(visitDefaultClause)
        val irContractWithin = contractWithin.map(visitContract)
        val res = new vercors.MethodDeclaration(visitStringNode(name),
                                            irReturn,
                                            parameters.map(visitParameter),
                                            irDims,
                                            irThrows,
                                            irDefault,
                                            irContractWithin,
                                            visitBlockOrSemicolon(body),
                                            srcLocation,
                                            precedingWhitespace)
        res.contractBefore = if(node.contractBefore.isEmpty) None else Some(visitContract(node.contractBefore.get))
        res.modifiers = reorderModifiers(node.modifiers.map(visitModifier))
        if (checkForNonNullModifier(node.modifiers)) {
          val expr = vercors.BinaryExpression(vercors.ResultExpression(srcLocation, " "),
            vercors.StringNode("!=", srcLocation),
            vercors.StringNode("null",srcLocation),
            srcLocation)
          val clause = vercors.Postcondition(expr, srcLocation,System.lineSeparator() + "    ")
          if(res.contractBefore.isDefined) {
            res.contractBefore.get.addClauseAfterContract[vercors.VerCorsTreeNode](clause, wrapInSpec, vercors.StringNode.apply)
          } else if(res.contractWithin.isDefined) {
            res.contractWithin.get.addClauseAfterContract[vercors.VerCorsTreeNode](clause, wrapInSpec, vercors.StringNode.apply)
          } else {
            val contract = vercors.Contract(Nil, srcLocation)
            if(returnType.isDefined) {
              contract.precedingWhitespace = returnType.get.precedingWhitespace
            }
            contract.addClauseAfterContract[vercors.VerCorsTreeNode](clause, wrapInSpec, vercors.StringNode.apply)
            res.contractBefore = Some(contract)
          }
        }
        val nonnulls = parameters.filter(x => checkForNonNullModifier(x.modifiers))
        if(nonnulls.nonEmpty) {
          var expr = vercors.BinaryExpression(vercors.StringNode(nonnulls.head.name.text, srcLocation, " "),
            vercors.StringNode("!=", srcLocation),
            vercors.StringNode("null",srcLocation),
            srcLocation)
          for(param <- nonnulls.tail) {
            val neq = vercors.BinaryExpression(vercors.StringNode(nonnulls.head.name.text, srcLocation, " "),
              vercors.StringNode("!=", srcLocation),
              vercors.StringNode("null",srcLocation),
              srcLocation)
            expr = vercors.BinaryExpression(expr, vercors.StringNode("&&",srcLocation, " "), neq, srcLocation)
          }
          val clause = vercors.Precondition(expr, srcLocation)
          if(res.contractBefore.isDefined) {
            res.contractBefore.get.addClauseBeforeContract[vercors.VerCorsTreeNode](clause, wrapInSpec, vercors.StringNode.apply)
          } else if(res.contractWithin.isDefined) {
            res.contractWithin.get.addClauseBeforeContract[vercors.VerCorsTreeNode](clause, wrapInSpec, vercors.StringNode.apply)
          } else {
            val contract = vercors.Contract(Nil, srcLocation)
            if(returnType.isDefined) {
              contract.precedingWhitespace = returnType.get.precedingWhitespace
            }
            contract.addClauseBeforeContract[vercors.VerCorsTreeNode](clause, wrapInSpec, vercors.StringNode.apply)
            res.contractBefore = Some(contract)
          }
        }
        res.copyWhitespace(node)
        res
    }
  }


  def visitGenericMethodDeclaration(node: ir.GenericMethodDeclaration): vercors.GenericMethodDeclaration = {
    node match {
      case ir.GenericMethodDeclaration(pName, typeParameters, pReturnType,
                                        pParameters, pDimensions, pThrows, pDefault, pContractWithin,
                                        pBody, pSrcLocation, pPrecedingWhitespace) =>
        val irReturn = pReturnType.map(visitTypeIndication)
        val irDims = pDimensions.map(visitStringNode)
        val irThrows = pThrows.map(visitThrowsClause)
        val irDefault = pDefault.map(visitDefaultClause)
        val irContractWithin = pContractWithin.map(visitContract)
        val res = vercors.GenericMethodDeclaration(visitStringNode(pName),
                                              visitTypeParameterClause(typeParameters),
                                              irReturn,
                                              pParameters.map(visitParameter),
                                              irDims,
                                              irThrows,
                                              irDefault,
                                              irContractWithin,
                                              visitBlockOrSemicolon(pBody),
                                              pSrcLocation,
                                              pPrecedingWhitespace)
//        res.contractBefore = if(node.contractBefore.isEmpty) None else Some(visitContract(node.contractBefore.get))
        res.modifiers = reorderModifiers(node.modifiers.map(visitModifier))
        res.copyWhitespace(node)
        res
    }
  }

  def visitThrowsClause(node: ir.ThrowsClause): vercors.ThrowsClause = {
    node match {
      case ir.ThrowsClause(exceptions, srcLocation, precedingWhitespace) =>
        val res = vercors.ThrowsClause(exceptions.map(visitStringNode),
                                  srcLocation,
                                  precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitTypeIndication(node: ir.TypeIndication): vercors.TypeIndication = {
    node match {
      case indication: ir.SpecTypeIndication => visitSpecTypeIndication(indication)
      case indication: ir.NoType => visitNoType(indication)
      case indication: ir.StringGenericTypeIndication => visitStringGenericTypeIndication(indication)
      case indication: ir.StringTypeIndication => visitStringTypeIndication(indication)
      case _ => ???
    }
  }

  def visitSpecTypeIndication(node: ir.SpecTypeIndication): vercors.SpecTypeIndication = {
    node match {
      case int: ir.IntegerSpecTypeIndication => visitIntegerSpecTypeIndication(int)
      case real: ir.RealSpecTypeIndication => visitRealSpecTypeIndication(real)
    }
  }

  def visitIntegerSpecTypeIndication(node: ir.IntegerSpecTypeIndication): vercors.IntegerSpecTypeIndication = {
    node match {
      case ir.IntegerSpecTypeIndication(fullTypeString, anns, typeAsString, location, whitespace) =>
        val res = vercors.IntegerSpecTypeIndication(annotations = anns.map(visitAnnotation), srcLocation = location, precedingWhitespace = whitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitRealSpecTypeIndication(node: ir.RealSpecTypeIndication): vercors.SpecTypeIndication = {
    throw new NotSupportedException("real type", node.srcLocation)
  }

  def visitStringTypeIndication(node: ir.StringTypeIndication): vercors.StringTypeIndication = {
    node match {
      case ir.StringTypeIndication(fullTypeString, annotations, typeAsString,
                                    pSrcLocation, pPrecedingWhitespace, pTrailingWhitespace) =>
        val res = new vercors.StringTypeIndication(fullTypeString,
                                              annotations.map(visitAnnotation),
                                              typeAsString,
                                              pSrcLocation,
                                              pPrecedingWhitespace,
                                              pTrailingWhitespace)
        res.modifiers = reorderModifiers(node.modifiers.map(visitModifier))
        res.copyWhitespace(node)
        res
    }
  }

  def visitStringGenericTypeIndication(node: ir.StringGenericTypeIndication): vercors.StringGenericTypeIndication = {
    node match {
      case ir.StringGenericTypeIndication(fullTypeString, typeParameters, annotations, typeAsString,
                                            pSrcLocation, pPrecedingWhitespace) =>
        val res = vercors.StringGenericTypeIndication(fullTypeString,
                                                  visitTypeParameterClause(typeParameters),
                                                  annotations.map(visitAnnotation),
                                                  typeAsString,
                                                  pSrcLocation,
                                                  pPrecedingWhitespace)
        res.modifiers = reorderModifiers(node.modifiers.map(visitModifier))
        res.copyWhitespace(node)
        res
    }
  }
  def visitNoType(node: ir.NoType): vercors.NoType = {
    node match {
      case ir.NoType(srcLocation) =>
        new vercors.NoType(srcLocation)
    }
  }

  def visitBlockOrSemicolon(node: ir.BlockOrSemicolon): vercors.BlockOrSemicolon = {
    node match {
      case block: ir.Block => visitBlock(block)
      case semi: ir.EmptyStatement => visitEmptyStatement(semi)
      case _ => ???
    }
  }

  def visitFieldDeclaration(node: ir.FieldDeclaration): vercors.FieldDeclaration = {
    node match {
      case ir.FieldDeclaration(fieldType, declarators, srcLocation, precedingWhitespace) =>
        val res = vercors.FieldDeclaration(visitTypeIndication(fieldType),
                                      declarators.map(visitVariableDeclarator),
                                      srcLocation,
                                      precedingWhitespace)
        res.modifiers = reorderModifiers(node.modifiers.map(visitModifier))
        res.copyWhitespace(node)
        res
    }
  }

  def visitVariableDeclarator(node: ir.VariableDeclarator): vercors.VariableDeclarator = {
    node match {
      case ir.VariableDeclarator(name, initializer, srcLocation, precedingWhitespace) =>
        val irInit = initializer.map(visitInitializer)
        val res = vercors.VariableDeclarator(visitStringNode(name),
                                        irInit,
                                        srcLocation,
                                        precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitInitializer(node: ir.Initializer): vercors.Initializer = {
    node match {
      case init: ir.ArrayInitializer => visitArrayInitializer(init)
      case init: ir.Expression => visitExpression(init)
      case _ => ???
    }
  }
  def visitArrayInitializer(node: ir.ArrayInitializer): vercors.ArrayInitializer = {
    node match {
      case ir.ArrayInitializer(initializers, trailingComma, srcLocation, precedingWhitespace) =>
        val irComma = trailingComma.map(visitStringNode)
        val res = vercors.ArrayInitializer(initializers.map(visitInitializer),
                                      irComma,
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitTypeArgumentClause(node: ir.TypeArgumentClause): vercors.TypeArgumentClause = {
    node match {
      case ir.TypeArgumentClause(parameters, srcLocation, precedingWhitespace) =>
        val res = vercors.TypeArgumentClause(parameters.map(visitTypeArgument),
                                        srcLocation,
                                        precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitTypeArgument(node: ir.TypeArgument): vercors.TypeArgument = {
    node match {
      case arg: ir.BoundedTypeArgument => visitBoundedTypeArgument(arg)
      case arg: ir.TypeIndication => visitTypeIndication(arg)
      case _ => ???
    }
  }
  def visitBoundedTypeArgument(node: ir.BoundedTypeArgument): vercors.BoundedTypeArgument = {
    node match {
      case ir.BoundedTypeArgument(annotations, typeSuper, typeExtends, srcLocation, precedingWhitespace) =>
        val irSuper = typeSuper.map(visitTypeIndication)
        val irExtends = typeExtends.map(visitTypeIndication)
        val res = vercors.BoundedTypeArgument(annotations.map(visitAnnotation),
                                          irSuper,
                                          irExtends,
                                          srcLocation,
                                          precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitParameter(node: ir.Parameter): vercors.Parameter = {
    node match {
      case varargs: ir.VarargsParameter =>
        visitVarargsParameter(varargs)
      case ir.Parameter(parameterType, name, srcLocation, precedingWhitespace) =>
        val res = vercors.Parameter(visitTypeIndication(parameterType),
                                visitStringNode(name),
                                srcLocation,
                                precedingWhitespace)
        res.modifiers = reorderModifiers(node.modifiers.map(visitModifierOfParam))
        res.copyWhitespace(node)
        res
    }
  }
  def visitVarargsParameter(node: ir.VarargsParameter): vercors.VarargsParameter = {
    node match {
      case ir.VarargsParameter(parameterType, annotations, name, srcLocation, precedingWhitespace) =>
        val res = vercors.VarargsParameter(visitTypeIndication(parameterType),
                                      annotations.map(visitAnnotation),
                                      visitStringNode(name),
                                      srcLocation,
                                      precedingWhitespace)
        res.modifiers = reorderModifiers(node.modifiers.map(visitModifierOfParam))
        res.copyWhitespace(node)
        res
    }
  }

  def visitAnnotation(node: ir.Annotation): vercors.Annotation = {
    node match {
      case ann: ir.StringAnnotation => visitStringAnnotation(ann)
      case _ => ???
    }
  }

  def visitStringAnnotation(node: ir.StringAnnotation): vercors.StringAnnotation = {
    node match {
      case ir.StringAnnotation(text, srcLocation, precedingWhitespace) =>
        val res = vercors.StringAnnotation(text,
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitAnnotationTypeDeclaration(node: ir.AnnotationTypeDeclaration): vercors.AnnotationTypeDeclaration = {
    node match {
      case ir.AnnotationTypeDeclaration(name, body, srcLocation, precedingWhitespace) =>
        val res = vercors.AnnotationTypeDeclaration(visitStringNode(name),
                                                visitBody(body),
                                                srcLocation,
                                                precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitDefaultClause(node: ir.DefaultClause): vercors.DefaultClause = {
    node match {
      case ir.DefaultClause(value, srcLocation, precedingWhitespace) =>
        val res = vercors.DefaultClause(visitStringNode(value),
                                    srcLocation,
                                    precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitStatement(node: ir.Statement): vercors.Statement = {
    try
      node match {
        case local: ir.LocalDeclaration => visitLocalDeclaration(local)
        case block: ir.Block => visitBlock(block)
        case assert: ir.JavaAssertion => visitJavaAssertion(assert)
        case ifStmt: ir.IfStatement => visitIfStatement(ifStmt)
        case forLoop: ir.ForLoop => visitForLoop(forLoop)
        case whileLoop: ir.WhileLoop => visitWhileLoop(whileLoop)
        case doLoop: ir.DoLoop => visitDoLoop(doLoop)
        case tryBlock: ir.TryBlock => visitTryBlock(tryBlock)
        case switch: ir.SwitchStatement => visitSwitchStatement(switch)
        case sync: ir.SynchronizedStatement => visitSynchronizedStatement(sync)
        case ret: ir.ReturnStatement => visitReturnStatement(ret)
        case throwStmt: ir.ThrowStatement => visitThrowStatement(throwStmt)
        case break: ir.BreakStatement => visitBreakStatement(break)
        case continue: ir.ContinueStatement => visitContinueStatement(continue)
        case empty: ir.EmptyStatement => visitEmptyStatement(empty)
        case expr: ir.ExpressionStatement => visitExpressionStatement(expr)
        case labeled: ir.LabeledStatement => visitLabeledStatement(labeled)
        case assert: ir.AssertStatement => visitAssertStatement(assert)
        case assume: ir.AssumeStatement => visitAssumeStatement(assume)
        case set: ir.SetStatement => visitSetStatement(set)
        case ghost: ir.GhostStatement => visitGhostStatement(ghost)
        case unreach: ir.UnreachableStatement => visitUnreachableStatement(unreach)
        case vercorsBlock: ir.JmlStatementBlock => visitJmlStatementBlock(vercorsBlock)
        case nontranslated: ir.NonTranslatedStatement => visitNonTranslatedStatement(nontranslated)
        case _ => ???
        // todo: ClassDecl, etc.
      }
    catch {
      case e: NotSupportedException =>
        val text = Helper.commentUnknownNode(node, s"${e.construct} not supported, whole Statement commented out")
        vercors.NonTranslatedStatement(text, node.srcLocation)
    }
  }

  def visitBlock(node: ir.Block): vercors.Block = {
    node match {
      case ir.Block(static, statements, srcLocation, precedingWhitespace) =>
        val irStatic = static.map(visitModifier)
        val res = vercors.Block(irStatic,
                            statements.map(visitStatement),
                            srcLocation,
                            precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitLocalDeclaration(node: ir.LocalDeclaration): vercors.LocalDeclaration = {
    node match {
      case ir.LocalDeclaration(modifiers, declType, declarators, isStatement, srcLocation, precedingWhitespace) =>
        val res = vercors.LocalDeclaration(reorderModifiers(modifiers.map(visitVariableModifier)),
                                      visitTypeIndication(declType),
                                      declarators.map(visitVariableDeclarator),
                                      isStatement,
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitJavaAssertion(node: ir.JavaAssertion): vercors.JavaAssertion = {
    node match {
      case ir.JavaAssertion(content, message, srcLocation, precedingWhitespace) =>
        val irMsg = message.map(visitExpression)
        val res = vercors.JavaAssertion(visitExpression(content),
                                    irMsg,
                                    srcLocation,
                                    precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitIfStatement(node: ir.IfStatement): vercors.IfStatement = {
    node match {
      case ir.IfStatement(condition, thenStatement, elseStatement, srcLocation, precedingWhitespace) =>
        val irElse = elseStatement.map(visitStatement)
        val res = vercors.IfStatement(visitExpression(condition),
                                  visitStatement(thenStatement),
                                  irElse,
                                  srcLocation,
                                  precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitForLoop(node: ir.ForLoop): vercors.ForLoop = {
    node match {
      case ir.ForLoop(contractBefore, control, contractWithin, body, srcLocation, precedingWhitespace) =>
        val irContractBefore = if (contractBefore.isEmpty) Nil else contractBefore.map(visitLoopSpecOrLabel)
        val irContractWithin = if (contractWithin.isEmpty) Nil else contractWithin.map(visitLoopSpecOrLabel)
        val res = vercors.ForLoop(irContractBefore,
                              visitForLoopControl(control),
                              irContractWithin,
                              visitStatement(body),
                              srcLocation,
                              precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitForLoopControl(node: ir.ForLoopControl): vercors.ForLoopControl = {
    node match {
      case init: ir.ForEachLoopControl => visitForEachLoopControl(init)
      case init: ir.RegularForLoopControl => visitRegularForLoopControl(init)
    }
  }
  def visitForInit(node: ir.ForInit): vercors.ForInit = {
    node match {
      case init: ir.ExpressionListForInit =>
        visitExpressionListForInit(init)
      case init: ir.LocalDeclaration =>
        visitLocalDeclaration(init)
      case _ =>
        ???
    }
  }
  def visitExpressionListForInit(node: ir.ExpressionListForInit): vercors.ExpressionListForInit = {
    node match {
      case ir.ExpressionListForInit(expressions, srcLocation, precedingWhitespace) =>
        val res = vercors.ExpressionListForInit(expressions.map(visitExpression),
                                          srcLocation,
                                          precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitRegularForLoopControl(node: ir.RegularForLoopControl): vercors.RegularForLoopControl = {
    node match {
      case ir.RegularForLoopControl(init, condition, update, srcLocation, precedingWhitespace) =>
        val irInit = init.map(visitForInit)
        val irCondition = condition.map(visitExpression)
        val res = vercors.RegularForLoopControl(irInit,
                                            irCondition,
                                            update.map(visitExpression),
                                            srcLocation,
                                            precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitForEachLoopControl(node: ir.ForEachLoopControl): vercors.ForEachLoopControl = {
    node match {
      case ir.ForEachLoopControl(controlType, variable, control, srcLocation, precedingWhitespace) =>
        val res = vercors.ForEachLoopControl(visitTypeIndication(controlType),
                                        visitStringNode(variable),
                                        visitExpression(control),
                                        srcLocation,
                                        precedingWhitespace)
        res.modifiers = node.modifiers.map(visitVariableModifier)
        res.copyWhitespace(node)
        res
    }
  }

  def visitWhileLoop(node: ir.WhileLoop): vercors.WhileLoop = {
    node match {
      case ir.WhileLoop(contractBefore, condition, contractWithin, body, srcLocation, precedingWhitespace) =>
        val irContractBefore = if (contractBefore.isEmpty) Nil else contractBefore.map(visitLoopSpecOrLabel)
        val irContractWithin = if (contractWithin.isEmpty) Nil else contractWithin.map(visitLoopSpecOrLabel)
        val res = vercors.WhileLoop(irContractBefore,
                                visitExpression(condition),
                                irContractWithin,
                                visitStatement(body),
                                srcLocation,
                                precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitDoLoop(node: ir.DoLoop): vercors.DoLoop = {
    node match {
      case ir.DoLoop(contractBefore, body, condition, srcLocation, precedingWhitespace) =>
        val irContractBefore = if (contractBefore.isEmpty) Nil else contractBefore.map(visitLoopSpecOrLabel)
        val res = vercors.DoLoop(irContractBefore,
                            visitStatement(body),
                            visitExpression(condition),
                            srcLocation,
                            precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitTryBlock(node: ir.TryBlock): vercors.TryBlock = {
    node match {
      case ir.TryBlock(resourceSpecification, tryBlock, catchClauses, finallyBlock, srcLocation, precedingWhitespace) =>
        val irResSpec = if (resourceSpecification.isEmpty) None
                        else Some(visitTryResourceSpecification(resourceSpecification.get))
        val irFinally = finallyBlock.map(visitBlock)
        val res = vercors.TryBlock(irResSpec,
                              visitBlock(tryBlock),
                              catchClauses.map(visitCatchClause),
                              irFinally,
                              srcLocation,
                              precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitCatchClause(node: ir.CatchClause): vercors.CatchClause = {
    node match {
      case ir.CatchClause(types, name, block, srcLocation, precedingWhitespace) =>
        val res = vercors.CatchClause(types.map(visitTypeIndication),
                                  visitStringNode(name),
                                  visitBlock(block),
                                  srcLocation,
                                  precedingWhitespace)
        res.modifiers = node.modifiers.map(visitVariableModifier)
        res.copyWhitespace(node)
        res
    }
  }
  def visitTryResourceSpecification(node: ir.TryResourceSpecification): vercors.TryResourceSpecification = {
    node match {
      case ir.TryResourceSpecification(resources, trailingSemicolon, srcLocation, precedingWhitespace) =>
        val irSemi = trailingSemicolon.map(visitStringNode)
        val res = vercors.TryResourceSpecification(resources.map(visitTryResource),
                                              irSemi,
                                              srcLocation,
                                              precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitTryResource(node: ir.TryResource): vercors.TryResource = {
    node match {
      case ir.TryResource(modifiers, resourceType, variable, expression, srcLocation, precedingWhitespace) =>
        val res = vercors.TryResource(modifiers.map(visitModifier),
                                  visitTypeIndication(resourceType),
                                  visitStringNode(variable),
                                  visitExpression(expression),
                                  srcLocation,
                                  precedingWhitespace)
        res.modifiers = reorderModifiers(node.modifiers.map(visitModifier))
        res.copyWhitespace(node)
        res
    }
  }

  def visitSwitchStatement(node: ir.SwitchStatement): vercors.SwitchStatement = {
    node match {
      case ir.SwitchStatement(condition, cases, srcLocation, precedingWhitespace) =>
        val res = vercors.SwitchStatement(visitExpression(condition),
                                      cases.map(visitSwitchCaseBlock),
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitSwitchCaseBlock(node: ir.SwitchCaseBlock): vercors.SwitchCaseBlock = {
    node match {
      case ir.SwitchCaseBlock(labels, statements, srcLocation, precedingWhitespace) =>
        val res = vercors.SwitchCaseBlock(labels.map(visitSwitchLabel),
                                      statements.map(visitStatement),
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitSwitchLabel(node: ir.SwitchLabel): vercors.SwitchLabel = {
    node match {
      case ir.SwitchLabel(expression, srcLocation, precedingWhitespace) =>
        val irExpr = expression.map(visitExpression)
        val res = vercors.SwitchLabel(irExpr,
                                  srcLocation,
                                  precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitSynchronizedStatement(node: ir.SynchronizedStatement): vercors.SynchronizedStatement = {
    node match {
      case ir.SynchronizedStatement(condition, body, srcLocation, precedingWhitespace) =>
        val res = vercors.SynchronizedStatement(visitExpression(condition),
                                            visitStatement(body),
                                            srcLocation,
                                            precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitReturnStatement(node: ir.ReturnStatement): vercors.ReturnStatement = {
    node match {
      case ir.ReturnStatement(expression, srcLocation, precedingWhitespace) =>
        val irExpr = expression.map(visitExpression)
        val res = vercors.ReturnStatement(irExpr,
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitThrowStatement(node: ir.ThrowStatement): vercors.ThrowStatement = {
    node match {
      case ir.ThrowStatement(expression, srcLocation, precedingWhitespace) =>
        val res = vercors.ThrowStatement(visitExpression(expression),
                                    srcLocation,
                                    precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitBreakStatement(node: ir.BreakStatement): vercors.BreakStatement = {
    node match {
      case ir.BreakStatement(label, srcLocation, precedingWhitespace) =>
        val irLabel = label.map(visitStringNode)
        val res = vercors.BreakStatement(irLabel,
                                    srcLocation,
                                    precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitContinueStatement(node: ir.ContinueStatement): vercors.ContinueStatement = {
    node match {
      case ir.ContinueStatement(label, srcLocation, precedingWhitespace) =>
        val irLabel = label.map(visitStringNode)
        val res = vercors.ContinueStatement(irLabel,
                                        srcLocation,
                                        precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitEmptyStatement(node: ir.EmptyStatement): vercors.EmptyStatement = {
    val res = new vercors.EmptyStatement(node.srcLocation)
    res.copyWhitespace(node)
    res
  }
  def visitExpressionStatement(node: ir.ExpressionStatement): vercors.ExpressionStatement = {
    node match {
      case ir.ExpressionStatement(expression, srcLocation, precedingWhitespace) =>
        val res = vercors.ExpressionStatement(visitExpression(expression),
                                          srcLocation,
                                          precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitLabeledStatement(node: ir.LabeledStatement): vercors.LabeledStatement = {
    node match {
      case ir.LabeledStatement(label, statement, srcLocation, precedingWhitespace) =>
        val res = vercors.LabeledStatement(visitLabel(label),
                                      visitStatement(statement),
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitLabel(node: ir.Label): vercors.Label = {
    node match {
      case ir.Label(label, srcLocation, precedingWhitespace) =>
        val res = vercors.Label(visitStringNode(label),
                            srcLocation,
                            precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitArguments(node: ir.Arguments): vercors.Arguments = {
    node match {
      case ir.Arguments(arguments, srcLocation, precedingWhitespace) =>
        val res = vercors.Arguments(arguments.map(visitExpression),
                                srcLocation,
                                precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }


  def visitExpression(node: ir.Expression): vercors.Expression = {
    node match {
      case expr: ir.NestedInvocationExpression => visitNestedInvocationExpression(expr)
      case expr: ir.BinaryExpression => visitBinaryExpression(expr)
      case expr: ir.InstanceofExpression => visitInstanceofExpression(expr)
      case expr: ir.ArrayIndexExpression => visitArrayIndexExpression(expr)
      case expr: ir.MethodCallExpression => visitMethodCallExpression(expr)
      case expr: ir.NewObjectExpression => visitNewObjectExpression(expr)
      case expr: ir.NewArrayExpression => visitNewArrayExpression(expr)
      case expr: ir.CastExpression => visitCastExpression(expr)
      case expr: ir.PostfixExpression => visitPostfixExpression(expr)
      case expr: ir.PrefixExpression => visitPrefixExpression(expr)
      case expr: ir.TernaryExpression => visitTernaryExpression(expr)
      case expr: ir.LambdaExpression => visitLambdaExpression(expr)
      case expr: ir.MethodReferenceExpression => visitMethodReferenceExpression(expr)
      case expr: ir.ParExpression => visitParExpression(expr)
      case expr: ir.Literal => visitLiteral(expr)
      case expr: ir.ClassName => visitClassName(expr)
      case expr: ir.GenericInvocationExpression => visitGenericInvocationExpression(expr)
      case expr: ir.OldExpression => visitOldExpression(expr)
      case expr: ir.PreExpression => visitPreExpression(expr)
      case expr: ir.ResultExpression => visitResultExpression(expr)
      case expr: ir.NotAssignedExpression => visitNotAssignedExpression(expr)
      case expr: ir.NotModifiedExpression => visitNotModifiedExpression(expr)
      case expr: ir.OnlyAccessedExpression => visitOnlyAccessedExpression(expr)
      case expr: ir.OnlyAssignedExpression => visitOnlyAssignedExpression(expr)
      case expr: ir.FreshExpression => visitFreshExpression(expr)
      case expr: ir.NonNullElementsExpression => visitNonNullElementsExpression(expr)
      case expr: ir.TypeOfExpression => visitTypeOfExpression(expr)
      case expr: ir.TypeExpression => visitTypeExpression(expr)
      case expr: ir.ForallExpression => visitForallExpression(expr)
      case expr: ir.ExistsExpression => visitExistsExpression(expr)
      case expr: ir.MaxQuantifiedExpression => visitMaxQuantifiedExpression(expr)
      case expr: ir.MinExpression => visitMinExpression(expr)
      case expr: ir.NumOfExpression => visitNumOfExpression(expr)
      case expr: ir.ProductExpression => visitProductExpression(expr)
      case expr: ir.SumExpression => visitSumExpression(expr)
      case expr: ir.StringNode =>
        if(expr.text.contains("\\")) {
          throw new NotSupportedException(s"reserved name \"${expr.text}\"", expr.srcLocation)
        }
        visitStringNode(expr)
      case expr: ir.SetComprehension => visitSetComprehension(expr)
      case expr: ir.ValuedPermissionExpression => visitValuedPermissionExpression(expr)
      case expr: ir.CurrentPermissionExpression => visitCurrentPermissionExpression(expr)
      case expr: ir.SomePermissionPredicate => visitSomePermissionPredicate(expr)
      case expr: ir.PermissionExpression => visitPermissionExpression(expr)
      case expr: ir.ScaledPredicateExpression => visitScaledPredicateExpression(expr)
      case expr: ir.UnfoldingExpression => visitUnfoldingExpression(expr)
      case expr: ir.ForallWithPermissionsExpression => visitForallWithPermissionsExpression(expr)
      case _ => ???
    }
    //todo: more?

  }

  def visitNestedInvocationExpression(node: ir.NestedInvocationExpression): vercors.NestedInvocationExpression = {
    node match {
      case ir.NestedInvocationExpression(left, right, srcLocation, precedingWhitespace) =>
        val res = vercors.NestedInvocationExpression(visitExpression(left),
                                                visitExpression(right),
                                                srcLocation,
                                                precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitBinaryExpression(node: ir.BinaryExpression): vercors.BinaryExpression = {
    node match {
      case ir.BinaryExpression(left, operator, right, srcLocation, precedingWhitespace) =>
        val relOps = List("<", ">", "<=", ">=")

        var leftVC = visitExpression(left)
        var rightVC = visitExpression(right)
        var res: vercors.BinaryExpression = null
        if(relOps.contains(operator.text)) {
          left match {
            case bin: ir.BinaryExpression if relOps.contains(bin.operator.text) =>
              val copiedMiddle = visitExpression(bin.right)
              rightVC = vercors.BinaryExpression(copiedMiddle, visitStringNode(operator),rightVC,srcLocation)
              res = vercors.BinaryExpression(leftVC,
                                        vercors.StringNode("&&",operator.srcLocation, " "),
                                        rightVC,
                                        srcLocation,
                                        precedingWhitespace)
            case _ =>
          }
          right match {
            case bin: ir.BinaryExpression if relOps.contains(bin.operator.text) =>
              val copiedMiddle = visitExpression(bin.left)
              leftVC = vercors.BinaryExpression(leftVC, visitStringNode(operator),copiedMiddle,srcLocation)
              res = vercors.BinaryExpression(leftVC,
                                        vercors.StringNode("&&",operator.srcLocation, " "),
                                        rightVC,
                                        srcLocation,
                                        precedingWhitespace)
            case _ =>
          }
        }
        if(res == null) {
          res = vercors.BinaryExpression(leftVC,
                                      visitStringNode(operator),
                                      rightVC,
                                      srcLocation,
                                      precedingWhitespace)
        }
        res.copyWhitespace(node)
        res
    }
  }

  def visitInstanceofExpression(node: ir.InstanceofExpression): vercors.InstanceofExpression = {
    node match {
      case ir.InstanceofExpression(left, right, srcLocation, precedingWhitespace) =>
        val res = vercors.InstanceofExpression(visitExpression(left),
                                                visitTypeIndication(right),
                                                srcLocation)
        res.copyWhitespace(node)
        res
    }
  }

  def visitArrayIndexExpression(node: ir.ArrayIndexExpression): vercors.ArrayIndexExpression = {
    node match {
      case ir.ArrayIndexExpression(arrayExpression, indexExpression, srcLocation, precedingWhitespace) =>
        val res = vercors.ArrayIndexExpression(visitExpression(arrayExpression),
                                          visitIndexExpression(indexExpression),
                                          srcLocation,
                                          precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitIndexExpression(node: ir.IndexExpression): vercors.IndexExpression = {
    node match {
      case idx: ir.IndexAllExpression => visitIndexAllExpression(idx)
      case idx: ir.RangeExpression => visitRangeExpression(idx)
      case idx: ir.Expression => visitExpression(idx)
      case _ => ???
    }
  }

  def visitIndexAllExpression(node: ir.IndexAllExpression): vercors.IndexAllExpression = {
    node match {
      case ir.IndexAllExpression(srcLocation, precedingWhitespace) =>
        val res = vercors.IndexAllExpression(srcLocation, precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitRangeExpression(node: ir.RangeExpression): vercors.RangeExpression = {
    node match {
      case ir.RangeExpression(start, end, srcLocation, precedingWhitespace) =>
        val res = vercors.RangeExpression(visitExpression(start),
                                      visitExpression(end),
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitMethodCallExpression(node: ir.MethodCallExpression): vercors.MethodCallExpression = {
    node match {
      case ir.MethodCallExpression(name, arguments, srcLocation, precedingWhitespace) =>
        val res = vercors.MethodCallExpression(visitStringNode(name),
                                          visitArguments(arguments),
                                          srcLocation,
                                          precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitNewObjectExpression(node: ir.NewObjectExpression): vercors.NewObjectExpression = {
    node match {
      case ir.NewObjectExpression(typeArguments, typeIndication, arguments, body, srcLocation, precedingWhitespace) =>
        val typeArgs = typeArguments.map(visitTypeArgumentClause)
        val irBody = body.map(visitBody)
        val res = vercors.NewObjectExpression(typeArgs,
                                          visitTypeIndication(typeIndication),
                                          visitArguments(arguments),
                                          irBody,
                                          srcLocation,
                                          precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitNewArrayExpression(node: ir.NewArrayExpression): vercors.NewArrayExpression = {
    node match {
      case ir.NewArrayExpression(typeIndication, dimensions, initializer, srcLocation, precedingWhitespace) =>
        val irInit = initializer.map(visitArrayInitializer)
        val res = vercors.NewArrayExpression(visitTypeIndication(typeIndication),
                                        dimensions.map(visitDimension),
                                        irInit,
                                        srcLocation,
                                        precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitDimension(node: ir.Dimension): vercors.Dimension = {
    node match {
      case ir.Dimension(sizeExpression, srcLocation, precedingWhitespace) =>
        val irSize = sizeExpression.map(visitExpression)
        val res = vercors.Dimension(irSize,
                                srcLocation,
                                precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitCastExpression(node: ir.CastExpression): vercors.CastExpression = {
    node match {
      case ir.CastExpression(annotations, typeIndications, expression, srcLocation, precedingWhitespace) =>
        val res = vercors.CastExpression(annotations.map(visitStringAnnotation),
                                    typeIndications.map(visitTypeIndication),
                                    visitExpression(expression),
                                    srcLocation,
                                    precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitPostfixExpression(node: ir.PostfixExpression): vercors.PostfixExpression = {
    node match {
      case ir.PostfixExpression(expression, postfix, srcLocation, precedingWhitespace) =>
        val res = vercors.PostfixExpression(visitExpression(expression),
                                        visitStringNode(postfix),
                                        srcLocation,
                                        precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitPrefixExpression(node: ir.PrefixExpression): vercors.PrefixExpression = {
    node match {
      case ir.PrefixExpression(expression, prefix, srcLocation, precedingWhitespace) =>
        val res = vercors.PrefixExpression(visitExpression(expression),
                                      visitStringNode(prefix),
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitTernaryExpression(node: ir.TernaryExpression): vercors.TernaryExpression = {
    node match {
      case ir.TernaryExpression(condition, thenExpression, elseExpression, srcLocation, precedingWhitespace) =>
        val res = vercors.TernaryExpression(visitExpression(condition),
                                        visitExpression(thenExpression),
                                        visitExpression(elseExpression),
                                        srcLocation,
                                        precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitLambdaExpression(node: ir.LambdaExpression): vercors.LambdaExpression = {
    node match {
      case ir.LambdaExpression(hasParentheses, parameters, body, srcLocation, precedingWhitespace) =>
        val res = vercors.LambdaExpression(hasParentheses,
                                      parameters.map(visitParameter),
                                      visitLambdaBody(body),
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitLambdaBody(node: ir.LambdaBody): vercors.LambdaBody = {
    node match {
      case body: ir.Expression => visitExpression(body)
      case _ => ???
    }
  }

  def visitMethodReferenceExpression(node: ir.MethodReferenceExpression): vercors.MethodReferenceExpression = {
    node match {
      case ir.MethodReferenceExpression(left, typeArgument, right, srcLocation, precedingWhitespace) =>
        val irArg = typeArgument.map(visitTypeArgumentClause)
        val res = vercors.MethodReferenceExpression(visitExpression(left),
                                                irArg,
                                                visitExpression(right),
                                                srcLocation,
                                                precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitParExpression(node: ir.ParExpression): vercors.ParExpression = {
    node match {
      case ir.ParExpression(expression, srcLocation, precedingWhitespace) =>
        val res = vercors.ParExpression(visitExpression(expression),
                                    srcLocation,
                                    precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitLiteral(node: ir.Literal): vercors.Literal = {
    node match {
      case ir.Literal(value, srcLocation, precedingWhitespace) =>
        if(value.text.contains("\\")) {
          throw new NotSupportedException(s"reserved name \"${value.text}\"", srcLocation)
        }
        val res = vercors.Literal(visitStringNode(value),
                              srcLocation,
                              precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitClassName(node: ir.ClassName): vercors.ClassName = {
    node match {
      case ir.ClassName(typeIndication, srcLocation, precedingWhitespace) =>
        val res = vercors.ClassName(visitTypeIndication(typeIndication),
                                srcLocation,
                                precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitGenericInvocationExpression(node: ir.GenericInvocationExpression): vercors.GenericInvocationExpression = {
    node match {
      case ir.GenericInvocationExpression(typeArgumentClause, name, arguments, srcLocation, precedingWhitespace) =>
        val irArgs = arguments.map(visitArguments)
        val res = vercors.GenericInvocationExpression(visitTypeArgumentClause(typeArgumentClause),
                                                  visitStringNode(name),
                                                  irArgs,
                                                  srcLocation,
                                                  precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitJmlModifierBlock(node: ir.JmlModifierBlock): vercors.JmlModifierBlock = {
    node match {
      case ir.JmlModifierBlock(startSpec, modifiers, endSpec, srcLocation, precedingWhitespace) =>
        def checkEmptyMod(modifier: vercors.Modifier): Boolean = {
          modifier match {
            case m: vercors.ModifierStringNode =>
              m.text.isEmpty
            case _ => false
          }
        }
        val irStart = startSpec.map(visitStringNode)
        val irEnd = endSpec.map(visitStringNode)
        val oldWrapInSpec = wrapInSpec
        val irModifiers = modifiers.map(visitModifier).filter(x => !checkEmptyMod(x))
        wrapInSpec = false
        val res = vercors.JmlModifierBlock(irStart,
          irModifiers,
          irEnd,
          srcLocation,
          precedingWhitespace)
        res.copyWhitespace(node)
        if(res.modifiers.isEmpty && res.startSpec.isDefined) {
          res.startSpec.get.text = res.startSpec.get.text.replace("@"," ")
        }
        wrapInSpec = oldWrapInSpec
        res
    }
  }

  def visitJmlModifierBlockForVariable(node: ir.JmlModifierBlock): vercors.Modifier = {
    node match {
      case ir.JmlModifierBlock(startSpec, modifiers, endSpec, srcLocation, precedingWhitespace) =>
        if(startSpec.isDefined) {
          val text = node.toString.replace("@", " ")
          System.err.println(s"VerCors does not support JML modifiers$srcLocation. Commented out.")
          new vercors.ModifierStringNode(text, srcLocation)
        } else {
          // VerCors does not see the "ghost" as a modifier, so that one is still allowed. Retain the first "ghost".
          if(modifiers.exists(x => x.isInstanceOf[ir.GhostModifier])) {
            if(modifiers.length > 1) {
              System.err.println(s"VerCors does not support JML modifiers$srcLocation, they are dropped")
            }
            visitGhostModifier(modifiers.filter(x => x.isInstanceOf[ir.GhostModifier]).head.asInstanceOf[ir.GhostModifier])
          } else {
            System.err.println(s"VerCors does not support JML modifiers$srcLocation, they are dropped")
            new vercors.ModifierStringNode("", srcLocation)
          }
        }
    }
  }
  def visitJmlModifierBlockForParameter(node: ir.JmlModifierBlock): vercors.Modifier = {
    node match {
      case ir.JmlModifierBlock(startSpec, modifiers, endSpec, srcLocation, precedingWhitespace) =>
        if(Helper.findModifier(modifiers, x => !x.isInstanceOf[ir.NullableModifier] && !x.isInstanceOf[ir.NonnullModifier])) {
          if(startSpec.isDefined) {
            val text = node.toString.replace("@", " ")
            System.err.println(s"VerCors does not support JML modifiers$srcLocation. Commented out.")
            new vercors.ModifierStringNode(text, srcLocation)
          } else {
            System.err.println(s"VerCors does not support JML modifiers$srcLocation, they are dropped")
            new vercors.ModifierStringNode("", srcLocation)
          }
        } else {
          new vercors.ModifierStringNode("", srcLocation)
        }
    }
  }

  def visitPureModifier(node: ir.PureModifier): vercors.PureModifier = {
    node match {
      case ir.PureModifier(srcLocation, precedingWhitespace) =>
        val res = vercors.PureModifier(srcLocation, precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitModelModifier(node: ir.ModelModifier): vercors.GhostModifier = {
    System.err.println("[info] Model modifier turned into Ghost.")
    val res = vercors.GhostModifier(node.srcLocation)
    res.copyWhitespace(node)
    res
  }
  def visitInstanceModifier(node: ir.InstanceModifier): vercors.ModifierStringNode = {
    node match {
      case ir.InstanceModifier(srcLocation, precedingWhitespace) =>
        System.err.println(s"[warning] Instance modifier not supported ${srcLocation.toString}, dropped.")
        val res = vercors.ModifierStringNode("", srcLocation)
        res.copyWhitespace(node)
        res
    }
  }

  def visitSpecMethodModifier(node: ir.SpecMethodModifier): vercors.SpecMethodModifier = {
    val res = vercors.SpecMethodModifier(node.srcLocation)
    res.copyWhitespace(node)
    res
  }
  def visitNullableModifier(node: ir.NullableModifier): vercors.ModifierStringNode = {
    // todo: comment out
    node match {
      case ir.NullableModifier(srcLocation, precedingWhitespace) =>
        val res = vercors.ModifierStringNode("", srcLocation, precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitNonnullModifier(node: ir.NonnullModifier): vercors.ModifierStringNode = {
    // todo: comment out
    node match {
      case ir.NonnullModifier(srcLocation, precedingWhitespace) =>
        val res = vercors.ModifierStringNode("", srcLocation, precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitGhostModifier(node: ir.GhostModifier): vercors.GhostModifier = {
    node match {
      case ir.GhostModifier(srcLocation, precedingWhitespace) =>
        val res = vercors.GhostModifier(srcLocation, precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitHelperModifier(node: ir.HelperModifier): vercors.ModifierStringNode = {
    // todo: comment out
    node match {
      case ir.HelperModifier(srcLocation, precedingWhitespace) =>
        val res = vercors.ModifierStringNode("", srcLocation, precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitMonitoredModifier(node: ir.MonitoredModifier): vercors.ModifierStringNode = {
    // todo: comment out
    node match {
      case ir.MonitoredModifier(srcLocation, precedingWhitespace) =>
        val res = vercors.ModifierStringNode("", srcLocation, precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  // check if there is a behavior-like construct in the contract, which needs translating
  def contractHasBehavior(clauses: List[ir.ContractClause]): Boolean = {
    if(clauses.exists(x => x.isInstanceOf[ir.Behavior])) {
      // there is an actual "behavior"
      true
    } else {
      val specCases = clauses.filter(x => x.isInstanceOf[ir.SpecCase])
      if(specCases.isEmpty) {
        // the whole contract is just a lightweight specification, no behavior-like things
        false
      } else if(specCases.size > 1) {
        // there are multiple spec cases
        true
      } else if (specCases != clauses) {
        // there are clauses besides the specCase
        true
      } else {
        // default behavior, triggered e.g. if there is exactly one lightweight spec case and no other clauses
        false
      }
    }
  }

  def addOldVariables(irClauses: List[ir.ContractClause], vcClauses: List[vercors.ContractClause]): List[vercors.ContractClause] = {
    val olds = irClauses.filter(x => x.isInstanceOf[ir.OldVariableDeclaration])
                        .map(x => x.asInstanceOf[ir.OldVariableDeclaration])
    if(olds.isEmpty) {
      vcClauses
    } else {
      def wrapWithOld(init: vercors.Initializer): vercors.Initializer = {
        init match {
          case a: vercors.ArrayInitializer =>
            val res = vercors.ArrayInitializer(a.initializers.map(wrapWithOld), a.trailingComma, a.srcLocation)
            res.copyWhitespace(a)
            res
          case e: vercors.Expression =>
            val o = vercors.OldExpression(e, None, init.srcLocation)
            o.oldKeyword.precedingWhitespace = " "
            o
          case _ =>
            // should be unreachable
            init
        }
      }
      def wrapWithLet(expr: vercors.Expression, useOld: Boolean=true): vercors.Expression = {
        var res = expr
        for (old <- olds) {
          for(decl <- old.declarators) {
            val v = visitVariableDeclarator(decl)
            val newInit = if(useOld) v.initializer.map(wrapWithOld)
                          else v.initializer
            val wrappedV = vercors.VariableDeclarator(v.name, newInit, v.srcLocation)
            wrappedV.copyWhitespace(v)
            val let = vercors.LetExpression(visitTypeIndication(old.typeIndication),
              wrappedV,
              expr,
              old.srcLocation)
            let.openParenthesis.precedingWhitespace = " "
            let.letKeyword.trailingWhitespace = " "
            res = let
          }
        }
        res
      }
      def addOld(clause: vercors.ContractClause): vercors.ContractClause = {
        clause match {
          case d: vercors.DecreasesClause =>
            val res = vercors.DecreasesClause(wrapWithLet(d.expression), d.srcLocation)
            res.startSpec = clause.startSpec
            res.endSpec = clause.endSpec
            res.copyWhitespace(d)
            res
          case e: vercors.ExceptionalPostcondition =>
            val res = vercors.ExceptionalPostcondition(e.exceptionType, e.exceptionName, Some(wrapWithLet(e.expression.get)), e.srcLocation)
            res.startSpec = clause.startSpec
            res.endSpec = clause.endSpec
            res.copyWhitespace(e)
            res
          case p: vercors.Postcondition =>
            val res = vercors.Postcondition(wrapWithLet(p.expression), p.srcLocation)
            res.startSpec = clause.startSpec
            res.endSpec = clause.endSpec
            res.copyWhitespace(p)
            res
          case p: vercors.Precondition =>
            val res = vercors.Precondition(wrapWithLet(p.expression, false), p.srcLocation)
            res.startSpec = clause.startSpec
            res.endSpec = clause.endSpec
            res.copyWhitespace(p)
            res
          case c: _ =>
            System.err.println(s"cannot add old expression to ${c.getClass.getName}${c.srcLocation}")
            c
        }
      }
      vcClauses.map(addOld)
    }
  }

  def visitContract(node: ir.Contract): vercors.Contract = {
    node match {
      case ir.Contract(clauses, srcLocation, precedingWhitespace) =>
        if(contractHasBehavior(clauses)) {
          val specCases = clauses.filter(x => x.isInstanceOf[ir.SpecCase]).map(x => x.asInstanceOf[ir.SpecCase])
          val vcClauses: ListBuffer[vercors.ContractClause] = ListBuffer()

          for(spec <- specCases) {
            for (clause <- spec.clauses) {
              clause match {
                case a: ir.AssignableClause => vcClauses.appendAll(visitAssignableClauseOfBehavior(a, spec))
                case a: ir.AccessibleClause => vcClauses.appendAll(visitAccessibleClauseOfBehavior(a, spec))
                case _ =>
              }
            }
          }
          if(vcClauses.nonEmpty) {
            vcClauses.head.precedingWhitespace =  System.lineSeparator() + "    // extracted from assignable/accessible clauses:" + System.lineSeparator()
              + vcClauses.head.precedingWhitespace
          }

          val pres = specCases.flatMap(x => x.clauses.filter(x => x.isInstanceOf[ir.Precondition]))
          if(pres.nonEmpty) {
            var condition: vercors.Expression = extractPrecondition(specCases.head.clauses)
            for(i <- 1 until specCases.length) {
              val right = extractPrecondition(specCases(i).clauses)
              condition = vercors.BinaryExpression(condition, vercors.StringNode("||", node.srcLocation, " "), right, condition.srcLocation)
            }
            val pre = vercors.Precondition(condition, node.srcLocation, " ")

//            val first = pres.head.asInstanceOf[ir.Precondition]
//            pre.copyWhitespace(first)
//            pre.startSpec = first.startSpec.map(visitStringNode)
//            pre.endSpec = first.endSpec.map(visitStringNode)
            pre.precedingWhitespace = System.lineSeparator() + "    // Combination of original behavior preconditions:" + System.lineSeparator() + "    "
              + pre.precedingWhitespace
            vcClauses.append(pre)
          }
          // todo: need to copy whitespace of speccases?
//          val startSpec = specCases.head.startSpec.map(visitStringNode)
////          val specCase = new vercors.SpecCase(startSpec,
////                                              vcClauses.toList,
////                                              clauses.last.endSpec.map(visitStringNode),
////                                              node.srcLocation)
//          if(vcClauses.nonEmpty && vcClauses.head.startSpec.isEmpty) {
//            vcClauses.head.startSpec = startSpec
//            vcClauses.last.endSpec = specCases.last.endSpec.map(visitStringNode)
//            vcClauses.last.trailingWhitespace = specCases.last.trailingWhitespace
//          }
//          startSpec match {
//            case Some(s) =>
//              s.precedingWhitespace = s.precedingWhitespace + System.lineSeparator() + "    // Combination of original behavior preconditions:" + System.lineSeparator() + "    "
//              } else {
//
//              }
//            case None => specCase.precedingWhitespace = System.lineSeparator() + "    // Combination of original behavior preconditions:" + System.lineSeparator() + specCase.precedingWhitespace
//          }
          val res = vercors.Contract(clauses.flatMap(x => visitContractClauseAsList(x)),
            srcLocation,
            precedingWhitespace)
          res.startSpec = node.startSpec.map(visitStringNode)
          res.endSpec = node.endSpec.map(visitStringNode)

          for(clause <- vcClauses.reverse) {
            res.addClauseBeforeContract[vercors.VerCorsTreeNode](clause, wrapInSpec, vercors.StringNode.apply)
          }
          res.clauses = addOldVariables(clauses, res.clauses)

          res.copyWhitespace(node)
          res
        } else {
          val res = vercors.Contract(clauses.flatMap(x => visitContractClauseAsList(x, false)),
                                      srcLocation,
                                      precedingWhitespace)
          res.clauses = addOldVariables(clauses, res.clauses)
          res.startSpec = node.startSpec.map(visitStringNode)
          res.endSpec = node.endSpec.map(visitStringNode)
          res.copyWhitespace(node)
          res

    }

    }
  }

  def visitSpecCase(node: ir.SpecCase, asBehavior:Boolean=true): List[vercors.ContractClause] = {
    node match {
      case spec: ir.NormalBehavior => visitNormalBehavior(spec)
      case spec: ir.ExceptionalBehavior => visitExceptionalBehavior(spec)
      case spec: ir.SpecCase =>
        if(asBehavior) {
          visitSpecCaseAsBehavior(spec)
        } else {
          visitSpecCaseSimple(spec)
        }
      case _ => ???
    }
}


  def visitLoopSpecOrLabel(node: ir.LoopSpecOrLabel): vercors.LoopSpecOrLabel = {
    node match {
      case l: ir.Label => visitLabel(l)
      case spec: ir.LoopSpecification => visitLoopSpecification(spec)
    }
  }

  def visitLoopSpecification(node: ir.LoopSpecification): vercors.LoopSpecification = {
    node match {
      case ir.LoopSpecification(startSpec, clauses, endSpec, srcLocation, precedingWhitespace) =>
        val vercorsClauses = clauses.map(visitLoopClause)
        val vercorsStart = startSpec.map(visitStringNode)
        val vercorsEnd = endSpec.map(visitStringNode)
        val res = vercors.LoopSpecification(vercorsStart, vercorsClauses, vercorsEnd, srcLocation, precedingWhitespace)
        res
    }
  }

  def visitContractClauseAsList(node: ir.ContractClause, asBehavior:Boolean=true): List[vercors.ContractClause] = {
    try
      val res = node match {
        case clause: ir.AssignableClause => List(visitAssignableClause(clause))
        case clause: ir.AccessibleClause => List(visitAccessibleClause(clause))
        case clause: ir.ForallVariableDeclaration => List(visitForallVariableDeclaration(clause))
        case clause: ir.OldVariableDeclaration =>
          // dealt with on higher level
          List()
        case clause: ir.Precondition => List(visitPrecondition(clause))
        case clause: ir.Postcondition => List(visitPostcondition(clause))
        case clause: ir.DivergeClause => List(visitDivergeClause(clause))
        case clause: ir.MeasuredByClause => List(visitMeasuredByClause(clause))
        case clause: ir.ExceptionalPostcondition => List(visitExceptionalPostcondition(clause))
        case clause: ir.SignalsOnlyClause => List(visitSignalsOnlyClause(clause))
        case clause: ir.NonTranslatedContractClause => List(visitNonTranslatedContractClause(clause))
        case clause: ir.SpecCase => visitSpecCase(clause, asBehavior)
        case _ => throw new NotSupportedException(node.getClass.getName, node.srcLocation)
      }
      if(res.nonEmpty && !node.isInstanceOf[ir.SpecCase]) {
        if(res.head.startSpec.isEmpty) { res.head.startSpec = node.startSpec.map(visitStringNode) }
        if(res.last.endSpec.isEmpty) { res.last.endSpec = node.endSpec.map(visitStringNode) }
      }
      res
    catch {
      case e: NotSupportedException =>
        val text = Helper.commentUnknownNode(node, s"${e.construct} not supported, whole ContractClause commented out")
        List(vercors.NonTranslatedContractClause(text, node.srcLocation))
    }
  }

  def visitAssignableClauseOfBehavior(node: ir.AssignableClause, specCase: ir.SpecCase): List[vercors.ContractClause] = {
    val ass = visitAssignableClause(node)
    if(ass.isInstanceOf[vercors.NonTranslatedContractClause]) {
      return List(ass)
    }
    val ctx = ass.asInstanceOf[vercors.Context]
    val pre = extractPrecondition(specCase.clauses)
    val old = vercors.OldExpression(pre, None, pre.srcLocation, pre.precedingWhitespace)
    pre.precedingWhitespace = ""
    val newPre = vercors.Precondition(vercors.BinaryExpression(pre, vercors.StringNode("==>", node.srcLocation, " "),
                                                                ctx.expression, node.srcLocation),
                                      ctx.srcLocation)
    newPre.copyWhitespace(ctx)
    newPre.requiresKeyword.copyWhitespace(ctx.contextKeyword)
    newPre.semicolon.copyWhitespace(ctx.semicolon)
    newPre.startSpec = ctx.startSpec
    newPre.endSpec = ctx.endSpec
    val newPost = vercors.Postcondition(vercors.BinaryExpression(old, vercors.StringNode("==>", node.srcLocation, " "),
      ctx.expression, node.srcLocation),
      ctx.srcLocation)
    newPost.copyWhitespace(ctx)
    newPost.ensuresKeyword.copyWhitespace(ctx.contextKeyword)
    newPost.semicolon.copyWhitespace(ctx.semicolon)
    newPost.startSpec = ctx.startSpec
    newPost.endSpec = ctx.endSpec
    List(newPre, newPost)
  }
  def visitAssignableClause(node: ir.AssignableClause): vercors.ContractClause = {
    node match {
      case ir.AssignableClause(expressions, srcLocation, precedingWhitespace) =>
        def find(expr: ir.Expression, str: String): Boolean = {
          expr match {
            case s: ir.StringNode => s.text.contains(str)
            case _ => false
          }
        }
        if(expressions.exists(x => find(x, "\\nothing"))) {
          val text = Helper.commentNode(node, node.assignableKeyword, "\\nothing is not directly supported, but is the default in VerCors. commented out.")
          return vercors.NonTranslatedContractClause(text, srcLocation, precedingWhitespace)
        }
        if(expressions.exists(x => find(x, "\\everything"))) {
          val text = Helper.commentNode(node, node.assignableKeyword, "\\everything is not supported in VerCors. commented out.")
          return vercors.NonTranslatedContractClause(text, srcLocation, precedingWhitespace)
        }
        if(expressions.exists(x => find(x, "\\not_specified"))) {
          val text = Helper.commentNode(node, node.assignableKeyword, "\\not_specified is not supported in VerCors. commented out.")
          return vercors.NonTranslatedContractClause(text, srcLocation, precedingWhitespace)
        }
        val perms:ListBuffer[vercors.Expression] = ListBuffer()
        for(expr <- expressions) {
          transformAccessReference(expr, perms, "write", srcLocation)
        }
        var combined: vercors.Expression = perms.head
        for(i <- 1 until perms.length) {
          combined = vercors.BinaryExpression(combined, vercors.StringNode("**", srcLocation, " "), perms(i), node.srcLocation)
        }
        val res = vercors.Context(combined, node.srcLocation, precedingWhitespace)
        res.contextKeyword.precedingWhitespace = node.assignableKeyword.precedingWhitespace
        res.semicolon.precedingWhitespace = node.semicolon.precedingWhitespace
        res.trailingWhitespace = node.trailingWhitespace
        res.startSpec = node.startSpec.map(visitStringNode)
        res.endSpec = node.endSpec.map(visitStringNode)
        res
    }
  }

  private def transformAccessReference(expr: ir.Expression, perms: ListBuffer[vercors.Expression], amount: String, srcLocation: SourceLocation) = {
    val guards: ListBuffer[vercors.Expression] = ListBuffer()
    val idents: ListBuffer[vercors.QuantifiedVarDeclaration] = ListBuffer()

    def makeForall(inner: vercors.Expression): Unit = {
      val perm = vercors.PermissionPredicate(inner, Some(vercors.StringNode(amount, expr.srcLocation, " ")), expr.srcLocation, " ")
      var guard: vercors.Expression = guards.head
      for (g <- guards.tail) {
        guard = vercors.BinaryExpression(guard, vercors.StringNode("&&", srcLocation, " "), g, srcLocation)
      }
      val forall = vercors.ForallStarExpression(idents.toList, Some(guard), perm, srcLocation)
      perms.append(forall)
    }

    def desugarArrays(expr: ir.Expression): vercors.Expression = {
      expr match {
        case ir.ArrayIndexExpression(arrayExpression, indexExpression, srcLocation, precedingWhitespace) =>
          def makeRange(arrExpr: vercors.Expression, lower: vercors.Expression, upper: vercors.Expression): vercors.Expression = {
            val idx = vercors.StringNode("i" + idents.length, srcLocation, " ")
            val intType = new vercors.StringTypeIndication("int", Nil, "int", srcLocation, " ")
            val quant = new vercors.QuantifiedVarDeclaration(intType, List(idx), srcLocation)
            idents.append(quant)
            val lowerG = vercors.BinaryExpression(lower,
              vercors.StringNode("<=", srcLocation, " "),
              idx,
              srcLocation)
            val upperG = vercors.BinaryExpression(idx,
              vercors.StringNode("<", srcLocation, " "),
              upper,
              srcLocation)
            val guard = vercors.BinaryExpression(lowerG,
              vercors.StringNode("&&", srcLocation, " "),
              upperG,
              srcLocation)
            guards.append(guard)
            val res = vercors.ArrayIndexExpression(arrExpr, idx, srcLocation)
            res.copyWhitespace(expr)
            makeForall(res)
            res
          }

          indexExpression match {
            case r: ir.RangeExpression =>
              val arrExpr = desugarArrays(arrayExpression)
              makeRange(arrExpr, visitExpression(r.start), visitExpression(r.end))
            case all: ir.IndexAllExpression =>
              val arrExpr = desugarArrays(arrayExpression)
              val upper = vercors.NestedInvocationExpression(arrExpr,
                vercors.StringNode("length", all.srcLocation),
                all.srcLocation)
              makeRange(arrExpr, vercors.StringNode("0", all.srcLocation), upper)
            case e: ir.Expression =>
              val res = vercors.ArrayIndexExpression(desugarArrays(arrayExpression), visitExpression(e), srcLocation)
              res.copyWhitespace(expr)
              res
          }
        case ir.NestedInvocationExpression(left, right, srcLocation, precedingWhitespace) =>
          if (right.isInstanceOf[ir.IndexAllExpression]) {
            throw new NotSupportedException("indexing all fields with '*'", srcLocation)
          }
          val res = vercors.NestedInvocationExpression(desugarArrays(left), visitExpression(right), srcLocation)
          res.copyWhitespace(expr)
          if (idents.nonEmpty) {
            makeForall(res)
          }
          res
        case e: ir.Expression => visitExpression(e)
      }
    }

    val perm = vercors.PermissionPredicate(desugarArrays(expr), Some(vercors.StringNode(amount, expr.srcLocation, " ")), expr.srcLocation, " ")
    if (idents.isEmpty) {
      perms.append(perm)
    }
  }

  def visitLoopModifiesClause(node: ir.LoopModifiesClause): vercors.ContractClause = {
    node match {
      case ir.LoopModifiesClause(expressions, srcLocation, precedingWhitespace) =>
        def find(expr: ir.Expression, str: String): Boolean = {
          expr match {
            case s: ir.StringNode => s.text.contains(str)
            case _ => false
          }
        }
        if(expressions.exists(x => find(x, "\\nothing"))) {
          val text = Helper.commentNode(node, node.loopModifiesKeyword, "\\nothing is not directly supported, but is the default in VerCors. commented out.")
          return vercors.NonTranslatedContractClause(text, srcLocation, precedingWhitespace)
        }
        if(expressions.exists(x => find(x, "\\everything"))) {
          val text = Helper.commentNode(node, node.loopModifiesKeyword, "\\everything is not supported in VerCors. commented out.")
          return vercors.NonTranslatedContractClause(text, srcLocation, precedingWhitespace)
        }
        if(expressions.exists(x => find(x, "\\not_specified"))) {
          val text = Helper.commentNode(node, node.loopModifiesKeyword, "\\not_specified is not supported in VerCors. commented out.")
          return vercors.NonTranslatedContractClause(text, srcLocation, precedingWhitespace)
        }
        val perms:ListBuffer[vercors.Expression] = ListBuffer()
        for(expr <- expressions) {
          transformAccessReference(expr, perms, "write", srcLocation)
        }
        var combined: vercors.Expression = perms.head
        for(i <- 1 until perms.length) {
          combined = vercors.BinaryExpression(combined, vercors.StringNode("**", srcLocation, " "), perms(i), node.srcLocation)
        }
        val res = vercors.LoopInvariant(combined, node.srcLocation, precedingWhitespace)
        res.invariantKeyword.precedingWhitespace = node.loopModifiesKeyword.precedingWhitespace
        res.semicolon.precedingWhitespace = node.semicolon.precedingWhitespace
        res.trailingWhitespace = node.trailingWhitespace
        res
    }
  }

  def visitAccessibleClauseOfBehavior(node: ir.AccessibleClause, specCase: ir.SpecCase): List[vercors.ContractClause] = {
    val clause = visitAccessibleClause(node)
    clause match {
      case _: vercors.NonTranslatedContractClause =>
        List(clause)
      case ctx: vercors.Context =>
        val pre = extractPrecondition(specCase.clauses)
        val old = vercors.OldExpression(pre, None, pre.srcLocation, pre.precedingWhitespace)
        pre.precedingWhitespace = ""
        val newPre = vercors.Precondition(vercors.BinaryExpression(pre, vercors.StringNode("==>", node.srcLocation, " "),
                                                                    ctx.expression, node.srcLocation),
                                          ctx.srcLocation)
        newPre.copyWhitespace(ctx)
        newPre.requiresKeyword.copyWhitespace(ctx.contextKeyword)
        newPre.semicolon.copyWhitespace(ctx.semicolon)
        newPre.startSpec = ctx.startSpec
        newPre.endSpec = ctx.endSpec
        val newPost = vercors.Postcondition(vercors.BinaryExpression(old, vercors.StringNode("==>", node.srcLocation, " "),
                                                                    ctx.expression, node.srcLocation),
                                            ctx.srcLocation)
        newPost.copyWhitespace(ctx)
        newPost.ensuresKeyword.copyWhitespace(ctx.contextKeyword)
        newPost.semicolon.copyWhitespace(ctx.semicolon)
        newPost.startSpec = ctx.startSpec
        newPost.endSpec = ctx.endSpec
        List(newPre, newPost)
    }
  }
  
  def visitAccessibleClause(node: ir.AccessibleClause): vercors.ContractClause = {
    node match {
      case ir.AccessibleClause(expressions, srcLocation, precedingWhitespace) =>
        def find(expr: ir.Expression, str: String): Boolean = {
          expr match {
            case s: ir.StringNode => s.text.contains(str)
            case _ => false
          }
        }
        if(expressions.exists(x => find(x, "\\nothing"))) {
          val text = Helper.commentNode(node, node.accessibleKeyword, "\\nothing is not directly supported, but is the default in VerCors. commented out.")
          return vercors.NonTranslatedContractClause(text, srcLocation, precedingWhitespace)
        }
        if(expressions.exists(x => find(x, "\\everything"))) {
          val text = Helper.commentNode(node, node.accessibleKeyword, "\\everything is not supported in VerCors. commented out.")
          return vercors.NonTranslatedContractClause(text, srcLocation, precedingWhitespace)
        }
        if(expressions.exists(x => find(x, "\\not_specified"))) {
          val text = Helper.commentNode(node, node.accessibleKeyword, "\\not_specified is not supported in VerCors. commented out.")
          return vercors.NonTranslatedContractClause(text, srcLocation, precedingWhitespace)
        }
        val perms:ListBuffer[vercors.Expression] = ListBuffer()
        for(expr <- expressions) {
          transformAccessReference(expr, perms, "read", srcLocation)
        }
        var combined: vercors.Expression = perms.head
        for(i <- 1 until perms.length) {
          combined = vercors.BinaryExpression(combined, vercors.StringNode("**", srcLocation, " "), perms(i), node.srcLocation)
        }
        val res = vercors.Context(combined, node.srcLocation, precedingWhitespace)
        res.contextKeyword.precedingWhitespace = node.accessibleKeyword.precedingWhitespace
        res.semicolon.precedingWhitespace = node.semicolon.precedingWhitespace
        res.trailingWhitespace = node.trailingWhitespace
        res.startSpec = node.startSpec.map(visitStringNode)
        res.endSpec = node.endSpec.map(visitStringNode)
        res
    }
  }

  def visitForallVariableDeclaration(node: ir.ForallVariableDeclaration): vercors.NonTranslatedContractClause = {
    val text = Helper.commentNode(node, node.forallKeyword, "Forall variable declaration not supported in VerCors. commented out")
    vercors.NonTranslatedContractClause(text, node.srcLocation, node.precedingWhitespace)
  }

  def visitNonTranslatedContractClause(node: ir.NonTranslatedContractClause): vercors.NonTranslatedContractClause = {
    node match {
      case ir.NonTranslatedContractClause(text, srcLocation, precedingWhitespace) =>
        val res = vercors.NonTranslatedContractClause(text,
          srcLocation,
          precedingWhitespace)
        res.copyWhitespace(node)
        res.startSpec = node.startSpec.map(visitStringNode)
        res.endSpec = node.endSpec.map(visitStringNode)
        res
    }
  }

  def visitPrecondition(node: ir.Precondition): vercors.Precondition = {
    node match {
      case ir.Precondition(expression, srcLocation, precedingWhitespace) =>
        val res = vercors.Precondition(visitExpression(expression),
                                  srcLocation,
                                  precedingWhitespace)
        res.copyWhitespace(node)
        res.startSpec = node.startSpec.map(visitStringNode)
        res.endSpec = node.endSpec.map(visitStringNode)
        res
    }
  }

  def visitPostcondition(node: ir.Postcondition): vercors.Postcondition = {
    node match {
      case ir.Postcondition(expression, srcLocation, precedingWhitespace) =>
        val res = vercors.Postcondition(visitExpression(expression),
                                    srcLocation,
                                    precedingWhitespace)
        res.copyWhitespace(node)
        res.startSpec = node.startSpec.map(visitStringNode)
        res.endSpec = node.endSpec.map(visitStringNode)
        res
    }
  }

  def visitDivergeClause(node: ir.DivergeClause): vercors.NonTranslatedContractClause = {
    val text = Helper.commentNode(node, node.divergesKeyword, "Diverge clause not supported in VerCors. commented out")
    vercors.NonTranslatedContractClause(text, node.srcLocation, node.precedingWhitespace)
  }

  def visitLoopClause(node: ir.LoopClause): vercors.ContractClause = {
    try
      node match {
        case clause: ir.LoopModifiesClause => visitLoopModifiesClause(clause)
        case clause: ir.LoopInvariant => visitLoopInvariant(clause)
        case clause: ir.DecreasesClause => visitDecreasesClause(clause)
        case clause: ir.NonTranslatedLoopClause => visitNonTranslatedLoopClause(clause)
        case _ => ???
      }
    catch {
      case e: NotSupportedException =>
        val e_txt = s"${e.construct} not supported, whole Loop Clause commented out. "
        val text = Helper.commentUnknownNode(node, e_txt)
        System.err.println(e.getMessage)
        vercors.NonTranslatedContractClause(text, node.srcLocation)
    }
  }

  def visitLoopInvariant(node: ir.LoopInvariant): vercors.LoopInvariant = {
    node match {
      case ir.LoopInvariant(expression, srcLocation, precedingWhitespace) =>
        val res = vercors.LoopInvariant(visitExpression(expression),
                                    srcLocation,
                                    precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitDecreasesClause(node: ir.DecreasesClause): vercors.DecreasesClause = {
    node match {
      case ir.DecreasesClause(expression, srcLocation, precedingWhitespace) =>
        val res = vercors.DecreasesClause(visitExpression(expression),
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitNonTranslatedLoopClause(node: ir.NonTranslatedLoopClause): vercors.NonTranslatedContractClause = {
    node match {
      case ir.NonTranslatedLoopClause(text, srcLocation, precedingWhitespace) =>
        val res = vercors.NonTranslatedContractClause(text,
          srcLocation,
          precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitMeasuredByClause(node: ir.MeasuredByClause): vercors.DecreasesClause = {
    node match {
      case ir.MeasuredByClause(expression, predicate, srcLocation, precedingWhitespace) =>
        val res = vercors.DecreasesClause(visitExpression(expression), srcLocation)
        res.copyWhitespace(node)
        res.decreasesKeyword.precedingWhitespace = node.measuredByKeyword.precedingWhitespace
        res.semicolon.precedingWhitespace = node.semicolon.precedingWhitespace
        if(predicate.isDefined) {
          System.err.println(s"VerCors does not support conditional termination measure${node.predicate.get.srcLocation}. Condition commented out.")
          res.semicolon.precedingWhitespace = " // " + Helper.commentUnknownNode(node.ifKeyword, "")
                              + Helper.commentUnknownNode(predicate.get, "") + res.semicolon.precedingWhitespace
        }
        res
    }
  }

  def visitExceptionalPostcondition(node: ir.ExceptionalPostcondition): vercors.ExceptionalPostcondition = {
    node match {
      case ir.ExceptionalPostcondition(exceptionType, exceptionName, expression, srcLocation, precedingWhitespace) =>
        val name: Some[vercors.StringNode] = exceptionName match {
          case Some(value) => Some(visitStringNode(value))
          case None => Some(vercors.StringNode("e", srcLocation, " "))
        }
        val expr = expression match {
          case Some(value) => Some(visitExpression(value))
          case None => Some(vercors.StringNode("true", srcLocation, " "))
        }
        val res = vercors.ExceptionalPostcondition(visitTypeIndication(exceptionType),
                                              name,
                                              expr,
                                              srcLocation,
                                              precedingWhitespace)
        res.copyWhitespace(node)
        res.startSpec = node.startSpec.map(visitStringNode)
        res.endSpec = node.endSpec.map(visitStringNode)
        res
    }
  }

  def visitSignalsOnlyClause(node: ir.SignalsOnlyClause): vercors.ContractClause = {
    node match {
      case ir.SignalsOnlyClause(exceptionTypes, srcLocation, precedingWhitespace) =>
        val name = vercors.StringNode("e", srcLocation, " ")
        val exceptionNames = exceptionTypes.map(x => x.typeAsString)
        while(exceptionNames.contains(name.text)) {
          name.text = "_" + name.text
        }
        if(exceptionTypes.nonEmpty) {
          val instance = vercors.InstanceofExpression(name.copy(), visitTypeIndication(exceptionTypes.head), srcLocation)
          instance.instanceofKeyword.precedingWhitespace = " "
          var predicate: vercors.Expression = instance
          for (e <- exceptionTypes.tail) {
            val instance = vercors.InstanceofExpression(name.copy(), visitTypeIndication(e), srcLocation)
            instance.instanceofKeyword.precedingWhitespace = " "
            predicate = vercors.BinaryExpression(predicate, vercors.StringNode("||", srcLocation, " "), instance, srcLocation)
          }
          val res = vercors.ExceptionalPostcondition(vercors.StringTypeIndication("java.lang.Exception",
                                                                                  Nil,
                                                                                  "java.lang.Exception",
                                                                                  srcLocation),
                                                    Some(name), Some(predicate), srcLocation)
          res.copyWhitespace(node)
          res.semicolon.copyWhitespace(node.semicolon)
          res.signalsKeyword.copyWhitespace(node.signalsOnlyKeyword)
          res.startSpec = node.startSpec.map(visitStringNode)
          res.endSpec = node.endSpec.map(visitStringNode)
          res
        } else {
          val text = Helper.commentNode(node, node.signalsOnlyKeyword, "signals_only not supported here. Commented.")
          val res = vercors.NonTranslatedContractClause(text, srcLocation)
          res.copyWhitespace(node)
          res
        }
    }
  }

  def extractPrecondition(clauses: List[ir.ContractClause]): vercors.Expression = {
    val pres: List[ir.Precondition] = clauses.filter(x => x.isInstanceOf[ir.Precondition]).map(x => x.asInstanceOf[ir.Precondition])
    val reqExprs: List[vercors.Expression] = pres.map(x => visitExpression(x.expression))

    if(reqExprs.isEmpty) {
      vercors.StringNode("true", null, " ")
    } else {
      var condition: vercors.Expression = reqExprs.head
      for (i <- 1 until reqExprs.length) {
        val parL = vercors.ParExpression(condition, condition.srcLocation)
        val parR = vercors.ParExpression(reqExprs(i), reqExprs(i).srcLocation, " ")
        condition = vercors.BinaryExpression(parL, vercors.StringNode("&&", parL.srcLocation, " "), parR, parL.srcLocation)
      }
      vercors.ParExpression(condition, condition.srcLocation, " ")
    }
  }

  def visitSpecCaseSimple(node: ir.SpecCase): List[vercors.ContractClause] = {
    node match {
      case ir.SpecCase(startSpec, hasAlso, clauses, endSpec, srcLocation, precedingWhitespace) =>
        var vcClauses = clauses.flatMap(x => visitContractClauseAsList(x, false))
        vcClauses = addOldVariables(clauses, vcClauses)
        if(vcClauses.isEmpty) {
          vcClauses = List(vercors.NonTranslatedContractClause("", srcLocation))
        }
        extractSpecFromSpecCase(node, vcClauses, false)
    }
  }

  private def extractSpecFromSpecCase(node: ir.SpecCase, vcClauses: List[vercors.ContractClause], asBehavior: Boolean): List[vercors.ContractClause] = {
    val startSpec = node.startSpec
    val endSpec = node.endSpec
    var clauses = vcClauses

    def behaviorWS(): String = {
      var extraWS = ""
      if (node.hasAlso) {
        extraWS += node.alsoKeyword.precedingWhitespace + node.alsoKeyword.trailingWhitespace
      }
      node match {
        case b: ir.Behavior =>
          b.privacy match {
            case Some(p) =>
              System.err.println(s"privacy modifier${p.srcLocation} cannot be translated to VerCors and is dropped")
              extraWS += p.precedingWhitespace + p.trailingWhitespace
            case None =>
          }
          extraWS += b.behaviorKeyword.precedingWhitespace + s"// from ${b.behaviorKeyword.text}:" + b.behaviorKeyword.trailingWhitespace
        case _ =>
          if(asBehavior) {
            extraWS += System.lineSeparator() + "    // from behavior-like specification case:" + System.lineSeparator()
          }
      }
      extraWS
    }


    if (startSpec.isDefined) {
      if (clauses.head.startSpec.isEmpty) {
        val vcSpec = startSpec.map(visitStringNode)
        val non = vercors.NonTranslatedContractClause(behaviorWS(), node.srcLocation, node.precedingWhitespace)
        non.startSpec = vcSpec
        clauses = clauses.prepended(non)
     } else {
        val extraWS = startSpec.get.precedingWhitespace + startSpec.get.trailingWhitespace
        val non = vercors.NonTranslatedContractClause(extraWS + behaviorWS(), node.srcLocation, node.precedingWhitespace)
        clauses = clauses.prepended(non)
      }
    } else {
      clauses.head.precedingWhitespace = node.precedingWhitespace + behaviorWS() + clauses.head.precedingWhitespace
    }
    if (endSpec.isDefined) {
      if (clauses.last.endSpec.isEmpty) {
        clauses.last.endSpec = endSpec.map(visitStringNode)
      } else {
        clauses.last.trailingWhitespace += endSpec.get.precedingWhitespace + endSpec.get.trailingWhitespace
      }
    }
    clauses.last.trailingWhitespace += node.trailingWhitespace
    clauses
  }

  def visitSpecCaseAsBehavior(node: ir.SpecCase): List[vercors.ContractClause] = {
    node match {
      case ir.SpecCase(startSpec, hasAlso, clauses, endSpec, srcLocation, precedingWhitespace) =>
        val extracted = extractPrecondition(clauses)
        val condition = vercors.OldExpression(extracted, None, extracted.srcLocation, extracted.precedingWhitespace)
        extracted.precedingWhitespace = ""

        val vcClauses: ListBuffer[vercors.ContractClause] = ListBuffer()
        for (clause <- clauses) {
          try
            clause match {
              case c: ir.Precondition =>
                var text = ""
                c.startSpec match {
                  case Some(value) => text += value.precedingWhitespace + value.trailingWhitespace
                  case None =>
                }
                val non = vercors.NonTranslatedContractClause(text + c.requiresKeyword.precedingWhitespace, c.srcLocation)
                c.endSpec match {
                  case Some(value) => non.text += value.precedingWhitespace + value.trailingWhitespace
                  case None =>
                }
                non.copyWhitespace(c)
                if(!"""\s*""".r.matches(non.toString)){
                  vcClauses.append(non)
                }
              case c: ir.AccessibleClause =>
                var text = ""
                c.startSpec match {
                  case Some(value) => text += value.precedingWhitespace + value.trailingWhitespace
                  case None =>
                }
                val non = vercors.NonTranslatedContractClause(text + c.accessibleKeyword.precedingWhitespace, c.srcLocation)
                c.endSpec match {
                  case Some(value) => non.text += value.precedingWhitespace + value.trailingWhitespace
                  case None =>
                }
                non.copyWhitespace(c)
                if(!"""\s*""".r.matches(non.toString)){
                  vcClauses.append(non)
                }
              case c: ir.AssignableClause =>
                var text = ""
                c.startSpec match {
                  case Some(value) => text += value.precedingWhitespace + value.trailingWhitespace
                  case None =>
                }
                val non = vercors.NonTranslatedContractClause(text + c.assignableKeyword.precedingWhitespace, c.srcLocation)
                c.endSpec match {
                  case Some(value) => non.text += value.precedingWhitespace + value.trailingWhitespace
                  case None =>
                }
                non.copyWhitespace(c)
                if(!"""\s*""".r.matches(non.toString)){
                  vcClauses.append(non)
                }
              case p: ir.Postcondition =>
                val expr = visitExpression(p.expression)
                val consequence = vercors.ParExpression(expr, expr.srcLocation, " ")
                val expression = vercors.BinaryExpression(condition, vercors.StringNode("==>", srcLocation, " "), consequence, expr.srcLocation)
                val newPost = vercors.Postcondition(expression, expr.srcLocation)
                newPost.copyWhitespace(p)
                newPost.startSpec = p.startSpec.map(visitStringNode)
                newPost.endSpec = p.endSpec.map(visitStringNode)
                vcClauses.append(newPost)
              case p: ir.ExceptionalPostcondition =>
                val vcPost = visitExceptionalPostcondition(p)
                if (p.expression.isDefined) {
                  val expr = vcPost.expression.get
                  val consequence = vercors.ParExpression(expr, expr.srcLocation, " ")
                  val expression = vercors.BinaryExpression(condition, vercors.StringNode("==>", srcLocation, " "), consequence, expr.srcLocation)
                  val newPost = vercors.ExceptionalPostcondition(vcPost.exceptionType, vcPost.exceptionName, Some(expression), vcPost.srcLocation)
                  newPost.copyWhitespace(vcPost)
                  newPost.startSpec = vcPost.startSpec
                  newPost.endSpec = vcPost.endSpec
                  vcClauses.append(newPost)
                } else {
                  vcClauses.append(vcPost)
                }
              case _ => addOldVariables(clauses, vcClauses.appendAll(visitContractClauseAsList(clause)).toList)
            }
          catch {
            case e: NotSupportedException =>
              val e_txt = s"${e.construct} not supported, whole Contract Clause commented out. "
              val text = Helper.commentUnknownNode(node, e_txt)
              System.err.println(e.getMessage)
              vcClauses.append(vercors.NonTranslatedContractClause(text, node.srcLocation))
          }
        }

        if(vcClauses.isEmpty) {
          vcClauses.append(vercors.NonTranslatedContractClause("", node.srcLocation))
        }
        extractSpecFromSpecCase(node, vcClauses.toList, true)
    }
  }
  def visitNormalBehavior(node: ir.NormalBehavior): List[vercors.ContractClause] = {
    val clauses = visitSpecCaseAsBehavior(node)
    val exType = new vercors.StringTypeIndication("java.lang.Exception", Nil, "java.lang.Exception", node.srcLocation)
    val condition = extractPrecondition(node.clauses)
    val old = vercors.OldExpression(condition, None, condition.srcLocation, condition.precedingWhitespace)
    condition.precedingWhitespace = ""
    val consequence = vercors.StringNode("false", node.srcLocation, " ")
    val expression = vercors.BinaryExpression(old, vercors.StringNode("==>", node.srcLocation, " "), consequence, node.srcLocation)
    val signals = vercors.ExceptionalPostcondition(exType,
                                                  Some(vercors.StringNode("e", node.srcLocation, " ")),
                                                  Some(expression),
                                                  node.srcLocation)
    signals.signalsKeyword.precedingWhitespace = " "
    if(clauses.isEmpty) {
      signals.startSpec = Some(vercors.StringNode("//@", node.srcLocation, System.lineSeparator() + "      "))
    } else {
      val last = clauses.last
      if(last.endSpec.isDefined) {
        if(last.startSpec.isDefined) {
          signals.startSpec = Some(last.startSpec.get.copy())
          signals.endSpec = Some(last.endSpec.get.copy())
        } else {
          signals.endSpec = last.endSpec
          last.endSpec = None
          signals.precedingWhitespace = System.lineSeparator() + "      "
        }
      } else {
        signals.precedingWhitespace = System.lineSeparator() + "      "
      }
    }
    clauses.appended(signals)
  }
  def visitExceptionalBehavior(node: ir.ExceptionalBehavior): List[vercors.ContractClause] = {
    val clauses = visitSpecCaseAsBehavior(node)
    val condition = extractPrecondition(node.clauses)
    val old = vercors.OldExpression(condition, None, condition.srcLocation, condition.precedingWhitespace)
    condition.precedingWhitespace = ""
    val consequence = vercors.StringNode("false", node.srcLocation, " ")
    val expression = vercors.BinaryExpression(old, vercors.StringNode("==>", node.srcLocation, " "), consequence, node.srcLocation)
    val ensures = vercors.Postcondition(expression, node.srcLocation)
    ensures.ensuresKeyword.precedingWhitespace = " "
    if(clauses.isEmpty) {
      ensures.startSpec = Some(vercors.StringNode("//@", node.srcLocation, System.lineSeparator() + "      "))
    } else {
      val last = clauses.last
      ensures.trailingWhitespace = last.trailingWhitespace
      last.trailingWhitespace = ""
      if(last.endSpec.isDefined) {
        if(last.startSpec.isDefined) {
          ensures.startSpec = Some(last.startSpec.get.copy())
          ensures.endSpec = Some(last.endSpec.get.copy())
        } else {
          ensures.endSpec = last.endSpec
          last.endSpec = None
          ensures.precedingWhitespace = System.lineSeparator() + "      "
        }
      } else {
        ensures.precedingWhitespace = System.lineSeparator() + "      "
      }
    }
    clauses.appended(ensures)
  }

  def visitNonTranslatedStatement(node: ir.NonTranslatedStatement): vercors.NonTranslatedStatement = {
    node match {
      case ir.NonTranslatedStatement(text, srcLocation, precedingWhitespace) =>
        val res = vercors.NonTranslatedStatement(text, srcLocation)
        res.copyWhitespace(node)
        res
    }
  }

  def visitAssertStatement(node: ir.AssertStatement): vercors.AssertStatement = {
    node match {
      case ir.AssertStatement(content, message, srcLocation, precedingWhitespace) =>
        val irMsg = message.map(visitExpression)
        val res = vercors.AssertStatement(visitExpression(content),
                                      irMsg,
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitAssumeStatement(node: ir.AssumeStatement): vercors.AssumeStatement = {
    node match {
      case ir.AssumeStatement(content, message, srcLocation, precedingWhitespace) =>
        val irMsg = message.map(visitExpression)
        val res = vercors.AssumeStatement(visitExpression(content),
                                      irMsg,
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitUnreachableStatement(node: ir.UnreachableStatement): vercors.NonTranslatedStatement = {
    node match {
      case ir.UnreachableStatement(srcLocation, precedingWhitespace) =>
        val text = Helper.commentNode(node, node.unreachableKeyword, "Unreachable statement not supported in VerCors")
        val res = vercors.NonTranslatedStatement(text, srcLocation, precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitSetStatement(node: ir.SetStatement): vercors.GhostStatement = {
    node match {
      case ir.SetStatement(assignment, srcLocation, precedingWhitespace) =>
        val expr = vercors.ExpressionStatement(visitExpression(assignment), srcLocation)
        expr.semicolon.copyWhitespace(node.semicolon)
        val res = vercors.GhostStatement(expr,
                                  srcLocation,
                                  precedingWhitespace)
        res.copyWhitespace(node)
        res.ghostKeyword.copyWhitespace(node.setKeyword)
        res
    }
  }
  def visitGhostStatement(node: ir.GhostStatement): vercors.GhostStatement = {
    node match {
      case ir.GhostStatement(statement, srcLocation, precedingWhitespace) =>
        val res = vercors.GhostStatement(visitStatement(statement), srcLocation)
        res.copyWhitespace(node)
        res
    }
  }

  def visitJmlStatementBlock(node: ir.JmlStatementBlock): vercors.JmlStatementBlock = {
    node match {
      case ir.JmlStatementBlock(startSpec, statements, endSpec, srcLocation, precedingWhitespace) =>
        val irStart = startSpec.map(visitStringNode)
        val irEnd = endSpec.map(visitStringNode)
        val oldWrapInSpec = wrapInSpec
        wrapInSpec = false
        val res = vercors.JmlStatementBlock(irStart,
                                        statements.map(visitStatement),
                                        irEnd,
                                        srcLocation,
                                        precedingWhitespace)
        res.copyWhitespace(node)
        wrapInSpec = oldWrapInSpec
        res
    }
  }

  def visitResultExpression(node: ir.ResultExpression): vercors.ResultExpression = {
    node match {
      case ir.ResultExpression(src, precedingWhitespace) =>
        val res = vercors.ResultExpression(src, precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitOldExpression(node: ir.OldExpression): vercors.OldExpression = {
    node match {
      case ir.OldExpression(expression, label, srcLocation, precedingWhitespace) =>
        val irLabel = label.map(visitStringNode)
        val res = vercors.OldExpression(visitExpression(expression),
                                    irLabel,
                                    srcLocation,
                                    precedingWhitespace)
        res.copyWhitespace(node)
        res.closeParenthesis.precedingWhitespace = node.comma.precedingWhitespace + res.closeParenthesis.precedingWhitespace
        if(node.label.isDefined) {
          res.label.get.precedingWhitespace = node.comma.trailingWhitespace + res.label.get.precedingWhitespace
        }
        res.comma.precedingWhitespace = ""
        res.comma.trailingWhitespace = ""
        res
    }
  }

  def visitPreExpression(node: ir.PreExpression): vercors.OldExpression = {
    node match {
      case ir.PreExpression(expression, srcLocation, precedingWhitespace) =>
        val res = vercors.OldExpression(visitExpression(expression),
                                    None,
                                    srcLocation,
                                    precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitNotAssignedExpression(node: ir.NotAssignedExpression): vercors.Expression = {
    throw new NotSupportedException("\\not_assigned expression", node.srcLocation)
  }

  def visitNotModifiedExpression(node: ir.NotModifiedExpression): vercors.Expression = {
    def makeOld(s: ir.Expression): vercors.Expression = {
      val old = vercors.OldExpression(visitExpression(s), None, s.srcLocation, " ")
      val eq = vercors.StringNode("==", node.srcLocation, " ")
      vercors.BinaryExpression(visitExpression(s), eq, old, node.srcLocation)
    }
    def combine(e1: vercors.Expression, e2: vercors.Expression) = {
      val and = vercors.StringNode("&&", node.srcLocation, " ")
      vercors.BinaryExpression(e1, and, e2, node.srcLocation)
    }
    if(node.storeRefs.isEmpty) {
      vercors.StringNode("true", node.srcLocation, " ")
    } else {
      var res = makeOld(node.storeRefs.head)
      for(s <- node.storeRefs.tail) {
        res = combine(res, makeOld(s))
      }
      res
    }
  }

  def visitOnlyAccessedExpression(node: ir.OnlyAccessedExpression): vercors.Expression = {
    throw new NotSupportedException("\\only_accessed expression", node.srcLocation)
  }

  def visitOnlyAssignedExpression(node: ir.OnlyAssignedExpression): vercors.Expression = {
    throw new NotSupportedException("\\only_assigned expression", node.srcLocation)
  }

  def visitFreshExpression(node: ir.FreshExpression): vercors.Expression = {
    throw new NotSupportedException("\\fresh expression", node.srcLocation)
  }

  def visitNonNullElementsExpression(node: ir.NonNullElementsExpression): vercors.Expression = {
    // turn into: myArray != null && (\forall int i; 0 <= i && i < myArray.length; myArray[i] != null)
    val neq = vercors.BinaryExpression(visitExpression(node.expression),
      vercors.StringNode("!=", node.srcLocation, " "),
      vercors.StringNode("null", node.srcLocation, " "),
      node.srcLocation)
    val i = vercors.StringNode("_i", node.srcLocation, " ")
    val vdecl = new vercors.QuantifiedVarDeclaration(
      new vercors.StringTypeIndication("int", Nil,"int", node.srcLocation, " "),
      List(i.copy()),
      node.srcLocation)
    val lower = vercors.BinaryExpression(vercors.StringNode("0", node.srcLocation, " "),
                                        vercors.StringNode("<=", node.srcLocation, " "),
                                        i.copy(),
                                        node.srcLocation)
    val upper = vercors.BinaryExpression(i.copy(),
                                        vercors.StringNode("<", node.srcLocation, " "),
      vercors.NestedInvocationExpression(visitExpression(node.expression), vercors.StringNode("length", node.srcLocation), node.srcLocation, " "),
      node.srcLocation)
    val guard = vercors.BinaryExpression(lower, vercors.StringNode("&&", node.srcLocation), upper, node.srcLocation)
    val idx = vercors.ArrayIndexExpression(visitExpression(node.expression), i.copy(), node.srcLocation)
    val idx_neq = vercors.BinaryExpression(idx,
      vercors.StringNode("!=", node.srcLocation, " "),
      vercors.StringNode("null", node.srcLocation, " "),
      node.srcLocation)
    val forall = vercors.ForallExpression(List(vdecl), Some(guard), idx_neq, node.srcLocation)
    vercors.BinaryExpression(neq, vercors.StringNode("&&", node.srcLocation), forall, node.srcLocation)
  }

  def visitTypeOfExpression(node: ir.TypeOfExpression): vercors.TypeOfExpression = {
    node match {
      case ir.TypeOfExpression(expression, srcLocation, precedingWhitespace) =>
        val res = vercors.TypeOfExpression(visitExpression(expression),
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitTypeExpression(node: ir.TypeExpression): vercors.TypeExpression = {
    node match {
      case ir.TypeExpression(indication, srcLocation, precedingWhitespace) =>
        val res = vercors.TypeExpression(visitTypeIndication(indication),
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }


  def visitForallExpression(node: ir.ForallExpression): vercors.ForallExpression = {
    node match {
      case ir.ForallExpression(declarations, guard, expression, srcLocation, precedingWhitespace) =>
        var irGuard = guard.map(visitExpression)
        if(irGuard.isEmpty) {
          irGuard = Some(vercors.StringNode("true", srcLocation, " "))
        }
        // nullable by default, special treatment for non-null
        val nonnulls = declarations.filter(x => x.isInstanceOf[ir.NonNullQuantifiedVarDeclaration])
        if(nonnulls.nonEmpty) {
          val par = vercors.ParExpression(irGuard.get, irGuard.get.srcLocation, " ")
          val vars = nonnulls.flatMap(x => x.declarators)
          var nullcheck = vercors.BinaryExpression(visitStringNode(vars.head),
                                                    vercors.StringNode("!=", srcLocation),
                                                    vercors.StringNode("null", srcLocation),
                                                    srcLocation, " ")
          for(v <- vars.tail) {
            val notnull = vercors.BinaryExpression(visitStringNode(v),
                                                  vercors.StringNode("!=", srcLocation),
                                                  vercors.StringNode("null", srcLocation),
                                                  srcLocation, " ")
            nullcheck = vercors.BinaryExpression(nullcheck, vercors.StringNode("&&", srcLocation, " "), notnull, srcLocation)
          }
          val newGuard = vercors.BinaryExpression(nullcheck, vercors.StringNode("&&", srcLocation, " "), par, srcLocation)
          irGuard = Some(newGuard)
        }
        val decls = declarations.flatMap(x => x.declarators.map(y => visitQuantifiedVarDeclaration(x,y)))
        val res = vercors.ForallExpression(decls,
                                          irGuard,
                                          visitExpression(expression),
                                          srcLocation)
        res.copyWhitespace(node)
        res


//        def nestForalls(declarators: List[ir.StringNode]): vercors.ForallExpression = {
//          var irGuard: Option[vercors.Expression] = None
//          var innerExpr: vercors.Expression = null
//          if(declarators.length > 1) {
//            innerExpr = nestForalls(declarators.tail)
//          } else {
//            innerExpr = visitExpression(expression)
//          }
//          if(declarators.length == 1) {
//            irGuard = guard.map(visitExpression)
//          }
//          if(irGuard.isEmpty) {
//            irGuard = Some(vercors.StringNode("true", srcLocation, " "))
//          }
//          val res = vercors.ForallExpression(visitQuantifiedVarDeclaration(declaration, declarators.head),
//                                            irGuard,
//                                            innerExpr,
//                                            srcLocation,
//                                            " ")
//          res
//        }
//
//        val res = nestForalls(declaration.declarators)
//        res.copyWhitespace(node)
//        res
    }
  }
  def visitForallWithPermissionsExpression(node: ir.ForallWithPermissionsExpression): vercors.ForallStarExpression = {
    node match {
      case ir.ForallWithPermissionsExpression(declarations, guard, expression, srcLocation, precedingWhitespace) =>
        var irGuard = guard.map(visitExpression)
        if(irGuard.isEmpty) {
          irGuard = Some(vercors.StringNode("true", srcLocation, " "))
        }
        val decls = declarations.flatMap(x => x.declarators.map(y => visitQuantifiedVarDeclaration(x,y)))
        val res = vercors.ForallStarExpression(decls,
          irGuard,
          visitExpression(expression),
          srcLocation)
        res.copyWhitespace(node)
        res


//        def nestForalls(declarators: List[ir.StringNode]): vercors.ForallStarExpression = {
//          var irGuard: Option[vercors.Expression] = None
//          var innerExpr: vercors.Expression = null
//          if(declarators.length > 1) {
//            innerExpr = nestForalls(declarators.tail)
//          } else {
//            innerExpr = visitExpression(expression)
//          }
//          if(declarators.length == 1) {
//            irGuard = guard.map(visitExpression)
//          }
//          if(irGuard.isEmpty) {
//            irGuard = Some(vercors.StringNode("true", srcLocation, " "))
//          }
//          val res = vercors.ForallStarExpression(visitQuantifiedVarDeclaration(declaration, declarators.head),
//                                                irGuard,
//                                                innerExpr,
//                                                srcLocation,
//                                                " ")
//          res
//        }
//
//        val res = nestForalls(declaration.declarators)
//        res.copyWhitespace(node)
//        res
    }
  }
  def visitExistsExpression(node: ir.ExistsExpression): vercors.ExistsExpression = {
    node match {
      case ir.ExistsExpression(declarations, guard, expression, srcLocation, precedingWhitespace) =>
        var irGuard = guard.map(visitExpression)
        if(irGuard.isEmpty) {
          irGuard = Some(vercors.StringNode("true", srcLocation, " "))
        }
        val decls = declarations.flatMap(x => x.declarators.map(y => visitQuantifiedVarDeclaration(x,y)))
        val res = vercors.ExistsExpression(decls,
          irGuard,
          visitExpression(expression),
          srcLocation)
        res.copyWhitespace(node)
        res


//        def nestExists(declarators: List[ir.StringNode]): vercors.ExistsExpression = {
//          var irGuard: Option[vercors.Expression] = None
//          var innerExpr: vercors.Expression = null
//          if(declarators.length > 1) {
//            innerExpr = nestExists(declarators.tail)
//          } else {
//            innerExpr = visitExpression(expression)
//          }
//          if(declarators.length == 1) {
//            irGuard = guard.map(visitExpression)
//          }
//          if(irGuard.isEmpty) {
//            irGuard = Some(vercors.StringNode("true", srcLocation, " "))
//          }
//          val res = vercors.ExistsExpression(visitQuantifiedVarDeclaration(declaration, declarators.head),
//                                              irGuard,
//                                              innerExpr,
//                                              srcLocation,
//                                              " ")
//          res
//        }
//
//        val res = nestExists(declaration.declarators)
//        res.copyWhitespace(node)
//        res
    }
  }
  def visitMaxQuantifiedExpression(node: ir.MaxQuantifiedExpression): vercors.Expression = {
    throw new NotSupportedException("\\max expression", node.srcLocation)
  }
  def visitMinExpression(node: ir.MinExpression): vercors.Expression = {
    throw new NotSupportedException("\\min expression", node.srcLocation)
  }
  def visitNumOfExpression(node: ir.NumOfExpression): vercors.Expression = {
    throw new NotSupportedException("\\num_of expression", node.srcLocation)
  }
  def visitProductExpression(node: ir.ProductExpression): vercors.Expression = {
    throw new NotSupportedException("\\product expression", node.srcLocation)
  }
  def visitSumExpression(node: ir.SumExpression): vercors.Expression = {
    throw new NotSupportedException("\\sum expression", node.srcLocation)
  }

  def visitQuantifiedVarDeclaration(node: ir.QuantifiedVarDeclaration, declaredVar: ir.StringNode): vercors.QuantifiedVarDeclaration = {
    node match {
      case ir.QuantifiedVarDeclaration(declType, _, srcLocation, precedingWhitespace) =>
        val declVar = visitStringNode(declaredVar)
        if (declVar.precedingWhitespace == "" && !declVar.text.startsWith(" ")) {
          declVar.precedingWhitespace = " "
        }
        val res = vercors.QuantifiedVarDeclaration(visitTypeIndication(declType),
                                              List(declVar),
                                              srcLocation,
                                              precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitSetComprehension(node: ir.SetComprehension): vercors.Expression = {
    node match {
      case ir.SetComprehension(typeIndication, declaration, containsExpression, filterExpression, srcLocation, precedingWhitespace) =>
        // todo: warning about limitations?
        val filter = vercors.BinaryExpression(visitExpression(containsExpression),
          vercors.StringNode("&&", srcLocation, " "),
          visitExpression(filterExpression),
          srcLocation)
        filter.operator.copyWhitespace(node.and)
        val res = vercors.VerCorsSetComprehension(visitTypeIndication(declaration.declType),
            visitStringNode(declaration.declarators.head),
            declaration.declarators.map(x => visitQuantifiedVarDeclaration(declaration, x)),
            filter,
            srcLocation)
        res.bar.copyWhitespace(node.bar)
        res.copyWhitespace(node)
        res
    }
  }

  def visitPermissionExpression(node: ir.PermissionExpression): vercors.PermissionPredicate = {
    node match {
      case ir.PermissionExpression(expr, amount, src, precedingWhitespace) =>
        val vAmount = if(amount.isDefined) Some(visitExpression(amount.get))
                      else Some(vercors.Literal(vercors.StringNode("write", src), src))
        val res = vercors.PermissionPredicate(visitExpression(expr), vAmount, src)
        res.copyWhitespace(node)
        res
    }
  }

  def visitValuedPermissionExpression(node: ir.ValuedPermissionExpression): vercors.ValuedPermissionPredicate = {
    node match {
      case ir.ValuedPermissionExpression(target, value, amount, srcLocation, precedingWhitespace) =>
        val res = vercors.ValuedPermissionPredicate(visitExpression(target),
                                                    visitExpression(value),
                                                    Some(visitExpression(amount.get)),
                                                    srcLocation,
                                                    precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitCurrentPermissionExpression(node: ir.CurrentPermissionExpression): vercors.CurrentPermissionExpression = {
    node match {
      case ir.CurrentPermissionExpression(target, srcLocation, precedingWhitespace) =>
        val res = vercors.CurrentPermissionExpression(visitExpression(target), srcLocation)
        res.copyWhitespace(node)
        res
    }
  }

  def visitSomePermissionPredicate(node: ir.SomePermissionPredicate): vercors.SomePermissionPredicate = {
    node match {
      case ir.SomePermissionPredicate(target, srcLocation, precedingWhitespace) =>
        val res = vercors.SomePermissionPredicate(visitExpression(target), srcLocation)
        res.copyWhitespace(node)
        res
    }
  }

  def visitScaledPredicateExpression(node: ir.ScaledPredicateExpression): vercors.ScaledPredicateExpression = {
    node match {
      case ir.ScaledPredicateExpression(amount, predicate, srcLocation, precedingWhitespace) =>
        val res = vercors.ScaledPredicateExpression(visitExpression(amount),
                                                    visitExpression(predicate),
                                                    srcLocation,
                                                    precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitUnfoldingExpression(node: ir.UnfoldingExpression): vercors.UnfoldingExpression = {
    node match {
      case ir.UnfoldingExpression(predicate, expression, srcLocation, precedingWhitespace) =>
        val res = vercors.UnfoldingExpression(visitExpression(predicate),
                                              visitExpression(expression),
                                              srcLocation,
                                              precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitSpecDeclarationBlock(node: ir.SpecDeclarationBlock): vercors.SpecDeclarationBlock = {
    node match {
      case ir.SpecDeclarationBlock(startSpec, declarations, endSpec, srcLocation, precedingWhitespace) =>
        val oldWrapInSpec = wrapInSpec
        wrapInSpec = false
        val res = vercors.SpecDeclarationBlock(startSpec.map(visitStringNode),
                                          declarations.map(visitSpecDeclaration),
                                          endSpec.map(visitStringNode),
                                          srcLocation,
                                          precedingWhitespace)
        res.copyWhitespace(node)
        wrapInSpec = oldWrapInSpec
        res
    }
  }

  def visitSpecDeclaration(node: ir.SpecDeclaration): vercors.SpecDeclaration = {
    try
      node match {
        case decl: ir.InvariantDeclaration => visitInvariantDeclaration(decl)
        case decl: ir.InitiallyDeclaration => visitInitiallyDeclaration(decl)
        case decl: ir.ReadableDeclaration => visitReadableDeclaration(decl)
        case decl: ir.WritableDeclaration => visitWritableDeclaration(decl)
        case decl: ir.MonitorsForDeclaration => visitMonitorsForDeclaration(decl)
        case decl: ir.AxiomDeclaration => visitAxiomDeclaration(decl)
        case decl: ir.SpecMethodDeclaration => visitSpecMethodDeclaration(decl)
        case decl: ir.GhostDeclaration => visitGhostDeclaration(decl)
        case decl: ir.ModelDeclaration => visitModelDeclaration(decl)
        case decl: ir.NonTranslatedSpecDeclaration => visitNonTranslatedSpecDeclaration(decl)
        case decl: ir.PredicateDeclaration => visitPredicateDeclaration(decl)
        case _ => ???
      }
    catch {
      case e: NotSupportedException =>
        val text = Helper.commentUnknownNode(node, s"${e.construct} not supported, whole SpecDeclaration commented out")
        vercors.NonTranslatedSpecDeclaration(text, node.srcLocation)
    }
  }

  def visitNonTranslatedSpecDeclaration(decl: ir.NonTranslatedSpecDeclaration): vercors.SpecDeclaration = {
    decl match {
      case ir.NonTranslatedSpecDeclaration(text, srcLocation, precedingWhitespace) =>
        vercors.NonTranslatedSpecDeclaration(text, srcLocation, precedingWhitespace)
    }
  }
  def visitInvariantDeclaration(node: ir.InvariantDeclaration): vercors.NonTranslatedSpecDeclaration = {
    val text = Helper.commentNode(node, node.invariantKeyword, "Invariant Clause transformed to pre-/postconditions. Commented out.")
    val res = vercors.NonTranslatedSpecDeclaration(text, node.srcLocation)
    res.copyWhitespace(node)
    res
  }
  def visitInitiallyDeclaration(node: ir.InitiallyDeclaration): vercors.NonTranslatedSpecDeclaration = {
    val text = Helper.commentNode(node, node.initiallyKeyword, "Initially Clause not supported in VerCors. Commented out.")
    val res = vercors.NonTranslatedSpecDeclaration(text, node.srcLocation)
    res.copyWhitespace(node)
    res
  }
  def visitReadableDeclaration(node: ir.ReadableDeclaration): vercors.NonTranslatedSpecDeclaration = {
    val text = Helper.commentNode(node, node.readableKeyword, "Readable-If Clause not supported in VerCors. Commented out.")
    val res = vercors.NonTranslatedSpecDeclaration(text, node.srcLocation)
    res.copyWhitespace(node)
    res
  }
  def visitWritableDeclaration(node: ir.WritableDeclaration): vercors.NonTranslatedSpecDeclaration = {
    val text = Helper.commentNode(node, node.writable, "Writable-If Clause not supported in VerCors. Commented out.")
    val res = vercors.NonTranslatedSpecDeclaration(text, node.srcLocation)
    res.copyWhitespace(node)
    res
  }
  def visitMonitorsForDeclaration(node: ir.MonitorsForDeclaration): vercors.NonTranslatedSpecDeclaration = {
    val text = Helper.commentNode(node, node.monitorsForKeyword, "Monitors-For Clause not supported in VerCors. Commented out.")
    val res = vercors.NonTranslatedSpecDeclaration(text, node.srcLocation)
    res.copyWhitespace(node)
    res
  }
  def visitAxiomDeclaration(node: ir.AxiomDeclaration): vercors.NonTranslatedSpecDeclaration = {
    val text = Helper.commentNode(node, node.axiomKeyword, "Axiom Clause not supported in VerCors. Commented out.")
    val res = vercors.NonTranslatedSpecDeclaration(text, node.srcLocation)
    res.copyWhitespace(node)
    res
  }
  def visitPredicateDeclaration(node: ir.PredicateDeclaration): vercors.PredicateDeclaration = {
    node match {
      case ir.PredicateDeclaration(name, tType, params, expr, srcLocation, precedingWhitespace) =>
        val res = vercors.PredicateDeclaration(visitStringNode(name),
                                              params.map(visitParameter),
                                              expr.map(visitExpression),
                                              srcLocation)
        res.copyWhitespace(node)
        res.tType.copyWhitespace(tType)
        res
    }
  }

  def visitSpecMethodDeclaration(node: ir.SpecMethodDeclaration): vercors.SpecMethodDeclaration = {
    node match {
      case ir.SpecMethodDeclaration(name, returnType, parameters, contractWithin,
                                    body, srcLocation, precedingWhitespace) =>
        val irReturn = returnType.map(visitTypeIndication)
        val res = vercors.SpecMethodDeclaration(visitStringNode(name),
                                                irReturn,
                                                parameters.map(visitParameter),
                                                contractWithin.map(visitContract),
                                                visitBlockOrSemicolon(body),
                                                srcLocation,
                                                precedingWhitespace)
        val contract = node.contractBefore.map(visitContract)
        // Remove start-/endspec if they are there since we are already in a spec
        if (contract.isDefined && contract.get.startSpec.isDefined) {
          val whitespace = contract.get.startSpec.get.precedingWhitespace
          contract.get.startSpec = None
          contract.get.precedingWhitespace = contract.get.precedingWhitespace.appendedAll(whitespace)
        }
        if (contract.isDefined && contract.get.endSpec.isDefined) {
          val whitespace = contract.get.endSpec.get.precedingWhitespace
          contract.get.endSpec = None
          contract.get.trailingWhitespace = contract.get.trailingWhitespace.prependedAll(whitespace)
        }
        res.contractBefore = contract
        res.modifiers = node.modifiers.map(visitModifier)
        if(!Helper.findModifier(res.modifiers, x=>x.isInstanceOf[vercors.GhostModifier] | x.isInstanceOf[vercors.SpecMethodModifier])) {
          val ghost = vercors.GhostModifier(srcLocation, " ")
          res.modifiers = res.modifiers.prepended(ghost)
        }
        res.copyWhitespace(node)
        res
    }
  }


}
