package languages.VerCors

import generic.{Helper, NotSupportedException}
import intermediateRepresentation.syntaxTree as ir
import languages.VerCors.syntaxTree as vercors

import scala.collection.mutable.ListBuffer


class ToIrVisitor {

  // store references to all variables occurring in e.g. contracts, in case they need renaming
  var variables: ListBuffer[ir.StringNode] = ListBuffer()


  def copyWhitespace(from: vercors.VerCorsTreeNode, to: ir.IrTreeNode): Unit = {
    to.precedingWhitespace = from.precedingWhitespace
    to.trailingWhitespace = from.trailingWhitespace
  }
  
  def visitStringNode(node: vercors.StringNode): ir.StringNode = {
    node match {
      case imp: vercors.ImpliesOperator => visitImpliesOperator(imp)
      case sep: vercors.SeparatingConjunctOperator => visitSeparatingConjunctOperator(sep)
      case sep: vercors.PermissionFractionOperator => visitPermissionFractionOperator(sep)
      case vercors.StringNode(text, src, pred, trail) =>
        val res = ir.StringNode(text, src, pred, trail)
        variables.append(res)
        res
    }
  }
  def visitImpliesOperator(node: vercors.ImpliesOperator): ir.ImpliesOperator = {
    node match {
      case vercors.ImpliesOperator(srcLocation, pred) =>
        val res = new ir.ImpliesOperator(srcLocation)
        res.copyWhitespace(node)
        res
    }
  }
  def visitSeparatingConjunctOperator(node: vercors.SeparatingConjunctOperator): ir.SeparatingConjunctOperator = {
    node match {
      case vercors.SeparatingConjunctOperator(srcLocation, pred) =>
        val res = new ir.SeparatingConjunctOperator(node.text, srcLocation)
        res.copyWhitespace(node)
        res
    }
  }
  def visitPermissionFractionOperator(node: vercors.PermissionFractionOperator): ir.StringNode = {
    node match {
      case vercors.PermissionFractionOperator(srcLocation, pred) =>
        val res = new ir.StringNode("/", srcLocation)
        res.copyWhitespace(node)
        res
    }
  }
  def visitModifierStringNode(node: vercors.ModifierStringNode): ir.ModifierStringNode = {
    node match {
      case vercors.ModifierStringNode(text, src, pred) =>
        val res = ir.ModifierStringNode(text, src, pred)
        res.copyWhitespace(node)
        res
    }
  }

  def visitCompilationUnit(node: vercors.CompilationUnit): ir.CompilationUnit = {
    DesugarVisitor().visitCompilationUnit(node)
    node match {
      case nt: vercors.NonTranslatedCompilationUnit =>
        visitNonTranslatedCompilationUnit(nt)
      case vercors.CompilationUnit(pkg, imports, declarations, srcLocation, precedingWhitespace) =>
        try
          var irPkg: Option[ir.PackageDeclaration] = None
          if (pkg.isDefined) {
            irPkg = Some(visitPackageDeclaration(pkg.get))
          }
          val res = ir.CompilationUnit(irPkg,
                                        imports.map(visitImportDeclaration),
                                        declarations.map(visitTypeDeclaration),
                                        srcLocation,
                                        precedingWhitespace)
          res.copyWhitespace(node)
          res
        catch {
          case e: NotSupportedException =>
            val text = Helper.commentUnknownNode(node, s"${e.construct} not supported, whole CompilationUnit commented out")
            ir.NonTranslatedCompilationUnit(text, node.srcLocation)
        }
    }
  }

  def visitNonTranslatedCompilationUnit(node: vercors.NonTranslatedCompilationUnit): ir.NonTranslatedCompilationUnit = {
    node match {
      case vercors.NonTranslatedCompilationUnit(text, srcLocation, precedingWhitespace) =>
        val res = ir.NonTranslatedCompilationUnit(text, srcLocation, precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitPackageDeclaration(node: vercors.PackageDeclaration): ir.PackageDeclaration = {
    node match {
      case vercors.PackageDeclaration(name, annotations, srcLocation, precedingWhitespace) =>
        val res = ir.PackageDeclaration(visitStringNode(name),
                                        annotations.map(visitStringAnnotation),
                                        srcLocation,
                                        precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitImportDeclaration(node: vercors.ImportDeclaration): ir.ImportDeclaration = {
    node match {
      case vercors.ImportDeclaration(name, srcLocation, precedingWhitespace) =>
        val res = ir.ImportDeclaration(visitStringNode(name),
                              srcLocation,
                              precedingWhitespace)
        res.modifiers = node.modifiers.map(visitModifier)
        res.copyWhitespace(node)
        res
    }
  }

  def visitTypeDeclaration(node: vercors.TypeDeclaration): ir.TypeDeclaration = {
    try
      node match {
        case decl: vercors.AnnotationTypeDeclaration => visitAnnotationTypeDeclaration(decl)
        case decl: vercors.ClassDeclaration => visitClassDeclaration(decl)
        case decl: vercors.EnumDeclaration => visitEnumDeclaration(decl)
        case decl: vercors.InterfaceDeclaration => visitInterfaceDeclaration(decl)
        case decl: vercors.EmptyStatement => visitEmptyStatement(decl)
        case decl: vercors.NonTranslatedTypeDeclaration => visitNonTranslatedTypeDeclaration(decl)
        case _ => ???
      }
    catch {
      case e: NotSupportedException =>
        val text = Helper.commentUnknownNode(node, s"${e.construct} not supported, whole TypeDeclaration commented out")
        ir.NonTranslatedTypeDeclaration(text, node.srcLocation)
    }
  }

  def visitNonTranslatedTypeDeclaration(node: vercors.NonTranslatedTypeDeclaration): ir.NonTranslatedTypeDeclaration = {
    node match {
      case vercors.NonTranslatedTypeDeclaration(text, srcLocation, precedingWhitespace) =>
        val res = ir.NonTranslatedTypeDeclaration(text, srcLocation, precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitGhostDeclaration(node: vercors.GhostDeclaration): ir.GhostDeclaration = {
    node match {
      case vercors.GhostDeclaration(ghost, decl, srcLocation, precedingWhitespace) =>
        val res = ir.GhostDeclaration(visitGhostModifier(ghost),
          visitMemberDeclaration(decl),
          srcLocation,
          precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitModifier(node: vercors.Modifier): ir.Modifier = {
    node match {
      case mod: vercors.StaticModifier => visitStaticModifier(mod)
      case mod: vercors.Annotation => visitAnnotation(mod)
      case mod: vercors.PureModifier => visitPureModifier(mod)
      case mod: vercors.GhostModifier => visitGhostModifier(mod)
      case mod: vercors.SpecMethodModifier => visitSpecMethodModifier(mod)
      case mod: vercors.InlineModifier => visitInlineModifier(mod)
      case mod: vercors.ThreadLocalModifier => visitThreadLocalModifier(mod)
      case mod: vercors.JmlModifierBlock => visitJmlModifierBlock(mod)
      case mod: vercors.ModifierStringNode => visitModifierStringNode(mod)
      case _ => ???
    }
    // todo: more?
  }
  def visitStaticModifier(node: vercors.StaticModifier): ir.StaticModifier = {
    node match {
      case vercors.StaticModifier(srcLocation, precedingWhitespace) =>
        val res = ir.StaticModifier(srcLocation, precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitGenericsSuperclassClause(node: vercors.GenericsSuperclassClause): ir.GenericsSuperclassClause = {
    node match {
      case vercors.GenericsSuperclassClause(superclasses, srcLocation, precedingWhitespace) =>
        val res = ir.GenericsSuperclassClause(superclasses.map(visitTypeIndication),
                                              srcLocation,
                                              precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitSuperclassClause(node: vercors.SuperclassClause): ir.SuperclassClause = {
    node match {
      case vercors.SuperclassClause(superclass, srcLocation, precedingWhitespace) =>
        val res = ir.SuperclassClause(visitTypeIndication(superclass),
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitMultiSuperclassClause(node: vercors.MultiSuperclassClause): ir.MultiSuperclassClause = {
    node match {
      case vercors.MultiSuperclassClause(superclasses, srcLocation, precedingWhitespace) =>
        val res = ir.MultiSuperclassClause(superclasses.map(visitTypeIndication),
                                            srcLocation,
                                            precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitInterfaceClause(node: vercors.InterfaceClause): ir.InterfaceClause = {
    node match {
      case vercors.InterfaceClause(interfaces, srcLocation, precedingWhitespace) =>
        val res = ir.InterfaceClause(interfaces.map(visitTypeIndication),
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitClassDeclaration(node: vercors.ClassDeclaration): ir.ClassDeclaration = {
    node match {
      case vercors.ClassDeclaration(name, typeParameters, superclass, interfaces, body, srcLocation, precedingWhitespace) =>
        val irTypeParameters = typeParameters.map(visitTypeParameterClause)
        val irSuperclass = superclass.map(visitSuperclassClause)
        val irInterfaces = interfaces.map(visitInterfaceClause)
        val res = ir.ClassDeclaration(visitStringNode(name),
                                      irTypeParameters,
                                      irSuperclass,
                                      irInterfaces,
                                      visitBody(body),
                                      srcLocation,
                                      precedingWhitespace)
        res.modifiers = node.modifiers.map(visitModifier)
        res.copyWhitespace(node)
        res
    }
  }

  def visitTypeParameterClause(node: vercors.TypeParameterClause): ir.TypeParameterClause = {
    node match {
      case vercors.TypeParameterClause(parameters, srcLocation, precedingWhitespace) =>
        val res = ir.TypeParameterClause(parameters.map(visitTypeParameter),
                                          srcLocation,
                                          precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitTypeParameter(node: vercors.TypeParameter): ir.TypeParameter = {
    node match {
      case vercors.TypeParameter(annotations, name, superclass, srcLocation, precedingWhitespace) =>
        val irSuperclass = superclass.map(visitGenericsSuperclassClause)
        val res = ir.TypeParameter(annotations.map(visitStringAnnotation),
                                    visitTypeIndication(name),
                                    irSuperclass,
                                    srcLocation,
                                    precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitEnumDeclaration(node: vercors.EnumDeclaration): ir.EnumDeclaration = {
    node match {
      case vercors.EnumDeclaration(name, interfaces, constants, declarations, comma, semicolon, srcLocation, precedingWhitespace) =>
        val irInterfaces = interfaces.map(visitInterfaceClause)
        val irComma = comma.map(visitStringNode)
        val irSemicolon = semicolon.map(visitStringNode)
        val res = ir.EnumDeclaration(visitStringNode(name),
                                      irInterfaces,
                                      constants.map(visitEnumConstant),
                                      declarations.map(visitMemberDeclaration),
                                      irComma,
                                      irSemicolon,
                                      srcLocation,
                                      precedingWhitespace)
        res.modifiers = node.modifiers.map(visitModifier)
        res.copyWhitespace(node)
        res
    }
  }
  def visitEnumConstant(node: vercors.EnumConstant): ir.EnumConstant = {
    node match {
      case vercors.EnumConstant(name, annotations, arguments, body, srcLocation, precedingWhitespace) =>
        val irArgs = arguments.map(visitArguments)
        val irBody = body.map(visitBody)
        val res = ir.EnumConstant(visitStringNode(name),
                                  annotations.map(visitStringAnnotation),
                                  irArgs,
                                  irBody,
                                  srcLocation,
                                  precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitInterfaceDeclaration(node: vercors.InterfaceDeclaration): ir.InterfaceDeclaration = {
    node match {
      case vercors.InterfaceDeclaration(name, typeParameter, superInterfaces, body, srcLocation, precedingWhitespace) =>
        val irTypeParameters = typeParameter.map(visitTypeParameterClause)
        val irSuperclass = superInterfaces.map(visitMultiSuperclassClause)
        val res = ir.InterfaceDeclaration(visitStringNode(name),
                                          irTypeParameters,
                                          irSuperclass,
                                          visitBody(body),
                                          srcLocation,
                                          precedingWhitespace)
        res.modifiers = node.modifiers.map(visitModifier)
        res.copyWhitespace(node)
        res
    }
  }

  def visitBody(node: vercors.Body): ir.Body = {
    node match {
      case vercors.Body(declarations, srcLocation, precedingWhitespace) =>
        val res = ir.Body(declarations.map(visitMemberDeclaration),
                          srcLocation,
                          precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitMemberDeclaration(node: vercors.MemberDeclaration): ir.MemberDeclaration = {
    try
      node match {
        case member: vercors.ClassDeclaration => visitClassDeclaration(member)
        case member: vercors.EnumDeclaration => visitEnumDeclaration(member)
        case member: vercors.InterfaceDeclaration => visitInterfaceDeclaration(member)
        case member: vercors.GenericMethodDeclaration => visitGenericMethodDeclaration(member)
        case member: vercors.MethodDeclaration => visitMethodDeclaration(member)
        case member: vercors.BlockOrSemicolon => visitBlockOrSemicolon(member)
        case member: vercors.FieldDeclaration => visitFieldDeclaration(member)
        case member: vercors.AnnotationTypeDeclaration => visitAnnotationTypeDeclaration(member)
        case member: vercors.SpecDeclaration => visitSpecDeclaration(member)
        case member: vercors.SpecDeclarationBlock => visitSpecDeclarationBlock(member)
        case member: vercors.NonTranslatedMemberDeclaration => visitNonTranslatedMemberDeclaration(member)
        case _ => ???
      }
    catch {
      case e: NotSupportedException =>
        val text = Helper.commentUnknownNode(node, s"${e.construct} not supported, whole MemberDeclaration commented out")
        ir.NonTranslatedMemberDeclaration(text, node.srcLocation)
    }
  }
  def visitNonTranslatedMemberDeclaration(node: vercors.NonTranslatedMemberDeclaration): ir.NonTranslatedMemberDeclaration = {
    node match {
      case vercors.NonTranslatedMemberDeclaration(text, srcLocation, precedingWhitespace) =>
        val res = ir.NonTranslatedMemberDeclaration(text, srcLocation, precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitMethodDeclaration(node: vercors.MethodDeclaration): ir.MethodDeclaration = {
    node match {
      case vercors.MethodDeclaration(name, returnType, parameters, dimensions, throws, default, contractWithin,
                                 body, srcLocation, precedingWhitespace) =>
        val irReturn = returnType.map(visitTypeIndication)
        val irDims = dimensions.map(visitStringNode)
        val irThrows = throws.map(visitThrowsClause)
        val irDefault = default.map(visitDefaultClause)
        variables.clear()
        val irContractWithin = contractWithin.map(visitContract)
        val irContractBefore = node.contractBefore.map(visitContract)
        if(returnType.isDefined) {
          // not supported for constructor, only for "real" methods
          renameGhostParameters(node)
        }
        val res = new ir.MethodDeclaration(visitStringNode(name),
                                            irReturn,
                                            parameters.map(visitParameter),
                                            irDims,
                                            irThrows,
                                            irDefault,
                                            irContractWithin,
                                            visitBlockOrSemicolon(body),
                                            srcLocation,
                                            precedingWhitespace)
        res.contractBefore = irContractBefore
        res.modifiers = node.modifiers.map(visitModifier)
        res.copyWhitespace(node)
        res
    }
  }


  def visitGenericMethodDeclaration(node: vercors.GenericMethodDeclaration): ir.GenericMethodDeclaration = {
    node match {
      case vercors.GenericMethodDeclaration(pName, typeParameters, pReturnType,
                                        pParameters, pDimensions, pThrows, pDefault, pContractWithin,
                                        pBody, pSrcLocation, pPrecedingWhitespace) =>
        val irReturn = pReturnType.map(visitTypeIndication)
        val irDims = pDimensions.map(visitStringNode)
        val irThrows = pThrows.map(visitThrowsClause)
        val irDefault = pDefault.map(visitDefaultClause)
        variables.clear()
        val irContractWithin = pContractWithin.map(visitContract)
        val irContractBefore = node.contractBefore.map(visitContract)
        if(pReturnType.isDefined) {
          // not supported for constructor, only for "real" methods
          renameGhostParameters(node)
        }
        val res = ir.GenericMethodDeclaration(visitStringNode(pName),
                                              visitTypeParameterClause(typeParameters),
                                              irReturn,
                                              pParameters.map(visitParameter),
                                              irDims,
                                              irThrows,
                                              irDefault,
                                              irContractWithin,
                                              visitBlockOrSemicolon(pBody),
                                              pSrcLocation,
                                              pPrecedingWhitespace)
        res.contractBefore = irContractBefore
        res.modifiers = node.modifiers.map(visitModifier)
        res.copyWhitespace(node)
        res
    }
  }

  def visitThrowsClause(node: vercors.ThrowsClause): ir.ThrowsClause = {
    node match {
      case vercors.ThrowsClause(exceptions, srcLocation, precedingWhitespace) =>
        val res = ir.ThrowsClause(exceptions.map(visitStringNode),
                                  srcLocation,
                                  precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitTypeIndication(node: vercors.TypeIndication): ir.TypeIndication = {
    node match {
      case indication: vercors.NoType => visitNoType(indication)
      case indication: vercors.StringGenericTypeIndication => visitStringGenericTypeIndication(indication)
      case indication: vercors.StringTypeIndication => visitStringTypeIndication(indication)
      case _ => ???
    }
  }
  def visitStringTypeIndication(node: vercors.StringTypeIndication): ir.StringTypeIndication = {
    node match {
      case vercors.StringTypeIndication(fullTypeString, annotations, typeAsString,
                                    pSrcLocation, pPrecedingWhitespace, pTrailingWhitespace) =>
        val res = new ir.StringTypeIndication(fullTypeString,
                                              annotations.map(visitAnnotation),
                                              typeAsString,
                                              pSrcLocation,
                                              pPrecedingWhitespace,
                                              pTrailingWhitespace)
        res.modifiers = node.modifiers.map(visitModifier)
        res.copyWhitespace(node)
        res
    }
  }

  def visitStringGenericTypeIndication(node: vercors.StringGenericTypeIndication): ir.StringGenericTypeIndication = {
    node match {
      case vercors.StringGenericTypeIndication(fullTypeString, typeParameters, annotations, typeAsString,
                                            pSrcLocation, pPrecedingWhitespace) =>
        val res = ir.StringGenericTypeIndication(fullTypeString,
                                                  visitTypeParameterClause(typeParameters),
                                                  annotations.map(visitAnnotation),
                                                  typeAsString,
                                                  pSrcLocation,
                                                  pPrecedingWhitespace)
        res.modifiers = node.modifiers.map(visitModifier)
        res.copyWhitespace(node)
        res
    }
  }
  def visitNoType(node: vercors.NoType): ir.NoType = {
    node match {
      case vercors.NoType(srcLocation) =>
        val res = new ir.NoType(srcLocation)
        res.copyWhitespace(node)
        res
    }
  }

  def visitBlockOrSemicolon(node: vercors.BlockOrSemicolon): ir.BlockOrSemicolon = {
    node match {
      case block: vercors.Block => visitBlock(block)
      case semi: vercors.EmptyStatement => visitEmptyStatement(semi)
      case _ => ???
    }
  }

  def visitFieldDeclaration(node: vercors.FieldDeclaration): ir.FieldDeclaration = {
    node match {
      case vercors.FieldDeclaration(fieldType, declarators, srcLocation, precedingWhitespace) =>
        val res = ir.FieldDeclaration(visitTypeIndication(fieldType),
                                      declarators.map(visitVariableDeclarator),
                                      srcLocation,
                                      precedingWhitespace)
        res.modifiers = node.modifiers.map(visitModifier)
        res.copyWhitespace(node)
        res
    }
  }

  def visitVariableDeclarator(node: vercors.VariableDeclarator): ir.VariableDeclarator = {
    node match {
      case vercors.VariableDeclarator(name, initializer, srcLocation, precedingWhitespace) =>
        val irInit = initializer.map(visitInitializer)
        val res = ir.VariableDeclarator(visitStringNode(name),
                                        irInit,
                                        srcLocation,
                                        precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitInitializer(node: vercors.Initializer): ir.Initializer = {
    node match {
      case init: vercors.ArrayInitializer => visitArrayInitializer(init)
      case init: vercors.Expression => visitExpression(init)
      case _ => ???
    }
  }
  def visitArrayInitializer(node: vercors.ArrayInitializer): ir.ArrayInitializer = {
    node match {
      case vercors.ArrayInitializer(initializers, trailingComma, srcLocation, precedingWhitespace) =>
        val irComma = trailingComma.map(visitStringNode)
        val res = ir.ArrayInitializer(initializers.map(visitInitializer),
                                      irComma,
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitTypeArgumentClause(node: vercors.TypeArgumentClause): ir.TypeArgumentClause = {
    node match {
      case vercors.TypeArgumentClause(parameters, srcLocation, precedingWhitespace) =>
        val res = ir.TypeArgumentClause(parameters.map(visitTypeArgument),
                                        srcLocation,
                                        precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitTypeArgument(node: vercors.TypeArgument): ir.TypeArgument = {
    node match {
      case arg: vercors.BoundedTypeArgument => visitBoundedTypeArgument(arg)
      case arg: vercors.TypeIndication => visitTypeIndication(arg)
      case _ => ???
    }
  }
  def visitBoundedTypeArgument(node: vercors.BoundedTypeArgument): ir.BoundedTypeArgument = {
    node match {
      case vercors.BoundedTypeArgument(annotations, typeSuper, typeExtends, srcLocation, precedingWhitespace) =>
        val irSuper = typeSuper.map(visitTypeIndication)
        val irExtends = typeExtends.map(visitTypeIndication)
        val res = ir.BoundedTypeArgument(annotations.map(visitAnnotation),
                                          irSuper,
                                          irExtends,
                                          srcLocation,
                                          precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitParameter(node: vercors.Parameter): ir.Parameter = {
    node match {
      case varargs: vercors.VarargsParameter =>
        visitVarargsParameter(varargs)
      case vercors.Parameter(parameterType, name, srcLocation, precedingWhitespace) =>
        val res = ir.Parameter(visitTypeIndication(parameterType),
                                visitStringNode(name),
                                srcLocation,
                                precedingWhitespace)
        res.modifiers = node.modifiers.map(visitModifier)
        res.copyWhitespace(node)
        res
    }
  }
  def visitVarargsParameter(node: vercors.VarargsParameter): ir.VarargsParameter = {
    node match {
      case vercors.VarargsParameter(parameterType, annotations, name, srcLocation, precedingWhitespace) =>
        val res = ir.VarargsParameter(visitTypeIndication(parameterType),
                                      annotations.map(visitAnnotation),
                                      visitStringNode(name),
                                      srcLocation,
                                      precedingWhitespace)
        res.modifiers = node.modifiers.map(visitModifier)
        res.copyWhitespace(node)
        res
    }
  }

  def visitAnnotation(node: vercors.Annotation): ir.Annotation = {
    node match {
      case ann: vercors.StringAnnotation => visitStringAnnotation(ann)
      case _ => ???
    }
  }

  def visitStringAnnotation(node: vercors.StringAnnotation): ir.StringAnnotation = {
    node match {
      case vercors.StringAnnotation(text, srcLocation, precedingWhitespace) =>
        val res = ir.StringAnnotation(text,
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitAnnotationTypeDeclaration(node: vercors.AnnotationTypeDeclaration): ir.AnnotationTypeDeclaration = {
    node match {
      case vercors.AnnotationTypeDeclaration(name, body, srcLocation, precedingWhitespace) =>
        val res = ir.AnnotationTypeDeclaration(visitStringNode(name),
                                                visitBody(body),
                                                srcLocation,
                                                precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitDefaultClause(node: vercors.DefaultClause): ir.DefaultClause = {
    node match {
      case vercors.DefaultClause(value, srcLocation, precedingWhitespace) =>
        val res = ir.DefaultClause(visitStringNode(value),
                                    srcLocation,
                                    precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitStatement(node: vercors.Statement): ir.Statement = {
    try
      node match {
        case local: vercors.LocalDeclaration => visitLocalDeclaration(local)
        case block: vercors.Block => visitBlock(block)
        case assert: vercors.JavaAssertion => visitJavaAssertion(assert)
        case ifStmt: vercors.IfStatement => visitIfStatement(ifStmt)
        case forLoop: vercors.ForLoop => visitForLoop(forLoop)
        case whileLoop: vercors.WhileLoop => visitWhileLoop(whileLoop)
        case doLoop: vercors.DoLoop => visitDoLoop(doLoop)
        case tryBlock: vercors.TryBlock => visitTryBlock(tryBlock)
        case switch: vercors.SwitchStatement => visitSwitchStatement(switch)
        case sync: vercors.SynchronizedStatement => visitSynchronizedStatement(sync)
        case ret: vercors.ReturnStatement => visitReturnStatement(ret)
        case throwStmt: vercors.ThrowStatement => visitThrowStatement(throwStmt)
        case break: vercors.BreakStatement => visitBreakStatement(break)
        case continue: vercors.ContinueStatement => visitContinueStatement(continue)
        case empty: vercors.EmptyStatement => visitEmptyStatement(empty)
        case expr: vercors.ExpressionStatement => visitExpressionStatement(expr)
        case labeled: vercors.LabeledStatement => visitLabeledStatement(labeled)
        case assert: vercors.AssertStatement => visitAssertStatement(assert)
        case assume: vercors.AssumeStatement => visitAssumeStatement(assume)
        case fold: vercors.FoldStatement => visitFoldStatement(fold)
        case unfold: vercors.UnfoldStatement => visitUnfoldStatement(unfold)
        case ghost: vercors.GhostStatement => visitGhostStatement(ghost)
        case jmlBlock: vercors.JmlStatementBlock => visitJmlStatementBlock(jmlBlock)
        case nonTranslated: vercors.NonTranslatedStatement => visitNonTranslatedStatement(nonTranslated)
        case _ => ???
        // todo: ClassDecl, etc
      }
    catch {
      case e: NotSupportedException =>
        val text = Helper.commentUnknownNode(node, s"${e.construct} not supported, whole Statement commented out")
        ir.NonTranslatedStatement(text, node.srcLocation)
    }
  }

  def visitBlock(node: vercors.Block): ir.Block = {
    node match {
      case vercors.Block(static, statements, srcLocation, precedingWhitespace) =>
        val irStatic = static.map(visitModifier)
        val res = ir.Block(irStatic,
                            statements.map(visitStatement),
                            srcLocation,
                            precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitLocalDeclaration(node: vercors.LocalDeclaration): ir.LocalDeclaration = {
    node match {
      case vercors.LocalDeclaration(modifiers, declType, declarators, isStatement, srcLocation, precedingWhitespace) =>
        val res = ir.LocalDeclaration(modifiers.map(visitModifier),
                                      visitTypeIndication(declType),
                                      declarators.map(visitVariableDeclarator),
                                      isStatement,
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitJavaAssertion(node: vercors.JavaAssertion): ir.JavaAssertion = {
    node match {
      case vercors.JavaAssertion(content, message, srcLocation, precedingWhitespace) =>
        val irMsg = message.map(visitExpression)
        val res = ir.JavaAssertion(visitExpression(content),
                                    irMsg,
                                    srcLocation,
                                    precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitIfStatement(node: vercors.IfStatement): ir.IfStatement = {
    node match {
      case vercors.IfStatement(condition, thenStatement, elseStatement, srcLocation, precedingWhitespace) =>
        val irElse = elseStatement.map(visitStatement)
        val res = ir.IfStatement(visitExpression(condition),
                                  visitStatement(thenStatement),
                                  irElse,
                                  srcLocation,
                                  precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitForLoop(node: vercors.ForLoop): ir.ForLoop = {
    node match {
      case vercors.ForLoop(contractBefore, control, contractWithin, body, srcLocation, precedingWhitespace) =>
        val irContractBefore = if (contractBefore.isEmpty) Nil else contractBefore.map(visitLoopSpecOrLabel)
        val irContractWithin = if (contractWithin.isEmpty) Nil else contractWithin.map(visitLoopSpecOrLabel)
        val res = ir.ForLoop(irContractBefore,
                              visitForLoopControl(control),
                              irContractWithin,
                              visitStatement(body),
                              srcLocation,
                              precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitForLoopControl(node: vercors.ForLoopControl): ir.ForLoopControl = {
    node match {
      case init: vercors.ForEachLoopControl => visitForEachLoopControl(init)
      case init: vercors.RegularForLoopControl => visitRegularForLoopControl(init)
    }
  }
  def visitForInit(node: vercors.ForInit): ir.ForInit = {
    node match {
      case init: vercors.ExpressionListForInit =>
        visitExpressionListForInit(init)
      case init: vercors.LocalDeclaration =>
        visitLocalDeclaration(init)
      case _ =>
        ???
    }
  }
  def visitExpressionListForInit(node: vercors.ExpressionListForInit): ir.ExpressionListForInit = {
    node match {
      case vercors.ExpressionListForInit(expressions, srcLocation, precedingWhitespace) =>
        val res = ir.ExpressionListForInit(expressions.map(visitExpression),
                                          srcLocation,
                                          precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitRegularForLoopControl(node: vercors.RegularForLoopControl): ir.RegularForLoopControl = {
    node match {
      case vercors.RegularForLoopControl(init, condition, update, srcLocation, precedingWhitespace) =>
        val irInit = init.map(visitForInit)
        val irCondition = condition.map(visitExpression)
        val res = ir.RegularForLoopControl(irInit,
                                            irCondition,
                                            update.map(visitExpression),
                                            srcLocation,
                                            precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitForEachLoopControl(node: vercors.ForEachLoopControl): ir.ForEachLoopControl = {
    node match {
      case vercors.ForEachLoopControl(controlType, variable, control, srcLocation, precedingWhitespace) =>
        val res = ir.ForEachLoopControl(visitTypeIndication(controlType),
                                        visitStringNode(variable),
                                        visitExpression(control),
                                        srcLocation,
                                        precedingWhitespace)
        res.modifiers = node.modifiers.map(visitModifier)
        res.copyWhitespace(node)
        res
    }
  }

  def visitWhileLoop(node: vercors.WhileLoop): ir.WhileLoop = {
    node match {
      case vercors.WhileLoop(contractBefore, condition, contractWithin, body, srcLocation, precedingWhitespace) =>
        val irContractBefore = if (contractBefore.isEmpty) Nil else contractBefore.map(visitLoopSpecOrLabel)
        val irContractWithin = if (contractWithin.isEmpty) Nil else contractWithin.map(visitLoopSpecOrLabel)
        val res = ir.WhileLoop(irContractBefore,
                                visitExpression(condition),
                                irContractWithin,
                                visitStatement(body),
                                srcLocation,
                                precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitDoLoop(node: vercors.DoLoop): ir.DoLoop = {
    node match {
      case vercors.DoLoop(contractBefore, body, condition, srcLocation, precedingWhitespace) =>
        val irContractBefore = if (contractBefore.isEmpty) Nil else contractBefore.map(visitLoopSpecOrLabel)
        val res = ir.DoLoop(irContractBefore,
                            visitStatement(body),
                            visitExpression(condition),
                            srcLocation,
                            precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitTryBlock(node: vercors.TryBlock): ir.TryBlock = {
    node match {
      case vercors.TryBlock(resourceSpecification, tryBlock, catchClauses, finallyBlock, srcLocation, precedingWhitespace) =>
        val irResSpec = if (resourceSpecification.isEmpty) None
                        else Some(visitTryResourceSpecification(resourceSpecification.get))
        val irFinally = finallyBlock.map(visitBlock)
        val res = ir.TryBlock(irResSpec,
                              visitBlock(tryBlock),
                              catchClauses.map(visitCatchClause),
                              irFinally,
                              srcLocation,
                              precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitCatchClause(node: vercors.CatchClause): ir.CatchClause = {
    node match {
      case vercors.CatchClause(types, name, block, srcLocation, precedingWhitespace) =>
        val res = ir.CatchClause(types.map(visitTypeIndication),
                                  visitStringNode(name),
                                  visitBlock(block),
                                  srcLocation,
                                  precedingWhitespace)
        res.modifiers = node.modifiers.map(visitModifier)
        res.copyWhitespace(node)
        res
    }
  }
  def visitTryResourceSpecification(node: vercors.TryResourceSpecification): ir.TryResourceSpecification = {
    node match {
      case vercors.TryResourceSpecification(resources, trailingSemicolon, srcLocation, precedingWhitespace) =>
        val irSemi = trailingSemicolon.map(visitStringNode)
        val res = ir.TryResourceSpecification(resources.map(visitTryResource),
                                              irSemi,
                                              srcLocation,
                                              precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitTryResource(node: vercors.TryResource): ir.TryResource = {
    node match {
      case vercors.TryResource(modifiers, resourceType, variable, expression, srcLocation, precedingWhitespace) =>
        val res = ir.TryResource(modifiers.map(visitModifier),
                                  visitTypeIndication(resourceType),
                                  visitStringNode(variable),
                                  visitExpression(expression),
                                  srcLocation,
                                  precedingWhitespace)
        res.modifiers = node.modifiers.map(visitModifier)
        res.copyWhitespace(node)
        res
    }
  }

  def visitSwitchStatement(node: vercors.SwitchStatement): ir.SwitchStatement = {
    node match {
      case vercors.SwitchStatement(condition, cases, srcLocation, precedingWhitespace) =>
        val res = ir.SwitchStatement(visitExpression(condition),
                                      cases.map(visitSwitchCaseBlock),
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitSwitchCaseBlock(node: vercors.SwitchCaseBlock): ir.SwitchCaseBlock = {
    node match {
      case vercors.SwitchCaseBlock(labels, statements, srcLocation, precedingWhitespace) =>
        val res = ir.SwitchCaseBlock(labels.map(visitSwitchLabel),
                                      statements.map(visitStatement),
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitSwitchLabel(node: vercors.SwitchLabel): ir.SwitchLabel = {
    node match {
      case vercors.SwitchLabel(expression, srcLocation, precedingWhitespace) =>
        val irExpr = expression.map(visitExpression)
        val res = ir.SwitchLabel(irExpr,
                                  srcLocation,
                                  precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitSynchronizedStatement(node: vercors.SynchronizedStatement): ir.SynchronizedStatement = {
    node match {
      case vercors.SynchronizedStatement(condition, body, srcLocation, precedingWhitespace) =>
        val res = ir.SynchronizedStatement(visitExpression(condition),
                                            visitStatement(body),
                                            srcLocation,
                                            precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitReturnStatement(node: vercors.ReturnStatement): ir.ReturnStatement = {
    node match {
      case vercors.ReturnStatement(expression, srcLocation, precedingWhitespace) =>
        val irExpr = expression.map(visitExpression)
        val res = ir.ReturnStatement(irExpr,
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitThrowStatement(node: vercors.ThrowStatement): ir.ThrowStatement = {
    node match {
      case vercors.ThrowStatement(expression, srcLocation, precedingWhitespace) =>
        val res = ir.ThrowStatement(visitExpression(expression),
                                    srcLocation,
                                    precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitBreakStatement(node: vercors.BreakStatement): ir.BreakStatement = {
    node match {
      case vercors.BreakStatement(label, srcLocation, precedingWhitespace) =>
        val irLabel = label.map(visitStringNode)
        val res = ir.BreakStatement(irLabel,
                                    srcLocation,
                                    precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitContinueStatement(node: vercors.ContinueStatement): ir.ContinueStatement = {
    node match {
      case vercors.ContinueStatement(label, srcLocation, precedingWhitespace) =>
        val irLabel = label.map(visitStringNode)
        val res = ir.ContinueStatement(irLabel,
                                        srcLocation,
                                        precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitEmptyStatement(node: vercors.EmptyStatement): ir.EmptyStatement = {
    val res = new ir.EmptyStatement(node.srcLocation)
    res.copyWhitespace(node)
    res
  }
  def visitExpressionStatement(node: vercors.ExpressionStatement): ir.ExpressionStatement = {
    node match {
      case vercors.ExpressionStatement(expression, srcLocation, precedingWhitespace) =>
        val res = ir.ExpressionStatement(visitExpression(expression),
                                          srcLocation,
                                          precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitLabeledStatement(node: vercors.LabeledStatement): ir.LabeledStatement = {
    node match {
      case vercors.LabeledStatement(label, statement, srcLocation, precedingWhitespace) =>
        val res = ir.LabeledStatement(visitLabel(label),
                                      visitStatement(statement),
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitLabel(node: vercors.Label): ir.Label = {
    node match {
      case vercors.Label(label, srcLocation, precedingWhitespace) =>
        val res = ir.Label(visitStringNode(label),
                            srcLocation,
                            precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitArguments(node: vercors.Arguments): ir.Arguments = {
    node match {
      case vercors.Arguments(arguments, srcLocation, precedingWhitespace) =>
        val res = ir.Arguments(arguments.map(visitExpression),
                                srcLocation,
                                precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }


  def visitExpression(node: vercors.Expression): ir.Expression = {
    node match {
      case expr: vercors.NestedInvocationExpression => visitNestedInvocationExpression(expr)
      case expr: vercors.BinaryExpression => visitBinaryExpression(expr)
      case expr: vercors.InstanceofExpression => visitInstanceofExpression(expr)
      case expr: vercors.ArrayIndexExpression => visitArrayIndexExpression(expr)
      case expr: vercors.MethodCallExpression => visitMethodCallExpression(expr)
      case expr: vercors.NewObjectExpression => visitNewObjectExpression(expr)
      case expr: vercors.NewArrayExpression => visitNewArrayExpression(expr)
      case expr: vercors.CastExpression => visitCastExpression(expr)
      case expr: vercors.PostfixExpression => visitPostfixExpression(expr)
      case expr: vercors.PrefixExpression => visitPrefixExpression(expr)
      case expr: vercors.TernaryExpression => visitTernaryExpression(expr)
      case expr: vercors.LambdaExpression => visitLambdaExpression(expr)
      case expr: vercors.MethodReferenceExpression => visitMethodReferenceExpression(expr)
      case expr: vercors.ParExpression => visitParExpression(expr)
      case expr: vercors.Literal => visitLiteral(expr)
      case expr: vercors.ClassName => visitClassName(expr)
      case expr: vercors.GenericInvocationExpression => visitGenericInvocationExpression(expr)
      case expr: vercors.ResultExpression => visitResultExpression(expr)
      case expr: vercors.VerCorsSpecialVariable => visitVerCorsSpecialVariable(expr)
      case expr: vercors.OldExpression => visitOldExpression(expr)
      case expr: vercors.TypeOfExpression => visitTypeOfExpression(expr)
      case expr: vercors.TypeExpression => visitTypeExpression(expr)
      case expr: vercors.ForallExpression => visitForallExpression(expr)
      case expr: vercors.ForallStarExpression => visitForallStarExpression(expr)
      case expr: vercors.ExistsExpression => visitExistsExpression(expr)
      case expr: vercors.LetExpression => visitLetExpression(expr)
      case expr: vercors.StringNode => visitStringNode(expr)
      case expr: vercors.ArrayExpression => visitArrayExpression(expr)
      case expr: vercors.ValuedPermissionPredicate => visitValuedPermissionPredicate(expr)
      case expr: vercors.SomePermissionPredicate => visitSomePermissionPredicate(expr)
      case expr: vercors.CurrentPermissionExpression => visitCurrentPermissionExpression(expr)
      case expr: vercors.ScaledPredicateExpression => visitScaledPredicateExpression(expr)
      case expr: vercors.PermissionPredicate => visitPermissionPredicate(expr)
      case expr: vercors.InlinePattern => visitInlinePattern(expr)
      case expr: vercors.UnfoldingExpression => visitUnfoldingExpression(expr)
      case expr: vercors.CardinalityExpression => visitCardinalityExpression(expr)
      case expr: vercors.OptionSomeExpression => visitOptionSomeExpression(expr)
      case expr: vercors.VerCorsSetComprehension => visitVerCorsSetComprehension(expr)
      case expr: vercors.TypedSetLiteral => visitTypedSetLiteral(expr)
      case expr: vercors.TypedSeqLiteral => visitTypedSeqLiteral(expr)
      case expr: vercors.TypedBagLiteral => visitTypedBagLiteral(expr)
      case _ => ???
    }
    //todo: more?

  }

  def visitNestedInvocationExpression(node: vercors.NestedInvocationExpression): ir.NestedInvocationExpression = {
    node match {
      case vercors.NestedInvocationExpression(left, right, srcLocation, precedingWhitespace) =>
        val res = ir.NestedInvocationExpression(visitExpression(left),
                                                visitExpression(right),
                                                srcLocation,
                                                precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitBinaryExpression(node: vercors.BinaryExpression): ir.BinaryExpression = {
    node match {
      case vercors.BinaryExpression(left, operator, right, srcLocation, precedingWhitespace) =>
        val res = ir.BinaryExpression(visitExpression(left),
                                      visitStringNode(operator),
                                      visitExpression(right),
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitInstanceofExpression(node: vercors.InstanceofExpression): ir.InstanceofExpression = {
    node match {
      case vercors.InstanceofExpression(left, right, srcLocation, precedingWhitespace) =>
        val res = ir.InstanceofExpression(visitExpression(left),
          visitTypeIndication(right),
          srcLocation)
        res.copyWhitespace(node)
        res
    }
  }

  def visitArrayIndexExpression(node: vercors.ArrayIndexExpression): ir.ArrayIndexExpression = {
    node match {
      case vercors.ArrayIndexExpression(arrayExpression, indexExpression, srcLocation, precedingWhitespace) =>
        val idx: ir.IndexExpression = indexExpression match {
          case vercors.PartialRangeExpression(start, end, iSrcLocation, iPrecedingWhitespace) =>
            if(start.isEmpty) {
              val res = ir.RangeExpression(ir.Literal(ir.StringNode("0", iSrcLocation), iSrcLocation),
                visitExpression(end.get),
                iSrcLocation)
              res.copyWhitespace(indexExpression)
              res
            } else if(end.isEmpty) {
              val irEnd = ir.NestedInvocationExpression(visitExpression(arrayExpression), ir.StringNode("size", iSrcLocation), iSrcLocation)
              val res = ir.RangeExpression(visitExpression(start.get), irEnd, iSrcLocation)
              res.copyWhitespace(indexExpression)
              res
            } else {
              val res = ir.RangeExpression(visitExpression(start.get), visitExpression(end.get), iSrcLocation)
              res.copyWhitespace(indexExpression)
              res
            }
          case _ => visitIndexExpression(indexExpression)
        }
        val res = ir.ArrayIndexExpression(visitExpression(arrayExpression),
                                          idx,
                                          srcLocation,
                                          precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitIndexExpression(node: vercors.IndexExpression): ir.IndexExpression = {
    node match {
      case idx: vercors.IndexAllExpression => visitIndexAllExpression(idx)
      case idx: vercors.RangeExpression => visitRangeExpression(idx)
      case idx: vercors.Expression => visitExpression(idx)
      case _ => ???
    }
  }

  def visitIndexAllExpression(node: vercors.IndexAllExpression): ir.IndexAllExpression = {
    node match {
      case vercors.IndexAllExpression(srcLocation, precedingWhitespace) =>
        val res = ir.IndexAllExpression(srcLocation, precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitRangeExpression(node: vercors.RangeExpression): ir.RangeExpression = {
    node match {
      case vercors.RangeExpression(start, end, srcLocation, precedingWhitespace) =>
        val res = ir.RangeExpression(visitExpression(start),
                                      visitExpression(end),
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitMethodCallExpression(node: vercors.MethodCallExpression): ir.MethodCallExpression = {
    node match {
      case vercors.MethodCallExpression(name, arguments, srcLocation, precedingWhitespace) =>
        val res = ir.MethodCallExpression(visitStringNode(name),
                                          visitArguments(arguments),
                                          srcLocation,
                                          precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitNewObjectExpression(node: vercors.NewObjectExpression): ir.NewObjectExpression = {
    node match {
      case vercors.NewObjectExpression(typeArguments, typeIndication, arguments, body, srcLocation, precedingWhitespace) =>
        val typeArgs = typeArguments.map(visitTypeArgumentClause)
        val irBody = body.map(visitBody)
        val res = ir.NewObjectExpression(typeArgs,
                                          visitTypeIndication(typeIndication),
                                          visitArguments(arguments),
                                          irBody,
                                          srcLocation,
                                          precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitNewArrayExpression(node: vercors.NewArrayExpression): ir.NewArrayExpression = {
    node match {
      case vercors.NewArrayExpression(typeIndication, dimensions, initializer, srcLocation, precedingWhitespace) =>
        val irInit = initializer.map(visitArrayInitializer)
        val res = ir.NewArrayExpression(visitTypeIndication(typeIndication),
                                        dimensions.map(visitDimension),
                                        irInit,
                                        srcLocation,
                                        precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitDimension(node: vercors.Dimension): ir.Dimension = {
    node match {
      case vercors.Dimension(sizeExpression, srcLocation, precedingWhitespace) =>
        val irSize = sizeExpression.map(visitExpression)
        val res = ir.Dimension(irSize,
                                srcLocation,
                                precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitCastExpression(node: vercors.CastExpression): ir.CastExpression = {
    node match {
      case vercors.CastExpression(annotations, typeIndications, expression, srcLocation, precedingWhitespace) =>
        val res = ir.CastExpression(annotations.map(visitStringAnnotation),
                                    typeIndications.map(visitTypeIndication),
                                    visitExpression(expression),
                                    srcLocation,
                                    precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitPostfixExpression(node: vercors.PostfixExpression): ir.PostfixExpression = {
    node match {
      case vercors.PostfixExpression(expression, postfix, srcLocation, precedingWhitespace) =>
        val res = ir.PostfixExpression(visitExpression(expression),
                                        visitStringNode(postfix),
                                        srcLocation,
                                        precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitPrefixExpression(node: vercors.PrefixExpression): ir.PrefixExpression = {
    node match {
      case vercors.PrefixExpression(expression, prefix, srcLocation, precedingWhitespace) =>
        val res = ir.PrefixExpression(visitExpression(expression),
                                      visitStringNode(prefix),
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitTernaryExpression(node: vercors.TernaryExpression): ir.TernaryExpression = {
    node match {
      case vercors.TernaryExpression(condition, thenExpression, elseExpression, srcLocation, precedingWhitespace) =>
        val res = ir.TernaryExpression(visitExpression(condition),
                                        visitExpression(thenExpression),
                                        visitExpression(elseExpression),
                                        srcLocation,
                                        precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitLambdaExpression(node: vercors.LambdaExpression): ir.LambdaExpression = {
    node match {
      case vercors.LambdaExpression(hasParentheses, parameters, body, srcLocation, precedingWhitespace) =>
        val res = ir.LambdaExpression(hasParentheses,
                                      parameters.map(visitParameter),
                                      visitLambdaBody(body),
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitLambdaBody(node: vercors.LambdaBody): ir.LambdaBody = {
    node match {
      case body: vercors.Expression => visitExpression(body)
      case _ => ???
    }
  }

  def visitMethodReferenceExpression(node: vercors.MethodReferenceExpression): ir.MethodReferenceExpression = {
    node match {
      case vercors.MethodReferenceExpression(left, typeArgument, right, srcLocation, precedingWhitespace) =>
        val irArg = typeArgument.map(visitTypeArgumentClause)
        val res = ir.MethodReferenceExpression(visitExpression(left),
                                                irArg,
                                                visitExpression(right),
                                                srcLocation,
                                                precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitParExpression(node: vercors.ParExpression): ir.ParExpression = {
    node match {
      case vercors.ParExpression(expression, srcLocation, precedingWhitespace) =>
        val res = ir.ParExpression(visitExpression(expression),
                                    srcLocation,
                                    precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitLiteral(node: vercors.Literal): ir.Literal = {
    node match {
      case vercors.Literal(value, srcLocation, precedingWhitespace) =>
        val res = ir.Literal(visitStringNode(value),
                              srcLocation,
                              precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitClassName(node: vercors.ClassName): ir.ClassName = {
    node match {
      case vercors.ClassName(typeIndication, srcLocation, precedingWhitespace) =>
        val res = ir.ClassName(visitTypeIndication(typeIndication),
                                srcLocation,
                                precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitGenericInvocationExpression(node: vercors.GenericInvocationExpression): ir.GenericInvocationExpression = {
    node match {
      case vercors.GenericInvocationExpression(typeArgumentClause, name, arguments, srcLocation, precedingWhitespace) =>
        val irArgs = arguments.map(visitArguments)
        val res = ir.GenericInvocationExpression(visitTypeArgumentClause(typeArgumentClause),
                                                  visitStringNode(name),
                                                  irArgs,
                                                  srcLocation,
                                                  precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitJmlModifierBlock(node: vercors.JmlModifierBlock): ir.JmlModifierBlock = {
    node match {
      case vercors.JmlModifierBlock(startSpec, modifiers, endSpec, srcLocation, precedingWhitespace) =>
        val irStart = startSpec.map(visitStringNode)
        val irEnd = endSpec.map(visitStringNode)
        val res = ir.JmlModifierBlock(irStart,
                                      modifiers.map(visitModifier),
                                      irEnd,
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitPureModifier(node: vercors.PureModifier): ir.PureModifier = {
    node match {
      case vercors.PureModifier(srcLocation, precedingWhitespace) =>
        val res = ir.PureModifier(srcLocation, precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitGhostModifier(node: vercors.GhostModifier): ir.GhostModifier = {
    node match {
      case vercors.GhostModifier(srcLocation, precedingWhitespace) =>
        val res = ir.GhostModifier(srcLocation, precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitSpecMethodModifier(node: vercors.SpecMethodModifier): ir.SpecMethodModifier = {
    node match {
      case vercors.SpecMethodModifier(srcLocation, precedingWhitespace) =>
        val res = ir.SpecMethodModifier(srcLocation, precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitInlineModifier(node: vercors.InlineModifier): ir.Modifier = {
    node match {
      case vercors.InlineModifier(srcLocation, precedingWhitespace) =>
        System.err.println(s"[warning] Inline modifier not supported ${srcLocation.toString}, dropped.")
        val res = ir.ModifierStringNode("", srcLocation)
        res.copyWhitespace(node)
        res
    }
  }
  def visitThreadLocalModifier(node: vercors.ThreadLocalModifier): ir.Modifier = {
    node match {
      case vercors.ThreadLocalModifier(srcLocation, precedingWhitespace) =>
        System.err.println(s"[warning] thread_local modifier not supported ${srcLocation.toString}, dropped.")
        val res = ir.ModifierStringNode("", srcLocation)
        res.copyWhitespace(node)
        res
    }
  }

  def visitContract(node: vercors.Contract): ir.Contract = {
    node match {
      case vercors.Contract(clauses, srcLocation, precedingWhitespace) =>
        val res = ir.Contract(clauses.map(visitContractClause),
                              srcLocation,
                              precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def renameGhostParameters(meth: vercors.MethodDeclaration): Unit = {
    var ghosts: List[vercors.GhostIO] = Nil
    meth.contractBefore match {
      case Some(contract) =>
        ghosts = contract.clauses.filter(x => x.isInstanceOf[vercors.GhostIO]).map(x => x.asInstanceOf[vercors.GhostIO])
      case _ =>
    }
    meth.contractWithin match {
      case Some(contract) =>
        ghosts = ghosts.appendedAll(contract.clauses.filter(x => x.isInstanceOf[vercors.GhostIO]).map(x => x.asInstanceOf[vercors.GhostIO]))
      case _ =>
    }
    if(ghosts.nonEmpty) {
      val ghostNames = ghosts.map(x => x.name.text)
      for(v <- variables) {
        if(ghostNames.contains(v.text)) {
          v.text = meth.name.text +"_"+ v.text
        }
      }
    }

  }

  def visitLoopSpecification(node: vercors.LoopSpecification): ir.LoopSpecification = {
    node match {
      case vercors.LoopSpecification(startSpec, clauses, endSpec, srcLocation, precedingWhitespace) =>
        val irStart = startSpec.map(visitStringNode)
        val irEnd = endSpec.map(visitStringNode)
        val res = ir.LoopSpecification(irStart,
                                      clauses.map(visitContractClauseOfLoop),
                                      irEnd,
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitLoopSpecOrLabel(node: vercors.LoopSpecOrLabel): ir.LoopSpecOrLabel = {
    node match {
      case l: vercors.Label => visitLabel(l)
      case spec: vercors.LoopSpecification => visitLoopSpecification(spec)
    }
  }

  def visitContractClause(node: vercors.ContractClause): ir.ContractClause = {
    try
      val res = node match {
        case clause: vercors.Precondition => visitPrecondition(clause)
        case clause: vercors.Postcondition => visitPostcondition(clause)
        case clause: vercors.ExceptionalPostcondition => visitExceptionalPostcondition(clause)
        case clause: vercors.DecreasesClause => visitDecreasesClauseOfMethod(clause)
        case clause: vercors.NonTranslatedContractClause => visitNonTranslatedContractClause(clause)
        case clause: vercors.GhostArgument => visitNonsupportedContractClause(clause, clause.givenKeyword)
        case clause: vercors.GhostReturn => visitNonsupportedContractClause(clause, clause.yieldsKeyword)
        case _ => ???
      }
      res.startSpec = node.startSpec.map(visitStringNode)
      res.endSpec = node.endSpec.map(visitStringNode)
      res
    catch {
      case e: NotSupportedException =>
        val text = Helper.commentUnknownNode(node, s"${e.construct} not supported, whole ContractClause commented out")
        ir.NonTranslatedContractClause(text, node.srcLocation)
    }
  }

  def visitContractClauseOfLoop(node: vercors.ContractClause): ir.LoopClause = {
    try
      node match {
        case inv: vercors.LoopInvariant =>
          visitLoopInvariant(inv)
        case dec: vercors.DecreasesClause =>
          visitDecreasesClause(dec)
        case vercors.NonTranslatedContractClause(text, srcLocation, precedingWhitespace) =>
          val res = ir.NonTranslatedLoopClause(text, srcLocation, precedingWhitespace)
          res.copyWhitespace(node.asInstanceOf[vercors.NonTranslatedContractClause])
          res
        case p: vercors.Precondition =>
          val err_text = "Warning: Loop Preconditions not supported by the Specification Translator. Commented out."
          System.err.println(err_text + p.srcLocation.toString)
          val text = Helper.commentNode(p, p.requiresKeyword, err_text)
          val res = ir.NonTranslatedLoopClause(text, p.srcLocation)
          res
        case p: vercors.Postcondition =>
          val err_text = "Warning: Loop Postconditions not supported by the Specification Translator. Commented out."
          System.err.println(err_text + p.srcLocation.toString)
          val text = Helper.commentNode(p, p.ensuresKeyword, err_text)
          val res = ir.NonTranslatedLoopClause(text, p.srcLocation)
          res
        case c: vercors.Context =>
          val err_text = "Warning: Loop Context not supported by the Specification Translator. Commented out."
          System.err.println(err_text + c.srcLocation.toString)
          val text = Helper.commentNode(c, c.contextKeyword, err_text)
          val res = ir.NonTranslatedLoopClause(text, c.srcLocation)
          res
        case c: vercors.ContextEverywhere =>
          val err_text = "Warning: Loop Context not supported by the Specification Translator. Commented out."
          System.err.println(err_text + c.srcLocation.toString)
          val text = Helper.commentNode(c, c.contextEverywhereKeyword, err_text)
          val res = ir.NonTranslatedLoopClause(text, c.srcLocation)
          res
        case _ =>
          val err_text = "Warning: Unexpected type of loop clause, not supported by the Specification Translator. Commented out."
          System.err.println(err_text + node.srcLocation.toString)
          val text = Helper.commentUnknownNode(node, err_text)
          val res = ir.NonTranslatedLoopClause(text, node.srcLocation)
          res
      }
    catch {
      case e: NotSupportedException =>
        val text = Helper.commentUnknownNode(node, s"${e.construct} not supported, whole LoopClause commented out")
        ir.NonTranslatedLoopClause(text, node.srcLocation)
    }
  }

  def visitNonTranslatedContractClause(node: vercors.NonTranslatedContractClause): ir.NonTranslatedContractClause = {
    node match {
      case vercors.NonTranslatedContractClause(text, srcLocation, precedingWhitespace) =>
        val res = ir.NonTranslatedContractClause(text,
          srcLocation,
          precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitNonsupportedContractClause(node: vercors.ContractClause, firstKeyword: generic.syntaxTree.TreeNode): ir.NonTranslatedContractClause = {
    val err_text = s"Warning: ${node.getClass.getName} not supported by the Specification Translator. Commented out."
    System.err.println(err_text + node.srcLocation.toString)
    val text = Helper.commentNode(node, firstKeyword, err_text)
    val res = ir.NonTranslatedContractClause(text, node.srcLocation)
    res
  }

  def visitPrecondition(node: vercors.Precondition): ir.Precondition = {
    node match {
      case vercors.Precondition(expression, srcLocation, precedingWhitespace) =>
        val res = ir.Precondition(visitExpression(expression),
                                  srcLocation,
                                  precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitPostcondition(node: vercors.Postcondition): ir.Postcondition = {
    node match {
      case vercors.Postcondition(expression, srcLocation, precedingWhitespace) =>
        val res = ir.Postcondition(visitExpression(expression),
                                    srcLocation,
                                    precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitLoopInvariant(node: vercors.LoopInvariant): ir.LoopInvariant = {
    node match {
      case vercors.LoopInvariant(expression, srcLocation, precedingWhitespace) =>
        val res = ir.LoopInvariant(visitExpression(expression),
                                    srcLocation,
                                    precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitDecreasesClause(node: vercors.DecreasesClause): ir.DecreasesClause = {
    node match {
      case vercors.DecreasesClause(expression, srcLocation, precedingWhitespace) =>
        val res = ir.DecreasesClause(visitExpression(expression),
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitDecreasesClauseOfMethod(node: vercors.DecreasesClause): ir.MeasuredByClause = {
    node match {
      case vercors.DecreasesClause(expression, srcLocation, precedingWhitespace) =>
        val res = ir.MeasuredByClause(visitExpression(expression),
                                      None,
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res.measuredByKeyword.precedingWhitespace = node.decreasesKeyword.precedingWhitespace
        res.semicolon.precedingWhitespace = node.semicolon.precedingWhitespace
        res
    }
  }

  def visitExceptionalPostcondition(node: vercors.ExceptionalPostcondition): ir.ExceptionalPostcondition = {
    node match {
      case vercors.ExceptionalPostcondition(exceptionType, exceptionName, expression, srcLocation, precedingWhitespace) =>
        val res = ir.ExceptionalPostcondition(visitTypeIndication(exceptionType),
                                              exceptionName.map(visitStringNode),
                                              expression.map(visitExpression),
                                              srcLocation,
                                              precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

//  def visitGhostArgument(node: vercors.GhostArgument): ir.ContractClause = {
//    node match {
//      case vercors.GhostArgument(argType, name, srcLocation, precedingWhitespace) =>
//
//    }
//  }

  def visitNonTranslatedStatement(node: vercors.NonTranslatedStatement): ir.NonTranslatedStatement = {
    node match {
      case vercors.NonTranslatedStatement(text, srcLocation, precedingWhitespace) =>
        val res = ir.NonTranslatedStatement(text,
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitAssertStatement(node: vercors.AssertStatement): ir.AssertStatement = {
    node match {
      case vercors.AssertStatement(content, message, srcLocation, precedingWhitespace) =>
        val irMsg = message.map(visitExpression)
        val res = ir.AssertStatement(visitExpression(content),
                                      irMsg,
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitAssumeStatement(node: vercors.AssumeStatement): ir.AssumeStatement = {
    node match {
      case vercors.AssumeStatement(content, message, srcLocation, precedingWhitespace) =>
        val irMsg = message.map(visitExpression)
        val res = ir.AssumeStatement(visitExpression(content),
                                      irMsg,
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitFoldStatement(node: vercors.FoldStatement): ir.AssertStatement = {
    node match {
      case vercors.FoldStatement(expression, srcLocation, precedingWhitespace) =>
        val res = ir.AssertStatement(visitExpression(expression),
                                    None,
                                    srcLocation,
                                    precedingWhitespace)
        res.copyWhitespace(node)
        res.assertKeyword.copyWhitespace(node.foldKeyword)
        res
    }
  }
  def visitUnfoldStatement(node: vercors.UnfoldStatement): ir.AssertStatement = {
    node match {
      case vercors.UnfoldStatement(expression, srcLocation, precedingWhitespace) =>
        val res = ir.AssertStatement(visitExpression(expression),
                                    None,
                                    srcLocation,
                                    precedingWhitespace)
        res.copyWhitespace(node)
        res.assertKeyword.copyWhitespace(node.unfoldKeyword)
        res
    }
  }
  def visitGhostStatement(node: vercors.GhostStatement): ir.GhostStatement = {
    node match {
      case vercors.GhostStatement(statement, srcLocation, precedingWhitespace) =>
        val res = ir.GhostStatement(visitStatement(statement), srcLocation)
        res.copyWhitespace(node)
        res
    }
  }

  def visitJmlStatementBlock(node: vercors.JmlStatementBlock): ir.JmlStatementBlock = {
    node match {
      case vercors.JmlStatementBlock(startSpec, statements, endSpec, srcLocation, precedingWhitespace) =>
        val irStart = startSpec.map(visitStringNode)
        val irEnd = endSpec.map(visitStringNode)
        val res = ir.JmlStatementBlock(irStart,
                                        statements.map(visitStatement),
                                        irEnd,
                                        srcLocation,
                                        precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitResultExpression(node: vercors.ResultExpression): ir.ResultExpression = {
    node match {
      case vercors.ResultExpression(src, precedingWhitespace) =>
        val res = ir.ResultExpression(src, precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitVerCorsSpecialVariable(node: vercors.VerCorsSpecialVariable): ir.Expression = {
    node match {
      case vercors.VerCorsSpecialVariable(text, srcLocation, precedingWhitespace) =>
        System.err.println(s"[Warning] VerCors special variable $text cannot be represented in other languages")
        // todo: comment out
        val res = ir.StringNode(text, srcLocation, precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitOldExpression(node: vercors.OldExpression): ir.OldExpression = {
    node match {
      case vercors.OldExpression(expression, label, srcLocation, precedingWhitespace) =>
        val irLabel = label.map(visitStringNode)
        val res = ir.OldExpression(visitExpression(expression),
                                    irLabel,
                                    srcLocation,
                                    precedingWhitespace)
        res.copyWhitespace(node)
        if(node.label.isDefined){
          res.oldKeyword.trailingWhitespace += node.openBracket.precedingWhitespace
          res.label.get.precedingWhitespace = node.openBracket.trailingWhitespace + res.label.get.precedingWhitespace
          res.label.get.trailingWhitespace += node.closeBracket.precedingWhitespace
          res.openParenthesis.precedingWhitespace = node.closeBracket.trailingWhitespace + res.openParenthesis.precedingWhitespace
        }
        res
    }
  }

  def visitTypeOfExpression(node: vercors.TypeOfExpression): ir.TypeOfExpression = {
    node match {
      case vercors.TypeOfExpression(expression, srcLocation, precedingWhitespace) =>
        val res = ir.TypeOfExpression(visitExpression(expression),
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitTypeExpression(node: vercors.TypeExpression): ir.TypeExpression = {
    node match {
      case vercors.TypeExpression(indication, srcLocation, precedingWhitespace) =>
        val res = ir.TypeExpression(visitTypeIndication(indication),
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }


  def visitForallExpression(node: vercors.ForallExpression): ir.ForallExpression = {
    node match {
      case vercors.ForallExpression(declarations, guard, expression, srcLocation, precedingWhitespace) =>
        val irGuard = guard.map(visitExpression)
        val res = ir.ForallExpression(declarations.map(visitQuantifiedVarDeclaration),
                                      irGuard,
                                      visitExpression(expression),
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitForallStarExpression(node: vercors.ForallStarExpression): ir.ForallWithPermissionsExpression = {
    node match {
      case vercors.ForallStarExpression(declarations, guard, expression, srcLocation, precedingWhitespace) =>
        val irGuard = guard.map(visitExpression)
        val res = ir.ForallWithPermissionsExpression(declarations.map(visitQuantifiedVarDeclaration),
                                      irGuard,
                                      visitExpression(expression),
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }
  def visitExistsExpression(node: vercors.ExistsExpression): ir.ExistsExpression = {
    node match {
      case vercors.ExistsExpression(declarations, guard, expression, srcLocation, precedingWhitespace) =>
        val irGuard = guard.map(visitExpression)
        val res = ir.ExistsExpression(declarations.map(visitQuantifiedVarDeclaration),
                                      irGuard,
                                      visitExpression(expression),
                                      srcLocation,
                                      precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitLetExpression(node: vercors.LetExpression): ir.Expression = {
    throw new NotSupportedException("Let-Expression", node.srcLocation)
  }


  def visitQuantifiedVarDeclaration(node: vercors.QuantifiedVarDeclaration): ir.QuantifiedVarDeclaration = {
    node match {
      case vercors.QuantifiedVarDeclaration(declType, declarators, srcLocation, precedingWhitespace) =>
        val res = ir.QuantifiedVarDeclaration(visitTypeIndication(declType),
                                              declarators.map(visitStringNode),
                                              srcLocation,
                                              precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitArrayExpression(node: vercors.ArrayExpression): ir.Expression = {
    node match {
      case vercors.ArrayExpression(name, size, srcLocation, precedingWhitespace) =>
        val notnull = ir.BinaryExpression(visitExpression(name),
          ir.StringNode("!=", srcLocation),
          ir.StringNode("null", srcLocation),
          srcLocation)
        val dot = ir.NestedInvocationExpression(visitExpression(name),
          ir.StringNode("length", srcLocation),
          srcLocation)
        val sizeExpr = ir.BinaryExpression(dot,
          ir.StringNode("==", srcLocation),
          visitExpression(size),
          srcLocation, " ")
        val res = ir.BinaryExpression(notnull,
          ir.StringNode("&&", srcLocation, " "),
          sizeExpr,
          srcLocation, " ")
        res.copyWhitespace(node)
        res.precedingWhitespace = res.precedingWhitespace + node.arrayKeyword.precedingWhitespace
        res
    }
  }
  
  def visitPermissionPredicate(node: vercors.PermissionPredicate): ir.Expression = {
    val desugared: vercors.Expression = node.target match {
      case idx: vercors.ArrayIndexExpression =>
        if(idx.indexExpression.isInstanceOf[vercors.IndexAllExpression]) {
          val arrayRef = idx.arrayExpression
          val intType = vercors.StringTypeIndication("int", Nil, "int", node.srcLocation, " ")
          val variable = vercors.StringNode("i", node.srcLocation, " ")
          val decl = vercors.QuantifiedVarDeclaration(intType, List(variable.copy()), node.srcLocation)
          val lowerBound = vercors.BinaryExpression(vercors.StringNode("0",node.srcLocation),
            vercors.StringNode("<=",node.srcLocation),
            variable.copy(),
            node.srcLocation)
          val length = vercors.BinaryExpression(arrayRef, vercors.StringNode(".", node.srcLocation),
            vercors.StringNode("length", node.srcLocation), node.srcLocation)
          val upperBound = vercors.BinaryExpression(variable.copy(),
            vercors.StringNode("<",node.srcLocation),
            length,
            node.srcLocation)
          val guard = Some(vercors.BinaryExpression(lowerBound, vercors.StringNode("&&", node.srcLocation),
            upperBound, node.srcLocation))
          val newTarget = vercors.ArrayIndexExpression(idx.arrayExpression, variable.copy(),
            idx.srcLocation, idx.precedingWhitespace)
          newTarget.trailingWhitespace = idx.trailingWhitespace
          val perm = vercors.PermissionPredicate(newTarget, node.amount,
            node.srcLocation, node.precedingWhitespace)
          perm.trailingWhitespace = node.trailingWhitespace
          vercors.ForallStarExpression(List(decl), guard, perm, node.srcLocation)
        } else {
          node
        }
      case _ => node
    }

    desugared match {
      case vercors.PermissionPredicate(target, amount, srcLocation, precedingWhitespace) =>
        val res = ir.PermissionExpression(visitExpression(target),
                                          amount.map(visitExpression),
                                          srcLocation,
                                          precedingWhitespace)
        res.copyWhitespace(node)
        res.precedingWhitespace += node.permKeyword.precedingWhitespace + node.permKeyword.trailingWhitespace
                                + node.openParenthesis.precedingWhitespace + node.openParenthesis.trailingWhitespace
        res.target.trailingWhitespace += node.comma.precedingWhitespace + node.comma.trailingWhitespace
        res.trailingWhitespace = node.closeParenthesis.precedingWhitespace + node.closeParenthesis.trailingWhitespace
                              + res.trailingWhitespace
        res
      case forall: vercors.ForallStarExpression =>
        visitForallStarExpression(forall)
    }
  }

  def visitValuedPermissionPredicate(node: vercors.ValuedPermissionPredicate): ir.ValuedPermissionExpression = {
    node match {
      case vercors.ValuedPermissionPredicate(target, value, amount, srcLocation, precedingWhitespace) =>
        val res = ir.ValuedPermissionExpression(visitExpression(target),
                                          visitExpression(value),
                                          Some(visitExpression(amount.get)),
                                          srcLocation,
                                          precedingWhitespace)
        res.copyWhitespace(node)
        res.precedingWhitespace += node.pointsToKeyword.precedingWhitespace + node.pointsToKeyword.trailingWhitespace
                                + node.openParenthesis.precedingWhitespace + node.openParenthesis.trailingWhitespace
        res.target.trailingWhitespace += node.firstComma.precedingWhitespace + node.firstComma.trailingWhitespace
        res.value.trailingWhitespace += node.secondComma.precedingWhitespace + node.secondComma.trailingWhitespace
        res.trailingWhitespace = node.closeParenthesis.precedingWhitespace + node.closeParenthesis.trailingWhitespace
                              + res.trailingWhitespace
        res
    }
  }

  def visitCurrentPermissionExpression(node: vercors.CurrentPermissionExpression): ir.CurrentPermissionExpression = {
    node match {
      case vercors.CurrentPermissionExpression(target, srcLocation, precedingWhitespace) =>
        val res = ir.CurrentPermissionExpression(visitExpression(target), srcLocation)
        res.copyWhitespace(node)
        res.precedingWhitespace += node.permKeyword.precedingWhitespace + node.permKeyword.trailingWhitespace
                                + node.openParenthesis.precedingWhitespace + node.openParenthesis.trailingWhitespace
        res.trailingWhitespace = node.closeParenthesis.precedingWhitespace + node.closeParenthesis.trailingWhitespace
                               + res.trailingWhitespace

        res
    }
  }

  def visitSomePermissionPredicate(node: vercors.SomePermissionPredicate): ir.SomePermissionPredicate = {
    node match {
      case vercors.SomePermissionPredicate(target, srcLocation, precedingWhitespace) =>
        val res = ir.SomePermissionPredicate(visitExpression(target), srcLocation)
        res.copyWhitespace(node)
        res.precedingWhitespace += node.valueKeyword.precedingWhitespace + node.valueKeyword.trailingWhitespace
                                + node.openParenthesis.precedingWhitespace + node.openParenthesis.trailingWhitespace
        res.trailingWhitespace = node.closeParenthesis.precedingWhitespace + node.closeParenthesis.trailingWhitespace
                              + res.trailingWhitespace

        res
    }
  }

  def visitScaledPredicateExpression(node: vercors.ScaledPredicateExpression): ir.ScaledPredicateExpression = {
    node match {
      case vercors.ScaledPredicateExpression(amount, predicate, srcLocation, precedingWhitespace) =>
        val res = ir.ScaledPredicateExpression(visitExpression(amount),
                                          visitExpression(predicate),
                                          srcLocation,
                                          precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitInlinePattern(node: vercors.InlinePattern): ir.Expression = {
    node match {
      case vercors.InlinePattern(openPattern, expression, srcLocation, precedingWhitespace) =>
        val res = visitExpression(expression)
        res.precedingWhitespace += openPattern.precedingWhitespace + openPattern.trailingWhitespace
        res.trailingWhitespace = node.closePattern.precedingWhitespace + node.closePattern.trailingWhitespace
                              + res.trailingWhitespace
        res
    }
  }

  def visitUnfoldingExpression(node: vercors.UnfoldingExpression): ir.UnfoldingExpression = {
    node match {
      case vercors.UnfoldingExpression(predicate, expression, srcLocation, precedingWhitespace) =>
        val res = ir.UnfoldingExpression(visitExpression(predicate),
                                          visitExpression(expression),
                                          srcLocation,
                                          precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitCardinalityExpression(node: vercors.CardinalityExpression): ir.Expression = {
    node match {
      case vercors.CardinalityExpression(expression, srcLocation, precedingWhitespace) =>
        val res = ir.NestedInvocationExpression(visitExpression(expression), ir.StringNode("size", srcLocation), srcLocation, precedingWhitespace)
        res.precedingWhitespace += node.leftBar.precedingWhitespace + node.leftBar.trailingWhitespace
        res.trailingWhitespace = node.rightBar.precedingWhitespace + node.rightBar.trailingWhitespace + node.trailingWhitespace
        res
    }
  }

  def visitOptionSomeExpression(node: vercors.OptionSomeExpression): ir.Expression = {
    val err_text = s"Warning: VerCors' Option type(${node.srcLocation.toString}) not supported by the Specification Translator. "
          + "Definition commented out, some usages may require manual intervention."
    System.err.println(err_text)
    throw new NotSupportedException("VerCors' Option type", node.srcLocation)
  }

  def visitVerCorsSetComprehension(node: vercors.VerCorsSetComprehension): ir.Expression = {
    System.err.println(s"[warning] VerCors set comprehension${node.srcLocation.toString} not supported by the Specification Translator, commented out")
    throw new NotSupportedException("VerCors set comprehension", node.srcLocation)
  }

  def visitTypedBagLiteral(node: vercors.TypedBagLiteral): ir.Expression = {
    val err_text = s"Warning: VerCors' \"bag\" type(${node.srcLocation.toString}) not supported by the Specification Translator. "
      + "Definition commented out, some usages may require manual intervention."
    System.err.println(err_text)
    throw new NotSupportedException("VerCors' \"seq\" type", node.srcLocation)
  }

  def visitTypedSeqLiteral(node: vercors.TypedSeqLiteral): ir.Expression = {
    val err_text = s"Warning: VerCors' \"seq\" type(${node.srcLocation.toString}) not supported by the Specification Translator. "
      + "Definition commented out, some usages may require manual intervention."
    System.err.println(err_text)
    throw new NotSupportedException("VerCors' \"seq\" type", node.srcLocation)
  }

  def visitTypedSetLiteral(node: vercors.TypedSetLiteral): ir.Expression = {
    val err_text = s"Warning: VerCors' \"set\" type(${node.srcLocation.toString}) not supported by the Specification Translator. "
      + "Definition commented out, some usages may require manual intervention."
    System.err.println(err_text)
    throw new NotSupportedException("VerCors' \"set\" type", node.srcLocation)
  }

  def visitSpecDeclarationBlock(node: vercors.SpecDeclarationBlock): ir.SpecDeclarationBlock = {
    node match {
      case vercors.SpecDeclarationBlock(startSpec, declarations, endSpec, srcLocation, precedingWhitespace) =>
        val res = ir.SpecDeclarationBlock(startSpec.map(visitStringNode),
                                          declarations.map(visitSpecDeclaration),
                                          endSpec.map(visitStringNode),
                                          srcLocation,
                                          precedingWhitespace)
        res.copyWhitespace(node)
        res
    }
  }

  def visitSpecDeclaration(node: vercors.SpecDeclaration): ir.SpecDeclaration = {
    try
      node match {
        case decl: vercors.SpecMethodDeclaration => visitSpecMethodDeclaration(decl)
        case decl: vercors.GhostDeclaration => visitGhostDeclaration(decl)
        case decl: vercors.NonTranslatedSpecDeclaration => visitNonTranslatedSpecDeclaration(decl)
        case decl: vercors.PredicateDeclaration => visitPredicateDeclaration(decl)
        case _ => ???
      }
    catch {
      case e: NotSupportedException =>
        val text = Helper.commentUnknownNode(node, s"${e.construct} not supported, whole SpecDeclaration commented out")
        ir.NonTranslatedSpecDeclaration(text, node.srcLocation)
    }
  }

  def visitNonTranslatedSpecDeclaration(decl: vercors.NonTranslatedSpecDeclaration): ir.SpecDeclaration = {
    decl match {
      case vercors.NonTranslatedSpecDeclaration(text, srcLocation, precedingWhitespace) =>
        ir.NonTranslatedSpecDeclaration(text, srcLocation, precedingWhitespace)
    }
  }
  def visitSpecMethodDeclaration(node: vercors.SpecMethodDeclaration): ir.SpecMethodDeclaration = {
    node match {
      case vercors.SpecMethodDeclaration(name, returnType, parameters, contractWithin,
                                        body, srcLocation, precedingWhitespace) =>
        val irReturn = returnType.map(visitTypeIndication)
        val res = ir.SpecMethodDeclaration(visitStringNode(name),
                                          irReturn,
                                          parameters.map(visitParameter),
                                          contractWithin.map(visitContract),
                                          visitBlockOrSemicolon(body),
                                          srcLocation,
                                          precedingWhitespace)
        res.contractBefore = node.contractBefore.map(visitContract)
        res.modifiers = node.modifiers.map(visitModifier)
        res.copyWhitespace(node)
        res
    }
  }

  def visitPredicateDeclaration(node: vercors.PredicateDeclaration): ir.PredicateDeclaration = {
    node match {
      case vercors.PredicateDeclaration(name, params, expr, srcLocation, precedingWhitespace) =>
        val res = ir.PredicateDeclaration(visitStringNode(name),
                                        visitTypeIndication(node.tType),
                                        params.map(visitParameter),
                                        expr.map(visitExpression),
                                        srcLocation)
        res.copyWhitespace(node)
        res
    }
  }
}
