package languages.VerCors

import generated.vercors.{JavaParser, JavaParserBaseVisitor}
import generic.{SourceLocation, NotSupportedException}
import languages.VerCors.syntaxTree.*
import org.antlr.v4.runtime.tree.{ParseTree, TerminalNode}
import org.antlr.v4.runtime.{CommonTokenStream, ParserRuleContext, Token}

import scala.jdk.CollectionConverters.*

class VerCorsParserVisitor(val tokens: CommonTokenStream, val filename: String)
      extends JavaParserBaseVisitor[VerCorsTreeNode] {

  val printer: ParserPrintVisitor = new ParserPrintVisitor(tokens)
  
  def getWhitespaceLeft(ctx: ParserRuleContext): String = {
    val hiddenTokens = tokens.getHiddenTokensToLeft(ctx.start.getTokenIndex)
    if (hiddenTokens == null) "" else hiddenTokens.asScala.map(wsToString).mkString
  }

  def wsToString(token: Token): String = {
    if (token.getType == JavaParser.ExtraAt) {
      token.getText.replace('@', ' ')
    } else {
      token.getText
    }
  }

  def getWhitespaceRight(ctx: ParserRuleContext): String = {
    val hiddenTokens = tokens.getHiddenTokensToRight(ctx.stop.getTokenIndex)
    if (hiddenTokens == null) "" else hiddenTokens.asScala.map(_.getText).mkString
  }

  def getWhitespaceLeft(node: TerminalNode): String = {
    val hiddenTokens = tokens.getHiddenTokensToLeft(node.getSymbol.getTokenIndex)
    if (hiddenTokens == null) "" else hiddenTokens.asScala.map(_.getText).mkString
  }

  def getWhitespaceRight(node: TerminalNode): String = {
    val hiddenTokens = tokens.getHiddenTokensToRight(node.getSymbol.getTokenIndex)
    if (hiddenTokens == null) "" else hiddenTokens.asScala.map(_.getText).mkString
  }

  def getWhitespaceLeft(token: Token): String = {
    val hiddenTokens = tokens.getHiddenTokensToLeft(token.getTokenIndex)
    if (hiddenTokens == null) "" else hiddenTokens.asScala.map(_.getText).mkString
  }

  def addWhitespaceRight[T <: VerCorsTreeNode](node: T, ctx: ParserRuleContext): T = {
    node.trailingWhitespace = getWhitespaceRight(ctx)
    node
  }

  def addWhitespaceRight[T <: VerCorsTreeNode](node: T, terminal: TerminalNode): T = {
    node.trailingWhitespace = getWhitespaceRight(terminal)
    node
  }

  def getSourceLocation(ctx: ParserRuleContext) : SourceLocation = {
    SourceLocation(filename, ctx.start.getLine, ctx.stop.getLine)
  }

  def getSourceLocation(node: TerminalNode) : SourceLocation = {
    SourceLocation(filename, node.getSymbol.getLine)
  }

  def getSourceLocation(token: Token) : SourceLocation = {
    SourceLocation(filename, token.getLine)
  }


  def visit(tree: ParserRuleContext): VerCorsTreeNode = {
    val res = super.visit(tree)
    if(res != null) {
      res
    } else {
      throw new NotSupportedException(tree.getText, getSourceLocation(tree))
    }
  }

  def visitTextNode(ctx: ParserRuleContext): StringNode = {
    val src = getSourceLocation(ctx)
    StringNode(ctx.getText, src, getWhitespaceLeft(ctx))
  }

  def visitTextNodeWithWS(ctx: ParserRuleContext): StringNode = {
    val src = getSourceLocation(ctx)
    StringNode(ctx.getText, src, getWhitespaceLeft(ctx), getWhitespaceRight(ctx))
  }

  override def visitJavaIdentifier(ctx: JavaParser.JavaIdentifierContext): StringNode = {
    visitTextNode(ctx)
  }
  def visitJavaIdentifierWithWS(ctx: JavaParser.JavaIdentifierContext): StringNode = {
    visitTextNodeWithWS(ctx)
  }

  def printNode(ctx: ParserRuleContext): StringNode = {
    val src = getSourceLocation(ctx)
    val text = printer.visitSkipFirstWS(ctx)
    StringNode(text, src, getWhitespaceLeft(ctx))
  }

  def printNodeWithWS(ctx: ParserRuleContext): StringNode = {
    val src = getSourceLocation(ctx)
    val text = printer.visitSkipFirstWS(ctx)
    StringNode(text, src, getWhitespaceLeft(ctx), getWhitespaceRight(ctx))
  }

  override def visitType(ctx: JavaParser.TypeContext): StringTypeIndication = {
    if (ctx.valType() != null) {
      throw new NotSupportedException(ctx.valType().getText, getSourceLocation(ctx))
    } else {
      val text = printer.visitSkipFirstWS(ctx)
      new StringTypeIndication(text, Nil, ctx.getText, getSourceLocation(ctx), getWhitespaceLeft(ctx))
    }
  }
  def visitTypeWithWS(ctx: JavaParser.TypeContext): StringTypeIndication = {
    if (ctx.valType() != null) {
      throw new NotSupportedException(ctx.valType().getText, getSourceLocation(ctx))
    } else {
      val text = printer.visitSkipFirstWS(ctx)
      new StringTypeIndication(text, Nil, ctx.getText, getSourceLocation(ctx),
                                getWhitespaceLeft(ctx), getWhitespaceRight(ctx))
    }
  }

  def visitTypeListAsList(ctx: JavaParser.TypeListContext): List[TypeIndication] = {
    if (ctx.typeList() == null) {
      val onlyType = visitType(ctx.`type`())
      List(onlyType)
    } else {
      val head = visitTypeWithWS(ctx.`type`())
      val tail = visitTypeListAsList(ctx.typeList())
      tail.prepended(head)
    }
  }

  def setNonTranslatedTreeNodeText[T <: generic.syntaxTree.StringNode](node: T,
                                                                       ctx: ParserRuleContext,
                                                                       exc: NotSupportedException): Unit = {
    val ws = getWhitespaceLeft(ctx)
    val asString = new ParserPrintVisitor(tokens,skipFirstHidden = true).visit(ctx)
    node.precedingWhitespace = ws
    node.text = s"// Warning: ${exc.construct} not supported by the Specification Translator. Respective parts commented out." + System.lineSeparator()
                + "// "
    node.text += asString.replace("\n", "\n//").replace("@", " ")
    System.err.println(exc.getMessage)
  }

  override def visitCompilationUnit(ctx: JavaParser.CompilationUnitContext): CompilationUnit = {
    try {
      val src = SourceLocation(filename, ctx.start.getLine, ctx.stop.getLine)
      var pkgDecl: Option[PackageDeclaration] = None
      if (ctx.packageDeclaration() != null) {
        pkgDecl = Some(visitPackageDeclaration(ctx.packageDeclaration()))
      }
      val imports = ctx.importDeclaration().asScala.map(visitImportDeclaration).toList
      val declarations = ctx.typeDeclaration().asScala.map(visitTypeDeclaration).toList
      val res = CompilationUnit(pkgDecl, imports, declarations, src)
      res.trailingWhitespace = getWhitespaceLeft(ctx.EOF())
      res
    } catch {
      case e: NotSupportedException => {
        val res = NonTranslatedCompilationUnit("", getSourceLocation(ctx))
        setNonTranslatedTreeNodeText(res, ctx, e)
        res.trailingWhitespace = getWhitespaceLeft(ctx.EOF())
        res
      }
    }
  }

  override def visitPackageDeclaration(ctx: JavaParser.PackageDeclarationContext): PackageDeclaration = {
    val src = getSourceLocation(ctx)
    val name = visitQualifiedName(ctx.qualifiedName())
    val annotations = ctx.annotation().asScala.map(visitAnnotation).toList
    val pkg = PackageDeclaration(name, annotations, src)
    pkg.packageKeyword.precedingWhitespace = getWhitespaceLeft(ctx.PACKAGE())
    pkg.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
    pkg
  }

  override def visitImportDeclaration(ctx: JavaParser.ImportDeclarationContext): ImportDeclaration = {
    val name = visitQualifiedName(ctx.qualifiedName())
    if (ctx.importAll() != null) {
      name.text += printer.visit(ctx.importAll())
    }
    var modifiers: List[Modifier] = Nil
    if (ctx.STATIC() != null) {
      val static = StaticModifier(getSourceLocation(ctx.STATIC()), getWhitespaceLeft(ctx.STATIC()))
      modifiers = modifiers.appended(static)
    }
    val imp = ImportDeclaration(name, getSourceLocation(ctx))
    imp.modifiers = modifiers
    imp.importKeyword.precedingWhitespace = getWhitespaceLeft(ctx.IMPORT())
    imp.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
    imp
  }

  override def visitTypeDeclaration(ctx: JavaParser.TypeDeclarationContext): TypeDeclaration  = {
    var res: TypeDeclaration = null
    if (ctx.SEMI() != null) {
      res = new EmptyStatement(getSourceLocation(ctx), getWhitespaceLeft(ctx.SEMI()))
    } else {
      try {
        val modifiers = ctx.classOrInterfaceModifier().asScala.map(visitClassOrInterfaceModifier).toList
        if (ctx.classDeclaration() != null) {
          val cls = visitClassDeclaration(ctx.classDeclaration())
          cls.modifiers = modifiers
          cls.srcLocation = getSourceLocation(ctx)
          res = cls
        } else if (ctx.enumDeclaration() != null) {
          val enumDecl = visitEnumDeclaration(ctx.enumDeclaration())
          enumDecl.modifiers = modifiers
          enumDecl.srcLocation = getSourceLocation(ctx)
          res = enumDecl
        } else if (ctx.interfaceDeclaration() != null) {
          val interface = visitInterfaceDeclaration(ctx.interfaceDeclaration())
          interface.modifiers = modifiers
          interface.srcLocation = getSourceLocation(ctx)
          res = interface
        } else if (ctx.annotationTypeDeclaration() != null) {
          val annotationType = visitAnnotationTypeDeclaration(ctx.annotationTypeDeclaration())
          annotationType.modifiers = modifiers
          annotationType.srcLocation = getSourceLocation(ctx)
          res = annotationType
        } else {
          println(s"Unknown type declaration: ${ctx.getText}")
        }
      } catch {
        case e: NotSupportedException =>
          val decl = NonTranslatedTypeDeclaration("", getSourceLocation(ctx))
          setNonTranslatedTreeNodeText(decl, ctx, e)
          res = decl
      }
    }
    res
  }

  override def visitModifier(ctx: JavaParser.ModifierContext): Modifier = {
    if(ctx.classOrInterfaceModifier() != null)
      visitClassOrInterfaceModifier(ctx.classOrInterfaceModifier())
    else if(ctx.valEmbedModifier() != null)
      visitValEmbedModifier(ctx.valEmbedModifier())
    else
      new ModifierStringNode(ctx.getText, getSourceLocation(ctx), getWhitespaceLeft(ctx))
  }

  override def visitClassOrInterfaceModifier(ctx: JavaParser.ClassOrInterfaceModifierContext): Modifier = {
    if (ctx.annotation() != null) {
      visitAnnotation(ctx.annotation())
    } else if(ctx.STATIC() != null) {
      StaticModifier(getSourceLocation(ctx), getWhitespaceLeft(ctx))
    } else {
      new ModifierStringNode(ctx.getText, getSourceLocation(ctx), getWhitespaceLeft(ctx))
    }
  }

  override def visitVariableModifier(ctx: JavaParser.VariableModifierContext): Modifier = {
    if (ctx.annotation() != null) {
      visitAnnotation(ctx.annotation())
    } else {
      new ModifierStringNode(ctx.getText, getSourceLocation(ctx), getWhitespaceLeft(ctx))
    }
  }

  override def visitClassDeclaration(ctx: JavaParser.ClassDeclarationContext): ClassDeclaration = {
    // todo: contract
    val name = visitJavaIdentifier(ctx.javaIdentifier())
    var sup: Option[SuperclassClause] = None
    if (ctx.ext() != null) {
      val ext = SuperclassClause(visitType(ctx.ext().`type`()), getSourceLocation(ctx.ext().`type`()))
      ext.extendsKeyword.precedingWhitespace = getWhitespaceLeft(ctx.ext().EXTENDS())
      sup = Option(ext)
    }
    var interfaces: Option[InterfaceClause] = None
    if (ctx.imp() != null) {
      val int = InterfaceClause(visitTypeListAsList(ctx.imp().typeList()), getSourceLocation(ctx.imp().typeList()))
      int.implementsKeyword.precedingWhitespace = getWhitespaceLeft(ctx.imp().IMPLEMENTS())
      interfaces = Option(int)
    }
    val typeParameterClause = Option(ctx.typeParameters()).map(visitTypeParameters)
    val body = visitClassBody(ctx.classBody())
    val res = ClassDeclaration(name, typeParameterClause, sup, interfaces, body, getSourceLocation(ctx))
    res.classKeyword.precedingWhitespace = getWhitespaceLeft(ctx.CLASS())
    res
  }

  override def visitTypeParameters(ctx: JavaParser.TypeParametersContext): TypeParameterClause = {
    val parameters = visitTypeParameterListAsList(ctx.typeParameterList())
    val clause = TypeParameterClause(parameters, getSourceLocation(ctx))
    clause.openAngle.precedingWhitespace = getWhitespaceLeft(ctx.LT())
    clause.closeAngle.precedingWhitespace = getWhitespaceLeft(ctx.GT())
    clause
  }

  def visitTypeParameterListAsList(ctx: JavaParser.TypeParameterListContext): List[TypeParameter] = {
    if (ctx.typeParameterList() == null) {
      val param = visitTypeParameter(ctx.typeParameter())
      List(param)
    } else {
      val front = visitTypeParameter(ctx.typeParameter())
      val back = visitTypeParameterListAsList(ctx.typeParameterList())
      front.trailingWhitespace = getWhitespaceLeft(ctx.COMMA())
      back.prepended(front)
    }
  }

  override def visitTypeParameter(ctx: JavaParser.TypeParameterContext): TypeParameter = {
    val name = visitJavaIdentifier(ctx.javaIdentifier())
    val typeName = new StringTypeIndication(name.text, Nil, ctx.javaIdentifier().getText, name.srcLocation)
    var superclass: Option[GenericsSuperclassClause] = None
    if (ctx.typeParameterBound() != null) {
      val sup = GenericsSuperclassClause(visitTypeBoundAsList(ctx.typeParameterBound().typeBound()),
                                         getSourceLocation(ctx.typeParameterBound().typeBound()))
      sup.extendsKeyword.precedingWhitespace = getWhitespaceLeft(ctx.typeParameterBound().EXTENDS())
      superclass = Option(sup)
    }
    val param = TypeParameter(Nil, typeName, superclass, getSourceLocation(ctx))
    param
  }

  def visitTypeBoundAsList(ctx: JavaParser.TypeBoundContext): List[TypeIndication] = {
    if (ctx.typeBound() == null) {
      val onlyType = visitType(ctx.`type`())
      List(onlyType)
    } else {
      val head = visitTypeWithWS(ctx.`type`())
      val tail = visitTypeBoundAsList(ctx.typeBound())
      tail.prepended(head)
    }
  }

  override def visitEnumDeclaration(ctx: JavaParser.EnumDeclarationContext): EnumDeclaration = {
    val name = visitJavaIdentifier(ctx.javaIdentifier())
    var interfaces: Option[InterfaceClause] = None
    if (ctx.imp() != null) {
      val int = InterfaceClause(visitTypeListAsList(ctx.imp().typeList()), getSourceLocation(ctx.imp().typeList()))
      int.implementsKeyword.precedingWhitespace = getWhitespaceLeft(ctx.imp().IMPLEMENTS())
      interfaces = Option(int)
    }
    var constants: List[EnumConstant] = Nil
    if (ctx.enumConstants() != null) {
      constants = visitEnumConstantsAsList(ctx.enumConstants())
    }
    var comma: Option[StringNode] = None
    if (ctx.COMMA() != null) {
      comma = Option(StringNode(",", getSourceLocation(ctx.COMMA())))
      comma.get.precedingWhitespace = getWhitespaceLeft(ctx.COMMA())
    }
    var decls: List[MemberDeclaration] = Nil
    var semi: Option[StringNode] = None
    if (ctx.enumBodyDeclarations() != null) {
      semi = Option(StringNode(";", getSourceLocation(ctx.enumBodyDeclarations().SEMI())))
      semi.get.precedingWhitespace = getWhitespaceLeft(ctx.enumBodyDeclarations().SEMI())
      decls = ctx.enumBodyDeclarations().classBodyDeclaration().asScala
                .map(visitClassBodyDeclaration).toList
    }
    val res = EnumDeclaration(name, interfaces, constants, decls, comma, semi, getSourceLocation(ctx))
    res.enumKeyword.precedingWhitespace = getWhitespaceLeft(ctx.ENUM())
    res.openBrace.precedingWhitespace = getWhitespaceLeft(ctx.LBRACE())
    res.closeBrace.precedingWhitespace = getWhitespaceLeft(ctx.RBRACE())
    res
  }

  def visitEnumConstantsAsList(ctx: JavaParser.EnumConstantsContext): List[EnumConstant] = {
    if (ctx.enumConstants() == null) {
      val onlyConst = visitEnumConstant(ctx.enumConstant())
      List(onlyConst)
    } else {
      val last = visitEnumConstant(ctx.enumConstant())
      val front = visitEnumConstantsAsList(ctx.enumConstants())
      front.last.trailingWhitespace += getWhitespaceRight(ctx.enumConstants())
      front.appended(last)
    }
  }

  override def visitEnumConstant(ctx: JavaParser.EnumConstantContext): EnumConstant = {
    val annotations = ctx.annotation().asScala.map(visitAnnotation).toList
    val name = visitJavaIdentifier(ctx.javaIdentifier())
    val args = Option(ctx.arguments()).map(visitArguments)
    val body = Option(ctx.classBody()).map(visitClassBody)
    val constant = EnumConstant(name, annotations, args, body, getSourceLocation(ctx))
    constant
  }

  override def visitInterfaceDeclaration(ctx: JavaParser.InterfaceDeclarationContext): InterfaceDeclaration = {
    val name = visitJavaIdentifier(ctx.javaIdentifier())
    var sup: Option[MultiSuperclassClause] = None
    if (ctx.intExt() != null) {
      val ext = MultiSuperclassClause(visitTypeListAsList(ctx.intExt().typeList()),
                                      getSourceLocation(ctx.intExt().typeList()))
      ext.extendsKeyword.precedingWhitespace = getWhitespaceLeft(ctx.intExt().EXTENDS())
      sup = Option(ext)
    }
    val typeParameterClause = Option(ctx.typeParameters()).map(visitTypeParameters)
    val body = visitInterfaceBody(ctx.interfaceBody())
    val res = InterfaceDeclaration(name, typeParameterClause, sup, body, getSourceLocation(ctx))
    res.interfaceKeyword.precedingWhitespace = getWhitespaceLeft(ctx.INTERFACE())
    res
  }

  override def visitClassBody(ctx: JavaParser.ClassBodyContext): Body = {
    val decls = ctx.classBodyDeclaration().asScala.map(visitClassBodyDeclaration).toList
    val body = Body(decls, getSourceLocation(ctx))
    body.openBrace.precedingWhitespace = getWhitespaceLeft(ctx.LBRACE())
    body.closeBrace.precedingWhitespace = getWhitespaceLeft(ctx.RBRACE())
    body
  }

  override def visitInterfaceBody(ctx: JavaParser.InterfaceBodyContext): Body = {
    val decls = ctx.interfaceBodyDeclaration().asScala.map(visitInterfaceBodyDeclaration).toList
    val body = Body(decls, getSourceLocation(ctx))
    body.openBrace.precedingWhitespace = getWhitespaceLeft(ctx.LBRACE())
    body.closeBrace.precedingWhitespace = getWhitespaceLeft(ctx.RBRACE())
    body
  }

  override def visitClassBodyDeclaration(ctx: JavaParser.ClassBodyDeclarationContext): MemberDeclaration = {
    var decl: MemberDeclaration = null
    if (ctx.SEMI() != null) {
      decl = new EmptyStatement(getSourceLocation(ctx), getWhitespaceLeft(ctx.SEMI()))
    } else if (ctx.block() != null) {
      decl = visitBlock(ctx.block())
      if (ctx.STATIC() != null) {
        val static = StaticModifier(getSourceLocation(ctx.STATIC()), getWhitespaceLeft(ctx.STATIC()))
        decl.modifiers = decl.modifiers.prepended(static)
      }
    } else if (ctx.memberDeclaration() != null) {
      decl = visitMemberDeclaration(ctx.memberDeclaration())
      decl.contractBefore = Option(ctx.valEmbedContract()).map(visitValEmbedContract)
      decl.modifiers = decl.modifiers.prependedAll(ctx.modifier().asScala.map(visitModifier))
    } else if (ctx.valEmbedClassDeclarationBlock() != null) {
      decl = visitValEmbedClassDeclarationBlock(ctx.valEmbedClassDeclarationBlock())
    } else {
      println(s"Error: unexpected alternative in ClassBodyDeclaration: ${ctx.getText}")
    }
    decl.srcLocation = getSourceLocation(ctx)
    decl
  }

  override def visitMemberDeclaration(ctx: JavaParser.MemberDeclarationContext): MemberDeclaration = {
    try {
      if(ctx.methodDeclaration() != null) visitMethodDeclaration(ctx.methodDeclaration())
      else if(ctx.genericMethodDeclaration() != null) visitGenericMethodDeclaration(ctx.genericMethodDeclaration())
      else if(ctx.fieldDeclaration() != null) visitFieldDeclaration(ctx.fieldDeclaration())
      else if(ctx.constructorDeclaration() != null) visitConstructorDeclaration(ctx.constructorDeclaration())
      else if(ctx.genericConstructorDeclaration() != null)
        visitGenericConstructorDeclaration(ctx.genericConstructorDeclaration())
      else if(ctx.interfaceDeclaration() != null)visitInterfaceDeclaration(ctx.interfaceDeclaration())
      else if(ctx.annotationTypeDeclaration() != null) visitAnnotationTypeDeclaration(ctx.annotationTypeDeclaration())
      else if(ctx.classDeclaration() != null) visitClassDeclaration(ctx.classDeclaration())
      else if(ctx.enumDeclaration() != null) visitEnumDeclaration(ctx.enumDeclaration())
      else {
        println(s"Error: unexpected alternative in MemberDeclaration: ${ctx.getText}")
        null
      }
    }catch {
      case e: NotSupportedException =>
        val commented = NonTranslatedMemberDeclaration("", getSourceLocation(ctx))
        setNonTranslatedTreeNodeText(commented, ctx, e)
        commented
    }
  }

  def setMethodDeclarationFields(ctx: JavaParser.MethodDeclarationContext, decl: MethodDeclaration): Unit = {
    decl.name = visitJavaIdentifier(ctx.javaIdentifier())
    decl.returnType = Option(visitTypeOrVoid(ctx.typeOrVoid()))
    if (ctx.formalParameters().formalParameterList() != null) {
      decl.parameters = visitFormalParameterListAsList(ctx.formalParameters().formalParameterList())
    }
    if (ctx.dims() != null) {
      decl.dimensions =  Option(printNode(ctx.dims()))
    }
    if (ctx.throwy() != null) {
      decl.throws = Option(ThrowsClause(visitQualifiedNameListAsList(ctx.throwy().qualifiedNameList()),
                                        getSourceLocation(ctx.throwy().qualifiedNameList())))
      decl.throws.get.throwsKeyword.precedingWhitespace = getWhitespaceLeft(ctx.throwy().THROWS())
    }
    decl.body = visitMethodBodyOrEmpty(ctx.methodBodyOrEmpty())
    decl.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.formalParameters().LPAREN())
    decl.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.formalParameters().RPAREN())
  }

  override def visitMethodDeclaration(ctx: JavaParser.MethodDeclarationContext): MethodDeclaration = {
    val decl = new MethodDeclaration(null, null, Nil, None, None, None, None,
                                     null, getSourceLocation(ctx))
    setMethodDeclarationFields(ctx, decl)
    decl
  }

  override def visitGenericMethodDeclaration(ctx: JavaParser.GenericMethodDeclarationContext):
        GenericMethodDeclaration = {
    val typeParameters = visitTypeParameters(ctx.typeParameters())
    val decl = GenericMethodDeclaration(null, typeParameters, null, Nil, None, None, None, None,
                                        null, getSourceLocation(ctx))
    setMethodDeclarationFields(ctx.methodDeclaration(), decl)
    decl
  }

  def setConstructorDeclarationFields(ctx: JavaParser.ConstructorDeclarationContext,
                                      decl: MethodDeclaration): Unit = {
    decl.name = visitJavaIdentifier(ctx.javaIdentifier())
    if (ctx.formalParameters().formalParameterList() != null) {
      decl.parameters = visitFormalParameterListAsList(ctx.formalParameters().formalParameterList())
    }
    if (ctx.throwy() != null) {
      decl.throws = Option(ThrowsClause(visitQualifiedNameListAsList(ctx.throwy().qualifiedNameList()),
                                        getSourceLocation(ctx.throwy().qualifiedNameList())))
      decl.throws.get.throwsKeyword.precedingWhitespace = getWhitespaceLeft(ctx.throwy().THROWS())
    }
    decl.body = visitBlock(ctx.constructorBody.block())
    decl.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.formalParameters().LPAREN())
    decl.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.formalParameters().RPAREN())
  }
  override def visitGenericConstructorDeclaration(ctx: JavaParser.GenericConstructorDeclarationContext):
                      GenericMethodDeclaration = {
    val typeParameters = visitTypeParameters(ctx.typeParameters())
    val decl = GenericMethodDeclaration(null, typeParameters, None, Nil, None, None, None, None,
                                        null, getSourceLocation(ctx))
    setConstructorDeclarationFields(ctx.constructorDeclaration(), decl)
    decl
  }
  override def visitConstructorDeclaration(ctx: JavaParser.ConstructorDeclarationContext):
              MethodDeclaration = {
    val decl = new MethodDeclaration(null, None, Nil, None, None, None, None,
                                     null, getSourceLocation(ctx))
    setConstructorDeclarationFields(ctx, decl)
    decl
  }

  override def visitFieldDeclaration(ctx: JavaParser.FieldDeclarationContext): FieldDeclaration = {
    val fieldType = visitType(ctx.`type`())
    val declarators = visitVariableDeclaratorsAsList(ctx.variableDeclarators())
    val field = FieldDeclaration(fieldType, declarators, getSourceLocation(ctx))
    field.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
    field
  }

  override def visitInterfaceBodyDeclaration(ctx: JavaParser.InterfaceBodyDeclarationContext):
                MemberDeclaration = {
    if(ctx.interfaceMemberDeclaration() != null) {
      val decl = visitInterfaceMemberDeclaration(ctx.interfaceMemberDeclaration())
      val modifiers = ctx.modifier().asScala.map(visitModifier)
      decl.modifiers = decl.modifiers.prependedAll(modifiers)
      decl.srcLocation = getSourceLocation(ctx)
      decl
    } else if(ctx.valEmbedClassDeclarationBlock() != null) {
      visitValEmbedClassDeclarationBlock(ctx.valEmbedClassDeclarationBlock())
    } else if(ctx.SEMI() != null) {
      new EmptyStatement(getSourceLocation(ctx), getWhitespaceLeft(ctx.SEMI()))
    } else {
      println(s"Error: unexpected alternative in InterfaceBodyDeclaration: ${ctx.getText}")
      null
    }
  }

  override def visitInterfaceMemberDeclaration(ctx: JavaParser.InterfaceMemberDeclarationContext):
                  MemberDeclaration = {
      if(ctx.constDeclaration() != null) visitConstDeclaration(ctx.constDeclaration())
      else if(ctx.interfaceMethodDeclaration() != null)
        visitInterfaceMethodDeclaration(ctx.interfaceMethodDeclaration())
      else if(ctx.genericInterfaceMethodDeclaration() != null)
        visitGenericInterfaceMethodDeclaration(ctx.genericInterfaceMethodDeclaration())
      else if(ctx.interfaceDeclaration() != null) visitInterfaceDeclaration(ctx.interfaceDeclaration())
      else if(ctx.annotationTypeDeclaration() != null) visitAnnotationTypeDeclaration(ctx.annotationTypeDeclaration())
      else if(ctx.classDeclaration() != null) visitClassDeclaration(ctx.classDeclaration())
      else if(ctx.enumDeclaration() != null) visitEnumDeclaration(ctx.enumDeclaration())
      else {
        println(s"Error: unexpected alternative in InterfaceMemberDeclaration: ${ctx.getText}")
        null
    }
  }

  override def visitConstDeclaration(ctx: JavaParser.ConstDeclarationContext): FieldDeclaration = {
    val fieldType = visitType(ctx.`type`())
    val decls = visitConstantDeclaratorListAsList(ctx.constantDeclaratorList())
    val field = FieldDeclaration(fieldType, decls, getSourceLocation(ctx))
    field.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
    field
  }

  def visitConstantDeclaratorListAsList(ctx: JavaParser.ConstantDeclaratorListContext): List[VariableDeclarator] = {
    if (ctx.constantDeclaratorList() == null) {
      val onlyDecl = visitConstantDeclarator(ctx.constantDeclarator())
      List(onlyDecl)
    } else {
      val head = visitConstantDeclarator(ctx.constantDeclarator())
      head.trailingWhitespace = getWhitespaceRight(ctx.constantDeclarator())
      val tail = visitConstantDeclaratorListAsList(ctx.constantDeclaratorList())
      tail.prepended(head)
    }
  }

  override def visitConstantDeclarator(ctx: JavaParser.ConstantDeclaratorContext): VariableDeclarator = {
    val name = visitJavaIdentifier(ctx.javaIdentifier())
    if (ctx.dims() != null) {
      name.text += printNode(ctx.dims()).text
    }
    val initializer = Option(visitVariableInitializer(ctx.variableInitializer()))
    val decl = VariableDeclarator(name, initializer, getSourceLocation(ctx))
    decl.equalSign.precedingWhitespace = getWhitespaceLeft(ctx.ASSIGN())
    decl
  }

  def setInterfaceMethodDeclarationFields(ctx: JavaParser.InterfaceMethodDeclarationContext,
                                          decl: MethodDeclaration): Unit = {
    decl.name = visitJavaIdentifier(ctx.javaIdentifier())
    decl.returnType = Option(visitTypeOrVoid(ctx.typeOrVoid()))
    decl.parameters = if (ctx.formalParameters().formalParameterList() == null) Nil
                      else visitFormalParameterListAsList(ctx.formalParameters().formalParameterList())
    decl.dimensions = Option(ctx.dims()).map(printNode)
    decl.throws = None
    if (ctx.throwy() != null) {
      decl.throws = Option(ThrowsClause(visitQualifiedNameListAsList(ctx.throwy().qualifiedNameList()),
                                        getSourceLocation(ctx.throwy().qualifiedNameList())))
      decl.throws.get.throwsKeyword.precedingWhitespace = getWhitespaceLeft(ctx.throwy().THROWS())
    }
    decl.body = new EmptyStatement(getSourceLocation(ctx), getWhitespaceLeft(ctx.SEMI()))
    decl.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.formalParameters().LPAREN())
    decl.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.formalParameters().RPAREN())
  }

  override def visitInterfaceMethodDeclaration(ctx: JavaParser.InterfaceMethodDeclarationContext):
            MethodDeclaration = {
    val decl = new MethodDeclaration(null, null, Nil, None, None, None, None,
                                     null, getSourceLocation(ctx))
    setInterfaceMethodDeclarationFields(ctx, decl)
    decl
  }

  override def visitGenericInterfaceMethodDeclaration(ctx: JavaParser.GenericInterfaceMethodDeclarationContext):
                GenericMethodDeclaration = {
    val typeParameters = visitTypeParameters(ctx.typeParameters())
    val decl = GenericMethodDeclaration(null, typeParameters, null, Nil, None, None, None, None,
                                        null, getSourceLocation(ctx))
    setInterfaceMethodDeclarationFields(ctx.interfaceMethodDeclaration(), decl)
    decl
  }

  def visitVariableDeclaratorsAsList(ctx: JavaParser.VariableDeclaratorsContext): List[VariableDeclarator] = {
    if (ctx.variableDeclarators() == null) {
      val onlyDecl = visitVariableDeclarator(ctx.variableDeclarator())
      List(onlyDecl)
    } else {
      val head = visitVariableDeclarator(ctx.variableDeclarator())
      head.trailingWhitespace = getWhitespaceRight(ctx.variableDeclarator())
      val tail = visitVariableDeclaratorsAsList(ctx.variableDeclarators())
      tail.prepended(head)
    }
  }

  override def visitVariableDeclarator(ctx: JavaParser.VariableDeclaratorContext): VariableDeclarator = {
    val name = visitVariableDeclaratorId(ctx.variableDeclaratorId())
    val initializer = Option(ctx.variableDeclaratorInit()).map(x=>visitVariableInitializer(x.variableInitializer()))
    val decl = VariableDeclarator(name, initializer, getSourceLocation(ctx))
    if (ctx.variableDeclaratorInit() != null) {
      decl.equalSign.precedingWhitespace = getWhitespaceLeft(ctx.variableDeclaratorInit().ASSIGN())
    }
    decl
  }

  override def visitVariableDeclaratorId(ctx: JavaParser.VariableDeclaratorIdContext): StringNode = {
    printNode(ctx)
  }

  override def visitVariableInitializer(ctx: JavaParser.VariableInitializerContext): Initializer = {
    if(ctx.arrayInitializer() != null) visitArrayInitializer(ctx.arrayInitializer())
    else if(ctx.expression() != null) visitExpression(ctx.expression())
    else {
        println(s"Error: unexpected alternative in VariableInitializer: ${ctx.getText}")
        null
    }
  }

  override def visitArrayInitializer(ctx: JavaParser.ArrayInitializerContext): ArrayInitializer = {
    val initializers = if(ctx.variableInitializerList() == null) Nil
                       else visitVariableInitializerListAsList(ctx.variableInitializerList())
    var comma: Option[StringNode] = None
    if (ctx.COMMA() != null) {
      val trailingComma: StringNode = StringNode(",", getSourceLocation(ctx.COMMA()))
      trailingComma.precedingWhitespace = getWhitespaceLeft(ctx.COMMA())
      comma = Option(trailingComma)
    }
    val ini = ArrayInitializer(initializers, comma, getSourceLocation(ctx))
    ini.openBrace.precedingWhitespace = getWhitespaceLeft(ctx.LBRACE())
    ini.closeBrace.precedingWhitespace = getWhitespaceLeft(ctx.RBRACE())
    ini
  }

  def visitVariableInitializerListAsList(ctx: JavaParser.VariableInitializerListContext): List[Initializer] = {
    if (ctx.variableInitializerList() == null) {
      val onlyInit = visitVariableInitializer(ctx.variableInitializer())
      List(onlyInit)
    } else {
      val head = visitVariableInitializer(ctx.variableInitializer())
      head.trailingWhitespace = getWhitespaceRight(ctx.variableInitializer())
      val tail = visitVariableInitializerListAsList(ctx.variableInitializerList())
      tail.prepended(head)
    }
  }

  override def visitTypeOrVoid(ctx: JavaParser.TypeOrVoidContext): TypeIndication = {
    if (ctx.`type`() != null) {
      visitType(ctx.`type`())
    } else {
      new StringTypeIndication(ctx.VOID().getText, Nil, ctx.VOID().getText,
                               getSourceLocation(ctx), getWhitespaceLeft(ctx.VOID()))
    }
  }

  def visitDimsAsList(ctx: JavaParser.DimsContext): List[Dimension] = {
    var dims: List[Dimension] = Nil
    for(d: JavaParser.DimContext <- ctx.dim().asScala) {
      val dim = Dimension(None, getSourceLocation(d))
      dim.openBracket.precedingWhitespace = getWhitespaceLeft(d.LBRACK())
      dim.closeBracket.precedingWhitespace = getWhitespaceLeft(d.RBRACK())
      dims = dims.appended(dim)
    }
    dims
  }

  override def visitClassOrInterfaceType(ctx: JavaParser.ClassOrInterfaceTypeContext): TypeIndication = {
    val text = printer.visitSkipFirstWS(ctx)
    new StringTypeIndication(text, Nil, ctx.getText, getSourceLocation(ctx), getWhitespaceLeft(ctx))
  }

  override def visitTypeArguments(ctx: JavaParser.TypeArgumentsContext): TypeArgumentClause = {
    val args = visitTypeArgumentListAsList(ctx.typeArgumentList())
    val clause = TypeArgumentClause(args, getSourceLocation(ctx))
    clause.openAngle.precedingWhitespace = getWhitespaceLeft(ctx.LT())
    clause.closeAngle.precedingWhitespace = getWhitespaceLeft(ctx.GT())
    clause
  }

  def visitTypeArgumentListAsList(ctx: JavaParser.TypeArgumentListContext): List[TypeArgument] = {
    if (ctx.typeArgumentList() == null) {
      val onlyType = visitTypeArgument(ctx.typeArgument())
      List(onlyType)
    } else {
      val head = visitTypeArgument(ctx.typeArgument())
      head.trailingWhitespace = getWhitespaceRight(ctx.typeArgument())
      val tail = visitTypeArgumentListAsList(ctx.typeArgumentList())
      tail.prepended(head)
    }
  }

  override def visitTypeArgument(ctx: JavaParser.TypeArgumentContext): TypeArgument = {
    if(ctx.`type`() != null) {
      visitType(ctx.`type`())
    } else {
      var superType: Option[TypeIndication] = None
      var extendsType: Option[TypeIndication] = None
      if (ctx.boundType() != null) {
        if (ctx.boundType().EXTENDS() != null) {
          extendsType = Some(visitType(ctx.boundType().`type`()))
        }
        if (ctx.boundType().SUPER() != null) {
          superType = Some(visitType(ctx.boundType().`type`()))
        }
      }
      val arg = BoundedTypeArgument(Nil, superType, extendsType, getSourceLocation(ctx))
      arg.questionMark.precedingWhitespace = getWhitespaceLeft(ctx.QUESTION())
      if (ctx.boundType() != null) {
        if (ctx.boundType().EXTENDS() != null) {
          arg.extendsKeyword.precedingWhitespace = getWhitespaceLeft(ctx.boundType().EXTENDS())
        }
        if (ctx.boundType().SUPER() != null) {
          arg.superKeyword.precedingWhitespace = getWhitespaceLeft(ctx.boundType().EXTENDS())
        }
      }
      arg
    }
  }

  def visitQualifiedNameListAsList(ctx: JavaParser.QualifiedNameListContext): List[StringNode] = {
    if (ctx.qualifiedNameList() == null) {
      val onlyName = visitQualifiedName(ctx.qualifiedName())
      List(onlyName)
    } else {
      val head = visitQualifiedNameWithWS(ctx.qualifiedName())
      val tail = visitQualifiedNameListAsList(ctx.qualifiedNameList())
      tail.prepended(head)
    }
  }

  def visitFormalParameterListAsList(ctx: JavaParser.FormalParameterListContext): List[Parameter] = {
    var params:List[Parameter] = Nil
    if (ctx.initFormalParameterList() != null) {
      params = visitInitFormalParameterListAsList(ctx.initFormalParameterList())
    }
    if(ctx.COMMA() != null) {
      params.last.trailingWhitespace = getWhitespaceLeft(ctx.COMMA())
    }
    if(ctx.varargsFormalParameter() != null) {
      params = params.appended(visitVarargsFormalParameter(ctx.varargsFormalParameter()))
    }
    params
  }

  def visitInitFormalParameterListAsList(ctx: JavaParser.InitFormalParameterListContext): List[Parameter] = {
    if (ctx.initFormalParameterList() == null) {
      val onlyParam = visitFormalParameter(ctx.formalParameter())
      List(onlyParam)
    } else {
      val head = visitFormalParameter(ctx.formalParameter())
      head.trailingWhitespace = getWhitespaceRight(ctx.formalParameter())
      val tail = visitInitFormalParameterListAsList(ctx.initFormalParameterList())
      tail.prepended(head)
    }
  }

  override def visitFormalParameter(ctx: JavaParser.FormalParameterContext): Parameter = {
    val modifiers = ctx.variableModifier().asScala.map(visitVariableModifier).toList
    val paramType = visitType(ctx.`type`())
    val name = visitVariableDeclaratorId(ctx.variableDeclaratorId())
    val param = new Parameter(paramType, name, getSourceLocation(ctx))
    param.modifiers = modifiers
    param
  }

  override def visitVarargsFormalParameter(ctx: JavaParser.VarargsFormalParameterContext): VarargsParameter = {
    val modifiers = ctx.variableModifier().asScala.map(visitVariableModifier).toList
    val paramType = visitType(ctx.`type`())
    val name = visitVariableDeclaratorId(ctx.variableDeclaratorId())
    val param = VarargsParameter(paramType, Nil, name, getSourceLocation(ctx))
    param.modifiers = modifiers
    param.dots.precedingWhitespace = getWhitespaceLeft(ctx.ELLIPSIS())
    param
  }

  override def visitMethodBodyOrEmpty(ctx: JavaParser.MethodBodyOrEmptyContext): BlockOrSemicolon = {
    if (ctx.SEMI() != null) {
      new EmptyStatement(getSourceLocation(ctx), getWhitespaceLeft(ctx.SEMI()))
    } else {
      visitBlock(ctx.methodBody().block())
    }
  }

  override def visitQualifiedName(ctx: JavaParser.QualifiedNameContext): StringNode = printNode(ctx)
  def visitQualifiedNameWithWS(ctx: JavaParser.QualifiedNameContext): StringNode = printNodeWithWS(ctx)
  def visitQualifiedNameWithWSAsTypeIndication(ctx: JavaParser.QualifiedNameContext): TypeIndication = {
    val text = printer.visitSkipFirstWS(ctx)
    new StringTypeIndication(text, Nil, ctx.getText, getSourceLocation(ctx),
                             getWhitespaceLeft(ctx), getWhitespaceRight(ctx))
  }


  override def visitLiteral(ctx: JavaParser.LiteralContext): Literal = {
    Literal(printNode(ctx), getSourceLocation(ctx))
  }

  // TODO: how much of the annotation-related ones do we need?
  override def visitAnnotation(ctx: JavaParser.AnnotationContext): StringAnnotation = {
    val src = getSourceLocation(ctx)
    val text = printer.visitSkipFirstWS(ctx)
    StringAnnotation(text, src, getWhitespaceLeft(ctx))
  }

  override def visitAnnotationTypeDeclaration(ctx: JavaParser.AnnotationTypeDeclarationContext):
            AnnotationTypeDeclaration = {
    // todo
    super.visitAnnotationTypeDeclaration(ctx).asInstanceOf[AnnotationTypeDeclaration]
  }


  override def visitBlock(ctx: JavaParser.BlockContext): Block = {
    val statements = ctx.blockStatement().asScala.map(visitBlockStatement).toList
    val block = Block(None, statements, getSourceLocation(ctx))
    block.openBrace.precedingWhitespace = getWhitespaceLeft(ctx.LBRACE())
    block.closeBrace.precedingWhitespace = getWhitespaceLeft(ctx.RBRACE())
    block
  }

  override def visitBlockStatement(ctx: JavaParser.BlockStatementContext): Statement = {
    if(ctx.localVariableDeclarationStatement() != null) {
      val decl = visitLocalVariableDeclaration(ctx.localVariableDeclarationStatement().localVariableDeclaration())
      decl.isStatement = true
      decl.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.localVariableDeclarationStatement().SEMI())
      decl
    } else if(ctx.statement() != null) {
      visitStatement(ctx.statement())
    } else if(ctx.typeDeclaration() != null) {
      visitTypeDeclaration(ctx.typeDeclaration())
    } else if(ctx.valEmbedStatementBlock() != null) {
      visitValEmbedStatementBlock(ctx.valEmbedStatementBlock())
    } else {
        println(s"Error: unexpected alternative in BlockStatement: ${ctx.getText}")
        null
    }
  }

  override def visitLocalVariableDeclaration(ctx: JavaParser.LocalVariableDeclarationContext):
            LocalDeclaration = {
    val modifiers = ctx.variableModifier().asScala.map(visitVariableModifier).toList
    val typeIndication = visitType(ctx.`type`())
    val declarators = visitVariableDeclaratorsAsList(ctx.variableDeclarators())
    val decl = LocalDeclaration(modifiers, typeIndication, declarators, isStatement = false, getSourceLocation(ctx))
    decl
  }

  def visitLoopAmalgamationAsList(ctx: JavaParser.LoopAmalgamationContext): List[LoopSpecOrLabel] = {
    if(ctx.valEmbedContract() != null) {
      val contract = visitValEmbedContractForLoop(ctx.valEmbedContract())
      val rest = visitLoopAmalgamationAsList(ctx.loopAmalgamation())
      rest.prependedAll(contract)
    } else if(ctx.loopLabel() != null) {
      val label = Label(visitJavaIdentifier(ctx.loopLabel().javaIdentifier()), getSourceLocation(ctx.loopLabel()))
      label.colon.precedingWhitespace = getWhitespaceLeft(ctx.loopLabel().COLON())
      val rest = visitLoopAmalgamationAsList(ctx.loopAmalgamation())
      rest.prepended(label)
    } else {
      Nil
    }
  }

  override def visitStatement(ctx: JavaParser.StatementContext): Statement = {
    try{
      if(ctx.ASSERT() != null) {
        val content = visitExpression(ctx.expression())
        val message = Option(ctx.assertMessage()).map(x=>visitExpression(x.expression()))
        val assertion = JavaAssertion(content, message, getSourceLocation(ctx))
        assertion.assertKeyword.precedingWhitespace = getWhitespaceLeft(ctx.ASSERT())
        if (ctx.assertMessage() != null) {
          assertion.colon.precedingWhitespace = getWhitespaceLeft(ctx.assertMessage().COLON())
        }
        assertion.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
        assertion
      } else if(ctx.IF() != null) {
        val cond = visitExpression(ctx.parExpression().expression())
        val thenStatement = visitStatement(ctx.statement())
        val elseStatement = Option(ctx.elseBlock()).map(x=>visitStatement(x.statement()))
        val ifStatement = IfStatement(cond, thenStatement, elseStatement, getSourceLocation(ctx))
        ifStatement.ifKeyword.precedingWhitespace = getWhitespaceLeft(ctx.IF())
        ifStatement.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.parExpression().LPAREN())
        ifStatement.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.parExpression().RPAREN())
        if (ctx.elseBlock() != null) {
          ifStatement.elseKeyword.precedingWhitespace = getWhitespaceLeft(ctx.elseBlock().ELSE())
        }
        ifStatement
      } else if(ctx.FOR() != null) {
        val amalgamation = visitLoopAmalgamationAsList(ctx.loopAmalgamation())
        var contract: List[LoopSpecOrLabel] = Nil
        if (ctx.valEmbedContract() != null) {
          contract = visitValEmbedContractForLoop(ctx.valEmbedContract())
        }
        val control = visitForControl(ctx.forControl())
        val body = visitStatement(ctx.statement())
        val loop = ForLoop(amalgamation, control, contract, body, getSourceLocation(ctx))
        loop.forKeyword.precedingWhitespace = getWhitespaceLeft(ctx.FOR())
        loop.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.LPAREN())
        loop.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.RPAREN())
        loop
      } else if(ctx.DO() != null) {
        val cond = visitExpression(ctx.parExpression().expression())
        val body = visitStatement(ctx.statement())
        val loop = DoLoop(Nil, body, cond, getSourceLocation(ctx))
        loop.doKeyword.precedingWhitespace = getWhitespaceLeft(ctx.DO())
        loop.whileKeyword.precedingWhitespace = getWhitespaceLeft(ctx.WHILE())
        loop.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.parExpression().LPAREN())
        loop.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.parExpression().RPAREN())
        loop.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
        loop
      } else if(ctx.WHILE() != null) {
        val amalgamation = visitLoopAmalgamationAsList(ctx.loopAmalgamation())
        var contract: List[LoopSpecOrLabel] = Nil
        if (ctx.valEmbedContract() != null) {
          contract = visitValEmbedContractForLoop(ctx.valEmbedContract())
        }
        val cond = visitExpression(ctx.parExpression().expression())
        val body = visitStatement(ctx.statement())
        val loop = WhileLoop(amalgamation, cond, contract, body, getSourceLocation(ctx))
        loop.whileKeyword.precedingWhitespace = getWhitespaceLeft(ctx.WHILE())
        loop.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.parExpression().LPAREN())
        loop.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.parExpression().RPAREN())
        loop
      } else if(ctx.resourceSpecification() != null) {
        val resource = visitResourceSpecification(ctx.resourceSpecification())
        val block = visitBlock(ctx.block())
        val catches = ctx.catchClause().asScala.map(visitCatchClause).toList
        val finallyBlock = Option(ctx.finallyBlock()).map(x=>visitBlock(x.block()))
        val tryBlock = TryBlock(Option(resource), block, catches, finallyBlock, getSourceLocation(ctx))
        tryBlock.tryKeyword.precedingWhitespace = getWhitespaceLeft(ctx.TRY())
        if (ctx.finallyBlock() != null) {
          tryBlock.finallyKeyword.precedingWhitespace = getWhitespaceLeft(ctx.finallyBlock().FINALLY())
        }
        tryBlock
      } else if (ctx.TRY() != null) {
        val block = visitBlock(ctx.block())
        val catches = ctx.catchClause().asScala.map(visitCatchClause).toList
        val finallyBlock = Option(ctx.finallyBlock()).map(x=>visitBlock(x.block()))
        val tryBlock = TryBlock(None, block, catches, finallyBlock, getSourceLocation(ctx))
        tryBlock.tryKeyword.precedingWhitespace = getWhitespaceLeft(ctx.TRY())
        if (ctx.finallyBlock() != null) {
          tryBlock.tryKeyword.precedingWhitespace = getWhitespaceLeft(ctx.finallyBlock().FINALLY())
        }
        tryBlock
      } else if(ctx.SWITCH() != null) {
        val cond = visitExpression(ctx.parExpression().expression())
        var blocks = ctx.switchBlockStatementGroup().asScala
          .map(visitSwitchBlockStatementGroup).toList
        if (!ctx.switchLabel().isEmpty) {
          val lastBlock = SwitchCaseBlock(ctx.switchLabel().asScala.map(visitSwitchLabel).toList, Nil,
            getSourceLocation(ctx.switchLabel(0)))
          blocks = blocks.appended(lastBlock)
        }
        val switchStatement = SwitchStatement(cond, blocks, getSourceLocation(ctx))
        switchStatement.switchKeyword.precedingWhitespace = getWhitespaceLeft(ctx.SWITCH())
        switchStatement.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.parExpression().LPAREN())
        switchStatement.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.parExpression().RPAREN())
        switchStatement.openBrace.precedingWhitespace = getWhitespaceLeft(ctx.LBRACE())
        switchStatement.closeBrace.precedingWhitespace = getWhitespaceLeft(ctx.RBRACE())
        switchStatement
      } else if(ctx.SYNCHRONIZED() != null) {
        val cond = visitExpression(ctx.parExpression().expression())
        val body = visitBlock(ctx.block())
        val synchronizedStatement = SynchronizedStatement(cond, body, getSourceLocation(ctx))
        synchronizedStatement.synchronizedKeyword.precedingWhitespace = getWhitespaceLeft(ctx.SYNCHRONIZED())
        synchronizedStatement.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.parExpression().LPAREN())
        synchronizedStatement.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.parExpression().RPAREN())
        synchronizedStatement
      } else if(ctx.RETURN() != null) {
        val res = Option(ctx.expression()).map(visitExpression)
        val returnStatement = ReturnStatement(res, getSourceLocation(ctx))
        returnStatement.returnKeyword.precedingWhitespace = getWhitespaceLeft(ctx.RETURN())
        returnStatement.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
        returnStatement
      } else if(ctx.THROW() != null) {
        val res = visitExpression(ctx.expression())
        val throwStatement = ThrowStatement(res, getSourceLocation(ctx))
        throwStatement.throwKeyword.precedingWhitespace = getWhitespaceLeft(ctx.THROW())
        throwStatement.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
        throwStatement
      } else if(ctx.BREAK() != null) {
        val res = Option(ctx.javaIdentifier()).map(visitJavaIdentifier)
        val break = BreakStatement(res, getSourceLocation(ctx))
        break.breakKeyword.precedingWhitespace = getWhitespaceLeft(ctx.BREAK())
        break.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
        break
      } else if(ctx.CONTINUE() != null) {
        val res = Option(ctx.javaIdentifier()).map(visitJavaIdentifier)
        val cont = ContinueStatement(res, getSourceLocation(ctx))
        cont.continueKeyword.precedingWhitespace = getWhitespaceLeft(ctx.CONTINUE())
        cont.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
        cont
      } else if(ctx.statementExpression() != null) {
        val expr = ExpressionStatement(visitExpression(ctx.statementExpression().expression()), getSourceLocation(ctx))
        expr.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
        expr
      } else if(ctx.valStatement() != null) {
        visitValStatement(ctx.valStatement())
      } else if (ctx.javaIdentifier() != null) {
        val label = visitJavaIdentifier(ctx.javaIdentifier())
        val lbl = Label(label, getSourceLocation(ctx.javaIdentifier()))
        lbl.colon.precedingWhitespace = getWhitespaceLeft(ctx.COLON())
        val stmt = visitStatement(ctx.statement())
        LabeledStatement(lbl, stmt, getSourceLocation(ctx))
        /// todo contract
      } else if(ctx.block() != null) {
        visitBlock(ctx.block())
      } else if(ctx.SEMI() != null) {
        new EmptyStatement(getSourceLocation(ctx), getWhitespaceLeft(ctx.SEMI()))
      } else {
        println(s"Error: unexpected alternative in Statement: ${ctx.getText}")
        null
      }
    }catch {
      case e: NotSupportedException =>
        val commented = NonTranslatedStatement("", getSourceLocation(ctx))
        setNonTranslatedTreeNodeText(commented, ctx, e)
        commented
    }
  }

  override def visitCatchClause(ctx: JavaParser.CatchClauseContext): CatchClause = {
    val modifiers = ctx.variableModifier().asScala.map(visitVariableModifier).toList
    val types = visitCatchTypeAsList(ctx.catchType())
    val name = visitJavaIdentifier(ctx.javaIdentifier())
    val block = visitBlock(ctx.block())
    val catchClause = CatchClause(types, name, block, getSourceLocation(ctx))
    catchClause.catchKeyword.precedingWhitespace = getWhitespaceLeft(ctx.CATCH())
    catchClause.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.LPAREN())
    catchClause.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.RPAREN())
    catchClause.modifiers = modifiers
    catchClause
  }

  def visitCatchTypeAsList(ctx: JavaParser.CatchTypeContext): List[TypeIndication] = {
    if (ctx.catchType() == null) {
      val onlyType = visitClassOrInterfaceType(ctx.classOrInterfaceType())
      onlyType.trailingWhitespace = ""
      List(onlyType)
    } else {
      val head = visitClassOrInterfaceType(ctx.classOrInterfaceType())
      val tail = visitCatchTypeAsList(ctx.catchType())
      tail.prepended(head)
    }
  }

  override def visitResourceSpecification(ctx: JavaParser.ResourceSpecificationContext):
          TryResourceSpecification = {
    val semi = if (ctx.SEMI() == null) None
               else Option(StringNode(ctx.SEMI().getText,
                           getSourceLocation(ctx.SEMI()), getWhitespaceLeft(ctx.SEMI())))
    val resources = visitResourcesAsList(ctx.resources())
    val res = TryResourceSpecification(resources, semi, getSourceLocation(ctx))
    res.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.LPAREN())
    res.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.RPAREN())
    res
  }

  def visitResourcesAsList(ctx: JavaParser.ResourcesContext): List[TryResource] = {
    if (ctx.resources() == null) {
      val onlyRes = visitResource(ctx.resource())
      List(onlyRes)
    } else {
      val head = visitResource(ctx.resource())
      head.trailingWhitespace = getWhitespaceRight(ctx.resource())
      val tail = visitResourcesAsList(ctx.resources())
      tail.prepended(head)
    }
  }

  override def visitResource(ctx: JavaParser.ResourceContext): TryResource = {
    val modifiers = ctx.variableModifier().asScala.map(visitVariableModifier).toList
    val rType = visitClassOrInterfaceType(ctx.classOrInterfaceType())
    val variable = visitVariableDeclaratorId(ctx.variableDeclaratorId())
    val expr = visitExpression(ctx.expression())
    val res = TryResource(modifiers, rType, variable, expr, getSourceLocation(ctx))
    res.equalsSign. precedingWhitespace = getWhitespaceLeft(ctx.ASSIGN())
    res
  }

  override def visitSwitchBlockStatementGroup(ctx: JavaParser.SwitchBlockStatementGroupContext):
                SwitchCaseBlock = {
    SwitchCaseBlock(ctx.switchLabel().asScala.map(visitSwitchLabel).toList,
                    ctx.blockStatement().asScala.map(visitBlockStatement).toList,
                    getSourceLocation(ctx))
  }

  override def visitSwitchLabel(ctx: JavaParser.SwitchLabelContext): SwitchLabel = {
    if (ctx.CASE() != null) {
      var cond: Expression = null
      if(ctx.constantExpression() != null) {
        cond = visitExpression(ctx.constantExpression().expression())
      } else {
        cond = visitJavaIdentifier(ctx.enumConstantName().javaIdentifier())
      }
      val label = SwitchLabel(Option(cond), getSourceLocation(ctx))
      label.caseKeyword.precedingWhitespace = getWhitespaceLeft(ctx.CASE())
      label.colon.precedingWhitespace = getWhitespaceLeft(ctx.COLON())
      label
    } else {
      val label = SwitchLabel(None, getSourceLocation(ctx))
      label.defaultKeyword.precedingWhitespace = getWhitespaceLeft(ctx.DEFAULT())
      label.colon.precedingWhitespace = getWhitespaceLeft(ctx.COLON())
      label
    }
  }

  override def visitForControl(ctx: JavaParser.ForControlContext): ForLoopControl = {
    if (ctx.enhancedForControl() != null) {
      visitEnhancedForControl(ctx.enhancedForControl())
    } else {
      val init = Option(ctx.forInit()).map(visitForInit)
      val cond = Option(ctx.expression()).map(visitExpression)
      val update = if(ctx.forUpdate().expressionList()==null) Nil
                   else visitExpressionListAsList(ctx.forUpdate().expressionList())
      val ctl = RegularForLoopControl(init, cond, update, getSourceLocation(ctx))
      ctl.firstSemicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI(0))
      ctl.secondSemicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI(1))
      ctl
    }
  }

  override def visitForInit(ctx: JavaParser.ForInitContext): ForInit = {
    if (ctx.localVariableDeclaration() != null) {
      visitLocalVariableDeclaration(ctx.localVariableDeclaration())
    } else {
      val exprs = visitExpressionListAsList(ctx.expressionList())
      ExpressionListForInit(exprs, getSourceLocation(ctx))
    }
  }

  override def visitEnhancedForControl(ctx: JavaParser.EnhancedForControlContext): ForEachLoopControl = {
    val modifiers = ctx.variableModifier().asScala.map(visitVariableModifier).toList
    val cType = visitType(ctx.`type`())
    val variable = visitVariableDeclaratorId(ctx.variableDeclaratorId())
    val control = visitExpression(ctx.expression())
    val ctl = ForEachLoopControl(cType, variable, control, getSourceLocation(ctx))
    ctl.modifiers = modifiers
    ctl.colon.precedingWhitespace = getWhitespaceLeft(ctx.COLON())
    ctl
  }


  override def visitParExpression(ctx: JavaParser.ParExpressionContext): Expression = {
    val parEx = ParExpression(visitExpression(ctx.expression()), getSourceLocation(ctx))
    parEx.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.LPAREN())
    parEx.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.RPAREN())
    parEx
  }

  def visitExpressionListAsList(ctx: JavaParser.ExpressionListContext): List[Expression] = {
    if (ctx.expressionList() == null) {
      val onlyExpr = visitExpression(ctx.expression())
      List(onlyExpr)
    } else {
      val head = visitExpression(ctx.expression())
      head.trailingWhitespace = getWhitespaceRight(ctx.expression())
      val tail = visitExpressionListAsList(ctx.expressionList())
      tail.prepended(head)
    }
  }

  override def visitExpression(ctx: JavaParser.ExpressionContext): Expression = {
    // todo: with/then
    visitExpr(ctx.expr())
  }

  def visitExpr(ctx: JavaParser.ExprContext): Expression = {
    visit(ctx).asInstanceOf[Expression]
  }

  override def visitJavaPrimary(ctx: JavaParser.JavaPrimaryContext): Expression = {
    val prim = visitPrimary(ctx.annotatedPrimary().primary())
    // todo: with/then
    prim
  }

  override def visitJavaDeref(ctx: JavaParser.JavaDerefContext): Expression = {
    val left = visitExpr(ctx.expr())
    val right = visitJavaIdentifier(ctx.javaIdentifier())
    val nested = NestedInvocationExpression(left, right, getSourceLocation(ctx))
    nested.dot.precedingWhitespace = getWhitespaceLeft(ctx.DOT())
    nested
  }

  override def visitJavaPinnedThis(ctx: JavaParser.JavaPinnedThisContext): Expression = {
    val left = visitExpr(ctx.expr())
    val right = StringNode(ctx.THIS().getText, getSourceLocation(ctx.THIS()), getWhitespaceLeft(ctx.THIS()))
    val nested = NestedInvocationExpression(left, right, getSourceLocation(ctx))
    nested.dot.precedingWhitespace = getWhitespaceLeft(ctx.DOT())
    nested
  }

  override def visitJavaPinnedOuterClassNew(ctx: JavaParser.JavaPinnedOuterClassNewContext): Expression = {
    val left = visitExpr(ctx.expr())
    val creator = visitInnerCreator(ctx.innerCreator())
    creator.newKeyword.precedingWhitespace = getWhitespaceLeft(ctx.NEW())
    if (ctx.nonWildcardTypeArguments() != null) {
      creator.typeArguments = Some(visitNonWildcardTypeArguments(ctx.nonWildcardTypeArguments()))
    }
    val nested = NestedInvocationExpression(left, creator, getSourceLocation(ctx))
    nested.dot.precedingWhitespace = getWhitespaceLeft(ctx.DOT())
    nested
  }

  override def visitJavaSuper(ctx: JavaParser.JavaSuperContext): Expression = {
    val left = visitExpr(ctx.expr())
    var right: Expression = null
    if (ctx.superSuffix().javaIdentifier() != null) {
      var text = ctx.SUPER().getText
      text += getWhitespaceLeft(ctx.superSuffix().DOT())
      text += ctx.superSuffix().DOT().getText
      text += getWhitespaceLeft(ctx.superSuffix().javaIdentifier())
      text += ctx.superSuffix().javaIdentifier().getText
      if (ctx.superSuffix().arguments()!=null) {
        val name = StringNode(text, getSourceLocation(ctx), getWhitespaceLeft(ctx.SUPER()))
        val args = visitArguments(ctx.superSuffix().arguments())
        right = MethodCallExpression(name, args, getSourceLocation(ctx))
      } else {
        right = StringNode(text, getSourceLocation(ctx), getWhitespaceLeft(ctx.SUPER()))
      }
    } else {
      val name = StringNode(ctx.SUPER().getText, getSourceLocation(ctx.SUPER()),
                            getWhitespaceLeft(ctx.SUPER()))
      val arguments = visitArguments(ctx.superSuffix().arguments())
      right = MethodCallExpression(name, arguments, getSourceLocation(ctx))
    }
    val nested = NestedInvocationExpression(left, right, getSourceLocation(ctx))
    nested.dot.precedingWhitespace = getWhitespaceLeft(ctx.DOT())
    nested
  }

  override def visitJavaGenericInvocation(ctx: JavaParser.JavaGenericInvocationContext): Expression = {
    val left = visitExpr(ctx.expr())
    val right = visitExplicitGenericInvocation(ctx.explicitGenericInvocation())
    val nested = NestedInvocationExpression(left, right, getSourceLocation(ctx))
    nested.dot.precedingWhitespace = getWhitespaceLeft(ctx.DOT())
    nested
  }

  override def visitJavaSubscript(ctx: JavaParser.JavaSubscriptContext): Expression = {
    val arrExpr = visitExpr(ctx.expr(0))
    val index = visitExpr(ctx.expr(1))
    val arr = ArrayIndexExpression(arrExpr, index, getSourceLocation(ctx))
    arr.openBracket.precedingWhitespace = getWhitespaceLeft(ctx.LBRACK())
    arr.closeBracket.precedingWhitespace = getWhitespaceLeft(ctx.RBRACK())
    arr
  }

  override def visitJavaNonNullInvocation(ctx: JavaParser.JavaNonNullInvocationContext): Expression = {
    val expr = visitExpr(ctx.expr())
    val id = visitJavaIdentifier(ctx.javaIdentifier())
    val args = visitArguments(ctx.arguments())
    val eq = BinaryExpression(expr,
                              StringNode("!=", getSourceLocation(ctx)),
                              StringNode("null", getSourceLocation(ctx)),
                              getSourceLocation(ctx))
    val call = MethodCallExpression(id, args, getSourceLocation(ctx))
    val nested = NestedInvocationExpression(expr, call, getSourceLocation(ctx))
    BinaryExpression(eq, StringNode("==>", getSourceLocation(ctx), " "), nested, getSourceLocation(ctx))
  }

  override def visitJavaInvocation(ctx: JavaParser.JavaInvocationContext): Expression = {
    val expr = visitExpr(ctx.expr())
    val id = visitJavaIdentifier(ctx.javaIdentifier())
    val args = visitArguments(ctx.arguments())
    // todo given/yields, prediateentrytype
    val call = MethodCallExpression(id, args, getSourceLocation(ctx))
    val nested = NestedInvocationExpression(expr, call, getSourceLocation(ctx))
    nested.dot.precedingWhitespace = getWhitespaceLeft(ctx.DOT())
    nested
  }

  override def visitJavaValPostfix(ctx: JavaParser.JavaValPostfixContext): Expression = {
    val expr = visitExpr(ctx.expr())
    if(ctx.postfixOp().valPostfix().LBRACK() != null) {
      val idx = visitValPostfix(ctx.postfixOp().valPostfix())
      val res = ArrayIndexExpression(expr, idx, getSourceLocation(ctx))
      res.openBracket.precedingWhitespace = getWhitespaceLeft(ctx.postfixOp().valPostfix().LBRACK())
      res.closeBracket.precedingWhitespace = getWhitespaceLeft(ctx.postfixOp().valPostfix().RBRACK())
      res
    } else {
      // postfix is coalesce operator '?.'
      val postfix = ctx.postfixOp().valPostfix()
      val eq = BinaryExpression(expr,
                                StringNode("!=", getSourceLocation(ctx)),
                                StringNode("null", getSourceLocation(ctx)),
                                getSourceLocation(ctx))

      val args = if(postfix.valExpressionList() == null) Nil
                else visitValExpressionListAsList(postfix.valExpressionList())
      val arg = Arguments(args, getSourceLocation(ctx))
      arg.openParenthesis.precedingWhitespace = getWhitespaceLeft(postfix.LPAREN())
      arg.closeParenthesis.precedingWhitespace = getWhitespaceLeft(postfix.RPAREN())
      val call = MethodCallExpression(visitJavaIdentifier(postfix.langId().javaIdentifier()), arg, getSourceLocation(ctx), " ")
      val nested = NestedInvocationExpression(expr, call, getSourceLocation(ctx))
      BinaryExpression(eq, StringNode("==>", getSourceLocation(ctx), " "), nested, getSourceLocation(ctx))
    }
  }

  override def visitJavaNew(ctx: JavaParser.JavaNewContext): Expression = {
    val newExpr = visitCreator(ctx.creator())
    newExpr.newKeyword.precedingWhitespace = getWhitespaceLeft(ctx.NEW())
    // todo: given/yields
    newExpr
  }

  override def visitJavaCast(ctx: JavaParser.JavaCastContext): Expression = {
    val types = visitTypeWithWS(ctx.`type`())
    val expr = visitExpr(ctx.expr())
    val cast = CastExpression(Nil, List(types), expr, getSourceLocation(ctx))
    cast.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.LPAREN())
    cast.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.RPAREN())
    cast
  }

  override def visitJavaPostfixIncDec(ctx: JavaParser.JavaPostfixIncDecContext): Expression = {
    val expr = visitExpr(ctx.expr())
    val postfixCtx = if(ctx.DEC() != null) ctx.DEC()
                     else if(ctx.INC() != null) ctx.INC()
                     else null // should be unreachable
    val op: StringNode = StringNode(postfixCtx.getText, getSourceLocation(postfixCtx), getWhitespaceLeft(postfixCtx))
    PostfixExpression(expr, op, getSourceLocation(ctx))
  }

  override def visitJavaPrefixOp(ctx: JavaParser.JavaPrefixOpContext): Expression = {
    val expr = visitExpr(ctx.expr())
    val prefixCtx = if(ctx.ADD() != null) ctx.ADD()
                    else if(ctx.SUB() != null) ctx.SUB()
                    else if(ctx.INC() != null) ctx.INC()
                    else if(ctx.DEC() != null) ctx.DEC()
                    else null // should be unreachable
    val op = StringNode(prefixCtx.getText, getSourceLocation(prefixCtx), getWhitespaceLeft(prefixCtx))
    PrefixExpression(expr, op, getSourceLocation(ctx))
  }

  override def visitJavaPrefixOp2(ctx: JavaParser.JavaPrefixOp2Context): Expression = {
    val expr = visitExpr(ctx.expr())
    val prefixCtx = if(ctx.TILDE() != null) ctx.TILDE()
                    else if(ctx.BANG() != null) ctx.BANG()
                    else null // should be unreachable
    val op = StringNode(prefixCtx.getText, getSourceLocation(prefixCtx), getWhitespaceLeft(prefixCtx))
    PrefixExpression(expr, op, getSourceLocation(ctx))
  }

  override def visitJavaValPrepend(ctx: JavaParser.JavaValPrependContext): Expression = {
    val left = visitExpr(ctx.expr(0))
    val right = visitExpr(ctx.expr(1))
    val prepCtx = ctx.prependOp().valPrependOp()
    val op = StringNode(prepCtx.getText, getSourceLocation(prepCtx), getWhitespaceLeft(prepCtx))
    BinaryExpression(left, op, right, getSourceLocation(ctx))
  }

  override def visitJavaMul(ctx: JavaParser.JavaMulContext): Expression = {
    val left = visitExpr(ctx.expr(0))
    val right = visitExpr(ctx.expr(1))
    var op: StringNode = null
    if(ctx.mulOp().valMulOp() != null) {
      op = PermissionFractionOperator(getSourceLocation(ctx.mulOp()), getWhitespaceLeft(ctx.mulOp()))
    } else {
      op = StringNode(ctx.mulOp().getText, getSourceLocation(ctx.mulOp()), getWhitespaceLeft(ctx.mulOp()))
    }
    BinaryExpression(left, op, right, getSourceLocation(ctx))
  }

  override def visitJavaAdd(ctx: JavaParser.JavaAddContext): Expression = {
    val left = visitExpr(ctx.expr(0))
    val right = visitExpr(ctx.expr(1))
    val addCtx = if(ctx.ADD() != null) ctx.ADD()
                else if(ctx.SUB() != null) ctx.SUB()
                else null // should be unreachable
    val op = StringNode(addCtx.getText, getSourceLocation(addCtx), getWhitespaceLeft(addCtx))
    BinaryExpression(left, op, right, getSourceLocation(ctx))
  }

  override def visitJavaShift(ctx: JavaParser.JavaShiftContext): Expression = {
    val left = visitExpr(ctx.expr(0))
    val right = visitExpr(ctx.expr(1))
    val op = StringNode(ctx.shiftOp().getText, getSourceLocation(ctx.shiftOp()), getWhitespaceLeft(ctx.shiftOp()))
    BinaryExpression(left, op, right, getSourceLocation(ctx))
  }

  override def visitJavaRel(ctx: JavaParser.JavaRelContext): Expression = {
    val left = visitExpr(ctx.expr(0))
    val right = visitExpr(ctx.expr(1))
    val op = StringNode(ctx.relOp().getText, getSourceLocation(ctx.relOp()), getWhitespaceLeft(ctx.relOp()))
    BinaryExpression(left, op, right, getSourceLocation(ctx))
  }

  override def visitJavaInstanceOf(ctx: JavaParser.JavaInstanceOfContext): Expression = {
    val left = visitExpr(ctx.expr())
    val right = visitType(ctx.`type`())
    val res = InstanceofExpression(left, right, getSourceLocation(ctx))
    res.instanceofKeyword.precedingWhitespace = getWhitespaceLeft(ctx.INSTANCEOF())
    res
  }

  override def visitJavaEquals(ctx: JavaParser.JavaEqualsContext): Expression = {
    val left = visitExpr(ctx.expr(0))
    val right = visitExpr(ctx.expr(1))
    val eqCtx = if(ctx.EQUAL() != null) ctx.EQUAL()
                else if(ctx.NOTEQUAL() != null) ctx.NOTEQUAL()
                else null // should be unreachable
    val op = StringNode(eqCtx.getText, getSourceLocation(eqCtx), getWhitespaceLeft(eqCtx))
    BinaryExpression(left, op, right, getSourceLocation(ctx))
  }

  override def visitJavaBitAnd(ctx: JavaParser.JavaBitAndContext): Expression = {
    val left = visitExpr(ctx.expr(0))
    val right = visitExpr(ctx.expr(1))
    val op = StringNode(ctx.BITAND().getText, getSourceLocation(ctx.BITAND()), getWhitespaceLeft(ctx.BITAND()))
    BinaryExpression(left, op, right, getSourceLocation(ctx))
  }

  override def visitJavaBitXor(ctx: JavaParser.JavaBitXorContext): Expression = {
    val left = visitExpr(ctx.expr(0))
    val right = visitExpr(ctx.expr(1))
    val op = StringNode(ctx.CARET().getText, getSourceLocation(ctx.CARET()), getWhitespaceLeft(ctx.CARET()))
    BinaryExpression(left, op, right, getSourceLocation(ctx))
  }

  override def visitJavaBitOr(ctx: JavaParser.JavaBitOrContext): Expression = {
    val left = visitExpr(ctx.expr(0))
    val right = visitExpr(ctx.expr(1))
    val op = StringNode(ctx.BITOR().getText, getSourceLocation(ctx.BITOR()), getWhitespaceLeft(ctx.BITOR()))
    BinaryExpression(left, op, right, getSourceLocation(ctx))
  }

  override def visitJavaAnd(ctx: JavaParser.JavaAndContext): Expression = {
    val left = visitExpr(ctx.expr(0))
    val right = visitExpr(ctx.expr(1))
    val op = if(ctx.andOp().valAndOp() != null) new SeparatingConjunctOperator(getSourceLocation(ctx.andOp()), getWhitespaceLeft(ctx.andOp()))
          else StringNode(ctx.andOp().getText, getSourceLocation(ctx.andOp()), getWhitespaceLeft(ctx.andOp()))
    BinaryExpression(left, op, right, getSourceLocation(ctx))
  }

  override def visitJavaOr(ctx: JavaParser.JavaOrContext): Expression = {
    val left = visitExpr(ctx.expr(0))
    val right = visitExpr(ctx.expr(1))
    val op = StringNode(ctx.OR().getText, getSourceLocation(ctx.OR()), getWhitespaceLeft(ctx.OR()))
    BinaryExpression(left, op, right, getSourceLocation(ctx))
  }

  override def visitJavaValImp(ctx: JavaParser.JavaValImpContext): Expression = {
    val left = visitExpr(ctx.expr(0))
    val right = visitExpr(ctx.expr(1))
    val op = if(ctx.impOp().valImpOp().IMPLIES() != null) new ImpliesOperator(getSourceLocation(ctx.impOp()), getWhitespaceLeft(ctx.impOp()))
            else StringNode(ctx.impOp().getText, getSourceLocation(ctx.impOp()), getWhitespaceLeft(ctx.impOp()))
    BinaryExpression(left, op, right, getSourceLocation(ctx))
  }

  override def visitJavaSelect(ctx: JavaParser.JavaSelectContext): Expression = {
    val cond = visitExpr(ctx.expr(0))
    val thenExpr = visitExpr(ctx.expr(1))
    val elseExpr = visitExpr(ctx.expr(2))
    val tern = TernaryExpression(cond, thenExpr, elseExpr, getSourceLocation(ctx))
    tern.questionmark.precedingWhitespace = getWhitespaceLeft(ctx.QUESTION())
    tern.colon.precedingWhitespace = getWhitespaceLeft(ctx.COLON())
    tern
  }

  override def visitJavaAssign(ctx: JavaParser.JavaAssignContext): Expression = {
    val left = visitExpr(ctx.expr(0))
    val right = visitExpr(ctx.expr(1))
    val op = StringNode(ctx.assignOp().getText, getSourceLocation(ctx.assignOp()), getWhitespaceLeft(ctx.assignOp()))
    BinaryExpression(left, op, right, getSourceLocation(ctx))
  }

  override def visitPrimary(ctx: JavaParser.PrimaryContext): Expression = {
    if (ctx.LPAREN() != null) {
      val expr = ParExpression(visitExpression(ctx.expression()), getSourceLocation(ctx))
      expr.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.LPAREN())
      expr.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.RPAREN())
      expr
    } else if (ctx.THIS() != null) {
      val name = StringNode(ctx.THIS().getText, getSourceLocation(ctx))
      Literal(name, getSourceLocation(ctx), getWhitespaceLeft(ctx.THIS()))
    } else if (ctx.SUPER() != null) {
      val name = StringNode(ctx.SUPER().getText, getSourceLocation(ctx))
      Literal(name, getSourceLocation(ctx), getWhitespaceLeft(ctx.SUPER()))
    } else if (ctx.literal() != null) {
      val name = StringNode(ctx.literal().getText, getSourceLocation(ctx))
      Literal(name, getSourceLocation(ctx), getWhitespaceLeft(ctx.literal()))
    } else if (ctx.arguments() != null) {
      val name = visitJavaIdentifier(ctx.javaIdentifier())
      val args = visitArguments(ctx.arguments())
      // todo given, yields, predicateentrytype (which is for predicate families)
      MethodCallExpression(name, args, getSourceLocation(ctx))
    } else if(ctx.javaIdentifier() != null) {
      visitJavaIdentifier(ctx.javaIdentifier())
    } else if(ctx.`type`() != null) {
      val typeInd = visitType(ctx.`type`())
      val expr = ClassName(typeInd, getSourceLocation(ctx))
      expr.dot.precedingWhitespace = getWhitespaceLeft(ctx.DOT())
      expr.classKeyword.precedingWhitespace = getWhitespaceLeft(ctx.CLASS())
      expr
    } else if(ctx.VOID() != null) {
      val typeInd = new StringTypeIndication(ctx.VOID().getText, Nil, ctx.VOID().getText,
                                             getSourceLocation(ctx), getWhitespaceLeft(ctx.VOID()))
      val expr = ClassName(typeInd, getSourceLocation(ctx))
      expr.dot.precedingWhitespace = getWhitespaceLeft(ctx.DOT())
      expr.classKeyword.precedingWhitespace = getWhitespaceLeft(ctx.CLASS())
      expr
    } else if(ctx.nonWildcardTypeArguments() != null) {
      val typeParam = visitNonWildcardTypeArguments(ctx.nonWildcardTypeArguments())
      if(ctx.constructorCall().THIS() != null) {
        val name = StringNode(ctx.THIS().getText, getSourceLocation(ctx.THIS()), getWhitespaceLeft(ctx.THIS()))
        val arguments = visitArguments(ctx.arguments())
        GenericInvocationExpression(typeParam, name, Some(arguments), getSourceLocation(ctx))
      } else {
        val explSuffix = ctx.constructorCall().explicitGenericInvocationSuffix()
        if (explSuffix.SUPER() != null) {
          if (explSuffix.superSuffix().javaIdentifier() != null) {
            var text = explSuffix.SUPER().getText
            text += getWhitespaceLeft(explSuffix.superSuffix().DOT())
            text += explSuffix.superSuffix().DOT().getText
            text += getWhitespaceLeft(explSuffix.superSuffix().javaIdentifier())
            text += explSuffix.superSuffix().javaIdentifier().getText
            val name = StringNode(text, getSourceLocation(explSuffix), getWhitespaceLeft(explSuffix))
            val arguments = Option(explSuffix.superSuffix().arguments()).map(visitArguments)
            GenericInvocationExpression(typeParam, name, arguments, getSourceLocation(ctx))
          } else {
            val sup = explSuffix.SUPER()
            val name = StringNode(sup.getText, getSourceLocation(sup), getWhitespaceLeft(sup))
            val arguments = visitArguments(explSuffix.superSuffix().arguments())
            GenericInvocationExpression(typeParam, name, Some(arguments), getSourceLocation(ctx))
          }
        } else {
          val name = visitJavaIdentifier(explSuffix.javaIdentifier())
          // todo: predicateentrytype (predicate families)
          val arguments = visitArguments(explSuffix.arguments())
          GenericInvocationExpression(typeParam, name, Some(arguments), getSourceLocation(ctx))
        }
      }
    } else if(ctx.valExpr() != null) {
      visitValExpr(ctx.valExpr())
    } else {
      println(s"Error: unexpected alternative in Primary: ${ctx.getText}")
      null
    }
  }

  override def visitCreator(ctx: JavaParser.CreatorContext): NewExpression = {
    val typeIndication = visitCreatedName(ctx.createdName())
    if (ctx.nonWildcardTypeArguments() != null
        || (ctx.creatorRest() != null && ctx.creatorRest().classCreatorRest() != null)) {
      val typeArgumentClause = Option(ctx.nonWildcardTypeArguments()).map(visitNonWildcardTypeArguments)
      val clsCreator = if(ctx.classCreatorRest() != null) ctx.classCreatorRest()
                       else ctx.creatorRest().classCreatorRest()
      val args = visitArguments(clsCreator.arguments())
      val body = Option(clsCreator.classBody()).map(visitClassBody)
      NewObjectExpression(typeArgumentClause, typeIndication, args, body, getSourceLocation(ctx))
    } else if(ctx.creatorRest().arrayCreatorRest() != null) {
      val arrCtx = ctx.creatorRest().arrayCreatorRest()
      var dimensions: List[Dimension] = Nil
      if(arrCtx.specifiedDims() != null) {
        dimensions = visitSpecifiedDimsAsList(arrCtx.specifiedDims())
      }
      if(arrCtx.dims() != null) {
        dimensions = dimensions.appendedAll(visitDimsAsList(arrCtx.dims()))
      }

      val initializer = Option(arrCtx.arrayInitializer()).map(visitArrayInitializer)
      NewArrayExpression(typeIndication, dimensions, initializer, getSourceLocation(ctx))
    } else {
      println(s"Error: unexpected alternative in Creator: ${ctx.getText}")
      null
    }
  }

  def visitSpecifiedDimsAsList(ctx: JavaParser.SpecifiedDimsContext): List[Dimension] = {
    val size: Option[Expression] = Some(visitExpression(ctx.specifiedDim().expression()))
    val dim = Dimension(size, getSourceLocation(ctx.specifiedDim()))
    dim.openBracket.precedingWhitespace = getWhitespaceLeft(ctx.specifiedDim().LBRACK())
    dim.closeBracket.precedingWhitespace = getWhitespaceLeft(ctx.specifiedDim().RBRACK())
    if(ctx.specifiedDims() != null) {
      visitSpecifiedDimsAsList(ctx.specifiedDims()).prepended(dim)
    } else {
      List(dim)
    }
  }

  override def visitCreatedName(ctx: JavaParser.CreatedNameContext): TypeIndication = {
    val text = printer.visitSkipFirstWS(ctx)
    new StringTypeIndication(text, Nil, ctx.getText, getSourceLocation(ctx), getWhitespaceLeft(ctx))
  }

  override def visitInnerCreator(ctx: JavaParser.InnerCreatorContext): NewObjectExpression = {
    var fulltext = printer.visitSkipFirstWS(ctx.javaIdentifier())
    var text = ctx.javaIdentifier().getText
    if (ctx.nonWildcardTypeArgumentsOrDiamond() != null) {
      fulltext += printer.visit(ctx.nonWildcardTypeArgumentsOrDiamond())
      text += ctx.nonWildcardTypeArgumentsOrDiamond().getText
    }
    val typeIndication = new StringTypeIndication(fulltext, Nil, text,
                                                  getSourceLocation(ctx.javaIdentifier()),
                                                  getWhitespaceLeft(ctx))
    val args = visitArguments(ctx.classCreatorRest().arguments())
    val body = Option(ctx.classCreatorRest().classBody()).map(visitClassBody)
    NewObjectExpression(None, typeIndication, args, body, getSourceLocation(ctx))
  }

  override def visitExplicitGenericInvocation(ctx: JavaParser.ExplicitGenericInvocationContext):
                  GenericInvocationExpression = {
    val typeParam = visitNonWildcardTypeArguments(ctx.nonWildcardTypeArguments())
    if (ctx.explicitGenericInvocationSuffix().SUPER() != null) {
      if (ctx.explicitGenericInvocationSuffix().superSuffix().javaIdentifier() != null) {
        var text = ctx.explicitGenericInvocationSuffix().SUPER().getText
        text += getWhitespaceLeft(ctx.explicitGenericInvocationSuffix().superSuffix().DOT())
        text += ctx.explicitGenericInvocationSuffix().superSuffix().DOT().getText
        text += getWhitespaceLeft(ctx.explicitGenericInvocationSuffix().superSuffix().javaIdentifier())
        text += ctx.explicitGenericInvocationSuffix().superSuffix().javaIdentifier().getText
        val name = StringNode(text, getSourceLocation(ctx), getWhitespaceLeft(ctx))
        val arguments = Option(ctx.explicitGenericInvocationSuffix().superSuffix().arguments()).map(visitArguments)
        GenericInvocationExpression(typeParam, name, arguments, getSourceLocation(ctx))
      } else {
        val sup = ctx.explicitGenericInvocationSuffix().SUPER()
        val name = StringNode(sup.getText, getSourceLocation(sup), getWhitespaceLeft(sup))
        val arguments = visitArguments(ctx.explicitGenericInvocationSuffix().superSuffix().arguments())
        GenericInvocationExpression(typeParam, name, Some(arguments), getSourceLocation(ctx))
      }
    } else {
      val name = visitJavaIdentifier(ctx.explicitGenericInvocationSuffix().javaIdentifier())
      // todo: predicateentrytype (predicate families)
      val arguments = visitArguments(ctx.explicitGenericInvocationSuffix().arguments())
      GenericInvocationExpression(typeParam, name, Some(arguments), getSourceLocation(ctx))
    }
  }

  override def visitNonWildcardTypeArguments(ctx: JavaParser.NonWildcardTypeArgumentsContext):
                TypeArgumentClause = {
    val types = visitTypeListAsList(ctx.typeList())
    val clause = TypeArgumentClause(types, getSourceLocation(ctx))
    clause.openAngle.precedingWhitespace = getWhitespaceLeft(ctx.LT())
    clause.closeAngle.precedingWhitespace = getWhitespaceLeft(ctx.GT())
    clause
  }

  override def visitArguments(ctx: JavaParser.ArgumentsContext): Arguments = {
    val args = if(ctx.expressionList() == null) Nil
                else visitExpressionListAsList(ctx.expressionList())
    val arguments = Arguments(args, getSourceLocation(ctx))
    arguments.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.LPAREN())
    arguments.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.RPAREN())
    arguments
  }


  def visitValExpressionListAsList(ctx: JavaParser.ValExpressionListContext): List[Expression] = {
    val expr = visitExpression(ctx.langExpr().expression())
    if(ctx.valExpressionList() == null) {
      List(expr)
    } else {
      expr.trailingWhitespace = getWhitespaceRight(ctx.langExpr())
      val remainder = visitValExpressionListAsList(ctx.valExpressionList())
      remainder.prepended(expr)
    }
  }

  def visitValIdListAsList(ctx: JavaParser.ValIdListContext): List[StringNode] = {
    val id = visitJavaIdentifier(ctx.langId().javaIdentifier())
    if(ctx.valIdList() == null) {
      List(id)
    } else {
      id.trailingWhitespace = getWhitespaceRight(ctx.langId())
      val remainder = visitValIdListAsList(ctx.valIdList())
      remainder.prepended(id)
    }
  }

  def visitValTypeListAsList(ctx: JavaParser.ValTypeListContext): List[TypeIndication] = {
    val typ = visitType(ctx.langType().`type`())
    if(ctx.valTypeList() == null) {
      List(typ)
    } else {
      typ.trailingWhitespace = getWhitespaceRight(ctx.langType())
      val remainder = visitValTypeListAsList(ctx.valTypeList())
      remainder.prepended(typ)
    }
  }

  override def visitValContractClause(ctx: JavaParser.ValContractClauseContext): ContractClause = {
    try {
      if (ctx.VAL_MODIFIES() != null || ctx.VAL_ACCESSIBLE() != null) {
        throw new NotSupportedException("contract clauses of process models", getSourceLocation(ctx))
      } else if(ctx.VAL_REQUIRES() != null) {
        val expr = visitExpression(ctx.langExpr().expression())
        val clause = Precondition(expr, getSourceLocation(ctx))
        clause.requiresKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_REQUIRES())
        clause.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
        clause
      } else if(ctx.VAL_ENSURES() != null) {
        val expr = visitExpression(ctx.langExpr().expression())
        val clause = Postcondition(expr, getSourceLocation(ctx))
        clause.ensuresKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_ENSURES())
        clause.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
        clause
      } else if(ctx.VAL_GIVEN() != null) {
        val name = visitJavaIdentifier(ctx.langId().javaIdentifier())
        val argType = visitType(ctx.langType().`type`())
        val ghost = GhostArgument(argType, name, getSourceLocation(ctx))
        ghost.givenKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_GIVEN())
        ghost.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
        ghost
      } else if(ctx.VAL_YIELDS() != null) {
        val name = visitJavaIdentifier(ctx.langId().javaIdentifier())
        val argType = visitType(ctx.langType().`type`())
        val ghost = GhostReturn(argType, name, getSourceLocation(ctx))
        ghost.yieldsKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_YIELDS())
        ghost.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
        ghost
      } else if(ctx.VAL_CONTEXT_EVERYWHERE() != null) {
        val expr = visitExpression(ctx.langExpr().expression())
        val clause = ContextEverywhere(expr, getSourceLocation(ctx))
        clause.contextEverywhereKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_CONTEXT_EVERYWHERE())
        clause.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
        clause
      } else if(ctx.VAL_CONTEXT() != null) {
        val expr = visitExpression(ctx.langExpr().expression())
        val clause = Context(expr, getSourceLocation(ctx))
        clause.contextKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_CONTEXT())
        clause.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
        clause
      } else if(ctx.VAL_LOOP_INVARIANT() != null) {
        val expr = visitExpression(ctx.langExpr().expression())
        val clause = LoopInvariant(expr, getSourceLocation(ctx))
        clause.invariantKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_LOOP_INVARIANT())
        clause.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
        clause
      } else if(ctx.VAL_KERNEL_INVARIANT() != null) {
        throw new NotSupportedException("kernel_invariant", getSourceLocation(ctx))
      } else if(ctx.VAL_SIGNALS() != null) {
        val exType = visitType(ctx.langType().`type`())
        val exId = visitJavaIdentifier(ctx.langId().javaIdentifier())
        val expr = visitExpression(ctx.langExpr().expression())
        val clause = ExceptionalPostcondition(exType, Some(exId), Some(expr), getSourceLocation(ctx))
        clause.signalsKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_SIGNALS())
        clause.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.LPAREN())
        clause.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.RPAREN())
        clause.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
        clause
      } else if(ctx.VAL_LOCK_INVARIANT() != null) {
        // todo
        ???
      } else if(ctx.VAL_DECREASES() != null) {
        var expr: Expression = null
        if(ctx.valDecreasesMeasure() == null) {
          // this means the function cannot be involved in recursion, and all called functions must terminate
          expr = Literal(StringNode("1", getSourceLocation(ctx), " "), getSourceLocation(ctx))
        } else if(ctx.valDecreasesMeasure().VAL_ASSUME() != null) {
          throw new NotSupportedException("Assumed termination", getSourceLocation(ctx))
        } else if(ctx.valDecreasesMeasure().valExpressionList().valExpressionList() != null) {
          throw new NotSupportedException("Multiple termination measures", getSourceLocation(ctx.valDecreasesMeasure()))
        }
        expr = visitExpression(ctx.valDecreasesMeasure().valExpressionList().langExpr().expression())
        val clause = DecreasesClause(expr, getSourceLocation(ctx))
        clause.decreasesKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_DECREASES())
        clause.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
        clause
      } else {
        println(s"Error: unexpected alternative in ContractClause: ${ctx.getText}")
        null
      }
    }catch {
      case e: NotSupportedException =>
        val commented = NonTranslatedContractClause("", getSourceLocation(ctx))
        setNonTranslatedTreeNodeText(commented, ctx, e)
        commented
    }
  }

  override def visitValBlock(ctx: JavaParser.ValBlockContext): Statement = {
    val statements = ctx.valStatement().asScala.map(visitValStatement).toList
    val block = Block(None, statements, getSourceLocation(ctx))
    block.openBrace.precedingWhitespace = getWhitespaceLeft(ctx.LBRACE())
    block.closeBrace.precedingWhitespace = getWhitespaceLeft(ctx.RBRACE())
    block
  }

  def visitValStatement(ctx: JavaParser.ValStatementContext): Statement = {
    try {
      visit(ctx).asInstanceOf[Statement]
    } catch {
      case e: NotSupportedException =>
        val commented = NonTranslatedStatement("", getSourceLocation(ctx))
        setNonTranslatedTreeNodeText(commented, ctx, e)
        commented
    }

  }

  /* todo */
//  override def visitValCreateWand(ctx: JavaParser.ValCreateWandContext): ??? = {
//    // todo
//  }
//
//  override def visitValQedWand(ctx: JavaParser.ValQedWandContext): ??? = {
//    // todo
//  }
//
//  override def visitValApplyWand(ctx: JavaParser.ValApplyWandContext): ??? = {
//    // todo
//  }
//
//  override def visitValUseWand(ctx: JavaParser.ValUseWandContext): ??? = {
//    // todo
//  }
//
  override def visitValFold(ctx: JavaParser.ValFoldContext): FoldStatement = {
    val expr = visitExpression(ctx.langExpr().expression())
    val fold = FoldStatement(expr, getSourceLocation(ctx))
    fold.foldKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_FOLD())
    fold.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
    fold
  }

  override def visitValUnfold(ctx: JavaParser.ValUnfoldContext): UnfoldStatement = {
    val expr = visitExpression(ctx.langExpr().expression())
    val fold = UnfoldStatement(expr, getSourceLocation(ctx))
    fold.unfoldKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_UNFOLD())
    fold.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
    fold
  }

//  override def visitValOpen(ctx: JavaParser.ValOpenContext): ??? = {
//    // todo
//  }
//
//  override def visitValClose(ctx: JavaParser.ValCloseContext): ??? = {
//    // todo
//  }

  override def visitValAssert(ctx: JavaParser.ValAssertContext): AssertStatement = {
    val content = visitExpression(ctx.langExpr().expression())
    val stmt = AssertStatement(content, None, getSourceLocation(ctx))
    stmt.assertKeyword.precedingWhitespace = getWhitespaceLeft(ctx.ASSERT())
    stmt.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
    stmt
  }

  override def visitValAssume(ctx: JavaParser.ValAssumeContext): AssumeStatement = {
    val content = visitExpression(ctx.langExpr().expression())
    val stmt = AssumeStatement(content, None, getSourceLocation(ctx))
    stmt.assumeKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_ASSUME())
    stmt.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
    stmt
  }

  override def visitValInhale(ctx: JavaParser.ValInhaleContext): AssumeStatement = {
    val expr = visitExpression(ctx.langExpr().expression())
    val res = AssumeStatement(expr, None, getSourceLocation(ctx))
    res.assumeKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_INHALE())
    res.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
    res
  }

  override def visitValExhale(ctx: JavaParser.ValExhaleContext): AssertStatement = {
    val expr = visitExpression(ctx.langExpr().expression())
    val res = AssertStatement(expr, None, getSourceLocation(ctx))
    res.assertKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_EXHALE())
    res.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
    res
  }

//  override def visitValLabel(ctx: JavaParser.ValLabelContext): ??? = {
//    // todo
//  }

  override def visitValRefute(ctx: JavaParser.ValRefuteContext): AssertStatement = {
    val expr = visitExpression(ctx.langExpr().expression())
    val parExpr = ParExpression(expr, getSourceLocation(ctx.langExpr()))
    val negExpr = PrefixExpression(parExpr, StringNode("!", getSourceLocation(ctx)), getSourceLocation(ctx))
    val ass = AssertStatement(negExpr, None, getSourceLocation(ctx))
    ass.assertKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_REFUTE())
    ass
  }

  override def visitValWitness(ctx: JavaParser.ValWitnessContext): Statement = {
    throw new NotSupportedException("witness", getSourceLocation(ctx))
  }

  override def visitValGhost(ctx: JavaParser.ValGhostContext): GhostStatement = {
    val stmt = visitBlockStatement(ctx.langStatement().blockStatement())
    val ghost = GhostStatement(stmt, getSourceLocation(ctx))
    ghost.ghostKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_GHOST())
    ghost
  }

//  override def visitValSend(ctx: JavaParser.ValSendContext): ??? = {
//    // todo
//  }
//
//  override def visitValRecv(ctx: JavaParser.ValRecv): ??? = {
//    // todo
//  }
//
//  override def visitValTransfer(ctx: JavaParser.ValTransferContext): ??? = {
//    // todo
//  }
//
//  override def visitValCslSubject(ctx: JavaParser.ValCslSubjectContext): ??? = {
//    // todo
//  }
//
//  override def visitValSpecIgnoreStart(ctx: JavaParser.ValSpecIgnoreStartContext): ??? = {
//    // todo
//  }
//
//  override def visitValSpecIgnoreEnd(ctx: JavaParser.ValSpecIgnoreEndContext): ??? = {
//    // todo
//  }
//
//  override def visitValActionModel(ctx: JavaParser.ValActionModelContext): ??? = {
//    // todo
//  }
//
//  override def visitValAtomic(ctx: JavaParser.ValAtomicContext): ??? = {
//    // todo
//  }

  override def visitValPostfix(ctx: JavaParser.ValPostfixContext): IndexExpression = {
    if(ctx.Arrow() != null) {
      // todo
      ???
    } else {
      val expr = visitExpression(ctx.langExpr(0).expression())
      if(ctx.langExpr().size() > 1) {
        val right = visitExpression(ctx.langExpr(1).expression())
        val range = RangeExpression(expr, right, getSourceLocation(ctx))
        range.dots.precedingWhitespace = getWhitespaceLeft(ctx.RANGE_TO())
        range
      } else {
        if (ctx.langExpr(0).getStart.getTokenIndex < ctx.RANGE_TO().getSymbol.getTokenIndex) {
          val range = PartialRangeExpression(Some(expr), None, getSourceLocation(ctx))
          range.dots.precedingWhitespace = getWhitespaceLeft(ctx.RANGE_TO())
          range
        } else {
          val range = PartialRangeExpression(None, Some(expr), getSourceLocation(ctx))
          range.dots.precedingWhitespace = getWhitespaceLeft(ctx.RANGE_TO())
          range
        }
      }
    }
  }

  override def visitValCardinality(ctx: JavaParser.ValCardinalityContext): Expression = {
    val expr = visitExpression(ctx.langExpr().expression())
    val card = CardinalityExpression(expr, getSourceLocation(ctx))
    card.leftBar.precedingWhitespace = getWhitespaceLeft(ctx.BITOR(0))
    card.rightBar.precedingWhitespace = getWhitespaceLeft(ctx.BITOR(1))
    card
  }

  override def visitValArrayValues(ctx: JavaParser.ValArrayValuesContext): Expression = {
    // todo: shall we support?
    ???
  }

  override def visitValSome(ctx: JavaParser.ValSomeContext): Expression = {
    val expr = visitExpression(ctx.langExpr().expression())
    val some = OptionSomeExpression(expr, getSourceLocation(ctx))
    some.keyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_SOME())
    some.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.LPAREN())
    some.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.RPAREN())
    some
  }

  // todo: Left/Right for Either
  // todo: set comprehension

  override def visitValTypedLiteralSeq(ctx: JavaParser.ValTypedLiteralSeqContext): Expression = {
    val innerType = visitType(ctx.langType().`type`())
    val vals = Option(ctx.valExpressionList()).flatMap(visitValExpressionListAsList).toList
    val seq = TypedSeqLiteral(innerType, vals, getSourceLocation(ctx))
    seq.adtType.precedingWhitespace = getWhitespaceLeft(ctx.VAL_SEQ())
    seq.openAngle.precedingWhitespace = getWhitespaceLeft(ctx.LT())
    seq.closeAngle.precedingWhitespace = getWhitespaceLeft(ctx.GT())
    seq.openBrace.precedingWhitespace = getWhitespaceLeft(ctx.LBRACE())
    seq.closeBrace.precedingWhitespace = getWhitespaceLeft(ctx.RBRACE())
    seq
  }

  override def visitValTypedLiteralSet(ctx: JavaParser.ValTypedLiteralSetContext): Expression = {
    val innerType = visitType(ctx.langType().`type`())
    val vals = Option(ctx.valExpressionList()).flatMap(visitValExpressionListAsList).toList
    val set = TypedSetLiteral(innerType, vals, getSourceLocation(ctx))
    set.adtType.precedingWhitespace = getWhitespaceLeft(ctx.VAL_SET())
    set.openAngle.precedingWhitespace = getWhitespaceLeft(ctx.LT())
    set.closeAngle.precedingWhitespace = getWhitespaceLeft(ctx.GT())
    set.openBrace.precedingWhitespace = getWhitespaceLeft(ctx.LBRACE())
    set.closeBrace.precedingWhitespace = getWhitespaceLeft(ctx.RBRACE())
    set
  }

  override def visitValTypedLiteralBag(ctx: JavaParser.ValTypedLiteralBagContext): Expression = {
    val innerType = visitType(ctx.langType().`type`())
    val vals = Option(ctx.valExpressionList()).flatMap(visitValExpressionListAsList).toList
    val bag = TypedBagLiteral(innerType, vals, getSourceLocation(ctx))
    bag.adtType.precedingWhitespace = getWhitespaceLeft(ctx.VAL_BAG())
    bag.openAngle.precedingWhitespace = getWhitespaceLeft(ctx.LT())
    bag.closeAngle.precedingWhitespace = getWhitespaceLeft(ctx.GT())
    bag.openBrace.precedingWhitespace = getWhitespaceLeft(ctx.LBRACE())
    bag.closeBrace.precedingWhitespace = getWhitespaceLeft(ctx.RBRACE())
    bag
  }

  // todo: Map, Tuple, Set-/Seq-/BagLiteral (also respective "empty" version), Range

  override def visitValRangeBinding(ctx: JavaParser.ValRangeBindingContext): QuantifiedVarDeclaration = {
    val ident = visitJavaIdentifier(ctx.langId().javaIdentifier())
    ident.trailingWhitespace = getWhitespaceLeft(ctx.ASSIGN())
    QuantifiedVarDeclaration(visitType(ctx.langType().`type`()),
                            List(ident),
                            getSourceLocation(ctx.langId()))
  }

  override def visitValNormalBinding(ctx: JavaParser.ValNormalBindingContext): QuantifiedVarDeclaration = {
    val typ = visitType(ctx.valArg().langType().`type`())
    val name = visitJavaIdentifier(ctx.valArg().langId().javaIdentifier())
    new QuantifiedVarDeclaration(typ, List(name), getSourceLocation(ctx.valArg()))
  }

  def visitValBindingsAsList(ctx: JavaParser.ValBindingsContext): List[QuantifiedVarDeclaration] = {
    val decl = visit(ctx.valBinding()).asInstanceOf[QuantifiedVarDeclaration]
    if(ctx.valBindings() == null) {
      List(decl)
    } else {
      val list = visitValBindingsAsList(ctx.valBindings())
      decl.trailingWhitespace = getWhitespaceLeft(ctx.COMMA())
      list.prepended(decl)
    }
  }

  def extractGuard(ctx: JavaParser.ValNormalBindingContext): Option[Expression] = {
    None
  }
  def extractGuard(ctx: JavaParser.ValRangeBindingContext): Option[Expression] = {
    val ident = visitJavaIdentifier(ctx.langId().javaIdentifier())
    val lower = BinaryExpression(visitExpression(ctx.langExpr(0).expression()),
      StringNode("<=", getSourceLocation(ctx), " "),
      ident,
      getSourceLocation(ctx))
    val upper = BinaryExpression(ident,
      StringNode("<", getSourceLocation(ctx), " "),
      visitExpression(ctx.langExpr(1).expression()),
      getSourceLocation(ctx))
    Some(BinaryExpression(lower,
      StringNode("&&", getSourceLocation(ctx), getWhitespaceLeft(ctx.RANGE_TO())),
      upper,
      getSourceLocation(ctx)))
  }

  def extractGuard(ctx: JavaParser.ValBindingsContext): Option[Expression] = {
    val rest = Option(ctx.valBindings()).flatMap(extractGuard)
    val guard = ctx.valBinding() match {
      case n: JavaParser.ValNormalBindingContext => extractGuard(n)
      case r: JavaParser.ValRangeBindingContext => extractGuard(r)
    }
    if(guard.isDefined) {
      if (rest.isDefined) {
        Some(BinaryExpression(guard.get, StringNode("&&", getSourceLocation(ctx), " "),
                              rest.get, getSourceLocation(ctx)))
      } else {
        guard
      }
    } else {
      rest
    }
  }

  override def visitValQuantifier(ctx: JavaParser.ValQuantifierContext): QuantifiedExpression = {
    val decls = visitValBindingsAsList(ctx.valBindings())
    var guard: Option[Expression] = extractGuard(ctx.valBindings())
    var body: Expression = null
    if(ctx.valBinderCont() != null) {
      val expr = visitExpression(ctx.langExpr().expression())
      if(guard.isEmpty) {
        guard = Some(expr)
      } else {
        guard = Some(BinaryExpression(guard.get, StringNode("&&", getSourceLocation(ctx), " "),
                                      expr, getSourceLocation(ctx)))

      }
      body = visitExpression(ctx.valBinderCont().langExpr().expression())
    } else {
      body = visitExpression(ctx.langExpr().expression())
    }
    var quant: QuantifiedExpression = null
    ctx.valBinderSymbol() match {
      case _: JavaParser.ValForallSymbContext =>
        quant = ForallExpression(decls, guard, body, getSourceLocation(ctx))
      case _: JavaParser.ValStarallSymbContext =>
        quant = ForallStarExpression(decls, guard, body, getSourceLocation(ctx))
      case _: JavaParser.ValExistsSymbContext =>
        quant = ExistsExpression(decls, guard, body, getSourceLocation(ctx))
      case _ =>
        // should be unreachable
        ???
    }
    quant.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.LPAREN())
    quant.quantifier.precedingWhitespace = getWhitespaceLeft(ctx.valBinderSymbol())
    quant.firstSemicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
    if(ctx.valBinderCont() != null) {
      quant.secondSemicolon.precedingWhitespace = getWhitespaceLeft(ctx.valBinderCont().SEMI())
    }
    quant.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.RPAREN())
    quant
  }

  override def visitValLet(ctx: JavaParser.ValLetContext): LetExpression = {
    val qType = visitType(ctx.langType().`type`())
    val name = visitJavaIdentifier(ctx.langId().javaIdentifier())
    val definition = visitExpression(ctx.langExpr(0).expression())
    val varDecl = VariableDeclarator(name, Some(definition), getSourceLocation(ctx))
    val expr = visitExpression(ctx.langExpr(1).expression())
    val let = LetExpression(qType, varDecl, expr, getSourceLocation(ctx))
    let.letKeyword.precedingWhitespace = getWhitespaceLeft(ctx.LET())
    let.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.LPAREN())
    let.semicolon.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
    let.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.RPAREN())
    let
  }


  override def visitValPerm(ctx: JavaParser.ValPermContext): PermissionPredicate = {
    val target = visitExpression(ctx.langExpr(0).expression())
    val amount = visitExpression(ctx.langExpr(1).expression())
    val perm = PermissionPredicate(target, Some(amount), getSourceLocation(ctx))
    perm.permKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_PERM())
    perm.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.LPAREN())
    perm.comma.precedingWhitespace = getWhitespaceLeft(ctx.COMMA())
    perm.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.RPAREN())
    perm
  }

  override def visitValCurPerm(ctx: JavaParser.ValCurPermContext): CurrentPermissionExpression = {
    val target = visitExpression(ctx.langExpr().expression())
    val perm = CurrentPermissionExpression(target, getSourceLocation(ctx))
    perm.permKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_PERM_VAL())
    perm.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.LPAREN())
    perm.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.RPAREN())
    perm
  }

  override def visitValValue(ctx: JavaParser.ValValueContext): SomePermissionPredicate = {
    val target = visitExpression(ctx.langExpr().expression())
    val perm = SomePermissionPredicate(target, getSourceLocation(ctx))
    perm.valueKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_VALUE())
    perm.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.LPAREN())
    perm.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.RPAREN())
    perm
  }

  override def visitValPointsTo(ctx: JavaParser.ValPointsToContext): ValuedPermissionPredicate = {
    val target = visitExpression(ctx.langExpr(0).expression())
    val value = visitExpression(ctx.langExpr(1).expression())
    val amount = visitExpression(ctx.langExpr(2).expression())
    val perm = ValuedPermissionPredicate(target, value, Some(amount), getSourceLocation(ctx))
    perm.pointsToKeyword.precedingWhitespace = getWhitespaceLeft(ctx.VAL_POINTS_TO())
    perm.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.LPAREN())
    perm.firstComma.precedingWhitespace = getWhitespaceLeft(ctx.COMMA(0))
    perm.secondComma.precedingWhitespace = getWhitespaceLeft(ctx.COMMA(1))
    perm.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.RPAREN())
    perm
  }

  override def visitValArray(ctx: JavaParser.ValArrayContext): ArrayExpression = {
    val name = visitExpression(ctx.langExpr(0).expression())
    val size = visitExpression(ctx.langExpr(1).expression())
    val arr = ArrayExpression(name, size, getSourceLocation(ctx))
    arr.arrayKeyword.precedingWhitespace = getWhitespaceLeft(ctx.ARRAY())
    arr.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.LPAREN())
    arr.comma.precedingWhitespace = getWhitespaceLeft(ctx.COMMA())
    arr.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.RPAREN())
    arr
  }

  // todo: Pointers, Matrix, HPerm, APerm, PrimaryVector, Reducible/Contribution, idle/running

  override def visitValExpr(ctx: JavaParser.ValExprContext): Expression = {
    if(ctx.valPrimary() != null) {
      visitValPrimary(ctx.valPrimary())
    } else {
      visitValKeywordExpr(ctx.valKeywordExpr())
    }
  }

  override def visitValAny(ctx: JavaParser.ValAnyContext): IndexAllExpression = {
    IndexAllExpression(getSourceLocation(ctx), getWhitespaceLeft(ctx.MUL()))
  }

  override def visitValScale(ctx: JavaParser.ValScaleContext): ScaledPredicateExpression = {
    val amount = visitExpression(ctx.langExpr(0).expression())
    val predicate = visitExpression(ctx.langExpr(1).expression())
    val scaled = ScaledPredicateExpression(amount, predicate, getSourceLocation(ctx))
    scaled.openBracket.precedingWhitespace = getWhitespaceLeft(ctx.LBRACK())
    scaled.closeBracket.precedingWhitespace = getWhitespaceLeft(ctx.RBRACK())
    scaled
  }

  override def visitValInlinePattern(ctx: JavaParser.ValInlinePatternContext): InlinePattern = {
    val expr = visitExpression(ctx.langExpr().expression())
    val open = StringNode(ctx.TRIGGER_OPEN().getText,
                          getSourceLocation(ctx.TRIGGER_OPEN()), getWhitespaceLeft(ctx.TRIGGER_OPEN()))
    val pattern = InlinePattern(open, expr, getSourceLocation(ctx))
    pattern.closePattern.precedingWhitespace = getWhitespaceLeft(ctx.TRIGGER_CLOSE())
    pattern
  }

  override def visitValUnfolding(ctx: JavaParser.ValUnfoldingContext): UnfoldingExpression = {
    val predicate = visitExpression(ctx.langExpr(0).expression())
    val expr = visitExpression(ctx.langExpr(1).expression())
    val unfold = UnfoldingExpression(predicate, expr, getSourceLocation(ctx))
    unfold.unfoldingKeyword.precedingWhitespace = getWhitespaceLeft(ctx.UNFOLDING())
    unfold.inKeyword.precedingWhitespace = getWhitespaceLeft(ctx.IN())
    unfold
  }

  override def visitValOld(ctx: JavaParser.ValOldContext): OldExpression = {
    val expr = visitExpression(ctx.langExpr().expression())
    val old = OldExpression(expr, None, getSourceLocation(ctx))
    old.oldKeyword.precedingWhitespace = getWhitespaceLeft(ctx.OLD())
    old.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.LPAREN())
    old.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.RPAREN())
    old
  }

  override def visitValOldLabeled(ctx: JavaParser.ValOldLabeledContext): OldExpression = {
    val expr = visitExpression(ctx.langExpr().expression())
    val label = visitJavaIdentifier(ctx.langId().javaIdentifier())
    val old = OldExpression(expr, Some(label), getSourceLocation(ctx))
    old.oldKeyword.precedingWhitespace = getWhitespaceLeft(ctx.OLD())
    old.openBracket.precedingWhitespace = getWhitespaceLeft(ctx.LBRACK())
    old.closeBracket.precedingWhitespace = getWhitespaceLeft(ctx.RBRACK())
    old.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.LPAREN())
    old.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.RPAREN())
    old
  }

  override def visitValTypeof(ctx: JavaParser.ValTypeofContext): TypeOfExpression = {
    val expr = visitExpression(ctx.langExpr().expression())
    val typeof = TypeOfExpression(expr, getSourceLocation(ctx))
    typeof.keyword.precedingWhitespace = getWhitespaceLeft(ctx.TYPEOF())
    typeof.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.LPAREN())
    typeof.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.RPAREN())
    typeof
  }

  override def visitValTypeValue(ctx: JavaParser.ValTypeValueContext): TypeExpression = {
    val ind = visitType(ctx.langType().`type`())
    val expr = TypeExpression(ind, getSourceLocation(ctx))
    expr.typeKeyword.precedingWhitespace = getWhitespaceLeft(ctx.TYPEVALUE())
    expr.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.LPAREN())
    expr.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.RPAREN())
    expr
  }

  override def visitValIdEscape(ctx: JavaParser.ValIdEscapeContext): StringNode = {
    printNode(ctx)
  }

  override def visitValPrimaryContext(ctx: JavaParser.ValPrimaryContextContext): Expression = {
    if(ctx.RESULT() != null) {
      ResultExpression(getSourceLocation(ctx), getWhitespaceLeft(ctx.RESULT()))
    } else {
      VerCorsSpecialVariable(ctx.getText, getSourceLocation(ctx), getWhitespaceLeft(ctx))
    }
  }

  // todo typeof, type, held, genericAdtInvocation

  def visitValPrimary(ctx: JavaParser.ValPrimaryContext): Expression = {
    visit(ctx).asInstanceOf[Expression]
  }
//    ctx.altNum match {
//      case 1 =>
//        StringNode(ctx.RESULT().getText, getSourceLocation(ctx.RESULT()),
//          getWhitespaceLeft(ctx.RESULT()))
//      case 2 =>
//        val expr = visitExpression(ctx.langExpr().expression())
//        if (ctx.OLD() != null) {
//          var label: Option[StringNode] = None
//          if(ctx.langId().size() > 0) {
//            label = Some(visitJavaIdentifier(ctx.langId(0).javaIdentifier()))
//          }
//          val old = OldExpression(expr, label, getSourceLocation(ctx))
//          old.oldKeyword.precedingWhitespace = getWhitespaceLeft(ctx.OLD())
//          old.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.LPAREN())
//          old.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.RPAREN())
//          if (ctx.COMMA().size() > 0) {
//            old.comma.precedingWhitespace = getWhitespaceLeft(ctx.COMMA(0))
//          }
//          old
//        } else {
//          val pre = PreExpression(expr, getSourceLocation(ctx))
//          pre.keyword.precedingWhitespace = getWhitespaceLeft(ctx.OLD())
//          pre.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.LPAREN())
//          pre.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.RPAREN())
//          pre
//        }
//      case 3 =>
//        val storeRefs = visitStoreRefListAsList(ctx.storeRefList())
//        val expr = NotAssignedExpression(storeRefs, getSourceLocation(ctx))
//        expr.keyword.precedingWhitespace = getWhitespaceLeft(ctx.NOT_ASSIGNED())
//        expr.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.LPAREN())
//        expr.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.RPAREN())
//        expr
//      case 4 =>
//        val storeRefs = visitStoreRefListAsList(ctx.storeRefList())
//        val expr = NotModifiedExpression(storeRefs, getSourceLocation(ctx))
//        expr.keyword.precedingWhitespace = getWhitespaceLeft(ctx.NOT_MODIFIED())
//        expr.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.LPAREN())
//        expr.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.RPAREN())
//        expr
//      case 5 =>
//        val storeRefs = visitStoreRefListAsList(ctx.storeRefList())
//        val expr = OnlyAccessedExpression(storeRefs, getSourceLocation(ctx))
//        expr.keyword.precedingWhitespace = getWhitespaceLeft(ctx.ONLY_ACCESSED())
//        expr.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.LPAREN())
//        expr.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.RPAREN())
//        expr
//      case 6 =>
//        val storeRefs = visitStoreRefListAsList(ctx.storeRefList())
//        val expr = OnlyAssignedExpression(storeRefs, getSourceLocation(ctx))
//        expr.keyword.precedingWhitespace = getWhitespaceLeft(ctx.ONLY_ASSIGNED())
//        expr.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.LPAREN())
//        expr.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.RPAREN())
//        expr
//      case 9 =>
//        val expressions = ctx.expressionList().expression().asScala.map(visitExpression).toList
//        val expr = FreshExpression(expressions, getSourceLocation(ctx))
//        expr.freshKeyword.precedingWhitespace = getWhitespaceLeft(ctx.FRESH())
//        expr.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.LPAREN())
//        expr.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.RPAREN())
//        expr
//      case 14 =>
//        val langExpr = visitExpression(ctx.langExpr().expression())
//        val expr = NonNullElementsExpression(langExpr, getSourceLocation(ctx))
//        expr.keyword.precedingWhitespace = getWhitespaceLeft(ctx.NONNULLELEMENTS())
//        expr.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.LPAREN())
//        expr.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.RPAREN())
//        expr
//      case 15 =>
//        val langExpr = visitExpression(ctx.langExpr().expression())
//        val expr = TypeOfExpression(langExpr, getSourceLocation(ctx))
//        expr.keyword.precedingWhitespace = getWhitespaceLeft(ctx.TYPEOF())
//        expr.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.LPAREN())
//        expr.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.RPAREN())
//        expr
//      case 24 =>
//        val quantExpr = visitSpecQuantifiedExpression(ctx.specQuantifiedExpression())
//        quantExpr.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.LPAREN())
//        quantExpr.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.RPAREN())
//        quantExpr
//      case 25 =>
//        val quantExpr =visitSpecQuantifiedExpression(ctx.specQuantifiedExpression())
//        quantExpr.openParenthesis.precedingWhitespace = quantExpr.quantifier.precedingWhitespace
//        quantExpr.quantifier.precedingWhitespace = ""
//        quantExpr
//      case 26 =>
//        StringNode(ctx.COUNT().getText, getSourceLocation(ctx.COUNT()), getWhitespaceLeft(ctx.COUNT()))
//      case 27 =>
//        StringNode(ctx.NOTHING().getText, getSourceLocation(ctx.NOTHING()), getWhitespaceLeft(ctx.NOTHING()))
//      case _ =>
//        StringNode("not yet implemented", null, " ")
//    }

  def visitValKeywordExpr(ctx: JavaParser.ValKeywordExprContext): Expression = {
    visit(ctx).asInstanceOf[Expression]
  }

  override def visitValNonePerm(ctx: JavaParser.ValNonePermContext): Literal = {
    Literal(StringNode("0", getSourceLocation(ctx), getWhitespaceLeft(ctx.NONE())), getSourceLocation(ctx))
  }

  override def visitValWrite(ctx: JavaParser.ValWriteContext): Literal = {
    Literal(StringNode("1", getSourceLocation(ctx), getWhitespaceLeft(ctx.WRITE())), getSourceLocation(ctx))
  }

  override def visitValRead(ctx: JavaParser.ValReadContext): StringNode = {
    StringNode(ctx.READ().getText, getSourceLocation(ctx), getWhitespaceLeft(ctx.READ()))
  }

  override def visitValNoneOption(ctx: JavaParser.ValNoneOptionContext): StringNode = {
    StringNode(ctx.OPTION_NONE().getText, getSourceLocation(ctx), getWhitespaceLeft(ctx.OPTION_NONE()))
  }

  override def visitValEmpty(ctx: JavaParser.ValEmptyContext): StringNode = {
    StringNode(ctx.EMPTY().getText, getSourceLocation(ctx), getWhitespaceLeft(ctx.EMPTY()))
  }

  override def visitValFalse(ctx: JavaParser.ValFalseContext): Literal = {
    Literal(StringNode(ctx.False().getText, getSourceLocation(ctx), getWhitespaceLeft(ctx.False())), getSourceLocation(ctx))
  }

  override def visitValTrue(ctx: JavaParser.ValTrueContext): Literal = {
    Literal(StringNode(ctx.True().getText, getSourceLocation(ctx), getWhitespaceLeft(ctx.True())), getSourceLocation(ctx))
  }

  def visitValGlobalDeclaration(ctx: JavaParser.ValGlobalDeclarationContext): SpecDeclaration = {
    try
      visit(ctx).asInstanceOf[SpecDeclaration]
    catch {
      case e: NotSupportedException =>
        val commented = NonTranslatedSpecDeclaration("", getSourceLocation(ctx))
        setNonTranslatedTreeNodeText(commented, ctx, e)
        commented
    }
  }
  def visitValClassDeclaration(ctx: JavaParser.ValClassDeclarationContext): SpecDeclaration = {
    try
      visit(ctx).asInstanceOf[SpecDeclaration]
    catch {
      case e: NotSupportedException =>
        val commented = NonTranslatedSpecDeclaration("", getSourceLocation(ctx))
        setNonTranslatedTreeNodeText(commented, ctx, e)
        commented
    }
  }

  override def visitValInstancePredicate(ctx: JavaParser.ValInstancePredicateContext): SpecDeclaration = {
    val modifiers = ctx.valModifier().asScala.map(visitValModifier).toList
    val name = visitJavaIdentifier(ctx.langId().javaIdentifier())
    var params: List[Parameter] = Nil
    if(ctx.valArgList() != null) {
      params = visitValArgListAsParameters(ctx.valArgList())
    }
    val body = visitValDef(ctx.valDef())
    val res = PredicateDeclaration(name, params, body.expr, getSourceLocation(ctx))
    res.modifiers = modifiers
    res.tType.precedingWhitespace = getWhitespaceLeft(ctx.VAL_RESOURCE())
    res.openParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.LPAREN())
    res.closeParenthesis.precedingWhitespace = getWhitespaceLeft(ctx.RPAREN())
    res.equalSign.copyWhitespace(body.eq)
    res.semicolon.copyWhitespace(body.semi)
    res
  }

  override def visitValInstanceFunction(ctx: JavaParser.ValInstanceFunctionContext): VerCorsTreeNode = {
    super.visitValInstanceFunction(ctx)
  }

  override def visitValInstanceGhostDecl(ctx: JavaParser.ValInstanceGhostDeclContext): SpecDeclaration = {
    val decl = visitClassBodyDeclaration(ctx.langClassDecl().classBodyDeclaration())
    val ghost = decl match {
      case meth: MethodDeclaration =>
        SpecMethodModifier(getSourceLocation(ctx.VAL_GHOST()), getWhitespaceLeft(ctx.VAL_GHOST()))
      case _ =>
        GhostModifier(getSourceLocation(ctx.VAL_GHOST()), getWhitespaceLeft(ctx.VAL_GHOST()))
    }
    GhostDeclaration(ghost, decl, getSourceLocation(ctx))
    // todo: typing: ClassBodyDecl vs SpecDecl
  }

  def visitValModifier(ctx: JavaParser.ValModifierContext): Modifier = {
    visit(ctx).asInstanceOf[Modifier]
  }

  case class Def(expr: Option[Expression]) extends VerCorsTreeNode {
    val semi = StringNode("", null)
    val eq = StringNode("", null)
    var precedingWhitespace = ""
    var srcLocation = null
    override def stringRepresentation = ""
  }

  def visitValDef(ctx: JavaParser.ValDefContext): Def = {
    visit(ctx).asInstanceOf[Def]
  }
  override def visitValAbstractBody(ctx: JavaParser.ValAbstractBodyContext): Def = {
    val res = new Def(None)
    res.semi.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
    res
  }

  override def visitValBody(ctx: JavaParser.ValBodyContext): Def = {
    val res = new Def(Some(visitExpression(ctx.langExpr().expression())))
    res.eq.precedingWhitespace = getWhitespaceLeft(ctx.ASSIGN())
    res.semi.precedingWhitespace = getWhitespaceLeft(ctx.SEMI())
    res
  }

  override def visitValModifier0(ctx: JavaParser.ValModifier0Context): Modifier = {
    if(ctx.VAL_PURE() != null) {
      PureModifier(getSourceLocation(ctx), getWhitespaceLeft(ctx))
    } else if(ctx.VAL_INLINE() != null) {
      new InlineModifier(getSourceLocation(ctx), getWhitespaceLeft(ctx))
    } else if(ctx.VAL_THREAD_LOCAL() != null) {
      new InlineModifier(getSourceLocation(ctx), getWhitespaceLeft(ctx))
    } else {
      println(s"Error: unexpected alternative in valModifier: ${ctx.getText}")
      null
    }
  }
  override def visitValStatic(ctx: JavaParser.ValStaticContext): Modifier = {
    val static = ctx.langStatic().STATIC()
    StaticModifier(getSourceLocation(static), getWhitespaceLeft(static))
  }

  def visitValArgListAsQuantifiedVarDeclarations(ctx: JavaParser.ValArgListContext): List[QuantifiedVarDeclaration] = {
    var list: List[QuantifiedVarDeclaration] = List()
    if(ctx.valArgList() != null) {
      list = visitValArgListAsQuantifiedVarDeclarations(ctx.valArgList())
    }
    val typ = visitType(ctx.valArg().langType().`type`())
    val name = visitJavaIdentifier(ctx.valArg().langId().javaIdentifier())
    val decl = new QuantifiedVarDeclaration(typ, List(name), getSourceLocation(ctx.valArg()))
    if(ctx.valArgList() != null) {
      decl.trailingWhitespace = getWhitespaceLeft(ctx.COMMA())
    }
    list.prepended(decl)
  }

  def visitValArgListAsParameters(ctx: JavaParser.ValArgListContext): List[Parameter] = {
    var list: List[Parameter] = List()
    if(ctx.valArgList() != null) {
      list = visitValArgListAsParameters(ctx.valArgList())
    }
    val typ = visitType(ctx.valArg().langType().`type`())
    val name = visitJavaIdentifier(ctx.valArg().langId().javaIdentifier())
    val decl = new Parameter(typ, name, getSourceLocation(ctx.valArg()))
    if(ctx.valArgList() != null) {
      decl.trailingWhitespace = getWhitespaceLeft(ctx.COMMA())
    }
    list.prepended(decl)
  }

  override def visitValEmbedContract(ctx: JavaParser.ValEmbedContractContext): Contract = {
    val specCases = ctx.valEmbedContractBlock().asScala.flatMap(visitValEmbedContractBlockAsList)
    Contract(specCases.toList, getSourceLocation(ctx))
  }

  def visitValEmbedContractForLoop(ctx: JavaParser.ValEmbedContractContext): List[LoopSpecification] = {
    ctx.valEmbedContractBlock().asScala.map(visitValEmbedContractBlockAsLoopSpec).toList
  }

  def visitValEmbedContractBlockAsList(ctx: JavaParser.ValEmbedContractBlockContext): List[ContractClause] = {
    val clauses = ctx.valContractClause().asScala.map(visitValContractClause).toList
    val start = Option(ctx.startSpec()).map(x=>StringNode(x.getText, getSourceLocation(x), getWhitespaceLeft(x)))
    val end = Option(ctx.endSpec()).map(x=>StringNode(x.getText, getSourceLocation(x), getWhitespaceLeft(x)))
    if(clauses.nonEmpty) {
      start match {
        case Some(s) =>
          s.trailingWhitespace = clauses.head.precedingWhitespace
          clauses.head.precedingWhitespace = ""
          clauses.head.startSpec = start
        case None =>
      }
      clauses.last.endSpec = end
      clauses
    } else {
      List(NonTranslatedContractClause(start.toString+end.toString, getSourceLocation(ctx)))
    }
  }

  def visitValEmbedContractBlockAsLoopSpec(ctx: JavaParser.ValEmbedContractBlockContext): LoopSpecification = {
    val clauses = ctx.valContractClause().asScala.map(visitValContractClause).toList
    val start = Option(ctx.startSpec()).map(x=>StringNode(x.getText, getSourceLocation(x), getWhitespaceLeft(x)))
    val end = Option(ctx.endSpec()).map(x=>StringNode(x.getText, getSourceLocation(x), getWhitespaceLeft(x)))
    LoopSpecification(start, clauses, end, getSourceLocation(ctx))
  }

  override def visitValEmbedStatementBlock(ctx: JavaParser.ValEmbedStatementBlockContext): JmlStatementBlock = {
    val statements = ctx.valStatement().asScala.map(visitValStatement).toList
    var startSpec: Option[StringNode] = None
    var endSpec: Option[StringNode] = None
    if (ctx.startSpec() != null) {
      startSpec = Some(StringNode(ctx.startSpec().getText,
        getSourceLocation(ctx.startSpec()), getWhitespaceLeft(ctx.startSpec())))
      endSpec = Some(StringNode(ctx.endSpec().getText,
        getSourceLocation(ctx.endSpec()), getWhitespaceLeft(ctx.endSpec())))
    }
    JmlStatementBlock(startSpec, statements, endSpec, getSourceLocation(ctx))
  }

  override def visitValEmbedClassDeclarationBlock(ctx: JavaParser.ValEmbedClassDeclarationBlockContext):
      SpecDeclarationBlock = {
    val decls = ctx.valClassDeclaration().asScala.map(visitValClassDeclaration)
    var startSpec: Option[StringNode] = None
    var endSpec: Option[StringNode] = None
    if (ctx.startSpec() != null) {
      startSpec = Some(StringNode(ctx.startSpec().getText,
        getSourceLocation(ctx.startSpec()), getWhitespaceLeft(ctx.startSpec())))
      endSpec = Some(StringNode(ctx.endSpec().getText,
        getSourceLocation(ctx.endSpec()), getWhitespaceLeft(ctx.endSpec())))
    }
    SpecDeclarationBlock(startSpec, decls.toList, endSpec, getSourceLocation(ctx))
  }

  override def visitValEmbedGlobalDeclarationBlock(ctx: JavaParser.ValEmbedGlobalDeclarationBlockContext):
      SpecDeclarationBlock = {
    ???
  }
//  SpecDeclarationBlock = {
//    val decls = ctx.valDeclaration().asScala.map(visitValDeclaration).toList
//    val startSpec = StringNode(ctx.startSpec().getText,
//      getSourceLocation(ctx.startSpec()), getWhitespaceLeft(ctx.startSpec()))
//    val endSpec = StringNode(ctx.endSpec().getText,
//      getSourceLocation(ctx.endSpec()), getWhitespaceLeft(ctx.endSpec()))
//    SpecDeclarationBlock(startSpec, decls, endSpec, getSourceLocation(ctx))
//  }

  override def visitValEmbedModifier(ctx: JavaParser.ValEmbedModifierContext): Modifier = {
    val modifier = visitValModifier(ctx.valModifier())
    val startSpec = Option(ctx.startSpec()).map(x=>StringNode(x.getText, getSourceLocation(x), getWhitespaceLeft(x)))
    val endSpec = Option(ctx.endSpec()).map(x=>StringNode(x.getText, getSourceLocation(x), getWhitespaceLeft(x)))
    JmlModifierBlock(startSpec, List(modifier), endSpec, getSourceLocation(ctx))
  }

}
