"""Script to translate specifications from one deductive verifier (Krakatoa, OpenJML or VerCors) to another for a set of examples, and running the target verifier afterwards on the successfully translated files, reporting the verification results.
Usage: execute "python3 evaluation.py" in commandline
When using OpenJML or VerCors for verification, make sure that either the command uses the correct java version OR that "java" refers to the correct java version.
"""

import os
import shlex
import subprocess
import sys
import statistics
import timeit

from run_verifier import OpenJML, VerCors, Krakatoa, Tool

################### 
#  configuration  #
################### 

# The following settings are used to call the translation tool and the verifiers. 
# They may need to be changed depending on your system.

# Translator settings
translator_command = "java -jar ../AnnotationTranslator/target/scala-3.1.0/AnnotationTranslator.jar -s {src} -t {target} {filename}"
output_dir = "./translated"

# how often should the translator be called on each file for statistical timing
NUM_REPS_TIMING = 5

### Verifier settings

# VerCors settings (requires Java 11+!):

# vercors_command = "sh ~/Documents/vercors/bin/vct {}"
vercors_command = "C:/Users/ArmborstL/Documents/vercors/vercors/bin/vct.cmd {}"
vercors_examples_dir = "VerCors-programs"

# Krakatoa settings:
# This script assumes that "CVC3", "CVC4" and "Alt-Ergo" are available as backend provers
krakatoa_command = "krakatoa -why3 'prove --prover {}' {}"
krakatoa_examples_dir = "Krakatoa-programs"

# OpenJML settings (requires Java 8!):
openjml_command = "java8 -jar C:/Users/ArmborstL/Documents/openjml/0-8-59/openjml.jar -esc -progress {}"
# openjml_command = "java -jar ~/Downloads/openjml/openjml.jar -esc -progress {}"
openjml_examples_dir = "OpenJML-programs"
openjml_examples_openjml_folder = "OpenJML"
openjml_examples_jmlspecs_folder = "jmlspecs"



############################## 
#  actual evaulation script  #
############################## 

# change working directory to the directory of this script
cur_dir = os.path.dirname(__file__)
os.chdir(cur_dir)

# dict storing the results in a compact way
results = {}


def translate(src: str, target: str, dir: str, shorten: bool=False):
  """ Translate all files in 'dir', run target verifier if translation is successful
  @param src: name of source language (from "OpenJML", "VerCors", "Krakatoa"; case-insensitive)
  @param target: name of target language (from "OpenJML", "VerCors", "Krakatoa"; case-insensitive)
  @param shorten: whether to print full output of translation tool or just first lines
  """
  print("----------\nTranslating from {} to {}:\n".format(src, target), flush=True)

  # set output_dir and verification tool according to 'target'
  tool: Tool = None
  output_path = ""
  if(target.lower() == "openjml"):
    output_path = os.path.join(output_dir, "OpenJMLProgram")
    tool = OpenJML(openjml_command, output_path)
  elif(target.lower() == "vercors"):
    output_path = os.path.join(output_dir, "VerCorsProgram")
    tool = VerCors(vercors_command, output_path)
  elif(target.lower() == "krakatoa"):
    output_path = os.path.join(output_dir, "KrakatoaProgram")
    tool = Krakatoa(krakatoa_command, output_path)

  for f in os.listdir(dir):
    if(not f.endswith(".java")):
      continue
    
    # translate f
    full_command = translator_command.format(src=src, target=target, 
                                            filename=os.path.join(dir, f))
    if os.name == 'posix':
      subproc_args = shlex.split(full_command)
    elif os.name == 'nt':
      subproc_args = full_command
    else:
      print("unclear operating system, no pre-processing of commands possible")
      subproc_args = full_command
    
    stdout = subprocess.PIPE if(shorten) else None
    stderr = subprocess.STDOUT if(shorten) else None
    res = subprocess.run(subproc_args, 
                         stdout=stdout, stderr=stderr, text=True)
    
    if(shorten):
      output = res.stdout.split('\n')
      for line in output[:3]:
        print(line.rstrip('\r'), flush=True)
    
    
    # timing analysis
    
    # wrapping translator call for timing library, repressing output
    def execute_timed():
      subprocess.call(subproc_args, 
                      stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
      
    # run wrapped call NUM_REPS_TIMING times, get vector of exec times
    times = timeit.repeat(execute_timed, repeat=NUM_REPS_TIMING, number=1)
    # format output
    mean = statistics.mean(times)
    dev = statistics.stdev(times)
    secs = "{:.2f} +/- {:.2f}({:2.0f}%)".format(mean, dev, 100*dev/mean)
    
    # # alternative way of measuring time, via profiler
    # # seems to give similar results, so let's stick with the timeit one
    # import cProfile
    # import pstats
    # prof = cProfile.Profile()
    # for i in range(NUM_REPS_TIMING):
      # prof.runctx("execute_timed()", locals=locals(), globals=globals())
    # print(pstats.Stats(prof).get_stats_profile()
                            # .func_profiles["call"]
                            # .percall_cumtime)
    
    
    # verify translation result if translation successful
    if(res.returncode == 0 and tool != None):
      print("----\nRunning {} on \"{}\":".format(target, f), flush=True)
      output = tool.run(os.path.join(output_path, f))
      for line in output:
        print(line, flush=True)
      
      # store result in table
      non_empty = [l for l in output if l.rstrip()]
      results[os.path.join(dir,f)] = ('+', secs, non_empty[-1] if non_empty else "")
    else:
      results[os.path.join(dir,f)] = ('-', secs, "")
    
    print("\n", flush=True)


""" Main method of script: translate VerCors -> OpenJML, and OpenJML -> VerCors """
if __name__ == "__main__":
  shorten = False
  if len(sys.argv) > 1:
    if(sys.argv[1].lower() == "--short"):
      shorten = True

  translate("Krakatoa", "OpenJML", krakatoa_examples_dir)
  translate("Krakatoa", "VerCors", krakatoa_examples_dir)
  translate("VerCors", "OpenJML", vercors_examples_dir, shorten)
  translate("OpenJML", "VerCors",
            os.path.join(openjml_examples_dir, openjml_examples_openjml_folder),
            shorten)
  translate("OpenJML", "VerCors",
            os.path.join(openjml_examples_dir, openjml_examples_jmlspecs_folder),
            shorten)

  # print result table
  print("\n File"+35*" "+"| Translation |     Time in s      | Tool execution")
  print(116*"-")
  for f, r in results.items():
    name = " " + f
    if len(name) > 39:
      name = "..." + f[-36:]
    print(name.ljust(40) + "|" + 6*" " + r[0] + 6*" " + "| "+str(r[1])+ " | " + r[2][:60])

