"""Script to run a deductive verifier (Krakatoa, OpenJML or VerCors) for a (set of) examples and prints output reporting the verification results.
Usage: execute "python3 run_verifier.py [verifier]" in commandline where [verifier] is replaced by either Krakatoa, OpenJML or VerCors (case insensitive)
When using OpenJML or VerCors, make sure that either the command uses the correct java version OR that "java" refers to the correct java version.
"""

import os
import sys
import yaml

from typing import List, Dict

# The following settings are used to call the verifiers. They may need to be changed depending on your system.

# Krakatoa settings:
# This script assumes that "CVC3", "CVC4" and "Alt-Ergo" are available as backend provers
krakatoa_command = "krakatoa -why3 'prove --prover {}' {}"
krakatoa_examples_dir = "Krakatoa-programs/"

# OpenJML settings:
openjml_command = "java -jar ~/Downloads/openjml/openjml.jar -esc -progress {}" # requires java 8!
openjml_examples_dir_openjml = "OpenJML-programs/OpenJML/"
openjml_examples_dir_jmlspecs = "OpenJML-programs/jmlspecs/"

# VerCors settings:
vercors_command = "C:/Users/ArmborstL/Documents/VerCors/vercors/bin/vct.cmd {}" # requires java 11+!
vercors_examples_dir = "{}/VerCors-programs/".format(os.path.dirname(__file__))



class Tool:
  """ abstract super class for the various tools """

  def __init__(self, command: str="", examples_dir: str=""):
    """ constructor
    @param command: how to invoke the tool 
    @param examples_dir: path to directory of the files to be verified
    """
    self.command = command
    self.examples_dir = examples_dir
    
    
  def run(self, filepath: str) -> List[str]:
    """ Run the tool on the given file
    @param filepath: full path to file
    @return: a list of output lines indicating whether the file could be verified or not.
    """
    print("Called 'run' on uninstantiated tool")
    return []

  def parse_results(self, tool_output: str) -> List[str]:
    """ Parse results from a single tool run. 
    @param tool_output: raw output after executing the tool for a specific file
    @return: a list of filtered output lines indicating what could (not) be proven
    """
    print("Called 'parse_results' on uninstantiated tool")
    return []

  def run_all(self) -> Dict[str, List[str]]:
    """ Run the Tool on all .java programs located in the Tool's examples_dir
    @return: a dictionary that contains the verification results per file
    """
    print("Called 'run_all' on uninstantiated tool")
    return {}

  def print_results(self, dict: Dict[str, List[str]]):
    """ Print results in the terminal
    @param dict: a dictionary which maps filenames to a list of results, e.g. '{"Creation.java" : ["Valid...", "Invalid..."]}'
    """
    print(yaml.dump(dict, sort_keys=False, default_flow_style=False))




class Krakatoa(Tool):
  def __init__(self, 
              command: str=krakatoa_command, 
              examples_dir: str=krakatoa_examples_dir):
    """ constructor
    @param command: how to invoke the tool (default: this script's 'krakatoa_command')
    @param examples_dir: path to directory of the files to be verified (default: this script's 'krakatoa_examples_dir')
    """
    self.command = command
    self.examples_dir = examples_dir

  def run(self, filepath: str, prover: str="Alt-Ergo") -> List[str]:
    """ Run Krakatoa on a single file with the required prover (prover selected based on file name)
    @param filepath: full path to file
    @param prover: name of the prover to be used (from "CVC3", "CVC4", "Alt-Ergo")
    @return: a list of which things could (not) be proven
    """
    full_command = self.command.format(prover, filepath)
    output_stream = os.popen(full_command)
    res = self.parse_results(output_stream.read())
    return res

  def parse_results(self, output: str) -> List[str]:
    """ Parse results from a Krakatoa run. 
    @param tool_output: raw output after executing the tool for a specific file
    @return: a list of filtered output lines indicating what could (not) be proven
    """
    output = output.split("\n")
    results = []
    for line in output:
      if "Jessie_program" in line:
        results.append((line.split("WP_parameter")[-1]).strip())
    return results

  def run_all(self) -> Dict[str, List[str]]:
    """ Run Krakatoa on all .java programs located in the Tool's examples_dir
    @return: a dictionary that contains the verification results per file
    """
    res = {}
    for filename in os.listdir(self.examples_dir):
      if filename.endswith(".java"):
        prover = ("CVC3" if filename == "Creation.java" 
                  else ("CVC4" if filename == "MyCosine.java" or filename == "TreeMax.java" 
                  else "Alt-Ergo"))
        filepath = os.path.join(self.examples_dir, filename)
        res[filename] = self.run(filepath)
    return res



class OpenJML(Tool):
  def __init__(self, 
              command: str=openjml_command, 
              examples_dir: str=openjml_examples_dir_openjml,
              examples_dir_spec: str=openjml_examples_dir_jmlspecs):
    """ constructor
    @param command: how to invoke the tool (default: this script's 'openjml_command')
    @param examples_dir: path to directory of the files to be verified (default: this script's 'openjml_examples_dir_openjml')
    @param examples_dir_spec: path to additional directory of files to be verified (default: this script's 'openjml_examples_dir_jmlspecs')
    """
    self.command = command
    self.examples_dir = examples_dir
    self.examples_dir_spec = examples_dir_spec

  def run_with_dir(self, filename: str, loc: str) -> List[str]:
    """ Run OpenJML on the given file
    @param filename: name of file (not path!)
    @param loc: directory where filename resides
    @return: a list of output lines indicating per method whether there were any warnings
    """
    return self.run(os.path.join(loc, filename))
      
  """
  # It returns a list indicating per method whether there were any warnings
  """
  def run(self, filepath: str) -> List[str]:
    """ Run OpenJML on the given file
    @param filepath: full path to file 
    @return: a list of output lines indicating per method whether there were any warnings
    """
    full_command = self.command.format(filepath)
    output_stream = os.popen(full_command)
    res = self.parse_results(output_stream.read())
    return res

  def parse_results(self, output: str) -> List[str]:
    """ Parse results from an OpenJML run. 
    @param tool_output: raw output after executing the tool for a specific file
    @return: a list of filtered output lines that say something about proofs that are completed
    """
    output = output.rstrip().split("\n")
    results = []
    for line in output:
      if "Completed proof of" in line:
        results.append(line)
      if "Operation not performed because of parse or type errors" in line:
        results.append(line)
    if output: 
      if "warning" in output[-1] or "error" in output[-1]:
        results.append(output[-1])
      elif "Operation not performed" not in output[-1]:
        results.append("apparently no warnings")
    else:
      results.append("no output")
    return results

  def run_all(self) -> Dict[str, List[str]]:
    """ Run OpenJML on all .java programs located in the Tool's examples_dir and examples_dir_spec
    @return: a dictionary that contains the verification results per file
    """
    res = {}
    for x in os.listdir(self.examples_dir):
      if x.endswith(".java"):
        res[x] = self.run_with_dir(x, self.examples_dir)
    for x in os.listdir(self.examples_dir_spec):
      if x.endswith(".java"):
        res[x] = self.run_with_dir(x, self.examples_dir_spec)
    return res



class VerCors(Tool):
  def __init__(self, 
              command: str=vercors_command, 
              examples_dir: str=vercors_examples_dir):
    """ constructor
    @param command: how to invoke the tool (default: this script's 'vercors_command')
    @param examples_dir: path to directory of the files to be verified (default: this script's 'vercors_examples_dir')
    """
    self.command = command
    self.examples_dir = examples_dir

  def run(self, filepath: str) -> List[str]:
    """ Run VerCors on the given file
    @param filepath: full path to file 
    @return: a list of output lines indicating whether the file could be verified or not.
    """
    full_command = self.command.format(filepath)
    output_stream = os.popen(full_command)
    res = self.parse_results(output_stream.read())
    return res

  def parse_results(self, output: str) -> List[str]:
    """ Parse results from a VerCors run. 
    @param tool_output: raw output after executing the tool for a specific file
    @return: a list of filtered output lines indicating the "Final verdict"
    """
    if ("The final verdict is Pass" in output) or ("Verification completed successfully" in output):
      return ["Success"]
    else:
      output = output.split("\n")
      if len(output) > 2:
        return [output[-3]]
      else:
        return ["Fail"]

  def run_all(self) -> Dict[str, List[str]]:
    """ Run VerCors on all .java programs located in the Tool's examples_dir
    @return: a dictionary that contains the verification results per file
    """
    res = {}
    for x in os.listdir(self.examples_dir):
      if x.endswith(".java"):
        res[x] = self.run(os.path.join(self.examples_dir, x))
    return res



"""
Main method, parses arguments and then calls the appropriate verifier and prints the results
"""
if __name__ == "__main__":
  if len(sys.argv) != 2:
    print("Usage: 'python3 run_verifier.py [verifier]' where [verifier] needs to be replaced by krakatoa, openjml or vercors.")
    sys.exit(1)
  
  verifier = sys.argv[1].lower()
  tool = None
  if verifier == "krakatoa":
    tool = Krakatoa()
  elif verifier == "openjml":
    tool = OpenJML()
  elif verifier == "vercors":
    tool = VerCors()
  else:
    raise ValueError("Unknown verifier. Please choose krakatoa, openjml or vercors.")
  
  res = tool.run_all()
  tool.print_results(res)
